
package com.ai_center.dragonbreath;

/** 
 * An interface that must be implemented by all constraint types.
 * This ensures that a constraint can appropriately be handled by a 
 * {@link GlobalSearchControl GlobalSearchControl}.
 *
 * @author Alexander Nareyek
 * @version $Revision: 2.0 $
 */


import com.ai_center.dragonbreath.costs.CostFunctionReference;


public interface Constraint {

    // *******************
    // ***** METHODS *****
    // *******************

    /** 
     * Returns a constraint's type ID. This must be a unique number
     * for each constraint type. However, different kinds of constraint
     * types (like conventional and structural constraint types) may use 
     * the same numbers.
     */
    int getTypeID();

    /** 
     * Queries the number of different costs available for the constraint.
     */
    int getNumberOfCosts();

    /** 
     * Queries specific costs of the constraint.
     * @param costID  The ID of the costs to be retrieved.
     */
    long getCosts(int costID);

    /** 
     * Queries the name of a specific cost type of the constraint.
     * @param costID  The ID of the costs for which the name is queried.
     */
    String getCostName(int costID);

    /** 
     * Sets a cost function reference for a constraint.
     * @param costID  The ID of the costs.
     * @param costFunctionReference  The reference to be set for the 
     *                               constraint.
     */
    void setCostFunctionReference(int costID,
				  CostFunctionReference costFunctionReference);

    /** 
     * Queries a cost function reference of a constraint.
     * @param costID  The ID of the related costs.
     */
    CostFunctionReference getCostFunctionReference(int costID);

    /** 
     * Initializes constraint-specific internal structures of a constraint 
     * instance. This function is called when the instance has been created.
     */
    void initializeInternalStructures();

    /** 
     * Initializes a specific cost type of the constraint. The function
     * ensures that the specified costs have a correct value and that the 
     * cost propagation has been done.
     * @param costID  The ID of the costs to be initialized.
     */
    void initializeCostType(int costID);

    /** 
     * Performs changes of the CSP in order to improve the constraint's cost.
     * @param costID  The ID of the costs to be improved.
     */
    void improveCosts(int costID);

}


/* 
 * DO _NOT_ EDIT ANYTHING BELOW THIS POINT! 
 * 
 * $Log: Constraint.java,v $
 * Revision 2.0  2002/09/17 00:33:06  nareyek
 * work on beta version began: resource type, state tasks and precondition tasks introduced
 *
 * Revision 1.17  2002/08/30 11:44:16  nareyek
 * current time object constraint introduced; basic foundations of task constraint added - not functional yet
 *
 * Revision 1.16  2002/06/11 22:56:23  nareyek
 * constraint initialization function added; inconsistent interval list for DiscreteResourceConstraint introduced
 *
 * Revision 1.15  2002/06/01 01:12:26  nareyek
 * initial cost type selection for graphics added
 *
 * Revision 1.14  2002/05/16 20:30:36  nareyek
 * more than one cost type per constraint can now be used
 *
 * Revision 1.13  2002/04/09 22:10:39  nareyek
 * introduced getTypeID() for Constraint interface
 *
 * Revision 1.12  2002/04/09 01:19:37  nareyek
 * Every cost item can now have a mapping; constraints can have multiple costs
 *
 * Revision 1.11  2002/04/04 09:25:36  nareyek
 * general constraint interface included; cost workflow improved
 * 
 * 
 */ 
