
package com.ai_center.dragonbreath;

/** 
 * Interface to a constraint solver / search engine. The solver
 * <ul>
 * <li> manages creation and changes to the search problem
 * <li> searches for consistent/optimal solutions
 * </ul>
 *
 * @author Alexander Nareyek
 * @version 0.3
 */


import java.math.BigInteger;

import com.ai_center.dragonbreath.costs.*;
import com.ai_center.dragonbreath.constraintgraph.*;


public interface ConstraintSolverInterface {

    // *******************
    // ***** METHODS *****
    // *******************

    /** 
     * Creates a new variable for the CSP.
     * @param name  The name for the variable.
     */
    Variable createVariable(String name);

    /** 
     * Deletes an existing variable of the CSP.
     * @param variable  The variable to be deleted.
     */
    void deleteVariable(Variable variable);

    /** 
     * Extends an existing embedding.
     * @param embedding  The related embedding.
     * @param node  The node to be added.
     * @param edgeID  ID of the edge to be added.
     * @param direction  Direction of the edge.
     * @see Registry
     */
    void extendEmbedding(Embedding embedding, Node node, int edgeID, 
			 boolean direction);

    /** 
     * Creates a new constraint instance for the CSP.
     * @param constraintID  The type of the constraint.
     * @param embedding  The base graph initialization for the constraint.
     * @see Registry
     */
    ConventionalConstraint createConstraint(int constraintID, 
					    Embedding embedding);

    /** 
     * Extends an existing constraint of the CSP.
     * @param constraint  The constraint to be extended.
     * @param embedding  The extension graph initialization for the constraint.
     * @see Registry
     */
    void extendConstraint(ConventionalConstraint constraint, 
			  Embedding embedding);

    /** 
     * Deletes an existing constraint of the CSP.
     * @param constraint  The constraint to be deleted.
     */
    void deleteConstraint(ConventionalConstraint constraint);

    /** 
     * Creates a new cost function. It is included at the highest 
     * importance level in the list of cost functions (but importance
     * being less than that of the function to guarantee structural 
     * satisfaction).
     * @param name  The name of the cost function.
     */
    CostFunction createCostFunction(String name);

    /** 
     * Executes one improvement/repair iteration. Returns true if an
     * improvement was executed and false if all constraints are already
     * satisfied.
     */
    boolean executeImprovementIteration();

}
