package com.ai_center.dragonbreath;


import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.EventListenerList;




/** Implements the RealTimeControl interface for the GlobalSearchControl. <p>
 

 @author Daniel Weiss
 @version $$
 */
public class DefaultRealTimeControl implements RealTimeControl
{


    /** An array of all breakpoint names. */
    private String[] breakpoints = Registry.breakpoints;

    /** An array with all states of the breakpoints (active/not active). */
    private boolean[] breakpointStates = new boolean[Registry.MAX_BREAKPOINT_NUMBER];

    /** A link to the GlobalSearchControl- */
    private GlobalSearchControl gsc;

    /** A list of EventListener which are informed about changes. */
    private EventListenerList listenerList = new EventListenerList();

    /** The mode of the present RealTimeControl- */
    private int mode = RealTimeControl.FASTFORWARD;

    /** Semaphore to jump to the next breakpoint. */
    private boolean nextStep = false;

    /** The pause between to breakpoints, when in PLAY mode. */
    private int pause = 1000; // wait 1 second for default

    /** Creates a DefaultRealTimeControl.*/
    public DefaultRealTimeControl()
    {
        breakpointStates[0] = true;
        // make first breakpoint always stop as default
        for (int i = 1; i < Registry.MAX_BREAKPOINT_NUMBER; i++)
        {
            breakpointStates[i] = false;
        }

    }

    /** Informs all changelisteners about a change..<p>
     
     @param changedObject The object which has changed.
     */
    protected synchronized void fireChange(Object changedObject)
    {

        ChangeEvent changeEvent = new ChangeEvent(changedObject);

        Object[] listeners = listenerList.getListenerList();
        ChangeListener listener = null;

        for (int i = listeners.length - 2; i >= 0; i -= 2)
        {
            if (listeners[i] == ChangeListener.class)
            {

                ((ChangeListener) listeners[i + 1]).stateChanged(changeEvent);

            }

        }
    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.dragonbreath.RealTimeControl#addChangeListener(javax.swing.event.ChangeListener)
     */
    public synchronized void addChangeListener(ChangeListener l)
    {
        listenerList.add(ChangeListener.class, (ChangeListener) l);

    }

    /* (non-Javadoc)
     * @see com.ai_center.dragonbreath.RealTimeControl#beakpoint(int, java.lang.String, java.lang.Object)
     */
    public synchronized void beakpoint(int breakpointNumber,
            Object changedObject)
    {
        if (breakpointStates[breakpointNumber] == true)
        {

            RTCMessage rtcMessage = new RTCMessage(breakpointNumber,
                    breakpoints[breakpointNumber], changedObject);
            fireChange(rtcMessage);

            // waiting time
            switch (mode)
            {
                case RealTimeControl.STEPPING :
                    // wait for next() call
                    nextStep = false;
                    while (nextStep == false)
                    {
                        try
                        {
                            wait();
                        } catch (InterruptedException e)
                        {
                        }
                    }
                    break;
                case RealTimeControl.PLAY :
                    // wait (pause)-milliseconds
                    try
                    {
                        wait(pause);
                    } catch (InterruptedException e)
                    {
                    }

                    break;

                case RealTimeControl.FASTFORWARD :
                default :
                    // don't wait
                    break;
            }
        }
    }

    /* (non-Javadoc)
     * @see com.ai_center.dragonbreath.RealTimeControl#getBreakpoints()
     */
    public String[] getBreakpoints()
    {

        return Registry.breakpoints;
    }

    /* (non-Javadoc)
     * @see com.ai_center.dragonbreath.RealTimeControl#getBreakpointState(int)
     */
    public boolean getBreakpointState(int breakpointNumber)
    {
        if (breakpointNumber >= Registry.MAX_BREAKPOINT_NUMBER)
        {
            throw new Error(
                    "DefaultRealTimeControl.java: Access on a undefined "
                            + "breakpoint state. breakpointNumber ist bigger than"
                            + " Registry.MAX_BREAKPOINT_NUMBER");
        }
        return breakpointStates[breakpointNumber];
    }

    /* (non-Javadoc)
     * @see com.ai_center.dragonbreath.RealTimeControl#getMode()
     */
    public int getMode()
    {

        return mode;
    }

    /* (non-Javadoc)
     * @see com.ai_center.dragonbreath.RealTimeControl#getPlayPauseTime()
     */
    public int getPlayPauseTime()
    {

        return pause;
    }

    /* (non-Javadoc)
     * @see com.ai_center.dragonbreath.RealTimeControl#next()
     */
    public synchronized void next()
    {
        nextStep = true;
        notify();

    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.dragonbreath.RealTimeControl#removeChangeListener(javax.swing.event.ChangeListener)
     */
    public synchronized void removeChangeListener(ChangeListener l)
    {
        listenerList.remove(ChangeListener.class, (ChangeListener) l);
    }

    /* (non-Javadoc)
     * @see com.ai_center.dragonbreath.RealTimeControl#setBeakpointActive(int, boolean)
     */
    public void setBeakpointActive(int breakpointNumber, boolean _active)
    {
        if (breakpointNumber >= Registry.MAX_BREAKPOINT_NUMBER)
        {
            throw new Error(
                    "DefaultRealTimeControl.java: Access on a undefined "
                            + "breakpoint state. breakpointNumber ist bigger than"
                            + " Registry.MAX_BREAKPOINT_NUMBER");
        }
        breakpointStates[breakpointNumber] = _active;

    }

    /* (non-Javadoc)
     * @see com.ai_center.dragonbreath.RealTimeControl#setMode(int)
     */
    public synchronized void setMode(int _mode)
    {
        mode = _mode;
        next();
        fireChange(this);

    }

    /* (non-Javadoc)
     * @see com.ai_center.dragonbreath.RealTimeControl#setPlayPauseTime(int)
     */
    public synchronized void setPlayPauseTime(int _pause)
    {
        pause = _pause;
        fireChange(this);

    }

}