
package com.ai_center.dragonbreath;

/**
 * Implements a local-search-based constraint solver.
 *
 * @author Alexander Nareyek
 * @version $Revision: 2.20 $
 *
 * @see dragonbreath
 */


import java.awt.*;
import java.awt.event.*;
import java.io.File;
import java.math.BigInteger;
import java.util.*;
import javax.swing.*;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.EventListenerList;

import org.jdom.JDOMException;

import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.constraints.IntegerConstraint;
import com.ai_center.dragonbreath.constraints.PreconditionTaskObjectConstraint;
import com.ai_center.dragonbreath.constraints.RangeConstraint;
import com.ai_center.dragonbreath.constraints.RangeConstraintGraphics;
import com.ai_center.dragonbreath.constraints.StateResourceConstraint;
import com.ai_center.dragonbreath.costs.*;
import com.ai_center.dragonbreath.loader.*;
import com.ai_center.dragonbreath.utilities.SetHashtable;


public class GlobalSearchControl implements ConstraintSolverInterface {

    // **********************
    // ***** ATTRIBUTES *****
    // **********************

    /**
     * Random generator.
     */
    public Random random;

    /**
     * Set of all created cost mappings.
     */
    protected java.util.List costMappings
        = Collections.synchronizedList( new ArrayList() );
    /**
     * The default cost mapping. Set in the constructor/initialization
     * method.
     */
    private CostMapping defaultCostMapping = null;;
    /**
     * Current cost mapping that was set.
     */
    protected CostMapping currentCostMapping = null;
    /**
     * Indicates whether the default cost mappings of cost functions are
     * to be used for constraint creation.
     */
    private boolean useCostMappingOfCostFunction = false;

    /**
     * List of the cost functions. Function 0 for maximization,
     * all other for minimization.
     */
    protected java.util.List costFunctions
        = Collections.synchronizedList( new LinkedList() );
    /**
     * Current cost function that was set.
     */
    protected CostFunction currentCostFunction = null;
    /**
     * Maximization cost function.
     */
    public CostFunction MAX_FUNCTION = null;
    /**
     * Default cost function.
     */
    protected CostFunction DEFAULT_FUNCTION = null;
    /**
     * SCSP cost function.
     */
    protected CostFunction SCSP_FUNCTION = null;
    /**
     * Number of defined cost functions.
     */
    protected int costFunctionNumber = 0;
    /**
     * Unsatisfied cost function of the highest priority level.
     */
    public CostFunction lastUnsatisfiedCostFunction = null;

    /**
     * Current cost mapping options that were set.
     */
    protected Object[] currentCostMappingOptions = null;

    /**
     * Set of the CSP's variables.
     */
    protected Set variables = Collections.synchronizedSet( new HashSet() );

    /**
     * List of the CSP's constraints.
     */
    protected Set constraints = Collections.synchronizedSet( new HashSet() );

    /**
     * SetHashtable of the CSP's constraints, from node types to sets
     * of constraint.
     */
    public SetHashtable constraintSHT = new SetHashtable();

    /**
     * The reference to be handed to the connected constraints when
     * a variable update must be propagated.
     */
    private ConnectionReference connectionReferenceUpdate
        = new ConnectionReference( null );
    /**
     * The reference to be handed to the connected constraints when
     * a new node connection must be propagated.
     */
    private ConnectionReference connectionReferenceConnection
        = new ConnectionReference( null );
    /**
     * The reference to be handed to the connected constraints when
     * a node disconnection must be propagated.
     */
    private ConnectionReference connectionReferenceDisconnection
        = new ConnectionReference( null );
    /**
     * A Connections object to be used as temporary copy for connection
     * enumerations.
     */
    private Connections connectionsCopy = new Connections( null );

    /**
     * Object for load and save operations.
     */
    private GraphLoader graphLoader = null;

    /**
     * Object for structural constraint control.
     */
    public com.ai_center.dragonbreath.structuralConstraintHandling.StructuralConstraintManagement scm;
    /**
     * Indicates whether the constraint graph is allowed to be
     * changed by improvement heuristics.
     */
    public boolean structuralChangesEnabled = true;

    /**
     * Frame for visual tracking.
     */
    public GlobalSearchControlGraphics graphObject = null;

    /**
     * The RealTimeControl interface.
     */
    private RealTimeControl realTimeControl = new DefaultRealTimeControl();

    /**
     * A list of EventListeners, which get informed about changes.
     */
    private EventListenerList listenerList = null;


    // ************************
    // ***** CONSTRUCTORS *****
    // ************************


    /**
     * Initializes the solver object.
     * (Shortcut function for using a currentCostMapping of
     * FactorCostMapping.)
     * @param randomSeed  The initial random seed.
     */
    public GlobalSearchControl( long randomSeed )
    {
        initialize( randomSeed , null );
    }

    /**
     * Initializes the solver object.
     * @param randomSeed  The initial random seed.
     * @param costMapping  The default cost mapping to be used. Pass "null"
     *     for creating a default of FactorCostMapping.
     */
    public GlobalSearchControl( long randomSeed , CostMapping costMapping )
    {
        initialize( randomSeed , costMapping );
    }


    // *******************
    // ***** METHODS *****
    // *******************


    /**
     * Initializes the solver object. Pay attention to reflect changes in
     * this function's code in clearCSP().
     * @param randomSeed  The initial random seed.
     * @param costMapping  The default cost mapping to be used. Pass "null"
     *     for a default of FactorCostMapping.
     */
    private void initialize( long randomSeed , CostMapping costMapping )
    {

        if ( costMapping == null )
        {
            defaultCostMapping
                = createCostMapping( Registry.FACTOR_COST_MAPPING );
        }
        else
        {
            defaultCostMapping = costMapping;
        }
        currentCostMapping = defaultCostMapping;
        useCostMappingOfCostFunction = false;

        random = new Random( randomSeed );
        graphLoader = new GraphLoader( this );
        scm = new com.ai_center.dragonbreath.structuralConstraintHandling.StructuralConstraintManagement( this );

        // Creating initial cost functions. DEFAULT_FUNCTION will
        // be initialized if used.

        synchronized( costFunctions )
        {
            MAX_FUNCTION = new CostFunction( "Maximization Function" ,
                this , 0 , currentCostMapping );
            MAX_FUNCTION.setConstraintSelectionMethod( CostFunction.SELECTION_RANDOM );
            costFunctions.add( MAX_FUNCTION );
            SCSP_FUNCTION = new CostFunction("Structural Costs" , this , Integer.MAX_VALUE , currentCostMapping );
            costFunctions.add( SCSP_FUNCTION );
            costFunctionNumber += 2;
        }
        currentCostFunction = null;

    }


    /**
     * Deletes all elements and cost information of the current CSP.
     */
    public void clearCSP()
    {

        if ( graphObject != null )
        {

            Set oldVariables;
            Set oldConstraints;
            synchronized( constraints )
            {
                synchronized( variables )
                {
                    oldVariables = variables;
                    variables = Collections.synchronizedSet( new HashSet() );
                    oldConstraints = constraints;
                    constraints = Collections.synchronizedSet( new HashSet() );
                }
            }

            Iterator it = oldVariables.iterator();
            Variable variable;
            while ( it.hasNext() )
            {
                variable = (Variable)it.next();
                if ( variable.graphObject instanceof NodeGraphics )
                {
                    graphObject.removeVariable( variable );
                }
            }

            it = oldConstraints.iterator();
            ConventionalConstraint constraint;
            while ( it.hasNext() )
            {
                constraint = (ConventionalConstraint)it.next();
                if ( constraint.graphObject instanceof NodeGraphics )
                {
                    graphObject.removeConstraint( constraint );
                }
            }

            graphObject.updateGraphics();
            graphObject.updateGlobalCosts( null );

        }
        else
        {
            variables.clear();
            constraints.clear();
        }

        synchronized( constraintSHT )
        {
            constraintSHT.clear();
        }

        // Clear all structural constraints

        scm.clearSCM();

        // Re-initializing cost mappings

        synchronized( costMappings )
        {
            costMappings.clear();
            costMappings.add( defaultCostMapping );
        }
        defaultCostMapping.reset();
        currentCostMapping = defaultCostMapping;
        useCostMappingOfCostFunction = false;

        // Re-initializing cost functions

        synchronized( costFunctions )
        {
            costFunctions.clear();
        }
        costFunctionNumber = 0;
        lastUnsatisfiedCostFunction = null;
        DEFAULT_FUNCTION = null;
        synchronized( costFunctions )
        {
            MAX_FUNCTION = new CostFunction( "Maximization Function" ,
                this , 0 , currentCostMapping );
            MAX_FUNCTION.setConstraintSelectionMethod( CostFunction.SELECTION_RANDOM );
            costFunctions.add( MAX_FUNCTION );
            SCSP_FUNCTION = new CostFunction( "Structural Costs" , this ,
                Integer.MAX_VALUE , currentCostMapping );
            costFunctions.add( SCSP_FUNCTION );
            costFunctionNumber += 2;
        }
        currentCostFunction = null;

    }


    // ***** CSP GRAPH SERVICE METHODS *****


    /**
     * Returns the total number of Variables and constraints.
     */
    public int graphSize()
    {
        return constraints.size() + variables.size();
    }


    /**
     * Checks if a constraint is connected to a variable with a
     * specific value.
     * @param constraint  The constraint to be checked.
     * @param edgeType  The edge type to the connected variable.
     * @param variableValue  The value that the connected variable has to have.
     */
    public boolean constraintHasVariableConnection(ConventionalConstraint
        constraint , int edgeType , double variableValue )
    {

        Variable variable;
        synchronized( constraint.connections )
        {
            variable = constraint.connections.getSpecificVariable( edgeType );
        }
        if ( variable != null )
        {
            return variable.value() == variableValue;
        }
        else
        {
            return false;
        }

    }


    /**
     * Retrieves a constraint of a specific type. If multiple constraints
     * of this type are existing, one of them is returned.
     * @param constraintID  The type ID of the constraint.
     */
    public ConventionalConstraint findConstraint( int constraintID )
    {

        // Retrieving matching constraints

        Object matchingConstraints;

        synchronized( constraintSHT )
        {
            matchingConstraints
                = constraintSHT.get( Registry.nodeType[ constraintID ] );
        }
        if ( matchingConstraints == null )
        {
            return null;
        }

        // Returning one of the matching constraints

        if ( matchingConstraints instanceof ConventionalConstraint)
        {
            return ( (ConventionalConstraint)matchingConstraints );
        }
        else
        {
            synchronized( matchingConstraints )
            {
                Iterator i = ( (HashSet)matchingConstraints ).iterator();
                return ( (ConventionalConstraint)i.next() );
            }
        }

    }


     /**
     * Searches the graph for constraints of a specific type.
     * @param constraintID  The type ID of the constraint.
     */
    public HashSet findConstraints( int constraintID )
    {
        // Retrieving potential constraints

        Object potentialConstraints;

        synchronized( constraintSHT )
        {
            potentialConstraints
                = constraintSHT.get( Registry.nodeType[ constraintID ] );
        }

        HashSet matchingConstraints = new HashSet();

        // Check if there is only a single potential constraint

        ConventionalConstraint constraint;

        if ( potentialConstraints instanceof ConventionalConstraint )
        {
            constraint = (ConventionalConstraint)potentialConstraints;
            matchingConstraints.add( constraint );

        }
        else
        {
            if ( potentialConstraints instanceof Set )
            {
                synchronized ( potentialConstraints )
                {
                    matchingConstraints.addAll( (Set)potentialConstraints );
                }
            }
        }
        return matchingConstraints;
    }


    /**
     * Searches the graph for a constraint of a specific type connected to
     * a variable with a specific value.
     * @param constraintID  The type ID of the constraint.
     * @param edgeType  The edge type of the connected variable.
     * @param variableValue  The value that the connected variable has to have.
     */
    public ConventionalConstraint findConstraint( int constraintID ,
        int edgeType , double variableValue )
    {

        // Retrieving potential constraints

        Object potentialConstraints;

        synchronized( constraintSHT )
        {
            potentialConstraints
                = constraintSHT.get( Registry.nodeType[ constraintID ] );
        }
        if ( potentialConstraints == null )
        {
            return null;
        }

        // Check if there is only a single potential constraint

        ConventionalConstraint constraint;

        if ( potentialConstraints instanceof ConventionalConstraint )
        {
            constraint = (ConventionalConstraint)potentialConstraints;
            if ( constraintHasVariableConnection( constraint , edgeType ,
                     variableValue ) )
            {
                return constraint;
            }
            else
            {
                return null;
            }
        }

        // Searching the HashSet of potential constraints

        synchronized( potentialConstraints )
        {
            Iterator i = ( (HashSet)potentialConstraints ).iterator();
            while ( i.hasNext() )
            {
                constraint = (ConventionalConstraint)i.next();
                if ( constraintHasVariableConnection( constraint , edgeType ,
                         variableValue ) )
                {
                    return constraint;
                }
            }
        }

        return null;

    }


    /**
     * Searches the graph for constraints of a specific type that are
     * connected to a variable with a specific value.
     * @param constraintID  The type ID of the constraint.
     * @param edgeType  The edge type of the connected variable.
     * @param variableValue  The value that the connected variable has to have.
     */
    public HashSet findConstraints( int constraintID , int edgeType ,
        double variableValue )
    {

        // Retrieving potential constraints

        Object potentialConstraints;

        synchronized( constraintSHT )
        {
            potentialConstraints
                = constraintSHT.get( Registry.nodeType[ constraintID ] );
        }
        if ( potentialConstraints == null )
        {
            return null;
        }

        HashSet matchingConstraints = new HashSet();

        // Check if there is only a single potential constraint

        ConventionalConstraint constraint;

        if ( potentialConstraints instanceof ConventionalConstraint )
        {
            constraint = (ConventionalConstraint)potentialConstraints;
            if ( constraintHasVariableConnection( constraint , edgeType ,
                     variableValue ) )
            {
                matchingConstraints.add( constraint );
                return matchingConstraints;
            }
            else
            {
                return null;
            }
        }

        // Searching the HashSet of potential constraints

        synchronized( potentialConstraints )
        {
            Iterator i = ( (HashSet)potentialConstraints ).iterator();
            while ( i.hasNext() )
            {
                constraint = (ConventionalConstraint)i.next();
                if ( constraintHasVariableConnection( constraint , edgeType ,
                         variableValue ) )
                {
                    matchingConstraints.add( constraint );
                }
            }
            if ( matchingConstraints.size() != 0 )
            {
                return matchingConstraints;
            }
        }

        return null;

    }


    /**
     * Removes a specific constraint after potentially begin disconnected
     * from another constraint. The constraint and all variables of the
     * constraint are deleted only if they are not connected to another
     * constraint. ATTENTION: For internal use only; no consistency checks
     * are being performed.
     * @param constraintToDelete  The constraint to be deleted.
     * @param otherConstraintID  The ID of the constraint to be potentially
     *                           disconnected from.
     * @param embedding  Embedding to be use for potential disconnection
     *                   (from constraint to be disconnected).
     */
    public void disconnectAndRemoveConstraint_internal(ConventionalConstraint
                                                       constraintToDelete,
                                                       int otherConstraintID,
                                                       Embedding embedding) {

        // Collecting disconnection and deletion objects

        int i;
        int connectionNumber;
        int connectionVarNumber;
        boolean deleteConstraint = false;
        ConventionalConstraint toBeDisconnected = null;
        Variable[] variableToBeDeleted = null;

        synchronized (constraintToDelete.connections) {
            connectionNumber = constraintToDelete.connections.number();
            connectionVarNumber =
                constraintToDelete.connections.numberOfVariables();

            if (connectionNumber == connectionVarNumber + 1) {
                toBeDisconnected = constraintToDelete.connections.getSpecificConstraint(otherConstraintID, Registry.PLAIN_EDGE);
                if (toBeDisconnected != null) {
                    connectionNumber--;
                }
            }

            // Finding out which variables are to be deleted

            if (connectionNumber == connectionVarNumber) {

                deleteConstraint = true;
                variableToBeDeleted = new Variable[connectionVarNumber];
                Variable variable;

                constraintToDelete.connections.initializeEnumeration(false, true, false, false, null, Registry.nodeType[Registry.VARIABLE], 0, false);
                for (i = 0; i < connectionVarNumber; i++) {
                    variable = (Variable) constraintToDelete.connections.getNext().node;
                    synchronized(variable.connections) {
                        if (variable.connections.number() == 1) {
                            variableToBeDeleted[i] = variable;
                        } else {
                            variableToBeDeleted[i] = null;
                        }
                    }
                }

            }
        }

        // Potential disconnection from the other constraint

        if (toBeDisconnected != null) {
            reduceConstraint_internal(toBeDisconnected, embedding);
        }

        // Potential deletion

        if (deleteConstraint) {

            // Base graph deletion

            deleteConstraint_internal(constraintToDelete);

            // Deleting variables if no other connections

            for (i = 0; i < connectionVarNumber; i++) {
                if (variableToBeDeleted[i] != null) {
                    deleteVariable_internal(variableToBeDeleted[i]);
                }
            }

        }

    }


    // ***** GRAPHICS *****

    /**
     * Initializes visual tracking.
     */
    public void createGraphics() {

        if (graphObject == null) {

            Variable v;
            ConventionalConstraint c;
            Iterator i;

            synchronized(variables) {
                synchronized(constraints) {

                    // Creating graphics objects for variables

                    i = variables.iterator();
                    while (i.hasNext()) {
                        v = (Variable) i.next();

                        if (v.graphObject instanceof NodeGraphicsCache) {
                            NodeGraphicsCache cache =
                                (NodeGraphicsCache) v.graphObject;
                            v.createGraphics();
                            ((NodeGraphics) v.graphObject).x = cache.x;
                            ((NodeGraphics) v.graphObject).y = cache.y;
                        } else {
                            v.createGraphics();
                        }
                    }

                    // Creating graphics objects for constraints

                    i = constraints.iterator();
                    while (i.hasNext()) {
                        c = (ConventionalConstraint) i.next();

                        if (c.graphObject instanceof NodeGraphicsCache) {
                            NodeGraphicsCache cache =
                                (NodeGraphicsCache) c.graphObject;
                            c.createGraphics();
                            ((NodeGraphics) c.graphObject).x = cache.x;
                            ((NodeGraphics) c.graphObject).y = cache.y;
                        } else {
                            c.createGraphics();
                        }
                    }

                }
            }

            // Creating graphics object for solver

            graphObject = new GlobalSearchControlGraphics(this);

            // Displaying variables

            i = variables.iterator();
            while (i.hasNext()) {
                v = (Variable) i.next();
                ((NodeGraphics) v.graphObject).updateSize();
                graphObject.displayVariable(v);
                ((NodeGraphics) v.graphObject).revalidate();
            }

            // Displaying constraints

            i = constraints.iterator();
            while (i.hasNext()) {
                c = (ConventionalConstraint) i.next();
                ((NodeGraphics) c.graphObject).updateSize();
                graphObject.displayConstraint(c);
                ((NodeGraphics) c.graphObject).revalidate();
            }

            if ((lastUnsatisfiedCostFunction != null) && !(lastUnsatisfiedCostFunction.objectiveCosts.equals(BigInteger.ZERO))) {
                graphObject.updateGlobalCosts(lastUnsatisfiedCostFunction);
            } else {
                graphObject.updateGlobalCosts(null);
            }
            graphObject.updateGraphics();

        }

    }

    /**
     * Terminate visual tracking.
     */
    public void deleteGraphics() {

        if (graphObject != null) {

            synchronized(variables) {
                synchronized(constraints) {

                    graphObject.setVisible(false);

                    // Creating graphical caches for variables

                    Variable v;
                    Iterator i = variables.iterator();
                    while (i.hasNext()) {
                        v = (Variable) i.next();

                        if (v.graphObject instanceof NodeGraphics) {
                            NodeGraphics ng = (NodeGraphics) v.graphObject;
                            v.graphObject =
                                (Object) new NodeGraphicsCache(ng.x, ng.y);
                        }

                    }

                    // Creating graphical caches for constraints

                    ConventionalConstraint c;
                    i = constraints.iterator();
                    while (i.hasNext()) {
                        c = (ConventionalConstraint) i.next();

                        if (c.graphObject instanceof NodeGraphics) {
                            NodeGraphics ng = (NodeGraphics) c.graphObject;
                            c.graphObject =
                                (Object) new NodeGraphicsCache(ng.x, ng.y);
                        }
                    }

                }
            }

            // Deleting graphical solver object

            graphObject = null;

        }

    }

    /**
    @return A link to the RealTimeControl.
    */
    public RealTimeControl getRealTimeControl()
    {
       return realTimeControl;
    }


    /** Adds a ChangeListener to observe this object.<p>

    @param l The ChangeListener to add.
    */
    public void addChangeListener(ChangeListener l)
    {
       if (listenerList == null)
       {
           listenerList = new EventListenerList();
       }
       listenerList.add(ChangeListener.class, (ChangeListener) l);

    }

    /** Removes a ChangeListener.<p>

    @param l The ChangeListener to be removed.
    */
    public void removeChangeListener(ChangeListener l)
    {
       if (listenerList != null)
       {
           listenerList.remove(ChangeListener.class, (ChangeListener) l);
       }
    }

    /** Informs all ChangeListeners about a change.
    @param changedObject The changed object.
    */
    protected void fireChange(Object changedObject)
    {
       if (listenerList != null)
       {
           ChangeEvent changeEvent = new ChangeEvent(changedObject);


           Object[] listeners = listenerList.getListenerList();
           ChangeListener listener = null;

           for (int i = listeners.length - 2; i >= 0; i -= 2)
           {
               if (listeners[i] == ChangeListener.class)
               {

                   ((ChangeListener) listeners[i + 1])
                           .stateChanged(changeEvent);

               }
           }
       }
    }

    // ***** FILE HANDLING *****

    /**
     * Load a structual constraint problem (SCP) from a specific file.
     * @param filename  Name of the SCP file.
     */
    public void loadSCP(java.lang.String filename) {
        loadSCP(new File(filename));
    }

    /**
     * Load a structual constraint problem (SCP) from a specific file.
     * @param file  File descriptor of the SCP file.
     */
    public void loadSCP(File file) {
        try {
            graphLoader.load(file);
        }
        catch (JDOMException ex) {
            if (ex.getCause() != null) {
                ex.getCause().printStackTrace();
            } else {
                ex.printStackTrace();
            }
        }
        catch (Exception exx) {
            exx.printStackTrace();
        }
    }

    /**
     * Save the current structual constraint problem (SCP) to a specific file.
     * @param name  Name of the file in which the SCP is to be saved.
     */
    public void saveSCP(java.lang.String name) {
        saveSCP(new File(name));
    }

    /**
     * Save the current structual constraint problem (SCP) to a specific file.
     * @param file  File descriptor of the file in which the SCP is to
     *              be saved.
     */
    public void saveSCP(File file) {
        try {
            graphLoader.save(file);
        }
        catch (JDOMException ex) {
            if (ex.getCause() != null) {
                ex.getCause().printStackTrace();
            } else {
                ex.printStackTrace();
            }
        }
        catch (Exception exx) {
            exx.printStackTrace();
        }
    }

     /**
     * Loads all the Structural Constraints in the file with the given
     * filename.
     * @param filename  The name of a *.sc file.
     * @return Returns whether the constraint was loaded or not.
     */
    public boolean loadStructuralConstraints(java.lang.String filename) {
        return scm.loadStructralConstraints(filename);
    }

    /**
     * Loads all the Structural Constraints in the file with the given
     * filename.
     * @param filename  The name of a *.sc file.
     * @return Returns whether the file was loaded or not.
     */
    public boolean loadStructuralConstraints(File file) {
        return scm.loadStructralConstraints(file);
    }

    /**
     * Returns a Collection of all Structural Constraints loaded.
     */
    public Collection getStructuralConstraints(){
        return scm.getStructuralConstraints();
    }


    // ***** VARIABLES *****

    /**
     * Creates a new variable for the CSP.
     * The related graphical component will only be constructed
     * if there are already graphics turned on. Use "_internal"
     * version if called without user interaction.
     * @param name  The name for the variable.
     */
    public Variable createVariable(String name) {
        // No checks need to be done
        return createVariable_unsafe(name);
    }

    /**
     * Version of createVariable for internal use.
     * @see createVariable(String)
     */
    public Variable createVariable_internal(String name) {
        if (Registry.SAFE_MODE) {
            return createVariable(name);
        } else {
            return createVariable_unsafe(name);
        }
    }

    /**
     * Variable creation without consistency checks.
     * @see createVariable(String)
     */
    private Variable createVariable_unsafe(String name) {

        Variable v = new Variable(name, this);
        synchronized(variables) {
            variables.add(v);
        }

        // Register the new node for StructuralConstraintProblemControl

        scm.registerNewNode(v);

        // Create variable's graphObject if graphics are turned on

        if (graphObject != null) {
            v.createGraphics();
            ((NodeGraphics) v.graphObject).updateSize();
            graphObject.displayVariable(v);
            ((NodeGraphics) v.graphObject).revalidate();
            graphObject.updateGraphics();
        }

        return v;
    }

    /**
     * Creates a variable with x/y coordinates for visual placement.
     * The related graphical component will be also constructed if the
     * graphics are turned on. Otherwise, a cache to store the graphical
     * information will be created. Use "_internal" version if called
     * without user interaction.
     * @param name  The name for the variable.
     * @param x  The x coordinate.
     * @param y  The y coordinate.
     */
    public Variable createVariable(String name, int x, int y) {
        // No checks need to be done
        return createVariable_unsafe(name, x, y);
    }

    /**
     * Version of createVariable (x/y) for internal use.
     * @see createVariable(String, int, int)
     */
    public Variable createVariable_internal(String name, int x, int y) {
        if (Registry.SAFE_MODE) {
            return createVariable(name, x, y);
        } else {
            return createVariable_unsafe(name, x, y);
        }
    }

    /**
     * Variable creation (x/y) without consistency checks.
     * @see createVariable(String, int, int)
     */
    private Variable createVariable_unsafe(String name, int x, int y) {

        Variable v = new Variable(name, this);
        synchronized(variables) {
            variables.add(v);
        }

        // Register the new node in StructuralConstraintProblemControl

        scm.registerNewNode(v);

        // Create variable's graphObject if graphics are turned on and
        // a graphics cache otherwise

        if (graphObject != null) {
            v.createGraphics();
            ((NodeGraphics) v.graphObject).x = x;
            ((NodeGraphics) v.graphObject).y = y;
            ((NodeGraphics) v.graphObject).updateSize();
            graphObject.displayVariable(v);
            ((NodeGraphics) v.graphObject).revalidate();
            graphObject.updateGraphics();
        } else {
            v.graphObject = (Object) new NodeGraphicsCache(x, y);
        }

        return v;
    }

    /**
     * Sets a variable's value. Use "_internal" version if called
     * without user interaction.
     * @param v  The affected variable.
     * @param i  A long value.
     */
    public void setVariable(Variable variable, long i) {
        synchronized(variables) {
            if (!variables.contains(variable)) {
                throw new Error("The variable \"" + variable +
                                "\" is not part of the CSP");
            }
        }
        setVariable_unsafe(variable, i);
    }

    /**
     * Variable setting (long) for internal use.
     * @see setVariable(Variable, long)
     */
    public void setVariable_internal(Variable variable, long i) {
        if (Registry.SAFE_MODE) {
            setVariable(variable, i);
        } else {
            setVariable_unsafe(variable, i);
        }
    }

    /**
     * Variable setting (long) without consistency checks.
     * @see setVariable(Variable, long)
     */
    private void setVariable_unsafe(Variable variable, long i) {
        double oldValue = variable.value();
        variable.set(i);
        setVariableUpdates(variable, oldValue, false);
    }

    /**
     * Sets the variable's value. Use "_internal" version if called
     * without user interaction.
     * @param variable  The affected variable.
     * @param r  A double value.
     */
    public void setVariable(Variable variable, double r) {
        synchronized(variables) {
            if (!variables.contains(variable)) {
                throw new Error("The variable \"" + variable +
                                "\" is not part of the CSP");
            }
        }
        setVariable_unsafe(variable, r);
    }

    /**
     * Variable setting (double) for internal use.
     * @see setVariable(Variable, double)
     */
    public void setVariable_internal(Variable variable, double r) {
        if (Registry.SAFE_MODE) {
            setVariable(variable, r);
        } else {
            setVariable_unsafe(variable, r);
        }
    }

    /**
     * Variable setting (double) without consistency checks.
     * @see setVariable(Variable, double)
     */
    private void setVariable_unsafe(Variable variable, double r) {
        double oldValue = variable.value();
        variable.set(r);
        setVariableUpdates(variable, oldValue, false);
    }

    /**
     * Deletes an existing variable of the CSP. Use "_internal"
     * version if called without user interaction.
     * @param variable  The variable to be deleted.
     */
    public void deleteVariable(Variable variable) {
        synchronized(variables) {
            if (!variables.contains(variable)) {
                throw new Error("The variable \"" + variable +
                                "\" is not part of the CSP");
            }
        }
        if (variable.connections.number() > 0) {
                throw new Error("The variable \"" + variable +
                                "\" is still connected to constraints");
        }
        deleteVariable_unsafe(variable);
    }

    /**
     * Variable deletion for internal use.
     * @see deleteVariable(Variable)
     */
    public void deleteVariable_internal(Variable variable) {
        if (Registry.SAFE_MODE) {
            deleteVariable(variable);
        } else {
            deleteVariable_unsafe(variable);
        }
    }

    /**
     * Variable deletion without consistency checks.
     * @see deleteVariable(Variable)
     */
    private void deleteVariable_unsafe(Variable variable) {
        synchronized(variables) {
            variables.remove(variable);
        }

        scm.registerDeletedNode(variable);

        if (variable.graphObject instanceof NodeGraphics) {
            graphObject.removeVariable(variable);
            graphObject.updateGraphics();
        }
    }

    /**
     * Returns a variable's description (name and value).
     * @param variable  The variable to be described.
     */
    public String getVariableDescription(Variable variable) {
        String s = null;
        synchronized(variables) {
            if (variables.contains(variable)) {
                s = variable.toString();
            } else {
                throw new Error("The variable \"" + variable +
                                "\" is not part of the CSP");
            }
        }
        return s;
    }

    /**
     * Returns the descriptions of all variables (names and value).
     */
    public StringBuffer getVariableDescriptions() {
        StringBuffer s = new StringBuffer();
        Variable variable;
        synchronized(variables) {
            Iterator i = variables.iterator();
            if (i.hasNext()) {
                variable = (Variable) i.next();
                s.append(variable.toString());
                while (i.hasNext()) {
                    variable = (Variable) i.next();
                    s.append("\n" + variable.toString());
                }
            }
        }
        return s;
    }

    /**
     * Is the given node part of the CSP-Graph?
     * @param n  The Node to be tested.
     * @return variables.contains(n) || constraints.contains(n)
     */
    public boolean containsNode(Node n) {
        if (n instanceof Variable) {
            return variables.contains(n);
        }
        return constraints.contains(n);
    }

    /**
     * Returns a Set with all the variables.
     * Produces a copy of the set in GlobalSearchControl, so when variables
     * are added or removed from GlobalSearchControl that is not reflected in
     * the returned copy.
     * @return a copy of variables
     */
    public Set getVariables() {
        synchronized(variables) {
            return new HashSet(variables);
        }
    }

    /**
     * Coordinates update for the addition of an integer
     * constraint.
     * @param variable  The affected variable.
     */
    public void addIntegerConstraint(Variable variable) {
        double oldValue = variable.value();
        variable.addIntegerConstraint();
        if (variable.integerConstraints == 1) {
            variable.set(oldValue);
            setVariableUpdates(variable, oldValue, true);
        }
    }

    /**
     * Coordinates update for the removal of an integer
     * constraint.
     * @param variable  The affected variable.
     */
    public void removeIntegerConstraint(Variable variable) {
        double oldValue = variable.value();
        variable.removeIntegerConstraint();
        if (variable.integerConstraints == 0) {
            variable.set(oldValue);
            setVariableUpdates(variable, oldValue, true);
        }
    }

    /**
     * Coordinates update for the addition of a range
     * constraint.
     * @param variable  The affected variable.
     * @param constraint  The affected range constraint.
     */
    public void addRangeConstraint(Variable variable,
                                   RangeConstraint constraint) {

        double oldValue = variable.value();
        variable.addRangeConstraint(constraint);

        // Update variable's value

        variable.ensureBounds();
        setVariableUpdates(variable, oldValue, false);

    }

    /**
     * Coordinates update for the removal of a range
     * constraint.
     * @param variable  The affected variable.
     * @param constraint  The affected range constraint.
     */
    public void removeRangeConstraint(Variable variable,
                                      RangeConstraint constraint) {
        variable.removeRangeConstraint(constraint);
    }

    /**
     * Propagates a variable's new value to connected nodes.
     * @param variable  The affected variable.
     * @param oldValue  The variable's old value.
     * @param typeChange  Indicates whether the variable's type has changed.
     */
    private void setVariableUpdates(Variable variable, double oldValue,
                                    boolean typeChange) {

        if ((variable.value() != oldValue) || typeChange) {

            // Update of the related constraints

            connectionReferenceUpdate.node = variable;
            connectionsCopy.initialize(variable.connections);

            ConnectedNode connectedNode;
            synchronized(connectionsCopy) {
                connectionsCopy.initializeEnumerationAll();
                while ((connectedNode = connectionsCopy.getNext()) != null) {
                    if (connectedNode.direction == Registry.OUTGOING) {
                        connectionReferenceUpdate.edgeID =
                            connectedNode.edgeID;
                        connectionReferenceUpdate.direction =
                            connectedNode.direction;
                        ((ConventionalConstraint) connectedNode.node).update(oldValue, connectionReferenceUpdate, connectionReferenceUpdate);
                    }
                }
            }

            // Update graphics

            if (variable.graphObject instanceof NodeGraphics) {
                ((VariableGraphics) variable.graphObject).updateValue();
                ((NodeGraphics) variable.graphObject).updateSize();
                graphObject.updateGraphics();
            }

        }
    }


    // ***** RANGE CONSTRAINTS *****

    /**
     * Sets the lower bound of a range constraint.
     * @param constraint  The affected range constraint.
     * @param lowerBound  The new lower bound.
     */
    public void setLowerBound(ConventionalConstraint constraint,
                              double lowerBound) {

        synchronized(constraints) {
            if (!constraints.contains(constraint)) {
                throw new Error("The constraint \"" + constraint +
                                "\" is not part of the CSP");
            }
        }
        if (!(constraint instanceof RangeConstraint)) {
            throw new Error("The constraint \"" + constraint +
                            "\" is no range constraint");
        }

        RangeConstraint rconstraint = ((RangeConstraint) constraint);
        double oldlb = rconstraint.lowerBound;
        rconstraint.setLowerBound(lowerBound);
        rconstraint.variable.checkLowerBounds(oldlb, lowerBound);

        // Update graphics

        if (constraint.graphObject instanceof NodeGraphics) {
            ((RangeConstraintGraphics) constraint.graphObject).updateRange();
            ((NodeGraphics) constraint.graphObject).updateSize();
            graphObject.updateGraphics();
        }

        // Update variable's value

        double oldValue = rconstraint.variable.value();
        rconstraint.variable.ensureBounds();
        setVariableUpdates(rconstraint.variable, oldValue, false);

    }

    /**
     * Sets the lower bound of a range constraint to -infinite.
     */
    public void deactivateLowerBound(ConventionalConstraint constraint) {

        synchronized(constraints) {
            if (!constraints.contains(constraint)) {
                throw new Error("The constraint \"" + constraint +
                                "\" is not part of the CSP");
            }
        }
        if (!(constraint instanceof RangeConstraint)) {
            throw new Error("The constraint \"" + constraint +
                            "\" is no range constraint");
        }

        RangeConstraint rconstraint = ((RangeConstraint) constraint);
        double oldlb = rconstraint.lowerBound;
        rconstraint.deactivateLowerBound();
        rconstraint.variable.checkLowerBounds(oldlb, Double.NEGATIVE_INFINITY);

        // Update graphics

        if (constraint.graphObject instanceof NodeGraphics) {
            ((RangeConstraintGraphics) constraint.graphObject).updateRange();
            ((NodeGraphics) constraint.graphObject).updateSize();
            graphObject.updateGraphics();
        }

    }

    /**
     * Sets the upper bound of a range constraint.
     * @param constraint  The affected range constraint.
     * @param upperBound  The new upper bound.
     */
    public void setUpperBound(ConventionalConstraint constraint,
                              double upperBound) {

        synchronized(constraints) {
            if (!constraints.contains(constraint)) {
                throw new Error("The constraint \"" + constraint +
                                "\" is not part of the CSP");
            }
        }
        if (!(constraint instanceof RangeConstraint)) {
            throw new Error("The constraint \"" + constraint +
                            "\" is no range constraint");
        }

        RangeConstraint rconstraint = ((RangeConstraint) constraint);
        double oldub = rconstraint.upperBound;
        rconstraint.setUpperBound(upperBound);
        rconstraint.variable.checkUpperBounds(oldub, upperBound);

        // Update graphics

        if (constraint.graphObject instanceof NodeGraphics) {
            ((RangeConstraintGraphics) constraint.graphObject).updateRange();
            ((NodeGraphics) constraint.graphObject).updateSize();
            graphObject.updateGraphics();
        }

        // Update variable's value

        double oldValue = rconstraint.variable.value();
        rconstraint.variable.ensureBounds();
        setVariableUpdates(rconstraint.variable, oldValue, false);

    }

    /**
     * Sets the upper bound of a range constraint to infinite.
     */
    public void deactivateUpperBound(ConventionalConstraint constraint) {

        synchronized(constraints) {
            if (!constraints.contains(constraint)) {
                throw new Error("The constraint \"" + constraint +
                                "\" is not part of the CSP");
            }
        }
        if (!(constraint instanceof RangeConstraint)) {
            throw new Error("The constraint \"" + constraint +
                            "\" is no range constraint");
        }

        RangeConstraint rconstraint = ((RangeConstraint) constraint);
        double oldub = rconstraint.upperBound;
        rconstraint.deactivateUpperBound();
        rconstraint.variable.checkUpperBounds(oldub, Double.POSITIVE_INFINITY);

        // Update graphics

        if (constraint.graphObject instanceof NodeGraphics) {
            ((RangeConstraintGraphics) constraint.graphObject).updateRange();
            ((NodeGraphics) constraint.graphObject).updateSize();
            graphObject.updateGraphics();
        }

    }

    /**
     * Sets lower and upper bound of a range constraint.
     * @param constraint  The affected range constraint.
     * @param lowerBound  The new lower bound.
     * @param upperBound  The new upper bound.
     */
    public void setBounds(ConventionalConstraint constraint,
                          double lowerBound, double upperBound) {

        synchronized(constraints) {
            if (!constraints.contains(constraint)) {
                throw new Error("The constraint \"" + constraint +
                                "\" is not part of the CSP");
            }
        }
        if (!(constraint instanceof RangeConstraint)) {
            throw new Error("The constraint \"" + constraint +
                            "\" is no range constraint");
        }

        RangeConstraint rconstraint = ((RangeConstraint) constraint);
        double oldlb = rconstraint.lowerBound;
        double oldub = rconstraint.upperBound;
        rconstraint.setBounds(lowerBound, upperBound);
        rconstraint.variable.checkBounds(oldlb, oldub, lowerBound, upperBound);

        // Update graphics

        if (constraint.graphObject instanceof NodeGraphics) {
            ((RangeConstraintGraphics) constraint.graphObject).updateRange();
            ((NodeGraphics) constraint.graphObject).updateSize();
            graphObject.updateGraphics();
        }

        // Update variable's value

        double oldValue = rconstraint.variable.value();
        rconstraint.variable.ensureBounds();
        setVariableUpdates(rconstraint.variable, oldValue, false);
    }


    // ***** EMBEDDINGS *****

    /**
     * Extension of an embedding. Use "_internal" version if called
     * without user interaction.
     * @param embedding  The related embedding.
     * @param node  The node to be added.
     * @param edgeID  ID of the edge to be added.
     * @param direction  Direction of the edge.
     * @see Registry
     */
    public void extendEmbedding(Embedding embedding, Node node, int edgeID,
                                boolean direction) {

        // Testing whether maximal connections are exceeded

        if (embedding.isClosed()) {
            throw new Error("The embedding \"" + embedding + "\" has not been defined to include that many extensions");
        }

        // Testing whether the given node is part of the CSP

        if (node instanceof ConventionalConstraint) {
            synchronized(constraints) {
                if (!constraints.contains((ConventionalConstraint) node)) {
                    throw new Error("The constraint \"" +
                                    (ConventionalConstraint) node +
                                    "\" is not part of the CSP");
                }
            }
        } else {
            if (node instanceof Variable) {
                synchronized(variables) {
                    if (!variables.contains((Variable) node)) {
                        throw new Error("The variable \"" + (Variable) node +
                                        "\" is not part of the CSP");
                    }
                }
            } else {
                throw new Error("The node \"" + node + "\" is neither a variable nor a conventional constraint");
            }
        }

        // Testing whether the given edge is valid

        if (edgeID >= Registry.edgeTypeNumber()) {
            throw new Error("The edge ID \"" + edgeID + "\" is not valid");
        }

        // Testing for correct direction

        if ((node instanceof Variable) && (direction == Registry.OUTGOING)) {
            throw new Error("Value propagation cannot proceed to variables");
        }
        if ((node instanceof ConventionalConstraint)
            && !(node instanceof ObjectConstraint)
            && (direction == Registry.INCOMING)) {
            throw new Error("Value propagation cannot proceed from conventional constraints");
        }

        extendEmbedding_unsafe(embedding, node, edgeID, direction);
    }

    /**
     * Embedding extension for internal use.
     * @see extendEmbedding(Embedding, Node, int, boolean)
     */
    public void extendEmbedding_internal(Embedding embedding, Node node,
                                         int edgeID, boolean direction) {
        if (Registry.SAFE_MODE) {
            extendEmbedding(embedding, node, edgeID, direction);
        } else {
            extendEmbedding_unsafe(embedding, node, edgeID, direction);
        }
    }

    /**
     * Embedding extension without consistency checks.
     * @see extendEmbedding(Embedding, Node, int, boolean)
     */
    private void extendEmbedding_unsafe(Embedding embedding, Node node,
                                        int edgeID, boolean direction) {
        embedding.addConnectedNode(node, node.typeID, edgeID, direction);
    }

    /**
     * Checks whether an embedding is valid. If a constraint is also
     * given, it is additionally checked whether the embedding is a
     * subset of the constraint's embedding.
     * @param embedding  The embedding to be checked.
     * @param constraint  The constraint to be checked ("null" if none).
     */
    private void checkEmbeddingValidity(Embedding embedding,
                                        ConventionalConstraint constraint) {

        // Testing whether the embedding has been filled

        if (!embedding.isClosed()) {
            throw new Error("The embedding \"" + embedding + "\" has not been extended up to its initial definition");
        }

        // Testing whether the nodes are still part of the CSP

        int i;
        ConnectedNode connectedNode;
        Node node;
        for (i = 0; i < embedding.connectedNodeNumber; i++) {
            connectedNode = embedding.connectedNode[i];
            node = connectedNode.node;

            // Testing existence

            if (node instanceof ConventionalConstraint) {
                synchronized(constraints) {
                    if (!constraints.contains((ConventionalConstraint) node)) {
                        throw new Error("The constraint \"" +
                                        (ConventionalConstraint) node +
                                        "\" is not part of the CSP");
                    }
                }
            } else {
                if (node instanceof Variable) {
                    synchronized(variables) {
                        if (!variables.contains((Variable) node)) {
                            throw new Error("The variable \"" +
                                            (Variable) node +
                                            "\" is not part of the CSP");
                        }
                    }
                } else {
                    throw new Error("The node \"" + node + "\" is neither a variable nor a conventional constraint");
                }
            }

            // Testing correspondence with the constraint's embedding

            if (constraint != null) {
                if (!constraint.connections.contains(node, Registry.getNodeType(node), connectedNode.edgeID, connectedNode.direction)) {
                    throw new Error("The embedding \"" + embedding + "\" is not a subset of the constraint's current embedding");
                }
            }

        }
    }


    // ***** CONSTRAINTS *****

    /**
     * Creates a new constraint instance for the CSP.
     * The related graphical component will only be constructed
     * if the graphics are already turned on. Use "_internal"
     * version if called without user interaction.
     * @param constraintID  The type of the constraint.
     * @param embedding  The base graph initialization for the constraint.
     * @param costID  The ID of the constraint's costs to be used.
     * @param costMappingOptions  Arbitrary object to pass information to the
     *                            costMapping.
     * @see Registry
     */
    public ConventionalConstraint createConstraint(int constraintID,
                                                   Embedding embedding,
                                                   int costID,
                                                   Object costMappingOptions) {
        verifyConstraintCreation(constraintID, embedding, costID,
                                 costMappingOptions);
        return createConstraint_unsafe(constraintID, embedding,
                                       costID, costMappingOptions);
    }

    /**
     * Shortcut function for constraint creation using no specific cost
     * mapping options.
     * @see createConstraint(int, Embedding, int, Object)
     */
    public ConventionalConstraint createConstraint(int constraintID, Embedding embedding, int costID) {
        return createConstraint(constraintID, embedding, costID, null);
    }
    /**
     * Shortcut function for constraint creation using the default cost ID.
     * @see createConstraint(int, Embedding, int, Object)
     */
    public ConventionalConstraint createConstraint(int constraintID, Embedding embedding, Object costMappingOptions) {
        return createConstraint(constraintID, embedding, 0, costMappingOptions);
    }
    /**
     * Shortcut function for constraint creation using the default cost ID
     * and no specific cost mapping options.
     * @see createConstraint(int, Embedding, int, Object)
     */
    public ConventionalConstraint createConstraint(int constraintID, Embedding embedding) {
        return createConstraint(constraintID, embedding, 0, null);
    }

    /**
     * Verifies that a constraint creation is correct.
     * @param constraintID  The type of the constraint.
     * @param embedding  The base graph initialization for the constraint.
     * @param costID  The ID of the constraint's costs to be used.
     * @param costMappingOptions  Arbitrary object to pass information to the
     *                            costMapping.
     * @see Registry
     */
    private void verifyConstraintCreation(int constraintID, Embedding embedding, int costID, Object costMappingOptions) {

        if (!Registry.isConstraint(constraintID)) {
            throw new Error("Invalid constraint ID: " + constraintID);
        }
        checkEmbeddingValidity(embedding, null);

        // Cost-related tests

        if ((costMappingOptions != null) || (costID != 0)) {
            Node nodeType = Registry.nodeType[constraintID];
            if ((nodeType instanceof ObjectConstraint)
                || (nodeType instanceof IntegerConstraint)
                || (nodeType instanceof RangeConstraint)) {
                if (costMappingOptions != null) {
                    throw new Error("Cost mapping options cannot be passed to this constraint type");
                } else {
                    throw new Error("A cost ID specification is not applicable for this constraint type");
                }
            } else {
                if ((costID < 0) || (costID >= ((Constraint) Registry.nodeType[constraintID]).getNumberOfCosts())) {
                    throw new Error("Invalid costID");
                }
            }
        }

        // Testing whether the base graph is correct

        if (!embedding.matchesType(((ConventionalConstraint) Registry.nodeType[constraintID]).getBaseGraph())) {
            throw new Error("The base graph of constraint type \"" + Registry.nodeTypeName[constraintID] + "\" does not match the given embedding");
        }

        // Testing whether maximal graphs of the extension graphs
        // constraints are violated

        int i;
        boolean correct = true;
        ConnectedNode connectedNode;
        ConventionalConstraint embeddingConstraint;
        for (i = 0; i < embedding.connectedNodeNumber; i++) {
            connectedNode = embedding.connectedNode[i];
            if (connectedNode.node instanceof ConventionalConstraint) {
                embeddingConstraint =
                    (ConventionalConstraint) connectedNode.node;
                if (embeddingConstraint.getMaximalConnectionNumber() <
                    embeddingConstraint.connections.number() +
                    embedding.numberOfNodeOccurences(embeddingConstraint)) {
                    correct = false;
                }
            }
        }
        if (!correct) {
            throw new Error("The maximal embedding of an extension node is exceeded");
        }

    }

    /**
     * Constraint creation for internal use.
     * @see createConstraint(int, Embedding, int, Object)
     */
    public ConventionalConstraint createConstraint_internal(int constraintID, Embedding embedding, int costID, Object costMappingOptions) {
        if (Registry.SAFE_MODE) {
            return createConstraint(constraintID, embedding, costID,
                                    costMappingOptions);
        } else {
            return createConstraint_unsafe(constraintID, embedding, costID,
                                           costMappingOptions);
        }
    }

    /**
     * Shortcut function for internal constraint creation using no specific
     * cost mapping options.
     * @see createConstraint_internal(int, Embedding, int, Object)
     */
    public ConventionalConstraint createConstraint_internal(int constraintID, Embedding embedding, int costID) {
        return createConstraint_internal(constraintID, embedding, costID, null);
    }
    /**
     * Shortcut function for internal constraint creation using the default
     * cost ID.
     * @see createConstraint_internal(int, Embedding, int, Object)
     */
    public ConventionalConstraint createConstraint_internal(int constraintID, Embedding embedding, Object costMappingOptions) {
        return createConstraint_internal(constraintID, embedding, 0, costMappingOptions);
    }
    /**
     * Shortcut function for internal constraint creation using the default
     * cost ID and no specific cost mapping options.
     * @see createConstraint_internal(int, Embedding, int, Object)
     */
    public ConventionalConstraint createConstraint_internal(int constraintID, Embedding embedding) {
        return createConstraint_internal(constraintID, embedding, 0, null);
    }

    /**
     * Constraint creation without consistency checks.
     * @see createConstraint(int, Embedding, int, Object)
     */
    private ConventionalConstraint createConstraint_unsafe(int constraintID, Embedding embedding, int costID, Object costMappingOptions) {

        // Creating the constraint

        ConventionalConstraint constraint =
            Registry.createConstraint(constraintID, this);

        Node nodeType = Registry.nodeType[constraintID];

        synchronized(constraints) {
            constraints.add(constraint);
        }
        synchronized(constraintSHT) {
            constraintSHT.add(nodeType, constraint);
        }

        // Notifying structural checking

        scm.registerNewNode(constraint);

        // Adding cost handling

        if (!(nodeType instanceof ObjectConstraint)
            && !(nodeType instanceof IntegerConstraint)
            && !(nodeType instanceof RangeConstraint)) {

            // Creating default cost function if needed

            if (currentCostFunction == null) {
                DEFAULT_FUNCTION = createCostFunction("Default Cost Function");
                currentCostFunction = DEFAULT_FUNCTION;
            }

            // Adding constraint costs' handling to cost function

            currentCostFunction.addConstraint(constraint, costID,
                                              currentCostMapping,
                                              costMappingOptions);

        }

        // Create constraint's graphical component if graphics are turned on

        if (graphObject != null) {
            constraint.createGraphics();
            ((NodeGraphics) constraint.graphObject).updateSize();
            graphObject.displayConstraint(constraint);
            ((NodeGraphics) constraint.graphObject).revalidate();
            graphObject.updateGraphics();
        }

        // Adding embedding edges

        addEmbedding(embedding, constraint);
        fireChange(constraint); // inform all changeListeners about the new constraint (inserted by PlanPresentation-Development)


        return constraint;
    }

    /**
     * Creates a new constraint instance with x/y coordinates for
     * visual placement. The related graphical component will be also
     * constructed if the graphics are turned on. Otherwise, a cache to
     * store the graphical information will be created. Use "_internal"
     * version if called without user interaction.
     * @param constraintID  The type of the constraint.
     * @param embedding  The base graph initialization for the constraint.
     * @param costID  The ID of the constraint's costs to be used.
     * @param costMappingOptions  Arbitrary object to pass information to the
     *                            costMapping.
     * @param x  The x coordinate.
     * @param y  The y coordinate.
     * @see Registry
     */
    public ConventionalConstraint createConstraint(int constraintID,
                                                   Embedding embedding,
                                                   int costID,
                                                   Object costMappingOptions,
                                                   int x, int y) {
        verifyConstraintCreation(constraintID, embedding, costID,
                                 costMappingOptions);
        return createConstraint_unsafe(constraintID, embedding,
                                       costID, costMappingOptions, x, y);
    }

    /**
     * Shortcut function for constraint creation with x/y coordinates for
     * visual placement using no specific cost mapping options.
     * @see createConstraint(int, Embedding, int, Object, int, int)
     */
    public ConventionalConstraint createConstraint(int constraintID, Embedding embedding, int costID, int x, int y) {
        return createConstraint(constraintID, embedding, costID, null, x, y);
    }
    /**
     * Shortcut function for constraint creation with x/y coordinates for
     * visual placement using the default cost ID.
     * @see createConstraint(int, Embedding, int, Object)
     */
    public ConventionalConstraint createConstraint(int constraintID, Embedding embedding, Object costMappingOptions, int x, int y) {
        return createConstraint(constraintID, embedding, 0, costMappingOptions, x, y);
    }
    /**
     * Shortcut function for constraint creation with x/y coordinates for
     * visual placement using the default cost ID and no specific cost
     * mapping options.
     * @see createConstraint(int, Embedding, int, Object, int, int)
     */
    public ConventionalConstraint createConstraint(int constraintID, Embedding embedding, int x, int y) {
        return createConstraint(constraintID, embedding, 0, null, x, y);
    }

    /**
     * Constraint creation (x/y) for internal use.
     * @see createConstraint(int, Embedding, int, Object, int, int)
     */
    public ConventionalConstraint createConstraint_internal(int constraintID, Embedding embedding, int costID, Object costMappingOptions, int x, int y) {
        if (Registry.SAFE_MODE) {
            return createConstraint(constraintID, embedding, costID,
                                    costMappingOptions, x, y);
        } else {
            return createConstraint_unsafe(constraintID, embedding, costID,
                                           costMappingOptions, x, y);
        }
    }

    /**
     * Shortcut function for internal constraint creation with x/y
     * coordinates for visual placement using no specific cost mapping
     * options.
     * @see createConstraint_internal(int, Embedding, int, Object, int, int)
     */
    public ConventionalConstraint createConstraint_internal(int constraintID, Embedding embedding, int costID, int x, int y) {
        return createConstraint_internal(constraintID, embedding, costID, null, x, y);
    }
    /**
     * Shortcut function for internal constraint creation with x/y
     * coordinates for visual placement using the default cost ID.
     * @see createConstraint_internal(int, Embedding, int, Object)
     */
    public ConventionalConstraint createConstraint_internal(int constraintID, Embedding embedding, Object costMappingOptions, int x, int y) {
        return createConstraint_internal(constraintID, embedding, 0, costMappingOptions, x, y);
    }

    /**
     * Shortcut function for internal constraint creation with x/y
     * coordinates for visual placement using the default cost ID and
     * no specific cost mapping options.
     * @see createConstraint_internal(int, Embedding, int, Object, int, int)
     */
    public ConventionalConstraint createConstraint_internal(int constraintID, Embedding embedding, int x, int y) {
        return createConstraint_internal(constraintID, embedding, 0, null, x, y);
    }

    /**
     * Shortcut function for internal constraint creation with x/y
     * coordinates for visual placement using the default cost ID and
     * no specific cost mapping options.
     * The ConnectedNode Objects of the embedding are used in the new Constraint.
     *
     * @see createConstraint_internal(int, Embedding, int, Object, int, int)
     */
    public ConventionalConstraint createConstraintUseEdgeObjects_internal( int constraintID, Embedding embedding, int x, int y) {
        if (Registry.SAFE_MODE) {
            verifyConstraintCreation(constraintID, embedding, 0, null);
        }
        // Creating the constraint

        ConventionalConstraint constraint =
            Registry.createConstraint(constraintID, this);

        Node nodeType = Registry.nodeType[constraintID];

        synchronized(constraints) {
            constraints.add(constraint);
        }
        synchronized(constraintSHT) {
            constraintSHT.add(nodeType, constraint);
        }

        // Notifying structural checking

        scm.registerNewNode(constraint);

        // Adding cost handling

        if (!(nodeType instanceof ObjectConstraint)
            && !(nodeType instanceof IntegerConstraint)
            && !(nodeType instanceof RangeConstraint)) {

            // Creating default cost function if needed

            if (currentCostFunction == null) {
                DEFAULT_FUNCTION = createCostFunction("Default Cost Function");
                currentCostFunction = DEFAULT_FUNCTION;
            }

            // Adding constraint costs' handling to cost function

            currentCostFunction.addConstraint(constraint, 0,
                                              currentCostMapping,
                                              null);

        }

        // Create constraint's graphObject if graphics are turned on and
        // a graphics cache otherwise

        if (graphObject != null) {
            constraint.createGraphics();
            ((NodeGraphics) constraint.graphObject).x = x;
            ((NodeGraphics) constraint.graphObject).y = y;
            ((NodeGraphics) constraint.graphObject).updateSize();
            graphObject.displayConstraint(constraint);
            ((NodeGraphics) constraint.graphObject).revalidate();
            graphObject.updateGraphics();
        } else {
            constraint.graphObject = (Object) new NodeGraphicsCache(x, y);
        }

        // Adding embedding edges

        addEmbeddingUseEdgeObjects(embedding, constraint);

        return constraint;
    }

    /**
     * Constraint creation (x/y) without consistency checks.
     * @see createConstraint(int, Embedding, int, Object, int, int)
     */
    private ConventionalConstraint createConstraint_unsafe(int constraintID, Embedding embedding, int costID, Object costMappingOptions, int x, int y) {

        // Creating the constraint

        ConventionalConstraint constraint =
            Registry.createConstraint(constraintID, this);

        Node nodeType = Registry.nodeType[constraintID];

        synchronized(constraints) {
            constraints.add(constraint);
        }
        synchronized(constraintSHT) {
            constraintSHT.add(nodeType, constraint);
        }

        // Notifying structural checking

        scm.registerNewNode(constraint);

        // Adding cost handling

        if (!(nodeType instanceof ObjectConstraint)
            && !(nodeType instanceof IntegerConstraint)
            && !(nodeType instanceof RangeConstraint)) {

            // Creating default cost function if needed

            if (currentCostFunction == null) {
                DEFAULT_FUNCTION = createCostFunction("Default Cost Function");
                currentCostFunction = DEFAULT_FUNCTION;
            }

            // Adding constraint costs' handling to cost function

            currentCostFunction.addConstraint(constraint, costID,
                                              currentCostMapping,
                                              costMappingOptions);

        }

        // Create constraint's graphObject if graphics are turned on and
        // a graphics cache otherwise

        if (graphObject != null) {
            constraint.createGraphics();
            ((NodeGraphics) constraint.graphObject).x = x;
            ((NodeGraphics) constraint.graphObject).y = y;
            ((NodeGraphics) constraint.graphObject).updateSize();
            graphObject.displayConstraint(constraint);
            ((NodeGraphics) constraint.graphObject).revalidate();
            graphObject.updateGraphics();
        } else {
            constraint.graphObject = (Object) new NodeGraphicsCache(x, y);
        }

        // Adding embedding edges

        addEmbedding(embedding, constraint);

        fireChange(constraint); // inform all changeListeners about the new constraint (inserted by PlanPresentation-Development)

        return constraint;
    }

    /**
     * Extends a constraint instance by an extension graph.
     * Use "_internal" version if called without user interaction.
     * @param constraint  The constraint to be extended.
     * @param embedding  The extension graph for the constraint.
     */
    public void extendConstraint(ConventionalConstraint constraint,
                                 Embedding embedding) {
        extendConstraintSafetyMeasures(constraint, embedding);


        extendConstraint_unsafe(constraint, embedding);
    }

    /**
     * Does the security stuff of extendConstraint
     * @param constraint  The constraint to be extended.
     * @param embedding  The extension graph for the constraint.
     */
    private void extendConstraintSafetyMeasures(ConventionalConstraint constraint,
                                    Embedding embedding) {
    synchronized(constraints) {
            if (!constraints.contains(constraint)) {
                throw new Error("The constraint \"" + constraint +
                                "\" is not part of the CSP");
            }
        }
        checkEmbeddingValidity(embedding, null);

        // Testing whether the constraint's maximal graph already exists

        if (constraint.getMaximalConnectionNumber() ==
            constraint.connections.number()) {
            throw new Error("The maximal embedding of the constraint \"" +
                            constraint + "\"  would be exceeded");
        }

        // Testing whether the extension is an extension graph

        int i;
        boolean correct = false;
        for (i = 0; (i < constraint.getExtensionGraphNumber()) &&
                 !correct; i++) {
            correct = embedding.matchesType(constraint.getExtensionGraph(i));
        }

        if (!correct) {
            throw new Error("The constraint's extension graphs do not match the given embedding");
        }

        ConnectedNode connectedNode;
        ConventionalConstraint embeddingConstraint;
        for (i = 0; i < embedding.connectedNodeNumber; i++) {
            connectedNode = embedding.connectedNode[i];

            // Testing for correct direction

            if (!(constraint instanceof ObjectConstraint)
                && (connectedNode.direction == Registry.OUTGOING)) {
                throw new Error("Value propagation cannot proceed from conventional constraints");
            }

            // Testing whether maximal graphs of the extension graph's
            // constraints are violated

            if (connectedNode.node instanceof ConventionalConstraint) {
                embeddingConstraint =
                    (ConventionalConstraint) connectedNode.node;
                if (embeddingConstraint.getMaximalConnectionNumber() <
                    embeddingConstraint.connections.number() +
                    embedding.numberOfNodeOccurences(embeddingConstraint)) {
                    throw new Error("The maximal embedding of an extension node would be exceeded");            }
            }
        }
    }

    /**
     * Constraint extension for internal use.
     * @see extendConstraint(ConventionalConstraint, Embedding)
     */
    public void extendConstraint_internal(ConventionalConstraint constraint,
                                          Embedding embedding) {
        if (Registry.SAFE_MODE) {
            extendConstraint(constraint, embedding);
        } else {
            extendConstraint_unsafe(constraint, embedding);
        }
    }

    /**
     * Constraint extension without consistency checks.
     * @see extendConstraint(ConventionalConstraint, Embedding)
     */
    private void extendConstraint_unsafe(ConventionalConstraint constraint,
                                         Embedding embedding) {

        addEmbedding(embedding, constraint);

        // Update graphics if turned on

        if (graphObject != null) {
            graphObject.updateGraphics();
        }

    }

    /**
     * Constraint extension for internal use.
     * Uses the ConnectedNode Objects given in the embedding, does not create new ones.
     * @see extendConstraint(ConventionalConstraint, Embedding)
     */
    public void extendConstraintUseEdgeObjects_internal(ConventionalConstraint constraint,
                                          Embedding embedding) {
        if (Registry.SAFE_MODE) {
            extendConstraintSafetyMeasures(constraint, embedding);
        }
        extendConstraintUseEdgeObjects_unsafe(constraint, embedding);
    }


    /**
     * Constraint extension without consistency checks.
     * Uses the ConnectedNode Objects given in the embedding, does not create new ones.
     * @see extendConstraint(ConventionalConstraint, Embedding)
     */
    private void extendConstraintUseEdgeObjects_unsafe(ConventionalConstraint constraint,
                                         Embedding embedding) {

        addEmbeddingUseEdgeObjects(embedding, constraint);

        // Update graphics if turned on

        if (graphObject != null) {
            graphObject.updateGraphics();
        }

    }

    /**
     * Reduces a constraint instance by a given extension graph.
     * Use "_internal" version if called without user interaction.
     * @param constraint  The constraint to be extended.
     * @param embedding  The extension graph part of the constraint.
     */
    public void reduceConstraint(ConventionalConstraint constraint,
                                 Embedding embedding) {

        synchronized(constraints) {
            if (!constraints.contains(constraint)) {
                throw new Error("The constraint \"" + constraint +
                                "\" is not part of the CSP");
            }
        }
        checkEmbeddingValidity(embedding, constraint);

        // Testing whether the embedding falls below the constraint's
        // base graph

        if (!constraint.keepsBaseGraph(embedding)) {
            throw new Error("A constraint cannot be reduced below its base graph");
        }

        // Testing whether the extension is an extension graph

        int i;
        boolean found = false;
        for (i = 0; (i < constraint.getExtensionGraphNumber())
                 && !found; i++) {
            found = embedding.matchesType(constraint.getExtensionGraph(i));
        }

        if (!found) {
            throw new Error("The constraint's extension graphs do not match the given embedding");
        }

        reduceConstraint_unsafe(constraint, embedding);
    }

    /**
     * Constraint reduction for internal use.
     * @see reduceConstraint(ConventionalConstraint, Embedding)
     */
    public void reduceConstraint_internal(ConventionalConstraint constraint,
                                          Embedding embedding) {
        if (Registry.SAFE_MODE) {
            reduceConstraint(constraint, embedding);
        } else {
            reduceConstraint_unsafe(constraint, embedding);
        }
    }

    /**
     * Constraint reduction without consistency checks.
     * @see reduceConstraint(ConventionalConstraint, Embedding)
     */
    private void reduceConstraint_unsafe(ConventionalConstraint constraint,
                                         Embedding embedding) {

        int i;
        for (i = 0; i < embedding.connectedNodeNumber; i++) {
            disconnectNode(constraint, embedding.connectedNode[i]);
        }

        // Update graphics if turned on

        if (graphObject != null) {
            graphObject.updateGraphics();
        }

    }

    /**
     * Deletes an existing constraint of the CSP.
     * Use "_internal" version if called without user interaction.
     * @param constraint  The constraint to be deleted.
     */
    public void deleteConstraint(ConventionalConstraint constraint) {

        synchronized(constraints) {
            if (!constraints.contains(constraint)) {
                throw new Error("The constraint \"" + constraint +
                                "\" is not part of the CSP");
            }
        }

        // Testing whether the constraint has been reduced to its base graph

        if (constraint.getBaseGraph().connectedNodeNumber !=
            constraint.connections.number()) {
            throw new Error("The constraint \"" + constraint +
                            "\" has not been reduced to its base graph");
        }

        deleteConstraint_unsafe(constraint);
    }

    /**
     * Constraint deletion for internal use.
     * @see deleteConstraint(ConventionalConstraint)
     */
    public void deleteConstraint_internal(ConventionalConstraint constraint) {
        if (Registry.SAFE_MODE) {
            deleteConstraint(constraint);
        } else {
            deleteConstraint_unsafe(constraint);
        }
    }

    /**
     * Constraint deletion without consistency checks.
     * @see deleteConstraint(ConventionalConstraint)
     */
    private void deleteConstraint_unsafe(ConventionalConstraint constraint) {

        // Update graph

        Connections connections = constraint.connections;
        int nodeNumber = connections.number();
        ConnectedNode[] connectedNode = new ConnectedNode[nodeNumber];
        int i;

        synchronized(connections) {
            connections.initializeEnumerationAll();
            for (i = 0; i < nodeNumber; i++) {
                connectedNode[i] = connections.getNext();
            }
        }
        for (i = 0; i < nodeNumber; i++) {
            disconnectNode(constraint, connectedNode[i]);
        }

        synchronized(constraints) {
            constraints.remove(constraint);
        }
        synchronized(constraintSHT) {
            constraintSHT.remove(Registry.nodeType[constraint.typeID],
                                 constraint);
        }

        // Notify structural constraint search control of changes

        scm.registerDeletedNode(constraint);

        // Update graphics

        if (constraint.graphObject instanceof NodeGraphics) {
            graphObject.removeConstraint(constraint);
            graphObject.updateGraphics();
        }

        // Update cost functions

        if (!(constraint instanceof ObjectConstraint)
            && !(constraint instanceof IntegerConstraint)
            && !(constraint instanceof RangeConstraint)) {

            int numberofCosts = constraint.getNumberOfCosts();
            CostFunctionReference costFunctionReference;

            for (i = 0; i < numberofCosts; i++) {
                costFunctionReference = constraint.getCostFunctionReference(i);
                if (costFunctionReference != null) {
                    costFunctionReference.costFunction.removeConstraint(constraint, i);
                }
            }

        }
                fireChange(constraint);
    }

    /**
     * Returns a copy of the set of constraints in the GlobalSearchControl.
     * Any changes to the list of constraints are not reflected in the copy!
     * @return a copy of constraints
     */
     public Set getConstraints() {
         synchronized(constraints) {
             return new HashSet(constraints);
         }
     }

    /**
     * Adding the edges of a constraint's base/extension graph.
     * @param embedding  The embedding to be added.
     * @param constraint  The related constraint.
     */
    private void addEmbedding(Embedding embedding,
                              ConventionalConstraint constraint) {
        int i;
        for (i = 0; i < embedding.connectedNodeNumber; i++) {
            connectNode(constraint, embedding.connectedNode[i].node,
                        embedding.connectedNode[i].edgeID,
                        embedding.connectedNode[i].direction);
        }
    }


    /**
     * Adding the edges of a constraint's base/extension graph.
     * Uses the ConnectedNode Objects given in the embedding, does not create new ones.
     *
     * @param embedding  The embedding to be added.
     * @param constraint  The related constraint.
     */
    private void addEmbeddingUseEdgeObjects(Embedding embedding,
                              ConventionalConstraint constraint) {
        int i;
        for (i = 0; i < embedding.connectedNodeNumber; i++) {
            connectNode(constraint, embedding.connectedNode[i]);
        }
    }

    /**
     * Connects a node to a constraint.
     * @param constraint  The constraint to be connected.
     * @param node  The node to be connected.
     * @param edgeType  The type of the connecting edge.
     * @param direction  The edge's direction WRT the constraint.
     * @see Registry
     */
    private void connectNode(ConventionalConstraint constraint, ConnectedNode edge) {

        // Adapting graph
        constraint.connections.add(edge);
        connectNodeUpdatePropagationAndGraphicsData(constraint, edge.node, edge.edgeID,
                                   edge.direction);
    }

    /**
     * Connects a node to a constraint.
     * @param constraint  The constraint to be connected.
     * @param node  The node to be connected.
     * @param edgeType  The type of the connecting edge.
     * @param direction  The edge's direction WRT the constraint.
     * @see Registry
     */
    private void connectNode(ConventionalConstraint constraint, Node node,
                             int edgeID, boolean direction) {

        // Adapting graph
     
        constraint.connections.add(node, Registry.nodeType[node.typeID],
                                   edgeID, direction);
        connectNodeUpdatePropagationAndGraphicsData(constraint, node, edgeID, direction);
    }

    /**
     * Updates the propagation data and the Graphics, when a new Edge was added.
     * @param constraint  The constraint to be connected.
     * @param node  The node to be connected.
     * @param edgeType  The type of the connecting edge.
     * @param direction  The edge's direction WRT the constraint.
     * @see Registry
     */
    private void connectNodeUpdatePropagationAndGraphicsData(ConventionalConstraint constraint, Node node,
                             int edgeID, boolean direction) {
        if ((node instanceof ConventionalConstraint)
            && (direction == Registry.OUTGOING)) {

            // Establishing propagation link if object constraint

            if (constraint instanceof ObjectConstraint) {
                ((ObjectConstraint) constraint).addConstraint(node, edgeID,
                                                              direction);
            }

            // Propagating connection to node

            connectionReferenceConnection.node = constraint;
            connectionReferenceConnection.direction = !direction;
            ((ConventionalConstraint) node).connectNode(connectionReferenceConnection, connectionReferenceConnection);

        } else {
            if (direction == Registry.INCOMING) {

                // Establishing propagation link if object constraint

                if (node instanceof ObjectConstraint) {
                    ((ObjectConstraint) node).addConstraint(constraint, edgeID,
                                                            !direction);
                }

                // Propagating connection to constraint

                connectionReferenceConnection.node = node;
                connectionReferenceConnection.edgeID = edgeID;
                connectionReferenceConnection.direction = direction;
                constraint.connectNode(connectionReferenceConnection,
                                       connectionReferenceConnection);

            }
        }

        // Updating graphics

        if ((constraint.graphObject instanceof NodeGraphics)
            && !(constraint instanceof IntegerConstraint)
            && !(constraint instanceof RangeConstraint)
            && !(constraint instanceof ObjectConstraint)) {
            ((ConventionalConstraintGraphics) constraint.graphObject).updateCosts();
            ((NodeGraphics) constraint.graphObject).updateSize();
        }

        // Notify scm of change of nodes

        scm.registerExpandedNode(constraint, new ConnectedNode(constraint, node.typeID, edgeID, direction));
        scm.registerExpandedNode(node, new ConnectedNode(node, constraint.typeID, edgeID, !direction));

    }

    /**
     * Disconnects a node from a constraint.
     * @param constraint  The constraint to be disconnected.
     * @param connectedNode  The connection to be removed.
     */
    private void disconnectNode(ConventionalConstraint constraint,
                                ConnectedNode connectedNode) {

        Node otherNode = connectedNode.node;

        if ((otherNode instanceof ConventionalConstraint)
            && (connectedNode.direction == Registry.OUTGOING)) {

            // Propagating disconnection to node

            connectionReferenceDisconnection.node = constraint;
            connectionReferenceDisconnection.direction =
                !connectedNode.direction;
            ((ConventionalConstraint) otherNode).disconnectNode(connectionReferenceDisconnection, connectionReferenceDisconnection);

            // Removing propagation link if object constraint

            if (constraint instanceof ObjectConstraint) {
                ((ObjectConstraint) constraint).removeConstraint(otherNode, connectedNode.edgeID, connectedNode.direction);
            }

        } else {
            if (connectedNode.direction == Registry.INCOMING) {

                // Propagating disconnection to constraint

                connectionReferenceDisconnection.node = otherNode;
                connectionReferenceDisconnection.edgeID = connectedNode.edgeID;
                connectionReferenceDisconnection.direction =
                    connectedNode.direction;
                constraint.disconnectNode(connectionReferenceDisconnection,
                                          connectionReferenceDisconnection);

                // Removing propagation link if object constraint

                if (otherNode instanceof ObjectConstraint) {
                    ((ObjectConstraint) otherNode).removeConstraint(constraint, connectedNode.edgeID, !connectedNode.direction);
                }

            }
        }

        // Adapting graph

        ConnectedNode removedCN = constraint.connections.remove(otherNode, Registry.nodeType[otherNode.typeID], connectedNode.edgeID, connectedNode.direction);

        // Updating graphics

        if ((constraint.graphObject instanceof NodeGraphics)
            && !(constraint instanceof IntegerConstraint)
            && !(constraint instanceof RangeConstraint)
            && !(constraint instanceof ObjectConstraint)) {
            ((ConventionalConstraintGraphics) constraint.graphObject).updateCosts();
            ((NodeGraphics) constraint.graphObject).updateSize();
        }

        if (removedCN != null) {
            // Register change at structural constraint search control (for
            // both sides of the connection)
            scm.registerReducedNode(constraint, removedCN);
            scm.registerReducedNode(otherNode, removedCN.other);
        }

    }


    // ***** COST FUNCTIONS *****

    /**
     * Creates and returns a new cost mapping instance.
     * @param costMappingID  The cost mapping's class ID.
     */
     public CostMapping createCostMapping(int costMappingID) {
         CostMapping cm = Registry.createCostMapping(costMappingID);
         synchronized(costMappings) {
             costMappings.add(cm);
         }
         return cm;
     }

    /**
     * Creates and returns a named new cost mapping instance.
     * @param costMappingID  The cost mapping's class ID.
     * @param name  The name to be set for the new cost mapping.
     */
     public CostMapping createCostMapping(int costMappingID, String name) {
         CostMapping cm = createCostMapping(costMappingID);
         cm.setName(name);
         return cm;
     }

     /**
      * Returns a copy of the costmappings.
      */
     public java.util.List getCostMappings() {
        synchronized(costMappings) {
            ArrayList cm = new ArrayList(costMappings);
            return cm;
         }
     }

     /**
      * Returns the CostMapping with the given name, if it is present,
      * else null.
      * @param name  The name of the searched mapping.
      */
     public CostMapping getCostMapping(String name) {
         synchronized(costMappings) {
             Iterator it =  costMappings.iterator();
             while (it.hasNext()) {
                 CostMapping cm = (CostMapping) it.next();
                 if (cm.getName().equals(name)) {
                     return cm;
                 }
             }
         }
         return null;
     }


    /**
     * Setting a cost mapping for following constraint creation. Use
     * "_internal" version if called without user interaction.
     */
    public void selectCostMapping(CostMapping costMapping) {
        synchronized(costMappings) {
            if (!costMappings.contains(costMapping)) {
                throw new Error("The cost mapping \"" + costMapping +
                                "\" is not part of the solver");
            }
        }
        selectCostMapping_unsafe(costMapping);
    }

    /**
     * Version of selectCostMapping for internal use.
     * @see selectCostMapping(CostMapping)
     */
    public void selectCostMapping_internal(CostMapping costMapping) {
        if (Registry.SAFE_MODE) {
            selectCostMapping(costMapping);
        } else {
            selectCostMapping_unsafe(costMapping);
        }
    }

    /**
     * Cost mapping selection without consistency checks.
     * @see selectCostMapping(CostMapping)
     */
    private void selectCostMapping_unsafe(CostMapping costMapping) {
        currentCostMapping = costMapping;
        useCostMappingOfCostFunction = false;
    }

    /**
     * Selects the solver's default cost mapping for following constraint
     * creations.
     */
    public void selectDefaultCostMapping() {
        currentCostMapping = defaultCostMapping;
        useCostMappingOfCostFunction = false;
    }

    /**
     * Demands the use of the current cost function's cost mapping for
     * following constraint creations.
     */
    public void selectCostMappingOfCostFunction() {
        if (currentCostFunction == null) {
            currentCostMapping = defaultCostMapping;
        } else {
            currentCostMapping = currentCostFunction.costMapping;
        }
        useCostMappingOfCostFunction = true;
    }

    /**
     * Returns the current cost mapping that was selected.
     */
    public CostMapping getSelectedCostMapping() {
        return currentCostMapping;
    }

    /**
     * Returns the default cost mapping.
     */
    public CostMapping getDefaultCostMapping() {
        return defaultCostMapping;
    }

    /**
     * Sets the given costMapping as default.
     */
    public void setDefaultCostMapping( CostMapping cm ) {
        defaultCostMapping = cm;
    }


    /**
     * Creates a new cost function. It is included at the highest
     * importance level in the list of cost functions (but importance
     * being less than that of the function to guarantee structural
     * satisfaction).
     * @param name  The name of the cost function.
     */
    public CostFunction createCostFunction(String name) {

        CostFunction costFunction;

        if (currentCostMapping == null) {
            costFunction = new CostFunction(name, this,
                                            costFunctionNumber - 1,
                                            defaultCostMapping);
        } else {
            costFunction = new CostFunction(name, this,
                                            costFunctionNumber - 1,
                                            currentCostMapping);
        }

        synchronized(costFunctions) {
            costFunctions.add(costFunctionNumber - 1, costFunction);
            costFunctionNumber++;
        }

        return costFunction;
    }

    /**
     * Returns the CostFunction with the given name, if it is present,
     * else null.
     * @param name  The name of the searched function.
     */
     public CostFunction getCostFunction(String name) {
         synchronized(costFunctions) {
             Iterator it =  costFunctions.iterator();
             while (it.hasNext()) {
                 CostFunction cf = (CostFunction) it.next();
                 if (cf.getName().equals(name)) {
                     return cf;
                 }
             }
         }
         return null;
     }

    /**
     * Returns a copy of the cost functions.
     * The list does not contain SCSP_FUNCTION and MAX_FUNCTION.
     */
    public java.util.List getAdditionalCostFunctions(){
        java.util.List result = new LinkedList();
        synchronized(costFunctions) {
            for (int i = 0; i < costFunctionNumber; i++) {
                CostFunction costFunction =
                    (CostFunction) costFunctions.get(i);
                if ( (costFunction != SCSP_FUNCTION)
                     && (costFunction != MAX_FUNCTION) ) {
                    result.add(costFunction);
                }
            }
        }
        return result;
    }

    /**
     * Setting a cost function for following constraint creation. Use
     * "_internal" version if called without user interaction.
     */
    public void selectCostFunction(CostFunction costFunction) {
        if ((costFunction != null) && (costFunction.gsc != this)) {
            throw new Error("The cost function \"" + costFunction +
                            "\" is not part of the solver");
        }
        if ((costFunction == null) && (DEFAULT_FUNCTION != null)) {
            throw new Error("No cost function given");
        }
        selectCostFunction_unsafe(costFunction);
    }

    /**
     * Version of selectCostFunction for internal use.
     * @see selectCostFunction(CostFunction)
     */
    public void selectCostFunction_internal(CostFunction costFunction) {
        if (Registry.SAFE_MODE) {
            selectCostFunction(costFunction);
        } else {
            selectCostFunction_unsafe(costFunction);
        }
    }

    /**
     * Cost function selection without consistency checks.
     * @see selectCostFunction(CostFunction)
     */
    private void selectCostFunction_unsafe(CostFunction costFunction) {
        currentCostFunction = costFunction;
        if (useCostMappingOfCostFunction) {
            if (costFunction != null) {
                currentCostMapping = currentCostFunction.costMapping;
            } else {
                currentCostMapping = defaultCostMapping;
            }
        }
    }

    /**
     * Returns the default cost function.
     */
    public CostFunction getDefaultCostFunction() {
        return DEFAULT_FUNCTION;
    }

    /**
     * Returns the cost function for structural costs.
     */
    public CostFunction getStructuralCostFunction() {
        return SCSP_FUNCTION;
    }

    /**
     * Returns the cost function for maximization.
     */
    public CostFunction getMaximizationCostFunction() {
        return MAX_FUNCTION;
    }

    /**
     * Returns the descriptions of all cost functions (names and
     * cost function value).
     */
    public StringBuffer getCostFunctionInfo() {
        StringBuffer s = new StringBuffer();
        int i;
        CostFunction costFunction;
        synchronized(costFunctions) {
            for (i = 0; i < costFunctionNumber; i++) {
                costFunction = (CostFunction) costFunctions.get(i);
                s.append(costFunction.toString() + "\n");
            }
        }
        return s;
    }

    /**
     * Adds further costs of a conventional constraint to the cost functions.
     * Use "_internal" version if called without user interaction.
     * @param constraint  The constraint whose costs are to be used.
     * @param costID  The ID of the constraint's costs to be used.
     * @param costMappingOptions  Arbitrary object to pass information to the
     *                            costMapping.
     */
    public void addCosts(ConventionalConstraint constraint, int costID,
                         Object costMappingOptions) {
        verifyCostAddition(constraint, costID);
        addCosts_unsafe(constraint, costID, costMappingOptions);
    }

    /**
     * Shortcut function for cost addition using no specific cost mapping
     * options.
     * @see addCosts(ConventionalConstraint, int, Object)
     */
    public void addCosts(ConventionalConstraint constraint, int costID) {
        addCosts(constraint, costID, null);
    }

    /**
     * Verifies that a cost addition is correct.
     * @param constraint  The constraint whose costs are to be used.
     * @param costID  The ID of the constraint's costs to be used.
     * @param costFunction  The cost function to which the costs are
     *                      to be assigned to.
     */
    private void verifyCostAddition(ConventionalConstraint constraint,
                                    int costID) {

        // Type-related tests

        if ((constraint instanceof ObjectConstraint)
            || (constraint instanceof IntegerConstraint)
            || (constraint instanceof RangeConstraint)) {
            throw new Error("A cost use is not applicable for this type of constraint");
        }

        // Testing whether the given constraint is valid and a part of the CSP

        synchronized(constraints) {
            if (!constraints.contains(constraint)) {
                throw new Error("The constraint \"" + constraint
                                + "\" is not part of the CSP");
            }
        }

        // Cost-related tests

        if ((costID < 0) || (costID >= constraint.getNumberOfCosts())) {
            throw new Error("Invalid costID");
        }

        if (constraint.costFunctionReference[costID] != null) {
            throw new Error("The costID \"" + costID
                            + "\" has already been used");
        }

    }

    /**
     * Cost addition for internal use.
     * @see addCosts(ConventionalConstraint, int, Object)
     */
    public void addCosts_internal(ConventionalConstraint constraint, int costID, Object costMappingOptions) {
        if (Registry.SAFE_MODE) {
            addCosts(constraint, costID, costMappingOptions);
        } else {
            addCosts_unsafe(constraint, costID, costMappingOptions);
        }
    }

    /**
     * Shortcut function for internal cost addition using no specific cost
     * mapping options.
     * @see addCosts_internal(ConventionalConstraint, int, Object)
     */
    public void addCosts_internal(ConventionalConstraint constraint,
                                  int costID) {
        addCosts_internal(constraint, costID, null);
    }

    /**
     * Cost addition without consistency checks.
     * @see addCosts(ConventionalConstraint, int, Object)
     */
    private void addCosts_unsafe(ConventionalConstraint constraint, int costID, Object costMappingOptions) {

        // Creating default cost function if needed

        if (currentCostFunction == null) {
            DEFAULT_FUNCTION = createCostFunction("Default Cost Function");
            currentCostFunction = DEFAULT_FUNCTION;
        }

        // Adding constraint costs' handling to cost function

        currentCostFunction.addConstraint(constraint, costID,
                                          currentCostMapping,
                                          costMappingOptions);

        // Updating graphical component if existent

        if (constraint.graphObject instanceof NodeGraphics) {
            ((ConventionalConstraintGraphics) constraint.graphObject).updateCosts();
            ((NodeGraphics) constraint.graphObject).updateSize();
        }

    }

    /**
     * Updates the information on the unsatisfied cost function with
     * highest priority.
     * @param costFunction  The cost function whose value has been changed.
     */
    public void registerCostFunctionUpdate(CostFunction costFunction) {

        // Pushing lastUnsatisfiedCostFunction back if necessary

        int oldLevel = -1;
        if (graphObject != null) {
            if (lastUnsatisfiedCostFunction != null) {
                oldLevel = lastUnsatisfiedCostFunction.level;
            }
        }

        if ((lastUnsatisfiedCostFunction == null) ||
            (costFunction.level > lastUnsatisfiedCostFunction.level)) {
            lastUnsatisfiedCostFunction = costFunction;
        }

        // Pulling lastUnsatisfiedCostFunction forward as much as possible

        int currentLevel = lastUnsatisfiedCostFunction.level;

        synchronized(costFunctions) {

            if ((lastUnsatisfiedCostFunction == SCSP_FUNCTION) && (lastUnsatisfiedCostFunction.objectiveCosts.equals(BigInteger.ZERO))) {
                currentLevel = costFunctionNumber - 2;
                lastUnsatisfiedCostFunction =
                    (CostFunction) costFunctions.get(currentLevel);
            }

            while ((currentLevel != 0) && (lastUnsatisfiedCostFunction.objectiveCosts.equals(BigInteger.ZERO))) {
                currentLevel--;
                lastUnsatisfiedCostFunction =
                    (CostFunction) costFunctions.get(currentLevel);
            }

        }

        // Updating graphics

        if ((graphObject != null) && (costFunction.level >= currentLevel)) {

            // Updating single constraints' cost graphics

            if (oldLevel != -1) {
                graphObject.updateConstraintCosts(oldLevel);
            }
            graphObject.updateConstraintCosts(currentLevel);

            // Updating global cost graphics

            if ((currentLevel == 0) && (lastUnsatisfiedCostFunction.objectiveCosts.equals(BigInteger.ZERO))) {
                graphObject.updateGlobalCosts(null);
            } else {
                graphObject.updateGlobalCosts(lastUnsatisfiedCostFunction);
            }

        }

    }


    // ***** IMPROVEMENT / SEARCH *****

    /**
     * Enables the constraint graph to be changed by improvement heuristics.
     */
    public void enableStructuralChanges() {
        structuralChangesEnabled = true;
        scm.switchOnRepair();
    }

    /**
     * Disables the constraint graph to be changed by improvement heuristics.
     */
    public void disableStructuralChanges() {
        structuralChangesEnabled = false;
        scm.switchOffRepair();
    }

    public boolean executeImprovementIteration() {

        // Updating structural evaluation

        scm.checkStructuralConstraints();

        // Checking if improvement is possible

        if (lastUnsatisfiedCostFunction == null) {
            // Already no costs because there is no cost function
            return false;
        }
        if ((lastUnsatisfiedCostFunction == SCSP_FUNCTION)
            && !structuralChangesEnabled) {
            // Structural inconsistency but not allowed be be repaired
            return false;
        }
        if (lastUnsatisfiedCostFunction.costFunctionReferences.numberOfElements == 0) {
            // No constraints have been assigned to the maximization function
            return false;
        }

        // Executing improvement step

        lastUnsatisfiedCostFunction.improveCosts();
        //
        //scm.checkStructuralConstraints();
        return true;
    }

    /**
     * Executes one improvement/repair iteration. Returns true if an
     * improvement was executed and false if all constraints of the given
     * cost function and of cost functions of higher priority are already
     * satisfied. Also returns false if structural inconsistency occured
     * but is not allowed be be repaired.
     * @param costFunction  The cost function of lowest priority to be
     *                      optimized.
     */
    public boolean executeImprovementIteration(CostFunction costFunction) {

        // Updating structural evaluation

        scm.checkStructuralConstraints();

        // Checking if improvement is possible

        if (lastUnsatisfiedCostFunction == null) {
            // Already no costs because there is no cost function
            return false;
        }
        if ((lastUnsatisfiedCostFunction == SCSP_FUNCTION)
            && !structuralChangesEnabled) {
            // Structural inconsistency but not allowed be be repaired
            return false;
        }
        if (lastUnsatisfiedCostFunction.costFunctionReferences.numberOfElements == 0) {
            // No constraints have been assigned to the maximization function
            return false;
        }
        if (lastUnsatisfiedCostFunction.level < costFunction.level) {
            // Given cost function already satisfied
            return false;
        }

        // Executing improvement step

        lastUnsatisfiedCostFunction.improveCosts();

        return true;
    }



}
