
package com.ai_center.dragonbreath;

/** 
 * Graphical interface for the global search control. 
 *
 * @author Alexander Nareyek
 * @version $Revision: 2.8 $
 * 
 * @see dragonbreath
 */


import java.awt.*;
import java.awt.event.*;
import java.io.File;
import java.lang.Math;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.*;
import javax.swing.*;
import javax.swing.filechooser.FileFilter;

import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.constraints.IntegerConstraint;
import com.ai_center.dragonbreath.constraints.RangeConstraint;
import com.ai_center.dragonbreath.constraints.ProportionConstraint;
import com.ai_center.dragonbreath.constraints.ExponentialConstraint;
import com.ai_center.dragonbreath.costs.*;
import com.ai_center.dragonbreath.utilities.ExtensionFileFilter;
import com.ai_center.dragonbreath.utilities.ListDialog;
import com.ai_center.dragonbreath.utilities.OrderedShiftDLLElement;
import com.ai_center.dragonbreath.structuralConstraintHandling.StructuralConstraintDisplayer;


public class GlobalSearchControlGraphics extends JFrame {

    // **********************
    // ***** ATTRIBUTES *****
    // **********************

    /** 
     * The Global Search Control reference.
     */
    private final GlobalSearchControl gsc;

    /**
     * Current cost mapping options that were set.
     */
    protected Object[] currentCostMappingOptions = null;

    /** 
     * Label for displaying the current cost function value.
     */
    private JLabel costsLabel;
 
    /** 
     * Tool buttons for CSP interaction.
     */
    private JButton[] toolButton = new JButton[9];

    /** 
     * Tool button constant representing that no interaction takes place.
     */
    private final int TOOLBUTTON_NONE = -1;
    /** 
     * Tool button constant representing the addition of variables.
     */
    private final int TOOLBUTTON_ADD_VARIABLE = 0;
    /** 
     * Tool button constant representing the addition of object constraints.
     */
    private final int TOOLBUTTON_ADD_OCONSTRAINT = 1;
    /** 
     * Tool button constant representing the addition of constraints.
     */
    private final int TOOLBUTTON_ADD_CONSTRAINT = 2;
    /** 
     * Tool button constant representing the addition of costs.
     */
    private final int TOOLBUTTON_ADD_COSTS = 3;
    /** 
     * Tool button constant representing the extension of constraints.
     */
    private final int TOOLBUTTON_EXTEND_CONSTRAINT = 4;
    /** 
     * Tool button constant representing the manipulation of nodes.
     */
    private final int TOOLBUTTON_EDIT_NODE = 5;
    /** 
     * Tool button constant representing the movement of nodes.
     */
    private final int TOOLBUTTON_MOVE_NODE = 6;
    /** 
     * Tool button constant representing the reduction of a constraint.
     */
    private final int TOOLBUTTON_REDUCE_CONSTRAINT = 7;
    /** 
     * Tool button constant representing the deletion of nodes.
     */
    private final int TOOLBUTTON_DELETE_NODE = 8;

        
    /** 
     * Contains variables that describe the current interaction state.
     */
    private class InteractionState {

	// Attributes

	/** 
	 * Describes the current user interaction mode.
	 * @see TOOLBUTTON_NONE
	 * @see TOOLBUTTON_ADD_VARIABLE
	 * @see TOOLBUTTON_ADD_OCONSTRAINT
	 * @see TOOLBUTTON_ADD_CONSTRAINT
	 * @see TOOLBUTTON_ADD_COSTS
	 * @see TOOLBUTTON_EXTEND_CONSTRAINT
	 * @see TOOLBUTTON_EDIT_NODE
	 * @see TOOLBUTTON_MOVE_NODE
	 * @see TOOLBUTTON_REDUCE_CONSTRAINT
	 * @see TOOLBUTTON_DELETE_NODE
	 */
	private int mode = TOOLBUTTON_NONE;
	/** 
	 * Counter for the step messages.
	 */
	private int stepCount = 0;
	/** 
	 * Currently selected interaction node of the CSP.
	 */
	private Object node = null;
	/** 
	 * Stores details about the current interaction mode.
	 */
	private int details = 0;
	/** 
	 * Stores a chosen extension ID.
	 */
	private int extension = 0;
	/** 
	 * Stores embedding graph elements.
	 */
	private Embedding embedding = null;
	/** 
	 * Stores X coordinate.
	 */
	private int x = 0;
	/** 
	 * Stores Y coordinate.
	 */
	private int y = 0;

	// Methods

	/** 
	 * Resets interaction-dependent values
	 */
	protected void reset() {
	    stepCount = 0;
	    node = null;
	    details = 0;
	    extension = 0;
	    embedding = null;
	    x = 0;
	    y = 0;
	}

    }

    /** 
     * The interaction state object.
     */
    private InteractionState interaction = new InteractionState();

    /** 
     * Panel for CSP display.
     */
    private CSPDisplay cspPanel;
    /** 
     * A queue of messages to be displayed.
     * @see messageLabel
     */
    private LinkedList messageQueue = new LinkedList();
    /** 
     * Stores whether the message label is highlighted.
     * @see messageLabel
     */
    private boolean messageHighlighted;
    /** 
     * Label for displaying messages.
     */
    private JLabel messageLabel;
    /** 
     * Panel of the message display.
     * @see messageLabel
     */
    private JPanel messagePanel;
    /** 
     * Timer for message blinking.
     * @see messageLabel
     */
    private javax.swing.Timer timer = null;
    // Message constants
    private final Message MESSAGE_OPERATION_ABORTED = new Message("OPERATION ABORTED!", Color.red);

    private final Message MESSAGE_LOAD_SCP_DONE = new Message("File successfully loaded!", Color.green);
    private final Message MESSAGE_SAVE_SCP_DONE = new Message("File successfully saved!", Color.green);

    private final Message MESSAGE_LOAD_SC_DONE = new Message("Structural constraint successfully loaded!", Color.green);
    private final Message MESSAGE_LOAD_SC_ERROR = new Message("Structural constraint could not be loaded!", Color.red);
    
    private final Message MESSAGE_SELECT_STRUCTURAL_CONSTRAINT = new Message("Select a Structural Constraint to display Instances.", Color.yellow);
    private final Message MESSAGE_NO_CONSTRAINT_LOADED = new Message("No Structural Constraint loaded.", Color.red);
    
    private final Message MESSAGE_SELECT_COSTFUNCTION_1 = new Message("SELECTING COST FUNCTION      Please select the cost function.", Color.yellow);
    private final Message MESSAGE_SELECT_COSTFUNCTION_DONE = new Message("Cost function successfully selected!", Color.green);

    private final Message MESSAGE_SELECT_COSTMAPPING_1 = new Message("SELECTING COST MAPPING      Please select the cost mapping.", Color.yellow);
    private final Message MESSAGE_SELECT_COSTMAPPING_DONE = new Message("Cost mapping successfully selected!", Color.green);

    private final Message MESSAGE_SET_COSTMAPPINGOPTIONS_FORMAT = new Message("WRONG VALUE FORMAT!", Color.red);
    private final Message MESSAGE_SET_COSTMAPPINGOPTIONS_DONE = new Message("Cost mapping options successfully set!", Color.green);

    private final Message MESSAGE_CREATE_COSTFUNCTION_1 = new Message("CREATING COST FUNCTION      Please input the new cost function's name.", Color.yellow);
    private final Message MESSAGE_CREATE_COSTFUNCTION_DONE = new Message("Cost function successfully created!", Color.green);

    private final Message MESSAGE_CREATE_COSTMAPPING_1 = new Message("CREATING COST MAPPING      Step 1 - Please select the cost mapping class.", Color.yellow);
    private final Message MESSAGE_CREATE_COSTMAPPING_2 = new Message("CREATING COST MAPPING      Step 2 - Please input the new cost mapping's name.", Color.yellow);
    private final Message MESSAGE_CREATE_COSTMAPPING_DONE = new Message("Cost mapping successfully created!", Color.green);
    private final Message MESSAGE_ENABLE_STRUCTURAL_CHANGES_DONE = new Message("Improvement heuristics were enabled to change the graph!", Color.green);
    private final Message MESSAGE_DISABLE_STRUCTURAL_CHANGES_DONE = new Message("Improvement heuristics were disabled to change the graph!", Color.green);

    private final Message MESSAGE_CHOOSE_OPERATION = new Message("Please choose an operation from the sidebar.", Color.yellow);

    private final Message MESSAGE_ADD_VARIABLE_1 = new Message("ADDING VARIABLES      If you want to add a new variable, please select a location for it.", Color.yellow);
    private final Message MESSAGE_ADD_VARIABLE_2 = new Message("ADDING VARIABLES      Step 2 - Please input the new variable's name.", Color.yellow);
    private final Message MESSAGE_ADD_VARIABLE_DONE = new Message("Variable successfully added!", Color.green);

    private final Message MESSAGE_ADD_OCONSTRAINT_1 = new Message("ADDING OBJECT CONSTRAINTS      If you want to add a new object constraint, please select a location for it.", Color.yellow);
    private final Message MESSAGE_ADD_OCONSTRAINT_2 = new Message("ADDING OBJECT CONSTRAINTS      Step 2 - Please select the object constraint's type.", Color.yellow);
    private final Message MESSAGE_ADD_OCONSTRAINT_DONE = new Message("Object constraint successfully added!", Color.green);

    private final Message MESSAGE_ADD_CONSTRAINT_1 = new Message("ADDING CONSTRAINTS      If you want to add a new constraint, please select a location for it.", Color.yellow);
    private final Message MESSAGE_ADD_CONSTRAINT_2 = new Message("ADDING CONSTRAINTS      Step 2 - Please select the constraint's type.", Color.yellow);
    private final Message MESSAGE_ADD_CONSTRAINT_DONE = new Message("Constraint successfully added!", Color.green);

    private final Message MESSAGE_ADD_COSTS_1 = new Message("ADDING COSTS      If you want to add additional costs of a constraint, please press its top button.", Color.yellow);
    private final Message MESSAGE_ADD_COSTS_2 = new Message("ADDING COSTS      Step 2 - Please select the cost type to be used.", Color.yellow);
    private final Message MESSAGE_ADD_COSTS_DONE = new Message("Costs successfully added!", Color.green);

    private final Message MESSAGE_EXTEND_CONSTRAINT_1 = new Message("EXTENDING CONSTRAINTS      If you want to extend a constraint, please press its top button.", Color.yellow);
    private final Message MESSAGE_EXTEND_CONSTRAINT_2 = new Message("EXTENDING CONSTRAINTS      Step 2 - Please select the extension.", Color.yellow);
    private final Message MESSAGE_EXTEND_CONSTRAINT_DONE = new Message("Constraint successfully extended!", Color.green);

    private final Message MESSAGE_EDIT_NODE_1 = new Message("EDITING NODES      If you want to edit a node, please press its top button.", Color.yellow);
    private final Message MESSAGE_EDIT_NODE_2 = new Message("EDITING NODES      Step 2 - Please input the variable's new value.", Color.yellow);
    private final Message MESSAGE_EDIT_NODE_2R = new Message("EDITING NODES      Step 2 - Please edit the constraint's parameters.", Color.yellow);
    private final Message MESSAGE_EDIT_NODE_DONE_PARTIAL = new Message("Node partially adapted!", Color.green);
    private final Message MESSAGE_EDIT_NODE_DONE = new Message("Node successfully adapted!", Color.green);

    private final Message MESSAGE_MOVE_NODE_1 = new Message("MOVING NODES      If you want to move a node, please press its top button.", Color.yellow);
    private final Message MESSAGE_MOVE_NODE_2 = new Message("MOVING NODES      Step 2 - Please select a new location for the node.", Color.yellow);
    private final Message MESSAGE_MOVE_NODE_DONE = new Message("Node successfully moved!", Color.green);

    private final Message MESSAGE_REDUCE_CONSTRAINT_1 = new Message("REDUCING CONSTRAINTS      If you want to reduce a constraint, please press its top button.", Color.yellow);
    private final Message MESSAGE_REDUCE_CONSTRAINT_2 = new Message("REDUCING CONSTRAINTS      Step 2 - Please select the reduction.", Color.yellow);
    private final Message MESSAGE_REDUCE_CONSTRAINT_DONE = new Message("Constraint successfully reduced!", Color.green);

    private final Message MESSAGE_DELETE_NODE_1 = new Message("DELETING NODES      If you want to delete a node, please press its top button.", Color.yellow);
    private final Message MESSAGE_DELETE_NODE_DONE = new Message("Node successfully deleted!", Color.green);

    private final Message MESSAGE_CLEAR_CSP_DONE = new Message("CSP elements successfully removed!", Color.green);

    /** 
     * Message that will be displayed to indicate the current 
     * interaction mode.
     * @see messageLabel
     */
    private Message defaultMessage = null;

    /** 
     * Main user interaction button.
     */
    private JButton mainButton;

    /**
     * For communication between solveButton and stopSolveButton.
     */
    private boolean solve = false;
    
    /** 
     * User interaction button for solving Problem.
     */
    private JButton solveButton;

    /** 
     * User interaction button to stop solving Problem.
     */
    private JButton stopSolveButton;
    
    /**
     * Timer used to execute the SolveTask.
     */
    private java.util.Timer utilTimer = new java.util.Timer();
    
    
    class SolveTask extends java.util.TimerTask {

        GlobalSearchControl myGSC;
        boolean running =false;

        /**
         * Constructor
         */
        protected SolveTask(GlobalSearchControl gsc) {
            myGSC = gsc;
        }
        
        /** 
         * Calls once the gsc.executeImprovementIteration()
         * @see     java.lang.Thread#run()
         */
        public void run() {
            running = true;
            synchronized (gsc) {
                if (! gsc.executeImprovementIteration() ) {
		    cancel();
                }
            }
        }
        
        /**
         * Is the task currently scheduled?
         */
        public boolean running() {
            return running;
        }
        
        public boolean cancel() {
            running = false;
            return super.cancel();
        }
        
    }

    /**
     * The Task that is called for subsequent gsc.executeImprovementIteration()
     */
    SolveTask solveTask;

    
    // ************************
    // ***** CONSTRUCTORS *****
    // ************************

    /** 
     * Creates basic graphical objects.
     * @param gsc_ref  The related GlobalSearchControl object.
     */
    protected GlobalSearchControlGraphics(GlobalSearchControl gsc_ref) {

	super("DragonBreath Engine (Version " + Registry.VERSION + ")");

	gsc = gsc_ref;
        solveTask = new SolveTask(gsc);
	JMenuBar menuBar;
	JMenu menu;
	JMenuItem menuItem;
	JPanel panel;
	JPanel panel2;
	JLabel label;
	JToolBar toolBar;
	JScrollPane scrollPane;
	ImageIcon i;

	// Top-level menu

	menuBar = new JMenuBar();
	setJMenuBar(menuBar);

	// Top-level menu: File

	menu = new JMenu("File");
	menu.setMnemonic(KeyEvent.VK_F);
	menuBar.add(menu);
        
        menuItem = new JMenuItem("Load", KeyEvent.VK_L);
	menuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_L, 
						       ActionEvent.ALT_MASK));
        menuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
                    JFileChooser fc = new JFileChooser(".");
                    
                    ExtensionFileFilter SCPfilter = new ExtensionFileFilter();
                    SCPfilter.addExtension("scp", false);
                    SCPfilter.setDescription("Structural Constraint Problem (*.scp)");
                    
                    fc.addChoosableFileFilter(SCPfilter);
                    
                    fc.setDialogTitle("Load Structural Constraint Problem");
                    
                    fc.setApproveButtonText("Load");
                    
                    int result = fc.showOpenDialog(new JFrame());
                    if (result == JFileChooser.APPROVE_OPTION) {
                        try {
                            File f = fc.getSelectedFile();
                            if (f != null) {
				gsc.loadSCP(f);
				MESSAGE_LOAD_SCP_DONE.activate();
                            }
                        }//try
                        catch (Exception exx) {
                            exx.printStackTrace();
                        }        

                    }//if
                }//actionPerformed
        });
	menu.add(menuItem);
        
        menuItem = new JMenuItem("Save", KeyEvent.VK_S);
	menuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_S, 
						       ActionEvent.ALT_MASK));
        menuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
                    JFileChooser fc = new JFileChooser(".");
		    
                    ExtensionFileFilter SCPfilter=new ExtensionFileFilter();
                    SCPfilter.addExtension("scp", false);
                    SCPfilter.setDescription("Structural Constraint Problem (*.scp)");
                    fc.addChoosableFileFilter(SCPfilter);
                    
                    fc.setDialogTitle("Save Structural Constraint Problem");
                    int result = fc.showSaveDialog(new JFrame());
                    if (result == JFileChooser.APPROVE_OPTION) {
                        try {
                            File f = fc.getSelectedFile();
                            if (f != null) {
                                String fileName = fc.getName(f);
				
                                if (!fc.accept(f)) {
				    // File wouldn't pass the filter: 
				    // wrong suffix

                                    FileFilter ff = fc.getFileFilter();
                                    if (ff == SCPfilter){
                                        fileName += ".scp";
                                    }
                                    f = new File(fc.getCurrentDirectory(), 
						 fileName);
                                }
                                gsc.saveSCP(f);
				MESSAGE_SAVE_SCP_DONE.activate();
                            }
                        }//try
                        catch (Exception exx) {
                            exx.printStackTrace();
                        }        
			
                    }//if
		}//actionPerformed	
	    });
	menu.add(menuItem);

        menu.addSeparator();

	menuItem = new JMenuItem("Clear", KeyEvent.VK_C);
	menuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_C, 
						       ActionEvent.ALT_MASK));
        menuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    clearCSP();
		}});
        menu.add(menuItem);
	menu.addSeparator();
	menuItem = new JMenuItem("Close", KeyEvent.VK_C);
	menuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Q, 
						       ActionEvent.ALT_MASK));
        menuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    gsc.deleteGraphics();
		}});
	menu.add(menuItem);
	menuItem = new JMenuItem("Exit", KeyEvent.VK_X);
	menuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_W, 
						       ActionEvent.ALT_MASK));
        menuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    System.exit(0);
		}});
	menu.add(menuItem);

	// Top-level menu: Costs

	menu = new JMenu("Costs");
	menu.setMnemonic(KeyEvent.VK_C);
	menuBar.add(menu);
        
        menuItem = new JMenuItem("Select Cost Function", KeyEvent.VK_F);
	menuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_F, 
						       ActionEvent.CTRL_MASK));
        menuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {

		    Message previousDefaultMessage = defaultMessage;
		    defaultMessage = MESSAGE_SELECT_COSTFUNCTION_1;
		    MESSAGE_SELECT_COSTFUNCTION_1.activate();

		    synchronized(gsc.costFunctions) {
			
			// Retrieving cost function names

			int cfNum = gsc.costFunctions.size();
			cfNum--;   // Excluding structural cost function
			String[] cfNames = new String[cfNum];

			Iterator it = gsc.costFunctions.listIterator(0);
			for (int i = 0; i < cfNum; i++) {
			    cfNames[i] = ((CostFunction) it.next()).getName();
			}
		    
			// Cost function selection

			ListDialog l = new ListDialog(cfNames, null, "Please select the cost function");
			int selectedCostFunction = l.select();
			if (selectedCostFunction != -1) {
			    gsc.selectCostFunction((CostFunction) gsc.costFunctions.get(selectedCostFunction));
			    MESSAGE_SELECT_COSTFUNCTION_DONE.activate();
			} else {
			    MESSAGE_OPERATION_ABORTED.activate();
			}

		    }

		    defaultMessage = previousDefaultMessage;

  		}});
	menu.add(menuItem);

        menuItem = new JMenuItem("Select Cost Mapping", KeyEvent.VK_M);
	menuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_M, 
						       ActionEvent.CTRL_MASK));
        menuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {

		    Message previousDefaultMessage = defaultMessage;
		    defaultMessage = MESSAGE_SELECT_COSTMAPPING_1;
		    MESSAGE_SELECT_COSTMAPPING_1.activate();

		    synchronized(gsc.costMappings) {
			
			// Retrieving cost mapping names

			int cmNum = gsc.costMappings.size();
			String[] cmNames = new String[cmNum];

			Iterator it = gsc.costMappings.iterator();
			for (int i = 0; i < cmNum; i++) {
			    cmNames[i] = ((CostMapping) it.next()).getName();
			}
			
			// Cost mapping selection

			ListDialog l = new ListDialog(cmNames, null, "Please select the cost mapping");
			int selectedCostMapping = l.select();
			if (selectedCostMapping != -1) {
			    gsc.selectCostMapping((CostMapping) gsc.costMappings.get(selectedCostMapping));
			    currentCostMappingOptions = null;
			    MESSAGE_SELECT_COSTMAPPING_DONE.activate();
			} else {
			    MESSAGE_OPERATION_ABORTED.activate();
			}

		    }

		    defaultMessage = previousDefaultMessage;

  		}});
	menu.add(menuItem);

        menuItem = new JMenuItem("Set Cost Mapping Options", KeyEvent.VK_O);
	menuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_O, 
						       ActionEvent.CTRL_MASK));
        menuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {

		    Message previousDefaultMessage = defaultMessage;
		    int[] optionType = gsc.currentCostMapping.getOptionTypes();
		    Object[] newCostMappingOptions = 
			new Object[optionType.length];
		    boolean canceled = false;

		    for (int i = 0; (i < optionType.length) && !canceled; 
			 i++) {

			// Requesting option value

			Message MESSAGE_SET_COSTMAPPINGOPTIONS_N = new Message("SETTING COST MAPPING OPTIONS      Step " + (i + 1) + " - Please input the  " + gsc.currentCostMapping.getOptionDescription(i) + "  option value for the cost mapping.", Color.yellow);
			MESSAGE_SET_COSTMAPPINGOPTIONS_N.activate();
			defaultMessage = MESSAGE_SET_COSTMAPPINGOPTIONS_N;

			String value = null;//JOptionPane.showInputDialog("Please input the  " + gsc.currentCostMapping.getOptionDescription(i) + "  option value ( " + Registry.typeDescription[optionType[i]] + " )", gsc.currentCostMapping.getOptionDefaultValue(i));

			// Testing and setting option value

			if (value != null) {
			    switch (optionType[i]) {
			    case Registry.TYPE_POSITIVE_BIGINTEGER: {
				boolean ok = true;
				BigInteger number = null;
				try {
				    number = new BigInteger(value);
				} catch (NumberFormatException nfe) {
				    ok = false;
				}
				if (!ok || (number.compareTo(BigInteger.ZERO) == -1)) {
				    canceled = true;
				    MESSAGE_SET_COSTMAPPINGOPTIONS_FORMAT.activate();
				} else {
				    newCostMappingOptions[i] = number;
				}
				break;
			    }
			    case Registry.TYPE_POSITIVE_BIGDECIMAL: {
				boolean ok = true;
				BigDecimal number = null;
				try {
				    number = new BigDecimal(value);
				} catch (NumberFormatException nfe) {
				    ok = false;
				}
				if (!ok || (number.compareTo(new BigDecimal(0.0)) == -1)) {
				    canceled = true;
				    MESSAGE_SET_COSTMAPPINGOPTIONS_FORMAT.activate();
				} else {
				    newCostMappingOptions[i] = number;
				}
				break;
			    }
			    default: {
				throw new Error("Option type not implemented");
			    }
			    }
			} else {
			    canceled = true;
			    MESSAGE_OPERATION_ABORTED.activate();
			}

		    }

		    if (!canceled) {
			currentCostMappingOptions = newCostMappingOptions;
			MESSAGE_SET_COSTMAPPINGOPTIONS_DONE.activate();
		    }

		    defaultMessage = previousDefaultMessage;

  		}});
	menu.add(menuItem);

        menu.addSeparator();

        menuItem = new JMenuItem("Create Cost Function", KeyEvent.VK_F);
	menuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_F, ActionEvent.SHIFT_MASK + ActionEvent.CTRL_MASK));
        menuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {

		    Message previousDefaultMessage = defaultMessage;
		    defaultMessage = MESSAGE_CREATE_COSTFUNCTION_1;
		    MESSAGE_CREATE_COSTFUNCTION_1.activate();

		    String value =  null;//JOptionPane.showInputDialog("Please input the cost function's name", "Cost Function");
		    if (value != null) {
			gsc.createCostFunction(value);
			MESSAGE_CREATE_COSTFUNCTION_DONE.activate();
		    } else {
			MESSAGE_OPERATION_ABORTED.activate();
		    }

		    defaultMessage = previousDefaultMessage;

  		}});
        menu.add(menuItem);

        menuItem = new JMenuItem("Create Cost Mapping", KeyEvent.VK_M);
	menuItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_M, ActionEvent.SHIFT_MASK + ActionEvent.CTRL_MASK));
        menuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {

		    Message previousDefaultMessage = defaultMessage;
		    defaultMessage = MESSAGE_CREATE_COSTMAPPING_1;
		    MESSAGE_CREATE_COSTMAPPING_1.activate();
		    
		    // Class selection

		    int cmNum = Registry.getCostMappingNumber();
		    String[] cmNames = new String[cmNum];
		    for (int i = 0; i < cmNum; i++) {
			cmNames[i] = Registry.costMappingName[i];
		    }
		    
		    ListDialog l = new ListDialog(cmNames, null, "Please select the cost mapping class");
		    int selectedCostMapping = l.select();
		    if (selectedCostMapping != -1) {

			// Name input selection

			defaultMessage = MESSAGE_CREATE_COSTMAPPING_2;
		        MESSAGE_CREATE_COSTMAPPING_2.activate();

			String value =  null;//JOptionPane.showInputDialog("Please input the cost mapping's name", Registry.costMappingName[selectedCostMapping]);
			if (value != null) {
			    CostMapping cm = gsc.createCostMapping(selectedCostMapping, value);
			    MESSAGE_CREATE_COSTMAPPING_DONE.activate();
			} else {
			    MESSAGE_OPERATION_ABORTED.activate();
			}

		    } else {
			MESSAGE_OPERATION_ABORTED.activate();
		    }
		    
		    defaultMessage = previousDefaultMessage;

  		}});
        menu.add(menuItem);

	// Top-level menu: Graph Structure

	menu = new JMenu("Graph Structure");
	menu.setMnemonic(KeyEvent.VK_G);
	menuBar.add(menu);
        
        menuItem = new JMenuItem("Enable Structural Changes", KeyEvent.VK_E);
        menuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    gsc.enableStructuralChanges();
		    MESSAGE_ENABLE_STRUCTURAL_CHANGES_DONE.activate();
  		}});
        menu.add(menuItem);

        menuItem = new JMenuItem("Disable Structural Changes", KeyEvent.VK_D);
        menuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    gsc.disableStructuralChanges();
		    MESSAGE_DISABLE_STRUCTURAL_CHANGES_DONE.activate();
  		}});
        menu.add(menuItem);

        menu.addSeparator();

        menuItem = new JMenuItem("Load Structural Constraint", KeyEvent.VK_L);
        menuItem.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent e) {
                    JFileChooser fc = new JFileChooser(".");
                    
                    ExtensionFileFilter SCPfilter = new ExtensionFileFilter();
                    SCPfilter.addExtension("sc", false);
                    SCPfilter.setDescription("Structural Constraint (*.sc)");
                    
                    fc.addChoosableFileFilter(SCPfilter);
                    
                    fc.setDialogTitle("Load Structural Constraint");
                    
                    fc.setApproveButtonText("Load");
                    
                    int result = fc.showOpenDialog(new JFrame());
                    if (result == JFileChooser.APPROVE_OPTION) {
                        try {
                            File f = fc.getSelectedFile();
                            if (f != null) {
                                if (gsc.loadStructuralConstraints(f)) {
                                    MESSAGE_LOAD_SC_DONE.activate();
                                } else {
                                    MESSAGE_LOAD_SC_ERROR.activate();
                                }
                            }
                        }//try
                        catch (Exception exx) {
                            exx.printStackTrace();
                        }        

                    }//if
                }//actionPerformed
        });
	menu.add(menuItem);
        
                 
        menuItem = new JMenuItem("Show Structural Constraints", KeyEvent.VK_S);
	
        menuItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
                       StructuralConstraintDisplayer.displayDialog(gsc.getStructuralConstraints());
                       
  		}});
	menu.add(menuItem);

	// Base panel

	panel = new JPanel();
	panel.setBorder(BorderFactory.createEmptyBorder(10, 30, 10, 30));
	panel.setLayout(new BoxLayout(panel, BoxLayout.Y_AXIS));

	// Costs display

	panel2 = new JPanel();
	panel2.setBorder(BorderFactory.createEmptyBorder(5, 0, 5, 0));
	panel2.setLayout(new FlowLayout());
	costsLabel = new JLabel("All Costs Resolved");
	panel2.add(costsLabel);
	panel.add(panel2, BorderLayout.CENTER);

	// CSP interaction area

	panel2 = new JPanel();
	panel2.setLayout(new BorderLayout());

	toolBar = new JToolBar();
	toolBar.setOrientation(JToolBar.VERTICAL);
	toolBar.setFloatable(false);

	i = new ImageIcon(this.getClass().getResource("images/tool-add-variables.gif"));
	toolButton[TOOLBUTTON_ADD_VARIABLE] = new JButton(i);
	toolButton[TOOLBUTTON_ADD_VARIABLE].setBackground(Color.gray);
	toolButton[TOOLBUTTON_ADD_VARIABLE].setToolTipText("Add variables");
        toolButton[TOOLBUTTON_ADD_VARIABLE].addActionListener(new 
	    ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    setInteractionMode(TOOLBUTTON_ADD_VARIABLE);
		}});
	toolBar.add(toolButton[TOOLBUTTON_ADD_VARIABLE]);

	i = new ImageIcon(this.getClass().getResource("images/tool-add-oconstraints.gif"));
	toolButton[TOOLBUTTON_ADD_OCONSTRAINT] = new JButton(i);
	toolButton[TOOLBUTTON_ADD_OCONSTRAINT].setBackground(Color.gray);
	toolButton[TOOLBUTTON_ADD_OCONSTRAINT].setToolTipText("Add object constraints");
        toolButton[TOOLBUTTON_ADD_OCONSTRAINT].addActionListener(new 
	    ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    setInteractionMode(TOOLBUTTON_ADD_OCONSTRAINT);
		}});
	toolBar.add(toolButton[TOOLBUTTON_ADD_OCONSTRAINT]);

	i = new ImageIcon(this.getClass().getResource("images/tool-add-constraints.gif"));
	toolButton[TOOLBUTTON_ADD_CONSTRAINT] = new JButton(i);
	toolButton[TOOLBUTTON_ADD_CONSTRAINT].setBackground(Color.gray);
	toolButton[TOOLBUTTON_ADD_CONSTRAINT].setToolTipText("Add constraints");
        toolButton[TOOLBUTTON_ADD_CONSTRAINT].addActionListener(new 
	    ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    setInteractionMode(TOOLBUTTON_ADD_CONSTRAINT);
		}});
	toolBar.add(toolButton[TOOLBUTTON_ADD_CONSTRAINT]);

	i = new ImageIcon(this.getClass().getResource("images/tool-add-costs.gif"));
	toolButton[TOOLBUTTON_ADD_COSTS] = new JButton(i);
	toolButton[TOOLBUTTON_ADD_COSTS].setBackground(Color.gray);
	toolButton[TOOLBUTTON_ADD_COSTS].setToolTipText("Add costs");
        toolButton[TOOLBUTTON_ADD_COSTS].addActionListener(new 
	    ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    setInteractionMode(TOOLBUTTON_ADD_COSTS);
		}});
	toolBar.add(toolButton[TOOLBUTTON_ADD_COSTS]);

	i = new ImageIcon(this.getClass().getResource("images/tool-extend-constraints.gif"));
	toolButton[TOOLBUTTON_EXTEND_CONSTRAINT] = new JButton(i);
	toolButton[TOOLBUTTON_EXTEND_CONSTRAINT].setBackground(Color.gray);
	toolButton[TOOLBUTTON_EXTEND_CONSTRAINT].setToolTipText("Extend constraints");
        toolButton[TOOLBUTTON_EXTEND_CONSTRAINT].addActionListener(new 
	    ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    setInteractionMode(TOOLBUTTON_EXTEND_CONSTRAINT);
		}});
	toolBar.add(toolButton[TOOLBUTTON_EXTEND_CONSTRAINT]);

	toolBar.add(Box.createRigidArea(new Dimension(0,10)));

	i = new ImageIcon(this.getClass().getResource("images/tool-move-nodes.gif"));
	toolButton[TOOLBUTTON_MOVE_NODE] = new JButton(i);
	toolButton[TOOLBUTTON_MOVE_NODE].setBackground(Color.gray);
	toolButton[TOOLBUTTON_MOVE_NODE].setToolTipText("Move nodes");
        toolButton[TOOLBUTTON_MOVE_NODE].addActionListener(new 
	    ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    setInteractionMode(TOOLBUTTON_MOVE_NODE);
		}});
	toolBar.add(toolButton[TOOLBUTTON_MOVE_NODE]);

	i = new ImageIcon(this.getClass().getResource("images/tool-edit.gif"));
	toolButton[TOOLBUTTON_EDIT_NODE] = new JButton(i);
	toolButton[TOOLBUTTON_EDIT_NODE].setBackground(Color.gray);
	toolButton[TOOLBUTTON_EDIT_NODE].setToolTipText("Edit nodes");
        toolButton[TOOLBUTTON_EDIT_NODE].addActionListener(new 
	    ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    setInteractionMode(TOOLBUTTON_EDIT_NODE);
		}});
	toolBar.add(toolButton[TOOLBUTTON_EDIT_NODE]);

	toolBar.add(Box.createRigidArea(new Dimension(0,10)));

	i = new ImageIcon(this.getClass().getResource("images/tool-reduce-constraints.gif"));
	toolButton[TOOLBUTTON_REDUCE_CONSTRAINT] = new JButton(i);
	toolButton[TOOLBUTTON_REDUCE_CONSTRAINT].setBackground(Color.gray);
	toolButton[TOOLBUTTON_REDUCE_CONSTRAINT].setToolTipText("Reduce constraints");
        toolButton[TOOLBUTTON_REDUCE_CONSTRAINT].addActionListener(new 
	    ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    setInteractionMode(TOOLBUTTON_REDUCE_CONSTRAINT);
		}});
	toolBar.add(toolButton[TOOLBUTTON_REDUCE_CONSTRAINT]);

	i = new ImageIcon(this.getClass().getResource("images/tool-delete-nodes.gif"));
	toolButton[TOOLBUTTON_DELETE_NODE] = new JButton(i);
	toolButton[TOOLBUTTON_DELETE_NODE].setBackground(Color.gray);
	toolButton[TOOLBUTTON_DELETE_NODE].setToolTipText("Delete nodes");
        toolButton[TOOLBUTTON_DELETE_NODE].addActionListener(new 
	    ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    setInteractionMode(TOOLBUTTON_DELETE_NODE);
		}});
	toolBar.add(toolButton[TOOLBUTTON_DELETE_NODE]);
        
	panel2.add(toolBar, BorderLayout.WEST);

	// CSP display area

	cspPanel = new CSPDisplay();
	cspPanel.setLayout(null);
	cspPanel.setBackground(Color.white);
	cspPanel.setPreferredSize(new Dimension(2000, 1500));
        cspPanel.addMouseListener(new MouseAdapter() {
		public void mouseClicked(MouseEvent e) {
		    cspAreaButtonPressed(e.getX(), e.getY());
		}});

	scrollPane = new JScrollPane(cspPanel, 
				     JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, 
				     JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
	scrollPane.setViewportBorder(BorderFactory.createLoweredBevelBorder());
	panel2.add(scrollPane, BorderLayout.CENTER);

	panel.add(panel2, BorderLayout.CENTER);

	// Message label

	panel2 = new JPanel();
	panel2.setBorder(BorderFactory.createEmptyBorder(10, 40, 3, 40));
	panel2.setLayout(new GridLayout(1, 1));
	messagePanel = new JPanel();
	messagePanel.setLayout(new FlowLayout());
	messagePanel.setBackground(Color.lightGray);
	messageLabel = new JLabel("Welcome to the DragonBreath Engine (Version " + Registry.VERSION + ")!", JLabel.CENTER);
	messageLabel.setForeground(Color.black);
	messageLabel.setBackground(Color.lightGray);
	messagePanel.add(messageLabel, BorderLayout.CENTER);
	panel2.add(messagePanel, BorderLayout.CENTER);
	panel.add(panel2, BorderLayout.CENTER);

	// User interaction buttons for starting/stopping search

	panel2 = new JPanel();
	panel2.setLayout(new FlowLayout());
	mainButton = new JButton("One Improvement Step");
	mainButton.setDefaultCapable(true);
	mainButton.setMnemonic(KeyEvent.VK_SPACE);
        mainButton.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    synchronized(GlobalSearchControlGraphics.this) { 
			gsc.executeImprovementIteration();
		    }
		}});
	panel2.add(mainButton, BorderLayout.WEST);
	panel.add(panel2, BorderLayout.CENTER);

	solveButton = new JButton("Optimize CSP");
	solveButton.setDefaultCapable(true);
	solveButton.setMnemonic(KeyEvent.VK_O);
        solveButton.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
                    if (! solveTask.running()) {
                        solveTask = new SolveTask(gsc);
                        utilTimer.schedule(solveTask, 0, 500);
                    }
		}});
	panel2.add(solveButton, BorderLayout.CENTER);

	stopSolveButton = new JButton("Stop Optimization");
	stopSolveButton.setDefaultCapable(true);
	stopSolveButton.setMnemonic(KeyEvent.VK_S);
        stopSolveButton.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
                    solveTask.cancel();
		}});
	panel2.add(stopSolveButton, BorderLayout.EAST);
        
	getContentPane().add(panel, BorderLayout.CENTER);

	addWindowListener(new WindowAdapter() {
		public void windowClosing(WindowEvent e) {
		    gsc.deleteGraphics();
		}});

        pack();
	setLocation(100, 100);
	setSize(new Dimension(1000, 700));
	setForeground(Color.darkGray);
	setBackground(Color.lightGray);
        setVisible(true);

    }


    // *******************
    // ***** METHODS *****
    // *******************
    
    /** 
     * Calls the update procedure for the CSP's graphics panel.
     */
    public void updateGraphics() {
        cspPanel.repaint();
    }

    /** 
     * Updates the displayed cost function value.
     * @param costFunction  The cost function of which the costs are to be 
     *                      displayed.
     */
    protected void updateGlobalCosts(CostFunction costFunction) {
	if (costFunction == null) {
	    costsLabel.setText("All Costs Resolved");
	} else {
	    costsLabel.setText(costFunction.getName() + ": " + 
			       costFunction.objectiveCosts);
	}
    }

    /**
     * Deletes all elements of the current CSP.
     */
    protected void clearCSP() {
        gsc.clearCSP();
        MESSAGE_CLEAR_CSP_DONE.activate();
    }


    // ***** TOOLBAR *****

    /** 
     * Choose interaction mode.
     * @param m  The interaction mode.
     * @see TOOLBUTTON constants.
     */
    protected void setInteractionMode(int m) {

	if (m != interaction.mode) {

	    if (interaction.mode != TOOLBUTTON_NONE) {
		toolButton[interaction.mode].setBackground(Color.gray);
	    }
	    interaction.mode = m;
	    toolButton[interaction.mode].setBackground(Color.white);

	    switch (interaction.mode) {

	    case TOOLBUTTON_ADD_VARIABLE: {
		cspPanel.setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR)); 
		enableVariableButtons(false);
		enableConstraintButtons(false);
		MESSAGE_ADD_VARIABLE_1.activate();
		defaultMessage = MESSAGE_ADD_VARIABLE_1;
		break;
	    }

	    case TOOLBUTTON_ADD_OCONSTRAINT: {
		cspPanel.setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR)); 
		enableVariableButtons(false);
		enableConstraintButtons(false);
		MESSAGE_ADD_OCONSTRAINT_1.activate();
		defaultMessage = MESSAGE_ADD_OCONSTRAINT_1;
		break;
	    }

	    case TOOLBUTTON_ADD_CONSTRAINT: {
		cspPanel.setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR)); 
		enableVariableButtons(false);
		enableConstraintButtons(false);
		MESSAGE_ADD_CONSTRAINT_1.activate();
		defaultMessage = MESSAGE_ADD_CONSTRAINT_1;
		break;
	    }

	    case TOOLBUTTON_ADD_COSTS: {
		cspPanel.setCursor(new Cursor(Cursor.HAND_CURSOR)); 
		enableVariableButtons(false);
		enableCostConstraintButtons();
		MESSAGE_ADD_COSTS_1.activate();
		defaultMessage = MESSAGE_ADD_COSTS_1;
		break;
	    }

	    case TOOLBUTTON_EXTEND_CONSTRAINT: {
		cspPanel.setCursor(new Cursor(Cursor.HAND_CURSOR)); 
		enableVariableButtons(false);
		enableExtensibleConstraintButtons();
		MESSAGE_EXTEND_CONSTRAINT_1.activate();
		defaultMessage = MESSAGE_EXTEND_CONSTRAINT_1;
		break;
	    }

	    case TOOLBUTTON_EDIT_NODE: {
		cspPanel.setCursor(new Cursor(Cursor.HAND_CURSOR)); 
		enableVariableButtons(true);
		enableConstraintButtonsEditing();
		MESSAGE_EDIT_NODE_1.activate();
		defaultMessage = MESSAGE_EDIT_NODE_1;
		break;
	    }

	    case TOOLBUTTON_MOVE_NODE: {
		cspPanel.setCursor(new Cursor(Cursor.HAND_CURSOR)); 
		enableVariableButtons(true);
		enableConstraintButtons(true);
		MESSAGE_MOVE_NODE_1.activate();
		defaultMessage = MESSAGE_MOVE_NODE_1;
		break;
	    }

	    case TOOLBUTTON_REDUCE_CONSTRAINT: {
		cspPanel.setCursor(new Cursor(Cursor.HAND_CURSOR)); 
		enableVariableButtons(false);
		enableMinimalConstraintButtons(false);
		MESSAGE_REDUCE_CONSTRAINT_1.activate();
		defaultMessage = MESSAGE_REDUCE_CONSTRAINT_1;
		break;
	    }

	    case TOOLBUTTON_DELETE_NODE: {
		cspPanel.setCursor(new Cursor(Cursor.HAND_CURSOR)); 
		enableVariableButtonsDeletion();
		enableMinimalConstraintButtons(true);
		MESSAGE_DELETE_NODE_1.activate();
		defaultMessage = MESSAGE_DELETE_NODE_1;
		break;
	    }
            
	    case TOOLBUTTON_NONE: {
		cspPanel.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
		enableVariableButtons(false);
		enableConstraintButtons(false);
		break;
	    }

	    default: {
		throw new Error("Invalid operation performed");
	    }

	    }
	}

	interaction.reset();

    }


    // ***** CSP GRAPHICS AREA *****

    /** 
     * Graphics object for displaying/manipulating the CSP.
     */
    private class CSPDisplay extends JPanel {
	public void paintComponent(Graphics g) {

	    synchronized(gsc.variables) {
		synchronized(gsc.constraints) {

		    if (gsc.graphObject != null) {

			// Paint background

			super.paintComponent(g); 
			
			g.setColor(Color.darkGray);
			FontMetrics fontMetrics = g.getFontMetrics();
			int yshift = fontMetrics.getHeight() / 2;
			
			ConventionalConstraint c;
			Connections connections;
			ConnectedNode connectedNode;
			
			Iterator i = gsc.constraints.iterator();
			while (i.hasNext()) {
			    c = (ConventionalConstraint) i.next();
			    connections = c.connections;
			    
			    synchronized(connections) {
				connections.initializeEnumerationAll();
				while ((connectedNode = connections.getNext()) != null) {
				    g.drawLine(((NodeGraphics) c.graphObject).x, ((NodeGraphics) c.graphObject).y, ((NodeGraphics) connectedNode.node.graphObject).x, ((NodeGraphics) connectedNode.node.graphObject).y);
				    if (connectedNode.edgeID != Registry.PLAIN_EDGE) {
					g.drawString(Registry.edgeTypeName[connectedNode.edgeID], (((NodeGraphics) c.graphObject).x + ((NodeGraphics) connectedNode.node.graphObject).x - fontMetrics.stringWidth(Registry.edgeTypeName[connectedNode.edgeID])) / 2, (((NodeGraphics) c.graphObject).y + ((NodeGraphics) connectedNode.node.graphObject).y) / 2 + yshift);
				    }
				}
			    }
			}
		    }

		}
	    }

	}
    }

    /** 
     * Lets the user make a selection of a constraint's cost type (if
     * more than one exists). Returns the selected cost ID or -1 if the
     * choice failed.
     * @param constraint  The constraint for which the selection is to be done.
     */
    private int selectCostID(Constraint constraint) {

	int cNum = constraint.getNumberOfCosts();

	if (cNum > 0) {

	    // Preparing name/ID retrieval

	    String[] cNames = new String[cNum];
	    int[] cID = new int[cNum];

	    int i = 0;
	    for (int id = 0; id < cNum; id++) {
		if (((Node) constraint).isType 
		    || (constraint.getCostFunctionReference(id) == null)) {
		    cNames[i] = constraint.getCostName(id);
		    cID[i] = id;
		    i++;
		}
	    }

	    if (i == 0) {
		return -1;
	    }
	    if (i == 1) {
		return cID[0];
	    }

	    // User's cost type selection

	    ListDialog l = new ListDialog(cNames, null, 
					  "Please select the cost type");
	    int selectedCosts = l.select();
	    if (selectedCosts != -1) {
		return cID[selectedCosts];
	    } else {
		return -1;
	    }

	} else {
	    throw new Error("Constraint has no cost types");
	}

    }

    /** 
     * Finalizes a constraint creation by the user's cost choice and
     * initiating the constraint's construction. Returns false if the 
     * creation was canceled.
     */
    private boolean finalizeConstraintCreation() {

	// User's cost type selection

	int costID = 0;

	if (((ConventionalConstraint) interaction.node).getNumberOfCosts() > 1) {
	    Message MESSAGE_ADD_CONSTRAINT_C = new Message("ADDING CONSTRAINTS      Step " + ++interaction.stepCount + " - Please select the cost type to be used.", Color.yellow);
	    defaultMessage = MESSAGE_ADD_CONSTRAINT_C;
	    MESSAGE_ADD_CONSTRAINT_C.activate();
	    costID = selectCostID((ConventionalConstraint) interaction.node);
	}
	
	if (costID != -1) {

	    // Constraint creation
       
	    if ((interaction.node instanceof ObjectConstraint) 
		|| (interaction.node instanceof IntegerConstraint) 
		|| (interaction.node instanceof RangeConstraint)) {
		gsc.createConstraint(((ConventionalConstraint) interaction.node).typeID, interaction.embedding, interaction.x, interaction.y);
	    } else {
		gsc.createConstraint(((ConventionalConstraint) interaction.node).typeID, interaction.embedding, costID, currentCostMappingOptions, interaction.x, interaction.y);
	    }

	    defaultMessage = MESSAGE_ADD_CONSTRAINT_1;
	    MESSAGE_ADD_CONSTRAINT_DONE.activate();
	    cspPanel.setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR));
	    interaction.reset();

	    return true;

	} else {

	    // Creation aborted 

	    defaultMessage = MESSAGE_ADD_CONSTRAINT_1;
	    MESSAGE_OPERATION_ABORTED.activate();
	    cspPanel.setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR)); 
	    interaction.reset();

	    return false;

	}

    }

    /** 
     * Determines response when the button is pressed in the CSP display area.
     * @param x  The x coordinate of the event.
     * @param y  The y coordinate of the event.
     */
    protected void cspAreaButtonPressed(int x, int y) {

	switch (interaction.mode) {

	case TOOLBUTTON_ADD_VARIABLE: {
	    defaultMessage = MESSAGE_ADD_VARIABLE_2;
	    MESSAGE_ADD_VARIABLE_2.activate();
	    String name = JOptionPane.showInputDialog("Please input the new variable's name");
	    if (name != null) {
		gsc.createVariable(name, x, y);
		MESSAGE_ADD_VARIABLE_DONE.activate();
	    } else {
		MESSAGE_OPERATION_ABORTED.activate();
	    }
	    defaultMessage = MESSAGE_ADD_VARIABLE_1;
	    break;
	}

	case TOOLBUTTON_ADD_OCONSTRAINT: {
	    if (interaction.node == null) {
		defaultMessage = MESSAGE_ADD_OCONSTRAINT_2;
		MESSAGE_ADD_OCONSTRAINT_2.activate();

		// Preparing name/ID retrieval

		int cNum = Registry.objectConstrantTypeNumber();
		String[] cNames = new String[cNum];
		int[] cID = new int[cNum];
		int cMax = Registry.nodeTypeNumber() - 1;
		int i = 0;
		for (int id = 0; id <= cMax; id++) {
		    if (Registry.nodeType[id] instanceof ObjectConstraint) {
			cNames[i] = Registry.nodeTypeName[id];
			cID[i] = id;
			i++;
		    }
		}

		// User's object constraint type selection

		ListDialog l = new ListDialog(cNames, null, "Please select the object constraint's type");
		int selectedConstraint = l.select();
		if (selectedConstraint != -1) {
		    selectedConstraint = cID[selectedConstraint];
		    interaction.node = Registry.nodeType[selectedConstraint];
		    Embedding baseGraph = ((ConventionalConstraint) interaction.node).getBaseGraph();
		    interaction.details = 0;
		    interaction.embedding = 
			new Embedding(baseGraph.connectedNodeNumber);
		    interaction.x = x;
		    interaction.y = y;
		    if (baseGraph.connectedNodeNumber == 0) {
			ConventionalConstraint c = gsc.createConstraint(((ConventionalConstraint) interaction.node).typeID, interaction.embedding, currentCostMappingOptions, interaction.x, interaction.y);
			defaultMessage = MESSAGE_ADD_OCONSTRAINT_1;
			MESSAGE_ADD_OCONSTRAINT_DONE.activate();
			cspPanel.setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR)); 
			interaction.reset();
		    } else {
			Message MESSAGE_ADD_OCONSTRAINT_N = new Message("ADDING OBJECT CONSTRAINTS      Step " + (3 + interaction.details) + " - Please select a  " + Registry.getConnectedNodeDescription(baseGraph.connectedNode[interaction.details]) + "  node for the object constraint.", Color.yellow);
			defaultMessage = MESSAGE_ADD_OCONSTRAINT_N;
			MESSAGE_ADD_OCONSTRAINT_N.activate();
			enableNodeTypeButtons(baseGraph.connectedNode[interaction.details].nodeTypeID);
			cspPanel.setCursor(new Cursor(Cursor.HAND_CURSOR));
		    }
		} else {
		    defaultMessage = MESSAGE_ADD_OCONSTRAINT_1;
		    MESSAGE_OPERATION_ABORTED.activate();
		    interaction.reset();
		}
	    } else {
		JOptionPane.showMessageDialog(null, "Please press the top button of the node to be connected.");
	    }
	    break;
	}

	case TOOLBUTTON_ADD_CONSTRAINT: {
	    if (interaction.node == null) {
		defaultMessage = MESSAGE_ADD_CONSTRAINT_2;
		MESSAGE_ADD_CONSTRAINT_2.activate();

		// Preparing name/ID retrieval

		int cNum = Registry.conventionalConstrantTypeNumber();
		String[] cNames = new String[cNum];
		int[] cID = new int[cNum];
		int cMax = Registry.nodeTypeNumber() - 1;
		int i = 0;
		for (int id = 1; id <= cMax; id++) {
		    if (!(Registry.nodeType[id] instanceof ObjectConstraint) &&
			!(Registry.nodeType[id] instanceof Variable)) {
			cNames[i] = Registry.nodeTypeName[id];
			cID[i] = id;
			i++;
		    }
		}

		// User's constraint type selection

		ListDialog l = new ListDialog(cNames, null, "Please select the constraint's type");
		int selectedConstraint = l.select();
		if (selectedConstraint != -1) {

		    selectedConstraint = cID[selectedConstraint];
		    interaction.node = Registry.nodeType[selectedConstraint];
		    Embedding baseGraph = ((ConventionalConstraint) interaction.node).getBaseGraph();
		    interaction.details = 0;
		    interaction.embedding = 
			new Embedding(baseGraph.connectedNodeNumber);
		    interaction.x = x;
		    interaction.y = y;
		    interaction.stepCount = 2;

		    if (baseGraph.connectedNodeNumber == 0) {
			finalizeConstraintCreation();
		    } else {
			Message MESSAGE_ADD_CONSTRAINT_N = new Message("ADDING CONSTRAINTS      Step " + ++interaction.stepCount + " - Please select a  " + Registry.getConnectedNodeDescription(baseGraph.connectedNode[interaction.details]) + "  node for the constraint.", Color.yellow);
			defaultMessage = MESSAGE_ADD_CONSTRAINT_N;
			MESSAGE_ADD_CONSTRAINT_N.activate();
			enableNodeTypeButtons(baseGraph.connectedNode[interaction.details].nodeTypeID);
			cspPanel.setCursor(new Cursor(Cursor.HAND_CURSOR)); 
		    }
		} else {
		    defaultMessage = MESSAGE_ADD_CONSTRAINT_1;
		    MESSAGE_OPERATION_ABORTED.activate();
		    interaction.reset();
		}
	    } else {
		JOptionPane.showMessageDialog(null, "Please press the top button of the node to be connected.");
	    }
	    break;
	}

	case TOOLBUTTON_EXTEND_CONSTRAINT: {
	    if (interaction.node == null) {
		JOptionPane.showMessageDialog(null, "Please press the top button of the constraint to be extended.");
	    } else {
		JOptionPane.showMessageDialog(null, "Please press the top button of the extension node.");
	    }
	    break;
	}

	case TOOLBUTTON_EDIT_NODE: {
	    JOptionPane.showMessageDialog(null, "Please press the top button of the node to be edited.");
	    break;
	}

	case TOOLBUTTON_MOVE_NODE: {
	    if (interaction.node == null) {
		JOptionPane.showMessageDialog(null, "Please press the top button of the node to be moved.");
	    } else {
		((NodeGraphics) ((Node) interaction.node).graphObject).x = x;
		((NodeGraphics) ((Node) interaction.node).graphObject).y = y;
		((NodeGraphics) ((Node) interaction.node).graphObject).updateSize();
		cspPanel.setCursor(new Cursor(Cursor.HAND_CURSOR)); 
		enableVariableButtons(true);
		enableConstraintButtons(true);
		((NodeGraphics) ((Node) interaction.node).graphObject).revalidate();
		updateGraphics();
		defaultMessage = MESSAGE_MOVE_NODE_1;
		MESSAGE_MOVE_NODE_DONE.activate();
		interaction.reset();
	    }
	    break;
	}

	case TOOLBUTTON_REDUCE_CONSTRAINT: {
	    if (interaction.node == null) {
		JOptionPane.showMessageDialog(null, "Please press the top button of the constraint to be reduced.");
	    } else {
		JOptionPane.showMessageDialog(null, "Please press the top button of the reduction node.");
	    }
	    break;
	}

	case TOOLBUTTON_DELETE_NODE: {
	    JOptionPane.showMessageDialog(null, "Please press the top button of the node to be deleted.");
	    break;
	}

	default: {
	    MESSAGE_CHOOSE_OPERATION.activate();
	}

	}
    }


    // ***** VARIABLES *****

    /** 
     * Displays the graphical representation of a variable.
     * @param v  The variable to be displayed
     */
    protected void displayVariable(Variable v) {
	Dimension d = ((NodeGraphics) v.graphObject).getPreferredSize();
	((NodeGraphics) v.graphObject).setBounds(((NodeGraphics) v.graphObject).x - d.width / 2, ((NodeGraphics) v.graphObject).y - d.height / 2, d.width, d.height);
	cspPanel.add(((NodeGraphics) v.graphObject));
    }

    /** 
     * Displays the graphical representation of a constraint.
     * @param constraint  The constraint to be displayed
     */
    protected void displayConstraint(ConventionalConstraint constraint) {
	Dimension d = 
	    ((NodeGraphics) constraint.graphObject).getPreferredSize();
	((NodeGraphics) constraint.graphObject).setBounds(((NodeGraphics) constraint.graphObject).x - d.width / 2, ((NodeGraphics) constraint.graphObject).y - d.height / 2, d.width, d.height);
	cspPanel.add(((NodeGraphics) constraint.graphObject));
    }

    /** 
     * Sets whether the variables' buttons are enabled.
     * @param b  True to enable the buttons, otherwise false.
     */
    protected void enableVariableButtons(boolean b) {
	Variable v;
	synchronized(gsc.variables) {
	    Iterator i = gsc.variables.iterator();
	    while (i.hasNext()) {
		v = (Variable) i.next();
                if ((v != null) && (v.graphObject != null)) {
                    ((NodeGraphics) v.graphObject).activator.setEnabled(b);
                }
	    }
	}
    }

    /** 
     * Enables variables' buttons for deletion.
     */
    protected void enableVariableButtonsDeletion() {
	Variable v;
	synchronized(gsc.variables) {
	    Iterator i = gsc.variables.iterator();
	    while (i.hasNext()) {
		v = (Variable) i.next();
		if (v.connections.isEmpty()) {
		    ((NodeGraphics) v.graphObject).activator.setEnabled(true);
		} else {
		    ((NodeGraphics) v.graphObject).activator.setEnabled(false);
		}
	    }
	}
    }

    /** 
     * Removes the graphical representation of a variable.
     * @param v  The variable to be removed.
     */
    protected void removeVariable(Variable v) {
	cspPanel.remove(((NodeGraphics) v.graphObject));
    }

    /** 
     * Determines response when a variable's button is pressed.
     * @param v  The affected variable.
     */
    public void variableButtonPressed(Variable v) {

	switch (interaction.mode) {

	case TOOLBUTTON_ADD_OCONSTRAINT: {
	    Embedding baseGraph = ((ConventionalConstraint) interaction.node).getBaseGraph();
	    ConnectedNode connectedNodeBG = 
		baseGraph.connectedNode[interaction.details];
	    gsc.extendEmbedding(interaction.embedding, v, 
				connectedNodeBG.edgeID, 
				connectedNodeBG.direction);
	    interaction.details++;
	    if (interaction.details == baseGraph.connectedNodeNumber) {
		enableVariableButtons(false);
		ConventionalConstraint c = gsc.createConstraint(((ConventionalConstraint) interaction.node).typeID, interaction.embedding, currentCostMappingOptions, interaction.x, interaction.y);

		defaultMessage = MESSAGE_ADD_OCONSTRAINT_1;
		MESSAGE_ADD_OCONSTRAINT_DONE.activate();
		cspPanel.setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR)); 
		interaction.reset();
	    } else {
		Message MESSAGE_ADD_OCONSTRAINT_N = new Message("ADDING OBJECT CONSTRAINTS      Step " + (3 + interaction.details) + " - Please select a  " + Registry.getConnectedNodeDescription(baseGraph.connectedNode[interaction.details]) + "  node for the object constraint.", Color.yellow);
		enableNodeTypeButtons(baseGraph.connectedNode[interaction.details].nodeTypeID);
		MESSAGE_ADD_OCONSTRAINT_N.activate();
		defaultMessage = MESSAGE_ADD_OCONSTRAINT_N;
	    }
	    break;
	}

	case TOOLBUTTON_ADD_CONSTRAINT: {
	    Embedding baseGraph = 
		((ConventionalConstraint) interaction.node).getBaseGraph();
	    ConnectedNode connectedNodeBG = 
		baseGraph.connectedNode[interaction.details];
	    gsc.extendEmbedding(interaction.embedding, v, 
				connectedNodeBG.edgeID, 
				connectedNodeBG.direction);
	    interaction.details++;
	    if (interaction.details == baseGraph.connectedNodeNumber) {
		enableVariableButtons(false);
		finalizeConstraintCreation();
	    } else {
		Message MESSAGE_ADD_CONSTRAINT_N = new Message("ADDING CONSTRAINTS      Step " + ++interaction.stepCount + " - Please select a  " + Registry.getConnectedNodeDescription(baseGraph.connectedNode[interaction.details]) + "  node for the constraint.", Color.yellow);
		defaultMessage = MESSAGE_ADD_CONSTRAINT_N;
		MESSAGE_ADD_CONSTRAINT_N.activate();
	    }
	    break;
	}

	case TOOLBUTTON_EXTEND_CONSTRAINT: {

	    // Handling a selected extension

	    enableVariableButtons(false);
	    Embedding extensionGraph = ((ConventionalConstraint) interaction.node).getExtensionGraph(interaction.extension);
	    ConnectedNode connectedNodeEG = 
		extensionGraph.connectedNode[interaction.details];
	    gsc.extendEmbedding(interaction.embedding, v, 
				connectedNodeEG.edgeID, 
				connectedNodeEG.direction);
	    interaction.details++;
	    if (interaction.details == extensionGraph.connectedNodeNumber) {
		gsc.extendConstraint((ConventionalConstraint) interaction.node, interaction.embedding);

		updateGraphics();
		defaultMessage = MESSAGE_EXTEND_CONSTRAINT_1;
		MESSAGE_EXTEND_CONSTRAINT_DONE.activate();
		enableExtensibleConstraintButtons();
		interaction.reset();
	    } else {
		Message MESSAGE_EXTEND_CONSTRAINT_N = new Message("EXTENDING CONSTRAINTS      Step " + ++interaction.stepCount + " - Please select a  " + Registry.getConnectedNodeDescription(extensionGraph.connectedNode[interaction.details]) + "  node for the constraint.", Color.yellow);
		defaultMessage = MESSAGE_EXTEND_CONSTRAINT_N;
		MESSAGE_EXTEND_CONSTRAINT_N.activate();
		enableNodeTypeButtons(extensionGraph.connectedNode[interaction.details].nodeTypeID);
	    }
	    break;
	}

	case TOOLBUTTON_EDIT_NODE: {
	    defaultMessage = MESSAGE_EDIT_NODE_2;
	    MESSAGE_EDIT_NODE_2.activate();
	    String value = JOptionPane.showInputDialog("Please input the variable's new value");
	    if (value != null) { 
		long i = 0;
		double d = 0.0;
		boolean okLong = true;
		boolean okDouble = true;
		try {
		    i = Long.parseLong(value);
		} catch (NumberFormatException e) {
		    okLong = false;
		}
		if (!okLong) try {
		    d = Double.parseDouble(value);
		} catch (NumberFormatException e) {
		    okDouble = false;
		    JOptionPane.showMessageDialog(null, "This is an inadmissible value!", null, JOptionPane.ERROR_MESSAGE);
		    MESSAGE_OPERATION_ABORTED.activate();
		}
		
		if (okLong) {
		    if ((v.integerLowerBound > i) 
			|| (v.integerUpperBound < i)) {
			JOptionPane.showMessageDialog(null, "Value is shifted because of range constraints!", null, JOptionPane.WARNING_MESSAGE);
		    }
		    gsc.setVariable(v, i);
		    MESSAGE_EDIT_NODE_DONE.activate();
		} else {
		    if (okDouble) {
			if (v.integerConstraints != 0) {
			    JOptionPane.showMessageDialog(null, "Value is transformed to an integer because of integer constraints!", null, JOptionPane.WARNING_MESSAGE);
			    if ((v.integerLowerBound > Math.round(d)) || 
				(v.integerUpperBound < Math.round(d))) {
				JOptionPane.showMessageDialog(null, "Value is shifted because of range constraints!", null, JOptionPane.WARNING_MESSAGE);
			    }
			} else {
			    if ((v.lowerBound > d) || (v.upperBound < d)) {
				JOptionPane.showMessageDialog(null, "Value is shifted because of range constraints!", null, JOptionPane.WARNING_MESSAGE);
			    }
			}
			gsc.setVariable(v, d);
			MESSAGE_EDIT_NODE_DONE.activate();
		    }
		}
	    } else {
		MESSAGE_OPERATION_ABORTED.activate();
	    }
	    defaultMessage = MESSAGE_EDIT_NODE_1;
	    break;
	}

	case TOOLBUTTON_MOVE_NODE: {
	    interaction.node = v;
	    enableVariableButtons(false);
	    enableConstraintButtons(false);
	    cspPanel.setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR)); 
	    defaultMessage = MESSAGE_MOVE_NODE_2;
	    MESSAGE_MOVE_NODE_2.activate();
	    break;
	}

	case TOOLBUTTON_REDUCE_CONSTRAINT: {

	    // Handling a selected extension

	    enableVariableButtons(false);
	    Embedding extensionGraph = ((ConventionalConstraint) interaction.node).getExtensionGraph(interaction.extension);
	    ConnectedNode connectedNodeEG = 
		extensionGraph.connectedNode[interaction.details];
	    gsc.extendEmbedding(interaction.embedding, v, 
				connectedNodeEG.edgeID, 
				connectedNodeEG.direction);
	    interaction.details++;
	    if (interaction.details == extensionGraph.connectedNodeNumber) {
		gsc.reduceConstraint((ConventionalConstraint) 
				     interaction.node, 
				     interaction.embedding);

		enableVariableButtons(false);
		enableMinimalConstraintButtons(false);
		updateGraphics();
		defaultMessage = MESSAGE_REDUCE_CONSTRAINT_1;
		MESSAGE_REDUCE_CONSTRAINT_DONE.activate();
		interaction.reset();
	    } else {
		Message MESSAGE_REDUCE_CONSTRAINT_N = new Message("REDUCING CONSTRAINTS      Step " + ++interaction.stepCount + " - Please select a  " + Registry.getConnectedNodeDescription(extensionGraph.connectedNode[interaction.details]) + "  node for the constraint.", Color.yellow);
		defaultMessage = MESSAGE_REDUCE_CONSTRAINT_N;
		MESSAGE_REDUCE_CONSTRAINT_N.activate();
		enableConnectedTypeButtons((Node) interaction.node, extensionGraph.connectedNode[interaction.details].nodeTypeID, extensionGraph.connectedNode[interaction.details].edgeID, extensionGraph.connectedNode[interaction.details].direction);
	    }
	    break;
	}

	case TOOLBUTTON_DELETE_NODE: {
	    if (JOptionPane.showConfirmDialog(null, "Do you really want to delete the node?", null, JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE) == JOptionPane.YES_OPTION) {
		gsc.deleteVariable(v);
		MESSAGE_DELETE_NODE_DONE.activate();
	    } else {
		MESSAGE_OPERATION_ABORTED.activate();
	    }
	    defaultMessage = MESSAGE_DELETE_NODE_1;
	    break;
	}

	default: {
	    throw new Error("Invalid operation performed");
	}

	}

    }


    // ***** CONSTRAINTS *****

    /** 
     * Updates the cost display of constraints that are connected to a
     * specific cost function.
     * @param level  The level of the cost function to be considered.
     */
    protected void updateConstraintCosts(int level) {

	OrderedShiftDLLElement listElement;
	if (level == Integer.MAX_VALUE) {
	    listElement = gsc.SCSP_FUNCTION.costFunctionReferences.first;
	} else {
	    listElement = ((CostFunction) gsc.costFunctions.get(level)).costFunctionReferences.first;
	}
	Constraint c;

	while (listElement != null) {
	    c = (Constraint) ((CostFunctionReference) listElement.element).constraint;
	    if ((c instanceof ConventionalConstraint) && 
		!(c instanceof ObjectConstraint)) {
		if (((ConventionalConstraint) c).graphObject instanceof ConventionalConstraintGraphics) {
		    ((ConventionalConstraintGraphics) ((ConventionalConstraint) c).graphObject).updateCosts();
		}
	    }
	    listElement = listElement.next;
	}
    }

    /** 
     * Sets whether the constraints' buttons are enabled.
     * @param b  True to enable the buttons, otherwise false.
     */
    protected void enableConstraintButtons(boolean b) {
	ConventionalConstraint c;
	synchronized(gsc.constraints) {
	    Iterator i = gsc.constraints.iterator();
	    while (i.hasNext()) {
		c = (ConventionalConstraint) i.next();
                if ((c != null) && (c.graphObject != null) && (((NodeGraphics) c.graphObject).activator != null)) {
                    ((NodeGraphics) c.graphObject).activator.setEnabled(b);
                }
	    }
	}
    }

    /** 
     * Tests whether a constraint has more than one cost type not connected 
     * to a cost function.
     * @param c  The constraint to be tested.
     */
    protected boolean moreThanOneCostChoice(ConventionalConstraint c) {

	int cNum = c.getNumberOfCosts();
	int unusedCosts = 0;

	for (int i = 0; (unusedCosts < 2) && (i < cNum); i++) {
	    if (c.getCostFunctionReference(i) == null) {
		unusedCosts++;
	    }
	}
	return (unusedCosts > 1);

    }

    /** 
     * Enables button of a constraint if it has at least one cost type
     * not connected to a cost function.
     * @param c  The constraint to be tested.
     */
    protected void enableCostConstraintButton(ConventionalConstraint c) {

	int cNum = c.getNumberOfCosts();
	boolean toBeEnabled = false;

	for (int i = 0; !toBeEnabled && (i < cNum); i++) {
	    if (c.getCostFunctionReference(i) == null) {
		toBeEnabled = true;
	    }
	}
	((NodeGraphics) c.graphObject).activator.setEnabled(toBeEnabled);

    }

    /** 
     * Enables buttons of constraints which have at least one cost type
     * not connected to a cost function.
     */
    protected void enableCostConstraintButtons() {
	synchronized(gsc.constraints) {
	    Iterator it = gsc.constraints.iterator();
	    while (it.hasNext()) {
		enableCostConstraintButton((ConventionalConstraint) it.next());
	    }
	}
    }

    /** 
     * Enables buttons of constraints that have only a
     * base graph embedding.
     * @param b  True to enable the buttons, false to enable 
     *           all other constraint buttons (but those with
     *           no extension graphs).
     */
    protected void enableMinimalConstraintButtons(boolean b) {
	ConventionalConstraint c;
	synchronized(gsc.constraints) {
	    Iterator i = gsc.constraints.iterator();
	    while (i.hasNext()) {
		c = (ConventionalConstraint) i.next();
		if (c.getBaseGraph().connectedNodeNumber == 
		    c.connections.number()) {
		    ((NodeGraphics) c.graphObject).activator.setEnabled(b);
		} else {
		    if (!b && (c.getExtensionGraphNumber() == 0)) {
			((NodeGraphics) c.graphObject).activator.setEnabled(false);
		    } else {
			((NodeGraphics) c.graphObject).activator.setEnabled(!b);
		    }
		}
	    }
	}
    }

    /** 
     * Enables buttons of constraints that can be extended.
     */
    protected void enableExtensibleConstraintButtons() {
	ConventionalConstraint c;
	synchronized(gsc.constraints) {
	    Iterator i = gsc.constraints.iterator();
	    while (i.hasNext()) {
		c = (ConventionalConstraint) i.next();
		if ((c.getExtensionGraphNumber() > 0) && 
		    (c.getMaximalConnectionNumber() 
		     > c.connections.number())) {
		    ((NodeGraphics) c.graphObject).activator.setEnabled(true);
		} else {
		    ((NodeGraphics) c.graphObject).activator.setEnabled(false);
		}
	    }
	}
    }

    /** 
     * Enables constraints' buttons for editing.
     */
    protected void enableConstraintButtonsEditing() {
	ConventionalConstraint c;
	synchronized(gsc.constraints) {
	    Iterator i = gsc.constraints.iterator();
	    while (i.hasNext()) {
		c = (ConventionalConstraint) i.next();
		if ((c instanceof RangeConstraint) || (c instanceof ExponentialConstraint)||(c instanceof ProportionConstraint)) {
		    ((NodeGraphics) c.graphObject).activator.setEnabled(true);
		} else {
		    ((NodeGraphics) c.graphObject).activator.setEnabled(false);
		}
	    }
	}
    }

    /** 
     * Enables buttons of a specific node type.
     * @param typeID  The node type.
     * @see Registry
     */
    protected void enableNodeTypeButtons(int typeID) {
	if (typeID == Registry.VARIABLE) {
	    enableConstraintButtons(false);
	    enableVariableButtons(true);
	} else {
	    enableVariableButtons(false);
	    ConventionalConstraint c;
	    synchronized(gsc.constraints) {
		Iterator i = gsc.constraints.iterator();
		while (i.hasNext()) {
		    c = (ConventionalConstraint) i.next();
		    if (c.typeID == typeID) {
			((NodeGraphics) c.graphObject).activator.setEnabled(true);
		    } else {
			((NodeGraphics) c.graphObject).activator.setEnabled(false);
		    }
		}
	    }
	}
    }

    /** 
     * Enables buttons that have a specific connection to a node.
     * @param node  The node from which the connection goes.
     * @param typeID  The node type of the connected element.
     * @param edgeID  Type of the edge.
     * @param direction  Direction of the edge.
     * @see Registry
     */
    protected void enableConnectedTypeButtons(Node node, int typeID, 
					      int edgeID, boolean direction) {
	enableVariableButtons(false);
	enableConstraintButtons(false);

	Connections connections = node.connections;;
	ConnectedNode connectedNode;

	synchronized(connections) {
	    connections.initializeEnumerationAll();
	    while ((connectedNode = connections.getNext()) != null) {
		if ((connectedNode.nodeTypeID == typeID) &&
		    (connectedNode.edgeID == edgeID) &&
		    (connectedNode.direction == direction)) {
		    ((NodeGraphics) connectedNode.node.graphObject).activator.setEnabled(true);
		}
	    }
	}
    }

    /** 
     * Removes the graphical representation of a constraint.
     * @param constraint  The constraint to be removed.
     */
    protected void removeConstraint(ConventionalConstraint constraint) {
	cspPanel.remove(((NodeGraphics) constraint.graphObject));
    }

    /** 
     * Determines response when a constraint's button is pressed.
     * @param constraint  The affected constraint.
     */
    public void constraintButtonPressed(ConventionalConstraint constraint) {

	switch (interaction.mode) {

	case TOOLBUTTON_ADD_COSTS: {
	    if (moreThanOneCostChoice(constraint)) {
		defaultMessage = MESSAGE_ADD_COSTS_2;
		MESSAGE_ADD_COSTS_2.activate();
	    }
	    
	    int costID = selectCostID(constraint);
	    if (costID != -1) {
		gsc.addCosts(constraint, costID, currentCostMappingOptions);
		enableCostConstraintButton(constraint);
		defaultMessage = MESSAGE_ADD_COSTS_1;
		MESSAGE_ADD_COSTS_DONE.activate();
	    } else {
		defaultMessage = MESSAGE_ADD_COSTS_1;
		MESSAGE_OPERATION_ABORTED.activate();
	    }
	    break;
	}

	case TOOLBUTTON_ADD_CONSTRAINT: {
	    throw new Error("Not yet implemented!");
	}

	case TOOLBUTTON_EXTEND_CONSTRAINT: {
	    enableConstraintButtons(false);

	    if (interaction.node == null) {

		interaction.node = constraint;
		boolean aborted = false;

		if (constraint.getExtensionGraphNumber() > 1) {

		    // Choosing one of the extensions

		    defaultMessage = MESSAGE_EXTEND_CONSTRAINT_2;
		    MESSAGE_EXTEND_CONSTRAINT_2.activate();
		    String[] cNames = 
			new String[constraint.getExtensionGraphNumber()];
		    int i, i2;
		    for (i = 0; i < constraint.getExtensionGraphNumber(); i++) {
			StringBuffer tmpString = new StringBuffer();
			for (i2 = 0; i2 < constraint.getExtensionGraph(i).connectedNodeNumber; i2++) {
			    if (i2 != 0) {
				tmpString.append(" , ");
			    }
			    tmpString.append(Registry.getConnectedNodeDescription(constraint.getExtensionGraph(i).connectedNode[i2]));
			}
			cNames[i] = tmpString.toString();
		    }
		    ListDialog l = new ListDialog(cNames, null, "Please select the extension type");
		    interaction.extension = l.select();
		    if (interaction.extension == -1) {
			aborted = true;
		    } else {
			interaction.stepCount = 3;
		    }
		} else {
		    interaction.extension = 0;
		    interaction.stepCount = 2;
		}
		if (!aborted) {
		    Embedding extensionGraph = ((ConventionalConstraint) interaction.node).getExtensionGraph(interaction.extension);
		    interaction.details = 0;
		    interaction.embedding = 
			new Embedding(extensionGraph.connectedNodeNumber);
		    Message MESSAGE_EXTEND_CONSTRAINT_N = new Message("EXTENDING CONSTRAINTS      Step " + interaction.stepCount + " - Please select a  " + Registry.getConnectedNodeDescription(extensionGraph.connectedNode[interaction.details]) + "  node for the constraint.", Color.yellow);
		    defaultMessage = MESSAGE_EXTEND_CONSTRAINT_N;
		    MESSAGE_EXTEND_CONSTRAINT_N.activate();
		    enableNodeTypeButtons(extensionGraph.connectedNode[interaction.details].nodeTypeID);
		} else {
		    defaultMessage = MESSAGE_EXTEND_CONSTRAINT_1;
		    MESSAGE_OPERATION_ABORTED.activate();
		    interaction.reset();
		    enableExtensibleConstraintButtons();
		}
	    } else {

		// Handling a selected extension

		enableConstraintButtons(false);
		Embedding extensionGraph = ((ConventionalConstraint) interaction.node).getExtensionGraph(interaction.extension);
		ConnectedNode connectedNodeEG = 
		    extensionGraph.connectedNode[interaction.details];
		gsc.extendEmbedding(interaction.embedding, constraint, 
				    connectedNodeEG.edgeID, 
				    connectedNodeEG.direction);
		interaction.details++;
		if (interaction.details 
		    == extensionGraph.connectedNodeNumber) {
		    gsc.extendConstraint((ConventionalConstraint) interaction.node, interaction.embedding);

		    updateGraphics();
		    defaultMessage = MESSAGE_EXTEND_CONSTRAINT_1;
		    MESSAGE_EXTEND_CONSTRAINT_DONE.activate();
		    interaction.reset();
		    enableExtensibleConstraintButtons();
		} else {
		    Message MESSAGE_EXTEND_CONSTRAINT_N = new Message("EXTENDING CONSTRAINTS      Step " + ++interaction.stepCount + " - Please select a  " + Registry.getConnectedNodeDescription(extensionGraph.connectedNode[interaction.details]) + "  node for the constraint.", Color.yellow);
		    defaultMessage = MESSAGE_EXTEND_CONSTRAINT_N;
		    MESSAGE_EXTEND_CONSTRAINT_N.activate();
		    enableNodeTypeButtons(extensionGraph.connectedNode[interaction.details].nodeTypeID);
		}
	    }
	    break;
	}

	case TOOLBUTTON_EDIT_NODE: {
            if (constraint instanceof ExponentialConstraint) {
                defaultMessage = MESSAGE_EDIT_NODE_2R;
                MESSAGE_EDIT_NODE_2R.activate();
                String value = JOptionPane.showInputDialog("Please input the new exponent");
                if (value != null) { 
                    double d = 1.0;
                    try {
                        d = Double.parseDouble(value);
                    }    catch (NumberFormatException e) {
                        JOptionPane.showMessageDialog(null, "This is an inadmissible value!", null, JOptionPane.ERROR_MESSAGE);
                        MESSAGE_OPERATION_ABORTED.activate();
                    }
                    ((ExponentialConstraint)constraint).setExponent(d);
                } else {
		MESSAGE_OPERATION_ABORTED.activate();
                }
            } else if (constraint instanceof ProportionConstraint) {
                defaultMessage = MESSAGE_EDIT_NODE_2R;
                MESSAGE_EDIT_NODE_2R.activate();
                String value = JOptionPane.showInputDialog("Please input the new proportion");
                if (value != null) { 
                    
                    double d = 1.0;
                    try {
                        d = Double.parseDouble(value);
                    }    catch (NumberFormatException e) {
                        JOptionPane.showMessageDialog(null, "This is an inadmissible value!", null, JOptionPane.ERROR_MESSAGE);
                        MESSAGE_OPERATION_ABORTED.activate();
                    }
                    ((ProportionConstraint)constraint).setCoefficient(d);
                } else {
		MESSAGE_OPERATION_ABORTED.activate();
                }
            } else if (constraint instanceof RangeConstraint) {

		// Creating user interaction window

		defaultMessage = MESSAGE_EDIT_NODE_2R;
		MESSAGE_EDIT_NODE_2R.activate();
		RangeConstraint rconstraint = (RangeConstraint) constraint;

		JPanel mainPanel = new JPanel();
		mainPanel.setLayout(new BoxLayout(mainPanel, 
						  BoxLayout.Y_AXIS));

		JLabel m = 
		    new JLabel("Please edit the constraint's range bounds");
		m.setAlignmentX(Component.LEFT_ALIGNMENT);
		m.setForeground(Color.black);
		mainPanel.add(m);
		mainPanel.add(Box.createRigidArea(new Dimension(0,8)));

		JPanel rangeInputPanel = new JPanel();
		rangeInputPanel.setLayout(new GridLayout(2,3));
		JLabel lowerBoundLabel = new JLabel("Lower bound");
		rangeInputPanel.add(lowerBoundLabel);
		final JTextField lowerBoundText;
		lowerBoundText = new
		    JTextField(Double.toString(rconstraint.lowerBound), 12);
		rangeInputPanel.add(lowerBoundText);
		JButton lowerBoundReset = new JButton("Deactivate");
		lowerBoundReset.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
			    lowerBoundText.setText(Double.toString(Double.NEGATIVE_INFINITY));
			}});
		rangeInputPanel.add(lowerBoundReset);
		JLabel upperBoundLabel = new JLabel("Upper bound");
		rangeInputPanel.add(upperBoundLabel);
		final JTextField upperBoundText;
		upperBoundText = new
		    JTextField(Double.toString(rconstraint.upperBound), 12);
		rangeInputPanel.add(upperBoundText);
		JButton upperBoundReset = new JButton("Deactivate");
		upperBoundReset.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
			    upperBoundText.setText(Double.toString(Double.POSITIVE_INFINITY));
			}});
		rangeInputPanel.add(upperBoundReset);

		rangeInputPanel.setAlignmentX(Component.LEFT_ALIGNMENT);
		mainPanel.add(rangeInputPanel);

		// Input evaluation

		if (JOptionPane.showConfirmDialog(null, mainPanel, null, JOptionPane.OK_CANCEL_OPTION) != JOptionPane.OK_OPTION) {
		    MESSAGE_OPERATION_ABORTED.activate();
		} else {
		    String lowerBoundString = lowerBoundText.getText();
		    String upperBoundString = upperBoundText.getText();
		    boolean successLB = true;
		    double lbound = 0.0;
		    try {
			lbound = Double.parseDouble(lowerBoundString);
		    } catch (NumberFormatException e) {
			if (lowerBoundString.equals("-Infinity")) {
			    lbound = Double.NEGATIVE_INFINITY;
			} else {
			    JOptionPane.showMessageDialog(null, "The given lower bound \"" + lowerBoundString + "\" is an inadmissible value!", null, JOptionPane.ERROR_MESSAGE);
			    successLB = false;
			}
		    }
		    boolean successUB = true;
		    double ubound = 0.0;
		    try {
			ubound = Double.parseDouble(upperBoundString);
		    } catch (NumberFormatException e) {
			if (upperBoundString.equals("Infinity")) {
			    ubound = Double.POSITIVE_INFINITY;
			} else {
			    JOptionPane.showMessageDialog(null, "The given upper bound \"" + upperBoundString + "\" is an inadmissible value!", null, JOptionPane.ERROR_MESSAGE);
			    successUB = false;
			}
		    }

		    successLB = successLB 
			&& (lbound != rconstraint.lowerBound);
		    successUB = successUB 
			&& (ubound != rconstraint.upperBound);

		    double newlb = successLB?lbound:rconstraint.lowerBound;
		    double newub = successUB?ubound:rconstraint.upperBound;
		    if (newlb > newub) {
			JOptionPane.showMessageDialog(null, "The lower bound \"" + lowerBoundString + "\" is larger than the upper bound \"" + upperBoundString + "\"!", null, JOptionPane.ERROR_MESSAGE);
			successLB = false;
			successUB = false;
		    }

		    if (successLB) {
			if (successUB) {
			    gsc.setBounds(constraint, lbound, ubound);
			    MESSAGE_EDIT_NODE_DONE.activate();
			} else {
			    gsc.setLowerBound(constraint, lbound);
			    MESSAGE_EDIT_NODE_DONE_PARTIAL.activate();
			}
		    } else { 
			if (successUB) {
			    gsc.setUpperBound(constraint, ubound);
			    MESSAGE_EDIT_NODE_DONE_PARTIAL.activate();
			} else {
			    MESSAGE_OPERATION_ABORTED.activate();
			}
		    }
		}
	    } else {
		throw new Error("Invalid operation performed");
	    }
	    defaultMessage = MESSAGE_EDIT_NODE_1;
	    break;
	}

	case TOOLBUTTON_MOVE_NODE: {
	    interaction.node = constraint;
	    enableVariableButtons(false);
	    enableConstraintButtons(false);
	    cspPanel.setCursor(new Cursor(Cursor.CROSSHAIR_CURSOR)); 
	    defaultMessage = MESSAGE_MOVE_NODE_2;
	    MESSAGE_MOVE_NODE_2.activate();
	    break;
	}

	case TOOLBUTTON_REDUCE_CONSTRAINT: {
	    enableConstraintButtons(false);

	    if (interaction.node == null) {

		interaction.node = constraint;
		boolean aborted = false;

		if (constraint.getExtensionGraphNumber() > 1) {

		    // Choosing one of the extensions

		    defaultMessage = MESSAGE_REDUCE_CONSTRAINT_2;
		    MESSAGE_REDUCE_CONSTRAINT_2.activate();
		    String[] cNames = 
			new String[constraint.getExtensionGraphNumber()];
		    int i, i2;
		    for (i = 0; i < constraint.getExtensionGraphNumber(); i++) {
			StringBuffer tmpString = new StringBuffer();
			for (i2 = 0; i2 < constraint.getExtensionGraph(i).connectedNodeNumber; i2++) {
			    if (i2 != 0) {
				tmpString.append(" , ");
			    }
			    tmpString.append(Registry.getConnectedNodeDescription(constraint.getExtensionGraph(i).connectedNode[i2]));
			}
			cNames[i] = tmpString.toString();
		    }
		    ListDialog l = new ListDialog(cNames, null, "Please select the reduction type");
		    interaction.extension = l.select();
		    if (interaction.extension == -1) {
			aborted = true;
		    } else {
			interaction.stepCount = 3;
		    }
		} else {
		    interaction.extension = 0;
		    interaction.stepCount = 2;
		}
		if (!aborted) {
		    Embedding extensionGraph = ((ConventionalConstraint) interaction.node).getExtensionGraph(interaction.extension);
		    interaction.details = 0;
		    interaction.embedding = 
			new Embedding(extensionGraph.connectedNodeNumber);
		    Message MESSAGE_REDUCE_CONSTRAINT_N = new Message("REDUCING CONSTRAINTS      Step " + interaction.stepCount + " - Please select a  " + Registry.getConnectedNodeDescription(extensionGraph.connectedNode[interaction.details]) + "  node of the constraint.", Color.yellow);
		    defaultMessage = MESSAGE_REDUCE_CONSTRAINT_N;
		    MESSAGE_REDUCE_CONSTRAINT_N.activate();
		    enableConnectedTypeButtons((Node) interaction.node, extensionGraph.connectedNode[interaction.details].nodeTypeID, extensionGraph.connectedNode[interaction.details].edgeID, extensionGraph.connectedNode[interaction.details].direction);
		} else {
		    defaultMessage = MESSAGE_REDUCE_CONSTRAINT_1;
		    MESSAGE_OPERATION_ABORTED.activate();
		    interaction.reset();
		    enableMinimalConstraintButtons(false);
		}
	    } else {

		// Handling a selected extension

		enableVariableButtons(false);
		enableConstraintButtons(false);
		Embedding extensionGraph = ((ConventionalConstraint) interaction.node).getExtensionGraph(interaction.extension);
		ConnectedNode connectedNodeEG = 
		    extensionGraph.connectedNode[interaction.details];
		gsc.extendEmbedding(interaction.embedding, constraint, 
				    connectedNodeEG.edgeID, 
				    connectedNodeEG.direction);
		interaction.details++;
		if (interaction.details 
		    == extensionGraph.connectedNodeNumber) {
		    gsc.reduceConstraint((ConventionalConstraint) interaction.node, interaction.embedding);

		    enableVariableButtons(false);
		    enableMinimalConstraintButtons(false);
		    updateGraphics();
		    defaultMessage = MESSAGE_REDUCE_CONSTRAINT_1;
		    MESSAGE_REDUCE_CONSTRAINT_DONE.activate();
		    interaction.reset();
		} else {
		    Message MESSAGE_REDUCE_CONSTRAINT_N = new Message("REDUCING CONSTRAINTS      Step " + ++interaction.stepCount + " - Please select a  " + Registry.getConnectedNodeDescription(extensionGraph.connectedNode[interaction.details]) + "  node for the constraint.", Color.yellow);
		    defaultMessage = MESSAGE_REDUCE_CONSTRAINT_N;
		    MESSAGE_REDUCE_CONSTRAINT_N.activate();
		    enableConnectedTypeButtons((Node) interaction.node, extensionGraph.connectedNode[interaction.details].nodeTypeID, extensionGraph.connectedNode[interaction.details].edgeID, extensionGraph.connectedNode[interaction.details].direction);
		}
	    }
	    break;
	}

	case TOOLBUTTON_DELETE_NODE: {
	    if (JOptionPane.showConfirmDialog(null, "Do you really want to delete the node?", null, JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE) == JOptionPane.YES_OPTION) {
		gsc.deleteConstraint(constraint);
		MESSAGE_DELETE_NODE_DONE.activate();
	    } else {
		MESSAGE_OPERATION_ABORTED.activate();
	    }
	    defaultMessage = MESSAGE_DELETE_NODE_1;
	    enableVariableButtonsDeletion();
	    enableMinimalConstraintButtons(true);
	    break;
	}

	default: {
	    throw new Error("Invalid operation performed");
	}

	}
    }


    // ***** MESSAGE FIELD *****

    /** 
     * Represents a message to be displayed for the user.
     */
    public class Message {
	protected String m;
	protected Color c;
	protected Message(String m, Color c) {
	    this.m = m;
	    this.c = c;
	}
	protected void activate() {
	    messageQueue.addLast(this);
	    if (timer == null) {
		messageHighlighted = false;
		timer = new javax.swing.Timer(0, new ActionListener() {
			public void actionPerformed(ActionEvent e) {
			    timer.stop();
			    if (messageHighlighted) {
				messagePanel.setBackground(Color.lightGray);
				messageHighlighted = false;
				if (messageQueue.size() > 0) {
				    timer.setInitialDelay(100 / (messageQueue.size() + 1));
				} else {
				    timer.setInitialDelay(2000);
				}
                                timer.start();
			    } else {	    
				if (messageQueue.size() > 0) {
				    Message next = 
					(Message)messageQueue.removeFirst();
				    messageLabel.setText(next.m);
				    messagePanel.setBackground(next.c);
				    messageHighlighted = true;
				    timer.setInitialDelay(100 / (messageQueue.size() + 1));
				    timer.start();
				} else {
				    if (defaultMessage != null) {
					messageLabel.setText(defaultMessage.m);
				    }
				    timer = null;
				}
			    }
			}});
		timer.setRepeats(false);
		timer.start();
	    } else {
		if (timer.getInitialDelay() > 100 / (messageQueue.size() + 1)) {
		    timer.stop();
		    timer.setInitialDelay(100 / (messageQueue.size() + 1));
		    timer.start();
		}
	    }
	}
    }

    /** 
     * Displays a message for the user.
     * @param s  The string to be displayed.
     */
    protected void displayMessage(String s, Color c) {
	messageLabel.setText(s);
	if (timer == null) {
	    messagePanel.setBackground(c);
	    timer = new javax.swing.Timer(100, new ActionListener() {
		    public void actionPerformed(ActionEvent e) {
			messagePanel.setBackground(Color.lightGray);
			timer.stop();
			timer = null;
		    }});
	    timer.start();
	}
    }

}
