
package com.ai_center.dragonbreath;


import java.util.*;

import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.costs.CostMapping;


/**
 * Registry for information and factory for CSP components.
 *
 * @author Alexander Nareyek
 * @version $Revision: 2.18 $
 *
 * @see dragonbreath
 */
public final class Registry
{


    // **********************
    // ***** ATTRIBUTES *****
    // **********************


    /* Start of customizable constants */

    /**
     * Version number of this DragonBreath engine.
     */
    public final static String VERSION = new String( "ALPHA-3" );

    /**
     * Declares whether internal consistency check should be
     * done. The consistency checks will always be done for
     * user inputs.
     */
    public final static boolean SAFE_MODE = true;

    /**
     * Declares whether the heuristics should print a description
     * of the changes they apply to the constraint graph.
     */
    public final static boolean LOG_CHANGES = false;

    /**
     * The maximal number of node types that can be registered.
     */
    public final static int MAX_NODE_TYPE_NUMBER = 512;

    /**
     * The maximal number of edge types that can be registered.
     */
    public final static int MAX_EDGE_TYPE_NUMBER = 512;

    /**
     * The maximal number of StructuralConstraint types that can be registered.
     */
    public final static int MAX_SC_TYPE_NUMBER = 256;

    /**
     * The maximal number of cost mappings that can be registered.
     */
    public final static int MAX_COST_MAPPING_NUMBER = 512;


   /**
    * The maximal number of breakpoints that can be registered.
    */
   public final static int MAX_BREAKPOINT_NUMBER = 100;

    /* End of customizable constants */


    /**
     * A handle to the unique Registry instance.
     */
    private static Registry _instance = null;


    // ***** Edges *****

     /**
     * Incoming edge direction constant.
     */
    public final static boolean INCOMING = true;
    /**
     * Outgoing edge direction constant.
     */
    public final static boolean OUTGOING = false;

    /**
     * Array of the edges' type names.
     */
    public static String[] edgeTypeName = new String[ MAX_EDGE_TYPE_NUMBER ];
    /**
     * Number of registered edge types.
     */
    private static int currentEdgeTypeNumber = 0;
    /**
     * Constants for registered edge types.
     */
    public final static int NULL_EDGE  = registerEdgeType( "noEdge" );
    public final static int ANY_EDGE  = registerEdgeType( "any" );
    public final static int PLAIN_EDGE = registerEdgeType( "plain" );
    public final static int BEGIN_EDGE = registerEdgeType( "begin" );
    public final static int DURATION_EDGE = registerEdgeType( "duration" );
    public final static int TEMPORALREFERENCE_EDGE
        = registerEdgeType( "temporal reference" );
    public final static int RESOURCETYPE_EDGE
        = registerEdgeType( "resource type" );
    public final static int OPERATION_EDGE
        = registerEdgeType( "operation" );
    public final static int CONTRIBUTION_EDGE
        = registerEdgeType( "contribution" );
    public final static int EXECUTIONSTATE_EDGE
        = registerEdgeType( "execution state" );
    public final static int STATE_EDGE = registerEdgeType( "state" );
    public final static int CONSUMPTION_EDGE
        = registerEdgeType( "consumption" );
    public final static int CAPACITY_EDGE = registerEdgeType( "capacity" );
    public final static int SUM_EDGE = registerEdgeType( "sum" );
    public final static int ADD_EDGE = registerEdgeType( "add" );
    public final static int MORE_EDGE = registerEdgeType( "more" );
    public final static int LESS_EDGE = registerEdgeType( "less" );
    public final static int CONSTANT_EDGE = registerEdgeType( "constant" );
    public final static int VARIABLE_EDGE = registerEdgeType( "variable" );
    public final static int ACTIONTYPE_EDGE
        = registerEdgeType( "action type" );
    public final static int PROPORTION_OUTPUT_EDGE = registerEdgeType( "=" );
    public final static int PROPORTION_INPUT_EDGE = registerEdgeType( "*" );
    public final static int EXPONENT_INPUT_EDGE = registerEdgeType( "power" );
    public final static int EXPONENT_OUTPUT_EDGE
        = registerEdgeType( "powerResult" );
    public final static int PRODUCT_EDGE = registerEdgeType( "product" );
    public final static int MULTIPLIER_EDGE = registerEdgeType( "multiplier" );


    // ***** Nodes *****

    /**
     * Type constant.
     */
    public final static boolean TYPE = true;
    /**
     * Instance constant.
     */
    public final static boolean INSTANCE = false;

    /**
     * Number of registered node types.
     */
    private static int currentNodeTypeNumber = 0;
    /**
     * Number of registered object constraint types.
     */
    private static int currentObjectConstraintTypeNumber = 0;
    /**
     * Array of available node types.
     */
    public static Node[] nodeType = new Node[ MAX_NODE_TYPE_NUMBER ];
    /**
     * Array of the nodes' type names.
     */
    public static String[] nodeTypeName = new String[ MAX_NODE_TYPE_NUMBER ];
    /**
     * Array of available node classes.
     */
    private static Class[] nodeTypeClass = new Class[ MAX_NODE_TYPE_NUMBER ];

    /**
     * Constants for pre-registered node types. PAY ATTENTION to sequence:
     * A constraint using other constraints must be placed after them!
     */

    public final static int ANY_NODE = produceNodeTypeID();
    public final static int VARIABLE = produceNodeTypeID();

    public final static int ACTION_OBJECTCONSTRAINT = produceNodeTypeID();
    public final static int ACTIONTASK_OBJECTCONSTRAINT = produceNodeTypeID();
    public final static int CURRENTTIME_OBJECTCONSTRAINT = produceNodeTypeID();
    public final static int NODE_OBJECTCONSTRAINT = produceNodeTypeID();
    public final static int PRECONDITIONTASK_OBJECTCONSTRAINT
        = produceNodeTypeID();
    public final static int STATETASK_OBJECTCONSTRAINT = produceNodeTypeID();

    public final static int ALLDIFFERENT_CONSTRAINT = produceNodeTypeID();
    public final static int DISCRETERESOURCE_CONSTRAINT = produceNodeTypeID();
    public final static int EVEN_CONSTRAINT = produceNodeTypeID();
    public final static int EXPONENTIAL_CONSTRAINT = produceNodeTypeID();
    public final static int INTEGER_CONSTRAINT = produceNodeTypeID();
    public final static int LESSEQUAL_CONSTRAINT = produceNodeTypeID();
    //    Not included to show dynamic loading in example file:
    //    public final static int ODD_CONSTRAINT = produceNodeTypeID();
    public final static int PERMUTATION_CONSTRAINT = produceNodeTypeID();
    public final static int PRODUCT_CONSTRAINT = produceNodeTypeID();
    public final static int PROPORTION_CONSTRAINT = produceNodeTypeID();
    public final static int RANGE_CONSTRAINT = produceNodeTypeID();
    public final static int STATERESOURCE_CONSTRAINT = produceNodeTypeID();
    public final static int SUM_CONSTRAINT = produceNodeTypeID();
    public final static int TASK_CONSTRAINT = produceNodeTypeID();
    public final static int TRANSITION_CONSTRAINT = produceNodeTypeID();

    static
    {
        registerNodeType( ANY_NODE , "com.ai_center.dragonbreath.constraintgraph.Node" );
        registerNodeType( VARIABLE , "com.ai_center.dragonbreath.constraintgraph.Variable" );
        registerNodeType( ACTION_OBJECTCONSTRAINT , "com.ai_center.dragonbreath.constraints.ActionObjectConstraint" );
        registerNodeType( ACTIONTASK_OBJECTCONSTRAINT , "com.ai_center.dragonbreath.constraints.ActionTaskObjectConstraint" );
        registerNodeType( CURRENTTIME_OBJECTCONSTRAINT , "com.ai_center.dragonbreath.constraints.CurrentTimeObjectConstraint" );
        registerNodeType( NODE_OBJECTCONSTRAINT , "com.ai_center.dragonbreath.constraints.NodeObjectConstraint" );
        registerNodeType( PRECONDITIONTASK_OBJECTCONSTRAINT , "com.ai_center.dragonbreath.constraints.PreconditionTaskObjectConstraint" );
        registerNodeType( STATETASK_OBJECTCONSTRAINT , "com.ai_center.dragonbreath.constraints.StateTaskObjectConstraint" );
        registerNodeType( ALLDIFFERENT_CONSTRAINT , "com.ai_center.dragonbreath.constraints.AllDifferentConstraint" );
        registerNodeType( DISCRETERESOURCE_CONSTRAINT , "com.ai_center.dragonbreath.constraints.DiscreteResourceConstraint" );
        registerNodeType( EVEN_CONSTRAINT , "com.ai_center.dragonbreath.constraints.EvenConstraint" );
        registerNodeType( EXPONENTIAL_CONSTRAINT , "com.ai_center.dragonbreath.constraints.ExponentialConstraint" );
        registerNodeType( INTEGER_CONSTRAINT , "com.ai_center.dragonbreath.constraints.IntegerConstraint" );
        registerNodeType( LESSEQUAL_CONSTRAINT , "com.ai_center.dragonbreath.constraints.LessEqualConstraint" );
        //    Not included to show dynamic loading in example file:
        //    registerNodeType( ODD_CONSTRAINT , "com.ai_center.dragonbreath.constraints.OddConstraint" );
        registerNodeType( PERMUTATION_CONSTRAINT , "com.ai_center.dragonbreath.constraints.PermutationConstraint" );
        registerNodeType( PRODUCT_CONSTRAINT , "com.ai_center.dragonbreath.constraints.ProductConstraint" );
        registerNodeType( PROPORTION_CONSTRAINT , "com.ai_center.dragonbreath.constraints.ProportionConstraint" );
        registerNodeType( RANGE_CONSTRAINT , "com.ai_center.dragonbreath.constraints.RangeConstraint" );
        registerNodeType( STATERESOURCE_CONSTRAINT , "com.ai_center.dragonbreath.constraints.StateResourceConstraint" );
        registerNodeType( SUM_CONSTRAINT , "com.ai_center.dragonbreath.constraints.SumConstraint" );
        registerNodeType( TASK_CONSTRAINT , "com.ai_center.dragonbreath.constraints.TaskConstraint" );
        registerNodeType( TRANSITION_CONSTRAINT , "com.ai_center.dragonbreath.constraints.TransitionConstraint" );
    }


    // ***** Breakpoints *****

    /**
     * Number of registered breapoints.
     */
    private static int currentBreakpointNumber = 0;
    /**
     * Array of registered breakpoints.
     */
    public static String[] breakpoints = new String[ MAX_BREAKPOINT_NUMBER ];

    public final static int IMPROVEMENT_BREAKPOINT
        = registerBreakpoint( "One Improvement Step" );
    public final static int SATISFACTION_BREAKPOINT
        = registerBreakpoint( "New Satisfactory Solution Found" );
    public final static int OPTIMUM_BREAKPOINT
        = registerBreakpoint( "New Optimal Solution Found" );


    // ***** Cost Mappings *****

    /**
     * Array of available cost mapping classes.
     */
    public static Class[] costMappingClass
        = new Class[ MAX_COST_MAPPING_NUMBER ];
    /**
     * Array of cost mapping class names.
     */
    public static String[] costMappingName
        = new String[ MAX_COST_MAPPING_NUMBER ];
     /**
     * Number of currently registered cost mapping classes.
     */
    private static int currentCostMappingNumber = 0;

    /**
     * Constants for pre-registered cost mappings.
     */
    public final static int FACTOR_COST_MAPPING = registerCostMapping( "com.ai_center.dragonbreath.costs.FactorCostMapping" , "Factorized Cost Mapping" );
    public final static int FACTOR_BINARY_COST_MAPPING = registerCostMapping( "com.ai_center.dragonbreath.costs.FactorBinaryCostMapping" , "Factorized Binarized Cost Mapping" );

     /**
     * Number of pre-registered cost mapping classes.
     */
    private static int initialCostMappingNumber = currentCostMappingNumber;

    /**
     * Type constants for cost mapping options
     */
    public final static int TYPE_POSITIVE_BIGINTEGER = 0;
    public final static int TYPE_POSITIVE_BIGDECIMAL = 1;
    /**
     * Descriptions of the type constants
     */
    public static String[] typeDescription = new String[ 2 ];
     static
    {
        typeDescription[ 0 ] = new String( "positive integer" );
        typeDescription[ 1 ] = new String( "positive decimal" );
    }


    // ***** Graph repair *****

    /**
     * The probability that the graphRepairer uses a node that is already
     * part of CSP-Graph and does not create a new one.
     */
    public final static double PROBABILITY_USE_EXISTING_NODE = 0.2;

    /**
     * The probability that the graphRepairer uses a node to repair a
     * constraint instance that was claimed by an other instance.
     */
    public final static double PROBABILITY_USE_USED_NODE = 0.1;

    // ***** Default Values for Structural Constraint Instances *****

    /**
     * The probability that a StructuralConstraintInstance destroys its
     * docking graph instead of repairing the testGraph.
     */
    public final static double PROBABILITY_REMOVE_DOCKING = 0.01;

    /**
     * The probability that - when a graphpart is destroyed during repair -
     * a node is removed and not an edge. Usually, it is better to remove
     * only edges because this changes less of the graph. However, we do
     * not want to exclude any possible graph, so a small probability that
     * a node is removed instead of an edge is retained.
     */
    public final static double PROBABILITY_REMOVE_NODE = 0.2;

    /**
     * The time that one repairStep for an instance is allowed to take.
     * After this time, the repair is interrupted and will be continued
     * in next search step.
     */
    public final static long MILLISECONDS_FOR_REPAIR = 5000;

    /**
     * If the testGraph could not be repaired in one step, this number
     * provides the probability of choosing the same testGraph again in the
     * next repair step.
     */
    public final static double PROBABILITY_KEEP_ON_REPAIRING_LAST = 0.5;

    /**
     * Modes of choosing a graphRepair node. A mode is set by calling
     * the function StructuralConstraintManagement.setNodeChooseMode().
     */
    public final static int ALLWAYS_CREATE = 0;
    public final static int NEVER_CREATE = 1;
    public final static int HEURISTIC_CREATE = 2;


    // ************************
    // ***** CONSTRUCTORS *****
    // ************************


    /**
     * Registry can only be used as a static class.
     */
    private Registry() {}


    // *******************
    // ***** METHODS *****
    // *******************


    // ***** Registration *****


    /**
     * @param string
     * @return
     */
    private static int registerBreakpoint( String name )
    {
        if ( currentBreakpointNumber == MAX_BREAKPOINT_NUMBER )
        {
            throw new Error( "No more breakpoint IDs available" );
        }
        breakpoints[ currentBreakpointNumber ] = name;
        return currentBreakpointNumber++ ;
    }


    /**
     * Registers a new edge type and returns the ID for this type.
     * @param name  The name for the new edge type.
     */
    public static int registerEdgeType( String name )
    {
        if ( currentEdgeTypeNumber == MAX_EDGE_TYPE_NUMBER )
        {
            throw new Error( "No more edge type IDs available" );
        }
        edgeTypeName[ currentEdgeTypeNumber ] = name;
        return currentEdgeTypeNumber++ ;
    }


    /**
     * Returns the number of registered edge types.
     */
    public static int edgeTypeNumber()
    {
        return currentEdgeTypeNumber;
    }


    /**
     * Reserves a node type ID and returns it.
     */
    public static int produceNodeTypeID()
    {
        if ( currentNodeTypeNumber == MAX_NODE_TYPE_NUMBER )
        {
            throw new Error( "No more node type IDs available" );
        }
        nodeTypeClass[ currentNodeTypeNumber ] = null;
        nodeType[ currentNodeTypeNumber ] = null;
        nodeTypeName[ currentNodeTypeNumber ] = null;
        return currentNodeTypeNumber++ ;
    }


    /**
     * Registers a new node type. Use produceNodeTypeID() to produce
     * a valid ID before.
     * @param id  The node type ID that is to be used.
     * @param classString  The node class' name path.
     */
    public static void registerNodeType( int id , String classString )
    {
        Class c = null;
        Node node  = null;

        if ( ( id < 0 ) || ( id >= currentNodeTypeNumber ) )
        {
            throw new Error( "Invalid node type ID" );
        }
        if ( nodeTypeClass[ id ] != null )
        {
            throw new Error( "Node type ID has already been used" );
        }

        // Creation of an instance

        try
        {
            c = Class.forName( classString );
            node = (Node)c.newInstance();
        }
        catch( Exception e )
        {
            e.printStackTrace();
        }

        // Instantiation of the instance

        if ( node instanceof ConventionalConstraint )
        {
            ( (ConventionalConstraint)node ).initialize( TYPE , null );
        }
        node.typeID = id;

        // Registration

        nodeTypeClass[ id ] = c;
        nodeType[ id ] = node;
        nodeTypeName[ id ] = node.getTypeName();

        if ( node instanceof ObjectConstraint )
        {
            currentObjectConstraintTypeNumber++ ;
        }
    }


    /**
     * Returns the number of registered node types.
     */
    public static int nodeTypeNumber()
    {
        return currentNodeTypeNumber;
    }


    /**
     * Returns the number of registered object constraint types.
     */
    public static int objectConstrantTypeNumber()
    {
        return currentObjectConstraintTypeNumber;
    }


    /**
     * Returns the number of registered conventional constraint types.
     */
    public static int conventionalConstrantTypeNumber()
    {
        return ( currentNodeTypeNumber
            - currentObjectConstraintTypeNumber - 1 );
    }


    /**
     * Reserves a cost mapping ID and registers a new cost mapping class
     * with it.
     * @param classString  The cost mapping class' name path.
     * @param name  The name for the new cost mapping.
     */
    public static int registerCostMapping( String classString , String name )
    {

        if ( currentCostMappingNumber == MAX_COST_MAPPING_NUMBER )
        {
            throw new Error( "No more cost mapping IDs available" );
        }

        int id = currentCostMappingNumber++ ;

        Class c = null;

        // Creation of an instance

        try
        {
            c = Class.forName( classString );
        }
        catch (Exception e )
        {
            e.printStackTrace();
        }

        // Registration

        costMappingClass[ id ] = c;
        costMappingName[ id ] = name;

        return id;
    }


    /**
     * If there is already a costmapping for the given name, the
     * costMappingID is returned, else -1.
     * @param name  The name of the tested costMapping.
     * @return costMappingID or -1
     */
    public static int getCostMappingID( String name )
    {
        for ( int i = 0 ; i < currentCostMappingNumber ; i++ )
        {
            if ( costMappingName[ i ].equals( name ) )
            {
                return i;
            }
        }
        return -1;
    }


    /**
     * Returns the number of registered cost mapping classes.
     */
    public static int getCostMappingNumber()
    {
        return currentCostMappingNumber;
    }


    /**
     * Returns the number of pre-registered cost mapping classes.
     */
    public static int getInitialCostMappingNumber()
    {
        return initialCostMappingNumber;
    }


    // ***** Factory methods *****


    /**
     * Returns a new instance of a constraint.
     * @param nodeTypeID  The constraint's type ID.
     * @param gsc  The related GlobalSearchControl solver object.
     */
    public static ConventionalConstraint createConstraint( int nodeTypeID , GlobalSearchControl gsc )
    {

        ConventionalConstraint constraint = null;
        if ( ( nodeTypeID < currentNodeTypeNumber ) && ( nodeTypeID > 0 ) )
        {
            try
            {
                constraint = (ConventionalConstraint)nodeTypeClass[ nodeTypeID ].newInstance();
            }
            catch( Exception e )
            {
                e.printStackTrace();
            }
        }
        else
        {
            throw new Error( "Requested constraint type is not registered" );
        }
        constraint.typeID = nodeTypeID;
        constraint.initialize( INSTANCE , gsc );
        return constraint;
    }


    /**
     * Returns a new instance of a cost mapping.
     * @param costMappingID  The cost mapping's class ID.
     */
    public static CostMapping createCostMapping( int costMappingID )
    {

        CostMapping cm = null;

        if ( costMappingID < currentCostMappingNumber )
        {
            try
            {
                cm = (CostMapping)costMappingClass[ costMappingID ].newInstance();
            }
            catch( Exception e )
            {
                e.printStackTrace();
            }
        }
        else
        {
            throw new Error( "Requested cost mapping class is not registered" );
        }

        cm.setName( costMappingName[ costMappingID ] );
        cm.setTypeName( costMappingName[ costMappingID ] );
        return cm;
    }


    // ***** Information Retrieval *****


    /**
     * Returns a textual description of a connected node.
     * @param connectedNode  The ConnectedNode to be described.
     */
    public static String getConnectedNodeDescription( ConnectedNode connectedNode )
    {
        String dir1 = ( connectedNode.direction == Registry.INCOMING ) ?
            "<-" : "--";
        String edge = ( connectedNode.edgeID == Registry.PLAIN_EDGE ) ?
            "" : edgeTypeName[ connectedNode.edgeID ];
        String dir2 = ( connectedNode.direction == Registry.INCOMING ) ?
            "--" : "->";
        return dir1 + edge + dir2
            + "[" + nodeTypeName[ connectedNode.nodeTypeID ] + "]";
    }


    /**
     * Tests if a given node type ID is valid.
     * @param id  The node type ID to be tested.
     */
    public static boolean isValidID( int id )
    {
        return ( ( id >= 0 ) && ( id < currentNodeTypeNumber )
            && ( nodeTypeClass[ id ] != null ) );
    }


    /**
     * Tests if a given ID refers to a constraint type.
     * @param id  The ID to be tested.
     */
    public static boolean isConstraint( int id )
    {
        if ( isValidID( id ) )
        {
            return ( nodeType[ id ] instanceof ConventionalConstraint );
        }
        else
        {
            return false;
        }
    }


    /**
     * Tests if the given node type ID refers to a type that is an object
     * constraint.
     * @param id  The ID to be tested.
     */
    public static boolean isObjectConstraint( int id )
    {
        if ( isValidID( id ) )
        {
            return ( nodeType[ id ] instanceof ObjectConstraint );
        }
        else
        {
            return false;
        }
    }


    /**
     * Tests if a given node type ID refers to a variable type.
     * @param nt NodeType
     */
    public static boolean isVariable( int id )
    {
        if ( isValidID( id ) )
        {
            return ( nodeType[ id ] instanceof Variable );
        }
        else
        {
            return false;
        }
    }


    /**
     * Return the node type ID for a given node type's name.
     * @return The typeID or -1 if name was not found.
     */
    public static int getNodeTypeID( String name )
    {
        for ( int i = 0 ; i < currentNodeTypeNumber ; i++ )
        {
            try
            {
                if ( nodeTypeName[ i ].compareTo( name ) == 0 )
                {
                    return i;
                }
            }
            catch ( java.lang.NullPointerException e )
            {
                return -1;
            }
        }
        return -1;
    }


    /**
     * Return the node type ID for a given edge type's name.
     * @return The edge type's ID or -1 if the name was not found.
     */
    public static int getEdgeTypeID( String name )
    {
        for ( int i = 0 ; i < currentEdgeTypeNumber ; i++ )
        {
            if ( edgeTypeName[ i ].compareTo( name ) == 0 ) {
                return i;
            }
        }
        return -1;
    }


    /**
     * Returns the constraint type for a given node.
     * @param c  The node of which the type is searched.
     * @return  The node that represents the searched type.
     */
    public static Node getNodeType( Node c )
    {
        return nodeType[ c.typeID ];
    }


}
