
package com.ai_center.dragonbreath.constraintgraph;

/** 
 * Connection information of constraints and variables.
 *
 * @author Alexander Nareyek
 * @version $Revision: 2.9 $
 * 
 * @see dragonbreath
 */


import java.util.*;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.EventListenerList;

import com.ai_center.dragonbreath.utilities.SetHashtable;
import com.ai_center.dragonbreath.Registry;


public class Connections {

    // **********************
    // ***** ATTRIBUTES *****
    // **********************

  
    /** 
     * SetHashtable from node type to hashset of connected nodes.
     */
    private SetHashtable nodeTypeSHT = new SetHashtable();

    /**
     * The node this connections belongs to.
     */
    public Node myNode;
    
    /** 
     * Search criteria in case of an enumeration.
     */
    private boolean isCriterionAll;
    private boolean isCriterionNode;
    private boolean isCriterionNodeType;
    private boolean isCriterionEdgeID;
    private boolean isCriterionDirection;
    private Node criterionNode;
    private Node criterionNodeType;
    private int criterionEdgeID;
    private boolean criterionDirection;

    /** 
     * Iterator over nodeTypeSHT keys in case of an enumeration.
     */
    private Iterator setHashtableIterator;

    /** 
     * Iterator over currentConnectedNodes hashset in case of an enumeration.
     */
    private Iterator connectedNodeIterator;
    
    /** A list of EventListeners, which get informed about changes. */
    private EventListenerList listenerList = null;



    // ************************
    // ***** CONSTRUCTORS *****
    // ************************

    /**
     * Constructs a Connections instance for given a node.
     * @param node  The node the connections are related to.
     */
    public Connections(Node n) {
        myNode = n;
    }
    

    // *******************
    // ***** METHODS *****
    // *******************
    
    /** 
     * Print function for the SetHashtable's content.
     */
    public String toString() {
	return nodeTypeSHT.toString();
    }

    /**
     * Initializes the Connections instance by copying the node and 
     * nodeTypeSHT of another Connections instance. Can be used for 
     * copy creation to enable parallel access.
     * @param connections  The Connections instance to be copied from.
     */
    public void initialize(Connections connections) {
        myNode = connections.myNode;
	nodeTypeSHT.clear();
        nodeTypeSHT.addAll(connections.nodeTypeSHT);
    }

    /** 
     * Adds a connection to a node.
     * @param node  The connected node.
     * @param nodeType  Type of the connected node.
     * @param edgeID  Type of the edge.
     * @param direction  Direction of the edge.
     */
    public void add(Node node, Node nodeType, int edgeID, 
		    boolean direction) {
	ConnectedNode connectedNode = new ConnectedNode(node, node.typeID,
							edgeID, direction);
	nodeTypeSHT.add(nodeType, connectedNode);
        connectedNode.other = 
	    new ConnectedNode(myNode, myNode.typeID, edgeID, !direction);
        connectedNode.other.other = connectedNode;
        node.connections.add(connectedNode.other);
        
        
        fireChange(connectedNode); // inform changelistener
    }

    /** 
     * Adds a connection to a node. 
     * Called by add(Node n, Node nodeType, int edgeID, 
		    boolean direction)
     * @param cn the Connected node added.
     */
    public void add(ConnectedNode cn){
        nodeTypeSHT.add(Registry.nodeType[cn.node.typeID], cn);
        if ( cn.other == null) {
            cn.other = new ConnectedNode(myNode, myNode.typeID, cn.edgeID, !cn.direction);
            cn.other.other = cn;
            cn.node.connections.add(cn.other);
        } else {
            if (! cn.node.connections.contains(cn.other)) {
                cn.node.connections.add(cn.other);
            }
        }
        
        fireChange(cn); // inform changeListeners
    }
    
    /** 
     * Removes a connection to a node.
     * @param node  The connected node.
     * @param nodeType  Type of the connected node.
     * @param edgeID  Type of the edge.
     * @param direction  Direction of the edge.
     * @return  the removed connectedNode
     */
    public ConnectedNode remove(Node node, Node nodeType, int edgeID, 
				boolean direction) {
	Object retrievedObject;

        ConnectedNode result = null;
	synchronized(nodeTypeSHT) {
	    if ((retrievedObject = nodeTypeSHT.get(nodeType)) == null) {
		throw new Error("A " + nodeType + " is not connected");
	    } else {
		if (retrievedObject instanceof HashSet) {
		    // List WRT the key object
		    Set connectedNodes = (Set) retrievedObject;
		    boolean found = false;
		    ConnectedNode connectedNode = null;
		
		    synchronized (connectedNodes) {
			Iterator connectedNodeIterator = 
			    connectedNodes.iterator();
		    
			while (!found && connectedNodeIterator.hasNext()) {
			    connectedNode = 
				(ConnectedNode) connectedNodeIterator.next();
			    found = (connectedNode.node == node && 
				     connectedNode.edgeID == edgeID && 
				     connectedNode.direction == direction);
			}
		    }
		    if (!found) {
			throw new Error("Connection not included");
		    }
		    synchronized (connectedNodes) {
                        result = connectedNode;
                        nodeTypeSHT.remove(nodeType, connectedNode);
                        connectedNode.node.connections.remove(connectedNode.other);
                        
		    }

		} else {
		    // Only one element WRT the key object
		    ConnectedNode connectedNode = 
			(ConnectedNode) retrievedObject;
		    if (Registry.SAFE_MODE && connectedNode.node == node && 
			connectedNode.edgeID == edgeID && 
			connectedNode.direction == direction) {
                       result = connectedNode;
		       nodeTypeSHT.remove(nodeType, connectedNode);
                       connectedNode.node.connections.remove(connectedNode.other);
                       
		    }
		}
	    }
	}
		fireChange(this);
        return result;
    }
    
    /** 
     * Removes a connection to a node.
     * @param cn  The node to be removed.
     */
    public void remove(ConnectedNode cn) {
	nodeTypeSHT.remove(Registry.nodeType[cn.node.typeID], cn);
	fireChange(this);
    }

    /**
     * Checks whether the connections includes the given connectioninstance.
     * @param cn  The ConnectedNode that should be contained.
     */
    public boolean contains(ConnectedNode cn) {
	Object retrievedObject;
        synchronized (nodeTypeSHT) {
	    if ((retrievedObject = 
		 nodeTypeSHT.get(Registry.nodeType[cn.nodeTypeID])) == null) {
		return false;
	    } else {
		if (retrievedObject instanceof HashSet) {
		    // List WRT the key object
		    Set connectedNodes = (Set) retrievedObject;
		    boolean found = false;
		    ConnectedNode connectedNode = null;
		    
		    synchronized (connectedNodes) {
			return connectedNodes.contains(cn);
		    }
		} else {
		    // Only one element WRT the key object
		    ConnectedNode connectedNode = 
			(ConnectedNode) retrievedObject;
		    return (cn == connectedNode);
		}
	    }
	}
    }
    
    /** 
     * Checks whether the connections includes the given connection.
     * @param node  The potentially connected node.
     * @param nodeType  Type of the potentially connected node.
     * @param edgeID  Type of the edge.
     * @param direction  Direction of the edge.
     */
    public boolean contains(Node node, Node nodeType, int edgeID, 
			    boolean direction) {
	Object retrievedObject;
	synchronized (nodeTypeSHT) {
	    if ((retrievedObject = nodeTypeSHT.get(nodeType)) == null) {
		throw new Error("A " + nodeType + " is not connected");
	    } else {
		if (retrievedObject instanceof HashSet) {
		    // List WRT the key object
		    Set connectedNodes = (Set) retrievedObject;
		    boolean found = false;
		    ConnectedNode connectedNode = null;
		    
		    synchronized (connectedNodes) {
			Iterator connectedNodeIterator = 
			    connectedNodes.iterator();
		    
			while (!found && connectedNodeIterator.hasNext()) {
			    connectedNode = 
				(ConnectedNode) connectedNodeIterator.next();
			    found = (connectedNode.node == node && 
				     connectedNode.edgeID == edgeID && 
				     connectedNode.direction == direction);
			}
		    }
		    return found;

		} else {
		    // Only one element WRT the key object
		    ConnectedNode connectedNode = 
			(ConnectedNode) retrievedObject;
		    return (connectedNode.node == node && 
			    connectedNode.edgeID == edgeID && 
			    connectedNode.direction == direction);
		}
	    }
	}
    }

    /**
     * Initializes further retrieval for ConnectedNodes that match certain 
     * search criteria.
     * Needs to be synchronized with further retrieval process.
     * @param criterionNode  Indicates if the node is a search criterion.
     * @param criterionNodeType  Indicates if the node type is a search criterion.
     * @param criterionEdgeID  Indicates if the edge ID is a search criterion.
     * @param criterionDirection  Indicates if the direction is a search criterion.
     * @param node  A potentially connected node.
     * @param nodeType  Type of a potentially connected node.
     * @param edgeID  Type of the edge.
     * @param direction  Direction of the edge.
     */
    public void initializeEnumeration(boolean isCriterionNode, 
				      boolean isCriterionNodeType, 
				      boolean isCriterionEdgeID, 
				      boolean isCriterionDirection, 
				      Node node, Node nodeType, 
				      int edgeID, boolean direction) {
	this.isCriterionNode = isCriterionNode;
	this.isCriterionNodeType = isCriterionNodeType;
	this.isCriterionEdgeID = isCriterionEdgeID;
	this.isCriterionDirection = isCriterionDirection;
	this.isCriterionAll = !(isCriterionNode || isCriterionNodeType || 
				isCriterionEdgeID || isCriterionDirection);

	criterionNode = node;
	criterionNodeType = nodeType;
	criterionEdgeID = edgeID;
	criterionDirection = direction;

	if (isCriterionNode) {
	    this.isCriterionNodeType = true;
	    criterionNodeType = Registry.nodeType[node.typeID];
	}

	setHashtableIterator = nodeTypeSHT.iterator();
	connectedNodeIterator = null;
    }

    /**
     * Initializes further retrieval for all ConnectedNodes.
     * Needs to be synchronized with further retrieval process.
     */
    public void initializeEnumerationAll() {
	isCriterionNode = false;
	isCriterionNodeType = false;
	isCriterionEdgeID = false;
	isCriterionDirection = false;
	isCriterionAll = true;
	
	setHashtableIterator = nodeTypeSHT.iterator();
	connectedNodeIterator = null;
    }
    
    /**
     * Checks whether the search criteria are fulfilled for a connected node.
     * @param connectedNode  The connectedNode under consideration.
     */
    private boolean criterionMatches(ConnectedNode connectedNode) {
	 boolean ok = true;
	 if (isCriterionNode && 
	     (connectedNode.node != criterionNode)) {
	     ok = false;
	 }
	 if (isCriterionEdgeID && 
	     (connectedNode.edgeID != criterionEdgeID)) {
	     ok = false;
	 }
	 if (isCriterionDirection && 
	     (connectedNode.direction != criterionDirection)) {
	     ok = false;
	 }
	 return ok;
    }

    /**
     * Returns the next ConnectedNode that matches certain search
     * criteria; see initializeEnumeration().
     * Needs to be synchronized with further retrieval process.
     */
    public ConnectedNode getNext() {
	Object retrievedObject = null;
	ConnectedNode connectedNode;
	if (setHashtableIterator != null) {
	    if (connectedNodeIterator == null) {
		// Starting to enumerate the next key's list
		
		// Retrieving reference object list
		if (isCriterionNodeType) {
		    if ((retrievedObject = nodeTypeSHT.get(criterionNodeType)) 
			== null) {
			return null;
		    }
		} else {
		    if (setHashtableIterator.hasNext()) {
                        try {
			retrievedObject = 
			    nodeTypeSHT.get(setHashtableIterator.next());
                        } catch ( Exception e) {
                            int a =0;
                        }
		    } else {
			return null;
		    }
		}
		
		// Retrieving reference object
		if (retrievedObject instanceof HashSet) {
		    connectedNodeIterator = 
			((HashSet) retrievedObject).iterator();
		    connectedNode = 
			(ConnectedNode) connectedNodeIterator.next();
		} else {
		    connectedNode = (ConnectedNode) retrievedObject;
		    if (isCriterionNodeType) {
			setHashtableIterator = null;
		    }
		}
		
	    } else {
		// Within an enumeration of a key's list
		
		// Checking if the next key must be selected
		if (! connectedNodeIterator.hasNext()) {
		    if (isCriterionNodeType) {
			setHashtableIterator = null;
			return null;
		    }
		    connectedNodeIterator = null;
		    return getNext();
		} else {
		    connectedNode = // <== here happened a nullPointerException !! 
                                    //but in the if statement the connectedNodeIterator must have been still ok!
			(ConnectedNode) connectedNodeIterator.next();
		}

	    }
            if (isCriterionNodeType) {
                if ( connectedNode.node.typeID != criterionNodeType.typeID ){
                    throw new Error("Wrong node type in enumeration.");
                }
            }
	    // Checking reference object
	    if (!isCriterionAll && 
		!criterionMatches((connectedNode))) {
		return getNext();
	    } else {
		return connectedNode;
	    }
	    
	} else {
	    return null;
	}
    }

    /**
     * Retrieves a unique connected variable. Returns "null" if no such
     * variable was connected, and if multiple matching variables were 
     * connected, a random one of them.
     * @param edgeID  Type of the edge to the variable.
     */
    public Variable getSpecificVariable(int edgeID) {

	ConnectedNode connectedNode;

	// Retrieving connected variables

	Object retrievedObject = 
	    nodeTypeSHT.get(Registry.nodeType[Registry.VARIABLE]);

	if (retrievedObject != null) {
	    
	    if (retrievedObject instanceof HashSet) {
		
		// Checking connected variables
		
		Iterator it = ((HashSet) retrievedObject).iterator();
		while (it.hasNext()) {
		    connectedNode = (ConnectedNode) it.next();
		    if (connectedNode.edgeID == edgeID) {
			return (Variable) connectedNode.node;
		    }
		}
		return null;
		
	    } else {
		
		// Only one variable was connected
		
		connectedNode = (ConnectedNode) retrievedObject;
		if (connectedNode.edgeID == edgeID) {
		    return (Variable) connectedNode.node;
		} else {
		    return null;
		}
		
	    }
	    
	} else {
	    
	    // No variable was connected
	    
	    return null;
	    
	}

    }

    /**
     * Retrieves a unique connected constraint. Returns "null" if no such
     * constraint was connected, and if multiple matching constraints were 
     * connected, a random one of them.
     * @param typeID  Type of the constraint.
     * @param edgeID  Type of the edge to the constraint.
     */
    public ConventionalConstraint getSpecificConstraint(int typeID, 
							int edgeID) {

	ConnectedNode connectedNode;

	// Retrieving specific connected constraints
	
	Object retrievedObject = nodeTypeSHT.get(Registry.nodeType[typeID]);
	
	if (retrievedObject != null) {
	    
	    if (retrievedObject instanceof HashSet) {
		
		// Checking specific connected constraints
		
		Iterator it = ((HashSet) retrievedObject).iterator();
		while (it.hasNext()) {
		    connectedNode = (ConnectedNode) it.next();
		    if (connectedNode.edgeID == edgeID) {
			return (ConventionalConstraint) connectedNode.node;
		    }
		}
		return null;
		
	    } else {
		
		// Only one specific constraint was connected
		
		connectedNode = (ConnectedNode) retrievedObject;
		if (connectedNode.edgeID == edgeID) {
		    return (ConventionalConstraint) connectedNode.node;
		} else {
		    return null;
		}
		
	    }
	    
	} else {
	    
	    // No specific constraint was connected
	    
	    return null;

	}

    }

    /**
     * Retrieves all connected nodes of a specific type. Returns null if no
     * corresponding node is connected, a ConnectedNode object is exactly
     * one node of this type is connected, and a HashSet of ConnectedNodes
     * for multiple matches.
     * @param typeID  Type of the nodes to be retrieved.
     */
    public Object getSpecificNodes(int typeID) {
	return nodeTypeSHT.get(Registry.nodeType[typeID]);
    }

    /** 
     * Returns whether at least one node is connected.
     */
    public boolean isEmpty() {
	synchronized (nodeTypeSHT) {
	    return (nodeTypeSHT.size() == 0);
	}
    }

    /** 
     * Returns the number of connected nodes.
     */
    public int number() {
	return nodeTypeSHT.size();
    }

    /**
     * Returns the number of connected variables.
     */
    public int numberOfVariables() {

        Object variables =
            nodeTypeSHT.get(Registry.nodeType[Registry.VARIABLE]);

        if (variables != null) {
            if (variables instanceof HashSet) {
		return ((HashSet) variables).size();
	    } else {
		return 1;
	    }
	} else {
	    return 0;
	}

    }
    
    
    /** Adds a ChangeListener to observe this object.<p>
    
    @param l The ChangeListener to add.
    */
   public void addChangeListener(ChangeListener l)
   {
       if (listenerList == null)
       {
           listenerList = new EventListenerList();
       }
       listenerList.add(ChangeListener.class, l);
   }

   /** Removes a ChangeListener.<p>
    
    @param l The ChangeListener to be removed.
    */
   public void removeChangeListener(ChangeListener l)
   {
       if (listenerList != null)
       {
           listenerList.remove(ChangeListener.class, l);
       }
   }

   /** Informs all ChangeListeners about a change.
     @param changedObject The changed object.
    */
   protected void fireChange(Object changedObject)
   {
       if (listenerList != null)
       {

           ChangeEvent simpleChangeEvent = new ChangeEvent(changedObject);


           Object[] listeners = listenerList.getListenerList();
           ChangeListener listener = null;

           for (int i = listeners.length - 2; i >= 0; i -= 2)
           {
               if (listeners[i] == ChangeListener.class)
               {

                   ((ChangeListener) listeners[i + 1])
                           .stateChanged(simpleChangeEvent);

               }

           }
       }
   }

}
