
package com.ai_center.dragonbreath.constraintgraph;

/** 
 * The superclass of all conventional constraints.
 *
 * @author Alexander Nareyek
 * @version $Revision: 2.4 $
 */


import java.util.Vector;

import com.ai_center.dragonbreath.Constraint;
import com.ai_center.dragonbreath.GlobalSearchControl;
import com.ai_center.dragonbreath.Registry;
import com.ai_center.dragonbreath.constraints.IntegerConstraint;
import com.ai_center.dragonbreath.constraints.RangeConstraint;
import com.ai_center.dragonbreath.costs.CostFunctionReference;



public abstract class ConventionalConstraint 
    extends Node implements Constraint {

    // **********************
    // ***** ATTRIBUTES *****
    // **********************

	 /** 
     * The constraint's subjective costs.
     */
    public long costs[];

    /** 
     * The constraint's cost function information;
     */
    public CostFunctionReference[] costFunctionReference;


    // *******************
    // ***** METHODS *****
    // *******************

    public int getTypeID() {
	return typeID;
    }

    public long getCosts(int costID) {
	return costs[costID];
    }

    public void setCostFunctionReference(int costID,
					 CostFunctionReference 
					 costFunctionReference) {
        
	this.costFunctionReference[costID] = costFunctionReference;
    }

    public CostFunctionReference getCostFunctionReference(int costID) {
	return costFunctionReference[costID];
    }

    /** 
     * Initializes the base and extension graphs.
     */
    public abstract void create_graphs();

    /** 
     * Returns the constraint type's base graph.
     */
    public abstract Embedding getBaseGraph();

    /** 
     * Returns the number of extension graph.
     */
    public abstract int getExtensionGraphNumber();

    /** 
     * Returns an extension graph.
     * @param graphID  The ID of the extension graph.
     */
    public abstract Embedding getExtensionGraph(int graphID);

    /** 
     * Returns the maximal number of connections allows.
     * This number can be derived from the SCSP's maximal 
     * embedding graph for the constraint. If there is
     * no maximum for the constraint, Integer.MAX_VALUE
     * is returned.
     */
    public abstract int getMaximalConnectionNumber();

    /**
     * Returns all Extensions that contain a ConnectedNode as cn.
     * @param cn  The connection that should be contained in the Extension
     * @return a Vector with the Extensions.
     */
    public Vector getMatchingExtensions( ConnectedNode cn) {
         return getMatchingExtensions(cn.nodeTypeID, cn.edgeID, cn.direction);
    }
    
    /**
     * Returns all Extensions that contain a ConnectedNode as cn.
     * @param cn  The connection that should be contained in the Extension
     * @return a Vector with the Extensions.
     */
    public Vector getMatchingExtensions( int nodeTypeID, int edgeID, boolean direction) {
        Vector result = new Vector();
        int exGrNr = getExtensionGraphNumber();
         for (int i = 0; i< exGrNr; i++) {
            Embedding e = getExtensionGraph(i);
            if ( e.containsConnectedNodeType(nodeTypeID, edgeID, direction) ) {
                result.add(e);
            }
         }
         return result;
    }
    
    public void initializeInternalStructures() {}

    /** 
     * Method to update the constraint's costs and structures when 
     * an included variable's value has changed.
     * @param oldValue  The old value of the variable.
     * @param first  The link to the changed variable.
     * @param last  The link to the last (object) constraint that
     *              propagated the change.
     */
    public abstract void update(double oldValue, ConnectionReference first,
				ConnectionReference last);

    /** 
     * Connects a node to a constraint.
     * @param first  The link to the connected node.
     * @param last  The link to the last (object) constraint that
     *              propagated the connection.
     */
    public abstract void connectNode(ConnectionReference first,
				     ConnectionReference last);

    /** 
     * Disconnects a node from a constraint.
     * @param first  The link to the node to be disconnected.
     * @param last  The link to the last (object) constraint that
     *              propagated the disconnection.
     */
    public abstract void disconnectNode(ConnectionReference first,
					ConnectionReference last);

    /** 
     * Initializes a new constraint.
     * @param isType  Indicates whether the constraint is to be a type or not.
     * @param gsc  The GlobalSearchControl solver object.
     */
    public void initialize(boolean isType, GlobalSearchControl gsc) {
	this.isType = isType;
	this.gsc = gsc;
	if (isType == Registry.TYPE) {
	    create_graphs();
	} else {
	    connections = new Connections(this);

	    if (!(this instanceof ObjectConstraint) 
		&& !(this instanceof IntegerConstraint) 
		&& !(this instanceof RangeConstraint)) {
		int numberOfCosts = getNumberOfCosts();
		costs = new long[numberOfCosts];
		costFunctionReference = 
		    new CostFunctionReference[numberOfCosts];
		for (int i = 0; i < numberOfCosts; i++) {
		    costs[i] = 0;
		    costFunctionReference[i] = null;
		}
	    }

	    initializeInternalStructures();
	}
    }

    /** 
     * Initializes visual tracking. Override this function in
     * subclasses if necessary.
     */
    public void createGraphics() {
	graphObject = new ConventionalConstraintGraphics(this, gsc);
    }

    /** 
     * Checks whether a constraint's embedding reduced by a 
     * reduction keeps the constraint's base graph. Also returns
     * false if the reduction does not match the constraint's 
     * embedding.
     * @param reduction  The reduction embedding (type).
     */
    public boolean keepsBaseGraph(Embedding reduction) {

	Embedding baseGraph = getBaseGraph();

	int toBeFound = reduction.connectedNodeNumber + 
	    baseGraph.connectedNodeNumber;
	int available = connections.number();
	if (available < toBeFound) {
	    return false;
	}
	int toBeFoundBG = baseGraph.connectedNodeNumber;

	int i, i2, i3;
	boolean[] usedBaseGraph = new boolean[baseGraph.connectedNodeNumber];
	for (i = 0; i < baseGraph.connectedNodeNumber; i++) {
	    usedBaseGraph[i] = false;
	}
	boolean[] usedReduction = new boolean[reduction.connectedNodeNumber];
	for (i = 0; i < reduction.connectedNodeNumber; i++) {
	    usedReduction[i] = false;
	}

	ConnectedNode connectedNodeC;
	ConnectedNode connectedNodeBG;
	boolean found;

	synchronized(connections) {
	    connections.initializeEnumerationAll();
	    while (toBeFound > 0 && 
		   ((connectedNodeC = connections.getNext()) != null)) {
		
		found = false;

		if (toBeFoundBG > 0) {
		    for (i = 0; i < baseGraph.connectedNodeNumber && !found; i++) {
			if (!usedBaseGraph[i]) {
			    connectedNodeBG = baseGraph.connectedNode[i];
			    if ((connectedNodeBG.nodeTypeID == 
				 connectedNodeC.nodeTypeID) && 
				(connectedNodeBG.edgeID == 
				 connectedNodeC.edgeID) && 
				(connectedNodeBG.direction == 
				 connectedNodeC.direction)) {
				usedBaseGraph[i] = true;
				found = true;
				toBeFound--;
				toBeFoundBG--;
			    }
			}
		    }
		}

		if (!found) {
		    for (i = 0; i < reduction.connectedNodeNumber && !found; i++) {
			if (!usedReduction[i]) {
			    connectedNodeBG = reduction.connectedNode[i];
			    if ((connectedNodeBG.nodeTypeID == 
				 connectedNodeC.nodeTypeID) && 
				(connectedNodeBG.edgeID == 
				 connectedNodeC.edgeID) && 
				(connectedNodeBG.direction == 
				 connectedNodeC.direction)) {
				usedReduction[i] = true;
				found = true;
				toBeFound--;
			    }
			}
		    }
		}

		available--;
		if (available < toBeFound) {
		    return false;
		}
	    }
	}

	return true;

    }

}

