
package com.ai_center.dragonbreath.constraintgraph;


import java.util.Vector;

import com.ai_center.dragonbreath.*;


/**
 * Structure to represent a constraint's embedding graph,
 * i.e., direct connections of a constraint.
 *
 * @author Alexander Nareyek
 * @version $Revision: 2.3 $
 */
public class Embedding
{


    // **********************
    // ***** ATTRIBUTES *****
    // **********************


    /**
     * Total number of connected nodes.
     */
    public final int connectedNodeNumber;

    /**
     * Number of currently connected nodes.
     */
    private int connectedNodeNumberNow = 0;

    /**
     * Set of connected nodes.
     */
    public final ConnectedNode[] connectedNode;

    /**
     * For each ConnectedNode in connectedNode, the associated values
     * determines whether the given edge may appear with the same node at
     * the end twice.
     */
    public boolean[] useNodeMultipleTimes;

    /**
     * Optional, if the embedding is a basegraph or an extension
     * the constraint this embedding is an extension of.
     */
    public ConventionalConstraint myConstraint;


    // ************************
    // ***** CONSTRUCTORS *****
    // ************************


    /**
     * Initializes the ConnectedNode array.
     * @param connectedNodeNumber  Total number of connected nodes.
     */
    public Embedding( int connectedNodeNumber )
    {
        this.connectedNodeNumber = connectedNodeNumber;
        connectedNode = new ConnectedNode[ connectedNodeNumber ];
        useNodeMultipleTimes = new boolean[ connectedNodeNumber ];
    }


    /**
     * Copies the information given in the example into this Embedding.
     * @param example  The pattern for this Embedding; has to be closed!
     */
    public Embedding( Embedding example )
    {
        this.connectedNodeNumber = example.connectedNodeNumber;
        useNodeMultipleTimes = new boolean[ connectedNodeNumber ];
        connectedNode = new ConnectedNode[ example.connectedNodeNumber ];
        for ( int i = 0 ; i < example.connectedNodeNumber ; i++ )
        {
            connectedNode[ i ]
                = new ConnectedNode( example.connectedNode[ i ] );
            useNodeMultipleTimes[ i ] = example.useNodeMultipleTimes[ i ];
            connectedNodeNumberNow++ ;
        }
    }


    // *******************
    // ***** METHODS *****
    // *******************


    /**
     * Addition of a connected node instance.
     * @param node  The node to be added.
     * @param nodeTypeID  Type of the node to be added.
     * @param edgeID  ID of the edge to be added.
     * @param direction  Direction of the edge.
     * @see Registry
     */
    public void addConnectedNode( Node node , int nodeTypeID , int edgeID ,
        boolean direction )
    {
        connectedNode[ connectedNodeNumberNow ]
            = new ConnectedNode( node , nodeTypeID , edgeID , direction );
        useNodeMultipleTimes[ connectedNodeNumberNow ] = true;
        connectedNodeNumberNow++ ;
    }


    /**
     * Addition of a connected node type.
     * @param nodeTypeID  Type of the node to be added.
     * @param edgeID  ID of the edge to be added.
     * @param direction  Direction of the edge.
     * @see Registry
     */
    public void addConnectedNode( int nodeTypeID , int edgeID ,
        boolean direction )
    {
        connectedNode[ connectedNodeNumberNow ]
            = new ConnectedNode( nodeTypeID , edgeID , direction );
        useNodeMultipleTimes[ connectedNodeNumberNow ] = true;
        connectedNodeNumberNow++ ;
    }


    /**
     * Addition of a ConnectedNode.
     * @param cn  The connectedNode to be added. Has to have a valid
     *     other member.
     */
    public void addConnectedNode( ConnectedNode cn )
    {
        connectedNode[ connectedNodeNumberNow ] = cn;
        useNodeMultipleTimes[ connectedNodeNumberNow ] = true;
        connectedNodeNumberNow++ ;
    }


    /**
     * Checks whether further connections can be added.
     */
    public boolean isClosed()
    {
        return ( connectedNodeNumberNow == connectedNodeNumber );
    }


    /**
     * Returns the number of edges to a special node.
     * @param node  The node to which egdes are counted.
     */
    public int numberOfNodeOccurences( Node node )
    {
        int i;
        int counter = 0;
        for ( i = 0 ; i < connectedNodeNumber ; i++ )
        {
            if ( connectedNode[ i ].node == node )
            {
                counter++ ;
            }
        }
        return counter;
    }


    /**
     * Checks whether an embedding type matches the embedding.
     * @param type  The embedding type.
     */
    public boolean matchesType( Embedding type )
    {

        if ( connectedNodeNumber != type.connectedNodeNumber )
        {
            return false;
        }

        int i, i2;
        boolean found = true;
        boolean[] used = new boolean[ connectedNodeNumber ];
        for ( i = 0 ; i < connectedNodeNumber ; i++ )
        {
            used[ i ] = false;
        }

        for ( i = 0 ; ( i < connectedNodeNumber ) && found ; i++ )
        {
            found = false;
            for ( i2 = 0 ; ( i2 < connectedNodeNumber ) && !found ; i2++ )
            {
                if ( !( used[ i2 ] ) )
                {
                    if ( found = ( ( connectedNode[ i ].nodeTypeID == type.connectedNode[ i2 ].nodeTypeID ) && ( connectedNode[ i ].edgeID == type.connectedNode[ i2 ].edgeID ) && ( connectedNode[ i ].direction == type.connectedNode[ i2 ].direction ) ) )
                    {
                        used[ i2 ] = true;
                    }
                }
            }
        }

        return found;
    }


    /**
     * Returns all the indices of all edges in this embedding that
     * have the same type as the given edge.
     * @return  A new Vector, which may be changed, with all the
     *     ConnectedNodes of the type.
     */
    public Vector allConnectedNodeTypeOccurences( ConnectedNode cn )
    {
        return allConnectedNodeTypeOccurences( cn.nodeTypeID , cn.edgeID ,
            cn.direction );
    }


    /**
     * Returns all the indices of all edges in this embedding that
     * have the given type.
     */
    public Vector allConnectedNodeTypeOccurences( int nodeTypeID , int edgeID ,
        boolean direction )
    {
        Vector result = new Vector();
        ConnectedNode cnInEmbedding;
         // iterate through the edges of the embedding to search for the cn
        for ( int cnNr = 0 ; ( cnNr < connectedNodeNumberNow ) ; cnNr++ )
        {
            cnInEmbedding = connectedNode[ cnNr ];
            if ( ( ( cnInEmbedding.nodeTypeID == nodeTypeID ) || ( nodeTypeID == Registry.ANY_NODE ) ) && ( ( cnInEmbedding.edgeID == edgeID ) || ( edgeID == Registry.ANY_EDGE ) ) && ( ( cnInEmbedding.direction == direction ) || ( edgeID == Registry.ANY_EDGE ) ) )
            {
                result.add( new Integer( cnNr ) );
            }
        }
        return result;
    }


    /**
     * Checks whether the embedding contains a ConnectedNode of the same
     * type as the given connected node. The same type means that direction,
     * edgeID and nodeTypeID are identical.
     * @param cn  The connected node.
     */
    public boolean containsConnectedNodeType( ConnectedNode cn )
    {
        return containsConnectedNodeType( cn.nodeTypeID , cn.edgeID ,
            cn.direction );
    }


    /**
     * Checks whether the embedding contains a ConnectedNode of the same
     * type as the given cn.
     * The same type means direction, edgeID and nodeTypeID are identical.
     * @param nodeTypeID  The type of the connected node.
     * @param edgeID  The type of the edge.
     * @param direction  The direction of the edge.
     * @return  If such an edge was found.
     */
    public boolean containsConnectedNodeType( int nodeTypeID , int edgeID ,
        boolean direction )
    {
        ConnectedNode cnInEmbedding;
         // iterate through the edges of the embedding to search for the cn
        for ( int cnNr = 0 ; ( cnNr < connectedNodeNumberNow ) ; cnNr++ )
        {
            cnInEmbedding = connectedNode[ cnNr ];
            if ( ( ( cnInEmbedding.nodeTypeID == nodeTypeID ) || ( nodeTypeID == Registry.ANY_NODE ) ) && ( ( cnInEmbedding.edgeID == edgeID )|| ( edgeID == Registry.ANY_EDGE ) ) && ( ( cnInEmbedding.direction  == direction ) || ( edgeID == Registry.ANY_EDGE ) ) )
            {
                return true;
            }
        }
        return false;
    }


     /**
     * Checks whether the embedding contains the given ConnectedNode.
     * @param cn  The connectedNode to be searched for.
     */
    public boolean containsConnectedNode( ConnectedNode cn )
    {
        ConnectedNode cnInEmbedding;
         // iterate through the edges of the embedding to search for the cn
        for ( int cnNr = 0 ; ( cnNr < connectedNodeNumberNow ); cnNr++ )
        {
            if ( connectedNode[ cnNr ] == cn )
            {
                return true;
            }
        }
        return false;
    }


    /**
     * When an Embedding was a Type its connections can be filled with
     * real edges by this function. One of the ConnectedNode that has the
     * same type as the given one, and is not yet instanciated (i.e., the
     * node member is null) is replaced by the given cn.
     * @param cn  The Connected Node that should be included in the embedding.
     * @return  Returns whether the operation was successful or not.
     */
    public boolean includeConnectedNode( ConnectedNode cn )
    {
        boolean result = false;
        int foundNr = 0;
        int cnNr;
         // iterate through the edges of the embedding to search for the cn
        for ( cnNr = 0 ; cnNr < connectedNodeNumberNow ; cnNr++ )
        {
            if ( ( connectedNode[ cnNr ].nodeTypeID == cn.nodeTypeID )
                && ( connectedNode[ cnNr ].edgeID == cn.edgeID )
                && ( connectedNode[ cnNr ].direction == cn.direction )
                && ( connectedNode[ cnNr ].node == null ) )
            {
                result = true;
                foundNr = cnNr;
            }
        }
        if ( result )
        {
            // Test whether the cn is used elsewhere too
            for ( ; cnNr < connectedNodeNumberNow ; cnNr++ )
            {
                if ( connectedNode[ cnNr ] == cn )
                {
                    result = false;
                }
            }
            if ( result )
            {
                // cn was not yet used
                connectedNode[ foundNr ] = cn;
            }
        }
        return result;
    }


}
