
package com.ai_center.dragonbreath.constraintgraph;

/** 
 * This is the base class for all object constraints.
 *
 * @author Alexander Nareyek
 * @version $Revision: 2.2 $
 */


import java.util.HashSet;
import java.util.Iterator;

import com.ai_center.dragonbreath.Registry;


public abstract class ObjectConstraint extends ConventionalConstraint {

    // **********************
    // ***** ATTRIBUTES *****
    // **********************

    /**
     * Set of connected nodes of directly connected constraints.
     */
    protected HashSet connectedNodes = new HashSet();

    /** 
     * The reference to be handed to the connected constraints when
     * a variable update must be propagated.
     */
    protected ConnectionReference connectionReferenceUpdate = 
	new ConnectionReference(this);

    /** 
     * The reference to be handed to the connected constraints when
     * a new node connection must be propagated.
     */
    protected ConnectionReference connectionReferenceConnection = 
	new ConnectionReference(this);

    /** 
     * The reference to be handed to the connected constraints when
     * a node disconnection must be propagated.
     */
    protected ConnectionReference connectionReferenceDisconnection = 
	new ConnectionReference(this);

    /** 
     * The variable of the last update propagation.
     */
    protected Variable lastChangeVariable = null;
    /** 
     * The value of the last update propagation.
     */
    protected double lastChangeValue = 0.0;
    /** 
     * The number of integer constraints of the last update 
     * propagation' variable.
     */
    protected double lastChangeIntC = 0;

    /** 
     * The node of the last connection propagation.
     */
    protected Node lastConnectionNode = null;

    /** 
     * The node of the last disconnection propagation.
     */
    protected Node lastDisconnectionNode = null;


    // *******************
    // ***** METHODS *****
    // *******************

    public void initializeCostType(int costID) {
	throw new Error("Invalid function for object constraints");
    }

    public void update(double oldValue, ConnectionReference first, 
		       ConnectionReference last) {
	
	
	
	Variable variable = (Variable) first.node;

	if ((lastChangeVariable != variable) || 
	    (lastChangeValue != variable.value()) || 
	    (lastChangeIntC != variable.integerConstraints)) {

            // Preventing from multiple calls in one update cycle

	    lastChangeVariable = variable;
	    lastChangeValue = variable.value();
	    lastChangeIntC = variable.integerConstraints;

            // Including own connection into reference chain

	    last.next = connectionReferenceUpdate;
	    connectionReferenceUpdate.previous = last;

            // Propagating change

            ConnectedNode connectedNode;
            Iterator it = connectedNodes.iterator();

            while (it.hasNext()) {
                connectedNode = (ConnectedNode) it.next();
		if (connectedNode.node != last.node) {
		    connectionReferenceUpdate.edgeID = 
			connectedNode.edgeID;
		    connectionReferenceUpdate.direction = 
			connectedNode.direction;
		    ((ConventionalConstraint) connectedNode.node).update(oldValue, first, connectionReferenceUpdate);
		}
	    }
	}
	fireChange(this);
    }

    /**
     * Announces a new directly connected constraint to the object constraint.
     * @param node  The newly connected constraint.
     * @param edgeID  The ID of the constraint's connecting edge.
     * @param direction  The edge's direction WRT the object constraint.
     */
    public void addConstraint(Node node, int edgeID, boolean direction) {
        connectedNodes.add(new ConnectedNode(node, node.typeID, 
					     edgeID, direction));
    }

    public void connectNode(ConnectionReference first, 
			    ConnectionReference last) {

	Node node = first.node;

	if (lastConnectionNode != node) {

	    // Preventing from multiple calls in one update cycle

	    lastConnectionNode = node;

	    // Including own connection into reference chain

	    last.next = connectionReferenceConnection;
	    connectionReferenceConnection.previous = last;

	    // Propagating change

	    ConnectedNode connectedNode;
            Iterator it = connectedNodes.iterator();

            while (it.hasNext()) {
		connectedNode = (ConnectedNode) it.next();
		if (connectedNode.node != last.node) {
		    connectionReferenceConnection.edgeID = 
			connectedNode.edgeID;
		    connectionReferenceConnection.direction = 
			connectedNode.direction;
		    ((ConventionalConstraint) connectedNode.node).connectNode(first, connectionReferenceConnection);
		}
	    }
	}

    }

    /**
     * Announces a disconnection of a directly connected constraint to the 
     * object constraint.
     * @param node  The constraint to be disconnected.
     * @param edgeID  The ID of the constraint's connecting edge.
     * @param direction  The edge's direction WRT the object constraint.
     */
    public void removeConstraint(Node node, int edgeID, boolean direction) {

	ConnectedNode connectedNode;
	Iterator it = connectedNodes.iterator();
	boolean notFound = true;

	while (notFound && it.hasNext()) {
	    connectedNode = (ConnectedNode) it.next();
	    notFound = ((connectedNode.node != node) 
			|| (connectedNode.edgeID != edgeID) 
			|| (connectedNode.direction != direction)); 
	}
	it.remove();

    }

    public void disconnectNode(ConnectionReference first, 
			       ConnectionReference last) {

        Node node = first.node;

	if (lastDisconnectionNode != node) {

            // Preventing from multiple calls in one update cycle

	    lastDisconnectionNode = node;

            // Including own connection into reference chain

	    last.next = connectionReferenceDisconnection;
	    connectionReferenceDisconnection.previous = last;

            // Propagating change

	    ConnectedNode connectedNode;
            Iterator it = connectedNodes.iterator();

            while (it.hasNext()) {
                connectedNode = (ConnectedNode) it.next();
		if (connectedNode.node != last.node) {
		    connectionReferenceDisconnection.edgeID = 
			connectedNode.edgeID;
		    connectionReferenceDisconnection.direction = 
			connectedNode.direction;
		    ((ConventionalConstraint) connectedNode.node).disconnectNode(first, connectionReferenceDisconnection);
		}
	    }
	}
	fireChange(this);

    }

    public void createGraphics() {
	graphObject = (ObjectConstraintGraphics) 
	    new ObjectConstraintGraphics(this, gsc);
    }

    public int getNumberOfCosts() {
	return 0;
    }

    public String getCostName(int costID) {
	throw new Error("Invalid function for object constraints");
    }

    public void improveCosts(int costID) {
	throw new Error("Invalid function for object constraints");
    }

}


/*
 * DO _NOT_ EDIT ANYTHING BELOW THIS POINT!
 *
 * $Log: ObjectConstraint.java,v $
 * Revision 2.2  2004/04/07 12:53:37  weissd
 * Added capability to observe the datastructures
 *
 * Revision 2.1  2002/10/01 01:17:44  nareyek
 * object constraint propagation overhauled; heuristics of resource constraints improved
 *
 * Revision 2.0  2002/09/17 00:33:06  nareyek
 * work on beta version began: resource type, state tasks and precondition tasks introduced
 *
 * Revision 1.8  2002/08/23 23:44:26  nareyek
 * object-, integer- and range-constraints get not included in cost functions anymore
 *
 * Revision 1.7  2002/06/01 01:12:28  nareyek
 * initial cost type selection for graphics added
 *
 * Revision 1.6  2002/05/21 22:24:10  nareyek
 * synchronization problem WRT propagation and Connections resolved
 *
 * Revision 1.5  2002/05/16 20:30:36  nareyek
 * more than one cost type per constraint can now be used
 *
 * Revision 1.4  2002/05/16 00:46:22  nareyek
 * IntervalList improved; basic internal structures of DiscreteResourceConstraint completed
 *
 * Revision 1.3  2002/04/09 01:19:37  nareyek
 * Every cost item can now have a mapping; constraints can have multiple costs
 *
 * Revision 1.2  2002/04/04 09:25:36  nareyek
 * general constraint interface included; cost workflow improved
 *
 * Revision 1.1  2002/04/04 04:55:58  nareyek
 * directories costs and constraintgraph introduced
 *
 * Revision 1.5  2002/04/04 03:04:58  nareyek
 * new constraint directory; VariableComparator removed and integrated into Variable
 *
 * Revision 1.4  2002/03/30 08:33:09  nareyek
 * basic cost-function functionality added
 *
 * Revision 1.3  2002/01/30 00:57:11  teutates
 * no message
 *
 * Revision 1.2  2001/11/19 17:42:24  nareyek
 * Connections class rewritten; added cvs log commands
 *
 *
 */
