
package com.ai_center.dragonbreath.constraintgraph;

/** 
 * Represents a variable of the CSP. 
 * 
 * @author Alexander Nareyek
 * @version $Revision: 2.5 $
 */


import java.lang.Math;
import java.util.*;

import org.jdom.Element;

import com.ai_center.dragonbreath.constraints.RangeConstraint;
import com.ai_center.dragonbreath.GlobalSearchControl;
import com.ai_center.dragonbreath.Registry;


public class Variable extends Node implements java.util.Comparator {

    // **********************
    // ***** ATTRIBUTES *****
    // **********************

    /** 
     * The variable type's name.
     */
    private static final String TYPE_NAME = new String("Variable");

    /** 
     * The variable's name.
     */
    private String name = TYPE_NAME; 

    // Bound handling

    /** 
     * Set of connected range constraints.
     */
    public Set rangeConstraints = Collections.synchronizedSet(new HashSet());
    /** 
     * The variable's real lower bound.
     */
    public double lowerBound = Double.NEGATIVE_INFINITY;
    /** 
     * The variable's real upper bound.
     */
    public double upperBound = Double.POSITIVE_INFINITY;
    /** 
     * The variable's integer lower bound.
     */
    public long integerLowerBound = (long) Math.ceil(lowerBound);
    /** 
     * The variable's integer upper bound.
     */
    public long integerUpperBound = (long) Math.floor(upperBound);

    // Integer constraint handling

    /** 
     * Counts how many constraints require the variable to be an 
     * integer. These integer constraints are considered to be hard, 
     * i.e., if at least one integer constraint exists only integer 
     * values can be assigned to the variable.
     * @see #addIntegerConstraint
     * @see #removeIntegerConstraint
     */
    public int integerConstraints = 0; 
    /** 
     * Represents variable's integer value if <code>integerConstraints > 0</code>.
     * @see #integerConstraints
     */
    public long integerValue = 0; 
    /** 
     * Represents variable's integer value if <code>integerConstraints = 0</code>.
     * @see #integerConstraints
     */
    public double realValue = 0.0;


    // ************************
    // ***** CONSTRUCTORS *****
    // ************************

    /** 
     * Type constructor.
     */
    public Variable() {
        typeID = Registry.VARIABLE;
    }

    /** 
     * Initializes the new variable.
     * @param name  The name for the variable.
     * @param gsc  The GlobalSearchControl solver object.
     */
    public Variable(String name, GlobalSearchControl gsc) {
	typeID = Registry.VARIABLE;
	this.name = name;
	this.gsc = gsc;
	connections = new Connections(this);
    }

    /**
     * Gets the Variable Name and the value from nodeData.
     * @param nodeData  The nodeData as stored in file. 
     */
    public void setAuxiliaryNodeData(Element nodeData) {
        if (nodeData != null) {
            java.lang.String newName = nodeData.getAttributeValue( "Name" );
            name = newName;
            if ((graphObject != null) && (graphObject instanceof VariableGraphics)) {
                ((VariableGraphics)graphObject).updateVariableName(name);
            }
            double value = java.lang.Double.parseDouble( nodeData.getAttributeValue( "Value"));
            gsc.setVariable( this, value);
        }
    }   
    /**
     * Before storing a node this function is called, and the node can add its own data to the XML-data.
     * Adds the name and the value of the Variable.
     * @param nodeData  Where the data should be stored.
     */
    public void getAuxiliaryNodeData(Element nodeData) {
        nodeData.setAttribute( "Name", getName());
        nodeData.setAttribute( "Value", java.lang.String.valueOf( value() ) );
    }
    
    // *******************
    // ***** METHODS *****
    // *******************

    public String getTypeName() {
	return TYPE_NAME;
    }

    /** 
     * Returns the variable's name.
     */
    public String getName() {
	return name;
    }

    /** 
     * Returns the variable's value.
     */
    public double value() {
	if (integerConstraints != 0) {
	    return(integerValue);
	} else {
	    return(realValue);
	}
    }

    /** 
     * Prints the variable's name and value.
     */
    public String toString() {
	if (integerConstraints != 0) {
	    return name + " = " + integerValue;
	} else {
	    return name + " = " + realValue;
	}
    }

     /** 
     * Sets the variable's value.
     * @param i  A long value.
     */
    public void set(long i) {
	if (integerConstraints != 0) {
	    integerValue = i;
	} else {
	    realValue = i;
	}
	ensureBounds();
	fireChange(this);
    }

    /** 
     * Sets the variable's value.
     * @param r  A real value.
     */
    public void set(double r) {
	if (integerConstraints != 0) {
	    integerValue = Math.round(r);
	} else {
	    realValue = r;
	}
	ensureBounds();
	fireChange(this);   
    }

    /** 
     * Returns whether the variable is an integer.
     */
    public boolean isInteger() {
	return (integerConstraints != 0);
    }

    /** 
     * Returns whether the variable is a constant (by way of
     * range constraints).
     */
    public boolean isConstant() {
	if (integerConstraints != 0) {
	    return (integerLowerBound == integerUpperBound);
	} else {
	    return (lowerBound == upperBound);
	}
    }

    /** 
     * Returns the variable's current lower bound.
     */
    public double lowerBound() {
	if (integerConstraints != 0) {
	    return(integerLowerBound);
	} else {
	    return(lowerBound);
	}
    }

    /** 
     * Returns the variable's current upper bound.
     */
    public double upperBound() {
	if (integerConstraints != 0) {
	    return(integerUpperBound);
	} else {
	    return(upperBound);
	}
    }

    /** 
     * Implements java.util.Comparator.
     */
    public int compare(java.lang.Object o1, java.lang.Object o2) {

	double v1Value = ((Variable) o1).value();
	double v2Value = ((Variable) o2).value();

        if (v1Value < v2Value) {
	    return -1;
	} else {
	    if (v1Value == v2Value) {
		return 0;
	    } else {
		return 1;
	    }
	} 

    }

    /** 
     * Ensures that the variable's value is within the
     * given bounds.
     */
    public void ensureBounds() {
	if (integerConstraints != 0) {
	    if (integerLowerBound > integerValue) {
		integerValue = integerLowerBound;
	    }
	    if (integerUpperBound < integerValue) {
		integerValue = integerUpperBound;
	    }
	} else {
	    if (lowerBound > realValue) {
		realValue = lowerBound;
	    }
	    if (upperBound < realValue) {
		realValue = upperBound;
	    }
	}
    }

   /** 
     * Initializes visual tracking.
     */
    public void createGraphics() {
	graphObject = new VariableGraphics(this, gsc);
    }

    /** 
     * Adds an integer constraint.
     * @see #integerConstraints
     */
    public void addIntegerConstraint() {
	integerConstraints++;
    }

    /** 
     * Removes an integer constraint.
     * @see #integerConstraints
     */
    public void removeIntegerConstraint() {
	integerConstraints--;
    }

    /** 
     * Restricts the variable's lower and upper bound.
     * @param lb  A lower bound restriction.
     * @param ub  An upper bound restriction.
     */
    public void restrictRange(double lb, double ub) {
	if (lb > lowerBound) {
	    lowerBound = lb;
	    integerLowerBound = (long) Math.ceil(lowerBound);
	}
	if (ub < upperBound) {
	    upperBound = ub;
	    integerUpperBound = (long) Math.floor(upperBound);
	}
    }

    /** 
     * Recalculates only the current lower bounds.
     */
    private void recalculateLowerBounds() {
	lowerBound = Double.NEGATIVE_INFINITY;

	RangeConstraint r;
	synchronized(rangeConstraints) {
	    Iterator i = rangeConstraints.iterator();
	    while (i.hasNext()) {
		r = (RangeConstraint) i.next();
		if (r.lowerBound > lowerBound) {
		    lowerBound = r.lowerBound;
		}
	    }
	}
	integerLowerBound = (long) Math.ceil(lowerBound);
    }

    /** 
     * Recalculates only the current upper bounds.
     */
    private void recalculateUpperBounds() {
	upperBound = Double.POSITIVE_INFINITY;

	RangeConstraint r;
	synchronized(rangeConstraints) {
	    Iterator i = rangeConstraints.iterator();
	    while (i.hasNext()) {
		r = (RangeConstraint) i.next();
		if (r.upperBound < upperBound) {
		    upperBound = r.upperBound;
		}
	    }
	}
	integerUpperBound = (long) Math.floor(upperBound);
    }

    /** 
     * Recalculates the current bounds.
     */
    private void recalculateBounds() {
	lowerBound = Double.NEGATIVE_INFINITY;
	upperBound = Double.POSITIVE_INFINITY;

	RangeConstraint r;
	synchronized(rangeConstraints) {
	    Iterator i = rangeConstraints.iterator();
	    while (i.hasNext()) {
		r = (RangeConstraint)i.next();
		restrictRange(r.lowerBound, r.upperBound);
	    }
	}
	integerLowerBound = (long) Math.ceil(lowerBound);
 	integerUpperBound = (long) Math.floor(upperBound);
   }

    /** 
     * Adds a range constraint.
     * @param constraint  The range constraint to be added. 
     */
    public void addRangeConstraint(RangeConstraint constraint) {
	synchronized(rangeConstraints) {
	    rangeConstraints.add(constraint);
	}
	restrictRange(constraint.lowerBound, constraint.upperBound);
    }

    /**
     *  Returns the rangeConstraints of the variable that determines the upperBound.
     *  Or null if no rangeConstraint exists.
     */
    public RangeConstraint getUpperBoundRangeConstraint() {
        synchronized(rangeConstraints) {
	    Iterator it = rangeConstraints.iterator();
            while (it.hasNext()) {
                RangeConstraint rc = (RangeConstraint) it.next();
                if (rc.upperBound == upperBound) {
                    return rc;
                }        
            }//while
        }
	return null;
    }
    
    /**
     *  Returns the rangeConstraints of the variable that determines the lowerBound.
     *  Or null if no rangeConstraint exists.
     */
    public RangeConstraint getLowerBoundRangeConstraint() {
        synchronized(rangeConstraints) {
	    Iterator it = rangeConstraints.iterator();
            while (it.hasNext()) {
                RangeConstraint rc = (RangeConstraint) it.next();
                if (rc.lowerBound == lowerBound) {
                    return rc;
                }        
            }//while
        }
	return null;
    }
    
    /** 
     * Removes a range constraint.
     * @param constraint  The range constraint to be removed. 
     */
    public void removeRangeConstraint(RangeConstraint constraint) {

	synchronized(rangeConstraints) {
	    rangeConstraints.remove(constraint);
	}

	if (constraint.lowerBound == lowerBound) {
	    if (constraint.upperBound == upperBound) {
		recalculateBounds();
	    } else {
		recalculateLowerBounds();
	    }
	} else {
	    if (constraint.upperBound == upperBound) {
		recalculateUpperBounds();
	    }
	}

    }

    /** 
     * Triggers bound recalculation if a range constraint's lower
     * bounds are changed in an important way.
     * @param oldlb  The constraint's former lower bound.
     * @param newlb  The constraint's new lower bound.
     */
    public void checkLowerBounds(double oldlb, double newlb) {
	if (newlb > lowerBound) {
	    lowerBound = newlb;
	    integerLowerBound = (long) Math.ceil(lowerBound);
	} else {
	    if (oldlb == lowerBound) {
		recalculateLowerBounds();
	    }
	}
    }

    /** 
     * Triggers bound recalculation if a range constraint's upper
     * bounds are changed in an important way.
     * @param oldub  The constraint's former upper bound.
     * @param newub  The constraint's new upper bound.
     */
    public void checkUpperBounds(double oldub, double newub) {
	if (newub < upperBound) {
	    upperBound = newub;
	    integerUpperBound = (long) Math.floor(upperBound);
	} else {
	    if (oldub == upperBound) {
		recalculateUpperBounds();
	    }
	}
    }
    
    /** 
     * Triggers bound recalculation if a range constraint's bounds
     * are changed in an important way.
     * @param oldlb  The constraint's former lower bound.
     * @param newlb  The constraint's new lower bound.
     * @param oldub  The constraint's former upper bound.
     * @param newub  The constraint's new upper bound.
     */
    public void checkBounds(double oldlb, double oldub, 
			    double newlb, double newub) {

	boolean lbMightHaveChanged = false;
	boolean ubMightHaveChanged = false;
	if (newlb > lowerBound) {
	    lowerBound = newlb;
	    integerLowerBound = (long) Math.ceil(lowerBound);
	} else {
	    if (oldlb == lowerBound) {
		lbMightHaveChanged = true;
	    }
	}
	if (newub < upperBound) {
	    upperBound = newub;
	    integerUpperBound = (long) Math.floor(upperBound);
	} else {
	    if (oldub == upperBound) {
		ubMightHaveChanged = true;
	    }
	}

	if (lbMightHaveChanged) {
	    if (ubMightHaveChanged) {
		recalculateBounds();
	    } else {
		recalculateLowerBounds();
	    }
	} else {
	    if (ubMightHaveChanged) {
		recalculateUpperBounds();
	    }
	}
    }

}


/* 
 * DO _NOT_ EDIT ANYTHING BELOW THIS POINT! 
 * 
 * $Log: Variable.java,v $
 * Revision 2.5  2004/04/07 12:53:37  weissd
 * Added capability to observe the datastructures
 *
 * Revision 2.4  2003/05/07 23:38:52  teutates
 * documentation for structural constraint handling updated
 *
 * Revision 2.3  2003/03/21 14:44:59  teutates
 * scExample added
 *
 * Revision 2.2  2003/02/21 16:55:23  teutates
 * improved auxiliary data load for nodes
 *
 * Revision 2.1  2003/02/21 01:32:59  teutates
 * structural repair initial release
 *
 * Revision 2.0  2002/09/17 00:33:06  nareyek
 * work on beta version began: resource type, state tasks and precondition tasks introduced
 *
 * Revision 1.3  2002/08/28 07:52:53  nareyek
 * engine can now also be run without a graphical display
 *
 * Revision 1.2  2002/08/08 22:17:35  nareyek
 * fixed minor issues
 *
 * Revision 1.1  2002/04/04 04:55:58  nareyek
 * directories costs and constraintgraph introduced
 *
 * Revision 1.4  2002/04/04 03:04:58  nareyek
 * new constraint directory; VariableComparator removed and integrated into Variable
 *
 * Revision 1.3  2002/02/04 18:31:28  teutates
 * connection between two representations of same edge added
 *
 * Revision 1.2  2001/11/22 18:55:02  nareyek
 * Major Registry revision: Dynamic constraint class addition; Registry is now a static object only
 * 
 * 
 */ 
