
package com.ai_center.dragonbreath.constraints;

/** 
 * This constraint enforces that all connected variables have
 * different values.
 *
 * @author Tim Baermann
 * @version $Revision: 2.1 $
 * @see Variable
 */


import java.lang.Math;
import java.util.*;
import java.util.Arrays;

import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.Registry;


public class AllDifferentConstraint extends ConventionalConstraint {

    // **********************
    // ***** ATTRIBUTES *****
    // **********************

    /** 
     * The constraint type's name.
     */
    private static final String TYPE_NAME = new String("AllDifferent");

    /** 
     * The constraint's base graph.
     */
    private static final Embedding BASE_GRAPH = new Embedding(2);
    
    /** 
     * The constraint's number of extension graphs.
     */
    private static final int EXTENSION_GRAPH_NUMBER = 1;
    /** 
     * The constraint's extension graphs.
     */
    private static final Embedding[] EXTENSION_GRAPH = 
	new Embedding[EXTENSION_GRAPH_NUMBER];
    
    /**
     * Maximal number of connections.
     */
    private static final int MAX_CONNECTION_NUMBER = 256;
    
    /** 
     * The number of variables to be different.
     */
    private int NumVariables = 0;

    /**
     * An ordered list of all Variables.
     */
    private Variable[] Ordered = new Variable[MAX_CONNECTION_NUMBER];
    
    /** 
     * The number of same variables.
     */
    private int NumSameVariables = 0;


    // *******************
    // ***** METHODS *****
    // *******************

    public void create_graphs() {

	// Creating the base graph
	
	BASE_GRAPH.addConnectedNode(Registry.VARIABLE, 
				    Registry.PLAIN_EDGE, Registry.INCOMING);
	BASE_GRAPH.addConnectedNode(Registry.VARIABLE, 
				    Registry.PLAIN_EDGE, Registry.INCOMING);

	// Creating the extension graph
        
	EXTENSION_GRAPH[0] = new Embedding(1);
	EXTENSION_GRAPH[0].addConnectedNode(Registry.VARIABLE, 
					    Registry.PLAIN_EDGE, 
					    Registry.INCOMING);
    }

    public int getNumberOfCosts() {
	return 1;
    }

    public String getCostName(int costID) {
	if (costID == 0) {
	    return new String("equality inconsistency");
	} else {
	    throw new Error("Invalid cost ID");
	}
    }

    public String getTypeName() {
	return TYPE_NAME;
    }

    public Embedding getBaseGraph() {
	return BASE_GRAPH;
    }

    public int getExtensionGraphNumber() {
	return EXTENSION_GRAPH_NUMBER;
    }

    public Embedding getExtensionGraph(int graphID) {
	return EXTENSION_GRAPH[graphID];
    }

    public int getMaximalConnectionNumber() {
	// There is no maximum for extensions
	return MAX_CONNECTION_NUMBER;
    }

    public void initializeCostType(int costID) {
	// Nothing to do
    }

    /**
     * Calculates how many of the variables in Ordered
     * are the same.
     */
    private void calcNumSameVariables() {
        int i;      
        NumSameVariables=0;
        for (i = 0; i < NumVariables-1; i++) {
	    if (Ordered[i].value() == Ordered[i+1].value())
		NumSameVariables++;
        }
	if (NumSameVariables > 0) NumSameVariables++;
   }
    
    /** 
     * Calculates the constraint's costs.
     */
    private void updateCosts() {
	calcNumSameVariables();
	costs[0] = (long) NumSameVariables;
	costFunctionReference[0].updateGlobalCosts();
    }

    public void update(double oldValue, ConnectionReference first, 
		       ConnectionReference last) {
	synchronized(Ordered){ 
            Arrays.sort(Ordered, 0, NumVariables, Ordered[0]);
	}
	updateCosts();
    }

    public void connectNode(ConnectionReference first, 
			    ConnectionReference last) {
	Variable variable = (Variable) first.node;
	synchronized(Ordered) {
            Ordered[NumVariables] = variable;
            NumVariables++;
            Arrays.sort(Ordered, 0, NumVariables, variable);
	}
        updateCosts();   
    }

    public void disconnectNode(ConnectionReference first, 
			       ConnectionReference last) {
	Variable variable = (Variable) first.node;
	synchronized(Ordered) {
	    for (int i = 0; i < NumVariables; i++) {
		if (Ordered[i] == variable){
		    Ordered[i] = Ordered[NumVariables-1];
		    break;
		}
	    }
	    NumVariables--;
	    if (NumVariables > 0) {
		Arrays.sort(Ordered, 0, NumVariables, Ordered[0]);
	    }
	}
	if (NumVariables != 0) updateCosts();
    }

    /**
     * Calculates an array with the number of duplicates for every variable
     * of Ordered.
     * @param a  The array to be filled.
     */
    public void calcNumDuplicates(int[] a) {
       
        int numDuplicates;
        double actVal;

        synchronized(Ordered) {
	    for (int i = 0; i < NumVariables; i++) {

		numDuplicates = 0;
		actVal=Ordered[i].value();

                if (a[i] == 0){

		    for (int n = 0; n < NumVariables; n++) {
			if (a[n] == 0) {
			    if (Ordered[n].value() == actVal) {
				numDuplicates++;
			    }
			}//if ai!=0
		    }//for

		    for (int n = 0; n<NumVariables; n++) {
			if (a[n] == 0) {
			    if (Ordered[n].value() == actVal) {
				a[n] = numDuplicates;
			    }
			}
		    }//for

                }//if ai!=0
            }//for
        }//syncronized
    }//calcNumDuplicates()
    
    public void improveCosts(int costID) {

	// Checking for valid cost ID

	if (costID != 0) {
	    throw new Error("The cost ID  \"" + costID + 
			    "\" is not supported by this constraint");
	}

	int i;
	int[] duplicates = new int[NumVariables];
	calcNumDuplicates(duplicates);
	int possibleChanges = 0;
	int range;

	synchronized(Ordered) {
	    for (i = 1; i < NumVariables; i++) { 
		if (((Variable) Ordered[i]).isConstant()) {
		    duplicates[i] = 0;
		}
		else possibleChanges += duplicates[i];
	    }
	}
    
	int chosenVariable = gsc.random.nextInt(possibleChanges);
	int pos = 0;

	while (chosenVariable >= 0) {
            while ((duplicates[pos] == 0) && (pos < NumVariables)) {
                pos++;
	    }
            chosenVariable -= duplicates[pos];
            pos++;
        } 

        Variable variable = (Variable) Ordered[pos-1];
        boolean done = false;

	//try until change
        while (!done) {
	    // the more variables exist, the wider is the potential shift
	    int change = (-NumVariables / 2) + 
		gsc.random.nextInt(NumVariables + 1);
	    double newVal = variable.value() + change;

	    if (variable.upperBound() >= newVal) {
		gsc.setVariable_internal(variable, newVal);
		done=true;
	    }
	    else if (variable.lowerBound() <= newVal) {
		gsc.setVariable_internal(variable, newVal);
		done = true;
	    }
        }//while !d
    }

}


/* 
 * DO _NOT_ EDIT ANYTHING BELOW THIS POINT! 
 * 
 * $Log: AllDifferentConstraint.java,v $
 * Revision 2.1  2003/02/21 01:32:59  teutates
 * structural repair initial release
 *
 * Revision 2.0  2002/09/17 00:33:06  nareyek
 * work on beta version began: resource type, state tasks and precondition tasks introduced
 *
 * Revision 1.6  2002/06/01 01:12:28  nareyek
 * initial cost type selection for graphics added
 *
 * Revision 1.5  2002/05/16 20:30:37  nareyek
 * more than one cost type per constraint can now be used
 *
 * Revision 1.4  2002/04/09 01:19:37  nareyek
 * Every cost item can now have a mapping; constraints can have multiple costs
 *
 * Revision 1.3  2002/04/04 09:25:36  nareyek
 * general constraint interface included; cost workflow improved
 *
 * Revision 1.2  2002/04/04 04:55:58  nareyek
 * directories costs and constraintgraph introduced
 *
 * Revision 1.1  2002/04/04 03:04:58  nareyek
 * new constraint directory; VariableComparator removed and integrated into Variable
 *
 * Revision 1.4  2002/03/30 08:33:09  nareyek
 * basic cost-function functionality added
 *
 * Revision 1.3  2001/11/22 18:55:00  nareyek
 * Major Registry revision: Dynamic constraint class addition; Registry is now a static object only
 * 
 * 
 */ 
