
package com.ai_center.dragonbreath.constraints;

/** 
 * This constraint enforces by its capacity overload cost function that the 
 * connected tasks' consumptions do not exceed the resource's capacity at 
 * any time. The makespan cost function returns the makespan, i.e., the time 
 * at which the last task finishes. The workload cost function sums the 
 * consumption over time for all tasks that start after the current time.
 * All connected tasks must have the same resource type variable's value as 
 * the constraint.
 *
 * @author Alexander Nareyek
 * @version $Revision: 2.18 $
 * 
 * @see ActionTaskObjectConstraint
 */


import java.util.*;

import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.constraints.discreteresourceconstraint.*;
import com.ai_center.dragonbreath.Registry;
import com.ai_center.dragonbreath.utilities.IntervalList;
import com.ai_center.dragonbreath.utilities.IntervalListElement;
import com.ai_center.dragonbreath.utilities.OrderedShiftDLLElement;


public class DiscreteResourceConstraint extends ConventionalConstraint {

    // **********************
    // ***** ATTRIBUTES *****
    // **********************

    /** The name of the DiscreteResource. */
    private String discreteResourceName;
    
    /** 
     * Constant for a task's execution state variable to indicate that the
     * processing of the task has not yet begun.
     */
    public static final int PENDING = 0;
    /** 
     * Constant for a task's execution state variable to indicate that the
     * task is currently processed.
     */
    public static final int IN_EXECUTION = 1;
    /** 
     * Constant for a task's execution state variable to indicate that the
     * task has already been processed.
     */
    public static final int DONE = 2;


    /***** ENGINE-INTEGRATION-RELATED VARIABLES *****/

    /** 
     * The constraint type's name.
     */
    private static final String TYPE_NAME = new String("Discrete Resource");

    /** 
     * The identifier for the constraint's capacity overload cost function.
     */
    public static final int CAPACITY_OVERLOAD = 0;
    /** 
     * The identifier for the constraint's makespan cost function.
     */
    public static final int MAKESPAN = 1;
    /** 
     * The identifier for the constraint's workload cost function.
     */
    public static final int WORKLOAD = 2;
    /** 
     * Descriptions of the cost types.
     */
    public static String[] costDescription = new String[3];
    static {
	costDescription[0] = new String("capacity overload");
	costDescription[1] = new String("makespan");
	costDescription[2] = new String("workload");
    }

    /** 
     * The constraint's base graph.
     */
    private static final Embedding BASE_GRAPH = new Embedding(3);
    /** 
     * The constraint's number of extension graphs.
     */
    private static final int EXTENSION_GRAPH_NUMBER = 1;
    /** 
     * The constraint's extension graphs.
     */
    private static final Embedding[] EXTENSION_GRAPH = 
	new Embedding[EXTENSION_GRAPH_NUMBER];


    /***** GENERAL INTERNAL STRUCTURES *****/

    /** 
     * The capacity variable.
     */
    private Variable capacityVariable = null;
    /** 
     * The internal copy of the capacity variable's value.
     */
    private long capacity = 0;

    /**
     * The internal copy of the current time variable's value.
     */
    private long currentTime = 0;
    /**
     * The interval including the current time.
     */
    private IntervalListElement currentTimeILE = null;

    /**
     * No task begin is moved before the current time plus this buffer
     * to enable more repair steps before a potential execution.
     */
    private static long repairTimeBuffer = 0;

    /** 
     * Indicates whether the constraint is currently changing CSP components
     * and the update procedure does not need to be called.
     */
    private boolean modificationModus = false;


    /***** INTERVAL MANAGEMENT *****/

    /** 
     * Temporal interval list in which all connected tasks are included.
     */
    private IntervalList intervalList = null;

    /**
     * List of intervals ordered according to their subjective capacity costs.
     */
    private OrderedCapacityCostsIntervalList capacityCostsIntervalList = null;

    /**
     * List of intervals ordered according to their number of claims (first
     * priority; highest first) and length (second priority; longest first).
     */
    private OrderedClaimsIntervalList claimsIntervalList = null;


    /***** TASK MANAGEMENT *****/

    /** 
     * Hashtable from ActionTaskObjectConstraints to DRCTasks.
     */
    private Hashtable taskConstraintsHT = null;

    /** 
     * The set of DRCTasks representing the connected TaskObjects.
     */
    private OrderedCapacityCostsTaskList taskList = null;
    /** 
     * Total of tasks' capacity cost variables.
     */
    private long taskCapacityCosts = 0;


    /***** SUPPORT/INFO STRUCTURES *****/

    /** 
     * Hashtable from operation IDs to strings.
     */
    private Hashtable operationDescriptions;


    // *******************
    // ***** METHODS *****
    // *******************

    public void create_graphs() {

	// Creating the base graph

	BASE_GRAPH.addConnectedNode(Registry.CURRENTTIME_OBJECTCONSTRAINT, 
				    Registry.PLAIN_EDGE, Registry.INCOMING);
	BASE_GRAPH.addConnectedNode(Registry.VARIABLE, 
				    Registry.RESOURCETYPE_EDGE, 
				    Registry.INCOMING);
	BASE_GRAPH.addConnectedNode(Registry.VARIABLE, 
				    Registry.CAPACITY_EDGE, Registry.INCOMING);

	// Creating the extension graph
        
	EXTENSION_GRAPH[0] = new Embedding(1);
	EXTENSION_GRAPH[0].addConnectedNode(Registry.ACTIONTASK_OBJECTCONSTRAINT, Registry.PLAIN_EDGE, Registry.INCOMING);

    }

    public int getNumberOfCosts() {
	return 3;
    }

    public String getCostName(int costID) {
	if ((costID >= 0) && (costID < 3)) {
	    return costDescription[costID];
	} else {
	    throw new Error("Invalid cost ID");
	}
    }

    public String getTypeName() {
	return TYPE_NAME;
    }

    public Embedding getBaseGraph() {
	return BASE_GRAPH;
    }

    public int getExtensionGraphNumber() {
	return EXTENSION_GRAPH_NUMBER;
    }

    public Embedding getExtensionGraph(int graphID) {
	return EXTENSION_GRAPH[graphID];
    }

    public int getMaximalConnectionNumber() {
	// There is no maximum for extensions
	return Integer.MAX_VALUE;
    }


    public void initializeInternalStructures() {

	// Initializing task structures

	taskList = new OrderedCapacityCostsTaskList();
	taskConstraintsHT = new Hashtable();

	// Initializing interval structures

	DRCInterval firstInterval = new DRCInterval();
	intervalList = new IntervalList(0, Long.MAX_VALUE, firstInterval);

	currentTimeILE = intervalList.getFirst();

	capacityCostsIntervalList = new OrderedCapacityCostsIntervalList();
	firstInterval.capacityCostsIntervalLE = capacityCostsIntervalList.addListElement((Object) intervalList.getFirst());

	claimsIntervalList = new OrderedClaimsIntervalList();
	firstInterval.claimsIntervalLE = 
	    claimsIntervalList.addListElement((Object) intervalList.getFirst());

    }

    public void initializeCostType(int costID) {
	switch (costID) {
	case MAKESPAN: {
	    costs[MAKESPAN] = intervalList.getLast().getBeginning();
	    break;
	}
	case WORKLOAD: {
	    OrderedShiftDLLElement currentLE = taskList.first;
	    DRCTask task = null;
	    long costBuffer = 0;
	    while (currentLE != null) {
		task = (DRCTask) currentLE.element;
		if (task.begin > currentTime) {
		    costBuffer += task.duration * task.consumption;
		}
		currentLE = currentLE.next;
	    }
	    costs[WORKLOAD] = costBuffer;
	    break;
	}
	}
	costFunctionReference[costID].updateGlobalCosts();
    }


    /***** TYPE CUSTOMIZATION *****/

    /** 
     * Sets a temporal buffer value for all DiscreteResourceConstraints
     * to enable more repair steps before a potential task execution. 
     * No task begin is moved by the constraints' heuristics before the 
     * current time plus this buffer.
     * @param buffer  The temporal buffer value.
     */
    public static void setRepairTimeBuffer(long buffer) {
	repairTimeBuffer = buffer;
    }


    /***** SUPPORT/INFO *****/

    /** 
     * Adds a description for an operation ID.
     * @param operationID  The operation ID to be described.
     * @param description  The description to be added.
     */
    public void addDescription(long operationID, String description) {
	if (operationDescriptions == null) {
	    operationDescriptions = new Hashtable();
	}
	operationDescriptions.put(new Long(operationID), description);
    }

	public String getDescription(long operationID) {
		if (operationDescriptions != null) {
			return (String) operationDescriptions.get(new Long(operationID));
		} else {
			return null;
		} 
	}

    /** Sets a name for the DiscreteResource.<p>
     @param name The name of the DiscreteResource
     */
    public void setName(String name)
    {
        discreteResourceName = name;

    }

    /** Gets the name of the DiscreteResource.<p>
     @return The name or unknown, if not set.
     */
    public String getName()
    {
        if (discreteResourceName == null)
        {
            return "Unkown";
        } else
        {
            return discreteResourceName;
        }

    }

    
    /** Returns the DRCTask for a give ActionTaskObjectConstraint.<p>
     
     @param actionTaskObjectConstraint The ActionTaskObjectConstraint to query.
     @return The corresponding DRCTask.
     */
    public DRCTask getDRCTask(
            ActionTaskObjectConstraint actionTaskObjectConstraint)
    {

        return (DRCTask) taskConstraintsHT.get(actionTaskObjectConstraint);
    }

    /**
     @return Returns the intervalList.
     */
    public IntervalList getIntervalList()
    {
        return intervalList;
    }

    /**
     @return Returns the taskList.
     */
    public OrderedCapacityCostsTaskList getTaskList()
    {
        return taskList;
    }
    
    /** Fires a change to make a initial update.<p> */
    public void fireIntervalListUpdate()
    {
        intervalList.fireIntervalListUpdate();

    }

    /***** INTERVAL AND COST HANDLING *****/

    /**
     * Returns a description of the involved actions.
     */
    public StringBuffer getLoadDescription() {

        StringBuffer s = new StringBuffer();
        IntervalListElement currentILE = intervalList.getFirst();
	DRCInterval interval;
	Iterator it;

	if (currentTime != currentILE.getBeginning()) {
	    s.append(" [" + currentILE.getBeginning() + "]");
	} else {
	    s.append(" [CT=" + currentTime + "]");
	}

        while (currentILE != null) {
	    interval = (DRCInterval) currentILE.getElement();
	    if ((currentTime > currentILE.getBeginning()) && (currentTime <= currentILE.getEnd())) {
		it = interval.tasks.iterator();
		if (!it.hasNext()) {
		    s.append(" <idle");
		} else {
		    if (operationDescriptions != null) {
			s.append(" <" + operationDescriptions.get(new Long((long) ((DRCTask) it.next()).operationVar.value())));
			while (it.hasNext()) {
			    s.append(", " + operationDescriptions.get(new Long((long) ((DRCTask) it.next()).operationVar.value())));
			}
		    } else {
			s.append(" <" + (long) ((DRCTask) it.next()).operationVar.value());
			while (it.hasNext()) {
			    s.append(", " + (long) ((DRCTask) it.next()).operationVar.value());
			}
		    }
		}
		s.append("> [CT=" + currentTime + "]");
	    }
	    if (currentTime != currentILE.getEnd()) {
		it = interval.tasks.iterator();
		if (!it.hasNext()) {
		    s.append(" <idle");
		} else {
		    if (operationDescriptions != null) {
			s.append(" <" + operationDescriptions.get(new Long((long) ((DRCTask) it.next()).operationVar.value())));
			while (it.hasNext()) {
			    s.append(", " + operationDescriptions.get(new Long((long) ((DRCTask) it.next()).operationVar.value())));
			}
		    } else {
			s.append(" <" + (long) ((DRCTask) it.next()).operationVar.value());
			while (it.hasNext()) {
			    s.append(", " + (long) ((DRCTask) it.next()).operationVar.value());
			}
		    }
		}
		s.append("> [" + currentILE.getEnd() + "]");
	    }
            currentILE = currentILE.getNext();
        }

	return s;

    }

    /** 
     * Calculates an interval's subjective capacity costs, sets the
     * interval's capacityCosts variable, updates costs[CAPACITY_OVERLOAD] 
     * and re-orders the capacity costs' and claims' interval structures.
     * Returns the capacity cost change.
     * @param ile  The interval list element to be evaluated.
     */
    private long updateIntervalCapacityCosts(IntervalListElement ile) {

	DRCInterval interval = (DRCInterval) ile.getElement();

	// Determining relevant interval duration (only after current time)

	long relevantDuration;

	if (currentTime <= ile.getBeginning()) {
	    relevantDuration = ile.getDuration();
	} else {
	    if (ile.getEnd() <= currentTime) {
		relevantDuration = 0;
	    } else {
		relevantDuration = ile.getEnd() - currentTime;
	    }
	}

	// Updating costs

	long costChange = -interval.capacityCosts;

	interval.calculateCapacityCosts(relevantDuration, capacity);

	costChange += interval.capacityCosts;
	costs[CAPACITY_OVERLOAD] += costChange;

	if (costChange != 0) {

	    // Updating tasks' capacity costs and task list

	    DRCTask task;
	    Iterator it = interval.tasks.iterator();
	    while (it.hasNext()) {
		task = (DRCTask) it.next();
		task.capacityCosts += costChange;
		taskList.updateListElementPosition(task.taskLE);
	    }

	    // Updating sum of tasks' capacity costs

	    taskCapacityCosts += interval.tasks.size() * costChange;

	    // Updating capacity costs' interval list

	    capacityCostsIntervalList.updateListElementPosition(interval.capacityCostsIntervalLE);

	}

	// Updating claims' interval list

	claimsIntervalList.updateListElementPosition(interval.claimsIntervalLE);

	return costChange;

    }

    /** 
     * Removes an interval's subjective capacity costs from 
     * costs[CAPACITY_OVERLOAD] and excludes all corresponding interval 
     * list elements. The previous (extended) interval list element is 
     * returned. Note that the ordering/costs of the extended previous 
     * interval element are not updated. Calling this procedure with the 
     * very first interval element will cause an error.
     * @param ile  The interval list element to be removed.
     */
    private IntervalListElement removeIntervalElement(IntervalListElement 
						      ile) {

	// Updating cost information

	DRCInterval interval = (DRCInterval) ile.getElement();
	long intervalCapacityCosts = interval.capacityCosts;
	costs[CAPACITY_OVERLOAD] -= intervalCapacityCosts;

	DRCTask task;
	Iterator it = interval.tasks.iterator();
	while (it.hasNext()) {
	    task = (DRCTask) it.next();
	    task.capacityCosts -= intervalCapacityCosts;
	    taskList.updateListElementPosition(task.taskLE);
	}

	taskCapacityCosts -= interval.tasks.size() * intervalCapacityCosts;

	// Removing interval structures

	capacityCostsIntervalList.removeListElement(interval.capacityCostsIntervalLE);
	interval.capacityCostsIntervalLE = null;

	claimsIntervalList.removeListElement(interval.claimsIntervalLE);
	interval.claimsIntervalLE = null;

	// Removing interval from interval list

	IntervalListElement extendedILE = 
	    intervalList.removeBasePoint(ile.getPrevious().getEnd(), ile.getPrevious());

	// Possibly updating current time reference

	if (currentTimeILE == ile) {
	    currentTimeILE = extendedILE;
	}

	return extendedILE;

    }

    /** 
     * Sets all lastInterval pointers of an IntervalListElement's tasks to 
     * a new IntervalListElement.
     * @param fromInterval  Interval from which the Pointers are moved away.
     * @param toInterval  Interval to which the Pointers are moved.
     */
    private void moveLastIntervalPointers(IntervalListElement fromILE,
					  IntervalListElement toILE) {
	DRCInterval interval = (DRCInterval) fromILE.getElement();
	DRCTask task;
	Iterator it = interval.tasks.iterator();
	while (it.hasNext()) {
	    task = (DRCTask) it.next();
	    if (task.lastInterval == fromILE) {
		task.lastInterval = toILE;
	    }
	}
    }


    /***** TASK HANDLING *****/

    /** 
     * Adds a task to the internal structures and updates costs.
     * @param task  The DRCTask task to be added.
     */
    private void addTask(DRCTask task) {

	// Checking cover interval

	if (task.begin < intervalList.getBeginning()) {
	    throw new Error("The given task \"" + task + "\" begins outside the interval list's cover interval [" + intervalList.getBeginning() + ", " + intervalList.getEnd() + "(");
	}
	if (task.end >= intervalList.getEnd()) {
	    throw new Error("The given task \"" + task + "\" ends outside the interval list's cover interval [" + intervalList.getBeginning() + ", " + intervalList.getEnd() + "(");
	}

	DRCInterval interval;

	// Splitting interval at task's begin point if necessary

	IntervalListElement firstILE = intervalList.addBasePoint(task.begin);

        // If interval was splitted, copying intervals for new list element 
	// and updating structures accordingly

	if (firstILE.getElement() == null) {

	    // Updating interval information

	    firstILE.setElement( (DRCInterval) ((DRCInterval) firstILE.getPrevious().getElement()).clone() );
	    interval = (DRCInterval) firstILE.getElement();

	    if (!interval.tasks.isEmpty()) {
		moveLastIntervalPointers(firstILE.getPrevious(), firstILE);
	    }

	    // Possibly updating current time reference

	    if ((currentTimeILE == firstILE.getPrevious()) 
		&& (firstILE.getBeginning() <= currentTime)) {
		currentTimeILE = firstILE;
	    }

	    // Updating cost structures

	    updateIntervalCapacityCosts(firstILE.getPrevious());

	    interval.capacityCostsIntervalLE = 
		capacityCostsIntervalList.addListElement((Object) firstILE);
	    interval.claimsIntervalLE = 
		claimsIntervalList.addListElement((Object) firstILE);

 	} else {

	    // Shifting firstILE such that firstILE starts with the task

	    if (firstILE.getEnd() <= task.begin) {
		firstILE = firstILE.getNext();
	    }

	}

	// Splitting interval at task's end point if necessary

	IntervalListElement lastILE = 
	    intervalList.addBasePointWithOffset(task.end, firstILE);

        // If interval was splitted, copying intervals for new list element 
	// and updating structures accordingly

	if (lastILE.getElement() == null) {

	    // Updating interval information

	    lastILE.setElement ( (DRCInterval) ((DRCInterval) lastILE.getPrevious().getElement()).clone());
	    interval = (DRCInterval) lastILE.getElement();
	    if (!interval.tasks.isEmpty()) {
		moveLastIntervalPointers(lastILE.getPrevious(), lastILE);
	    }

	    // Possibly updating current time reference

	    if ((currentTimeILE == lastILE.getPrevious()) 
		&& (lastILE.getBeginning() <= currentTime)) {
		currentTimeILE = lastILE;
	    }

	    // Updating cost structures

	    interval.capacityCostsIntervalLE = 
		capacityCostsIntervalList.addListElement((Object) lastILE);
	    interval.claimsIntervalLE = 
		claimsIntervalList.addListElement((Object) lastILE);

	    updateIntervalCapacityCosts(lastILE);

 	} else {

	    // Shifting lastILE such that lastILE is after the task

	    if (task.end > lastILE.getBeginning()) {
		lastILE = lastILE.getNext();
	    }

	}

        // Setting task's interval reference

        task.firstInterval = firstILE;
	task.lastInterval = lastILE.getPrevious();

	// Adding task information to interval list elements and updating
	// costs and orderings of interval-related lists

	while (firstILE != lastILE) {
	    interval = (DRCInterval) firstILE.getElement();

	    interval.tasks.add(task);
	    task.capacityCosts += interval.capacityCosts;
	    taskCapacityCosts += interval.capacityCosts;
	    interval.claims += task.consumption;

	    if (updateIntervalCapacityCosts(firstILE) == 0) {
		taskList.updateListElementPosition(task.taskLE);
	    }

	    firstILE = firstILE.getNext();
	}

        // Updating workload costs

	if ((costFunctionReference[WORKLOAD] != null) 
	    && (task.begin > currentTime)) {
	    costs[WORKLOAD] += task.duration * task.consumption;
	}
	fireChange(this);
    }

    /** 
     * Removes a task from the internal structures and updates costs.
     * @param task  The DRCTask task to be removed.
     */
    private void removeTask(DRCTask task) {

	IntervalListElement firstILE = task.firstInterval;
	IntervalListElement lastILE = task.lastInterval;
	DRCInterval interval;

	// Removing task information from task's first interval list element

	interval = (DRCInterval) firstILE.getElement();

	interval.tasks.remove(task);
	interval.claims -= task.consumption;

	// Merging task's first interval if necessary

	if (firstILE.getPrevious() != null) {
	    if (((DRCInterval) firstILE.getPrevious().getElement()).tasks.equals(interval.tasks)) {
		if (!interval.tasks.isEmpty()) {
		    moveLastIntervalPointers(firstILE, firstILE.getPrevious());
		}
		if (firstILE == lastILE) {
		    lastILE = firstILE.getPrevious();
		}
		firstILE = removeIntervalElement(firstILE);
	    }
	}

	// Removing task information from task's last interval list element

	interval = (DRCInterval) lastILE.getElement();

	if (lastILE != firstILE) {
	    interval.tasks.remove(task);
	    interval.claims -= task.consumption;
	}

	// Merging task's last interval if necessary

	if (lastILE.getNext() != null) {
	    if (((DRCInterval) lastILE.getNext().getElement()).tasks.equals(interval.tasks)) {
		if (!interval.tasks.isEmpty()) {
		    moveLastIntervalPointers(lastILE.getNext(), lastILE);
		}
		removeIntervalElement(lastILE.getNext());
	    }
	}

	// Updating costs and orderings of interval-related lists

	IntervalListElement currentILE = firstILE;
	IntervalListElement afterILE = lastILE.getNext();
	boolean reorder = true;

	while (currentILE != afterILE) {
	    interval = (DRCInterval) currentILE.getElement();

	    if ((currentILE != firstILE) && (currentILE != lastILE)) {
		interval.tasks.remove(task);
		interval.claims -= task.consumption;
	    }

	    updateIntervalCapacityCosts(currentILE);

	    currentILE = currentILE.getNext();
	}

	taskCapacityCosts -= task.capacityCosts;
	task.capacityCosts = 0;

        // Updating the task's interval references

        task.firstInterval = null;
	task.lastInterval = null;

        // Updating workload costs

	if ((costFunctionReference[WORKLOAD] != null) 
	    && (task.begin > currentTime)) {
	    costs[WORKLOAD] -= task.duration * task.consumption;
	}
	fireChange(this);

    }


    // ***** CAPACITY/CONSUMPTION HANDLING *****

    /** 
     * Resets the internal capacity variable and updates the constraint's 
     * capacity costs and related structures.
     * @param newCapacity  The new value of the capacity.
     */
    private void updateCapacityVariable(long newCapacity) {

	// Updating capacity variable

	capacity = newCapacity;

	// Updating costs and orderings of interval-related lists

	IntervalListElement currentILE = intervalList.getFirst();
	DRCInterval interval;

	while (currentILE != null) {
	    interval = (DRCInterval) currentILE.getElement();
	    if (!interval.tasks.isEmpty()) {
		updateIntervalCapacityCosts(currentILE);
	    }
	    currentILE = currentILE.getNext();
	}

    }

    /** 
     * Updates the capacity costs and related structures for a task's 
     * consumption change.
     * @param task  The task for which the consumption has been updated.
     * @param oldConsumption  The old value of the task's consumption.
     */
    private void propagateConsumptionChange(DRCTask task, 
					    long oldConsumption) {

        IntervalListElement firstILE = task.firstInterval;
        IntervalListElement afterILE = task.lastInterval.getNext();
	DRCInterval interval;
	long newConsumption = task.consumption;

	while (firstILE != afterILE) {
	    interval = (DRCInterval) firstILE.getElement();
	    interval.claims += newConsumption - oldConsumption;
	    updateIntervalCapacityCosts(firstILE);
	    firstILE = firstILE.getNext();
	}

       // Updating workload costs

	if ((costFunctionReference[WORKLOAD] != null) 
	    && (task.begin > currentTime)) {
	    costs[WORKLOAD] += 
		task.duration * (task.consumption - oldConsumption);
	}

    }


    // ***** ONGOING TIME HANDLING *****

    /** 
     * Returns the set of variables of operations that should currently be 
     * executed, i.e., the operation variables of tasks that are part of 
     * the current time's interval and have an execution state of 
     * IN_EXECUTION.
     */
    public ArrayList retrieveTasksToBeExecuted() {

	DRCInterval interval = (DRCInterval) currentTimeILE.getElement();
	DRCTask task;

	ArrayList taskList = new ArrayList(interval.tasks.size());

	Iterator it = interval.tasks.iterator();
	while (it.hasNext()) {
	    task = (DRCTask) it.next();
	    if (task.executionState 
		== DiscreteResourceConstraint.IN_EXECUTION) {
		taskList.add(task.operationVar);
	    }
	}

	return taskList;

    }

    /** 
     * Updates internal structures and costs if the current time has changed.
     * @param oldTime  The old value of the current time.
     */
    private void propagateCurrentTimeChange(long oldTime) {

	updateIntervalCapacityCosts(currentTimeILE);

	if (costFunctionReference[WORKLOAD] != null) {

	    Iterator it;
	    DRCTask task;
	    long costBuffer = 0;

	    if (oldTime < currentTime) {
		while (currentTimeILE.getEnd() <= currentTime) {

		    // Moving current time interval forward
		    
		    currentTimeILE = currentTimeILE.getNext();
		    updateIntervalCapacityCosts(currentTimeILE);

		    it = ((DRCInterval) currentTimeILE.getElement()).tasks.iterator();
		    while (it.hasNext()) {
			task = (DRCTask) it.next();
			if (task.begin == currentTimeILE.getBeginning()) {
			    costBuffer -= task.duration * task.consumption;
			}
		    }

		}
	    } else {
		while (currentTime < currentTimeILE.getBeginning()) {

		    // Moving current time interval backward
		    
		    it = ((DRCInterval) currentTimeILE.getElement()).tasks.iterator();
		    while (it.hasNext()) {
			task = (DRCTask) it.next();
			if (task.begin == currentTimeILE.getBeginning()) {
			    costBuffer += task.duration * task.consumption;
			}
		    }

		    currentTimeILE = currentTimeILE.getPrevious();
		    updateIntervalCapacityCosts(currentTimeILE);

		}
	    } 

	    if (costBuffer != 0) {
		costs[WORKLOAD] += costBuffer;
		costFunctionReference[WORKLOAD].updateGlobalCosts();
	    }

	} else {

	    if (oldTime < currentTime) {
		while (currentTimeILE.getEnd() <= currentTime) {

		    // Moving current time interval forward
		    
		    currentTimeILE = currentTimeILE.getNext();
		    updateIntervalCapacityCosts(currentTimeILE);

		}
	    } else {
		while (currentTime < currentTimeILE.getBeginning()) {

		    // Moving current time interval backward
		    
		    currentTimeILE = currentTimeILE.getPrevious();
		    updateIntervalCapacityCosts(currentTimeILE);

		}
	    } 

	}

    }


    // ***** EXTERNAL DYNAMICS *****

    public void update(double oldValue, ConnectionReference first, 
		       ConnectionReference last) {

	if (!modificationModus 
	    && (((Variable) first.node).value() != oldValue)) {

	    if (last.node instanceof CurrentTimeObjectConstraint) {
		
		long oldTime = currentTime;

		// Current time variable was updated

		currentTime = (long) ((Variable) first.node).value();

		// Checking for a valid current time value
		
		if (currentTime < 0) {
		    throw new Error("Invalid value \"" + currentTime
				    + "\" for current time");
		}

		propagateCurrentTimeChange(oldTime);

		// Updating the capacity's cost function
		
		if (costFunctionReference[CAPACITY_OVERLOAD] != null) {
		    costFunctionReference[CAPACITY_OVERLOAD].updateGlobalCosts();
		}

	    } else {

		if (last.edgeID == Registry.CAPACITY_EDGE) {
		    
		    // Capacity variable was updated
		    
		    long newCapacity = (long) capacityVariable.value();

		    // Checking for a valid capacity value

		    if (newCapacity < 0) {
			throw new Error("Invalid value \"" + newCapacity
					+ "\" for capacity");
		    } 

		    updateCapacityVariable(newCapacity);
		    
		    // Updating the capacity's cost function
		    
		    if (costFunctionReference[CAPACITY_OVERLOAD] != null) {
			costFunctionReference[CAPACITY_OVERLOAD].updateGlobalCosts();
		    }
		    
		} else {
		    
		    if (last.node instanceof ActionTaskObjectConstraint) {
			
			if ((first.edgeID == Registry.BEGIN_EDGE) 
			    || (first.edgeID == Registry.DURATION_EDGE)) {
			    
			    // A task's begin or duration variable has 
			    // been changed

			    ActionTaskObjectConstraint taskConstraint = 
				(ActionTaskObjectConstraint) last.node;
			    
			    // Fetching related DRCTask for the 
			    // ActionTaskObjectConstraint
			    
			    DRCTask task = (DRCTask) taskConstraintsHT.get(taskConstraint);
			    
			    // Removing task
			    
			    removeTask(task);

			    // Checking for a valid duration/begin values and 
			    // updating internal task variables
		    
			    if (first.edgeID == Registry.DURATION_EDGE) {
				task.duration = 
				    (long) task.durationVar.value();
				if (task.duration <= 0) {
				    throw new Error("Invalid value \"" 
						    + task.duration
						    + "\" for task duration");
				}
			    } else {
				task.begin = (long) task.beginVar.value();
				if (task.begin < 0) {
				    throw new Error("Invalid value \"" 
						    + task.begin
						    + "\" for task begin");
				}
			    }
			    task.end = task.begin + task.duration;
		    
			    // Re-inserting task
			    
			    addTask(task);

			    // Updating the capacity's cost function
			    
			    if (costFunctionReference[CAPACITY_OVERLOAD] != null) {
				costFunctionReference[CAPACITY_OVERLOAD].updateGlobalCosts();
			    }
			    
			    // Updating the makespan's cost function
			    
			    if ((costFunctionReference[MAKESPAN] != null) && (costs[MAKESPAN] != intervalList.getLast().getBeginning())) {
				costs[MAKESPAN] = intervalList.getLast().getBeginning();
				costFunctionReference[MAKESPAN].updateGlobalCosts();
			    }
			    
			    // Updating the workload cost function

			    if (costFunctionReference[WORKLOAD] != null) {
				costFunctionReference[WORKLOAD].updateGlobalCosts();
			    }

			} else {
			    
			    if (first.edgeID == Registry.CONSUMPTION_EDGE) {
				
				// A task's consumption variable has 
				// been changed
				
				ActionTaskObjectConstraint taskConstraint = 
				    (ActionTaskObjectConstraint) last.node;
				
				// Fetching related DRCTask for the 
				// ActionTaskObjectConstraint
				
				DRCTask task = (DRCTask) taskConstraintsHT.get(taskConstraint);
				
				// Checking for a valid consumption value and 
				// updating internal variable
				
				long oldConsumption = task.consumption;
				task.consumption = 
				    (long) task.consumptionVar.value();
				
				if (task.consumption < 0) {
				    throw new Error("Invalid value \"" + task.consumption + "\" for task's consumption");
				}
				
				// Updating internal structures

				propagateConsumptionChange(task, 
							   oldConsumption);
				
				// Updating the capacity's cost function
				
				if (costFunctionReference[CAPACITY_OVERLOAD] != null) {
				    costFunctionReference[CAPACITY_OVERLOAD].updateGlobalCosts();
				}

				// Updating the workload cost function

				if (costFunctionReference[WORKLOAD] != null) {
				    costFunctionReference[WORKLOAD].updateGlobalCosts();
				}

			    } else {
				
				if (first.edgeID == 
				    Registry.EXECUTIONSTATE_EDGE) {
				    
				    // A task's execution state variable 
				    // has been changed

				    ActionTaskObjectConstraint taskConstraint = (ActionTaskObjectConstraint) last.node;
				    
				    // Fetching related DRCTask for the 
				    // ActionTaskObjectConstraint
		    
				    DRCTask task = (DRCTask) taskConstraintsHT.get(taskConstraint);

				    // Updating internal value copy

				    task.executionState = 
					(int) task.executionStateVar.value();

				    // Checking for a valid execution state 
				    // value
			    
				    if ((task.executionState != PENDING) && (task.executionState != IN_EXECUTION) && (task.executionState != DONE)) {
					throw new Error("Invalid value \"" + task.executionState + "\" for task's execution state");
				    }

				}

			    }
			}
		    }
		}
	    }

	}
	fireChange(this);
    }

    public void connectNode(ConnectionReference first, 
			    ConnectionReference last) {

	if (last.node instanceof CurrentTimeObjectConstraint) {
	    
	    // Current time variable/constraint was connected
	    
	    Node currentTimeConstraint = last.node;
	    
	    // Fetching current time's variable
	    
	    Variable variable;
	    synchronized(currentTimeConstraint.connections) {
		variable = currentTimeConstraint.connections.getSpecificVariable(Registry.PLAIN_EDGE);
	    }
	    if (variable == null) {
		throw new Error("The connected CurrentTimeObjectConstraint \"" + currentTimeConstraint + "\" has no variable");
	    }
	    
	    currentTime = (long) variable.value();

	    // Checking for a valid current time value

	    if (currentTime < 0) {
		throw new Error("Invalid value \"" + currentTime
				+ "\" for current time");
	    }

	} else {

	    if (last.edgeID == Registry.CAPACITY_EDGE) {

		// Capacity variable was connected
		
		capacityVariable = (Variable) last.node;
		modificationModus = true;
		gsc.addIntegerConstraint(capacityVariable);
		modificationModus = false;
		long newCapacity = (long) capacityVariable.value();

		// Checking for a valid capacity value

		if (newCapacity < 0) {
		    throw new Error("Invalid value \"" + newCapacity
				    + "\" for capacity");
		} 

		updateCapacityVariable(newCapacity);
	
	    } else {
		
		if (last.edgeID == Registry.RESOURCETYPE_EDGE) {
		    
		    // Resource type variable was connected
		    
		    // TODO
		    
		} else {
		    
		    if (last.node instanceof ActionTaskObjectConstraint) {
			
			// ActionTaskObjectConstraint was connected
			
			ActionTaskObjectConstraint taskConstraint = 
			    (ActionTaskObjectConstraint) last.node;
			
			// Checking whether task has already been connected
			
			DRCTask newTask = 
			    (DRCTask) taskConstraintsHT.get(taskConstraint);
			if (newTask != null) {
			    throw new Error("The ActionTaskObjectConstraint \"" + taskConstraint + "\" has already been connected");
			}
			
			// Creating a DRCTask for the 
			// ActionTaskObjectConstraint
			
			newTask = new DRCTask(taskConstraint);
			newTask.taskLE = taskList.addListElement(newTask);
			taskConstraintsHT.put(taskConstraint, newTask);
			
			// Fetching task's variables
			
			Variable tcBeginVar;
			Variable tcDurationVar;
			Variable tcConsumptionVar;
			Variable tcExecutionStateVar;
			Variable tcOperationVar;
			Variable tcResourceTypeVar;

			synchronized(taskConstraint.connections) {
			    tcBeginVar = taskConstraint.connections.getSpecificVariable(Registry.BEGIN_EDGE);
			    tcDurationVar = taskConstraint.connections.getSpecificVariable(Registry.DURATION_EDGE);
			    tcConsumptionVar = taskConstraint.connections.getSpecificVariable(Registry.CONSUMPTION_EDGE);
			    tcExecutionStateVar = taskConstraint.connections.getSpecificVariable(Registry.EXECUTIONSTATE_EDGE);
			    tcOperationVar = taskConstraint.connections.getSpecificVariable(Registry.OPERATION_EDGE);
			    tcResourceTypeVar = taskConstraint.connections.getSpecificVariable(Registry.RESOURCETYPE_EDGE);
			}
			if (tcBeginVar == null) {
			    throw new Error("The connected ActionTaskObjectConstraint \"" + taskConstraint + "\" has no begin variable");
			}
			if (tcDurationVar == null) {
			    throw new Error("The connected ActionTaskObjectConstraint \"" + taskConstraint + "\" has no duration variable");
			}
			if (tcConsumptionVar == null) {
			    throw new Error("The connected ActionTaskObjectConstraint \"" + taskConstraint + "\" has no consumption variable");
			}
			if (tcExecutionStateVar == null) {
			    throw new Error("The connected ActionTaskObjectConstraint \"" + taskConstraint + "\" has no execution state variable");
			}
			if (tcOperationVar == null) {
			    throw new Error("The connected ActionTaskObjectConstraint \"" + taskConstraint + "\" has no operation variable");
			}
			if (tcResourceTypeVar == null) {
			    throw new Error("The connected ActionTaskObjectConstraint \"" + taskConstraint + "\" has no resource type variable");
			}

			// Processing task's begin variable

			modificationModus = true;
			gsc.addIntegerConstraint(tcBeginVar);
			modificationModus = false;
			newTask.beginVar = tcBeginVar;
			newTask.begin = (long) tcBeginVar.value();

			// Processing task's duration variable
			
			modificationModus = true;
			gsc.addIntegerConstraint(tcDurationVar);
			modificationModus = false;
			newTask.durationVar = tcDurationVar;
			newTask.duration = (long) tcDurationVar.value();

			// Checking for a valid begin/duration value
			
			if (newTask.begin < 0) {
			    throw new Error("Invalid value \"" + newTask.begin
					    + "\" for task begin");
			}
    			if (newTask.duration <= 0) {
			    throw new Error("Invalid value \"" 
					    + newTask.duration
					    + "\" for task duration");
			}

			// Setting internal DRCTask task's end variable
			
			newTask.end = newTask.begin + newTask.duration; 
		    
			// Processing task's consumption variable
		
			modificationModus = true;
			gsc.addIntegerConstraint(tcConsumptionVar);
			modificationModus = false;
			newTask.consumptionVar = tcConsumptionVar;
			newTask.consumption = (long) tcConsumptionVar.value();

			// Checking for a valid consumption value
			
			if (newTask.consumption < 0) {
			    throw new Error("Invalid value \"" 
					    + newTask.consumption
					    + "\" for task's consumption");
			}
			
			// Processing task's execution state variable
						
			modificationModus = true;
			gsc.addIntegerConstraint(tcExecutionStateVar);
			modificationModus = false;
			newTask.executionStateVar = tcExecutionStateVar;
			newTask.executionState = 
			    (int) tcExecutionStateVar.value();

			// Checking for a valid execution state value
			
			if ((newTask.executionState != PENDING) 
			    && (newTask.executionState != IN_EXECUTION) 
			    && (newTask.executionState != DONE)) {
			    throw new Error("Invalid value \"" + newTask.executionState + "\" for task's execution state");
			}
			
			// Processing task's operation variable
			
      			modificationModus = true;
			gsc.addIntegerConstraint(tcOperationVar);
			modificationModus = false;
			newTask.operationVar = tcOperationVar;

			// Processing task's resource type variable
						
			modificationModus = true;
			gsc.addIntegerConstraint(tcResourceTypeVar);
			modificationModus = false;
			newTask.resourceTypeVar = tcResourceTypeVar;

			// Updating interval list structures and costs
			
			addTask(newTask);
			
			// Updating the makespan's cost function
			
			if ((costFunctionReference[MAKESPAN] != null) && (costs[MAKESPAN] != intervalList.getLast().getBeginning())) {
			    costs[MAKESPAN] = intervalList.getLast().getBeginning();
			    costFunctionReference[MAKESPAN].updateGlobalCosts();
			}
			
			// Updating the workload cost function
			
			if (costFunctionReference[WORKLOAD] != null) {
			    costFunctionReference[WORKLOAD].updateGlobalCosts();
			}

		    } else {
			throw new Error("Invalid connection of \"" + first.node + "\" to the DiscreteResourceConstraint");
		    }
		}
	    }
	
	    // Updating the capacity's cost function
	    
	    if (costFunctionReference[CAPACITY_OVERLOAD] != null) {
		costFunctionReference[CAPACITY_OVERLOAD].updateGlobalCosts();
	    }

	}

    }

    public void disconnectNode(ConnectionReference first, 
			       ConnectionReference last) {

	if (last.edgeID == Registry.CAPACITY_EDGE) {

	    // Capacity variable was disconnected

	    capacityVariable = null;
	    modificationModus = true;
	    gsc.removeIntegerConstraint((Variable) last.node);
	    modificationModus = false;

	} else {

	    if (last.edgeID == Registry.RESOURCETYPE_EDGE) {
	    
		// Resource type variable was disconnected

		// TODO

	    } else {
		
		if (last.node instanceof ActionTaskObjectConstraint) {
		    
		    // ActionTaskObjectConstraint was disconnected
		    
		    ActionTaskObjectConstraint taskConstraint = 
			(ActionTaskObjectConstraint) last.node;
		    
		    // Removing related DRCTask for the 
		    // ActionTaskObjectConstraint

		    DRCTask task = 
			(DRCTask) taskConstraintsHT.get(taskConstraint);
		    taskList.removeListElement(task.taskLE);
		    task.taskLE = null;
		    taskConstraintsHT.remove(taskConstraint);
		    
		    // Removing IntegerConstraints from the variables
		    
		    modificationModus = true;
		    gsc.removeIntegerConstraint(task.beginVar);
		    gsc.removeIntegerConstraint(task.durationVar);
		    gsc.removeIntegerConstraint(task.operationVar);
		    gsc.removeIntegerConstraint(task.executionStateVar);
		    gsc.removeIntegerConstraint(task.consumptionVar);
		    gsc.removeIntegerConstraint(task.resourceTypeVar);
		    modificationModus = false;
		    
		    // Updating interval list structures and costs
		    
		    removeTask(task);
		    
		    // Updating the makespan's cost function
		    
		    if ((costFunctionReference[MAKESPAN] != null) 
			&& (costs[MAKESPAN] != intervalList.getLast().getBeginning())) {
			costs[MAKESPAN] = intervalList.getLast().getBeginning();
			costFunctionReference[MAKESPAN].updateGlobalCosts();
		    }

		    // Updating the workload cost function
		    
		    if (costFunctionReference[WORKLOAD] != null) {
			costFunctionReference[WORKLOAD].updateGlobalCosts();
		    }

		}
	    }
	}

	// Updating the capacity's cost function
	
	if (costFunctionReference[CAPACITY_OVERLOAD] != null) {
	    costFunctionReference[CAPACITY_OVERLOAD].updateGlobalCosts();
	}

    }


    // ***** IMPROVEMENT (Service methods) *****

    /** 
     * Disconnects and removes an action task from the constraint if
     * all of the corresponding action's action tasks have a pending 
     * execution state. Returns whether the deletion was successful.
     * @param task  The task to be removed.
     */
    private boolean deleteTask(DRCTask task) {

	// Checking for valid execution states of action's tasks

	ConventionalConstraint actionConstraint;
	Object tasks;

	synchronized(task.taskConstraint.connections) {
	    actionConstraint = (ConventionalConstraint) task.taskConstraint.connections.getSpecificConstraint(Registry.ACTION_OBJECTCONSTRAINT, Registry.PLAIN_EDGE);
	}
	if (actionConstraint != null) {

	    // Retrieving action's action tasks

	    synchronized(actionConstraint.connections) {
		tasks = actionConstraint.connections.getSpecificNodes(Registry.ACTIONTASK_OBJECTCONSTRAINT);
	    }
	    if (tasks instanceof HashSet) {
		tasks = (HashSet) ((HashSet) tasks).clone();
	    } else {
		if (tasks instanceof ConnectedNode) {
		    Object node = tasks;
		    tasks = new HashSet();
		    ((HashSet) tasks).add(node);
		}
	    }

	    // Checking for valid execution states

	    if (tasks != null) {

		ActionTaskObjectConstraint actionTask;
		Variable executionVariable;
		Iterator it = ((HashSet) tasks).iterator();

		while (it.hasNext()) {
		    actionTask = (ActionTaskObjectConstraint) ((ConnectedNode) it.next()).node;
		    synchronized(actionTask.connections) {
			executionVariable = actionTask.connections.getSpecificVariable(Registry.EXECUTIONSTATE_EDGE);
		    }
		    if (((long) executionVariable.value()) 
			!= DiscreteResourceConstraint.PENDING) {
			return false;
		    }
		}

	    }

	}

	// Disconnection from the discrete resource constraint

	Embedding e = new Embedding(1);
	gsc.extendEmbedding_internal(e, task.taskConstraint, 
				     Registry.PLAIN_EDGE, Registry.INCOMING);
	gsc.reduceConstraint_internal(this, e);

        // Deletion and disconnection from the task constraint

        gsc.disconnectAndRemoveConstraint_internal(task.taskConstraint, Registry.ACTION_OBJECTCONSTRAINT, e);

	return true;

    }

    /** 
     * Calculates the maximal distance a task can be shifted to the
     * previous or following interval. The (relative) shift is stored 
     * in the task's tempValue variable and returned.
     */
    private long calculateMaxShiftDistance(IntervalListElement ile, 
					   DRCTask task) {

	DRCInterval interval = (DRCInterval) ile.getElement();

	long minTime = currentTime + repairTimeBuffer;
        long maxEnd = (long) task.beginVar.upperBound();
        if (Long.MAX_VALUE - maxEnd > task.duration) {
            maxEnd = task.duration + maxEnd;
        } else {
            maxEnd = Long.MAX_VALUE;
        }

	// Calculating how many claims a better interval would maximally have

        long maxClaims = interval.claims - task.consumption - 1;
	if (maxClaims < 0) {
	    maxClaims = 0;
	}

	// Checking task's previous interval

	boolean previousIntervalOK = false;
	IntervalListElement previousILE = task.firstInterval.getPrevious();
	if (previousILE != null) {
	    interval = (DRCInterval) previousILE.getElement();
	    if ((minTime < previousILE.getEnd())
		&& (interval.claims <= maxClaims)) {
		previousIntervalOK = true;
	    }
	}

	// Checking task's following interval

	boolean nextIntervalOK = false;
	IntervalListElement nextILE = task.lastInterval.getNext();
	if (nextILE != null) {
	    interval = (DRCInterval) nextILE.getElement();
	    if ((minTime + task.duration <= nextILE.getEnd()) && (maxEnd > nextILE.getBeginning()) && (interval.claims <= maxClaims)) {
		nextIntervalOK = true;
	    }
	}

	// Choosing one shift direction

	if (previousIntervalOK && nextIntervalOK) {

	    long IntervalMiddle = (ile.getBeginning() + ile.getEnd()) / 2;
	    long TaskMiddle = (task.begin + task.end) / 2;

	    if (TaskMiddle < IntervalMiddle) {
		nextIntervalOK = false;
	    } else {
		if (TaskMiddle > IntervalMiddle) {
		    previousIntervalOK = false;
		} else {
		    if (gsc.random.nextInt(2) == 0) {
			nextIntervalOK = false;
		    } else {
			previousIntervalOK = false;
		    }
		}
	    }

	}

	// Storing the shift value

	if (previousIntervalOK) {

	    long newBeginning = task.lastInterval.getBeginning() - task.duration;
	    if (previousILE.getBeginning() > newBeginning) {
		newBeginning = previousILE.getBeginning();
	    }
	    if (newBeginning < minTime) {
		newBeginning = minTime;
	    }

	    task.tempValue = newBeginning - task.begin;

	} else {
	    if (nextIntervalOK) {

		long newBeginning = task.firstInterval.getEnd();
		if (nextILE.getEnd() - task.duration < newBeginning) {
		    newBeginning = nextILE.getEnd() - task.duration;
		}
		if (minTime <= newBeginning) {
		    task.tempValue = newBeginning - task.begin;
		} else {
		    task.tempValue = 0;
		}

	    } else {
		task.tempValue = 0;
	    }
	}

        return task.tempValue;

    }

    /** 
     * A task that involved capacity costs is moved to a position
     * decreasing these costs. Returns whether the move was successful.
     */
    private boolean repositionTask(DRCTask task) {

	long taskDuration = task.duration; 
	long taskDuration90 = (taskDuration * 9) / 10; 
	long taskDuration60 = (taskDuration * 6) / 10; 
	long intervalDuration; 

	long currentPreference;
	long cumulatedPreference = 0;

	// Calculating how many claims an interval without capacity costs 
	// can maximally have

        long maxClaims;
        if (task.consumption <= capacity) {
	    maxClaims = capacity - task.consumption;
	} else {
	    maxClaims = 0;
	}

	// Analyzing better intervals by moving in claims interval list 
	// from last interval to beginning as long as no capacity costs 
        // would be caused

	OrderedShiftDLLElement currentLE = claimsIntervalList.last;
        IntervalListElement currentILE = 
	    (IntervalListElement) currentLE.element;
	DRCInterval interval = (DRCInterval) currentILE.getElement();

	long minTime = currentTime + repairTimeBuffer;
	long maxEnd = (long) task.beginVar.upperBound();
	if (Long.MAX_VALUE - maxEnd > taskDuration) {
	    maxEnd = taskDuration + maxEnd;
	} else {
	    maxEnd = Long.MAX_VALUE;
	}
	if (minTime + taskDuration > maxEnd) {
	    return false;
	}

	do {

	    if ((minTime < currentILE.getEnd()) 
		&& (currentILE.getBeginning() < maxEnd)) {

		// Analyzing preference for this interval

		if (minTime <= currentILE.getBeginning()) {
		    intervalDuration = currentILE.getDuration();
		} else {
		    intervalDuration = currentILE.getEnd() - minTime;
		}
		if (maxEnd < currentILE.getEnd()) {
		    intervalDuration -= currentILE.getEnd() - maxEnd;
		}

		if (intervalDuration >= taskDuration) {
		    currentPreference = 10;
		} else {
		    if (intervalDuration < taskDuration60) {
			currentPreference = 2;
		    } else {
			if (intervalDuration < taskDuration90) {
			    currentPreference = 4;
			} else {
			    currentPreference = 7;
			}
		    }
		}
		cumulatedPreference += currentPreference;
		
		// Storing preference value for later choice retrieval
		
		interval.tempValue = currentPreference;
	    } else {
		interval.tempValue = 0;
	    }

	    // Advancing to previous interval
		
	    currentLE = currentLE.previous;
	    currentILE = (IntervalListElement) currentLE.element;
	    interval = (DRCInterval) currentILE.getElement();

	} while (interval.claims <= maxClaims);

	if (cumulatedPreference > 0) {
	    
	    currentLE = currentLE.next;
	    
	    // Random selection of an interval according to the preference 
	    // values
	    
	    long chosenPreference = 
		Math.abs(gsc.random.nextLong()) % cumulatedPreference;
	    boolean found = false;
	    
	    while (!found) {
		currentILE = (IntervalListElement) currentLE.element;

		interval = (DRCInterval) currentILE.getElement();
		
		chosenPreference -= interval.tempValue;
		if (!(found = (chosenPreference < 0))) {
		    currentLE = currentLE.next;
		}
	    }

	    // Executing the move

	    if (currentILE.getBeginning() >= minTime) {
		gsc.setVariable_internal(task.beginVar, currentILE.getBeginning());
	    } else {
		gsc.setVariable_internal(task.beginVar, minTime);
	    }

	    return true;

	} else {
	    return false;
	}

    }


    // ***** CAPACITY IMPROVEMENT (Heuristics) *****

    /** 
     * An inconsistent interval is selected and one of its tasks is shifted
     * to the previous or following interval. Returns whether the heuristic 
     * was able to succeed.
     */
    private boolean heuristic_ShiftTask() {
	
        // Selecting an inconsistent interval

	OrderedShiftDLLElement currentLE = capacityCostsIntervalList.first;
	DRCInterval interval = null;

	long chosenCosts = Math.abs(gsc.random.nextLong()) 
	    % costs[CAPACITY_OVERLOAD];
	boolean found = false;

	while (!found) {
	    interval = (DRCInterval) ((IntervalListElement) currentLE.element).getElement();
	    chosenCosts -= interval.capacityCosts;
	    if (!(found = (chosenCosts < 0))) {
		currentLE = currentLE.next;
	    }
	}

	IntervalListElement currentILE = 
	    (IntervalListElement) currentLE.element;

	// Analyzing the interval's tasks

	DRCTask task = null;
	long cumulativeShifts = 0;

	Iterator it = interval.tasks.iterator();

	while (it.hasNext()) {
	    task = (DRCTask) it.next();
	    if (task.executionState == DiscreteResourceConstraint.PENDING) {
		cumulativeShifts += 
		    Math.abs(calculateMaxShiftDistance(currentILE, task));
	    }
	}

	if (cumulativeShifts > 0) {

	    // Selecting one of the interval's tasks to be moved

	    long chosenShift = 
		Math.abs(gsc.random.nextLong()) % cumulativeShifts;
	    it = interval.tasks.iterator();
	    found = false;

	    while (!found) {
		task = (DRCTask) it.next();
		if (task.executionState 
		    == DiscreteResourceConstraint.PENDING) {
		    chosenShift -= Math.abs(task.tempValue);
		}
		found = (chosenShift < 0);
	    }

	    // Shifting the task

	    if (Registry.LOG_CHANGES) {
		if (operationDescriptions != null) {
		    System.out.println("DRC ShiftTask: " + operationDescriptions.get(new Long((long) task.operationVar.value())));
		} else {
		    System.out.println("DRC ShiftTask: " + task);
		}
	    }

	    gsc.setVariable_internal(task.beginVar, 
				     task.begin + task.tempValue);

	    return true; 

	} else {
	    return false; 
	}

    }

    /** 
     * An inconsistent interval is selected and one of its tasks is moved 
     * to an interval such that no capacity costs are caused. Returns 
     * whether the heuristic was able to succeed.
     */
    private boolean heuristic_ImproveInterval() {
	
        // Selecting an inconsistent interval

	OrderedShiftDLLElement currentLE = capacityCostsIntervalList.first;
	DRCInterval interval = null;

	long chosenCosts = Math.abs(gsc.random.nextLong()) 
	    % costs[CAPACITY_OVERLOAD];
	boolean found = false;

	while (!found) {
	    interval = (DRCInterval) ((IntervalListElement) currentLE.element).getElement();
	    chosenCosts -= interval.capacityCosts;
	    if (!(found = (chosenCosts < 0))) {
		currentLE = currentLE.next;
	    }
	}

	// Selecting one of the interval's tasks to be moved

	DRCTask task = null;

	Iterator it = interval.tasks.iterator();
	long chosenClaims = Math.abs(gsc.random.nextLong()) % interval.claims;
	found = false;

	while (!found) {
	    task = (DRCTask) it.next();
	    chosenClaims -= task.consumption;
	    found = (chosenClaims < 0);
	}

	// Moving the task to a better position

	if (Registry.LOG_CHANGES) {
	    if (operationDescriptions != null) {
		System.out.println("DRC ImproveInterval: " + operationDescriptions.get(new Long((long) task.operationVar.value())));
	    } else {
		System.out.println("DRC ImproveInterval: " + task);
	    }
	}

	if (task.executionState == DiscreteResourceConstraint.PENDING) {
	    return repositionTask(task);
	} else {
	    if (Registry.LOG_CHANGES) {
		System.out.println("failed");
	    }
	    return false;
	}

    }

    /** 
     * An inconsistent task is selected and moved to an interval such that 
     * no capacity costs are caused. Returns whether the heuristic was able 
     * to succeed.
     */
    private boolean heuristic_ImproveTask() {
	
        // Selecting an inconsistent task

	OrderedShiftDLLElement currentLE = taskList.first;
	DRCTask task = null;

	long chosenCosts = Math.abs(gsc.random.nextLong()) % taskCapacityCosts;
	boolean found = false;

	while (!found) {
	    task = (DRCTask) currentLE.element;
	    chosenCosts -= task.capacityCosts;
	    if (!(found = (chosenCosts < 0))) {
		currentLE = currentLE.next;
	    }
	}

	// Moving the task to a better position

	if (Registry.LOG_CHANGES) {
	    if (operationDescriptions != null) {
		System.out.println("DRC ImproveTask: " + operationDescriptions.get(new Long((long) task.operationVar.value())));
	    } else {
		System.out.println("DRC ImproveTask: " + task);
	    }
	}

	if (task.executionState == DiscreteResourceConstraint.PENDING) {
	    return repositionTask(task);
	} else {
	    if (Registry.LOG_CHANGES) {
		System.out.println("failed");
	    }
	    return false;
	}

    }

    /** 
     * An inconsistent task is selected and deleted. 
     * Returns whether the heuristic was able to succeed.
     */
    private boolean heuristic_DeleteTask() {
	
        // Selecting an inconsistent task

	OrderedShiftDLLElement currentLE = taskList.first;
	DRCTask task = null;

	long chosenCosts = Math.abs(gsc.random.nextLong()) % taskCapacityCosts;
	boolean found = false;

	while (!found) {
	    task = (DRCTask) currentLE.element;
	    chosenCosts -= task.capacityCosts;
	    if (!(found = (chosenCosts < 0))) {
		currentLE = currentLE.next;
	    }
	}

	// Deleting the task

	boolean success = deleteTask(task);
	if (Registry.LOG_CHANGES && success) {
	    if (operationDescriptions != null) {
		System.out.println("DRC DeleteTask: " + operationDescriptions.get(new Long((long) task.operationVar.value())));
	    } else {
		System.out.println("DRC DeleteTask: " + task);
	    }
	}
	return success;

    }

    /** 
     * Improving the current capacity costs.
     */
    private void improveCapacity() {

	int r = gsc.random.nextInt(100);

	if (gsc.structuralChangesEnabled && (r < 10)) {
	    heuristic_DeleteTask();
	} else {
	    if (r < 75) {
		heuristic_ShiftTask();
	    } else {
		if (r < 80) {
		    heuristic_ImproveInterval();
		} else {
		    heuristic_ImproveTask();
		}
	    }
	}

    }


    // ***** MAKESPAN IMPROVEMENT *****

    /** 
     * Improving the current makespan costs by deleting one of the tasks 
     * that end at the makespan time. Returns whether the heuristic was 
     * able to succeed.
     */
    private boolean heuristic_DeleteLatestTask() {

	// Choosing a task of the last interval

	DRCInterval interval = 
	    (DRCInterval) intervalList.getLast().getPrevious().getElement();
	DRCTask task = null;

	int chosenTask = gsc.random.nextInt(interval.tasks.size());
	Iterator it = interval.tasks.iterator();

	while (chosenTask-- >= 0) {
	    task = (DRCTask) it.next();
	}

	// Deleting the task

	boolean success = deleteTask(task);
	if (Registry.LOG_CHANGES && success) {
	    if (operationDescriptions != null) {
		System.out.println("DRC DeleteLatestTask: " + operationDescriptions.get(new Long((long) task.operationVar.value())));
	    } else {
		System.out.println("DRC DeleteLatestTask: " + task);
	    }
	}
	return success;

    }

    /** 
     * Improving the current makespan costs by shifting one of the tasks 
     * that end at the makespan time. Returns whether the heuristic was 
     * able to succeed.
     */
    private boolean heuristic_ShiftLatestTask() {

	DRCInterval interval = 
	    (DRCInterval) intervalList.getLast().getPrevious().getElement();

	// Choosing a task

	int chosenTask = gsc.random.nextInt(interval.tasks.size());
	Iterator it = interval.tasks.iterator();
	DRCTask task = null;

	while (chosenTask-- >= 0) {
	    task = (DRCTask) it.next();
	}

	// Shifting the task

	if (Registry.LOG_CHANGES) {
	    if (operationDescriptions != null) {
		System.out.println("DRC ShiftLatestTask: " + operationDescriptions.get(new Long((long) task.operationVar.value())));
	    } else {
		System.out.println("DRC ShiftLatestTask: " + task);
	    }
	}

	if ((task.executionState == DiscreteResourceConstraint.PENDING) 
	    && (task.begin > currentTime + repairTimeBuffer)) {
	    gsc.setVariable_internal(task.beginVar, task.begin - 1);
	    return true;
	} else {
	    if (Registry.LOG_CHANGES) {
		System.out.println("failed");
	    }
	    return false;
	}

    }

    /** 
     * Improving the current makespan costs.
     */
    private void improveMakespan() {

	if (gsc.structuralChangesEnabled && (gsc.random.nextInt(100) == 0)) {
	    heuristic_DeleteLatestTask();
	} else {
	    if (!heuristic_ShiftLatestTask()) {
		if (gsc.structuralChangesEnabled) {
		    heuristic_DeleteLatestTask();
		}
	    }
	}

    }

    // ***** WORKLOAD IMPROVEMENT *****

    /** 
     * Improving the current workload by deleting one of the tasks after 
     * the current time. Returns whether the heuristic was able to succeed.
     */
    private boolean heuristic_DeleteWorkloadTask() {

	// Selecting a task after the current time
	
	OrderedShiftDLLElement currentLE = taskList.first;
	DRCTask task = null;
	
	long chosenCosts = Math.abs(gsc.random.nextLong()) % costs[WORKLOAD];
	boolean found = false;
	
	while (!found) {
	    task = (DRCTask) currentLE.element;
	    if (task.begin > currentTime) {
		chosenCosts -= task.duration * task.consumption;
	    }
	    if (!(found = (chosenCosts < 0))) {
		currentLE = currentLE.next;
	    }
	}
	
	// Deleting the task
	
	boolean success = deleteTask(task);
	if (Registry.LOG_CHANGES && success) {
	    if (operationDescriptions != null) {
		System.out.println("DRC DeleteWorkloadTask: " + operationDescriptions.get(new Long((long) task.operationVar.value())));
	    } else {
		System.out.println("DRC DeleteWorkloadTask: " + task);
	    }
	}
	return success;

    }

    /** 
     * Improving the current workload costs.
     */
    private void improveWorkload() {
	if (gsc.structuralChangesEnabled) {
	    heuristic_DeleteWorkloadTask();
	}
    }

    public void improveCosts(int costID) {

	switch (costID) {
	case CAPACITY_OVERLOAD: {
	    improveCapacity();
	    break;
	}
	case MAKESPAN: {
	    improveMakespan();
	    break;
	}
	case WORKLOAD: {
	    improveWorkload();
	    break;
	}
	default: {
	    throw new Error("The cost ID \"" + costID + 
			    "\" is not supported by this constraint");	
	}
	}

    }

}
