
package com.ai_center.dragonbreath.constraints;

import java.lang.Math;
import java.util.*;

import org.jdom.Element;
import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.Registry;

/** 
 * This constraint forces the that the two connected Variables are a multiple of each other.
 * The coefficient can be determined at creation.
 *
 * @author tim baermann
 * @version $Revision: 1.1 $
 * @see Variable
 */

public class ExponentialConstraint extends ConventionalConstraint {
    
 
    // **********************
    // ***** ATTRIBUTES *****
    // **********************

    /** 
     * The constraint type's name.
     */
    private static final String TYPE_NAME = new String("Exponential");

    /** 
     * The constraint's base graph.
     */
    private static final Embedding BASE_GRAPH = new Embedding(2);
    
    /** 
     * The constraint's number of extension graphs.
     */
    private static final int EXTENSION_GRAPH_NUMBER = 0;
    
    /** 
     * input^exponent = output
     */
    private Variable output = null;
    
    /** 
     * input^exponent = output
     */
    private Variable input = null;
    
    /** 
     * The exponent.
     */
    private double exponent = 1;
    
    /**
     * Is the exponent even?
     */
    private boolean expIsEven = false;
    /**
     * Is the exponent integer?
     */
    private boolean expIsInteger = false;
    
    private double epsilon = 0.00000000001; //allowed error when computing integers from doubles
    /**
     * The current variable value of input.
     */
    private double inputValue = 0;
    
    /**
     * The current variable value of output
     */
    private double outputValue = 0;
    
    // *******************
    // ***** METHODS *****
    // *******************
    /**
     * Sets the Coefficient.
     * @param nodeData  The nodeData as stored in file. 
     */
    public void setAuxiliaryNodeData(Element nodeData) {    
        if (nodeData != null) {
            String cS = nodeData.getAttributeValue( "Exponent");
            if (cS != null) {
                double exponent = java.lang.Double.parseDouble(cS);
                setExponent(exponent);
            }
        }
    }
    
     /**
     * Before storing a node this function is called, and the node can add its own data to the XML-data.
     * Adds the Coefficient.
     * @param nodeData  Where the data should be stored.
     */
    public void getAuxiliaryNodeData(Element nodeData) {
        nodeData.setAttribute("Exponent", java.lang.String.valueOf(exponent) );
    }
    
    /**
     * Sets how many times output should be input
     */
    public void setExponent(double newExponent) {
        exponent = newExponent;
        expIsInteger = exponent == Math.round(exponent);
        expIsEven = exponent / 2 == Math.round(exponent/2);
        if ((graphObject != null) && (graphObject instanceof ExponentialConstraintGraphics)) {
            ((ExponentialConstraintGraphics)graphObject).updateExponent();
        }
        updateCosts();
    }
    
    /** 
     * Adding specialized graphics for the constraint.
     */
    public void createGraphics() {
	graphObject = (ConventionalConstraintGraphics) 
	    new ExponentialConstraintGraphics(this, gsc);
    }
    
    /**
     * Returns the coefficient
     */
    public double getExponent() {
        return exponent;
    }
    public void create_graphs() {

	// Creating the base graph
	
	BASE_GRAPH.addConnectedNode(Registry.VARIABLE, 
				    Registry.EXPONENT_OUTPUT_EDGE, Registry.INCOMING);
	BASE_GRAPH.addConnectedNode(Registry.VARIABLE, 
				    Registry.EXPONENT_INPUT_EDGE, Registry.INCOMING);
    }

    public int getNumberOfCosts() {
	return 1;
    }

    public String getCostName(int costID) {
	if (costID == 0) {
	    return new String("proportion inconsistency");
	} else {
	    throw new Error("Invalid cost ID");
	}
    }

    public String getTypeName() {
	return TYPE_NAME; 
    }

    public Embedding getBaseGraph() {
	return BASE_GRAPH;
    }

    public int getExtensionGraphNumber() {
	return EXTENSION_GRAPH_NUMBER;
    }

    public Embedding getExtensionGraph(int graphID) {
	return null;
    }

    public int getMaximalConnectionNumber() {
	return 2;
    }

    public void initializeCostType(int costID) {
	// Nothing to do
    }

    /** 
     * Calculates the constraint's costs.
     */
    private void updateCosts() {
        if (input == null || output == null) {
            costs[0] = 0;
        } else if (input.isInteger() && output.isInteger()) {
            costs[0] = (long) Math.round(Math.abs((Math.pow(inputValue, exponent) - outputValue)));
        } else {
            costs[0] = (long) Math.ceil(Math.abs(Math.pow(inputValue, exponent) - outputValue));
        }
	costFunctionReference[0].updateGlobalCosts();
    }

    public void update(double oldValue, ConnectionReference first, 
			  ConnectionReference last) {
	Variable variable = (Variable) first.node;
	if (variable == input) {
	    inputValue = variable.value();
	} else {
	    outputValue = variable.value();
	}
	updateCosts();
    }

    public void connectNode(ConnectionReference first, 
			       ConnectionReference last) {
	Variable variable = (Variable) first.node;
	if (first.edgeID == Registry.EXPONENT_OUTPUT_EDGE) {
	    output = variable;
	    outputValue = variable.value();
	    if (input != null) updateCosts();
	} else {
	    input = variable;
	    inputValue = variable.value();
	    if (output != null) updateCosts();
	}
    }

    public void disconnectNode(ConnectionReference first, 
				  ConnectionReference last) {

	if (first.edgeID == Registry.PROPORTION_OUTPUT_EDGE) {
	    output = null;
	    outputValue = 0;
	} else {
	    Variable variable = (Variable) first.node;
	    input = null;
            inputValue = 0;
	}
	updateCosts();
    }

    public void improveCosts(int costID) {

	// Checking for valid cost ID

	if (costID != 0) {
	    throw new Error("The cost ID  \"" + costID + 
			    "\" is not supported by this constraint");
	}
        double newOutputValue = outputValue; 
        double newInputValue = inputValue; 
        double oldInputValue = inputValue; 
        double oldOutputValue = outputValue; 
        if (exponent == 0) {
            gsc.setVariable_internal(output, 1);
            return;
        } else if (gsc.random.nextDouble() > 0.5) {
                newOutputValue = changeOutput(newInputValue);
                if (java.lang.Double.isNaN(newOutputValue)) {
                    newOutputValue = oldOutputValue;
                    newInputValue = changeInput(newOutputValue);
                }
                
        } else {
            //change input
            newInputValue = changeInput(newOutputValue);
            if ( java.lang.Double.isNaN(newInputValue)) {
                newInputValue = oldInputValue;
                newOutputValue = changeOutput(newInputValue);
            }
        }
        
        
        //now handle existing bounds
        boolean solved = false;
        while (! solved) {
            
            solved = true;
            if (output.upperBound() < newOutputValue) {
                newOutputValue = output.upperBound();
                newInputValue  = changeInput(newOutputValue);
                

                if ((input.lowerBound() > newInputValue) ) {// upperBound is right anyway
                     removeOneRange();
                     solved = false;
                }
            }
            if (solved && (output.lowerBound() > newOutputValue)) {//first solve upperbound
                newOutputValue = output.lowerBound();
                newInputValue = changeInput(newOutputValue);
                if (input.upperBound() < newInputValue) {// lowerBound is right anyway
                     removeOneRange();
                     solved = false;
                }
            }
            if (solved && (input.lowerBound() > newInputValue)) {//first solve upperbound
                newInputValue = input.lowerBound();
                newOutputValue = changeOutput(newInputValue);
                if (output.upperBound() < newOutputValue) {// lowerBound is right anyway
                     removeOneRange();
                     solved = false;
                }
            }
            if (solved && (input.upperBound() < newInputValue)) {//first solve upperbound
                newInputValue = input.lowerBound();
                newOutputValue = changeOutput(newInputValue);
                if (output.lowerBound() > newOutputValue) {// lowerBound is right anyway
                     removeOneRange();
                     solved = false;
                }
            }
            if (java.lang.Double.isNaN(newOutputValue) || ( java.lang.Double.isNaN(newInputValue))) {
                newOutputValue = 1;
                newInputValue = 1;
            }
        }
        gsc.setVariable_internal(input, newInputValue);
        gsc.setVariable_internal(output, newOutputValue);
        
    }
    
    /**
     * Returns NaN, if the input has to be changed
     */
    double changeOutput(double in) {
        if (output.isInteger()) {
            throw new Error("Exponential Constraint doesn`t work with integer variables!");
        }
        if (in > 0) {
            return Math.pow(in, exponent);
        } else if (Math.abs(exponent) > 1) {
            if (expIsInteger) {
                return Math.pow(in, exponent);
            } 
        }
        return Double.NaN;
    }
    
    double changeInput(double out) {
        if (input.isInteger()) {
           throw new Error("Exponential Constraint doesn`t work with integer variables!");
        }
        if (out  > 0) {
            return Math.pow(out , 1 / exponent);
        } else if (Math.abs(exponent) < 1) {
            if (Math.round (1 / exponent) - (1 / exponent) < epsilon) {
                return Math.pow(out , Math.round (1 / exponent));
            } 
        } else if (! expIsEven) {
            return - Math.pow(-out , 1 / exponent);
        }
        return Double.NaN;
    }
    
    /**
     * If the range Constraints contradict each other this function is called.
     * It removes the RangeConstraint of one of the two variables, which one is 
     * determined by random.
     */
    void removeOneRange() {
        if (gsc.random.nextDouble() > 0.5) {
            if (outputValue > inputValue) {
                RangeConstraint rc = output.getUpperBoundRangeConstraint();
                gsc.removeRangeConstraint(output, rc);
                gsc.deleteConstraint_internal(rc);
            } else {
                RangeConstraint rc = output.getLowerBoundRangeConstraint();
                gsc.removeRangeConstraint(output, rc);
                gsc.deleteConstraint_internal(rc);
            }
        } else {
            if (outputValue < inputValue) {
                RangeConstraint rc = input.getUpperBoundRangeConstraint();
                gsc.removeRangeConstraint(input, rc);
                gsc.deleteConstraint_internal(rc);
            } else {
                RangeConstraint rc = input.getLowerBoundRangeConstraint();
                gsc.removeRangeConstraint(input, rc);
                gsc.deleteConstraint_internal(rc);
            }
        }
    }

}

