
package com.ai_center.dragonbreath.constraints;

/**
 * Constraint to force the first variable to be less than or equal to 
 * the second variable
 *
 * @author Marcel Lemke
 * @version $Revision: 2.1 $
 */


import java.lang.Math;
import java.util.*;

import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.Registry;


public class LessEqualConstraint extends ConventionalConstraint {

    // **********************
    // ***** ATTRIBUTES *****
    // **********************

    /**
     * The constraint type's name.
     */
    private static final String TYPE_NAME = new String("Less/Equal");

    /**
     * The constraint's base graph.
     */
    private static final Embedding BASE_GRAPH = new Embedding(2);

    /**
     * The variable with higher value.
     */
    private Variable moreVariable = null;
    /**
     * The variable with lower value.
     */
    private Variable lessVariable = null;
    /**
     * The number of variables to be added.
     */
    private int addVariableNumber = 0;

    /**
     * The value of the more variable.
     */
    private double more = 0;
    /**
     * The value of the less variable.
     */
    private double less = 0;


    // *******************
    // ***** METHODS *****
    // *******************

    public void create_graphs() {

	// Creating the base graph

	BASE_GRAPH.addConnectedNode(Registry.VARIABLE,
				    Registry.LESS_EDGE, Registry.INCOMING);
	BASE_GRAPH.addConnectedNode(Registry.VARIABLE,
				    Registry.MORE_EDGE, Registry.INCOMING);

    }

    public int getNumberOfCosts() {
	return 1;
    }

    public String getCostName(int costID) {
	if (costID == 0) {
	    return new String("shift distance inconsistency");
	} else {
	    throw new Error("Invalid cost ID");
	}
    }

    public String getTypeName() {
	return TYPE_NAME;
    }

    public Embedding getBaseGraph() {
	return BASE_GRAPH;
    }

    public int getExtensionGraphNumber() {
	return 0;
    }

    public Embedding getExtensionGraph(int graphID) {
	return null;
    }

    public int getMaximalConnectionNumber() {
	// There are no extensions
	return BASE_GRAPH.connectedNodeNumber;
    }

    public void initializeCostType(int costID) {
	// Nothing to do
    }

    /**
     * Calculates the constraint's costs.
     */
    private void updateCosts() {
	costs[0] = (long) Math.ceil(less>more?less-more:0);
	costFunctionReference[0].updateGlobalCosts();
    }

    public void update(double oldValue, ConnectionReference first,
		       ConnectionReference last) {
	Variable variable = (Variable) first.node;
	if (variable == moreVariable) {
	    more = moreVariable.value();
	} else {
	    less = lessVariable.value();
	}
	updateCosts();

    }

    public void connectNode(ConnectionReference first,
			    ConnectionReference last) {

	Variable variable = (Variable) first.node;
	if (first.edgeID == Registry.MORE_EDGE) {
	    moreVariable = variable;
	    more = moreVariable.value();
	    if (lessVariable != null) updateCosts();
	} else {
	    lessVariable = variable;
	    less = lessVariable.value();
	    if (moreVariable != null) updateCosts();
	}

    }

    public void disconnectNode(ConnectionReference first,
			       ConnectionReference last) {

	if (first.edgeID == Registry.MORE_EDGE) {
	    moreVariable = null;
	    more = 0;
	} else {
	    lessVariable = null;
	    less = 0;
	}
	updateCosts();

    }

    public void improveCosts(int costID) {

	// Checking for valid cost ID

	if (costID != 0) {
	    throw new Error("The cost ID  \"" + costID + 
			    "\" is not supported by this constraint");
	}

	int modify[] = new int[3];
        int modifyNow = 0;

        if (!lessVariable.isConstant() && !moreVariable.isConstant()) {
	    modify[0] = gsc.random.nextInt(2);        // choose randomly whether to change more or less first
	    modify[1] = 1-modify[0];                  // then try to improve cost changing the other variable
	    modify[2] = -1;                           // do break iteration
        } else {
	    if (lessVariable.isConstant()) {      // check cases where at least one variable is constant
		if (moreVariable.isConstant()) return;  // can't improve anything if both are constants
		modify[0] =  1;                         // modify moreVariable only
		modify[1] = -1;
	    } else {
		modify[0] = 0;                          // modify lessVariable only
		modify[1] = -1;
	    }
        }
	
        do {
	    if (modify[modifyNow] == 0) { // check whether to change the less or the more variable
		// if we can't set less to be equal to more, because its lowerbound is smaller than more:
		// set less to its lower bound!
		gsc.setVariable_internal(lessVariable,
					 (lessVariable.lowerBound()>more) ? lessVariable.lowerBound() : more);
	    } else {
		// if we can't set more to be equal to less, because its upperbound is lower than less:
		// set more to its upper bound!
		gsc.setVariable_internal(moreVariable, (moreVariable.upperBound()<less) ? moreVariable.upperBound() : less);
          }

          modifyNow++;
	} while ((costs[0] != 0) && (modify[modifyNow] != -1)); // repeat if costs still not zero and we can still change sth.
    }

}


/*
 * DO _NOT_ EDIT ANYTHING BELOW THIS POINT!
 *
 * $Log: LessEqualConstraint.java,v $
 * Revision 2.1  2002/10/24 03:45:01  nareyek
 * propagation of new values/connection/disconnection does only follow edge direction now
 *
 * Revision 2.0  2002/09/17 00:33:06  nareyek
 * work on beta version began: resource type, state tasks and precondition tasks introduced
 *
 * Revision 1.6  2002/06/01 01:12:28  nareyek
 * initial cost type selection for graphics added
 *
 * Revision 1.5  2002/05/16 20:30:37  nareyek
 * more than one cost type per constraint can now be used
 *
 * Revision 1.4  2002/04/09 01:19:37  nareyek
 * Every cost item can now have a mapping; constraints can have multiple costs
 *
 * Revision 1.3  2002/04/04 09:25:36  nareyek
 * general constraint interface included; cost workflow improved
 *
 * Revision 1.2  2002/04/04 04:55:58  nareyek
 * directories costs and constraintgraph introduced
 *
 * Revision 1.1  2002/04/04 03:04:58  nareyek
 * new constraint directory; VariableComparator removed and integrated into Variable
 *
 * Revision 1.5  2002/03/30 08:33:09  nareyek
 * basic cost-function functionality added
 *
 * Revision 1.4  2001/12/18 15:17:54  sonnebln
 * no message
 *
 * Revision 1.3  2001/12/18 08:18:20  sonnebln
 * no message
 *
 * Revision 1.2  2001/11/22 18:55:02  nareyek
 * Major Registry revision: Dynamic constraint class addition; Registry is now a static object only
 *
 *
 */
