
package com.ai_center.dragonbreath.constraints;

/** 
 * This constraint forces a bag of variables to be a permutation
 * of another bag (WARNING: highly inefficient solution with only
 * one deterministic heuristics - may easily lead to deadlocks).
 *
 * @author Marcelo Carvalho
 * @version $Revision: 2.1 $
 *
 * @see Variable
 */


import java.lang.Math;
import java.util.*;

import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.Registry;


public class PermutationConstraint extends ConventionalConstraint {

    // **********************
    // ***** ATTRIBUTES *****
    // **********************

    /** 
     * The constraint type's name.
     */
    private static final String TYPE_NAME = new String("Permutation");

    /** 
     * The constraint's base graph.
     */
    private static final Embedding BASE_GRAPH = new Embedding(2);
    /** 
     * The constraint's number of extension graphs.
     */
    private static final int EXTENSION_GRAPH_NUMBER = 1;
    /** 
     * The constraint's extension graphs.
     */
    private static final Embedding[] EXTENSION_GRAPH = 
	new Embedding[EXTENSION_GRAPH_NUMBER];

    /** 
     * The first list of variables (CONSTANT).
     */
    private java.util.List constList = 
	Collections.synchronizedList(new LinkedList());
    /** 
     * The second list of variables.
     */
    private java.util.List varList = 
	Collections.synchronizedList(new LinkedList());
    /** 
     * Auxiliar list: contains the absolute difference of the variables of
     * first and second list.
     */
    private java.util.List auxiliarList = 
	Collections.synchronizedList(new LinkedList());

    /** 
     * The number of variables of the lists.
     */
    private int listNumber = 0;


    // *******************
    // ***** METHODS *****
    // *******************

    /**
     * Creates the auxiliar list.
     */
    private void createAuxiliarList() {
	int i;
	// Clears the list
	auxiliarList.clear();
	// Building the list
	for (i = 0; i < listNumber; i++) {
	    Variable aux1 = (Variable) constList.get(i);
	    Variable aux2 = (Variable) varList.get(i);
	    Variable diff = new Variable();
	    diff.set(Math.abs(aux1.value() - aux2.value()));
	    auxiliarList.add(diff);
        }
    }

    public void create_graphs() {

	// Creating the base graph
	
	BASE_GRAPH.addConnectedNode(Registry.VARIABLE, 
				    Registry.CONSTANT_EDGE, Registry.INCOMING);
	BASE_GRAPH.addConnectedNode(Registry.VARIABLE, 
				    Registry.VARIABLE_EDGE, Registry.INCOMING);
	
	// Creating the extension graph
	
	EXTENSION_GRAPH[0] = new Embedding(2);
	EXTENSION_GRAPH[0].addConnectedNode(Registry.VARIABLE, 
					    Registry.CONSTANT_EDGE, 
					    Registry.INCOMING);
	EXTENSION_GRAPH[0].addConnectedNode(Registry.VARIABLE, 
					    Registry.VARIABLE_EDGE, 
					    Registry.INCOMING);	
	
    }

    public int getNumberOfCosts() {
	return 1;
    }

    public String getCostName(int costID) {
	if (costID == 0) {
	    return new String("Permutation Violation");
	} else {
	    throw new Error("Invalid cost ID");
	}
    }

    public String getTypeName() {
	return TYPE_NAME;
    }

    public Embedding getBaseGraph() {
	return BASE_GRAPH;
    }

    public int getExtensionGraphNumber() {
	return EXTENSION_GRAPH_NUMBER;
    }

    public Embedding getExtensionGraph(int graphID) {
	return EXTENSION_GRAPH[graphID];
    }

    public int getMaximalConnectionNumber() {
	// There is no maximum for extensions
	return Integer.MAX_VALUE;
    }

    public void initializeCostType(int costID) {
	// Nothing to do
    }

    /** 
     * Calculates the constraint's costs.
     */
    private void updateCosts() {
	int i;
	double permutationCost = 0;
	// Sum list elements
	for (i = 0; i < listNumber; i++) {
	    Variable elementDiff = (Variable) auxiliarList.get(i);
	    permutationCost = permutationCost + elementDiff.value();
	}
	costs[0] = (long) Math.ceil(permutationCost);
	costFunctionReference[0].updateGlobalCosts();
    }

    public void update(double oldValue, ConnectionReference first, 
		       ConnectionReference last) {
	// Changing in constant list
	if (first.edgeID == Registry.CONSTANT_EDGE) {
	    // Sort list
	    Collections.sort(constList, new Variable());
	}
	// Changing in variable list
	if (first.edgeID == Registry.VARIABLE_EDGE) {
	    // Sort list
	    Collections.sort(varList, new Variable());
	}	
	createAuxiliarList();
	updateCosts();
    }

    public void connectNode(ConnectionReference first, 
			    ConnectionReference last) {
	Variable variable1 = (Variable) first.node;
	// Connecting to constant list
	if (first.edgeID == Registry.CONSTANT_EDGE) {
	    synchronized(constList) {
		constList.add(variable1);
	    }	  
	    // Sort list
	    Collections.sort(constList, new Variable());
	    // Verify if it is the second call
	    if (constList.size() == varList.size()) {
		// New size of the list
		listNumber++;
		// Creates auxiliar list
		createAuxiliarList();
		// Updating
		updateCosts();
	    }
	}
	// Connecting to variable list
	if (first.edgeID == Registry.VARIABLE_EDGE) {
	    synchronized(varList) {
		varList.add(variable1);
	    }
	    // Sort list
	    Collections.sort(varList, new Variable());
	    // Verify if it is the second call
	    if (constList.size() == varList.size()) {
		// New size of the list
		listNumber++;
		// Creates auxiliar list
		createAuxiliarList();
		// Updating
		updateCosts();
	    }
	}	
    }

    public void disconnectNode(ConnectionReference first, 
			       ConnectionReference last) {
	Variable variable1 = (Variable) first.node;
	// Disconnecting from constant list
	if (first.edgeID == Registry.CONSTANT_EDGE) {
	    synchronized(constList) {
		constList.remove(variable1);		
	    }
	    // Verify if is the second call
	    if (constList.size() == varList.size()) {
		// New size
		listNumber--;
		// Creates auxiliar list
		createAuxiliarList();
		// Updating
		updateCosts();
	    }
	}
	// Disconnecting from variable list
	if (first.edgeID == Registry.VARIABLE_EDGE) {
	    synchronized(varList) {
		varList.remove(variable1);		
	    }
	    // Verify if is the second call
	    if (constList.size() == varList.size()) {
		// New size
		listNumber--;
		// Creates auxiliar list
		createAuxiliarList();
		// Updating
		updateCosts();
	    }
	}
    }

    public void improveCosts(int costID) {

	// Checking for valid cost ID

	if (costID != 0) {
	    throw new Error("The cost ID  \"" + costID + 
			    "\" is not supported by this constraint");
	}

	int i;
	double max = 0.0;
	int posmax = 0;
	// Searching for the maximum value of the auxiliar list
	synchronized(auxiliarList) {
	    for (i = 0; i < listNumber; i++) {
		Variable elementDiff = (Variable) auxiliarList.get(i);
		if (elementDiff.value() > max) {
		    max = elementDiff.value();
		    posmax = i;
		}
	    }
	}
	
	// If maximum is greater than zero, then change variable List
	if (max > 0) {
	    // Changing the Ith value of variable list to be equal to
	    // the Ith value of the constant list
	    Variable changeVar = (Variable) varList.get(posmax);
	    Variable baseVar = (Variable) constList.get(posmax);
	    gsc.setVariable_internal(changeVar, baseVar.value());
	    // Sort the variable list
	    Collections.sort(varList, new Variable());
	    // Build the auxiliar list
	    createAuxiliarList();				
	}
    }

}
