package com.ai_center.dragonbreath.constraints;
 
/**
 * This constraint forces the product over a set of variables
 * to equal to another variable.
 *
 * @author Yiqi Pan
 * @version $Revision: 1.6 $
 * @see Variable
 */


import java.lang.Math;
import java.util.*;

import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.Registry;


public class ProductConstraint extends ConventionalConstraint 
{

	// **********************
	// ***** ATTRIBUTES *****
	// **********************
	
	/**
	* The constraint type's name.
	*/
	private static final String TYPE_NAME = new String( "Product" );
	
	/**
	* The constraint's base graph.
	*/
	private static final Embedding BASE_GRAPH = new Embedding(2);
	
	/**
	* The constraint's number of extension graphs.
	*/
	private static final int EXTENSION_GRAPH_NUMBER = 1;
	
	/**
	* The constraint's extension graphs.
	*/
	private static final Embedding[] EXTENSION_GRAPH =
		new Embedding[EXTENSION_GRAPH_NUMBER];
	
	/**
	* The product variable.
	*/
	private Variable productVariable = null;
	
	/**
	* The list of variables to be multiplied.
	*/
	private java.util.List multiplyVariables =
		Collections.synchronizedList(new LinkedList());
	
	/**
	* The number of variables to be multiplied.
	*/
	private int multiplyVariableNumber = 0;
	
	/**
	* The value of the product variable.
	*/
	private double product = 0;
	
	/**
	* The product of the multiplied variables.
	*/
	private double productVariables = 0;
	
	
	// *******************
	// ***** METHODS *****
	// *******************
	
	public void create_graphs() 
	{
	
	// Creating the base graph
	BASE_GRAPH.addConnectedNode( Registry.VARIABLE, Registry.PRODUCT_EDGE, Registry.INCOMING );
	BASE_GRAPH.addConnectedNode( Registry.VARIABLE, Registry.MULTIPLIER_EDGE, Registry.INCOMING );
	
	// Creating the extension graph
	
	EXTENSION_GRAPH[0] = new Embedding(1);
	EXTENSION_GRAPH[0].addConnectedNode( Registry.VARIABLE, Registry.MULTIPLIER_EDGE, Registry.INCOMING);
	
	}

	public int getNumberOfCosts() 
	{
		return 1;
	}
	
	public String getCostName( int costID ) 
	{
		if ( costID == 0 ) 
		{
		    return new String( "multiplicative difference inconsistency" );
		} 
		else 
		{
		    throw new Error( "Invalid cost ID" );
		}
	}
	
	public String getTypeName() 
	{
		return TYPE_NAME;
	}
	
	public Embedding getBaseGraph() 
	{
		return BASE_GRAPH;
	}
	
	public int getExtensionGraphNumber() 
	{
		return EXTENSION_GRAPH_NUMBER;
	}
	
	public Embedding getExtensionGraph( int graphID ) 
	{
		return EXTENSION_GRAPH [ graphID ];
	}
	
	public int getMaximalConnectionNumber() 
	{
		// There is no maximum for extensions
		return Integer.MAX_VALUE;
	}
	
	public void initializeCostType( int costID ) 
	{
		// Nothing to do
	}
	
	
	/**
	* 	Calculates the constraint's costs.
	*/
	private void updateCosts() 
	{
		costs[0] = (long) Math.ceil( Math.abs( product - productVariables ) );
		costFunctionReference[0].updateGlobalCosts();
	}


	/**
	* 	Recalculates the product of all the variables in the constraint.
	*/
	private void recalculate() 
	{
		Variable var;
		Iterator itr = multiplyVariables.iterator();
		    
		if ( ! itr.hasNext() )
		{
			// There is no variable, just set productVariables = 0
		    	productVariables = 0;
		}
		else
		{
			productVariables = 1;
	
			while ( itr.hasNext() ) 
			{
				var = (Variable)itr.next();
				
				// when improveCosts() is working, the productVariable is in the list. 
				//it should to be ignored			
				if ( productVariable != var ) 
				{
					productVariables *= var.value();
				}
			}
		}
	}

	/**
	*	Update the product of varibles
	*/
	public void update( double oldValue, ConnectionReference first, ConnectionReference last ) 
        {
	        Variable variable = (Variable) first.node;
	        
	        if (variable == productVariable) 
	        {
	        	product = productVariable.value();
	        }
	        else 
	        {
	        	if ( oldValue == 0 ) 
	        	{  
	        		// change from zero to non-zero, so recalculate
	                	recalculate();
	            	}
	            	else 
	            	{
	                	if ( variable.value() == 0 ) 
	                	{  
	                		// new value is 0, so just set productVariables to 0
	                    		productVariables = 0;
	                	}
	                	else 
	                	{
	                    		productVariables = productVariables / oldValue * variable.value();
	                	}
			}
	        }
	        updateCosts();
	}

	/**
	*	Connect a new node to the constraint
	*/
	public void connectNode( ConnectionReference first, ConnectionReference last )
	{
		Variable variable = (Variable) first.node;
		double fVal = variable.value();
		
		if (first.edgeID == Registry.PRODUCT_EDGE) 
		{
			productVariable = variable;
			product = fVal;
			if ( multiplyVariableNumber >= 1 ) 
			{
				updateCosts();
			}
		}
		else 
		{
			multiplyVariables.add( variable );
			multiplyVariableNumber++;
			
			if ( multiplyVariableNumber == 1 ) 
			{  	
				// this is the first multiply variable
				productVariables = fVal;
			}
			else 
			{
				productVariables = productVariables * fVal;
			}
			if ( productVariable != null ) 
			{
				updateCosts();
			}
		}
	}

	/**
	*	Disconnect a node from the contraint
	*/
	public void disconnectNode( ConnectionReference first, ConnectionReference last ) 
	{                       	
		if (first.edgeID == Registry.PRODUCT_EDGE) 
		{
			productVariable = null;
			product = 0;
		}
		else {
			Variable variable = (Variable) first.node;
			double fVal = variable.value();
			multiplyVariables.remove( variable );
			multiplyVariableNumber--;			
			if ( fVal != 0 ) 
			{
				productVariables = productVariables / fVal;
			}
			else 
			{  
				// deal with zero case
				recalculate();
			}
		}
		updateCosts();
	}

	/**
	*  	Optimize the varibles
	*/
	public void improveCosts( int costID ) 
	{	
		// Checking for valid cost ID
		
		if ( costID != 0 ) 
		{
		    throw new Error("The cost ID  \"" + costID + 
		                    "\" is not supported by this constraint");
		}
		
		int i, j;
		Variable var;
		double v;
		boolean bFinished = false;
		
		// Choose variable randomly from the list of multiplyVariables. 
		// The chosen variable is updated and moved to
		// the end of the list. Every time, the maximal random value decreased by 1.
		
		multiplyVariables.add( productVariable );
		
		for ( j = multiplyVariableNumber; ( j >= 0 ) && !bFinished; j-- ) 
		{
			i = gsc.random.nextInt( j + 1 );
			var = (Variable)multiplyVariables.get( i );
			
			if ( !var.isConstant() ) 
			{
				if ( var == productVariable ) 
				{
					// The product variable will be changed
					if ( product > productVariables	) 
					{	
						// product variable needs to be decreased	
						if ( productVariable.lowerBound() < product ) 
						{
							if ( productVariables < productVariable.lowerBound() ) 
							{
								v = productVariable.lowerBound();
							}
							else 
							{
								v = productVariables;
								bFinished = true;
							}
							gsc.setVariable_internal( productVariable, v );
				                }
		            		}
		            		else 
		            		{						
						// product variable needs to be increased						
						if ( productVariable.upperBound() > product ) 
						{
							if ( productVariables > productVariable.upperBound() ) 
							{
								v = productVariable.upperBound();
							}
							else 
							{
								v = productVariables;
								bFinished = true;
							}
							gsc.setVariable_internal( productVariable, v );
						}
						
					}
				}
				else 
				{
					// One of the multiplied variables will be changed
					v = var.value();
					
					if ( productVariables == 0 ) 
					{		
				        	if ( v == 0 ) 
				        	{						
							// set the current variable to a non-zero value
							if ( ( var.upperBound() >= 1) && ( var.lowerBound () <= 1 ) )
							{
								gsc.setVariable_internal(var, 1);
							}
							else 
							{
								// find random non-zero value for the current variable
								v = 0;
								while ( v == 0 ) 
								{
							    		v = gsc.random.nextDouble() * 
							    			( var.upperBound() - var.lowerBound() ) + 
							    			var.lowerBound();
							    		gsc.setVariable_internal( var, v );
								}
							}
							
						}
				
				    	}
					else 
					{					
						/***************************************************/
						/* Deciding whether the variable should be         */
						/* increased or decreased:                         */
						/*                                                 */
						/*  ProductVariables >  product     v    var       */
						/*       >0               >0       >0    dec       */
						/*       >0               >0       <0    inc       */
						/*       <0               <0       >0    inc       */
						/*       <0               <0       <0    dec       */
						/*       >0               <0       >0    dec       */
						/*       >0               <0       <0    inc       */
						/*                                                 */
						/*  ProductVariables <  product     v    var       */
						/*       >0               >0       >0    inc       */
						/*       >0               >0       <0    dec       */
						/*       <0               <0       >0    dec       */
						/*       <0               <0       <0    inc       */
						/*       <0               >0       >0    dec       */
						/*       <0               >0       <0    inc       */
						/***************************************************/
						
						boolean bInc = true;
						
						if ( productVariables > product ) 
						{
							if ( productVariables > 0 ) 
							{
								if ( v > 0 ) 
								{
									bInc = false;
								}
							}
							else 
							{
								if ( v < 0 ) 
								{
									bInc = false;
								}
							}
						}
						else 
						{
							if ( productVariables < 0 ) 
							{
								if ( v > 0 ) 
								{
									bInc = false;
								}
							}
							else 
							{
								if ( v < 0 ) 
								{
									bInc = false;
								}
							}
						}
						
						// Do the increase/decrease
						
						v = ( v * product ) / productVariables;
						
						if ( bInc ) 
						{  
							// variable needs to be increased
							if ( var.upperBound() >= v ) 
							{
								bFinished = true;
							}
							else 
							{
								v = var.upperBound();
							}
						}
						else 
						{  
							// variable needs to be decreased
							if ( v >= var.lowerBound() ) 
							{
								bFinished = true;
							}
							else 
							{
								v = var.lowerBound();
							}
						}	
						gsc.setVariable_internal( var, v );
					}
				}
			}
	
			// move the chosen variable to the end of the list
			multiplyVariables.remove( i );
			multiplyVariables.add( var );
		}	
		multiplyVariables.remove( productVariable );
	}
}
