
package com.ai_center.dragonbreath.constraints;

import java.lang.Math;
import java.util.*;

import org.jdom.Element;
import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.Registry;

/** 
 * This constraint forces the that the two connected Variables are a multiple of each other.
 * The coefficient can be determined at creation.
 *
 * @author tim baermann
 * @version $Revision: 2.4 $
 * @see Variable
 */

public class ProportionConstraint extends ConventionalConstraint {
    
 
    // **********************
    // ***** ATTRIBUTES *****
    // **********************

    /** 
     * The constraint type's name.
     */
    private static final String TYPE_NAME = new String("Proportion");

    /** 
     * The constraint's base graph.
     */
    private static final Embedding BASE_GRAPH = new Embedding(2);
    
    /** 
     * The constraint's number of extension graphs.
     */
    private static final int EXTENSION_GRAPH_NUMBER = 0;
    
    /** 
     * The product of the input and the coefficient
     */
    private Variable output = null;
    
    /** 
     * The  output / coeff = input.
     */
    private Variable input = null;
    
    /** 
     * The Coefficient.
     */
    private double coeff = 1;
    
    /**
     * The current variable value of input.
     */
    private double inputValue = 0;
    
    /**
     * The current variable value of output
     */
    private double outputValue = 0;
    
    // *******************
    // ***** METHODS *****
    // *******************
    /**
     * Sets the Coefficient.
     * @param nodeData  The nodeData as stored in file. 
     */
    public void setAuxiliaryNodeData(Element nodeData) {    
        if (nodeData != null) {
            String cS = nodeData.getAttributeValue( "Coefficient");
            if (cS != null) {
                double coefficient = java.lang.Double.parseDouble(cS);
                setCoefficient(coefficient);
            }
        }
    }
    
     /**
     * Before storing a node this function is called, and the node can add its own data to the XML-data.
     * Adds the Coefficient.
     * @param nodeData  Where the data should be stored.
     */
    public void getAuxiliaryNodeData(Element nodeData) {
        nodeData.setAttribute("Coefficient", java.lang.String.valueOf( coeff) );
    }
    
    /**
     * Sets how many times output should be input
     */
    public void setCoefficient(double newCoefficient) {
        coeff = newCoefficient;
        if ((graphObject != null) && (graphObject instanceof ProportionConstraintGraphics)) {
            ((ProportionConstraintGraphics)graphObject).updateCoefficient();
        }
        updateCosts();
    }
    
    /** 
     * Adding specialized graphics for the constraint.
     */
    public void createGraphics() {
	graphObject = (ConventionalConstraintGraphics) 
	    new ProportionConstraintGraphics(this, gsc);
    }
    
    /**
     * Returns the coefficient
     */
    public double getCoefficient() {
        return coeff;
    }
    public void create_graphs() {

	// Creating the base graph
	
	BASE_GRAPH.addConnectedNode(Registry.VARIABLE, 
				    Registry.PROPORTION_OUTPUT_EDGE, Registry.INCOMING);
	BASE_GRAPH.addConnectedNode(Registry.VARIABLE, 
				    Registry.PROPORTION_INPUT_EDGE, Registry.INCOMING);
    }

    public int getNumberOfCosts() {
	return 1;
    }

    public String getCostName(int costID) {
	if (costID == 0) {
	    return new String("proportion inconsistency");
	} else {
	    throw new Error("Invalid cost ID");
	}
    }

    public String getTypeName() {
	return TYPE_NAME; 
    }

    public Embedding getBaseGraph() {
	return BASE_GRAPH;
    }

    public int getExtensionGraphNumber() {
	return EXTENSION_GRAPH_NUMBER;
    }

    public Embedding getExtensionGraph(int graphID) {
	return null;
    }

    public int getMaximalConnectionNumber() {
	return 2;
    }

    public void initializeCostType(int costID) {
	// Nothing to do
    }

    /** 
     * Calculates the constraint's costs.
     */
    private void updateCosts() {
        if (input == null || output == null) {
            costs[0] = 0;
        } else if (input.isInteger() && output.isInteger()) {
            costs[0] = (long) Math.round(Math.abs((inputValue * coeff) - outputValue));
        } else {
            costs[0] = (long) Math.ceil(Math.abs((inputValue * coeff) - outputValue));
        }
	costFunctionReference[0].updateGlobalCosts();
    }

    public void update(double oldValue, ConnectionReference first, 
			  ConnectionReference last) {
	Variable variable = (Variable) first.node;
	if (variable == input) {
	    inputValue = variable.value();
	} else {
	    outputValue = variable.value();
	}
	updateCosts();
    }

    public void connectNode(ConnectionReference first, 
			       ConnectionReference last) {
	Variable variable = (Variable) first.node;
	if (first.edgeID == Registry.PROPORTION_OUTPUT_EDGE) {
	    output = variable;
	    outputValue = variable.value();
	    if (input != null) updateCosts();
	} else {
	    input = variable;
	    inputValue = variable.value();
	    if (output != null) updateCosts();
	}
    }

    public void disconnectNode(ConnectionReference first, 
				  ConnectionReference last) {

	if (first.edgeID == Registry.PROPORTION_OUTPUT_EDGE) {
	    output = null;
	    outputValue = 0;
	} else {
	    Variable variable = (Variable) first.node;
	    input = null;
            inputValue = 0;
	}
	updateCosts();
    }

    public void improveCosts(int costID) {

	// Checking for valid cost ID

	if (costID != 0) {
	    throw new Error("The cost ID  \"" + costID + 
			    "\" is not supported by this constraint");
	}

        double newOutputValue = outputValue; 
        double newInputValue = inputValue; 
        if (coeff == 0) {
            gsc.setVariable_internal(output, 0);
        } else if (gsc.random.nextDouble() > 0.5) {
            //change output
            if (! output.isInteger()) {
                newOutputValue = inputValue * coeff;
                newInputValue = inputValue;
            } else {
                newOutputValue = Math.ceil(inputValue * coeff);
                newInputValue = Math.ceil(newOutputValue / coeff);
            }
        } else {
            //change input
            if (gsc.random.nextDouble() < 0.01) {// in one percent of all cases simply choose the solution 0
                newInputValue = 0;
            } else if (! input.isInteger()) {
                newInputValue = outputValue / coeff;
                newOutputValue = outputValue;
            } else {
                newInputValue = Math.ceil(outputValue / coeff);
                newOutputValue = Math.ceil(newInputValue * coeff);
            }
                
        }
        //now handle existing bounds
        boolean solved = false;
        while (! solved) {
            solved = true;
            if (output.upperBound() < newOutputValue) {
                newOutputValue = output.upperBound();
                if (! input.isInteger()) {
                    newInputValue  = newOutputValue / coeff;
                } else {
                    newInputValue = Math.ceil(outputValue / coeff);
                }

                if (input.lowerBound() > newInputValue) {// upperBound is right anyway
                     removeOneRange();
                     solved = false;
                }
            }
            if (solved && (output.lowerBound() > newOutputValue)) {//first solve upperbound
                newOutputValue = output.lowerBound();
                if (! input.isInteger()) {
                   newInputValue  = newOutputValue / coeff;
                } else {
                    newInputValue = Math.ceil(outputValue / coeff);
                }

                if (input.upperBound() < newInputValue) {// lowerBound is right anyway
                     removeOneRange();
                     solved = false;
                }
            }
        }
        if (input.isInteger()) {
            gsc.setVariable_internal(input, (int) newInputValue);
        } else {
            gsc.setVariable_internal(input, newInputValue);
        }
        if (output.isInteger()) {
            gsc.setVariable_internal(output, (int) newOutputValue);
        } else {
            gsc.setVariable_internal(output, newOutputValue);
        }
    }
    
    /**
     * If the range Constraints contradict each other this function is called.
     * It removes the RangeConstraint of one of the two variables, which one is 
     * determined by random.
     */
    void removeOneRange() {
        if (gsc.random.nextDouble() > 0.5) {
            if (outputValue > inputValue) {
                RangeConstraint rc = output.getUpperBoundRangeConstraint();
                gsc.removeRangeConstraint(output, rc);
                gsc.deleteConstraint_internal(rc);
            } else {
                RangeConstraint rc = output.getLowerBoundRangeConstraint();
                gsc.removeRangeConstraint(output, rc);
                gsc.deleteConstraint_internal(rc);
            }
        } else {
            if (outputValue < inputValue) {
                RangeConstraint rc = input.getUpperBoundRangeConstraint();
                gsc.removeRangeConstraint(input, rc);
                gsc.deleteConstraint_internal(rc);
            } else {
                RangeConstraint rc = input.getLowerBoundRangeConstraint();
                gsc.removeRangeConstraint(input, rc);
                gsc.deleteConstraint_internal(rc);
            }
        }
    }

}

