
package com.ai_center.dragonbreath.constraints;

/** 
 * This unary constraint forces the affected variable to
 * take a value within a certain range. The constraint is 
 * hard, i.e., it directly affects the variable and is 
 * always satisfied.
 *
 * @author Alexander Nareyek
 * @version $Revision: 2.2 $
 * @see Variable
 */


import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.Registry;

import org.jdom.Element;

public class RangeConstraint extends ConventionalConstraint {

    // **********************
    // ***** ATTRIBUTES *****
    // **********************

    /** 
     * The constraint type's name.
     */
    private static final String TYPE_NAME = new String("Range");

    /** 
     * The constraint's base graph.
     */
    private static final Embedding BASE_GRAPH = new Embedding(1);

    /** 
     * The variable to be affected.
     */
    public Variable variable = null;

    // Bound variables

    /** 
     * The variable's lower bound.
     */
    public double lowerBound = Double.NEGATIVE_INFINITY;

    /** 
     * The variable's upper bound.
     */
    public double upperBound = Double.POSITIVE_INFINITY;


    // ***********************
    // ***** DESTRUCTORS *****
    // ***********************

    /** 
     * Removes the constraint.
     */
    public void finalize() {
	if (!isType && variable != null) {
	    gsc.removeRangeConstraint(variable, this);
	}
    }


    // *******************
    // ***** METHODS *****
    // *******************
     /**
     * Sets the boundaries.
     * @param nodeData  The nodeData as stored in file. 
     */
    public void setAuxiliaryNodeData(Element nodeData) {
        if (nodeData != null) {
            String sl = nodeData.getAttributeValue( "RangeL");
            if (sl != null) {
                double rangeL = java.lang.Double.parseDouble(sl);
                gsc.setLowerBound(this, rangeL);
            } 
            String su = nodeData.getAttributeValue( "RangeU");
            if (su != null) {
                double rangeU = java.lang.Double.parseDouble(su);
                gsc.setUpperBound(this, rangeU);
            }
        }
    }
    
     /**
     * Before storing a node this function is called, and the node can add its own data to the XML-data.
     * Adds the name and the value of the Variable.
     * @param nodeData  Where the 
     */
    public void getAuxiliaryNodeData(Element nodeData) {
        nodeData.setAttribute("RangeL",java.lang.String.valueOf(lowerBound));
        nodeData.setAttribute("RangeU", java.lang.String.valueOf(upperBound));
    }
    
    public void create_graphs() {

	// Creating the base graph
	
	BASE_GRAPH.addConnectedNode(Registry.VARIABLE, 
				    Registry.PLAIN_EDGE, Registry.INCOMING);
    }

    public int getNumberOfCosts() {
	return 0;
    }

    public String getCostName(int costID) {
	throw new Error("Invalid function for this constraint type");
    }

    public String getTypeName() {
	return TYPE_NAME;
    }

    public Embedding getBaseGraph() {
	return BASE_GRAPH;
    }

    public int getExtensionGraphNumber() {
	return 0;
    }

    public Embedding getExtensionGraph(int graphID) {
	return null;
    }

    public int getMaximalConnectionNumber() {
	// Making the constraint non-extensible
	return BASE_GRAPH.connectedNodeNumber;
    }

    public void initializeCostType(int costID) {
	throw new Error("Invalid function for this constraint type");
    }

    public void update(double oldValue, ConnectionReference first, 
		       ConnectionReference last) {
    }

    public void connectNode(ConnectionReference first, 
			    ConnectionReference last) {
	Variable variable = (Variable) first.node;
	this.variable = variable;
	gsc.addRangeConstraint(variable, this);
    }

    public void disconnectNode(ConnectionReference first, 
			       ConnectionReference last) {
	this.variable = null;
	gsc.removeRangeConstraint((Variable) first.node, this);
    }

    public void improveCosts(int costID) {
	throw new Error("Invalid function for this constraint type");
    }

    /** 
     * Adding specialized graphics for the constraint.
     */
    public void createGraphics() {
	graphObject = (ConventionalConstraintGraphics) 
	    new RangeConstraintGraphics(this, gsc);
    }

    /** 
     * Sets the bounds of the range constraint.
     * @param lbound  The new lower bound.
     * @param ubound  The new upper bound.
     */
    public void setBounds(double lbound, double ubound) {
	lowerBound = lbound;
	upperBound = ubound;
	if (lowerBound > upperBound) {
	    throw new Error("The lower bound is larger than the upper bound");
	}
    }

    /** 
     * Sets the lower bound of the range constraint.
     * @param bound  The new lower bound.
     */
    public void setLowerBound(double bound) {
	lowerBound = bound;
	if (lowerBound > upperBound) {
	    throw new Error("The given lower bound is larger than the upper bound");
	}
    }

    /** 
     * Sets the upper bound of the range constraint.
     * @param bound  The new upper bound.
     */
    public void setUpperBound(double bound) {
	upperBound = bound;
	if (lowerBound > upperBound) {
	    throw new Error("The lower bound is larger than the given upper bound");
	}
    }

    /** 
     * Deactivates the lower bound of the range constraint.
     */
    public void deactivateLowerBound() {
	lowerBound = Double.NEGATIVE_INFINITY;
    }

    /** 
     * Deactivates the upper bound of the range constraint.
     */
    public void deactivateUpperBound() {
	upperBound = Double.POSITIVE_INFINITY;
    }

}


/* 
 * DO _NOT_ EDIT ANYTHING BELOW THIS POINT! 
 * 
 * $Log: RangeConstraint.java,v $
 * Revision 2.2  2003/02/24 21:41:58  teutates
 * no message
 *
 * Revision 2.1  2003/02/21 16:55:23  teutates
 * improved auxiliary data load for nodes
 *
 * Revision 2.0  2002/09/17 00:33:06  nareyek
 * work on beta version began: resource type, state tasks and precondition tasks introduced
 *
 * Revision 1.7  2002/08/23 23:44:26  nareyek
 * object-, integer- and range-constraints get not included in cost functions anymore
 *
 * Revision 1.6  2002/06/01 01:12:28  nareyek
 * initial cost type selection for graphics added
 *
 * Revision 1.5  2002/05/16 20:30:37  nareyek
 * more than one cost type per constraint can now be used
 *
 * Revision 1.4  2002/04/09 01:19:37  nareyek
 * Every cost item can now have a mapping; constraints can have multiple costs
 *
 * Revision 1.3  2002/04/04 09:25:36  nareyek
 * general constraint interface included; cost workflow improved
 *
 * Revision 1.2  2002/04/04 04:55:58  nareyek
 * directories costs and constraintgraph introduced
 *
 * Revision 1.1  2002/04/04 03:04:58  nareyek
 * new constraint directory; VariableComparator removed and integrated into Variable
 *
 * Revision 1.3  2002/03/30 08:33:09  nareyek
 * basic cost-function functionality added
 *
 * Revision 1.2  2001/11/22 18:55:02  nareyek
 * Major Registry revision: Dynamic constraint class addition; Registry is now a static object only
 * 
 * 
 */ 
