
package com.ai_center.dragonbreath.constraints;


import java.util.*;

import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.constraints.stateresourceconstraint.*;
import com.ai_center.dragonbreath.Registry;


/**
 * This constraint projects the state of a property over time. State
 * tasks influence the state development, and precondition tasks can be
 * used to check the state for specific conditions.
 * The cost component TYPE_INCONSISTENCY returns costs if the resource
 * type variable does not have a valid value.
 * The cost component PRECONDITION_FAILURE returns costs is connected
 * precondition tasks' tests are not satisfied by the state projection.
 * The amount of costs is computed by a heuristic measure that checks how
 * many state changes would have to be added to satisfy a precondition.
 * The cost component STATIC_GOAL_FAILURE returns costs if additionally
 * given tests - similar to preconditions - are not satisfied.
 * It must be ensured that there are no TYPE_INCONSISTENCY costs when an
 * improvement of PRECONDITION_FAILURE or STATIC_GOAL_FAILURE is to be
 * called.
 *
 * @author Alexander Nareyek
 * @version $Revision: 2.19 $
 *
 * @see Variable
 * @see CurrentTimeObjectConstraint
 * @see PreconditionTaskObjectConstraint
 * @see StateTaskObjectConstraint
 */
public class StateResourceConstraint extends ConventionalConstraint
{


    // **********************
    // ***** ATTRIBUTES *****
    // **********************


    /**
     * Constant representing the sub-class of a symbolic state resource
     * constraint.
     */
    public static final int SYMBOLIC_SRC = 0;

    /**
     * Constant representing the sub-class of a numerical heap state
     * resource constraint.
     */
    public static final int NUMERICALHEAP_SRC = 1;


    /***** ENGINE-INTEGRATION-RELATED VARIABLES *****/

    /**
     * The constraint type's name.
     */
    private static final String TYPE_NAME = new String( "State Resource" );

    /**
     * The identifier for the constraint's type-related cost function.
     */
    public static final int TYPE_INCONSISTENCY = 0;
    /**
     * The identifier for the constraint's precondition-related cost component.
     */
    public static final int PRECONDITION_FAILURE = 1;
    /**
     * The identifier for the constraint's static goal-related cost component.
     */
    public static final int STATIC_GOAL_FAILURE = 2;
    /**
     * Descriptions of the cost components.
     */
    public static String[] costDescription = new String[ 3 ];
    static
    {
        costDescription[ TYPE_INCONSISTENCY ]
            = new String( "type inconsistency" );
        costDescription[ PRECONDITION_FAILURE ]
            = new String( "precondition failure" );
        costDescription[ STATIC_GOAL_FAILURE ]
            = new String( "static goal failure" );
    }

    /**
     * The constraint's base graph.
     */
    private static final Embedding BASE_GRAPH = new Embedding( 3 );
    /**
     * The constraint's number of extension graphs.
     */
    private static final int EXTENSION_GRAPH_NUMBER = 2;
    /**
     * The constraint's extension graphs.
     */
    private static final Embedding[] EXTENSION_GRAPH
        = new Embedding[ EXTENSION_GRAPH_NUMBER ];


    /***** TYPE-RELATED VARIABLES *****/


    /**
     * The number of supported sub-classes.
     */
    private final static int SUBCLASS_NUMBER = 2;

    /**
     * Array of available sub-classes.
     */
    private static Class[] subClassHandle = new Class[ SUBCLASS_NUMBER ];
    static
    {
        registerSubClass( SYMBOLIC_SRC , "com.ai_center.dragonbreath.constraints.stateresourceconstraint.SymbolicSRC" );
        registerSubClass( NUMERICALHEAP_SRC , "com.ai_center.dragonbreath.constraints.stateresourceconstraint.NumericalHeapSRC" );
    }

    /**
     * The maximal number of resource types that can be registered.
     */
    private final static int MAX_RESOURCE_TYPE_NUMBER = 1024;

    /**
     * Array of available resource types' sub-class IDs.
     */
    private static int[] resourceTypeSubClassID
        = new int[ MAX_RESOURCE_TYPE_NUMBER ];
    /**
     * Array of definition objects of the available resource types.
     */
    private static SRCResourceType[] resourceTypeDefinition
        = new SRCResourceType[ MAX_RESOURCE_TYPE_NUMBER ];

    /**
     * Number of registered resource types.
     */
    private static int currentResourceTypeNumber = 0;


    /***** GENERAL INTERNAL STRUCTURES *****/

    /**
     * Indicates whether the constraint is currently changing CSP components
     * and the update procedure does not need to be called.
     */
    public boolean modificationModus = false;

    /**
     * Indicates whether the constraint is currently deleting itself and
     * and the disconnection procedure does not need to be called.
     */
    private boolean alive = true;

    /**
     * The state variable.
     */
    public Variable stateVariable = null;
    /**
     * The internal copy of the state variable's value.
     */
    public double state = 0;

    /**
     * The connected current time object constraint.
     */
    public ConventionalConstraint currentTimeConstraint = null;
    /**
     * The internal copy of the current time variable's value.
     */
    public long currentTime = 0;

    /**
     * The specialized sub-class constraint.
     */
    public SRCInterface subClass = null;

    /**
     * The resource type variable.
     */
    public Variable resourceTypeVariable = null;
    /**
     * The internal copy of the resource type variable's value.
     */
    private long resourceType = 0;

    /**
     * Hashtable from StateTaskObjectConstraints to specialized
     * SRCStateTasks.
     */
    private Hashtable stConstraintsHT = null;
    /**
     * A collection view of the specialized state task instances
     * of stConstraintsHT.
     */
    private Collection stateTaskSet = null;
    /**
     * The number of connected state tasks.
     */
    public long stateTaskNumber = 0;

    /**
     * Hashtable from PreconditionTaskObjectConstraints to specialized
     * SRCPreconditionTasks.
     */
    private Hashtable ptConstraintsHT = null;
    /**
     * A collection view of the specialized precondition task instances
     * of ptConstraintsHT.
     */
    private Collection preconditionTaskSet = null;
    /**
     * The number of connected precondition tasks.
     */
    private long preconditionTaskNumber = 0;


    // *******************
    // ***** METHODS *****
    // *******************


    public void create_graphs()
    {

        // Creating the base graph

        BASE_GRAPH.addConnectedNode( Registry.CURRENTTIME_OBJECTCONSTRAINT ,
            Registry.PLAIN_EDGE , Registry.INCOMING );
        BASE_GRAPH.addConnectedNode( Registry.VARIABLE ,
            Registry.RESOURCETYPE_EDGE , Registry.INCOMING );
        BASE_GRAPH.addConnectedNode( Registry.VARIABLE ,
            Registry.STATE_EDGE , Registry.INCOMING );

        // Creating the extension graph

        EXTENSION_GRAPH[ 0 ] = new Embedding( 1 );
        EXTENSION_GRAPH[ 0 ].addConnectedNode( Registry.STATETASK_OBJECTCONSTRAINT, Registry.PLAIN_EDGE , Registry.INCOMING );

        EXTENSION_GRAPH[ 1 ] = new Embedding( 1 );
        EXTENSION_GRAPH[ 1 ].addConnectedNode( Registry.PRECONDITIONTASK_OBJECTCONSTRAINT , Registry.PLAIN_EDGE , Registry.INCOMING );

    }


    public int getNumberOfCosts()
    {
        return 3;
    }


    public String getCostName( int costID )
    {
        if ( ( costID >= 0 ) && ( costID < 4 ) )
        {
            return costDescription[ costID ];
        }
        else
        {
            throw new Error( "Invalid cost ID" );
        }
    }


    public String getTypeName()
    {
        return TYPE_NAME;
    }


    public Embedding getBaseGraph()
    {
        return BASE_GRAPH;
    }


    public int getExtensionGraphNumber()
    {
        return EXTENSION_GRAPH_NUMBER;
    }


    public Embedding getExtensionGraph( int graphID )
    {
        return EXTENSION_GRAPH[ graphID ];
    }


    public int getMaximalConnectionNumber()
    {
        // There is no maximum for extensions
        return Integer.MAX_VALUE;
    }


    public void initializeInternalStructures()
    {
        ptConstraintsHT = new Hashtable();
        preconditionTaskSet = ptConstraintsHT.values();
        stConstraintsHT = new Hashtable();
        stateTaskSet = stConstraintsHT.values();
    }


    public void initializeCostType( int costID )
    {
        costFunctionReference[ costID ].updateGlobalCosts();
    }


    /***** RESOURCE-TYPE HANDLING *****/

    /**
     * Registers a possible resource type. This function is to be called
     * from the sub-class constraint. Users should call the respective
     * sub-class function instead.
     * @param subclassID  The ID of the new resource type's constraint
     *                    sub-class.
     * @param definitionObject  A reference to the sub-class' internal
     *                          definition object.
     */
    public static int registerNewResourceType( int subclassID ,
        SRCResourceType definitionObject )
    {
        if ( ( subclassID < 0 ) || ( subclassID >= SUBCLASS_NUMBER ) )
        {
            throw new Error( "A sub-class with ID \"" + subclassID
                + "\" is not supported" );
        }
        if ( currentResourceTypeNumber == MAX_RESOURCE_TYPE_NUMBER )
        {
            throw new Error( "No more resource type IDs available" );
        }
        resourceTypeSubClassID[ currentResourceTypeNumber ] = subclassID;
        resourceTypeDefinition[ currentResourceTypeNumber ] = definitionObject;
        return currentResourceTypeNumber++ ;
    }


    /**
     * Verifies a resource type ID and returns the corresponding resource
     * type definition.
     * @param resourceTypeID  The resource type ID to be verified.
     * @param subClassID  The ID of the sub-class is tested to correspond.
     */
    public static SRCResourceType getResourceTypeDefinition( int resourceTypeID , int subClassID )
    {
        if ( ( resourceTypeID >=0 )
            && ( resourceTypeID < currentResourceTypeNumber ) )
        {
            if ( resourceTypeSubClassID[ resourceTypeID ] == subClassID )
            {
                return resourceTypeDefinition[ resourceTypeID ];
            }
            else
            {
                throw new Error( "Wrong sub-class for resource type ID" );
            }

        }
        else
        {
            throw new Error( "Resource type ID \"" + resourceTypeID
                + "\" has not been registered" );
        }
    }


    /**
     * Returns the corresponding resource type definition.
     * @param resourceTypeID  The ID of the resource type to be retrieved.
     */
    public static SRCResourceType getResourceTypeDefinition( int resourceTypeID )
    {
        if ( ( resourceTypeID >=0 )
            && ( resourceTypeID < currentResourceTypeNumber ) )
        {
            return resourceTypeDefinition[ resourceTypeID ];
        }
        else
        {
           throw new Error( "Resource type ID \"" + resourceTypeID + "\" has not been registered" );
        }
    }


    /**
     * Returns the subclass ID for a resource type ID.
     * @param resourceTypeID  The resource type ID of the subclass.
     */
    public static int getResourceTypeSubClassID( int resourceTypeID , int subClassID )
    {
        if ( (resourceTypeID >=0 )
            && ( resourceTypeID < currentResourceTypeNumber ) )
        {
            return resourceTypeSubClassID[ resourceTypeID ];
        }
        else
        {
            throw new Error( "Resource type ID \"" + resourceTypeID + "\" has not been registered" );
        }
    }


    /**
     * Updates the sub-class object and updates costs regarding the resource
     * type.
     */
    private void updateResourceType()
    {

        Iterator it;

        // Removing old sub-class

        if ( subClass != null )
        {

            // Removing state tasks from sub-class

            it = stateTaskSet.iterator();
            while ( it.hasNext() )
            {
                subClass.excludeStateTask( (SRCStateTask)it.next() );
            }

            // Removing precondition tasks from sub-class

            it = preconditionTaskSet.iterator();
            while ( it.hasNext() )
            {
                subClass.excludePreconditionTask( (SRCPreconditionTask)it.next() );
            }

            // Deleting sub-class object

            subClass.terminate();
            subClass = null;

        }

        if ( (resourceType >=0 )
            && ( resourceType < currentResourceTypeNumber ) )
        {

            // Creating new sub-class object

            try
            {
                subClass = (SRCInterface)subClassHandle[ resourceTypeSubClassID[ (int)resourceType ] ].newInstance();
            }
            catch( Exception e )
            {
                e.printStackTrace();
            }
            subClass.initialize( this , gsc ,
                resourceTypeDefinition[ (int)resourceType ] );

            // Updating type-related cost information

            if ( costs[ TYPE_INCONSISTENCY ] != 0 )
            {
                costs[ TYPE_INCONSISTENCY ] = 0;
                if ( costFunctionReference[ TYPE_INCONSISTENCY ] != null )
                {
                    costFunctionReference[ TYPE_INCONSISTENCY ].updateGlobalCosts();
                }
            }

            // Announcing connected state tasks to sub-class

            Hashtable newConstraintsHT = new Hashtable();
            it = stateTaskSet.iterator();
            StateTaskObjectConstraint stateTaskConstraint;
            SRCStateTask oldStateTask;
            SRCStateTask newStateTask;

            while ( it.hasNext() )
            {
                oldStateTask = (SRCStateTask)it.next();
                stateTaskConstraint = oldStateTask.taskConstraint;
                newStateTask
                    = subClass.includeStateTask( stateTaskConstraint );
                newConstraintsHT.put( stateTaskConstraint , newStateTask );
                newStateTask.taskConstraint = stateTaskConstraint;
                newStateTask.resourceTypeVariable
                    = oldStateTask.resourceTypeVariable;
                newStateTask.resourceType = oldStateTask.resourceType;
            }
            stConstraintsHT = newConstraintsHT;
            stateTaskSet = stConstraintsHT.values();

            // Announcing connected precondition tasks to sub-class

            newConstraintsHT = new Hashtable();
            it = preconditionTaskSet.iterator();
            PreconditionTaskObjectConstraint preconditionTaskConstraint;
            SRCPreconditionTask oldPreconditionTask;
            SRCPreconditionTask newPreconditionTask;

            while ( it.hasNext() )
            {
                oldPreconditionTask = (SRCPreconditionTask)it.next();
                preconditionTaskConstraint
                    = oldPreconditionTask.taskConstraint;
                newPreconditionTask = subClass.includePreconditionTask( preconditionTaskConstraint );
                newConstraintsHT.put( preconditionTaskConstraint , newPreconditionTask );
                newPreconditionTask.taskConstraint
                    = preconditionTaskConstraint;
                newPreconditionTask.resourceTypeVariable
                    = oldPreconditionTask.resourceTypeVariable;
                newPreconditionTask.resourceType
                    = oldPreconditionTask.resourceType;
            }
            ptConstraintsHT = newConstraintsHT;
            preconditionTaskSet = ptConstraintsHT.values();

        }
        else
        {

            // Updating type-related cost information

            if ( costs[ TYPE_INCONSISTENCY ] == 0 )
            {
                costs[ TYPE_INCONSISTENCY ] = 1;
                if ( costFunctionReference[ TYPE_INCONSISTENCY ] != null )
                {
                    costFunctionReference[ TYPE_INCONSISTENCY ].updateGlobalCosts();
                }
            }

        }

    }


    /***** SUB-CLASS HANDLING *****/


    /**
     * Registers a new sub-class.
     * @param ID  The sub-class ID that is to be used.
     * @param classString  The class' name path.
     */
    private static void registerSubClass( int ID , String classString )
    {

        if ( ( ID < 0 ) || ( ID >= SUBCLASS_NUMBER ) )
        {
            throw new Error( "Invalid sub-class ID" );
        }
        if ( subClassHandle[ ID ] != null )
        {
            throw new Error( "Sub-class ID has already been used" );
        }

        // Retrieving the sub-class handle

        try
        {
            subClassHandle[ ID ] = Class.forName( classString );
        }
        catch( Exception e )
        {
            e.printStackTrace();
        }

    }


    /***** SERVICE METHODS *****/


    /**
     * Returns the name of the current subclass.
     * @return  The name of the subclass, or "StateResourceConstraint" if
     *     no subclass is set.
     */
    public String getName()
    {
        if ( subClass != null )
        {
            return resourceTypeDefinition[ (int)resourceType ].getName();
        }
        else
        {
            return new String( "StateResourceConstraint" );
        }
    }


    /**
     * Returns the corresponding SRCPreconditionTask for a
     * PreconditionTaskObjectConstraint.<p>
     * @param ptConstraint  The PreconditionTaskObjectConstraint to query.
     * @return  The corresponding SRCPreconditionTask.
     */
    public SRCPreconditionTask getPreconditionTask( PreconditionTaskObjectConstraint ptConstraint )
    {
        return (SRCPreconditionTask)ptConstraintsHT.get( ptConstraint );
    }


    /**
     * Returns the corresponding SRCStateTask for a
     * StateTaskObjectConstraint.<p>
     * @param stateTaskObjectConstraint  The StateTaskObjectConstraint to query.
     * @return  The corresponding SRCStateTask.
     */
    public SRCStateTask getStateTask( StateTaskObjectConstraint stateTaskObjectConstraint )
    {
        return (SRCStateTask)stConstraintsHT.get( stateTaskObjectConstraint );

    }


    /**
     * Checks if the given precondition task is fullfilled.<p>
     * Returns 0 if fullfilled and 1 if not.
     * -1 will be returned if this task is not known.
     * @param ptConstraint  The PreconditionTaskObjectConstraint to be checked.
     * @return  0 if fullfilled, 1 if not, and -1 if this constraint is not known.
     */
    public int checkPrecondition( PreconditionTaskObjectConstraint ptConstraint )
    {
        SRCPreconditionTask task
            = (SRCPreconditionTask)ptConstraintsHT.get( ptConstraint );
        if ( task != null )
        {
            return task.checkPrecondition();
        }
        else
        {
            return -1;
        }

    }


    /**
     * Retrieves the projected state at a specific time.
     * @param time  The time at which the state is to be checked.
     * @return  The state variable's value at the given time.
     *     Returns -1 if there currently is no valid resource type.
     */
    public double getStateAtTime( long time )
    {

        if ( subClass != null )
        {
            return subClass.getStateAtTime( time );
        }
        else
        {
            return -1;
        }

    }


    /**
     * Returns a heuristic value for how difficult it would be to realize
     * a specific state at a specific time. The larger the value, the more
     * difficult would a realization be.
     * @param stateValue  The state that is to be realized.
     * @param time  The time at which the state is to be realized.
     * @return  The heuristic difficulty value. Returns -1 if there
     *     currently is no valid resource type.
     */
    public int getDifficultyForStateAtTime( double stateValue , long time )
    {

        if ( subClass != null )
        {
            return subClass.getDifficultyForStateAtTime( stateValue , time );
        }
        else
        {
            return -1;
        }

    }


    /**
     * Returns a heuristic value for how difficult it would be to repair
     * a specific contribution's consequence at a specific time with
     * respect to following preconditions. The larger the value, the more
     * difficult would a repair be.
     * @param contributionValue  The contribution that would occur.
     * @param time  The time at which the contribution would occur.
     * @return  The heuristic difficulty value (can be negative for a
     *     positive influence). Returns Integer.MAX_VALUE if there
     *     currently is no valid resource type.
     */
    public int getDifficultyForContributionAtTime( double contributionValue ,
        long time )
    {

        if ( subClass != null )
        {
            return subClass.getDifficultyForContributionAtTime( contributionValue , time );
        }
        else
        {
            return Integer.MAX_VALUE;
        }

    }


    // ***** EXTERNAL DYNAMICS *****


    public void update( double oldValue , ConnectionReference first ,
        ConnectionReference last )
    {

        if ( !modificationModus
            && ( ( (Variable)first.node ).value() != oldValue ) )
        {

            if ( last.node instanceof CurrentTimeObjectConstraint )
            {

                // Current time variable was updated

                currentTime = (long)( (Variable)first.node ).value();

                // Checking for a valid current time value

                if ( currentTime < 0 )
                {
                    throw new Error( "Invalid value \"" + currentTime
                        + "\" for current time" );
                }

                // Propagating update to sub-class

                if ( subClass != null )
                {
                    subClass.updateCurrentTime( (long)oldValue );
                }

            }
            else
            {

                if ( last.edgeID == Registry.STATE_EDGE )
                {

                    // State variable was updated

                    state = stateVariable.value();

                    // Propagating update to sub-class

                    if ( subClass != null )
                    {
                        subClass.updateCurrentState( oldValue );
                    }

                }
                else
                {

                    if ( last.edgeID == Registry.RESOURCETYPE_EDGE )
                    {

                        // Resource type variable was updated

                        resourceType = (long)resourceTypeVariable.value();
                        updateResourceType();

                    }
                    else
                    {

                        if ( last.node instanceof StateTaskObjectConstraint )
                        {

                            StateTaskObjectConstraint taskConstraint
                                = (StateTaskObjectConstraint)last.node;
                            SRCStateTask task = (SRCStateTask)stConstraintsHT.get( taskConstraint );

                            if ( first.edgeID == Registry.RESOURCETYPE_EDGE )
                            {

                                // A state task's resource type variable was
                                // updated

                                task.resourceType
                                    = (long)task.resourceTypeVariable.value();

                            }
                            else
                            {

                                // Propagating update to sub-class

                                if ( subClass != null )
                                {
                                    subClass.updateStateTask( task ,
                                        first.edgeID );
                                }

                            }

                        }
                        else
                        {

                            if ( last.node instanceof
                                PreconditionTaskObjectConstraint )
                            {

                                PreconditionTaskObjectConstraint taskConstraint = (PreconditionTaskObjectConstraint)last.node;
                                SRCPreconditionTask task = (SRCPreconditionTask)ptConstraintsHT.get( taskConstraint );

                                if ( first.edgeID == Registry.RESOURCETYPE_EDGE )
                                {

                                    // A precondition task's resource type
                                    // variable was updated

                                    task.resourceType = (long)task.resourceTypeVariable.value();

                                }
                                else
                                {

                                    // Propagating update to sub-class

                                    if ( subClass != null )
                                    {
                                        subClass.updatePreconditionTask( task , first.edgeID );
                                    }

                                }

                            }
                        }
                    }
                }
            }
        }

    }


    public void connectNode( ConnectionReference first ,
        ConnectionReference last )
    {

        if ( last.node instanceof CurrentTimeObjectConstraint )
        {

            // Current time variable/constraint was connected

            currentTimeConstraint = (ConventionalConstraint)last.node;

            // Fetching current time's variable

            Variable variable;
            synchronized( currentTimeConstraint.connections )
            {
                variable = currentTimeConstraint.connections.getSpecificVariable( Registry.PLAIN_EDGE );
            }
            if ( variable == null )
            {
                throw new Error( "The connected CurrentTimeObjectConstraint \"" + currentTimeConstraint + "\" has no variable" );
            }

            currentTime = (long)variable.value();

            // Checking for a valid current time value

            if ( currentTime < 0 )
            {
                throw new Error( "Invalid value \"" + currentTime
                    + "\" for current time" );
            }

            // Triggering sub-class creation if base graph is completed

            if ( connections.number() == 3 )
            {
                updateResourceType();
            }

        }
        else
        {

            if ( last.edgeID == Registry.STATE_EDGE )
            {

                // State variable was connected

                stateVariable = (Variable)last.node;
                state = stateVariable.value();

                // Triggering sub-class creation if base graph is completed

                if ( connections.number() == 3 )
                {
                    updateResourceType();
                }

            }
            else
            {

                if ( last.edgeID == Registry.RESOURCETYPE_EDGE )
                {

                    // Resource type variable was connected

                    resourceTypeVariable = (Variable)last.node;
                    modificationModus = true;
                    gsc.addIntegerConstraint( resourceTypeVariable );
                    modificationModus = false;
                    resourceType = (long)resourceTypeVariable.value();

                    // Triggering sub-class creation if base graph is completed

                    if ( connections.number() == 3 )
                    {
                        updateResourceType();
                    }

                }
                else
                {

                    if ( last.node instanceof StateTaskObjectConstraint )
                    {

                        // A state task was connected

                        StateTaskObjectConstraint taskConstraint
                            = (StateTaskObjectConstraint)last.node;

                        // Checking whether task has already been connected

                        SRCStateTask task = (SRCStateTask)stConstraintsHT.get( taskConstraint );
                        if ( task != null )
                        {
                            throw new Error( "The StateTaskObjectConstraint \""
                                + taskConstraint
                                + "\" has already been connected" );
                        }

                        // Storing task information in an SRCStateTask

                        if ( subClass != null )
                        {
                            task = subClass.includeStateTask( taskConstraint );
                        }
                        else
                        {
                            task = new SRCStateTask();
                        }
                        stConstraintsHT.put( taskConstraint , task );
                        stateTaskNumber++ ;
                        task.taskConstraint = taskConstraint;

                        // Fetching state task's resource type variable

                        Variable tcResourceTypeVariable;
                        synchronized( taskConstraint.connections )
                        {
                            tcResourceTypeVariable = taskConstraint.connections.getSpecificVariable( Registry.RESOURCETYPE_EDGE );
                        }
                        if ( tcResourceTypeVariable == null )
                        {
                            throw new Error( "The connected StateTaskObjectConstraint \"" + taskConstraint + "\" has no resource type variable" );
                        }

                        modificationModus = true;
                        gsc.addIntegerConstraint( tcResourceTypeVariable );
                        modificationModus = false;
                        task.resourceTypeVariable = tcResourceTypeVariable;
                        task.resourceType
                            = (long)tcResourceTypeVariable.value();

                    }
                    else
                    {

                        if ( last.node instanceof
                            PreconditionTaskObjectConstraint )
                        {

                            // A precondition task was connected

                            PreconditionTaskObjectConstraint taskConstraint
                                = (PreconditionTaskObjectConstraint)last.node;

                            // Checking whether task has already been connected

                            SRCPreconditionTask task = (SRCPreconditionTask)ptConstraintsHT.get( taskConstraint );

                            if ( task != null )
                            {
                                throw new Error( "The PreconditionTaskObjectConstraint \"" + taskConstraint + "\" has already been connected" );
                            }

                            // Storing task information in an
                            // SRCPreconditionTask

                            if ( subClass != null )
                            {
                                task = subClass.includePreconditionTask( taskConstraint );
                            }
                            else
                            {
                                task = new SRCPreconditionTask();
                            }
                            ptConstraintsHT.put( taskConstraint , task );
                            preconditionTaskNumber++ ;
                            task.taskConstraint = taskConstraint;

                            // Fetching precondition task's resource type variable

                            Variable tcResourceTypeVariable;
                            synchronized( taskConstraint.connections )
                            {
                                tcResourceTypeVariable = taskConstraint.connections.getSpecificVariable( Registry.RESOURCETYPE_EDGE );
                            }
                            if ( tcResourceTypeVariable == null )
                            {
                                throw new Error( "The connected PreconditionTaskObjectConstraint \"" + taskConstraint + "\" has no resource type variable" );
                            }

                            modificationModus = true;
                            gsc.addIntegerConstraint( tcResourceTypeVariable );
                            modificationModus = false;
                            task.resourceTypeVariable = tcResourceTypeVariable;
                            task.resourceType
                                = (long)tcResourceTypeVariable.value();

                        }
                        else
                        {
                            throw new Error( "Invalid connection of \"" + first.node + "\" to the DiscreteResourceConstraint" );
                        }
                    }
                }
            }
        }
        fireChange( this );
    }


    public void disconnectNode( ConnectionReference first ,
        ConnectionReference last )
    {

        if ( alive )
        {

            if ( last.edgeID == Registry.STATE_EDGE )
            {

                // State variable was disconnected

                // Propagating constraint's deletion to sub-class

                if ( subClass != null )
                {
                    subClass.terminate();
                }


            }
            else
            {

                if ( last.edgeID == Registry.RESOURCETYPE_EDGE )
                {

                    // Resource type variable was disconnected

                    modificationModus = true;
                    gsc.removeIntegerConstraint( resourceTypeVariable );
                    modificationModus = false;

                }
                else
                {

                    if ( last.node instanceof StateTaskObjectConstraint )
                    {

                        // StateTaskObjectConstraint was disconnected

                        StateTaskObjectConstraint taskConstraint
                            = (StateTaskObjectConstraint)last.node;

                        // Removing related SRCStateTask for the
                        // StateTaskObjectConstraint

                        SRCStateTask task = (SRCStateTask)stConstraintsHT.get( taskConstraint );
                        stConstraintsHT.remove( taskConstraint );
                        stateTaskNumber-- ;

                        // Removing IntegerConstraints from the resource
                        // type variable

                        modificationModus = true;
                        gsc.removeIntegerConstraint( task.resourceTypeVariable );
                        modificationModus = false;

                        // Propagating disconnection to sub-class

                        if ( subClass != null )
                        {
                            subClass.excludeStateTask( task );
                        }

                    }
                    else
                    {

                        if ( last.node instanceof
                            PreconditionTaskObjectConstraint )
                        {

                            // PreconditionTaskObjectConstraint was
                            // disconnected

                            PreconditionTaskObjectConstraint taskConstraint
                                = (PreconditionTaskObjectConstraint)last.node;

                            // Removing related SRCPreconditionTask for the
                            // PreconditionTaskObjectConstraint

                            SRCPreconditionTask task = (SRCPreconditionTask)ptConstraintsHT.get( taskConstraint );
                            ptConstraintsHT.remove( taskConstraint );
                            preconditionTaskNumber-- ;

                            // Removing IntegerConstraints from the resource
                            // type variable

                            modificationModus = true;
                            gsc.removeIntegerConstraint( task.resourceTypeVariable );
                            modificationModus = false;

                            // Propagating disconnection to sub-class

                            if ( subClass != null )
                            {
                                subClass.excludePreconditionTask( task );
                            }

                        }
                    }
                }
            }

        }

    }


    // ***** IMPROVEMENT (Service methods) *****

    /**
     * Disconnects and removes a state/precondition task from the constraint
     * if all of the corresponding action's action tasks have a pending
     * execution state. Returns whether the deletion was successful.
     * @param taskConstraint  The task to be removed.
     * @param forceRemoval  If true, the execution states of the action's
     *                      action tasks will not be considered.
     */
    public boolean removeTask(ConventionalConstraint taskConstraint ,
        boolean forceRemoval )
    {

        if ( !forceRemoval )
        {

            // Checking for valid execution states of action's tasks

            ConventionalConstraint actionConstraint;
            Object tasks;

            synchronized( taskConstraint.connections )
            {
                actionConstraint = (ConventionalConstraint)taskConstraint.connections.getSpecificConstraint( Registry.ACTION_OBJECTCONSTRAINT , Registry.PLAIN_EDGE );
            }
            if ( actionConstraint != null )
            {

                // Retrieving action's action tasks


                synchronized( actionConstraint.connections )
                {
                    tasks = actionConstraint.connections.getSpecificNodes( Registry.ACTIONTASK_OBJECTCONSTRAINT );
                }

                if ( tasks instanceof HashSet )
                {
                    tasks = (HashSet)( (HashSet)tasks ).clone();
                }
                else
                {
                    if ( tasks instanceof ConnectedNode )
                    {

                        Object node = tasks;
                        tasks = new HashSet();
                        ( (HashSet)tasks ).add( node );
                    }
                }

                // Checking for valid execution states

                if ( tasks != null )
                {

                    ActionTaskObjectConstraint actionTask;
                    Variable executionVariable;
                    Iterator it = ( (HashSet)tasks ).iterator();

                    while ( it.hasNext() )
                    {
                        actionTask = (ActionTaskObjectConstraint)( (ConnectedNode)it.next() ).node;
                        synchronized( actionTask.connections )
                        {
                            executionVariable = actionTask.connections.getSpecificVariable( Registry.EXECUTIONSTATE_EDGE );
                        }
                        if ( ( (long)executionVariable.value() )
                            != DiscreteResourceConstraint.PENDING )
                        {
                            return false;
                        }
                    }

                }
            }
        }

        // Disconnection from the state resource constraint

        Embedding e = new Embedding( 1 );
        gsc.extendEmbedding_internal( e , taskConstraint ,
            Registry.PLAIN_EDGE , Registry.INCOMING );
        gsc.reduceConstraint_internal( this , e );

        // Deletion and disconnection from the action object

        gsc.disconnectAndRemoveConstraint_internal( taskConstraint , Registry.ACTION_OBJECTCONSTRAINT , e );

        return true;

    }


    /**
     * Removes the state resource constraint (including all base graph
     * elements but the current time).
     */
    public void removeStateResourceConstraint()
    {

        alive = false;
        ConventionalConstraint taskConstraint;

        // Disconnecting/deleting all precondition tasks

        Iterator it = preconditionTaskSet.iterator();
        while ( it.hasNext() )
        {
            removeTask( ( (SRCPreconditionTask)it.next() ).taskConstraint , true );
        }

        // Disconnecting/deleting all state tasks

        it = stateTaskSet.iterator();
        while ( it.hasNext() )
        {
            removeTask( ( (SRCStateTask)it.next() ).taskConstraint , true );
        }

        // Self-destruction

        gsc.deleteConstraint_internal( this );

        // Resource type variable deletion

        int connectionNumber;
        synchronized( resourceTypeVariable.connections )
        {
            connectionNumber = resourceTypeVariable.connections.number();
        }
        if ( connectionNumber == 0 )
        {
            gsc.deleteVariable_internal( resourceTypeVariable );
        }

    }


    // ***** TYPE IMPROVEMENT (Heuristics) *****

    /**
     * Improves the current type costs by deleting itself.
     */
    private void improveTypeInconsistency()
    {
        if ( gsc.structuralChangesEnabled )
        {
            removeStateResourceConstraint();
            if ( Registry.LOG_CHANGES )
            {
                System.out.println( "improveTypeInconsistency: removeStateResourceConstraint" );
            }
        }
        else
        {
            if ( currentResourceTypeNumber > 0 )
            {
                gsc.setVariable( resourceTypeVariable ,
                    gsc.random.nextInt( currentResourceTypeNumber ) );
                if ( Registry.LOG_CHANGES )
                {
                    System.out.println( "improveTypeInconsistency: resource type set to " + resourceTypeDefinition[ (int)resourceType ].getName() );
                }
            }
            else
            {
                if ( Registry.LOG_CHANGES )
                {
                    System.out.println( "improveTypeInconsistency: improvement failed - no resource types defined" );
                }
            }
        }
    }


    public void improveCosts( int costID )
    {

        if ( Registry.LOG_CHANGES )
        {
            if ( costs[ TYPE_INCONSISTENCY ] == 0 )
            {
                System.out.println( "Improvement by " + resourceTypeDefinition[ (int)resourceType ].getName() + ":" );
            }
            else
            {
                System.out.println( "Improvement by (not yet typed) SRC:" );
            }
        }

        switch ( costID )
        {
        case PRECONDITION_FAILURE:
            {
                if ( costs[ TYPE_INCONSISTENCY ] == 0 )
                {
                    subClass.improvePreconditionInconsistency();
                }
                else
                {
                    throw new Error( "The type inconsistency costs were not connected to a cost function!" );
                }
                break;
            }
        case STATIC_GOAL_FAILURE:
            {
                if ( costs[ TYPE_INCONSISTENCY ] == 0 )
                {
                    subClass.improveGoalInconsistency();
                }
                else
                {
                    throw new Error( "The type inconsistency costs were not connected to a cost function!" );
                }
                break;
            }
        case TYPE_INCONSISTENCY:
            {
                improveTypeInconsistency();
                break;
            }
        default:
            {
                throw new Error( "The cost ID \"" + costID +
                    "\" is not supported by this constraint" );
            }
        }

    }


}
