
package com.ai_center.dragonbreath.constraints;

/** 
 * This constraint forces the sum over a set of variables 
 * to equal another variable.
 *
 * @author Alexander Nareyek
 * @version $Revision: 2.2 $
 * @see Variable
 */


import java.lang.Math;
import java.util.*;

import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.Registry;


public class SumConstraint extends ConventionalConstraint {

    // **********************
    // ***** ATTRIBUTES *****
    // **********************

    /** 
     * The constraint type's name.
     */
    private static final String TYPE_NAME = new String("Sum");

    /** 
     * The constraint's base graph.
     */
    private static final Embedding BASE_GRAPH = new Embedding(2);
    /** 
     * The constraint's number of extension graphs.
     */
    private static final int EXTENSION_GRAPH_NUMBER = 1;
    
    /** 
     * The constraint's extension graphs.
     */
    private static final Embedding[] EXTENSION_GRAPH = 
	new Embedding[EXTENSION_GRAPH_NUMBER];

    /** 
     * The sum variable.
     */
    private Variable sumVariable = null;
    /** 
     * The list of variables to be added.
     */
    private java.util.List addVariables = 
	Collections.synchronizedList(new LinkedList());
    /** 
     * The number of variables to be added.
     */
    private int addVariableNumber = 0;

    /** 
     * The value of the sum variable.
     */
    private double sum = 0;
    /** 
     * The sum of the added variables.
     */
    private double sumVariables = 0;


    // *******************
    // ***** METHODS *****
    // *******************

    public void create_graphs() {

	// Creating the base graph
	
	BASE_GRAPH.addConnectedNode(Registry.VARIABLE, 
				    Registry.SUM_EDGE, Registry.INCOMING);
	BASE_GRAPH.addConnectedNode(Registry.VARIABLE, 
				    Registry.ADD_EDGE, Registry.INCOMING);
	
	// Creating the extension graph
	
	EXTENSION_GRAPH[0] = new Embedding(1);
	EXTENSION_GRAPH[0].addConnectedNode(Registry.VARIABLE, 
					    Registry.ADD_EDGE, 
					    Registry.INCOMING);

    }

    public int getNumberOfCosts() {
	return 1;
    }

    public String getCostName(int costID) {
	if (costID == 0) {
	    return new String("difference inconsistency");
	} else {
	    throw new Error("Invalid cost ID");
	}
    }

    public String getTypeName() {
	return TYPE_NAME;
    }

    public Embedding getBaseGraph() {
	return BASE_GRAPH;
    }

    public int getExtensionGraphNumber() {
	return EXTENSION_GRAPH_NUMBER;
    }

    public Embedding getExtensionGraph(int graphID) {
	return EXTENSION_GRAPH[graphID];
    }

    public int getMaximalConnectionNumber() {
	// There is no maximum for extensions
	return Integer.MAX_VALUE;
    }

    public void initializeCostType(int costID) {
	// Nothing to do
    }

    /** 
     * Calculates the constraint's costs.
     */
    private void updateCosts() {
	costs[0] = (long) Math.ceil(Math.abs(sum - sumVariables));
	costFunctionReference[0].updateGlobalCosts();
    }

    public void update(double oldValue, ConnectionReference first, 
			  ConnectionReference last) {
	Variable variable = (Variable) first.node;
	if (variable == sumVariable) {
	    sum = sumVariable.value();
	} else {
	    sumVariables = sumVariables - oldValue + variable.value();
	}
	updateCosts();
    }

    public void connectNode(ConnectionReference first, 
			       ConnectionReference last) {

	Variable variable = (Variable) first.node;
	if (first.edgeID == Registry.SUM_EDGE) {
	    sumVariable = variable;
	    sum = sumVariable.value();
	    boolean structureOK;
	    synchronized(addVariables) {
		structureOK = (addVariables.size() >= 1);
	    }
	    if (structureOK) updateCosts();
	} else {
	    synchronized(addVariables) {
		addVariables.add(variable);
		addVariableNumber++;
	    }
	    sumVariables = sumVariables + variable.value();
	    if (sumVariable != null) updateCosts();
	}

    }

    public void disconnectNode(ConnectionReference first, 
				  ConnectionReference last) {

	if (first.edgeID == Registry.SUM_EDGE) {
	    sumVariable = null;
	    sum = 0;
	} else {
	    Variable variable = (Variable) first.node;
	    synchronized(addVariables) {
		addVariables.remove(variable);
		addVariableNumber--;
	    }
	    sumVariables = sumVariables - variable.value();
	}
	updateCosts();

    }

    public void improveCosts(int costID) {

	// Checking for valid cost ID

	if (costID != 0) {
	    throw new Error("The cost ID  \"" + costID + 
			    "\" is not supported by this constraint");
	}

	int i;
	boolean[] possibleChange = new boolean[addVariableNumber + 1];
	int possibleChanges = 0;
	if (possibleChange[addVariableNumber] = !sumVariable.isConstant()) {
	    possibleChanges++;
	}
	synchronized(addVariables) {
	    for (i = 0; i < addVariableNumber; i++) {
		if (possibleChange[i] = 
		    !((Variable) addVariables.get(i)).isConstant()) {
		    possibleChanges++;
		}
	    }
	}

	boolean ready = false;
	int chosenVariable;

	while (!ready && possibleChanges > 0) {
	    chosenVariable = gsc.random.nextInt(possibleChanges);

	    i = -1;
	    do {
		while (!possibleChange[++i]);
	    } while (chosenVariable-- > 0);

            if (i == addVariableNumber) {
		if (sum > sumVariables) {
		    if (sumVariable.lowerBound() < sum) {
			gsc.setVariable_internal(sumVariable, sumVariables);
			ready = true;
		    }
		} else {
		    if (sumVariable.upperBound() > sum) {
			gsc.setVariable_internal(sumVariable, sumVariables);
			ready = true;
		    }
		}
	    } else {
		Variable var = (Variable) addVariables.get(i);
		double v = var.value();
		if (sumVariables > sum) {
		    if (var.lowerBound() < v) {
			gsc.setVariable_internal(var, v + sum - sumVariables);
			ready = true;
		    }
		} else {
		    if (var.upperBound() > v) {
			gsc.setVariable_internal(var, v + sum - sumVariables);
			ready = true;
		    }
		}
	    }

            possibleChanges--;
	    possibleChange[i] = false;
	}

    }

}


