
package com.ai_center.dragonbreath.constraints;


import java.util.*;

import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.constraints.taskconstraint.*;
import com.ai_center.dragonbreath.Registry;
import com.ai_center.dragonbreath.utilities.ObjectIntSelection;
import com.ai_center.dragonbreath.utilities.OrderedShiftDLLElement;


/**
 * This constraint enforces that the connected tasks (via the action object
 * constraint) match a given configuration, and that the required temporal
 * relations are fulfilled. No improvement heuristic shifts an action or
 * state task before the current time. ATTENTION: An addition of action
 * types after constraint instance creation is not yet supported.
 * The cost component CONFIGURATION_INCONSISTENCY returns costs if
 * the connected tasks do not match a given configuration.
 * The cost component TEMPORAL_INCONSISTENCY calculates for each
 * temporal relation the minimal value that a variable would have
 * to be changed such that the relation is satisfied, and sums these
 * values. Costs related to incorrect execution states of action tasks
 * are added here as well.
 * The cost component TRANSITION_TEMPORAL_INCONSISTENCY calculates the
 * temporal inconsistencies for actions that are world transitions of the
 * transition constraint.
 * Note that the tasks that are initially connected to the action object
 * constraint will be taken as heuristic guidance on what could/should
 * be changed.
 *
 * @author Alexander Nareyek
 * @version $Revision: 2.32 $
 *
 * @see CurrentTimeObjectConstraint
 * @see ActionObjectConstraint
 * @see TransitionConstraint
 */
public class TaskConstraint extends ConventionalConstraint
{


    // **********************
    // ***** ATTRIBUTES *****
    // **********************


    /***** CONSTANTS *****/


    /**
     * Constant representing the temporal reference of a precondition/state
     * task.
     */
    public final static int TEMPORAL_REFERENCE = 0;
    /**
     * Constant representing the begin of an action task.
     */
    public final static int BEGIN = 0;
    /**
     * Constant representing the duration of an action task.
     */
    public final static int DURATION = 1;
    /**
     * Constant representing the end of an action task.
     */
    public final static int END = 2;

    /**
     * Constant representing the operator "less than".
     */
    public static final int LESS_THAN = 0;
    /**
     * Constant representing the operator "less or equals".
     */
    public static final int LESS_OR_EQUALS = 1;
    /**
     * Constant representing the operator "equals".
     */
    public static final int EQUALS = 2;
    /**
     * Constant representing the operator "greater or equals".
     */
    public static final int GREATER_OR_EQUALS = 3;
    /**
     * Constant representing the operator "greater than".
     */
    public static final int GREATER_THAN = 4;


    /***** ENGINE-INTEGRATION-RELATED VARIABLES *****/


    /**
     * The constraint type's name.
     */
    private static final String TYPE_NAME = new String( "Task Constraint" );

    /**
     * The identifier for the constraint's configuration-related cost
     * component.
     */
    public static final int CONFIGURATION_INCONSISTENCY = 0;
    /**
     * The identifier for the constraint's cost component related to the
     * inconsistency of temporal relations.
     */
    public static final int TEMPORAL_INCONSISTENCY = 1;
    /**
     * The identifier for the constraint's cost component related to the
     * inconsistency of transitions' temporal relations.
     */
    public static final int TRANSITION_TEMPORAL_INCONSISTENCY = 2;
    /**
     * Descriptions of the cost components.
     */
    public static String[] costDescription = new String[ 3 ];
    static
    {
        costDescription[ 0 ] = new String( "configuration inconsistency" );
        costDescription[ 1 ] = new String( "temporal inconsistency" );
        costDescription[ 2 ]
            = new String( "transitions' temporal inconsistency" );
    }

    /**
     * The constraint's base graph.
     */
    private static final Embedding BASE_GRAPH = new Embedding( 2 );


    /***** TYPE-RELATED VARIABLES *****/


    /**
     * The maximal number of action types that can be registered.
     */
    private final static int MAX_ACTION_TYPE_NUMBER = 512;

    /**
     * Number of registered action types.
     */
    private static int currentActionTypeNumber = 0;

    /**
     * Array of available action types.
     */
    private static TCActionType[] actionType
        = new TCActionType[ MAX_ACTION_TYPE_NUMBER ];

    /**
     * No action task's begin is moved before the current time
     * plus this buffer (stored in variable earliestStartingTime) to
     * enable more repair steps before a potential execution.
     */
    private static long repairTimeBuffer = 0;

    /**
     * Indicates if the execution states of action tasks can be changed.
     */
    private static boolean executionStateChangesEnabled = true;

    /**
     * Maximum of tasks for a valid action type.
     */
    private static long maxTasks = 0;

    /**
     * Structure for choosing alternatives (re-used among different
     * instances). Attention: Cannot be used in parallel!
     */
    private static ObjectIntSelection objectSelection;


    /***** GENERAL INTERNAL STRUCTURES *****/


    /**
     * The cost part of the temporal inconsistency related to the execution
     * states of action tasks.
     */
    private long executionStateCosts = 0;

    /**
     * Indicates whether the constraint is currently changing CSP components
     * and the update procedure does not need to be called.
     */
    private boolean modificationModus = false;

    /**
     * The internal copy of the current time variable's value.
     */
    private long currentTime = 0;
    /**
     * The earliest possible starting time to which an action task
     * can be moved by the constraint's heuristics.
     */
    private long earliestStartingTime = 0;

    /**
     * The connected action object.
     */
    private ConventionalConstraint actionConstraint;
    /**
     * The action type variable of the action object.
     */
    private Variable actionTypeVar;
    /**
     * The internal copy of the action type variable's value.
     */
    private int actionTypeID = -1;

    /**
     * List of relations ordered according to their subjective relation costs.
     */
    private OrderedTemporalRelationCostsRelationList
        temporalRelationsList = null;
    /**
     * Indicates which cost ID is to be used for violations of the
     * temporal relations.
     */
    private int temporalRelationCostID = TEMPORAL_INCONSISTENCY;

    /**
     * Hashtable from PreconditionTaskObjectConstraints to
     * TCPreconditionTaskInstances.
     */
    private Hashtable ptConstraintsHT = null;
    /**
     * A collection view of the precondition task instances of ptConstraintsHT.
     */
    private Collection preconditionTaskSet = null;
    /**
     * The number of connected precondition tasks.
     */
    private long preconditionTaskNumber = 0;
    /**
     * The number of externally connected precondition tasks.
     */
    private long externalPreconditionTaskNumber = 0;

    /**
     * Hashtable from ActionTaskObjectConstraints to TCActionTaskInstances.
     */
    private Hashtable atConstraintsHT = null;
    /**
     * A collection view of the action task instances of atConstraintsHT.
     */
    private Collection actionTaskSet = null;
    /**
     * The number of connected action tasks.
     */
    private long actionTaskNumber = 0;
    /**
     * The number of externally connected action tasks.
     */
    private long externalActionTaskNumber = 0;

    /**
     * Hashtable from StateTaskObjectConstraints to TCStateTaskInstances.
     */
    private Hashtable stConstraintsHT = null;
    /**
     * A collection view of the state task instances of stConstraintsHT.
     */
    private Collection stateTaskSet = null;
    /**
     * The number of connected state tasks.
     */
    private long stateTaskNumber = 0;
    /**
     * The number of externally connected state tasks.
     */
    private long externalStateTaskNumber = 0;

    /**
     * One of the tasks that was initially connected to the action.
     */
    private Object initiallyConnectedTask = null;

    /**
     * List of possibly matching action types, i.e., action types that can
     * be reached by an extension of the current structures.
     */
    private LinkedList possibleActionTypes = new LinkedList();

    /**
     * List of not matching action types, i.e., action types that cannot be
     * reached by an extension of the current structures.
     */
    private LinkedList impossibleActionTypes = new LinkedList();


    /***** TRANSITION-RELATED VARIABLES *****/


    /**
     * Indicates the reference precondition task. Only necessary for world
     * transitions, i.e., if the tasks are non-movable. Is "null" if
     * the action is no world transition that is connected to the
     * TransitionConstraint.
     */
    private TCPreconditionTaskInstance referenceTask;

    /**
     * The time point of the reference precondition task.
     */
    private long referenceTaskTime;


    /***** HEURISTICS-RELATED VARIABLES *****/


    /**
     * Indicates if a specific valid action type has been externally set.
     */
    private boolean actionTypeGiven = false;

    /**
     * The ID of a matching action type (-1 if no type is matching).
     */
    private int matchingActionTypeID = -1;

    /**
     * The ID of an action type to be established as desired configuration
     * (-1 if not yet determined).
     */
    private int desiredActionTypeID = -1;
    /**
     * The time point at which the desired action type's orientation task
     * should be placed (if there were initially connected tasks).
     */
    private long desiredTemporalAnchor = Long.MIN_VALUE;

    /**
     * An array indicating if a specific action type is OK or heuristically
     * undesired (though potentially possible). Action types are marked as
     * undesired if:
     *  - A specific action type was externally changed to a new one.
     *  - An action type involves a task with has equal properties as one
     *     that was externally deleted from the action.
     *  - An action type does not involve a task with equal properties as
     *     one that was externally deleted from the action.
     */
    private boolean[] actionTypeOK = new boolean[ MAX_ACTION_TYPE_NUMBER ];

    /**
     * Indicates if the difference lists to the desired configuration were
     * established.
     */
    private boolean differenceListsEstablished = false;
    /**
     * The difference list of task instances to be removed for the desired
     * configuration.
     */
    private LinkedList differenceRemovalList;
    /**
     * The difference list of task types to be added for the desired
     * configuration.
     */
    private LinkedList differenceAdditionList;

    /**
     * Indicates whether the current change was made by a heuristic
     * of the constraint.
     */
    private boolean ownHeuristicChange = false;

    /**
     * Variable that counts the number of external changes of tasks'
     * temporal variables.
     */
    private int externalTaskMoveCounter;


    // *******************
    // ***** METHODS *****
    // *******************


    public void create_graphs()
    {

        // Creating the base graph

        BASE_GRAPH.addConnectedNode( Registry.CURRENTTIME_OBJECTCONSTRAINT , Registry.PLAIN_EDGE , Registry.INCOMING );
        BASE_GRAPH.addConnectedNode( Registry.ACTION_OBJECTCONSTRAINT , Registry.PLAIN_EDGE , Registry.INCOMING );

    }


    public int getNumberOfCosts()
    {
        return 3;
    }


    public String getCostName( int costID )
    {
        if ( costID < 3 )
        {
            return costDescription[ costID ];
        }
        else
        {
            throw new Error( "Invalid cost ID" );
        }
    }


    public String getTypeName()
    {
        return TYPE_NAME;
    }


    public Embedding getBaseGraph()
    {
        return BASE_GRAPH;
    }


    public int getExtensionGraphNumber()
    {
        return 0;
    }


    public Embedding getExtensionGraph( int graphID )
    {
        return null;
    }


    public int getMaximalConnectionNumber()
    {
        // There is no maximum for extensions
        return Integer.MAX_VALUE;
    }


    public void initializeInternalStructures()
    {
        int i;
        for ( i = 0 ; i < currentActionTypeNumber ; i++ )
        {
            possibleActionTypes.add( actionType[ i ] );
        }
        for ( i = 0 ; i < MAX_ACTION_TYPE_NUMBER ; i++ )
        {
            actionTypeOK[ i ] = true;
        }

        temporalRelationsList = new OrderedTemporalRelationCostsRelationList();

        ptConstraintsHT = new Hashtable();
        preconditionTaskSet = ptConstraintsHT.values();
        atConstraintsHT = new Hashtable();
        actionTaskSet = atConstraintsHT.values();
        stConstraintsHT = new Hashtable();
        stateTaskSet = stConstraintsHT.values();

        objectSelection = new ObjectIntSelection( gsc );

    }


    public void initializeCostType( int costID )
    {
        costFunctionReference[ costID ].updateGlobalCosts();
    }


    /***** TYPE CUSTOMIZATION *****/


    /**
     * Sets a temporal buffer value for all TaskConstraints to enable more
     * repair steps before a potential action task execution. No action
     * task begin is moved by the constraints' heuristics before the
     * current time plus this buffer. ATTENTION: A call after a creation
     * of constraint instances will only get propagated to these instances
     * after the current time has changed. Use an instance's function
     * "noteRepairTimeBufferUpdate()" in addition if an immediate
     * propagation is needed.
     * @param buffer  The temporal buffer value.
     */
    public static void setRepairTimeBuffer( long buffer )
    {
        repairTimeBuffer = buffer;
    }


    /**
     * Sets whether the execution states of action tasks can be changed by
     * the constraint. If only external changes are possible, actions will
     * only be deleted if all its action tasks have non-pending states.
     * @param check  Indicates if the execution states can be changed.
     */
    public static void enableExecutionStateChanges( boolean enableSwitch )
    {
        executionStateChangesEnabled = enableSwitch;
    }


    /**
     * Reserves an action type ID and returns it.
     */
    public static int registerNewActionType()
    {
        if ( currentActionTypeNumber == MAX_ACTION_TYPE_NUMBER )
        {
            throw new Error( "No more action type IDs available" );
        }
        actionType[ currentActionTypeNumber ]
            = new TCActionType( currentActionTypeNumber );
        return currentActionTypeNumber++ ;
    }


    /**
     * Reserves an action type ID and adds a name to it.
     * @param actionName  The name of the action.
     */
    public static int registerNewActionType( String actionName )
    {
        int actionTypeID = registerNewActionType();
        addNameToActionType( actionName , actionTypeID );
        return actionTypeID;
    }


    /**
     * Sets a name for an action type ID.
     * @param actionName  The name of the action.
     * @param actionTypeID  The ID of the action type to be named.
     */
    public static synchronized void addNameToActionType( String actionName , int actionTypeID )
    {

        if ( ( actionTypeID > currentActionTypeNumber )
            || ( actionTypeID < 0 ) )
        {
            throw new Error( "Invalid action type ID \""
                + actionTypeID + "\"" );
        }
        actionType[ actionTypeID ].actionName = actionName;

    }


    /**
     * Returns a name for the given actionTypeID or null if no name is given
     * to this actionTypeID.<p>
     * @param actionTypeID  The ID of the action type to be named.
     * @return  The name of the action.
     */
    public static synchronized String getNameForActionType( int actionTypeForNameID )
    {
        if ( ( actionTypeForNameID > currentActionTypeNumber )
            || ( actionTypeForNameID < 0 ) )
        {
            return null;
        } else {
            return (String)actionType[ actionTypeForNameID ].actionName;
        }

    }


    /**
     * Returns the number of action types that have been reserved so far.
     * @return  The current actionType number.
     */
    public static int getCurrentActionTypeNumber()
    {
        return currentActionTypeNumber;
    }


    /**
     * Returns a TCActionType for a given actionTypeID.<p>
     *  @param actionTypeID The actionType number to query.
     *  @return The corresponding TCActionType.
     */
    public static TCActionType getTCActionType( int actionTypeID )
    {
        return actionType[ actionTypeID ];
    }


    /**
     * Reserves an action type ID for a transition and returns it. All
     * connected tasks will be announced to the TransitionConstraint class.
     */
    public static int registerNewTransitionActionType()
    {
        if ( currentActionTypeNumber == MAX_ACTION_TYPE_NUMBER )
        {
            throw new Error( "No more action type IDs available" );
        }
        actionType[ currentActionTypeNumber ]
            = new TCActionType( currentActionTypeNumber );
        actionType[ currentActionTypeNumber ].transitionID = TransitionConstraint.registerNewTransitionType( currentActionTypeNumber );
        return currentActionTypeNumber++ ;
    }


    /**
     * Adds a precondition task to an action type. Returns an object handle
     * to identify the action type's task.
     * @param actionTypeID  The ID of the action type to be extended.
     * @param resourceType  The required resource type for the task.
     * @param state  The required state to be checked for by the task.
     */
    public static TCTask addPreconditionTaskToActionType( int actionTypeID , long resourceType , long state )
    {

        if ( ( actionTypeID < 0 )
            || ( actionTypeID >= currentActionTypeNumber ) )
        {
            throw new Error( "Invalid action type ID \""
                + actionTypeID + "\"" );
        }

        TCPreconditionTask preconditionTask
            = new TCPreconditionTask( resourceType , state );
        actionType[ actionTypeID ].preconditionTasks.add( preconditionTask );
        actionType[ actionTypeID ].preconditionTaskNumber++ ;

        long taskNumber = actionType[ actionTypeID ].preconditionTaskNumber
            + actionType[ actionTypeID ].actionTaskNumber
            + actionType[ actionTypeID ].stateTaskNumber;
        if ( taskNumber > maxTasks )
        {
            maxTasks = taskNumber;
        }

        // First task becomes the reference task for relative positioning

        if ( ( actionType[ actionTypeID ].preconditionTaskNumber == 1 ) && ( actionType[ actionTypeID ].actionTaskNumber == 0 ) && ( actionType[ actionTypeID ].stateTaskNumber == 0 ) )
        {
            preconditionTask.shiftMin = 0;
            preconditionTask.shiftMax = 0;
            preconditionTask.relativeShift = 0;
        }

        // Notifying TransitionConstraint if the action is a transition

        if ( actionType[ actionTypeID ].transitionID != -1 )
        {
            TransitionConstraint.addPreconditionTaskToTransitionType( actionType[ actionTypeID ].transitionID , resourceType , state );
        }

        return preconditionTask;

    }


    /**
     * Adds an action task to an action type. Returns an object handle
     * to identify the action type's task.
     * @param actionTypeID  The ID of the action type to be extended.
     * @param resourceType  The required resource type for the action task.
     * @param operation  The required operation for the action task.
     * @param duration  The required duration for the action task.
     * @param consumption  The required consumption for the action task.
     */
    public static TCTask addActionTaskToActionType( int actionTypeID , long resourceType , long operation , long duration , double consumption )
    {

        if ( (actionTypeID < 0 )
            || ( actionTypeID >= currentActionTypeNumber ) )
        {
            throw new Error( "Invalid action type ID" );
        }

        TCActionTask actionTask = new TCActionTask( resourceType , operation , duration , consumption );
        actionType[ actionTypeID ].actionTasks.add( actionTask );
        actionType[ actionTypeID ].actionTaskNumber++ ;

        long taskNumber = actionType[ actionTypeID ].preconditionTaskNumber
            + actionType[ actionTypeID ].actionTaskNumber
            + actionType[ actionTypeID ].stateTaskNumber;
        if ( taskNumber > maxTasks )
        {
            maxTasks = taskNumber;
        }

        // First task becomes the reference task for relative positioning

        if ( ( actionType[ actionTypeID ].preconditionTaskNumber == 0 ) && ( actionType[ actionTypeID ].actionTaskNumber == 1 ) && ( actionType[ actionTypeID ].stateTaskNumber == 0 ) )
        {
            actionTask.shiftMin = 0;
            actionTask.shiftMax = 0;
            actionTask.relativeShift = 0;
        }

        return actionTask;

    }


    /**
     * Adds a state task to an action type. Returns an object handle
     * to identify the action type's task.
     * @param actionTypeID  The ID of the action type to be extended.
     * @param resourceType  The required resource type for the task.
     * @param contribution  The required contribution for the state task.
     */
    public static TCTask addStateTaskToActionType( int actionTypeID , long resourceType , long contribution )
    {

        if ( ( actionTypeID < 0 )
            || ( actionTypeID >= currentActionTypeNumber ) )
        {
            throw new Error( "Invalid action type ID" );
        }

        TCStateTask stateTask = new TCStateTask( resourceType , contribution );
        actionType[ actionTypeID ].stateTasks.add( stateTask );
        actionType[ actionTypeID ].stateTaskNumber++ ;

        long taskNumber = actionType[ actionTypeID ].preconditionTaskNumber
            + actionType[ actionTypeID ].actionTaskNumber
            + actionType[ actionTypeID ].stateTaskNumber;
        if ( taskNumber > maxTasks )
        {
            maxTasks = taskNumber;
        }

        // First task becomes the reference task for relative positioning

        if ( ( actionType[ actionTypeID ].preconditionTaskNumber == 0 ) && ( actionType[ actionTypeID ].actionTaskNumber == 0 ) && ( actionType[ actionTypeID ].stateTaskNumber == 1 ) )
        {
            stateTask.shiftMin = 0;
            stateTask.shiftMax = 0;
            stateTask.relativeShift = 0;
        }

        // Notifying TransitionConstraint if the action is a transition

        if ( actionType[ actionTypeID ].transitionID != -1 )
        {
            TransitionConstraint.addStateTaskToTransitionType( actionType[ actionTypeID ].transitionID , resourceType , contribution );
        }

        return stateTask;

    }


    /**
     * Propagates an update of a task's bounds to another task that is
     * related by a temporal relation.
     * @param temporalRelation  The temporal relation that is to be used for
     *     the propagation.
     * @param variable1Changed  Indicates whether the first variable has been
     *     changed (second variable otherwise).
     * @param temporalRelations  Set of temporal relations that need to be
     *     considered for further propagation.
     */
    public static void propagateBounds( TCTemporalRelation temporalRelation ,
         boolean variable1Changed , HashSet temporalRelations )
    {

        // Only considering a propagation if the other variable is not
        // a constant

        TCTask task1 = (TCTask)temporalRelation.task1;
        TCTask task2 = (TCTask)temporalRelation.task2;
        int variableID1 = temporalRelation.variableID1;
        int variableID2 = temporalRelation.variableID2;
        long newMin;
        long newMax;
        boolean propagated = false;

        if ( !variable1Changed )
        {

            // Switching task info if variable 2 was changed

            TCTask taskTemp = task1;
            task1 = task2;
            task2 = taskTemp;
            int variableIDTemp = variableID1;
            variableID1 = variableID2;
            variableID2 = variableIDTemp;

        }

        // Determining new min bound

        if ( task1.shiftMin != Long.MIN_VALUE )
        {

            if ( variable1Changed
                || ( temporalRelation.comparator == EQUALS ) )
            {

                // Calculating loosest possible min bound

                if ( temporalRelation.comparator == EQUALS )
                {
                    if ( variable1Changed )
                    {
                        newMin = task1.shiftMin - temporalRelation.constant;
                    }
                    else
                    {
                        newMin = task1.shiftMin + temporalRelation.constant;
                    }
                }
                else
                {
                    newMin = task1.shiftMin - temporalRelation.constant + 1;
                }

                if ( variableID1 == END )
                {
                    newMin += ( (TCActionTask)task1 ).duration;
                }
                if ( variableID2 == END )
                {
                    newMin -= ( (TCActionTask)task2 ).duration;
                }

                // Adapting current min bound if necessary

                if ( newMin > task2.shiftMin )
                {
                    task2.shiftMin = newMin;
                    propagated = true;
                }

            }

        }

        // Determining new max bound

        if ( task1.shiftMax != Long.MAX_VALUE )
        {

            if ( !variable1Changed
                || ( temporalRelation.comparator == EQUALS ) )
            {

                // Calculating loosest possible max bound

                if ( temporalRelation.comparator == EQUALS )
                {
                    if ( variable1Changed )
                    {
                        newMax = task1.shiftMax - temporalRelation.constant;
                    }
                    else
                    {
                        newMax = task1.shiftMax + temporalRelation.constant;
                    }
                }
                else
                    {
                        newMax = task1.shiftMax
                            + temporalRelation.constant - 1;
                    }

                if ( variableID1 == END )
                {
                    newMax += ( (TCActionTask)task1 ).duration;
                }
                if ( variableID2 == END )
                {
                    newMax -= ( (TCActionTask)task2 ).duration;
                }

                // Adapting current max bound if necessary

                if ( newMax < task2.shiftMax )
                {
                    task2.shiftMax = newMax;
                    propagated = true;
                }

            }

        }

        if ( propagated )
        {

            if ( task2.shiftMax < task2.shiftMin )
            {
                throw new Error( "Temporally inconsistent action defintion" );
            }

            // Adapting relative shift of task

            if ( task2.shiftMax == Long.MAX_VALUE )
            {
                task2.relativeShift = task2.shiftMin;
            }
            else
            {
                if ( task2.shiftMin == Long.MIN_VALUE )
                {
                    task2.relativeShift = task2.shiftMax;
                }
                else
                {
                    task2.relativeShift
                        = (task2.shiftMin + task2.shiftMax ) / 2;
                }
            }

            // Recursive propagation call to related temporal relations

            Iterator it = temporalRelations.iterator();
            TCTemporalRelation next;
            while ( it.hasNext() )
            {
                next = (TCTemporalRelation)it.next();
                if ( next.task1 == task2 )
                {
                    if ( next.task2.shiftMin != next.task2.shiftMax )
                    {
                        propagateBounds( next , true , temporalRelations );
                    }
                }
                else
                {
                    if ( next.task2 == task2 )
                    {
                        if ( next.task1.shiftMin != next.task1.shiftMax )
                        {
                            propagateBounds( next , false ,
                                temporalRelations );
                        }
                    }
                }
            }

        }

    }


    /**
     * Adds a temporal relation to an action type.
     * The relation has the form "var1 comparator var2 + const".
     * @param actionTypeID  The ID of the action type to be extended.
     * @param task1  The task of the first variable.
     * @param edgeID1  Indicates the first task's variable to be involved.
     * @param comparator  The relation's comparator.
     * @param task2  The task of the second variable.
     * @param edgeID2  Indicates the second task's variable to be involved.
     * @param constant The constant to be added to the second variable.
     */
    public static void addTemporalRelationToActionType( int actionTypeID , TCTask task1 , int edgeID1 , int comparator , TCTask task2 , int edgeID2 , long constant )
    {

        if ( ( actionTypeID < 0 )
            || ( actionTypeID >= currentActionTypeNumber ) )
        {
            throw new Error( "Invalid action type ID" );
        }

        // Checking whether first and second task have been registered for
        // the action type

        if ( !actionType[ actionTypeID ].preconditionTasks.contains( task1 )
            && !actionType[ actionTypeID ].actionTasks.contains( task1 )
            && !actionType[ actionTypeID ].stateTasks.contains( task1 ) )
        {
            throw new Error( "The task \"" + task1 + "\" has not been registered for action type ID " + actionTypeID );
        }

        if ( !actionType[ actionTypeID ].preconditionTasks.contains( task2 )
            && !actionType[ actionTypeID ].actionTasks.contains( task2 )
            && !actionType[ actionTypeID ].stateTasks.contains( task2 ) )
        {
            throw new Error( "The task \"" + task2 + "\" has not been registered for action type ID " + actionTypeID );
        }

        // Checking whether edgeIDs are valid

        if ( task1 instanceof TCActionTask )
        {
            if ( ( edgeID1 != BEGIN ) && ( edgeID1 != END ) )
            {
                throw new Error( "Invalid variable indicator has been used" );
            }
        }
        else
        {
            if ( edgeID1 != TEMPORAL_REFERENCE )
            {
                throw new Error( "Invalid variable indicator has been used" );
            }
        }

        if ( task2 instanceof TCActionTask )
        {
            if ( ( edgeID2 != BEGIN ) && ( edgeID2 != END ) )
            {
                throw new Error( "Invalid variable indicator has been used" );
            }
        }
        else
        {
            if ( edgeID2 != TEMPORAL_REFERENCE )
            {
                throw new Error( "Invalid variable indicator has been used" );
            }
        }

        // Transforming LESS_OR_EQUALS, GREATER_THAN and
        // GREATER_OR_EQUALS comparators

        if ( comparator == LESS_OR_EQUALS )
        {
            constant++ ;
            comparator = LESS_THAN;
        }
        if ( ( comparator == GREATER_THAN)
            || ( comparator == GREATER_OR_EQUALS ) )
        {
            if ( comparator == GREATER_OR_EQUALS )
            {
                constant-- ;
            }
            TCTask taskCache = task1;
            task1 = task2;
            task2 = taskCache;
            int edgeIDCache = edgeID1;
            edgeID1 = edgeID2;
            edgeID2 = edgeIDCache;
            constant = -constant;
            comparator = LESS_THAN;
        }

        // Adding temporal relation object

        TCTemporalRelation temporalRelation = new TCTemporalRelation( task1 , edgeID1 , comparator , task2 , edgeID2 , constant );
        actionType[ actionTypeID ].temporalRelations.add( temporalRelation );

        // Propagating bounds for relative distances

        if ( task1.shiftMin != task1.shiftMax )
        {
            propagateBounds( temporalRelation , false , actionType[ actionTypeID ].temporalRelations );
        }
        if ( task2.shiftMin != task2.shiftMax )
        {
            propagateBounds( temporalRelation , true , actionType[ actionTypeID ].temporalRelations );
        }

    }


    /***** INFO OUTPUT *****/


    /**
     * Prints the relations.
     */
    public String toString()
    {

        StringBuffer s = new StringBuffer();

        if ( temporalRelationsList != null )
        {

            OrderedShiftDLLElement currentLE = temporalRelationsList.first;
            TCTemporalRelationInstance relation;

            while ( currentLE != null )
            {
                relation = (TCTemporalRelationInstance)currentLE.element;
                s.append( relation.costs + " - " + relation.toString() + "\n" );
                currentLE = currentLE.next;
            }

        }
        else
        {
            s.append( "no relations existent" );
        }

        return s.toString();

    }


    /***** ACTION TYPE MATCHING *****/


    /**
     * Compares the current configuration of connected tasks to an action
     * type. Returns 1 if there is a full match, -1 if there no configuration
     * extension possible to result in a match, and 0 otherwise.
     * @param ID  The ID of the action type to be compared.
     */
    private int compareConfigurationToActionType( int ID )
    {

        TCActionType selectedActionType = actionType[ ID ];

        // Simple pre-check if task numbers are not equal

        if ( ( preconditionTaskNumber
            != selectedActionType.preconditionTaskNumber)
            || ( actionTaskNumber != selectedActionType.actionTaskNumber )
            || ( stateTaskNumber != selectedActionType.stateTaskNumber ) )
        {

            if ( ( preconditionTaskNumber
                <= selectedActionType.preconditionTaskNumber )
                && ( actionTaskNumber <= selectedActionType.actionTaskNumber )
                && ( stateTaskNumber <= selectedActionType.stateTaskNumber ) )
            {
                return 0;
            }
            else
            {
                return -1;
            }
        }

        // Checking precondition task instances

        Iterator itInstance;
        Iterator itType;
        boolean found;

        TCPreconditionTask preconditionTask;
        TCPreconditionTaskInstance preconditionTaskInstance = null;

        // Initializing instance markings

        itInstance = preconditionTaskSet.iterator();
        while ( itInstance.hasNext() )
        {
            ( (TCPreconditionTaskInstance)itInstance.next() ).found = null;
        }

        // Iterating through action type's precondition tasks

        itType = selectedActionType.preconditionTasks.iterator();

        while ( itType.hasNext() )
        {
            preconditionTask = (TCPreconditionTask)itType.next();

            // Searching for corresponding precondition task instance

            itInstance = preconditionTaskSet.iterator();
            found = false;

            while ( !found && itInstance.hasNext() )
            {
                preconditionTaskInstance
                    = (TCPreconditionTaskInstance)itInstance.next();
                if ( preconditionTaskInstance.found == null )
                {
                    found = ( ( preconditionTaskInstance.resourceType == preconditionTask.resourceType ) && ( preconditionTaskInstance.state == preconditionTask.state ) );
                }
            }

            if ( found ) {
                preconditionTaskInstance.found = preconditionTask;
            }
            else
            {
                return -1;
            }
        }

        // Checking action task instances

        TCActionTask actionTask;
        TCActionTaskInstance actionTaskInstance = null;

        // Initializing instance markings

        itInstance = actionTaskSet.iterator();
        while ( itInstance.hasNext() )
        {
            ( (TCActionTaskInstance)itInstance.next() ).found = null;
        }

        // Iterating through action type's action tasks

        itType = selectedActionType.actionTasks.iterator();

        while ( itType.hasNext() )
        {
            actionTask = (TCActionTask)itType.next();

            // Searching for corresponding action task instance

            itInstance = actionTaskSet.iterator();
            found = false;

            while ( !found && itInstance.hasNext() )
            {
                actionTaskInstance = (TCActionTaskInstance)itInstance.next();
                if ( actionTaskInstance.found == null )
                {
                    found = ( ( actionTaskInstance.resourceType == actionTask.resourceType ) && ( actionTaskInstance.operation == actionTask.operation ) && ( actionTaskInstance.tempValue[ DURATION ] == actionTask.duration ) && ( actionTaskInstance.consumption == actionTask.consumption ) );
                }
            }

            if ( found )
            {
                actionTaskInstance.found = actionTask;
            }
            else
            {
                return -1;
            }
        }

        // Checking state task instances

        TCStateTask stateTask;
        TCStateTaskInstance stateTaskInstance = null;

        // Initializing instance markings

        itInstance = stateTaskSet.iterator();
        while ( itInstance.hasNext() )
        {
            ( (TCStateTaskInstance)itInstance.next() ).found = null;
        }

        // Iterating through action type's state tasks

        itType = selectedActionType.stateTasks.iterator();

        while ( itType.hasNext() )
        {
            stateTask = (TCStateTask)itType.next();

            // Searching for corresponding state task instance

            itInstance = stateTaskSet.iterator();
            found = false;

            while ( !found && itInstance.hasNext() )
            {
                stateTaskInstance = (TCStateTaskInstance)itInstance.next();
                if ( stateTaskInstance.found == null )
                {
                    found = ( ( stateTaskInstance.resourceType == stateTask.resourceType ) && ( stateTaskInstance.contribution == stateTask.contribution ) );
                }
            }

            if ( found )
            {
                stateTaskInstance.found = stateTask;
            }
            else
            {
                return -1;
            }
        }

        return 1;

    }


    /**
     * Checks if the currect connections represent a valid action type
     * configuration. If no specific action type was externally provided
     * by the corresponding variable, checked action types with impossible
     * matches are moved to the list of impossible action types.
     */
    private void checkForValidActionType()
    {

        boolean matched = false;
        TCActionType selectedActionType = null;

        if ( actionTypeGiven )
        {

            if ( actionTypeOK[ actionTypeID ] )
            {
                matched = ( compareConfigurationToActionType( actionTypeID ) == 1 );
                selectedActionType = actionType[ actionTypeID ];
            }
            else
            {
                matched = false;
            }

        }
        else
        {

            // Searching for a action type match with the current configuration

            int matchResult;
            Iterator it = possibleActionTypes.iterator();

            while ( !matched && it.hasNext() )
            {

                selectedActionType = (TCActionType)it.next();
                if ( actionTypeOK[ selectedActionType.ID ] )
                {
                    matchResult = compareConfigurationToActionType( selectedActionType.ID );
                }
                else
                {
                    matchResult = -1;
                }

                if ( matchResult == 1 )
                {
                    matched = true;
                }
                else
                {
                    matched = false;
                    if ( matchResult == -1 )
                    {
                        impossibleActionTypes.add( selectedActionType );
                        it.remove();
                    }
                }

            }

        }

        // Determining configuration costs

        int newCosts;

        if ( matched )
        {
            matchingActionTypeID = selectedActionType.ID;
            if ( matchingActionTypeID == actionTypeID )
            {
                newCosts = 0;
            }
            else
            {
                newCosts = 1;
            }
        }
        else
        {
            matchingActionTypeID = -1;
            newCosts = 1;
        }

        // Updating non-configuration-related structures and costs if
        // necessary

        if ( newCosts == 0 )
        {

            long oldTRCosts = costs[ temporalRelationCostID ];

            // Synthesizing temporal relations

            clearTemporalRelationsList();
            costs[ temporalRelationCostID ] = 0;
            if ( referenceTask != null )
            {
                temporalRelationCostID = TRANSITION_TEMPORAL_INCONSISTENCY;
            }
            else
            {
                temporalRelationCostID = TEMPORAL_INCONSISTENCY;
            }
            synthesizeTemporalRelations();

            // Evaluating execution states of action tasks

            executionStateCosts = 0;
            TCActionTaskInstance actionTaskInstance;

            Iterator it = actionTaskSet.iterator();
            while ( it.hasNext() )
            {
                actionTaskInstance = (TCActionTaskInstance)it.next();
                actionTaskInstance.executionStateCosts = 0;
                executionStateCosts
                    += evaluateExecutionState( actionTaskInstance );
            }
            costs[ temporalRelationCostID ] += executionStateCosts;

            if ( oldTRCosts != costs[ temporalRelationCostID ] ) {
                if ( costFunctionReference[ temporalRelationCostID ] != null )
                {
                    costFunctionReference[ temporalRelationCostID ].updateGlobalCosts();
                }
            }

        }
        else
        {
            if ( costs[ temporalRelationCostID ] > 0 )
            {
                costs[ temporalRelationCostID ] = 0;
                if ( costFunctionReference[ temporalRelationCostID ] != null )
                {
                    costFunctionReference[ temporalRelationCostID ].updateGlobalCosts();
                }
            }
        }

        // Setting new configuration costs

        if ( newCosts != costs[ CONFIGURATION_INCONSISTENCY ] )
        {
            costs[ CONFIGURATION_INCONSISTENCY ] = newCosts;
            if ( costFunctionReference[ CONFIGURATION_INCONSISTENCY ] != null )
            {
                costFunctionReference[ CONFIGURATION_INCONSISTENCY ].updateGlobalCosts();
            }
        }

    }


    /**
     * Moves now possibly matching action types from the set of not
     * matching to the set of possibly matching action types.
     */
    private void reactivatePossibleActionTypes()
    {

        TCActionType selectedActionType = null;
        Iterator it = impossibleActionTypes.iterator();

        while ( it.hasNext() )
        {
            selectedActionType = (TCActionType)it.next();
            if ( actionTypeOK[ selectedActionType.ID ]
                && ( preconditionTaskNumber <= selectedActionType.preconditionTaskNumber )
                && ( actionTaskNumber <= selectedActionType.actionTaskNumber )
                &&  (stateTaskNumber <= selectedActionType.stateTaskNumber ) )
            {
                possibleActionTypes.add( selectedActionType );
                it.remove();
            }
        }

     }


    /***** TEMPORAL RELATION HANDLING *****/


    /**
     * Causes an internal update of the earliest possible starting time
     * to which action tasks can be moved by the constraint's heuristics.
     * Only necessary if the function "setRepairTimeBuffer(long buffer)"
     * has been called after the constraint instance's creation. If the
     * current time is changed, this update is made automatically.
     */
    public void noteRepairTimeBufferUpdate()
    {
        earliestStartingTime = currentTime + repairTimeBuffer;
    }


    /**
     * Deletes all temporal relations.
     */
    private void clearTemporalRelationsList()
    {

        // Clearing list of temporal relations

        temporalRelationsList.clear();

        // Clearing references from all task instances

        Iterator it;

        it = preconditionTaskSet.iterator();
        while ( it.hasNext() )
        {
            ( (TCPreconditionTaskInstance)it.next() ).temporalRelations.clear();
        }

        it = actionTaskSet.iterator();
        while ( it.hasNext() )
        {
            ( (TCActionTaskInstance)it.next() ).temporalRelations.clear();
        }

        it = stateTaskSet.iterator();
        while ( it.hasNext() )
        {
            ( (TCStateTaskInstance)it.next() ).temporalRelations.clear();
        }

    }


    /**
     * Adds a temporal relation instance to the internal structures and
     * updates costs.
     * @param temporalRelation  The temporal relation blueprint.
     * @param task1  The task instance to be used for the first task slot.
     * @param task2  The task instance to be used for the second task slot.
     */
    private void addTemporalRelation( TCTemporalRelation temporalRelation , Object task1 , Object task2 )
    {

        // Creating temporal relation object

        TCTemporalRelationInstance temporalRelationInstance = new TCTemporalRelationInstance( task1 , temporalRelation.variableID1 , temporalRelation.comparator , task2 , temporalRelation.variableID2 , temporalRelation.constant );

        // Register temporal relation in tasks and retrieving maximal
        // externalTaskMoveCounts.

        int externalTaskMoveCount1;

        if ( task1 instanceof TCPreconditionTaskInstance )
        {
            ( (TCPreconditionTaskInstance)task1 ).temporalRelations.add( temporalRelationInstance );
            externalTaskMoveCount1 = ( (TCPreconditionTaskInstance)task1 ).externalTaskMoveCount;
        }
        else
        {
            if ( task1 instanceof TCActionTaskInstance )
            {
                ( (TCActionTaskInstance)task1 ).temporalRelations.add( temporalRelationInstance );
                externalTaskMoveCount1 = ( (TCActionTaskInstance)task1 ).externalTaskMoveCount;
            }
            else
            {
                ( (TCStateTaskInstance)task1 ).temporalRelations.add( temporalRelationInstance );
                externalTaskMoveCount1 = ( (TCStateTaskInstance)task1 ).externalTaskMoveCount;
            }
        }

        if ( task2 instanceof TCPreconditionTaskInstance )
        {
            ( (TCPreconditionTaskInstance)task2 ).temporalRelations.add( temporalRelationInstance );
            if ( externalTaskMoveCount1
                > ( (TCPreconditionTaskInstance)task2 ).externalTaskMoveCount)
            {
                temporalRelationInstance.externalTaskMoveCount
                    = externalTaskMoveCount1;
            }
            else
            {
                temporalRelationInstance.externalTaskMoveCount
                    = ( (TCPreconditionTaskInstance)task2 ).externalTaskMoveCount;
            }
        }
        else
        {
            if ( task2 instanceof TCActionTaskInstance )
            {
                ( (TCActionTaskInstance)task2 ).temporalRelations.add( temporalRelationInstance );
                if ( externalTaskMoveCount1 > ( (TCActionTaskInstance)task2 ).externalTaskMoveCount )
                {
                    temporalRelationInstance.externalTaskMoveCount
                        = externalTaskMoveCount1;
                }
                else
                {
                    temporalRelationInstance.externalTaskMoveCount = ( (TCActionTaskInstance)task2 ).externalTaskMoveCount;
                }
            }
            else
            {
                ( (TCStateTaskInstance)task2 ).temporalRelations.add( temporalRelationInstance );
                if ( externalTaskMoveCount1 > ( (TCStateTaskInstance)task2 ).externalTaskMoveCount )
                {
                    temporalRelationInstance.externalTaskMoveCount
                        = externalTaskMoveCount1;
                }
                else
                {
                    temporalRelationInstance.externalTaskMoveCount = ( (TCStateTaskInstance)task2 ).externalTaskMoveCount;
                }
            }
        }

        // Adding costs

        long costChange = temporalRelationInstance.calculateCosts();
        temporalRelationInstance.temporalRelationLE = temporalRelationsList.addListElement( (Object)temporalRelationInstance );

        costs[ temporalRelationCostID ] += costChange;

    }


    /**
     * Synthesizes the temporal relations of the current action type.
     */
    private void synthesizeTemporalRelations()
    {

        Hashtable taskMapping = new Hashtable();

        // Preparing task mapping

        Iterator it;
        Object taskInstance;

        it = preconditionTaskSet.iterator();
        while ( it.hasNext() )
        {
            taskInstance = it.next();
            taskMapping.put( ( (TCPreconditionTaskInstance)taskInstance ).found , taskInstance );
        }

        it = actionTaskSet.iterator();
        while ( it.hasNext() )
        {
            taskInstance = it.next();
            taskMapping.put( ( (TCActionTaskInstance)taskInstance ).found , taskInstance );
        }

        it = stateTaskSet.iterator();
        while ( it.hasNext() )
        {
            taskInstance = it.next();
            taskMapping.put( ( (TCStateTaskInstance)taskInstance ).found , taskInstance );
        }

        // Synthesizing temporal relations

        it = actionType[ actionTypeID ].temporalRelations.iterator();
        TCTemporalRelation temporalRelation;

        while ( it.hasNext() )
        {
            temporalRelation = (TCTemporalRelation)it.next();
            addTemporalRelation(temporalRelation ,
                taskMapping.get( temporalRelation.task1 ) ,
                taskMapping.get( temporalRelation.task2 ) );
        }

    }


    /**
     * Updates all related temporal relations of a precondition or state task
     * in case the task's temporal reference has been changed.
     * @param task  The task that was modified.
     * @param externalTaskMoveCount  The task's externalTaskMoveCount.
     */
    private void propagatePSTaskUpdate( Object task , int externalTaskMoveCount )
    {

        long costChange = 0;

        // Updating affected temporal relations

        TCTemporalRelationInstance temporalRelation;
        Iterator it;
        if ( task instanceof TCPreconditionTaskInstance )
        {
            it = ( (TCPreconditionTaskInstance)task ).temporalRelations.iterator();
        }
        else
        {
            it = ( (TCStateTaskInstance)task ).temporalRelations.iterator();
        }

        while ( it.hasNext() )
        {
            temporalRelation = (TCTemporalRelationInstance)it.next();
            if ( !ownHeuristicChange )
            {
                temporalRelation.externalTaskMoveCount
                    = externalTaskMoveCount;
            }
            costChange += temporalRelation.calculateCosts();
            temporalRelationsList.updateListElementPosition( temporalRelation.temporalRelationLE );
        }

        // Updating costs

        if ( costChange != 0 )
        {
            costs[ temporalRelationCostID ] += costChange;
            if ( costFunctionReference[ temporalRelationCostID ] != null )
            {
                costFunctionReference[ temporalRelationCostID ].updateGlobalCosts();
            }
        }

    }


    /**
     * Updates all related temporal relations of an action task in case the
     * task's begin/duration has been changed.
     * @param task  The action task that was modified.
     * @param variableID  Indicates if begin or duration has been changed.
     */
    private void propagateActionTaskUpdate( TCActionTaskInstance task , int variableID )
    {

        long costChange = 0;

        // Updating affected temporal relations

        TCTemporalRelationInstance temporalRelation;
        Iterator it = task.temporalRelations.iterator();

        while ( it.hasNext() )
        {
            temporalRelation = (TCTemporalRelationInstance)it.next();
            if ( ( (temporalRelation.task1 == task )
                && ( ( variableID == BEGIN )
                || ( temporalRelation.variableID1 == END ) ) )
                || ( ( temporalRelation.task2 == task )
                && ( ( variableID == BEGIN )
                || ( temporalRelation.variableID2 == END ) ) ) )
            {

                if ( !ownHeuristicChange )
                {
                    temporalRelation.externalTaskMoveCount
                        = externalTaskMoveCounter;
                }
                costChange += temporalRelation.calculateCosts();
                temporalRelationsList.updateListElementPosition( temporalRelation.temporalRelationLE );

            }
        }

        // Updating costs

        if ( costChange != 0 )
        {
            costs[ temporalRelationCostID ] += costChange;
            if ( costFunctionReference[ temporalRelationCostID ] != null )
            {
                costFunctionReference[ temporalRelationCostID ].updateGlobalCosts();
            }
        }

    }


    /**
     * Sets the externalTaskMoveCounter to the lowest possible value and
     * updates related tasks' and temporal relations' values.
     */
    private void resetExternalTaskMoveCounter()
    {

        externalTaskMoveCounter = -1;
        int oldCount = -1;

        OrderedShiftDLLElement currentLE = temporalRelationsList.last;
        TCTemporalRelationInstance temporalRelation;

        while ( currentLE != null )
        {
            temporalRelation = (TCTemporalRelationInstance)currentLE.element;

            if ( temporalRelation.externalTaskMoveCount != oldCount )
            {
                externalTaskMoveCounter++ ;
                oldCount = temporalRelation.externalTaskMoveCount;
            }

            // Skipping the update of the reference task

            if ( oldCount != Integer.MAX_VALUE )
            {

                // Update temporal relation

                temporalRelation.externalTaskMoveCount
                    = externalTaskMoveCounter;

                // Update temporal relation's tasks

                if ( temporalRelation.task1 instanceof TCPreconditionTaskInstance )
                {
                    if ( ( (TCPreconditionTaskInstance)temporalRelation.task1 ).externalTaskMoveCount == oldCount )
                    {
                        ( (TCPreconditionTaskInstance)temporalRelation.task1 ).externalTaskMoveCount = externalTaskMoveCounter;
                    }
                }
                else
                {
                    if ( temporalRelation.task1 instanceof TCActionTaskInstance )
                    {
                        if ( ( (TCActionTaskInstance)temporalRelation.task1 ).externalTaskMoveCount == oldCount )
                        {
                            ( (TCActionTaskInstance)temporalRelation.task1 ).externalTaskMoveCount = externalTaskMoveCounter;
                        }
                    }
                    else
                    {
                        if ( ( (TCStateTaskInstance)temporalRelation.task1 ).externalTaskMoveCount == oldCount )
                        {
                            ( (TCStateTaskInstance)temporalRelation.task1 ).externalTaskMoveCount = externalTaskMoveCounter;
                        }
                    }
                }

                if ( temporalRelation.task2 instanceof TCPreconditionTaskInstance )
                {
                    if ( ( (TCPreconditionTaskInstance)temporalRelation.task2 ).externalTaskMoveCount == oldCount )
                    {
                        ( (TCPreconditionTaskInstance)temporalRelation.task2 ).externalTaskMoveCount = externalTaskMoveCounter;
                    }
                }
                else
                {
                    if ( temporalRelation.task2 instanceof TCActionTaskInstance )
                    {
                        if ( ( (TCActionTaskInstance)temporalRelation.task2 ).externalTaskMoveCount == oldCount )
                        {
                            ( (TCActionTaskInstance)temporalRelation.task2 ).externalTaskMoveCount = externalTaskMoveCounter;
                        }
                    }
                    else
                    {
                        if ( ( (TCStateTaskInstance)temporalRelation.task2).externalTaskMoveCount == oldCount )
                        {
                            ( (TCStateTaskInstance)temporalRelation.task2 ).externalTaskMoveCount = externalTaskMoveCounter;
                        }
                    }
                }
            }

            currentLE = currentLE.previous;
        }
    }


    /**
     * Evaluates/updates an action task instance's costs related to the
     * task's execution state. Returns the cost change.
     * @param task  The action task to be analyzed.
     */
    private long evaluateExecutionState( TCActionTaskInstance task )
    {

        long oldCosts = task.executionStateCosts;

        if ( task.executionState == DiscreteResourceConstraint.PENDING )
        {

            if ( currentTime < task.tempValue[ BEGIN ] )
            {
                task.executionStateCosts = 0;
            }
            else
            {
                task.executionStateCosts
                    = currentTime - task.tempValue[ BEGIN ] + 1;
            }

        }
        else
        {
            if ( task.executionState == DiscreteResourceConstraint.DONE )
            {

                if ( task.tempValue[ END ] <= currentTime )
                {
                    task.executionStateCosts = 0;
                }
                else
                {
                    task.executionStateCosts
                        = task.tempValue[ END ] - currentTime;
                }

            }
            else
            {

                if ( task.tempValue[ BEGIN ] <= currentTime )
                {
                    if ( currentTime < task.tempValue[ END ] )
                    {
                        task.executionStateCosts = 0;
                    }
                    else
                    {
                        task.executionStateCosts
                            = currentTime - task.tempValue[ END ] + 1;
                    }
                }
                else
                {
                    task.executionStateCosts
                        = task.tempValue[ BEGIN ] - currentTime;
                }

            }
        }

        return task.executionStateCosts - oldCosts;

    }


    /***** EXTERNAL DYNAMICS *****/


    /**
     * Adds a precondition task to internal structures.
     * @param taskConstraint  The precondition task that was connected.
     * @return  The newly created internal precondition task instance.
     */
    private TCPreconditionTaskInstance addPreconditionTask( PreconditionTaskObjectConstraint taskConstraint )
    {

        // Checking whether task has already been connected

        TCPreconditionTaskInstance newTask = (TCPreconditionTaskInstance)ptConstraintsHT.get( taskConstraint );
        if ( newTask != null )
        {
            throw new Error( "The PreconditionTaskObjectConstraint \""
                + taskConstraint + "\" has already been connected" );
        }

        // Creating a TCPreconditionTaskInstance for the
        // PreconditionTaskObjectConstraint

        newTask = new TCPreconditionTaskInstance();
        newTask.constraint = taskConstraint;
        ptConstraintsHT.put( taskConstraint , newTask );
        preconditionTaskNumber++ ;
        if ( !ownHeuristicChange )
        {
            newTask.externallyConnected = true;
            newTask.externalTaskMoveCount = externalTaskMoveCounter;
            externalPreconditionTaskNumber++ ;
        }
        else
        {
            newTask.externallyConnected = false;
        }

        // Fetching precondition task's variables

        Variable tcTempVariable;
        Variable tcResourceTypeVariable;
        Variable tcStateVariable;

        synchronized( taskConstraint.connections )
        {
            tcTempVariable = taskConstraint.connections.getSpecificVariable( Registry.TEMPORALREFERENCE_EDGE );
            tcResourceTypeVariable = taskConstraint.connections.getSpecificVariable( Registry.RESOURCETYPE_EDGE );
            tcStateVariable = taskConstraint.connections.getSpecificVariable( Registry.STATE_EDGE );
        }
        if ( tcTempVariable == null )
        {
            throw new Error( "The connected PreconditionTaskObjectConstraint \"" + taskConstraint + "\" has no temporal reference variable" );
        }
        if ( tcResourceTypeVariable == null )
        {
            throw new Error( "The connected PreconditionTaskObjectConstraint \"" + taskConstraint + "\" has no resource type variable" );
        }
        if ( tcStateVariable == null )
        {
            throw new Error( "The connected PreconditionTaskObjectConstraint \"" + taskConstraint + "\" has no state variable" );
        }

        // Processing precondition task's temporal reference variable

        modificationModus = true;
        gsc.addIntegerConstraint( tcTempVariable );
        modificationModus = false;
        newTask.tempVariable = tcTempVariable;
        newTask.tempValue = (long)tcTempVariable.value();

        // Processing precondition task's resource type variable

        modificationModus = true;
        gsc.addIntegerConstraint( tcResourceTypeVariable );
        modificationModus = false;
        newTask.resourceTypeVariable = tcResourceTypeVariable;
        newTask.resourceType = (long)tcResourceTypeVariable.value();

        // Processing precondition task's state variable

        modificationModus = true;
        gsc.addIntegerConstraint( tcStateVariable );
        modificationModus = false;
        newTask.stateVariable = tcStateVariable;
        newTask.state = (long)tcStateVariable.value();

        // If task was externally connected, marking actions that do not
        // include a task with similar properties as undesired

        if ( !ownHeuristicChange )
        {
            Iterator it;
            TCPreconditionTask actionTypeTask;
            boolean found;

            for ( int i = 0 ; i < currentActionTypeNumber ; i++ )
            {
                if ( actionTypeOK[ i ] )
                {
                    found = false;
                    it = actionType[ i ].preconditionTasks.iterator();
                    while ( !found && it.hasNext() )
                    {
                        actionTypeTask = (TCPreconditionTask)it.next();
                        found = ( ( actionTypeTask.resourceType == newTask.resourceType ) && ( actionTypeTask.state == newTask.state ) );
                    }
                    if ( !found )
                    {
                        actionTypeOK[ i ] = false;
                    }
                }
            }
        }

        return newTask;

    }


    /**
     * Adds an action task to internal structures.
     * @param taskConstraint  The action task that was connected.
     * @return  The newly created internal action task instance.
     */
    private TCActionTaskInstance addActionTask( ActionTaskObjectConstraint taskConstraint )
    {

        // Checking whether task has already been connected

        TCActionTaskInstance newTask
            = (TCActionTaskInstance)atConstraintsHT.get( taskConstraint );
        if ( newTask != null )
        {
            throw new Error("The ActionTaskObjectConstraint \""
                + taskConstraint + "\" has already been connected" );
        }

        // Creating a TCActionTaskInstance for the
        // ActionTaskObjectConstraint

        newTask = new TCActionTaskInstance();
        newTask.constraint = taskConstraint;
        atConstraintsHT.put( taskConstraint , newTask );
        actionTaskNumber++ ;
        if ( !ownHeuristicChange )
        {
            newTask.externallyConnected = true;
            newTask.externalTaskMoveCount = externalTaskMoveCounter;
            externalActionTaskNumber++ ;
        }
        else
        {
            newTask.externallyConnected = false;
        }

        // Fetching action task's variables

        Variable tcTempVariableBegin;
        Variable tcTempVariableEnd;
        Variable tcResourceTypeVariable;
        Variable tcOperationVariable;
        Variable tcExecutionStateVariable;
        Variable tcConsumptionVariable;

        synchronized( taskConstraint.connections )
        {
            tcTempVariableBegin = taskConstraint.connections.getSpecificVariable( Registry.BEGIN_EDGE );
            tcTempVariableEnd = taskConstraint.connections.getSpecificVariable( Registry.DURATION_EDGE );
            tcResourceTypeVariable = taskConstraint.connections.getSpecificVariable( Registry.RESOURCETYPE_EDGE );
            tcOperationVariable = taskConstraint.connections.getSpecificVariable( Registry.OPERATION_EDGE );
            tcExecutionStateVariable = taskConstraint.connections.getSpecificVariable( Registry.EXECUTIONSTATE_EDGE );
            tcConsumptionVariable = taskConstraint.connections.getSpecificVariable( Registry.CONSUMPTION_EDGE );
        }
        if ( tcTempVariableBegin == null )
        {
            throw new Error( "The connected ActionTaskObjectConstraint \"" + taskConstraint + "\" has no begin variable" );
        }
        if ( tcTempVariableEnd == null )
        {
            throw new Error( "The connected ActionTaskObjectConstraint \"" + taskConstraint + "\" has no duration variable" );
        }
        if ( tcResourceTypeVariable == null )
        {
            throw new Error( "The connected ActionTaskObjectConstraint \"" + taskConstraint + "\" has no resource type variable" );
        }
        if ( tcOperationVariable == null )
        {
            throw new Error( "The connected ActionTaskObjectConstraint \"" + taskConstraint + "\" has no operation variable" );
        }
        if ( tcExecutionStateVariable == null )
        {
            throw new Error( "The connected ActionTaskObjectConstraint \"" + taskConstraint + "\" has no execution state variable" );
        }
        if ( tcConsumptionVariable == null )
        {
            throw new Error( "The connected ActionTaskObjectConstraint \"" + taskConstraint + "\" has no consumption variable" );
        }

        // Processing action task's begin variable

        modificationModus = true;
        gsc.addIntegerConstraint( tcTempVariableBegin );
        modificationModus = false;
        newTask.tempVariable[ BEGIN ] = tcTempVariableBegin;
        newTask.tempValue[ BEGIN ] = (long)tcTempVariableBegin.value();

        // Processing action task's duration variable

        modificationModus = true;
        gsc.addIntegerConstraint( tcTempVariableEnd );
        modificationModus = false;
        newTask.tempVariable[ DURATION ] = tcTempVariableEnd;
        newTask.tempValue[ DURATION ] = (long)tcTempVariableEnd.value();

        // Setting internal TCActionTaskInstance task's end variable

        newTask.tempValue[ END ] = newTask.tempValue[ BEGIN ]
            + newTask.tempValue[ DURATION ];

        // Processing action task's resource type variable

        modificationModus = true;
        gsc.addIntegerConstraint( tcResourceTypeVariable );
        modificationModus = false;
        newTask.resourceTypeVariable = tcResourceTypeVariable;
        newTask.resourceType = (long)tcResourceTypeVariable.value();

        // Processing action task's operation variable

        modificationModus = true;
        gsc.addIntegerConstraint( tcOperationVariable );
        modificationModus = false;
        newTask.operationVariable = tcOperationVariable;
        newTask.operation = (long)tcOperationVariable.value();

        // Processing action task's execution state variable

        modificationModus = true;
        gsc.addIntegerConstraint( tcExecutionStateVariable );
        modificationModus = false;
        newTask.executionStateVariable = tcExecutionStateVariable;
        newTask.executionState = (int)tcExecutionStateVariable.value();

        // Checking for a valid execution state value

        if ( ( newTask.executionState != DiscreteResourceConstraint.PENDING ) && ( newTask.executionState != DiscreteResourceConstraint.IN_EXECUTION ) && ( newTask.executionState != DiscreteResourceConstraint.DONE ) )
        {
            throw new Error( "Invalid value \"" + newTask.executionState + "\" for task's execution state" );
        }

        // Processing action task's consumption variable

        newTask.consumptionVariable = tcConsumptionVariable;
        newTask.consumption = tcConsumptionVariable.value();

        // If task was externally connected, marking actions that do not
        // include a task with similar properties as undesired

        if ( !ownHeuristicChange )
        {
            Iterator it;
            TCActionTask actionTypeTask;
            boolean found;

            for ( int i = 0 ; i < currentActionTypeNumber ; i++ )
            {
                if ( actionTypeOK[ i ] )
                {
                    found = false;
                    it = actionType[ i ].actionTasks.iterator();
                    while ( !found && it.hasNext() )
                    {
                        actionTypeTask = (TCActionTask)it.next();
                        found = ( ( actionTypeTask.resourceType == newTask.resourceType ) && ( actionTypeTask.operation == newTask.operation ) && ( actionTypeTask.duration == newTask.tempValue[ DURATION ] ) && ( actionTypeTask.consumption == newTask.consumption ) );
                    }
                    if ( !found )
                    {
                        actionTypeOK[ i ] = false;
                    }
                }
            }
        }

        return newTask;

    }


    /**
     * Adds a state task to internal structures.
     * @param taskConstraint  The state task that was connected.
     * @return  The newly created internal state task instance.
     */
    private TCStateTaskInstance addStateTask( StateTaskObjectConstraint taskConstraint )
    {

        // Checking whether task has already been connected

        TCStateTaskInstance newTask
            = (TCStateTaskInstance)stConstraintsHT.get(taskConstraint);
        if ( newTask != null )
        {
            throw new Error( "The StateTaskObjectConstraint \""
                + taskConstraint + "\" has already been connected" );
        }

        // Creating a TCStateTaskInstance for the StateTaskObjectConstraint

        newTask = new TCStateTaskInstance();
        newTask.constraint = taskConstraint;
        stConstraintsHT.put( taskConstraint , newTask );
        stateTaskNumber++ ;
        if ( !ownHeuristicChange )
        {
            newTask.externallyConnected = true;
            newTask.externalTaskMoveCount = externalTaskMoveCounter;
            externalStateTaskNumber++ ;
        }
        else
        {
            newTask.externallyConnected = false;
        }

        // Fetching state task's variables

        Variable tcTempVariable;
        Variable tcResourceTypeVariable;
        Variable tcContributionVariable;

        synchronized( taskConstraint.connections )
        {
            tcTempVariable = taskConstraint.connections.getSpecificVariable( Registry.TEMPORALREFERENCE_EDGE );
            tcResourceTypeVariable = taskConstraint.connections.getSpecificVariable( Registry.RESOURCETYPE_EDGE );
            tcContributionVariable = taskConstraint.connections.getSpecificVariable( Registry.CONTRIBUTION_EDGE );
        }
        if ( tcTempVariable == null )
        {
            throw new Error( "The connected StateTaskObjectConstraint \"" + taskConstraint + "\" has no temporal reference variable" );
        }
        if ( tcResourceTypeVariable == null )
        {
            throw new Error( "The connected StateTaskObjectConstraint \"" + taskConstraint + "\" has no resource type variable" );
        }
        if ( tcContributionVariable == null )
        {
            throw new Error( "The connected StateTaskObjectConstraint \"" + taskConstraint + "\" has no state variable" );
        }

        // Processing state task's temporal reference variable

        modificationModus = true;
        gsc.addIntegerConstraint( tcTempVariable );
        modificationModus = false;
        newTask.tempVariable = tcTempVariable;
        newTask.tempValue = (long)tcTempVariable.value();

        // Processing state task's resource type variable

        modificationModus = true;
        gsc.addIntegerConstraint( tcResourceTypeVariable );
        modificationModus = false;
        newTask.resourceTypeVariable = tcResourceTypeVariable;
        newTask.resourceType = (long)tcResourceTypeVariable.value();

        // Processing state task's contribution variable

        modificationModus = true;
        gsc.addIntegerConstraint( tcContributionVariable );
        modificationModus = false;
        newTask.contributionVariable = tcContributionVariable;
        newTask.contribution = (long)tcContributionVariable.value();

        // If task was externally connected, marking actions that do not
        // include a task with similar properties as undesired

        if ( !ownHeuristicChange )
        {
            Iterator it;
            TCStateTask actionTypeTask;
            boolean found;

            for ( int i = 0 ; i < currentActionTypeNumber ; i++ )
            {
                if ( actionTypeOK[ i ] )
                {
                    found = false;
                    it = actionType[ i ].stateTasks.iterator();
                    while ( !found && it.hasNext() )
                    {
                        actionTypeTask = (TCStateTask)it.next();
                        found = ( ( actionTypeTask.resourceType == newTask.resourceType ) && ( actionTypeTask.contribution == newTask.contribution ) );
                    }
                    if ( !found )
                    {
                        actionTypeOK[ i ] = false;
                    }
                }
            }
        }

        return newTask;

    }


    /**
     * Excludes action types with a similar number of specific precondition
     * tasks from further consideration.
     * @param task  The specific task to be compared with.
     */
    private void excludePreconditionTaskInstance( TCPreconditionTaskInstance task )
    {

        // Counting how often similar instances are present

        long counter = 0;

        Iterator itInstance;
        TCPreconditionTaskInstance preconditionTaskInstance;

        itInstance = ptConstraintsHT.values().iterator();
        while ( itInstance.hasNext() )
        {
            preconditionTaskInstance
                = (TCPreconditionTaskInstance)itInstance.next();
            if ( ( preconditionTaskInstance.resourceType == task.resourceType )
                && ( preconditionTaskInstance.state == task.state ) )
            {
                counter++ ;
            }
        }

        // Marking action types with an equal precondition task
        // configuration as undesired

        long typeCounter;
        Iterator it;
        TCPreconditionTask actionTypeTask;

        for ( int i = 0 ; i < currentActionTypeNumber ; i++ )
        {
            if ( actionTypeOK[ i ] )
            {
                typeCounter = counter;
                it = actionType[ i ].preconditionTasks.iterator();
                while ( it.hasNext() )
                {
                    actionTypeTask = (TCPreconditionTask)it.next();
                    if ( ( actionTypeTask.resourceType == task.resourceType )
                        && ( actionTypeTask.state == task.state ) )
                    {
                        typeCounter-- ;
                    }
                }
                if ( typeCounter == 0 )
                {
                    actionTypeOK[ i ] = false;
                }
            }
        }

        desiredActionTypeID = -1;

    }


    /**
     * Excludes action types with a similar number of specific action
     * tasks from further consideration.
     * @param task  The specific task to be compared with.
     */
    private void excludeActionTaskInstance( TCActionTaskInstance task )
    {

        long counter = 0;

        Iterator itInstance;
        TCActionTaskInstance actionTaskInstance;

        itInstance = atConstraintsHT.values().iterator();
        while ( itInstance.hasNext() )
        {
            actionTaskInstance = (TCActionTaskInstance)itInstance.next();
            if ( ( actionTaskInstance.resourceType == task.resourceType ) && ( actionTaskInstance.operation == task.operation ) && ( actionTaskInstance.tempValue[ BEGIN ] == task.tempValue[ BEGIN ] ) && ( actionTaskInstance.consumption == task.consumption ) )
            {
                counter++ ;
            }
        }

        // Marking action types with an equal action task configuration
        // as undesired

        long typeCounter;
        Iterator it;
        TCActionTask actionTypeTask;

        for ( int i = 0 ; i < currentActionTypeNumber ; i++ )
        {
            if ( actionTypeOK[ i ] )
            {
                typeCounter = counter;
                it = actionType[ i ].actionTasks.iterator();
                while ( it.hasNext() )
                {
                    actionTypeTask = (TCActionTask)it.next();
                    if ( ( actionTypeTask.resourceType == task.resourceType ) && ( actionTypeTask.operation == task.operation ) && ( actionTypeTask.duration == task.tempValue[ DURATION ] ) && ( actionTypeTask.consumption == task.consumption ) )
                    {
                        typeCounter-- ;
                    }
                }
                if ( typeCounter == 0 )
                {
                    actionTypeOK[ i ] = false;
                }
            }
        }

        desiredActionTypeID = -1;

    }


    /**
     * Excludes action types with a similar number of specific state
     * tasks from further consideration.
     * @param task  The specific task to be compared with.
     */
    private void excludeStateTaskInstance( TCStateTaskInstance task )
    {

        long counter = 0;

        Iterator itInstance;
        TCStateTaskInstance stateTaskInstance;

        itInstance = stateTaskSet.iterator();
        while ( itInstance.hasNext() )
        {
            stateTaskInstance = (TCStateTaskInstance)itInstance.next();
            if ( ( stateTaskInstance.resourceType == task.resourceType )
                && ( stateTaskInstance.contribution == task.contribution ) )
            {
                counter++ ;
            }
        }

        // Marking action types with an equal state task configuration
        // as undesired

        long typeCounter;
        Iterator it;
        TCStateTask actionTypeTask;

        for ( int i = 0 ; i < currentActionTypeNumber ; i++ )
        {
            if ( actionTypeOK[ i ] )
            {
                typeCounter = counter;
                it = actionType[ i ].stateTasks.iterator();
                while ( it.hasNext() )
                {
                    actionTypeTask = (TCStateTask)it.next();
                    if ( ( actionTypeTask.resourceType == task.resourceType ) && ( actionTypeTask.contribution == task.contribution ) )
                    {
                        typeCounter-- ;
                    }
                }
                if ( typeCounter == 0 )
                {
                    actionTypeOK[ i ] = false;
                }
            }
        }

        desiredActionTypeID = -1;

    }


    public void update( double oldValue , ConnectionReference first , ConnectionReference last )
    {

        if ( !modificationModus
            && ( ( (Variable)first.node ).value() != oldValue ) )
        {

            if ( last.node instanceof CurrentTimeObjectConstraint )
            {

                // Current time variable was updated

                currentTime = (long)( (Variable)first.node ).value();
                earliestStartingTime = currentTime + repairTimeBuffer;

                // Re-evaluation execution states of action tasks

                if ( costs[ CONFIGURATION_INCONSISTENCY ] == 0 )
                {

                    long costChange = 0;

                    Iterator it = actionTaskSet.iterator();
                    while ( it.hasNext() )
                    {
                        costChange += evaluateExecutionState( (TCActionTaskInstance)it.next() );
                    }

                    // Propagation cost changes

                    if ( costChange != 0 )
                    {
                        executionStateCosts += costChange;
                        costs[ temporalRelationCostID ] += costChange;
                        if ( costFunctionReference[ temporalRelationCostID ] != null )
                        {
                            costFunctionReference[ temporalRelationCostID ].updateGlobalCosts();
                        }
                    }

                }

            }
            else
            {

                if ( first.edgeID == Registry.ACTIONTYPE_EDGE )
                {

                    // Action type variable was updated

                    if ( !ownHeuristicChange )
                    {

                        // Marking previous action type as undesired

                        if ( ( actionTypeID >= 0 )
                            && ( actionTypeID < currentActionTypeNumber ) )
                        {
                            actionTypeOK[ actionTypeID ] = false;
                        }

                    }

                    // Updating internal value copy

                    actionTypeID = (int)actionTypeVar.value();

                    // Updating structural consistency

                    if ( !ownHeuristicChange )
                    {
                        actionTypeGiven = ( ( actionTypeID >= 0 )  && ( actionTypeID < currentActionTypeNumber ) );
                        desiredActionTypeID = -1;
                    }
                    reactivatePossibleActionTypes();
                    checkForValidActionType();

                }
                else
                {

                    if ( last.node instanceof ActionObjectConstraint )
                    {

                        last = last.previous;

                        if ( last.node instanceof
                            PreconditionTaskObjectConstraint )
                        {

                            PreconditionTaskObjectConstraint taskConstraint
                                = (PreconditionTaskObjectConstraint)last.node;

                            // Fetching related TCPreconditionTaskInstance
                            // for the PreconditionTaskObjectConstraint

                            TCPreconditionTaskInstance task = (TCPreconditionTaskInstance)ptConstraintsHT.get( taskConstraint );

                            if ( first.edgeID ==
                                Registry.TEMPORALREFERENCE_EDGE )
                            {

                                // Updating internal structures & costs

                                task.tempValue
                                    = (long)task.tempVariable.value();

                                if ( ( costs[ CONFIGURATION_INCONSISTENCY ] == 0 ) && ( task != referenceTask ) )
                                {
                                    if ( !ownHeuristicChange )
                                    {
                                        if ( externalTaskMoveCounter
                                            == Integer.MAX_VALUE - 1 )
                                        {
                                            resetExternalTaskMoveCounter();
                                        }
                                        task.externalTaskMoveCount
                                            = ++externalTaskMoveCounter;
                                    }
                                    propagatePSTaskUpdate( task , externalTaskMoveCounter );
                                }

                            }
                            else
                            {

                                if ( first.edgeID == Registry.RESOURCETYPE_EDGE )
                                {

                                    if ( !ownHeuristicChange )
                                    {
                                        excludePreconditionTaskInstance( task );
                                    }

                                    // Updating internal value copy

                                    task.resourceType = (long)task.resourceTypeVariable.value();

                                    // Updating structural consistency

                                    reactivatePossibleActionTypes();
                                    checkForValidActionType();

                                }
                                else
                                {

                                    if ( first.edgeID == Registry.STATE_EDGE )
                                    {

                                        if ( !ownHeuristicChange )
                                        {
                                            excludePreconditionTaskInstance( task );
                                        }

                                        // Updating internal value copy

                                        task.state
                                            = (long)task.stateVariable.value();

                                        // Updating structural consistency

                                        reactivatePossibleActionTypes();
                                        checkForValidActionType();

                                    }

                                }

                            }

                        }
                        else
                        {

                            if ( last.node instanceof
                                ActionTaskObjectConstraint )
                            {

                                ActionTaskObjectConstraint taskConstraint
                                    = (ActionTaskObjectConstraint)last.node;

                                // Fetching related TCActionTaskInstance for
                                // the ActionTaskObjectConstraint

                                TCActionTaskInstance task = (TCActionTaskInstance)atConstraintsHT.get( taskConstraint );

                                if ( first.edgeID == Registry.BEGIN_EDGE )
                                {

                                    // Updating internal structures & costs

                                    task.tempValue[ BEGIN ] = (long)task.tempVariable[ BEGIN ].value();
                                    task.tempValue[ END ] = task.tempValue[ BEGIN ] + task.tempValue[ DURATION ];

                                    if ( costs[ CONFIGURATION_INCONSISTENCY ] == 0 )
                                    {

                                        if ( !ownHeuristicChange )
                                        {
                                            if ( externalTaskMoveCounter
                                                == Integer.MAX_VALUE - 1 )
                                            {
                                                resetExternalTaskMoveCounter();
                                            }
                                            task.externalTaskMoveCount
                                                = ++externalTaskMoveCounter;
                                        }

                                        // Updating temporal relation costs

                                        propagateActionTaskUpdate( task , BEGIN );

                                        // Updating execution state's costs

                                        long costChange
                                            = evaluateExecutionState( task );
                                        if ( costChange != 0 )
                                        {
                                            executionStateCosts += costChange;
                                            costs[temporalRelationCostID]
                                                += costChange;
                                            if ( costFunctionReference[ temporalRelationCostID ] != null )
                                            {
                                                costFunctionReference[ temporalRelationCostID ].updateGlobalCosts();
                                            }
                                        }

                                    }


                                }
                                else
                                {

                                    if ( ( first.edgeID == Registry.DURATION_EDGE ) || ( first.edgeID == Registry.RESOURCETYPE_EDGE ) || ( first.edgeID == Registry.OPERATION_EDGE ) || ( first.edgeID == Registry.CONSUMPTION_EDGE ) )
                                    {

                                        if ( !ownHeuristicChange )
                                        {
                                            excludeActionTaskInstance( task );
                                        }

                                        // Updating internal value copy

                                        if ( first.edgeID == Registry.DURATION_EDGE )
                                        {

                                            task.tempValue[ DURATION ] = (long)task.tempVariable[ DURATION ].value();
                                            task.tempValue[ END ]
                                                = task.tempValue[ BEGIN ]
                                                + task.tempValue[ DURATION ];

                                            // Temporal relations are not
                                            // updated because action type
                                            // will be changed

                                        }
                                        else
                                        {
                                            if ( first.edgeID == Registry.RESOURCETYPE_EDGE )
                                            {
                                                task.resourceType = (long)task.resourceTypeVariable.value();
                                            }
                                            else
                                            {
                                                if ( first.edgeID == Registry.OPERATION_EDGE )
                                                {
                                                    task.operation = (long)task.operationVariable.value();
                                                } else {
                                                    task.consumption = (long)task.consumptionVariable.value();
                                                }
                                            }
                                        }

                                        // Updating structural consistency

                                        reactivatePossibleActionTypes();
                                        checkForValidActionType();

                                    }
                                    else
                                    {

                                        if ( first.edgeID
                                            == Registry.EXECUTIONSTATE_EDGE )
                                        {

                                            task.executionState = (int)task.executionStateVariable.value();

                                            // Checking for a valid execution
                                            // state value

                                            if ( ( task.executionState != DiscreteResourceConstraint.PENDING ) && ( task.executionState != DiscreteResourceConstraint.IN_EXECUTION)  && ( task.executionState != DiscreteResourceConstraint.DONE ) )
                                            {
                                                throw new Error( "Invalid value \"" + task.executionState + "\" for task's execution state" );
                                            }

                                            // Updating execution state's costs

                                            if ( costs[ CONFIGURATION_INCONSISTENCY ] == 0 )
                                            {
                                                long costChange = evaluateExecutionState( task );
                                                if ( costChange != 0 )
                                                {
                                                    executionStateCosts
                                                        += costChange;
                                                    costs[ temporalRelationCostID ] += costChange;
                                                    if ( costFunctionReference[ temporalRelationCostID ] != null )
                                                    {
                                                        costFunctionReference[ temporalRelationCostID ].updateGlobalCosts();
                                                    }
                                                }
                                            }

                                        }

                                    }

                                }

                            } else {

                                if ( last.node instanceof
                                    StateTaskObjectConstraint )
                                {

                                    StateTaskObjectConstraint taskConstraint
                                        = (StateTaskObjectConstraint)last.node;

                                    // Fetching related TCStateTaskInstance
                                    // for the StateTaskObjectConstraint

                                    TCStateTaskInstance task = (TCStateTaskInstance)stConstraintsHT.get( taskConstraint );

                                    if ( first.edgeID == Registry.TEMPORALREFERENCE_EDGE )
                                    {

                                        // Updating internal structures & costs

                                        task.tempValue
                                            = (long)task.tempVariable.value();

                                        if ( costs[ CONFIGURATION_INCONSISTENCY ] == 0 )
                                        {
                                            if ( !ownHeuristicChange )
                                            {
                                                if ( externalTaskMoveCounter
                                                    == Integer.MAX_VALUE - 1 )
                                                {
                                                    resetExternalTaskMoveCounter();
                                                }
                                                task.externalTaskMoveCount = ++externalTaskMoveCounter;
                                            }
                                            propagatePSTaskUpdate (task , externalTaskMoveCounter );
                                        }

                                    }
                                    else
                                    {

                                        if ( first.edgeID == Registry.RESOURCETYPE_EDGE )
                                        {

                                            if ( !ownHeuristicChange )
                                            {
                                                excludeStateTaskInstance( task );
                                            }

                                            // Updating internal value copy

                                            task.resourceType = (long)task.resourceTypeVariable.value();

                                            // Updating structural consistency

                                            reactivatePossibleActionTypes();
                                            checkForValidActionType();

                                        }
                                        else
                                        {

                                            if ( first.edgeID == Registry.CONTRIBUTION_EDGE )
                                            {

                                                if ( !ownHeuristicChange )
                                                {
                                                    excludeStateTaskInstance( task );
                                                }

                                                // Updating internal value copy

                                                task.contribution = (long)task.contributionVariable.value();

                                                // Updating structural
                                                // consistency

                                                reactivatePossibleActionTypes();
                                                checkForValidActionType();

                                            }

                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }

    }


    public void connectNode( ConnectionReference first , ConnectionReference last )
    {

        if ( last.node instanceof CurrentTimeObjectConstraint )
        {

            // Current time variable/constraint was connected

            Node currentTimeConstraint = last.node;

            // Fetching current time's variable

            Variable variable;
            synchronized( currentTimeConstraint.connections )
            {
                variable = currentTimeConstraint.connections.getSpecificVariable( Registry.PLAIN_EDGE );
            }
            if ( variable == null )
            {
                throw new Error( "The connected CurrentTimeObjectConstraint \"" + currentTimeConstraint + "\" has no variable" );
            }

            currentTime = (long)variable.value();
            earliestStartingTime = currentTime + repairTimeBuffer;

        }
        else
        {

            if ( first.node instanceof ActionObjectConstraint )
            {

                // Action object was connected

                actionConstraint = (ConventionalConstraint)first.node;

                // Fetching action's action type variable

                synchronized( actionConstraint.connections )
                {
                    actionTypeVar = actionConstraint.connections.getSpecificVariable( Registry.ACTIONTYPE_EDGE );
                }
                if ( actionTypeVar == null )
                {
                    throw new Error( "The connected ActionObjectConstraint \"" + actionConstraint + "\" has no action type variable" );
                }

                // Processing action type variable

                modificationModus = true;
                gsc.addIntegerConstraint( actionTypeVar );
                modificationModus = false;
                actionTypeID = (int)actionTypeVar.value();

                actionTypeGiven
                    = ( ( actionTypeID >= 0 )
                     && ( actionTypeID < currentActionTypeNumber ) );
                desiredActionTypeID = -1;

                externalTaskMoveCounter = 1;

                // Fetching and processing all connected precondition tasks

                Iterator it;
                HashSet taskSet = null;
                Object retrievedTaskSet = null;

                synchronized( actionConstraint.connections )
                {
                    retrievedTaskSet = actionConstraint.connections.getSpecificNodes( Registry.PRECONDITIONTASK_OBJECTCONSTRAINT );
                    if ( retrievedTaskSet instanceof HashSet )
                    {
                        taskSet
                            = (HashSet)( (HashSet)retrievedTaskSet ).clone();
                    }

                }

                if ( retrievedTaskSet != null )
                {

                    if ( taskSet == null )
                    {
                        referenceTask = addPreconditionTask( (PreconditionTaskObjectConstraint)( (ConnectedNode)retrievedTaskSet ).node );
                    }
                    else
                    {
                        it = taskSet.iterator();
                        while ( it.hasNext() )
                        {
                            referenceTask = addPreconditionTask( (PreconditionTaskObjectConstraint)( (ConnectedNode)it.next() ).node );
                        }
                    }

                    initiallyConnectedTask = referenceTask;

                }

                // Updating reference task info if the action is a world
                // transition

                if ( actionTypeGiven
                    && ( actionType[ actionTypeID ].transitionID != -1 ) )
                {

                    Object transitionConstraint;

                    synchronized( actionConstraint.connections )
                    {
                        transitionConstraint = actionConstraint.connections.getSpecificNodes( Registry.TRANSITION_CONSTRAINT );
                    }

                    if ( transitionConstraint != null )
                    {
                        if ( referenceTask == null )
                        {
                            throw new Error( "No precondition task for static time reference connected to the action object" );
                        }
                        referenceTaskTime = referenceTask.tempValue;
                        referenceTask.externalTaskMoveCount
                            = Integer.MAX_VALUE;
                    }
                    else
                    {
                        referenceTask = null;
                    }

                }
                else
                {
                    referenceTask = null;
                }

                // Fetching and processing all connected action tasks

                taskSet = null;
                synchronized( actionConstraint.connections )
                {
                    retrievedTaskSet = actionConstraint.connections.getSpecificNodes( Registry.ACTIONTASK_OBJECTCONSTRAINT );
                    if ( retrievedTaskSet instanceof HashSet )
                    {
                        taskSet
                            = (HashSet)( (HashSet) retrievedTaskSet ).clone();
                    }

                }
                if ( retrievedTaskSet != null )
                {
                    if ( taskSet == null )
                    {

                        initiallyConnectedTask = addActionTask( (ActionTaskObjectConstraint)( (ConnectedNode)retrievedTaskSet ).node );
                    }
                    else
                    {
                        it = taskSet.iterator();
                        while ( it.hasNext() )
                        {
                            initiallyConnectedTask = addActionTask( (ActionTaskObjectConstraint)( (ConnectedNode)it.next() ).node );
                        }
                    }
                }

                // Fetching and processing all connected state tasks

                taskSet = null;
                synchronized( actionConstraint.connections )
                {
                    retrievedTaskSet = actionConstraint.connections.getSpecificNodes( Registry.STATETASK_OBJECTCONSTRAINT );
                    if ( retrievedTaskSet instanceof HashSet )
                    {
                        taskSet
                            = (HashSet)( (HashSet)retrievedTaskSet ).clone();
                    }

                }
                if ( retrievedTaskSet != null )
                {
                    if ( taskSet == null )
                    {
                        initiallyConnectedTask = addStateTask( (StateTaskObjectConstraint)( (ConnectedNode)retrievedTaskSet ).node);
                    }
                    else
                    {
                        it = taskSet.iterator();
                        while ( it.hasNext() )
                        {
                            initiallyConnectedTask = addStateTask( (StateTaskObjectConstraint)( (ConnectedNode)it.next() ).node);
                        }
                    }
                }

                // Updating structural consistency

                checkForValidActionType();

            }
            else
            {

                // Increasing external task change count

                if ( !ownHeuristicChange )
                {
                    if ( externalTaskMoveCounter == Integer.MAX_VALUE - 1 )
                    {
                        resetExternalTaskMoveCounter();
                    }
                    externalTaskMoveCounter++ ;
                }

                if ( first.node instanceof PreconditionTaskObjectConstraint )
                {

                    // PreconditionTaskObjectConstraint was connected

                    addPreconditionTask( (PreconditionTaskObjectConstraint)first.node );

                    // Updating structural consistency

                    if ( !ownHeuristicChange )
                    {
                        desiredActionTypeID = -1;
                    }
                    checkForValidActionType();

                }
                else
                {

                    if ( first.node instanceof ActionTaskObjectConstraint )
                    {

                        // ActionTaskObjectConstraint was connected

                        addActionTask( (ActionTaskObjectConstraint)first.node );

                        // Updating structural consistency

                        if ( !ownHeuristicChange )
                        {
                            desiredActionTypeID = -1;
                        }
                        checkForValidActionType();

                    }
                    else
                    {

                        if ( first.node instanceof StateTaskObjectConstraint )
                        {

                            // StateTaskObjectConstraint was connected

                            addStateTask( (StateTaskObjectConstraint)first.node );

                            // Updating structural consistency

                            if ( !ownHeuristicChange )
                            {
                                desiredActionTypeID = -1;
                            }
                            checkForValidActionType();

                        }
                        else
                        {
                            throw new Error( "Invalid connection of \""
                                + first.node + "\" to the TaskConstraint" );
                        }
                    }
                }
            }
        }

    }


    public void disconnectNode( ConnectionReference first , ConnectionReference last )
    {

        if ( first.node instanceof ActionObjectConstraint )
        {

            // Action object was disconnected

            modificationModus = true;

            // Removing integer constraint from action type variable

            gsc.removeIntegerConstraint( actionTypeVar );

            // Removing integer constraints from all connected
            // precondition tasks' variables

            TCPreconditionTaskInstance preconditionTaskInstance;
            Iterator it = preconditionTaskSet.iterator();
            while ( it.hasNext() )
            {
                preconditionTaskInstance
                    = (TCPreconditionTaskInstance)it.next();
                gsc.removeIntegerConstraint( preconditionTaskInstance.tempVariable );
                gsc.removeIntegerConstraint( preconditionTaskInstance.resourceTypeVariable );
                gsc.removeIntegerConstraint( preconditionTaskInstance.stateVariable );
            }

            // Removing integer constraints from all connected
            // action tasks' variables

            TCActionTaskInstance actionTaskInstance;
            it = actionTaskSet.iterator();
            while ( it.hasNext() )
            {
                actionTaskInstance = (TCActionTaskInstance)it.next();
                gsc.removeIntegerConstraint( actionTaskInstance.tempVariable[ BEGIN ] );
                gsc.removeIntegerConstraint( actionTaskInstance.tempVariable[ DURATION ] );
                gsc.removeIntegerConstraint( actionTaskInstance.resourceTypeVariable );
                gsc.removeIntegerConstraint( actionTaskInstance.operationVariable );
                gsc.removeIntegerConstraint( actionTaskInstance.executionStateVariable );
            }

            // Removing integer constraints from all connected
            // state tasks' variables

            TCStateTaskInstance stateTaskInstance;
            it = stateTaskSet.iterator();
            while ( it.hasNext() )
            {
                stateTaskInstance = (TCStateTaskInstance)it.next();
                gsc.removeIntegerConstraint( stateTaskInstance.tempVariable );
                gsc.removeIntegerConstraint( stateTaskInstance.resourceTypeVariable );
                gsc.removeIntegerConstraint( stateTaskInstance.contributionVariable );
            }

            modificationModus = false;

        }
        else
        {

            if ( first.node instanceof PreconditionTaskObjectConstraint )
            {

                // PreconditionTaskObjectConstraint was disconnected

                PreconditionTaskObjectConstraint taskConstraint
                    = (PreconditionTaskObjectConstraint)first.node;

                // Removing related TCPreconditionTaskInstance for the
                // PreconditionTaskObjectConstraint

                TCPreconditionTaskInstance task = (TCPreconditionTaskInstance)ptConstraintsHT.get( taskConstraint );
                if ( !ownHeuristicChange )
                {
                    excludePreconditionTaskInstance( task );
                }
                ptConstraintsHT.remove( taskConstraint );
                preconditionTaskNumber-- ;
                if ( task.externallyConnected )
                {
                    externalPreconditionTaskNumber-- ;
                }

                // Removing IntegerConstraints from the variables

                modificationModus = true;
                gsc.removeIntegerConstraint( task.tempVariable );
                gsc.removeIntegerConstraint( task.resourceTypeVariable );
                gsc.removeIntegerConstraint( task.stateVariable );
                modificationModus = false;

                // Updating structural consistency

                reactivatePossibleActionTypes();
                checkForValidActionType();

            }
            else
            {

                if ( first.node instanceof ActionTaskObjectConstraint )
                {

                    // ActionTaskObjectConstraint was disconnected

                    ActionTaskObjectConstraint taskConstraint
                        = (ActionTaskObjectConstraint)first.node;

                    // Removing related TCActionTaskInstance for the
                    // ActionTaskObjectConstraint

                    TCActionTaskInstance task = (TCActionTaskInstance)atConstraintsHT.get( taskConstraint );
                    if ( !ownHeuristicChange )
                    {
                        excludeActionTaskInstance( task );
                    }
                    atConstraintsHT.remove( taskConstraint );
                    actionTaskNumber-- ;
                    if ( task.externallyConnected )
                    {
                        externalActionTaskNumber-- ;
                    }

                    // Removing IntegerConstraints from the variables

                    modificationModus = true;
                    gsc.removeIntegerConstraint( task.tempVariable[ BEGIN ] );
                    gsc.removeIntegerConstraint( task.tempVariable[ DURATION ] );
                    gsc.removeIntegerConstraint( task.resourceTypeVariable );
                    gsc.removeIntegerConstraint( task.operationVariable );
                    gsc.removeIntegerConstraint( task.executionStateVariable );
                    modificationModus = false;

                    // Updating structural consistency

                    reactivatePossibleActionTypes();
                    checkForValidActionType();

                }
                else
                {

                    if ( first.node instanceof StateTaskObjectConstraint )
                    {

                        // StateTaskObjectConstraint was disconnected

                        StateTaskObjectConstraint taskConstraint
                            = (StateTaskObjectConstraint)first.node;

                        // Removing related TCStateTaskInstance for the
                        // StateTaskObjectConstraint

                        TCStateTaskInstance task = (TCStateTaskInstance)stConstraintsHT.get( taskConstraint );
                        if ( !ownHeuristicChange )
                        {
                            excludeStateTaskInstance( task );
                        }
                        stConstraintsHT.remove( taskConstraint );
                        stateTaskNumber-- ;
                        if ( task.externallyConnected )
                        {
                            externalStateTaskNumber-- ;
                        }

                        // Removing IntegerConstraints from the variables

                        modificationModus = true;
                        gsc.removeIntegerConstraint( task.tempVariable );
                        gsc.removeIntegerConstraint( task.resourceTypeVariable );
                        gsc.removeIntegerConstraint( task.contributionVariable );
                        modificationModus = false;

                        // Updating structural consistency

                        reactivatePossibleActionTypes();
                        checkForValidActionType();

                    }
                }
            }
        }

    }


    // ***** IMPROVEMENT (Service methods) *****


    /**
     * Adds a precondition task to the constraint.
     * @param preconditionTask  The task blueprint for the task.
     */
    private void addPreconditionTask( TCPreconditionTask preconditionTask )
    {

        // Adding and initializing new variables

        Variable temporalReferenceVariable
            = gsc.createVariable_internal( "temporal reference" );
        if ( desiredTemporalAnchor >= 0 )
        {
            if ( Registry.LOG_CHANGES )
            {
                System.out.println( "adding precondition task " + preconditionTask + " at " + ( desiredTemporalAnchor + preconditionTask.relativeShift ) );
            }
            gsc.setVariable_internal( temporalReferenceVariable ,
                desiredTemporalAnchor + preconditionTask.relativeShift );
        }
        else
        {
            if ( Registry.LOG_CHANGES )
            {
                System.out.println( "adding precondition task " + preconditionTask + " at " + earliestStartingTime );
            }
            gsc.setVariable_internal( temporalReferenceVariable ,
                earliestStartingTime );
        }
        Variable stateVariable = gsc.createVariable_internal( "state" );
        gsc.setVariable_internal( stateVariable , preconditionTask.state );

        // Fetching/establishing resource type variable of corresponding
        // state resource constraint

        Variable resourceTypeVariable;
        Embedding e;

        ConventionalConstraint src = gsc.findConstraint( Registry.STATERESOURCE_CONSTRAINT , Registry.RESOURCETYPE_EDGE , preconditionTask.resourceType );

        if ( src == null )
        {

            // Creating a new StateResourceConstraint and resource
            // type variable

            resourceTypeVariable
                = gsc.createVariable_internal( "resource type" );
            gsc.setVariable_internal( resourceTypeVariable ,
                preconditionTask.resourceType );
            Variable stateVariableSRC
                = gsc.createVariable_internal( "state" );
            ConventionalConstraint currentTimeObjectConstraint;
            synchronized( gsc.constraintSHT )
            {
                currentTimeObjectConstraint = (ConventionalConstraint)gsc.constraintSHT.get( Registry.nodeType[ Registry.CURRENTTIME_OBJECTCONSTRAINT ] );
            }

            e = new Embedding( 3 );
            gsc.extendEmbedding_internal( e , resourceTypeVariable ,
                Registry.RESOURCETYPE_EDGE , Registry.INCOMING );
            gsc.extendEmbedding_internal( e , stateVariableSRC ,
                Registry.STATE_EDGE , Registry.INCOMING );
            gsc.extendEmbedding_internal( e , currentTimeObjectConstraint ,
                Registry.PLAIN_EDGE , Registry.INCOMING );

            gsc.selectCostFunction( gsc.getStructuralCostFunction() );
            src = gsc.createConstraint_internal( Registry.STATERESOURCE_CONSTRAINT , e );

            gsc.selectCostFunction( gsc.getDefaultCostFunction() );
            gsc.addCosts( src , StateResourceConstraint.PRECONDITION_FAILURE );

        }
        else
        {

            // Fetching state resource constraint's resource type variable

            synchronized( src.connections )
            {
                resourceTypeVariable = src.connections.getSpecificVariable( Registry.RESOURCETYPE_EDGE );
            }

        }

        // Creating the precondition task

        e = new Embedding( 3 );
        gsc.extendEmbedding_internal( e , temporalReferenceVariable ,
            Registry.TEMPORALREFERENCE_EDGE , Registry.INCOMING );
        gsc.extendEmbedding_internal( e , stateVariable ,
            Registry.STATE_EDGE , Registry.INCOMING );
        gsc.extendEmbedding_internal( e , resourceTypeVariable ,
            Registry.RESOURCETYPE_EDGE , Registry.INCOMING );
        ConventionalConstraint preconditionConstraint = gsc.createConstraint_internal( Registry.PRECONDITIONTASK_OBJECTCONSTRAINT , e );

        // Connecting the precondition task

        e = new Embedding( 1 );
        gsc.extendEmbedding_internal( e , preconditionConstraint ,
            Registry.PLAIN_EDGE , Registry.INCOMING );
        gsc.extendConstraint_internal( actionConstraint , e );
        gsc.extendConstraint_internal( src , e );

    }


    /**
     * Adds a action task to the constraint.
     * @param actionTask  The task blueprint for the task.
     */
    private void addActionTask( TCActionTask actionTask )
    {

        // Adding and initializing new variables

        Variable operationVariable = gsc.createVariable_internal( "operation" );
        gsc.setVariable_internal( operationVariable , actionTask.operation );
        Variable beginVariable = gsc.createVariable_internal( "begin" );
        if ( desiredTemporalAnchor >= earliestStartingTime )
        {
            if ( Registry.LOG_CHANGES )
            {
                System.out.println( "adding action task " + actionTask + " at " + ( desiredTemporalAnchor + actionTask.relativeShift ) );
            }
            gsc.setVariable_internal( beginVariable ,
                desiredTemporalAnchor + actionTask.relativeShift );
        }
        else
        {
            if ( Registry.LOG_CHANGES )
            {
                System.out.println( "adding action task " + actionTask + " at " + earliestStartingTime );
            }
            gsc.setVariable_internal( beginVariable , earliestStartingTime );
        }
        Variable durationVariable = gsc.createVariable_internal( "duration" );
        gsc.setVariable_internal( durationVariable , actionTask.duration );
        Variable consumptionVariable
            = gsc.createVariable_internal( "consumption" );
        gsc.setVariable_internal( consumptionVariable ,
            actionTask.consumption );
        Variable executionStateVariable
            = gsc.createVariable_internal( "execution state" );
        gsc.setVariable_internal( executionStateVariable ,
            DiscreteResourceConstraint.PENDING );

        // Fetching/establishing resource type variable of corresponding
        // discrete resource constraint

        Variable resourceTypeVariable;
        Embedding e;

        ConventionalConstraint drc = gsc.findConstraint( Registry.DISCRETERESOURCE_CONSTRAINT , Registry.RESOURCETYPE_EDGE , actionTask.resourceType );

        if ( drc == null )
        {

            // Creating a new DiscreteResourceConstraint and resource
            // type variable

            resourceTypeVariable
                = gsc.createVariable_internal( "resource type" );
            gsc.setVariable_internal( resourceTypeVariable ,
                actionTask.resourceType );
            Variable capacityVariable
                = gsc.createVariable_internal( "capacity" );
            gsc.setVariable_internal( capacityVariable , 1 );
            ConventionalConstraint currentTimeObjectConstraint;
            synchronized( gsc.constraintSHT )
            {
                currentTimeObjectConstraint = (ConventionalConstraint)gsc.constraintSHT.get( Registry.nodeType[ Registry.CURRENTTIME_OBJECTCONSTRAINT ] );
            }

            e = new Embedding( 3 );
            gsc.extendEmbedding_internal( e , resourceTypeVariable ,
                Registry.RESOURCETYPE_EDGE , Registry.INCOMING );
            gsc.extendEmbedding_internal( e , capacityVariable ,
                Registry.CAPACITY_EDGE , Registry.INCOMING );
            gsc.extendEmbedding_internal( e , currentTimeObjectConstraint ,
                Registry.PLAIN_EDGE , Registry.INCOMING );
            drc = gsc.createConstraint_internal( Registry.DISCRETERESOURCE_CONSTRAINT , e );

        }
        else
        {

            // Fetching discrete resource constraint's resource type variable

            synchronized( drc.connections )
            {
                resourceTypeVariable = drc.connections.getSpecificVariable( Registry.RESOURCETYPE_EDGE );
            }

        }

        // Creating the action task

        e = new Embedding( 6 );
        gsc.extendEmbedding_internal( e , operationVariable ,
            Registry.OPERATION_EDGE , Registry.INCOMING );
        gsc.extendEmbedding_internal( e , beginVariable ,
            Registry.BEGIN_EDGE , Registry.INCOMING );
        gsc.extendEmbedding_internal( e , durationVariable ,
            Registry.DURATION_EDGE , Registry.INCOMING );
        gsc.extendEmbedding_internal( e , consumptionVariable ,
            Registry.CONSUMPTION_EDGE , Registry.INCOMING );
        gsc.extendEmbedding_internal( e , resourceTypeVariable ,
            Registry.RESOURCETYPE_EDGE , Registry.INCOMING );
        gsc.extendEmbedding_internal( e , executionStateVariable ,
            Registry.EXECUTIONSTATE_EDGE , Registry.INCOMING );

        ConventionalConstraint actionTaskConstraint = gsc.createConstraint_internal( Registry.ACTIONTASK_OBJECTCONSTRAINT , e );

        // Connecting the action task

        e = new Embedding( 1 );
        gsc.extendEmbedding_internal( e , actionTaskConstraint ,
            Registry.PLAIN_EDGE , Registry.INCOMING );
        gsc.extendConstraint_internal( actionConstraint , e );
        gsc.extendConstraint_internal( drc , e );

    }


    /**
     * Adds a state task to the constraint.
     * @param stateTask  The task blueprint for the task.
     */
    private void addStateTask( TCStateTask stateTask )
    {

        // Adding and initializing new variables

        Variable temporalReferenceVariable
            = gsc.createVariable_internal( "temporal reference" );
        if ( desiredTemporalAnchor >= earliestStartingTime )
        {
            if ( Registry.LOG_CHANGES )
            {
                System.out.println( "adding state task " + stateTask + " at " + ( desiredTemporalAnchor + stateTask.relativeShift ) );
            }
            gsc.setVariable_internal( temporalReferenceVariable ,
                desiredTemporalAnchor + stateTask.relativeShift );
        }
        else
        {
            if ( Registry.LOG_CHANGES )
            {
                System.out.println( "adding state task " + stateTask + " at " + earliestStartingTime );
            }
            gsc.setVariable_internal( temporalReferenceVariable ,
                earliestStartingTime );
        }
        Variable contributionVariable
            = gsc.createVariable_internal( "contribution" );
        gsc.setVariable_internal( contributionVariable ,
            stateTask.contribution );

        // Fetching/establishing resource type variable of corresponding
        // state resource constraint

        Variable resourceTypeVariable;
        Embedding e;

        ConventionalConstraint src = gsc.findConstraint( Registry.STATERESOURCE_CONSTRAINT , Registry.RESOURCETYPE_EDGE , stateTask.resourceType );

        if ( src == null )
        {

            // Creating a new StateResourceConstraint and resource
            // type variable

            resourceTypeVariable
                = gsc.createVariable_internal( "resource type" );
            gsc.setVariable_internal( resourceTypeVariable ,
                stateTask.resourceType );
            Variable stateVariable
                = gsc.createVariable_internal( "state" );
            ConventionalConstraint currentTimeObjectConstraint;
            synchronized( gsc.constraintSHT )
            {
                currentTimeObjectConstraint = (ConventionalConstraint)gsc.constraintSHT.get( Registry.nodeType[ Registry.CURRENTTIME_OBJECTCONSTRAINT ] );
            }

            e = new Embedding( 3 );
            gsc.extendEmbedding_internal( e , resourceTypeVariable ,
                Registry.RESOURCETYPE_EDGE , Registry.INCOMING );
            gsc.extendEmbedding_internal( e , stateVariable ,
                Registry.STATE_EDGE , Registry.INCOMING );
            gsc.extendEmbedding_internal( e , currentTimeObjectConstraint ,
                Registry.PLAIN_EDGE , Registry.INCOMING );

            gsc.selectCostFunction( gsc.getStructuralCostFunction() );
            src = gsc.createConstraint_internal( Registry.STATERESOURCE_CONSTRAINT , e );

            gsc.selectCostFunction( gsc.getDefaultCostFunction() );
            gsc.addCosts( src , StateResourceConstraint.PRECONDITION_FAILURE );

        }
        else
        {

            // Fetching state resource constraint's resource type variable

            synchronized( src.connections )
            {
                resourceTypeVariable = src.connections.getSpecificVariable( Registry.RESOURCETYPE_EDGE );
            }

        }

        // Creating the state task

        e = new Embedding( 3 );
        gsc.extendEmbedding_internal( e , temporalReferenceVariable ,
            Registry.TEMPORALREFERENCE_EDGE , Registry.INCOMING );
        gsc.extendEmbedding_internal( e , contributionVariable ,
            Registry.CONTRIBUTION_EDGE , Registry.INCOMING );
        gsc.extendEmbedding_internal( e , resourceTypeVariable ,
            Registry.RESOURCETYPE_EDGE , Registry.INCOMING );
        ConventionalConstraint stateTaskConstraint = gsc.createConstraint_internal( Registry.STATETASK_OBJECTCONSTRAINT , e );

        // Connecting the state task

        e = new Embedding( 1 );
        gsc.extendEmbedding_internal( e , stateTaskConstraint ,
            Registry.PLAIN_EDGE , Registry.INCOMING );
        gsc.extendConstraint_internal( actionConstraint , e );
        gsc.extendConstraint_internal( src , e );

    }


    /**
     * Disconnects and removes a task from the action object.
     * @param taskConstraint  The task to be removed.
     * @param otherConstraintID  The ID of the constraint to be potentially
     *                           disconnected from.
     */
    private void removeTask( ConventionalConstraint taskConstraint , int otherConstraintID )
    {

        // Disconnection from the action object

        Embedding e = new Embedding( 1 );
        gsc.extendEmbedding_internal( e , taskConstraint ,
            Registry.PLAIN_EDGE , Registry.INCOMING );
        gsc.reduceConstraint_internal( actionConstraint , e );

        // Deletion and disconnection from the task constraint

        gsc.disconnectAndRemoveConstraint_internal( taskConstraint ,
            otherConstraintID , e );

    }


    /**
     * Removes the task constraint (including the action object and all
     * tasks).
     */
    public void removeTaskConstraint() {

        // Self-destruction

        gsc.deleteConstraint_internal( this );

        // Potentially disconnecting transition constraint

        ConventionalConstraint transitionConstraint;
        synchronized( actionConstraint.connections )
        {
           transitionConstraint = actionConstraint.connections.getSpecificConstraint( Registry.TRANSITION_CONSTRAINT , Registry.PLAIN_EDGE );
        }
        if ( transitionConstraint != null )
        {
            Embedding e = new Embedding( 1 );
            gsc.extendEmbedding_internal( e , actionConstraint ,
                Registry.PLAIN_EDGE , Registry.INCOMING );
            gsc.reduceConstraint_internal( transitionConstraint , e );
        }

        // Disconnecting/deleting all precondition tasks

        Iterator it = preconditionTaskSet.iterator();
        while ( it.hasNext() )
        {
            removeTask( ( (TCPreconditionTaskInstance)it.next() ).constraint ,
                Registry.STATERESOURCE_CONSTRAINT );
        }

        // Disconnecting/deleting all action tasks

        it = actionTaskSet.iterator();
        while ( it.hasNext() )
        {
            removeTask( ( (TCActionTaskInstance)it.next() ).constraint ,
                Registry.DISCRETERESOURCE_CONSTRAINT );
        }

        // Disconnecting/deleting all state tasks

        it = stateTaskSet.iterator();
        while ( it.hasNext() )
        {
            removeTask( ( (TCStateTaskInstance)it.next() ).constraint ,
                Registry.STATERESOURCE_CONSTRAINT );
        }

        // Action object deletion

        gsc.deleteConstraint_internal( actionConstraint );

        // Action type variable deletion

        int connectionNumber;
        synchronized( actionTypeVar.connections )
        {
            connectionNumber = actionTypeVar.connections.number();
        }
        if ( connectionNumber == 0 )
        {
            gsc.deleteVariable_internal( actionTypeVar );
        }

    }


    /**
     * Repairs a temporal relation by shifting one of the tasks' begin
     * variables. Returns whether the procedure was able to succeed.
     * @param temporalRelation  The temporal relation to be repaired.
     * @param correctVar1OK  If the first task should be moved.
     * @param correctVar2OK  If the second task should be moved.
     */
    private boolean repairTemporalRelation( TCTemporalRelationInstance
        temporalRelation , boolean correctVar1OK , boolean correctVar2OK )
    {

        int taskToBeRepaired;
        long modifier = 0;
        long newValue;
        long shift = 0;
        long otherValue;
        Variable var1;
        Variable var2;
        long minTime1;
        long minTime2;
        long upperBound1;
        long upperBound2;
        boolean change = false;

        if ( correctVar1OK && correctVar2OK )
        {
            taskToBeRepaired = gsc.random.nextInt( 2 );
        }
        else
        {
            if ( correctVar1OK )
            {
                taskToBeRepaired = 0;
            }
            else
            {
                taskToBeRepaired = 1;
            }
        }

        // Repairing temporal relation

        if ( taskToBeRepaired == 0 )
        {

            if ( temporalRelation.variableID1 == END )
            {
                modifier = -( (TCActionTaskInstance)temporalRelation.task1 ).tempValue[ DURATION ];
            }
            if ( temporalRelation.comparator == LESS_THAN )
            {
                modifier-- ;
            }
            newValue = temporalRelation.constant + modifier;

            if ( temporalRelation.task2 instanceof TCPreconditionTaskInstance )
            {
                otherValue = ( (TCPreconditionTaskInstance)temporalRelation.task2 ).tempValue;
                var2 = ( (TCPreconditionTaskInstance)temporalRelation.task2 ).tempVariable;
                minTime2 = 0;
            }
            else
            {
                if ( temporalRelation.task2 instanceof TCActionTaskInstance )
                {
                    otherValue = ( (TCActionTaskInstance)temporalRelation.task2 ).tempValue[ temporalRelation.variableID2 ];
                    var2 = ( (TCActionTaskInstance)temporalRelation.task2 ).tempVariable[ BEGIN ];
                    minTime2 = earliestStartingTime;
                }
                else
                {
                    otherValue = ( (TCStateTaskInstance)temporalRelation.task2 ).tempValue;
                    var2 = ( (TCStateTaskInstance)temporalRelation.task2 ).tempVariable;
                    minTime2 = currentTime;
                }
            }
            newValue += otherValue;

            if ( temporalRelation.task1 instanceof TCPreconditionTaskInstance )
            {
                var1 = ( (TCPreconditionTaskInstance)temporalRelation.task1 ).tempVariable;
                ( (TCPreconditionTaskInstance)temporalRelation.task1 ).repaired = true;
                minTime1 = 0;
            }
            else
            {
                if ( temporalRelation.task1 instanceof TCActionTaskInstance )
                {
                    var1 = ( (TCActionTaskInstance)temporalRelation.task1 ).tempVariable[ BEGIN ];
                    ( (TCActionTaskInstance)temporalRelation.task1 ).repaired = true;
                    minTime1 = earliestStartingTime;
                }
                else
                {
                    var1 = ( (TCStateTaskInstance)temporalRelation.task1 ).tempVariable;
                    ( (TCStateTaskInstance)temporalRelation.task1 ).repaired = true;
                    minTime1 = currentTime;
                }
            }

            if ( minTime1 > newValue )
            {
                shift = minTime1 - newValue;
            }
            if ( otherValue + shift < minTime2 )
            {
                shift = minTime2 - otherValue;
            }
            upperBound1 = (long)var1.upperBound();
            if ( upperBound1 < newValue + shift )
            {
                shift = upperBound1 - newValue;
            }
            upperBound2 = (long)var2.upperBound();
            if ( upperBound2 < otherValue + shift )
            {
                shift = upperBound2 - otherValue;
            }

            if ( ( shift != 0 ) && correctVar2OK )
            {
                if ( var2.value() != otherValue + shift )
                {
                    gsc.setVariable_internal( var2 , otherValue + shift );
                    change = true;
                }
            }
            if ( var1.value() != newValue + shift )
            {
                gsc.setVariable_internal( var1 , newValue + shift );
                change = true;
            }

        }
        else
        {

            if ( temporalRelation.variableID2 == END )
            {
                modifier = -( (TCActionTaskInstance)temporalRelation.task2 ).tempValue[ DURATION ];
            }
            if ( temporalRelation.comparator == LESS_THAN )
            {
                modifier++ ;
            }
            newValue = modifier - temporalRelation.constant;

            if ( temporalRelation.task1 instanceof TCPreconditionTaskInstance )
            {
                otherValue = ( (TCPreconditionTaskInstance)temporalRelation.task1 ).tempValue;
                var1 = ( (TCPreconditionTaskInstance)temporalRelation.task1 ).tempVariable;
                minTime1 = 0;
            }
            else
            {
                if ( temporalRelation.task1 instanceof TCActionTaskInstance )
                {
                    otherValue = ( (TCActionTaskInstance)temporalRelation.task1 ).tempValue[ temporalRelation.variableID1 ];
                    var1 = ( (TCActionTaskInstance)temporalRelation.task1 ).tempVariable[ BEGIN ];
                    minTime1 = earliestStartingTime;
                }
                else
                {
                    otherValue = ( (TCStateTaskInstance)temporalRelation.task1 ).tempValue;
                    var1 = ( (TCStateTaskInstance)temporalRelation.task1 ).tempVariable;
                    minTime1 = currentTime;
                }
            }
            newValue += otherValue;

            if ( temporalRelation.task2 instanceof TCPreconditionTaskInstance )
            {
                var2 = ( (TCPreconditionTaskInstance)temporalRelation.task2 ).tempVariable;
                ( (TCPreconditionTaskInstance)temporalRelation.task2 ).repaired = true;
                minTime2 = 0;
            }
            else
            {
                if ( temporalRelation.task2 instanceof TCActionTaskInstance )
                {
                    var2 = ( (TCActionTaskInstance)temporalRelation.task2 ).tempVariable[ BEGIN ];
                    ( (TCActionTaskInstance)temporalRelation.task2 ).repaired = true;
                    minTime2 = earliestStartingTime;
                }
                else
                {
                    var2 = ( (TCStateTaskInstance)temporalRelation.task2 ).tempVariable;
                    ( (TCStateTaskInstance)temporalRelation.task2 ).repaired = true;
                    minTime2 = currentTime;
                }
            }

            if ( minTime2 > newValue )
            {
                shift = minTime2 - newValue;
            }
            if ( otherValue + shift < minTime1 )
            {
                shift = minTime1 - otherValue;
            }
            upperBound2 = (long)var2.upperBound();
            if ( upperBound2 < newValue + shift )
            {
                shift = upperBound2 - newValue;
            }
            upperBound1 = (long)var1.upperBound();
            if ( upperBound1 < otherValue + shift )
            {
                shift = upperBound1 - otherValue;
            }

            if ( ( shift != 0 ) && correctVar1OK )
            {
                if ( var1.value() != otherValue + shift )
                {
                    gsc.setVariable_internal( var1 , otherValue + shift );
                    change = true;
                }
            }
            if ( var2.value() != newValue + shift )
            {
                gsc.setVariable_internal( var2 , newValue + shift );
                change = true;
            }

        }

        return change;

    }


    // ***** (TRANSITION_)TEMPORAL_INCONSISTENCY IMPROVEMENT (Heuristics) *****


    /**
     * Improving the current temporal inconsistency costs by repairing
     * a temporal relation. Returns whether the heuristic was able to
     * succeed.
     * @param chosenCosts  The selection of the costs to be improved.
     */
    private boolean heuristic_RepairTemporalRelation( long chosenCosts )
    {

        // Selecting an inconsistent temporal relation

        OrderedShiftDLLElement currentLE = temporalRelationsList.first;
        TCTemporalRelationInstance temporalRelation = null;
        boolean notFound = true;

        while ( notFound )
        {
            temporalRelation = (TCTemporalRelationInstance)currentLE.element;
            chosenCosts -= temporalRelation.costs;
            if ( notFound = ( chosenCosts >= 0 ) )
            {
                currentLE = currentLE.next;
            }
        }

        if ( Registry.LOG_CHANGES )
        {
            System.out.println( "TC RepairTemporalRelation: "
                + temporalRelation );
        }

        // Repair chosen temporal relation (task is selected by random)

        boolean OK1 = ( temporalRelation.task1 != referenceTask );
        boolean OK2 = ( temporalRelation.task2 != referenceTask );

        if ( !executionStateChangesEnabled )
        {
            if ( ( temporalRelation.task1 instanceof TCActionTaskInstance ) && ( ( (TCActionTaskInstance)temporalRelation.task1 ).executionState != DiscreteResourceConstraint.PENDING ) )
            {
                OK1 = false;
            }
            if ( ( temporalRelation.task2 instanceof TCActionTaskInstance ) && ( ( (TCActionTaskInstance)temporalRelation.task2 ).executionState != DiscreteResourceConstraint.PENDING ) )
            {
                OK2 = false;
            }
        }

        return repairTemporalRelation( temporalRelation , OK1 , OK2 );

    }


    /**
     * Improving the current temporal inconsistency costs by repairing
     * all temporal relations. Returns whether the heuristic was able to
     * succeed.
     */
    private boolean heuristic_RepairAllTemporalRelations()
    {

        if ( Registry.LOG_CHANGES )
        {
            System.out.println( "TC RepairAllTemporalRelations" );
        }

        OrderedShiftDLLElement firstLE = null;
        OrderedShiftDLLElement currentLE = temporalRelationsList.first;
        int currentPriority;
        long currentCosts;
        TCTemporalRelationInstance temporalRelation;
        boolean notReady;
        boolean OK1;
        boolean OK2;
        int task1priority;
        int task2priority;

        // Marking tasks as not repaired

        Iterator it = preconditionTaskSet.iterator();
        while ( it.hasNext() )
        {
            ( (TCPreconditionTaskInstance)it.next() ).repaired = false;
        }
        it = actionTaskSet.iterator();
        while ( it.hasNext() )
        {
            ( (TCActionTaskInstance)it.next() ).repaired = false;
        }
        it = stateTaskSet.iterator();
        while ( it.hasNext() )
        {
            ( (TCStateTaskInstance)it.next() ).repaired = false;
        }

        // Blocking a repair of the reference task

        if ( referenceTask != null )
        {
            referenceTask.repaired = true;
        }

        // Marking temporal relations as not repaired

        do
        {
            ( (TCTemporalRelationInstance)currentLE.element ).repaired = false;
            currentLE = currentLE.next;
        } while ( currentLE != null );

        while ( true )
        {

            // Summing costs for current priority

            currentLE = temporalRelationsList.first;
            currentPriority = -1;
            notReady = true;
            currentCosts = 0;

            do
            {
                temporalRelation
                    = (TCTemporalRelationInstance)currentLE.element;
                if ( !temporalRelation.repaired )
                {
                    if ( temporalRelation.externalTaskMoveCount
                        == currentPriority )
                    {
                        currentCosts += temporalRelation.costs;
                    }
                    else
                    {
                        if ( currentPriority == -1 )
                        {
                            if ( temporalRelation.costs != 0 )
                            {
                                currentPriority
                                    = temporalRelation.externalTaskMoveCount;
                                firstLE = currentLE;
                                currentCosts = temporalRelation.costs;
                            }
                        }
                        else
                        {
                            notReady = false;
                        }
                    }
                }
                currentLE = currentLE.next;
            } while ( notReady && ( currentLE != null ) );

            if ( ( currentLE == null ) && ( currentCosts == 0 ) )
            {
                return true;
            }

            // Choosing one of the current priority's temporal relations

            currentCosts = Math.abs( gsc.random.nextLong() ) % currentCosts;
            currentLE = firstLE;
            notReady = true;

            while ( notReady )
            {
                temporalRelation
                    = (TCTemporalRelationInstance)currentLE.element;
                if ( !temporalRelation.repaired )
                {
                    currentCosts -= temporalRelation.costs;
                    if ( notReady = ( currentCosts >= 0 ) )
                    {
                        currentLE = currentLE.next;
                    }
                }
                else
                {
                    currentLE = currentLE.next;
                }
            }
            temporalRelation.repaired = true;

            // Repair chosen temporal relation

            if ( temporalRelation.task1 instanceof TCPreconditionTaskInstance )
            {
                OK1 = !( (TCPreconditionTaskInstance)temporalRelation.task1 ).repaired;
                task1priority = ( (TCPreconditionTaskInstance)temporalRelation.task1 ).externalTaskMoveCount;
            }
            else
            {
                if ( temporalRelation.task1 instanceof TCActionTaskInstance )
                {
                    OK1 = !( (TCActionTaskInstance)temporalRelation.task1 ).repaired;
                    if ( !executionStateChangesEnabled && ( ( (TCActionTaskInstance)temporalRelation.task1 ).executionState != DiscreteResourceConstraint.PENDING ) )
                    {
                        OK1 = false;
                    }

                    task1priority = ( (TCActionTaskInstance)temporalRelation.task1 ).externalTaskMoveCount;
                }
                else
                {
                    OK1 = !( (TCStateTaskInstance)temporalRelation.task1 ).repaired;
                    task1priority = ( (TCStateTaskInstance)temporalRelation.task1 ).externalTaskMoveCount;
                }
            }

            if ( temporalRelation.task2 instanceof TCPreconditionTaskInstance )
            {
                OK2 = !( (TCPreconditionTaskInstance)temporalRelation.task2 ).repaired;
                task2priority = ( (TCPreconditionTaskInstance)temporalRelation.task2 ).externalTaskMoveCount;
            }
            else
            {
                if ( temporalRelation.task2 instanceof TCActionTaskInstance )
                {
                    OK2 = !( (TCActionTaskInstance)temporalRelation.task2 ).repaired;
                    if ( !executionStateChangesEnabled && ( ( (TCActionTaskInstance)temporalRelation.task2 ).executionState != DiscreteResourceConstraint.PENDING ) )
                    {
                        OK2 = false;
                    }
                    task2priority = ( (TCActionTaskInstance)temporalRelation.task2 ).externalTaskMoveCount;
                }
                else
                {
                    OK2 = !( (TCStateTaskInstance)temporalRelation.task2 ).repaired;
                    task2priority = ( (TCStateTaskInstance)temporalRelation.task2 ).externalTaskMoveCount;
                }
            }

            if ( OK1 || OK2 )
            {
                if ( OK1 && OK2 )
                {
                    if ( task1priority > task2priority )
                    {
                        if ( !repairTemporalRelation( temporalRelation ,
                                 false , true ) )
                        {
                            repairTemporalRelation( temporalRelation ,
                                true,  false );
                        }
                    }
                    else
                    {
                        if ( task1priority < task2priority )
                        {
                            if ( !repairTemporalRelation( temporalRelation ,
                                     true , false ) )
                            {
                                repairTemporalRelation( temporalRelation ,
                                    false , true );
                            }
                        }
                        else
                        {
                            repairTemporalRelation( temporalRelation ,
                                true , true );
                        }
                    }
                }
                else
                {
                    repairTemporalRelation( temporalRelation , OK1 , OK2 );
                }
            }

        }

    }


    /**
     * An action task's execution state is adapted in order to improving
     * related costs. Returns whether the heuristic was able to succeed.
     * @param chosenCosts  The selection of the costs to be improved.
     */
    private boolean heuristic_ChangeExecutionState( long chosenCosts )
    {

        // Selecting an inconsistent action task

        TCActionTaskInstance task = null;
        Iterator it = actionTaskSet.iterator();

        while ( ( chosenCosts >= 0 ) && it.hasNext() )
        {
            task = (TCActionTaskInstance)it.next();
            chosenCosts -= task.executionStateCosts;
        }

        if ( Registry.LOG_CHANGES )
        {
            System.out.println( "TC ChangeExecutionState: " + task );
        }

        if ( currentTime < task.tempValue[ BEGIN ] )
        {
            gsc.setVariable_internal( task.executionStateVariable ,
                DiscreteResourceConstraint.PENDING );
        }
        else
        {
            if ( task.tempValue[ END ] <= currentTime )
            {
                gsc.setVariable_internal(task.executionStateVariable ,
                    DiscreteResourceConstraint.DONE );
            }
            else
            {
                gsc.setVariable_internal( task.executionStateVariable ,
                    DiscreteResourceConstraint.IN_EXECUTION );
            }
        }

        return true;

    }


    /**
     * Moves an action state with an invalid execution state in order to
     * improving related costs. Returns whether the heuristic was able to
     * succeed.
     * @param chosenCosts  The selection of the costs to be improved.
     */
    private boolean heuristic_ExecutionStateMove( long chosenCosts )
    {

        // Selecting an inconsistent action task

        TCActionTaskInstance task = null;
        Iterator it = actionTaskSet.iterator();

        while ( ( chosenCosts >= 0 ) && it.hasNext() )
        {
            task = (TCActionTaskInstance)it.next();
            chosenCosts -= task.executionStateCosts;
        }

        if ( Registry.LOG_CHANGES )
        {
            System.out.println( "TC ExecutionStateMove: " + task );
        }

        if ( task.executionState == DiscreteResourceConstraint.PENDING )
        {
            gsc.setVariable_internal( task.tempVariable[ BEGIN ] ,
                earliestStartingTime );
        }
        else
        {
            if ( task.executionState == DiscreteResourceConstraint.DONE )
            {
                gsc.setVariable_internal( task.tempVariable[ BEGIN ] ,
                    currentTime - task.tempValue[ DURATION ] );
            }
            else
            {
                gsc.setVariable_internal( task.tempVariable[ BEGIN ] ,
                    currentTime - task.tempValue[ DURATION ] + 1 );
            }
        }

        return true;

    }


    /**
     * Improving the current costs of temporal inconsistencies.
     */
    private void improveTemporalInconsistency() {

        if ( Registry.LOG_CHANGES )
        {
            System.out.println( "improving temporal inconsistency:" );
        }

        long chosenCosts = Math.abs( gsc.random.nextLong() )
            % costs[ temporalRelationCostID ];

        if ( chosenCosts >= executionStateCosts )
        {
            /*
            heuristic_RepairTemporalRelation(chosenCosts
                                             - executionStateCosts);
            */
            heuristic_RepairAllTemporalRelations();
        }
        else
        {
            if ( executionStateChangesEnabled )
            {
                heuristic_ChangeExecutionState( chosenCosts );
            }
            else
            {
                heuristic_ExecutionStateMove( chosenCosts );
            }
        }

    }


    // ***** CONFIGURATION_INCONSISTENCY IMPROVEMENT (Heuristics) *****


    /**
     * Compares the externally added tasks to an action type. Returns -1
     * if an externally added task is not part of the action type, and a
     * distance measure otherwise.
     * @param ID  The ID of the action type to be compared.
     */
    private long actionTypeIncludesExternalConfiguration( int ID )
    {

        TCActionType selectedActionType = actionType[ ID ];

        // Simple pre-check if task numbers are less

        if ( ( externalPreconditionTaskNumber > selectedActionType.preconditionTaskNumber ) || ( externalActionTaskNumber > selectedActionType.actionTaskNumber ) || ( externalStateTaskNumber > selectedActionType.stateTaskNumber ) )
        {
            return -1;
        }

        // Checking precondition task instances

        Iterator itInstance;
        Iterator itType;
        boolean found;
        long foundExternalTaskMatches = 0;
        long ptDistance = 0;

        TCPreconditionTask preconditionTask;
        TCPreconditionTaskInstance preconditionTaskInstance = null;

        // Initializing instance markings

        itInstance = preconditionTaskSet.iterator();
        while ( itInstance.hasNext() )
        {
            ( (TCPreconditionTaskInstance)itInstance.next() ).found = null;
        }

        // Iterating through action type's precondition tasks

        itType = selectedActionType.preconditionTasks.iterator();

        while ( itType.hasNext() )
        {
            preconditionTask = (TCPreconditionTask)itType.next();

            // Searching for corresponding precondition task instance

            itInstance = preconditionTaskSet.iterator();
            found = false;

            while ( !found && itInstance.hasNext() )
            {
                preconditionTaskInstance
                    = (TCPreconditionTaskInstance)itInstance.next();
                if ( preconditionTaskInstance.found == null )
                {
                    found = ( ( preconditionTaskInstance.resourceType == preconditionTask.resourceType ) && ( preconditionTaskInstance.state == preconditionTask.state ) );
                }
            }

            if ( found )
            {
                preconditionTaskInstance.found = preconditionTask;
                if ( preconditionTaskInstance.externallyConnected )
                {
                    foundExternalTaskMatches++ ;
                }
            }
            else
            {
                ptDistance++ ;
            }
        }

        if ( foundExternalTaskMatches < externalPreconditionTaskNumber )
        {
            return -1;
        }
        ptDistance += preconditionTaskNumber
            - selectedActionType.preconditionTaskNumber + ptDistance;

        // Checking action task instances

        foundExternalTaskMatches = 0;
        long atDistance = 0;

        TCActionTask actionTask;
        TCActionTaskInstance actionTaskInstance = null;

        // Initializing instance markings

        itInstance = actionTaskSet.iterator();
        while ( itInstance.hasNext() )
        {
            ( (TCActionTaskInstance)itInstance.next() ).found = null;
        }

        // Iterating through action type's action tasks

        itType = selectedActionType.actionTasks.iterator();

        while ( itType.hasNext() )
        {
            actionTask = (TCActionTask)itType.next();

            // Searching for corresponding action task instance

            itInstance = actionTaskSet.iterator();
            found = false;

            while ( !found && itInstance.hasNext() )
            {
                actionTaskInstance = (TCActionTaskInstance)itInstance.next();
                if ( actionTaskInstance.found == null )
                {
                    found = ( ( actionTaskInstance.resourceType == actionTask.resourceType ) && ( actionTaskInstance.operation == actionTask.operation ) && ( actionTaskInstance.tempValue[ DURATION ] == actionTask.duration ) && ( actionTaskInstance.consumption == actionTask.consumption ) );
                }
            }

            if ( found )
            {
                actionTaskInstance.found = actionTask;
                if ( actionTaskInstance.externallyConnected )
                {
                    foundExternalTaskMatches++ ;
                }
            }
            else
            {
                atDistance++ ;
            }
        }

        if ( foundExternalTaskMatches < externalActionTaskNumber )
        {
            return -1;
        }
        atDistance += actionTaskNumber
            - selectedActionType.actionTaskNumber + atDistance;

        // Checking state task instances

        foundExternalTaskMatches = 0;
        long stDistance = 0;

        TCStateTask stateTask;
        TCStateTaskInstance stateTaskInstance = null;

        // Initializing instance markings

        itInstance = stateTaskSet.iterator();
        while ( itInstance.hasNext() )
        {
            ( (TCStateTaskInstance)itInstance.next() ).found = null;
        }

        // Iterating through action type's state tasks

        itType = selectedActionType.stateTasks.iterator();

        while ( itType.hasNext() )
        {
            stateTask = (TCStateTask)itType.next();

            // Searching for corresponding state task instance

            itInstance = stateTaskSet.iterator();
            found = false;

            while ( !found && itInstance.hasNext() )
            {
                stateTaskInstance
                    = (TCStateTaskInstance)itInstance.next();
                if ( stateTaskInstance.found == null )
                {
                    found = ( ( stateTaskInstance.resourceType == stateTask.resourceType ) && ( stateTaskInstance.contribution == stateTask.contribution ) );
                }
            }

            if ( found )
            {
                stateTaskInstance.found = stateTask;
                if ( stateTaskInstance.externallyConnected )
                {
                    foundExternalTaskMatches++ ;
                }
            }
            else
            {
                stDistance++ ;
            }
        }

        if ( foundExternalTaskMatches < externalStateTaskNumber )
        {
            return -1;
        }
        stDistance += stateTaskNumber
            - selectedActionType.stateTaskNumber + stDistance;

        return ptDistance + atDistance + stDistance;

    }


    /**
     * Sets the temporal anchor variable according to the initially connected
     * tasks and the given action type. If no valid temporal reference can
     * be determined, a value of Long.MIN_VALUE is assigned to the variable.
     * The temporal anchor value will also be temporarily saved in the action
     * type.
     * @param actionType  The action type with which the temporal anchor is
     *     to be aligned with.
     */
    private void setDesiredTemporalAnchor( TCActionType actionType )
    {

        if ( initiallyConnectedTask != null )
        {

            Iterator itType;
            boolean found;

            // Determining reference time (searching in action type
            // for a task that is similar to initially connected task)

            if ( initiallyConnectedTask instanceof TCStateTaskInstance )
            {

                TCStateTask stateTask;
                TCStateTaskInstance stateTaskInstance
                    = (TCStateTaskInstance)initiallyConnectedTask;
                itType = actionType.stateTasks.iterator();
                found = false;

                while ( !found )
                {
                    stateTask = (TCStateTask)itType.next();
                    if ( ( stateTask.resourceType == stateTaskInstance.resourceType ) && ( stateTask.contribution == stateTaskInstance.contribution ) )
                    {
                        found = true;
                        if ( stateTask.relativeShift != Long.MIN_VALUE )
                        {
                            actionType.temporalAnchor = stateTaskInstance.tempValue - stateTask.relativeShift;
                        }
                        else
                        {
                            actionType.temporalAnchor = Long.MIN_VALUE;
                        }
                    }
                }

            }
            else
            {
                if ( initiallyConnectedTask
                    instanceof TCPreconditionTaskInstance )
                {

                    TCPreconditionTask preconditionTask;
                    TCPreconditionTaskInstance preconditionTaskInstance
                        = (TCPreconditionTaskInstance)initiallyConnectedTask;
                    itType = actionType.preconditionTasks.iterator();
                    found = false;

                    while ( !found )
                    {
                        preconditionTask = (TCPreconditionTask)itType.next();
                        if ( ( preconditionTask.resourceType == preconditionTaskInstance.resourceType ) && ( preconditionTask.state == preconditionTaskInstance.state ) )
                        {
                            found = true;
                            if ( preconditionTask.relativeShift
                                != Long.MIN_VALUE )
                            {
                                actionType.temporalAnchor = preconditionTaskInstance.tempValue - preconditionTask.relativeShift;
                            }
                            else
                            {
                                actionType.temporalAnchor = Long.MIN_VALUE;
                            }
                        }
                    }

                }
                else
                {
                    if ( initiallyConnectedTask
                        instanceof TCActionTaskInstance )
                    {

                        TCActionTask actionTask;
                        TCActionTaskInstance actionTaskInstance
                            = (TCActionTaskInstance)initiallyConnectedTask;
                        itType = actionType.actionTasks.iterator();
                        found = false;

                        while ( !found )
                        {
                            actionTask = (TCActionTask)itType.next();
                            if ( ( actionTask.resourceType == actionTaskInstance.resourceType ) && ( actionTask.operation == actionTaskInstance.operation ) && ( actionTask.duration == actionTaskInstance.tempValue[ DURATION ] ) && ( actionTask.consumption == actionTaskInstance.consumption ) )
                            {
                                found = true;
                                if ( actionTask.relativeShift
                                    != Long.MIN_VALUE )
                                {
                                    actionType.temporalAnchor = actionTaskInstance.tempValue[ BEGIN ] - actionTask.relativeShift;
                                }
                                else
                                {
                                    actionType.temporalAnchor = Long.MIN_VALUE;
                                }
                            }
                        }

                    }
                }
            }
        }
        else
        {
            actionType.temporalAnchor = Long.MIN_VALUE;
        }

        desiredTemporalAnchor = actionType.temporalAnchor;

    }


    /**
     * Selects one of the action types as the desired configuration.
     * The ID of the selected action type will be stored in the
     * "desiredActionTypeID" variable. A value of -1 indicates that
     * the action should be deleted.
     */
    private void chooseDesiredActionTypeID()
    {

        desiredTemporalAnchor = Long.MIN_VALUE;

        if ( initiallyConnectedTask != null )
        {

            // Selection based on involved costs if initially connected
            // tasks make it possible to have a reference time point

            objectSelection.clear();
            long difficulty;

            Iterator itType;
            StateResourceConstraint src;
            TCStateTask stateTask;
            TCPreconditionTask preconditionTask;
            TCActionTask actionTask;
            long taskTime;

            for ( int i = 0 ; i < currentActionTypeNumber ; i++ )
            {
                if ( actionTypeOK[ i ] )
                {

                    setDesiredTemporalAnchor( actionType[ i ] );

                    if ( desiredTemporalAnchor != Long.MIN_VALUE )
                    {

                        // Calculating difficulties for desired action types

                        difficulty = 0;

                        // Adding precondition task difficulty

                        itType = actionType[ i ].preconditionTasks.iterator();
                        while ( itType.hasNext() )
                        {
                            preconditionTask
                                = (TCPreconditionTask)itType.next();
                            src = (StateResourceConstraint)( gsc.findConstraint( Registry.STATERESOURCE_CONSTRAINT , Registry.RESOURCETYPE_EDGE , preconditionTask.resourceType ) );
                            taskTime = desiredTemporalAnchor + preconditionTask.relativeShift;
                            if ( taskTime < 0 )
                            {
                                taskTime = 0;
                            }
                            difficulty += src.getDifficultyForStateAtTime( preconditionTask.state , taskTime );
                       }

                        // Adding difficulty in consequence of state tasks

                        itType = actionType[ i ].stateTasks.iterator();
                        while ( itType.hasNext() )
                        {
                            stateTask = (TCStateTask)itType.next();
                            src = (StateResourceConstraint)( gsc.findConstraint( Registry.STATERESOURCE_CONSTRAINT , Registry.RESOURCETYPE_EDGE , stateTask.resourceType ) );
                            taskTime = desiredTemporalAnchor + stateTask.relativeShift;
                            if ( taskTime < currentTime )
                            {
                                taskTime = currentTime;
                            }
                            difficulty += src.getDifficultyForContributionAtTime( stateTask.contribution , taskTime );
                        }

                        // Adding action type to selection

                        if ( difficulty < 0 )
                        {
                            // If the difficulty is better than neutral, add
                            // the action type twice
                            objectSelection.addElementAntiproportionally( actionType[ i ] , 0 );
                            objectSelection.addElementAntiproportionally( actionType[ i ] , 0 );
                        }
                        else
                        {
                            if ( difficulty <= Integer.MAX_VALUE )
                            {
                                objectSelection.addElementAntiproportionally( actionType[ i ] , (int)difficulty );
                            }
                            else
                            {
                                objectSelection.addElementAntiproportionally( actionType[ i ] , Integer.MAX_VALUE );
                            }
                        }

                    }
                    else
                    {

                        // If no temporal reference is available, add the
                        // action type with a "default" difficulty of 1
                        // (for the effort involved in finding appropriate
                        // time points).

                        objectSelection.addElementAntiproportionally( actionType[ i ] , 1 );

                    }

                }
            }

        }
        else
        {

            // Selecting an action type according to the least number of
            // necessary configuration changes

            // Evaluating action types

            objectSelection.clear();
            long distance;

            for ( int i = 0 ; i < currentActionTypeNumber ; i++ )
            {
                if ( actionTypeOK[ i ] )
                {
                    distance = actionTypeIncludesExternalConfiguration( i );
                    if ( distance != -1 )
                    {
                        distance = distance * distance;
                        if ( distance <= Integer.MAX_VALUE )
                        {
                            objectSelection.addElementAntiproportionally( actionType[ i ] , (int)distance );
                        }
                        else
                        {
                            objectSelection.addElementAntiproportionally( actionType[ i ] , Integer.MAX_VALUE );
                        }
                    }
                }
            }

        }

        // Choosing one of the desired action types. A result value
        // of -1 indicates that the TaskConstraint should be deleted.

        TCActionType desiredActionType
            = (TCActionType)objectSelection.chooseElement();
        if ( desiredActionType != null )
        {
            desiredActionTypeID = desiredActionType.ID;
            desiredTemporalAnchor = desiredActionType.temporalAnchor;
        }
        else
        {
            desiredActionTypeID = -1;
            desiredTemporalAnchor = Long.MIN_VALUE;
        }

    }


    /**
     * Computes the difference lists from the current configuration to
     * the desired configuration.
     */
    private void establishDifferenceLists() {

        differenceRemovalList = new LinkedList();
        differenceAdditionList = new LinkedList();

        TCActionType selectedActionType = actionType[ desiredActionTypeID ];

        // Checking precondition task instances

        Iterator itInstance;
        Iterator itType;
        boolean found;

        TCPreconditionTask preconditionTask;
        TCPreconditionTaskInstance preconditionTaskInstance = null;

        // Initializing instance markings

        itInstance = preconditionTaskSet.iterator();
        while ( itInstance.hasNext() )
        {
            ( (TCPreconditionTaskInstance)itInstance.next() ).found = null;
        }

        // Iterating through action type's precondition tasks

        itType = selectedActionType.preconditionTasks.iterator();

        while ( itType.hasNext() )
        {
            preconditionTask = (TCPreconditionTask)itType.next();

            // Searching for corresponding precondition task instance

            itInstance = preconditionTaskSet.iterator();
            found = false;

            while ( !found && itInstance.hasNext() )
            {
                preconditionTaskInstance
                    = (TCPreconditionTaskInstance)itInstance.next();
                if ( preconditionTaskInstance.found == null )
                {
                    found = ( ( preconditionTaskInstance.resourceType == preconditionTask.resourceType ) && ( preconditionTaskInstance.state == preconditionTask.state ) );
                }
            }

            if ( found )
            {
                preconditionTaskInstance.found = preconditionTask;
            }
            else
            {
                differenceAdditionList.add( preconditionTask );
            }
        }

        // Checking for superfluous tasks

        itInstance = preconditionTaskSet.iterator();
        while ( itInstance.hasNext() )
        {
            preconditionTaskInstance
                = (TCPreconditionTaskInstance)itInstance.next();
            if ( preconditionTaskInstance.found == null )
            {
                differenceRemovalList.add( preconditionTaskInstance );
            }
        }

        // Checking action task instances

        TCActionTask actionTask;
        TCActionTaskInstance actionTaskInstance = null;

        // Initializing instance markings

        itInstance = actionTaskSet.iterator();
        while ( itInstance.hasNext() )
        {
            ( (TCActionTaskInstance)itInstance.next() ).found = null;
        }

        // Iterating through action type's action tasks

        itType = selectedActionType.actionTasks.iterator();

        while ( itType.hasNext() )
        {
            actionTask = (TCActionTask)itType.next();

            // Searching for corresponding action task instance

            itInstance = actionTaskSet.iterator();
            found = false;

            while ( !found && itInstance.hasNext() )
            {
                actionTaskInstance = (TCActionTaskInstance)itInstance.next();
                if ( actionTaskInstance.found == null )
                {
                    found = ( ( actionTaskInstance.resourceType == actionTask.resourceType ) && ( actionTaskInstance.operation == actionTask.operation ) && ( actionTaskInstance.tempValue[ DURATION ] == actionTask.duration ) && ( actionTaskInstance.consumption == actionTask.consumption ) );
                }
            }

            if ( found )
            {
                actionTaskInstance.found = actionTask;
            }
            else
            {
                differenceAdditionList.add( actionTask );
            }
        }

        // Checking for superfluous tasks

        itInstance = actionTaskSet.iterator();
        while ( itInstance.hasNext() )
        {
            actionTaskInstance = (TCActionTaskInstance)itInstance.next();
            if ( actionTaskInstance.found == null )
            {
                differenceRemovalList.add( actionTaskInstance );
            }
        }

        // Checking state task instances

        TCStateTask stateTask;
        TCStateTaskInstance stateTaskInstance = null;

        // Initializing instance markings

        itInstance = stateTaskSet.iterator();
        while ( itInstance.hasNext() )
        {
            ( (TCStateTaskInstance)itInstance.next() ).found = null;
        }

        // Iterating through action type's state tasks

        itType = selectedActionType.stateTasks.iterator();

        while ( itType.hasNext() )
        {
            stateTask = (TCStateTask)itType.next();

            // Searching for corresponding state task instance

            itInstance = stateTaskSet.iterator();
            found = false;

            while ( !found && itInstance.hasNext() )
            {
                stateTaskInstance = (TCStateTaskInstance)itInstance.next();
                if ( stateTaskInstance.found == null )
                {
                    found = ( ( stateTaskInstance.resourceType == stateTask.resourceType ) && ( stateTaskInstance.contribution == stateTask.contribution ) );
                }
            }

            if ( found )
            {
                stateTaskInstance.found = stateTask;
            }
            else
            {
                differenceAdditionList.add( stateTask );
            }
        }

        // Checking for superfluous tasks

        itInstance = stateTaskSet.iterator();
        while ( itInstance.hasNext() )
        {
            stateTaskInstance = (TCStateTaskInstance)itInstance.next();
            if ( stateTaskInstance.found == null )
            {
                differenceRemovalList.add( stateTaskInstance );
            }
        }

    }


    /**
     * Sets the action type variable to the currently matching action
     * type ID. Returns whether the heuristic was able to succeed.
     */
    private boolean heuristic_RepairActionType()
    {
        if ( Registry.LOG_CHANGES )
        {
            System.out.println( "heuristic_RepairActionType: changing action type to \"" + getNameForActionType( matchingActionTypeID ) + "\"" );
        }
        gsc.setVariable_internal( actionTypeVar , matchingActionTypeID );
        return true;
    }


    /**
     * Changes the configuration toward a possible action type. Returns
     * whether the heuristic was able to succeed.
     */
    private boolean heuristic_RepairConfiguration()
    {

        // Updating which action type is desired

        if ( Registry.LOG_CHANGES )
        {
            System.out.print( "heuristic_RepairConfiguration: " );
        }

        if ( desiredActionTypeID == -1 )
        {
            if ( actionTypeGiven )
            {
                if ( actionTypeOK[ actionTypeID ] )
                {
                    desiredActionTypeID = actionTypeID;
                    setDesiredTemporalAnchor( actionType[ actionTypeID ] );
                }
            }
            else
            {
                chooseDesiredActionTypeID();
            }
            if ( desiredActionTypeID == -1 )
            {
                if ( Registry.LOG_CHANGES )
                {
                    System.out.println( "removing action" );
                }
                removeTaskConstraint();
                return true;
            }
            differenceListsEstablished = false;
        }

        if ( desiredActionTypeID == actionTypeID )
        {

            // Determining difference to desired action type if not already
            // done

            if ( !differenceListsEstablished )
            {
                establishDifferenceLists();
            }

            // Changing configuration one step toward desired action type

            int addSize = differenceAdditionList.size();
            int removeSize = differenceRemovalList.size();
            int requiredChanges = addSize + removeSize;
            int choice = gsc.random.nextInt( requiredChanges );

            if ( choice < addSize )
            {

                // Adding a task needed for the desired configuration

                Object task = differenceAdditionList.get( choice );

                if ( task instanceof TCPreconditionTask )
                {
                    addPreconditionTask( (TCPreconditionTask)task );
                }
                else
                {
                    if ( task instanceof TCActionTask )
                    {
                        addActionTask( (TCActionTask)task );
                    }
                    else
                    {
                        addStateTask( (TCStateTask)task );
                    }
                }

            }
            else
            {

                // Removing a superfluous connected task

                Object taskInstance = differenceRemovalList.get( choice
                    - addSize );

                if ( taskInstance instanceof TCPreconditionTaskInstance )
                {
                    if ( Registry.LOG_CHANGES )
                    {
                        System.out.println( "removing precondition task" );
                    }
                    removeTask( ( (TCPreconditionTaskInstance)taskInstance ).constraint , Registry.STATERESOURCE_CONSTRAINT );
                }
                else
                {
                    if ( taskInstance instanceof TCActionTaskInstance )
                    {
                        if ( Registry.LOG_CHANGES )
                        {
                            System.out.println( "removing action task" );
                        }
                        removeTask( ( (TCActionTaskInstance)taskInstance ).constraint , Registry.DISCRETERESOURCE_CONSTRAINT );
                    }
                    else
                    {
                        if ( Registry.LOG_CHANGES )
                        {
                            System.out.println( "removing state task" );
                        }
                        removeTask( ( (TCStateTaskInstance)taskInstance ).constraint , Registry.STATERESOURCE_CONSTRAINT );
                    }
                }

            }

            // Repairing temporal relations if the desired task configuration
            // was established

            if ( requiredChanges == 1 )
            {
                if ( costs[ temporalRelationCostID ] > executionStateCosts )
                {
                    heuristic_RepairAllTemporalRelations();
                }
            }

            return true;

        }
        else
        {

            if ( Registry.LOG_CHANGES )
            {
                System.out.println( "changing action type to \""
                    + getNameForActionType( desiredActionTypeID ) + "\"" );
            }
            gsc.setVariable_internal( actionTypeVar , desiredActionTypeID );
            return true;

        }

    }


    /**
     * Improving the current costs of configuration inconsistencies.
     */
    private void improveConfigurationInconsistency()
    {

        if ( Registry.LOG_CHANGES )
        {
            System.out.println( "improving configuration inconsistency:" );
        }

        boolean actionTypeRepairEnabled
            = ( matchingActionTypeID != -1 ) && !actionTypeGiven;
        boolean configurationRepairEnabled
            = ( matchingActionTypeID == -1 ) && gsc.structuralChangesEnabled;

        if ( actionTypeRepairEnabled )
        {
            heuristic_RepairActionType();
        }
        else
        {
            if ( configurationRepairEnabled )
            {
                heuristic_RepairConfiguration();
            }
        }

    }


    public void improveCosts( int costID )
    {

        if ( Registry.LOG_CHANGES )
        {
            System.out.println( "Improvement by (not yet named) TC:" );
        }

        ownHeuristicChange = true;

        switch ( costID )
        {
        case TEMPORAL_INCONSISTENCY:
        case TRANSITION_TEMPORAL_INCONSISTENCY:
            {
                if ( costs[ CONFIGURATION_INCONSISTENCY ] == 0 )
                {
                    improveTemporalInconsistency();
                }
                else
                {
                    throw new Error( "The configuration inconsistency costs were not connected to a cost function!" );
                }
                break;
            }
        case CONFIGURATION_INCONSISTENCY:
            {
                improveConfigurationInconsistency();
                break;
            }
        default:
            {
                throw new Error( "The cost ID \"" + costID +
                    "\" is not supported by this constraint" );
            }
        }

        ownHeuristicChange = false;

    }


}
