
package com.ai_center.dragonbreath.constraints;


import java.lang.Class;
import java.util.*;

import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.constraints.stateresourceconstraint.*;
import com.ai_center.dragonbreath.constraints.stateresourceconstraint.symbolicsrc.SymSRCEvent;
import com.ai_center.dragonbreath.constraints.stateresourceconstraint.symbolicsrc.SymSRCState;
import com.ai_center.dragonbreath.constraints.transitionconstraint.*;
import com.ai_center.dragonbreath.Registry;
import com.ai_center.dragonbreath.utilities.IntervalList;
import com.ai_center.dragonbreath.utilities.IntervalListElement;


/**
 * Specific state transitions that are imposed by the environment are
 * handled by this constraint. StateResourceConstraints announce relevant
 * states to this constraint and if a specific state configuration is
 * fulfilled, the constraint generates a new ActionConstraint representing
 * the transition consequences. ATTENTION: An addition of transition types
 * after constraint instance creation is not yet supported.
 * The cost component TRANSITION_INCONSISTENCY returns costs if not all
 * occurring transitions are realized or transitions that are no more valid
 * are still realized.
 *
 * @author Alexander Nareyek
 * @version $Revision: 2.11 $
 *
 * @see CurrentTimeObjectConstraint
 * @see StateResourceConstraint
 */
public class TransitionConstraint extends ConventionalConstraint
{


    // **********************
    // ***** ATTRIBUTES *****
    // **********************


    /***** ENGINE-INTEGRATION-RELATED VARIABLES *****/

    /**
     * The constraint type's name.
     */
    private static final String TYPE_NAME
        = new String( "Transition Constraint" );

    /**
     * The identifier for the constraint's transition-related cost component.
     */
    public static final int TRANSITION_INCONSISTENCY = 0;

    /**
     * The constraint's base graph.
     */
    private static final Embedding BASE_GRAPH = new Embedding( 1 );
    /**
     * The constraint's number of extension graphs.
     */
    private static final int EXTENSION_GRAPH_NUMBER = 1;
    /**
     * The constraint's extension graphs.
     */
    private static final Embedding[] EXTENSION_GRAPH
        = new Embedding[ EXTENSION_GRAPH_NUMBER ];


    /***** TYPE-RELATED VARIABLES *****/

    /**
     * The maximal number of transition types that can be registered.
     */
    private final static int MAX_TRANSITION_TYPE_NUMBER = 256;

    /**
     * Number of registered transition types.
     */
    private static int currentTransitionTypeNumber = 0;

    /**
     * Array of available transition types.
     */
    private static TCTransitionType[] transitionType
        = new TCTransitionType[ MAX_TRANSITION_TYPE_NUMBER ];

    /**
     * Contains the actionTypeIDs that are Transitions.
     */
    private static HashSet actionTypeIDHash = new HashSet();

    /**
     * Hashtable from state resources' state objects to the transition
     * constraint's (SymSRCState->TCStateType).
     */
    private static Hashtable stateMapping = new Hashtable();
    /**
     * Hashtable from state resources' event objects to the transition
     * constraint's.
     */
    private static Hashtable eventMapping = new Hashtable();


    /***** GENERAL INTERNAL STRUCTURES *****/

    /**
     * Indicates whether the constraint is currently changing CSP components
     * and the update procedure does not need to be called.
     */
    public boolean modificationModus = false;

    /**
     * The connected current time object constraint.
     */
    public ConventionalConstraint currentTimeConstraint = null;
    /**
     * The internal copy of the current time variable's value.
     */
    public long currentTime = 0;

    /**
     * Temporal interval list with all information about watched states.
     */
    private IntervalList intervalList = null;

    /**
     * Set of transitions to be added.
     */
    private HashSet transitionAdditionSet = new HashSet();
    /**
     * Set of transitions to be removed.
     */
    private HashSet transitionRemovalSet = new HashSet();

    /**
     * Hashtable from actionConstraints to existing transitions.
     */
    private Hashtable existingTransitions = new Hashtable();


    /***** HEURISTICS-RELATED STRUCTURES *****/

    /**
     * The transition that is currently being removed.
     */
    private TCTransition transitionUnderRemoval = null;


    // *******************
    // ***** METHODS *****
    // *******************


    public void create_graphs()
    {

        // Creating the base graph

        BASE_GRAPH.addConnectedNode( Registry.CURRENTTIME_OBJECTCONSTRAINT ,
            Registry.PLAIN_EDGE , Registry.INCOMING );

        // Creating the extension graph

        EXTENSION_GRAPH[ 0 ] = new Embedding( 1 );
        EXTENSION_GRAPH[ 0 ].addConnectedNode( Registry.ACTION_OBJECTCONSTRAINT , Registry.PLAIN_EDGE , Registry.INCOMING );

    }


    public int getNumberOfCosts()
    {
        return 1;
    }


    public String getCostName( int costID )
    {
        if ( costID == 0 )
        {
            return new String( "transition inconsistency" );
        }
        else
        {
            throw new Error( "Invalid cost ID" );
        }
    }


    public String getTypeName()
    {
        return TYPE_NAME;
    }


    public Embedding getBaseGraph()
    {
        return BASE_GRAPH;
    }


    public int getExtensionGraphNumber()
    {
        return EXTENSION_GRAPH_NUMBER;
    }


    public Embedding getExtensionGraph( int graphID )
    {
        return EXTENSION_GRAPH[ graphID ];
    }


    public int getMaximalConnectionNumber()
    {
        // There is no maximum for extensions
        return Integer.MAX_VALUE;
    }


    public void initializeInternalStructures()
    {

        // Initializing interval structure

        TCInterval firstInterval = new TCInterval();
        firstInterval.states = null;
        intervalList = new IntervalList( 0 , Long.MAX_VALUE , firstInterval );

    }


    public void initializeCostType( int costID )
    {
        costFunctionReference[ costID ].updateGlobalCosts();
    }


    /***** REGISTRATION OF TRANSITIONS *****/


    /**
     * Reserves a transition type ID and returns it.
     * @param actionTypeID  The ID of the related TaskConstraint's action type.
     */
    public static int registerNewTransitionType( int actionTypeID )
    {
        if ( currentTransitionTypeNumber == MAX_TRANSITION_TYPE_NUMBER )
        {
            throw new Error( "No more transition type IDs available" );
        }

        actionTypeIDHash.add( new Integer( actionTypeID ) );

        transitionType[ currentTransitionTypeNumber ] = new TCTransitionType ( currentTransitionTypeNumber , actionTypeID );
        return currentTransitionTypeNumber++ ;
    }


    /** Checks a given actionTypeID if it is a Transition or not.<p>

       @param actionTypeID The actionTypeID to check.
       @return True if this is a transition or false if not.
     */
    public static boolean isTransition( int actionTypeID )
    {
        if ( actionTypeIDHash.contains( new Integer( actionTypeID ) ) )
        {
            return true;
        }
        else
        {
            return false;
        }
    }


    /**
     * Adds a precondition task to a transition type.
     * @param transitionTypeID  The ID of the transition type to be extended.
     * @param resourceType  The required resource type for the task.
     * @param state  The required state to be checked for by the task.
     */
    public static void addPreconditionTaskToTransitionType( int transitionTypeID , long resourceType , long state )
    {

        if ( ( transitionTypeID < 0 ) || ( transitionTypeID >= currentTransitionTypeNumber ) )
        {
            throw new Error( "Invalid transition type ID" );
        }

        // Creating state object for precondition

        SymSRCState stateObject
            = SymbolicSRC.retrieveState( resourceType , state );

        TCStateType preconditionTask
            = (TCStateType)stateMapping.get( stateObject );
        if ( preconditionTask == null )
        {
            preconditionTask = new TCStateType( resourceType , state );
            stateMapping.put( stateObject , preconditionTask );
        }

        // Adding state as precondition for transition

        transitionType[ transitionTypeID ].preconditionTasks.add( preconditionTask );
        transitionType[ transitionTypeID ].preconditionTaskNumber++ ;
        preconditionTask.transitionTypes.add( transitionType[ transitionTypeID ] );

    }


    /**
     * Adds a state task to a transition type.
     * @param transitionTypeID  The ID of the transition type to be extended.
     * @param resourceType  The required resource type for the task.
     * @param event  The event produced by the state task.
     */
    public static void addStateTaskToTransitionType( int transitionTypeID , long resourceType , long event )
    {

        if ( ( transitionTypeID < 0 ) || ( transitionTypeID >= currentTransitionTypeNumber ) )
        {
            throw new Error( "Invalid transition type ID" );
        }

        // Creating contribution object for state change

        SymSRCEvent eventObject
            = SymbolicSRC.retrieveContribution( resourceType , event );

        TCEventType stateTask = (TCEventType)eventMapping.get( eventObject );
        if ( stateTask == null )
        {
            stateTask = new TCEventType( resourceType , event );
            eventMapping.put( eventObject , stateTask );
        }

        // Adding event as effect of transition

        transitionType[ transitionTypeID ].stateTasks.add( stateTask );
        transitionType[ transitionTypeID ].stateTaskNumber++ ;

    }


    /**
     * Retrieves a list of states to be watched for a specific state
     * resource type.
     * @param resourceType  The states' state resource type.
     */
    public static HashSet getStateWatchList( long resourceType )
    {

        HashSet stateIDList = new HashSet();
        Iterator it;
        TCStateType preconditionTask;

        for ( int i = 0 ; i < currentTransitionTypeNumber ; i++ )
        {
            it = transitionType[ i ].preconditionTasks.iterator();
            while ( it.hasNext() )
            {
                preconditionTask = (TCStateType)it.next();
                if ( preconditionTask.resourceType == resourceType )
                {
                    stateIDList.add( new Long( preconditionTask.state ) );
                }
            }
        }

        return stateIDList;

    }


    /***** INTERVAL LIST *****/


    /**
     * Returns a description of the state projection.
     */
    public StringBuffer getStateProjectionDescription()
    {

        StringBuffer s = new StringBuffer( "World.StateWatch: " );
        IntervalListElement currentILE = intervalList.getFirst();
        TCInterval interval;

        s.append( " [" + currentILE.getBeginning() + "]" );

        while ( currentILE != null )
        {
            interval = (TCInterval)currentILE.getElement();
            if ( interval.states != null )
            {
                s.append( " " + interval.toString()
                         + " [" + currentILE.getEnd() + "]" );
            }
            else
            {
                s.append( " {} [" + currentILE.getEnd() + "]" );
            }
            currentILE = currentILE.getNext();
        }

        return s;

    }


    // ***** STATE WATCH *****


    /**
     * Handles a StateResourceConstraint's notification that a watched
     * state occurs.
     * @param src  The state resource constraint that is announcing the state.
     * @param state  The state that occurred.
     * @param beginning  The beginning of the occurrence.
     * @param end  The end of the occurrence.
     */
    public void registerStateOccurrence( StateResourceConstraint src , SymSRCState state , long beginning , long end )
    {

        long oldTRCosts = costs[ TRANSITION_INCONSISTENCY ];
        TCStateType newStateType = (TCStateType)stateMapping.get( state );
        TCState newState = new TCState( src , newStateType );

        // Splitting interval at occurence's begin point if necessary

        IntervalListElement firstILE = intervalList.addBasePoint( beginning );

        // If interval was splitted, copying intervals for new list element
        // and updating structures accordingly

        if ( firstILE.getElement() == null )
        {

            // Updating interval information

            firstILE.setElement( (TCInterval)((TCInterval)firstILE.getPrevious().getElement()).clone() );

        }
        else
        {

            // Shifting firstILE such that firstILE starts with the
            // occurrence

            if ( firstILE.getEnd() <= beginning )
            {
                firstILE = firstILE.getNext();
            }

        }

        // Splitting interval at occurrence's end point if necessary

        IntervalListElement lastILE
            = intervalList.addBasePointWithOffset( end , firstILE );
        IntervalListElement lastAdditionILE = lastILE;

        // If interval was splitted, copying intervals for new list element
        // and updating structures accordingly

        if ( lastILE.getElement() == null )
        {

            // Updating interval information

            lastILE.setElement( (TCInterval)((TCInterval)lastILE.getPrevious().getElement()).clone() );
            lastILE = lastILE.getPrevious();

        }
        else
        {

            // Shifting lastAdditionILE such that lastAdditionILE is after
            // the occurrence

            if ( end > lastILE.getBeginning() )
            {
                lastAdditionILE = lastILE.getNext();
            }

        }

        // Adding occurrence information to interval list elements

        IntervalListElement currentILE = firstILE;
        TCInterval interval;

        while ( currentILE != lastAdditionILE )
        {
            interval = (TCInterval)currentILE.getElement();
            if ( interval.states == null )
            {
                interval.states = new HashSet();
            }
            interval.states.add( newState );
            if ( currentILE != firstILE )
            {
                reevaluateForStateAddition( currentILE , newStateType );
            }
            currentILE = currentILE.getNext();
        }

        // Adding begin state information for first interval

        boolean modifyOK = true;

        if ( firstILE.getPrevious() != null )
        {
            interval = (TCInterval)firstILE.getPrevious().getElement();
            if ( ( interval.states != null )
                && ( interval.states.contains( newState ) ) )
            {
                modifyOK = false;
            }
        }

        if ( modifyOK )
        {
            interval = (TCInterval)firstILE.getElement();
            if ( interval.newStates == null )
            {
                interval.newStates = new HashSet();
            }
            interval.newStates.add( newState );
            reevaluateForNewStateAddition( firstILE , newStateType );
        }
        else
        {
            reevaluateForStateAddition( firstILE , newStateType );
        }

        // Possibly removing begin state information from successor

        modifyOK = false;

        if ( lastILE.getNext() != null )
        {
            interval = (TCInterval)lastILE.getNext().getElement();
            if ( ( interval.newStates != null )
                && ( interval.newStates.contains( newState ) ) )
            {
                modifyOK = true;
            }
        }

        if ( modifyOK )
        {
            interval = (TCInterval)lastILE.getNext().getElement();
            interval.newStates.remove( newState );
            if ( interval.newStates.isEmpty() )
            {
                interval.newStates = null;
            }
            reevaluateForStateRemoval( lastILE.getNext() , newStateType );
        }

        // Possibly merging first and/or last interval

        possiblyMergeNeighborIntervals( firstILE , lastILE );

        // Propagation of cost changes

        if ( oldTRCosts != costs[ TRANSITION_INCONSISTENCY ] )
        {
            costFunctionReference[ TRANSITION_INCONSISTENCY ].updateGlobalCosts();
        }

    }


    /**
     * Handles a StateResourceConstraint's notification that a watched
     * state does not occur anymore.
     * @param src  The state resource constraint that is announcing the
     *     state loss.
     * @param state  The state that occurred.
     * @param beginning  The beginning of the removed occurrence.
     * @param end  The end of the removed occurrence.
     */
    public void registerStateLoss( StateResourceConstraint src , SymSRCState state , long beginning , long end )
    {

        long oldTRCosts = costs[ TRANSITION_INCONSISTENCY ];
        TCStateType oldState = (TCStateType)stateMapping.get( state );

        // Splitting interval at loss' begin point if necessary

        IntervalListElement firstILE = intervalList.addBasePoint( beginning );

        // If interval was splitted, copying intervals for new list element
        // and updating structures accordingly

        if ( firstILE.getElement() == null )
        {

            // Updating interval information

            firstILE.setElement( (TCInterval)( (TCInterval)firstILE.getPrevious().getElement() ).clone() );

        }
        else
        {

            // Shifting firstILE such that firstILE starts with the loss

            if ( firstILE.getEnd() <= beginning )
            {
                firstILE = firstILE.getNext();
            }

        }

        // Splitting interval at loss' end point if necessary

        IntervalListElement lastILE
            = intervalList.addBasePointWithOffset( end , firstILE );
        IntervalListElement lastAdditionILE = lastILE;

        // If interval was splitted, copying intervals for new list element
        // and updating structures accordingly

        if ( lastILE.getElement() == null )
        {

            // Updating interval information

            lastILE.setElement( (TCInterval)( (TCInterval)lastILE.getPrevious().getElement() ).clone() );
            lastILE = lastILE.getPrevious();

        }
        else
        {

            // Shifting lastAdditionILE such that lastAdditionILE is after
            // the loss

            if ( end > lastILE.getBeginning() )
            {
                lastAdditionILE = lastILE.getNext();
            }

        }

        // Removing state occurrence information from interval list elements

        IntervalListElement currentILE = firstILE;
        TCInterval interval;

        while ( currentILE != lastAdditionILE )
        {
            interval = (TCInterval)currentILE.getElement();
            interval.states.remove( oldState );
            if ( interval.states.isEmpty() )
            {
                interval.states = null;
            }
            reevaluateForStateRemoval( currentILE , oldState );
            currentILE = currentILE.getNext();
        }

        // Removing begin state information

        interval = (TCInterval)firstILE.getElement();
        if ( interval.newStates != null )
        {
            if ( interval.newStates.remove( oldState ) )
            {
                if ( interval.newStates.isEmpty() )
                {
                    interval.newStates = null;
                }
            }
        }

        // Possibly adding begin state information to successor

        boolean modifyOK = false;

        if ( lastILE.getNext() != null )
        {
            interval = (TCInterval)lastILE.getNext().getElement();
            if ( (interval.states != null) && ( interval.states.contains( oldState ) ) )
            {
                modifyOK = true;
            }
        }

        if ( modifyOK ) {
            interval = (TCInterval)lastILE.getNext().getElement();
            if ( interval.newStates == null )
            {
                interval.newStates = new HashSet();
            }
            interval.newStates.add( oldState );
            reevaluateForNewStateAddition( lastILE.getNext() , oldState );
        }

        // Possibly merging first and/or last interval

        possiblyMergeNeighborIntervals( firstILE , lastILE );

        // Propagation cost changes

        if ( oldTRCosts != costs[ TRANSITION_INCONSISTENCY ] )
        {
            costFunctionReference[ TRANSITION_INCONSISTENCY ].updateGlobalCosts();
        }

    }


    /**
     * Merges the first interval with its predecessor and/or the last
     * interval with its successor if they are equal with respect to
     * the included states.
     * @param firstILE  The first interval of the change.
     * @param lastILE  The last interval of the change.
     */
    private void possiblyMergeNeighborIntervals( IntervalListElement firstILE , IntervalListElement lastILE )
    {

        IntervalListElement otherILE;
        TCInterval otherInterval;

        // Possibly merging first interval

        TCInterval firstInterval = (TCInterval)firstILE.getElement();
        otherILE = firstILE.getPrevious();
        if ( otherILE != null )
        {
            otherInterval = (TCInterval)otherILE.getElement();
            if ( ( ( firstInterval.states == null )
                && ( otherInterval.states == null ) )
                || ( ( firstInterval.states != null )
                && ( firstInterval.states.equals( otherInterval.states ) ) ) )
            {
                intervalList.removeBasePoint( otherILE.getEnd() , otherILE );
                if ( firstILE == lastILE )
                {
                    lastILE = otherILE;
                }
                firstILE = otherILE;
            }
        }

        // Possibly merging last interval

        TCInterval lastInterval = (TCInterval)lastILE.getElement();
        otherILE = lastILE.getNext();
        if ( otherILE != null )
        {
            otherInterval = (TCInterval)otherILE.getElement();
            if ( ( ( lastInterval.states == null )
                && ( otherInterval.states == null ) )
                || ( ( lastInterval.states != null )
                && ( lastInterval.states.equals(otherInterval.states ) ) ) )
            {
                intervalList.removeBasePoint( lastILE.getEnd() , lastILE );
            }
        }

    }


    /**
     * Checks an interval for transition changes if a new state has
     * been added (the occurrence of the state beginning in a predecessor
     * interval).
     * @param ile  The affected interval.
     * @param state  The added state.
     */
    private void reevaluateForStateAddition( IntervalListElement ile , TCStateType state )
    {

        TCInterval interval = (TCInterval)ile.getElement();

        if ( interval.newStates != null )
        {

            int stateNumber = interval.states.size();
            HashSet transitionCandidates
                = (HashSet)state.transitionTypes.clone();

            // Iteration through transition candidates

            Iterator itTransitionType = transitionCandidates.iterator();
            TCTransitionType transitionType;
            Iterator itState;
            TCStateType typeState;
            boolean correct;
            boolean newStateIncluded;
            Iterator itTransition;
            boolean notFound;

            while ( itTransitionType.hasNext() )
            {

                transitionType = (TCTransitionType)itTransitionType.next();

                if ( transitionType.preconditionTaskNumber <= stateNumber )
                {

                    // Iterating through transition type's states

                    itState = transitionType.preconditionTasks.iterator();
                    correct = true;
                    newStateIncluded = false;

                    while ( correct && itState.hasNext() )
                    {
                        typeState = (TCStateType)itState.next();
                        if ( !interval.states.contains( typeState ) )
                        {
                            correct = false;
                        }
                        else
                        {
                            if ( !newStateIncluded )
                            {
                                newStateIncluded
                                    = interval.newStates.contains( typeState );
                            }
                        }
                    }

                    // Potentially adding transition

                    if ( correct && newStateIncluded )
                    {

                        // Checking if already in removal set

                        TCTransition transition = null;
                        itTransition = transitionRemovalSet.iterator();
                        notFound = true;
                        while ( notFound && itTransition.hasNext() )
                        {
                            transition = (TCTransition)itTransition.next();
                            if ( ( transition.ile.getBeginning() == ile.getBeginning() ) && ( transition.transitionType == transitionType ) )
                            {
                                notFound = false;
                            }
                        }

                        // Adding transition

                        if ( interval.transitions == null )
                        {
                            interval.transitions = new HashSet();
                        }
                        if ( notFound )
                        {
                            transition
                                = new TCTransition( transitionType , ile );
                            transitionAdditionSet.add( transition );
                            costs[ TRANSITION_INCONSISTENCY ]++ ;
                        }
                        else
                        {
                            transitionRemovalSet.remove( transition );
                            transition.ile = ile;
                            costs[ TRANSITION_INCONSISTENCY ]-- ;
                        }
                        interval.transitions.add( transition );

                    }

                }
            }
        }

    }


    /**
     * Checks an interval for transition changes if a new state has
     * been added (the occurrence of the state beginning with this
     * interval).
     * @param ile  The affected interval.
     * @param state  The added state.
     */
    private void reevaluateForNewStateAddition( IntervalListElement ile , TCStateType state )
    {

        TCInterval interval = (TCInterval)ile.getElement();

        int stateNumber = interval.states.size();
        HashSet transitionCandidates = (HashSet)state.transitionTypes.clone();

        // Iteration through transition candidates

        Iterator itTransitionType = transitionCandidates.iterator();
        TCTransitionType transitionType;
        Iterator itState;
        TCStateType typeState;
        boolean correct;
        Iterator itTransition;
        boolean notFound;

        while ( itTransitionType.hasNext() )
        {

            transitionType = (TCTransitionType)itTransitionType.next();

            if ( transitionType.preconditionTaskNumber <= stateNumber )
            {

                // Iterating through transition type's states

                itState = transitionType.preconditionTasks.iterator();
                correct = true;

                while ( correct && itState.hasNext() )
                {
                    correct
                        = interval.states.contains( (TCStateType)itState.next() );
                }

                // Potentially adding transition

                if ( correct )
                {

                    // Checking if already in removal set

                    TCTransition transition = null;
                    itTransition = transitionRemovalSet.iterator();
                    notFound = true;
                    while ( notFound && itTransition.hasNext() )
                    {
                        transition = (TCTransition)itTransition.next();
                        if ( ( transition.ile.getBeginning() == ile.getBeginning() ) && ( transition.transitionType == transitionType ) )
                        {
                            notFound = false;
                        }
                    }

                    // Adding transition

                    if ( interval.transitions == null )
                    {
                        interval.transitions = new HashSet();
                    }
                    if ( notFound )
                    {
                        transition = new TCTransition( transitionType , ile );
                        transitionAdditionSet.add( transition );
                        costs[ TRANSITION_INCONSISTENCY ]++ ;
                    }
                    else
                    {
                        transitionRemovalSet.remove( transition );
                        transition.ile = ile;
                        costs[ TRANSITION_INCONSISTENCY ]-- ;
                    }
                    interval.transitions.add( transition );

                }

            }
        }

    }


    /**
     * Checks an interval for transition changes if a state has
     * been removed.
     * @param ile  The affected interval.
     * @param state  The removed state.
     */
    private void reevaluateForStateRemoval( IntervalListElement ile , TCStateType state )
    {

        TCInterval interval = (TCInterval)ile.getElement();

        if ( interval.transitions != null )
        {

            // Iterating through existing transitions

            Iterator it = interval.transitions.iterator();
            TCTransition transition;

            while ( it.hasNext() ) {
                transition = (TCTransition)it.next();
                if ( transition.transitionType.preconditionTasks.contains( state ) )
                {
                    if ( transition.actionConstraint != null )
                    {
                        transitionRemovalSet.add( transition );
                        costs[ TRANSITION_INCONSISTENCY ]++ ;
                    } else {
                        transitionAdditionSet.remove( transition );
                        costs[ TRANSITION_INCONSISTENCY ]-- ;
                    }
                    it.remove();
                }
            }

            if ( interval.transitions.isEmpty() )
            {
                interval.transitions = null;
            }

        }

    }


    // ***** EXTERNAL DYNAMICS *****


    public void update( double oldValue , ConnectionReference first , ConnectionReference last )
    {

        if ( !modificationModus
            && ( ( (Variable)first.node).value() != oldValue ) )
        {

            if ( last.node instanceof CurrentTimeObjectConstraint )
            {

                // Current time variable was updated

                currentTime = (long)((Variable)first.node).value();

                // Checking for a valid current time value

                if ( currentTime < 0 )
                {
                    throw new Error( "Invalid value \"" + currentTime
                        + "\" for current time" );
                }

            }
        }

    }


    public void connectNode( ConnectionReference first , ConnectionReference last )
    {

        if ( last.node instanceof CurrentTimeObjectConstraint )
        {

            // Current time variable/constraint was connected

            currentTimeConstraint = (ConventionalConstraint)last.node;

            // Fetching current time's variable

            Variable variable;
            synchronized( currentTimeConstraint.connections )
            {
                variable = currentTimeConstraint.connections.getSpecificVariable( Registry.PLAIN_EDGE );
            }
            if ( variable == null )
            {
                throw new Error( "The connected CurrentTimeObjectConstraint \"" + currentTimeConstraint + "\" has no variable" );
            }

            currentTime = (long)variable.value();

            // Checking for a valid current time value

            if ( currentTime < 0 ) {
                throw new Error( "Invalid value \"" + currentTime
                    + "\" for current time" );
            }

        }

    }


    public void disconnectNode( ConnectionReference first , ConnectionReference last )
    {

        if ( first.node instanceof ActionObjectConstraint )
        {

            // Action of a transition was disconnected

            TCTransition transition
                = (TCTransition)existingTransitions.get( first.node );
            if ( transition != null )
            {

                existingTransitions.remove( transition.actionConstraint );

                if ( transitionUnderRemoval == transition )
                {

                    // Removal was initiated by the constraint itself

                    costs[ TRANSITION_INCONSISTENCY ]-- ;

                }
                else
                {
                    if ( transitionRemovalSet.contains( transition ) )
                    {

                        // Updating structures for this supportive removal

                        transitionRemovalSet.remove( transition );
                        costs[ TRANSITION_INCONSISTENCY ]-- ;

                    }
                    else
                    {

                        // Initiating rebuild if the transition is still valid

                        transition.actionConstraint = null;
                        transitionAdditionSet.add( transition );
                        costs[ TRANSITION_INCONSISTENCY ]++ ;

                    }
                }
            }
            else
            {
                throw new Error( "The action object constraint \""
                    + first.node + "\" has no associated transition" );
            }

            // Cost propagation

            costFunctionReference[ TRANSITION_INCONSISTENCY ].updateGlobalCosts();

        }

    }


    // ***** IMPROVEMENT (Service methods) *****


    // ***** DELEGATION PROCEDURES *****


    /**
     * Deletes an event. Returns whether the heuristic was able
     * to succeed.
     */
    public boolean delegatedHeuristic_DeleteEvent( ActionObjectConstraint taskConstraint )
    {

        if ( Registry.LOG_CHANGES )
        {
            System.out.println( "Improvement by TransitionConstraint:" );
            System.out.print( "delegatedHeuristic_DeleteEvent: " );
        }

        // Retrieving corresponding transition

        TCTransition transition
            = (TCTransition)existingTransitions.get( taskConstraint );
        if ( transition == null )
        {
            throw new Error( "The action object constraint \""
                + taskConstraint + "\" has no associated transition" );
        }

        // Randomly choosing one of the states that triggered the
        // precondition

        long chosenStateID = Math.abs( gsc.random.nextLong() )
            % transition.transitionType.preconditionTaskNumber;
        Iterator it = transition.transitionType.preconditionTasks.iterator();
        TCStateType state = (TCStateType)it.next();
        while ( chosenStateID != 0 )
        {
            state = (TCStateType)it.next();
            chosenStateID-- ;
        }

        // Searching backward in time until the chosen state has been
        // reached

        IntervalListElement currentILE = transition.ile;
        TCInterval interval = (TCInterval)currentILE.getElement();
        while ( !interval.newStates.contains(state) )
        {
            currentILE = currentILE.getPrevious();
            interval = (TCInterval)currentILE.getElement();
        }

        // Notifying state resource to delete the

        if ( Registry.LOG_CHANGES )
        {
            System.out.println( "delegated deletion of " + state + " @"
                + currentILE.getBeginning()
                + " to state resource constraint" );
        }

        return false;
//xxxx

    }


    // ***** TRANSITION INCONSISTENCY IMPROVEMENT (Heuristics) *****


    /**
     * Adds a required transition. Returns whether the heuristic was able
     * to succeed.
     */
    private boolean heuristic_AddTransition( TCTransition transition )
    {

        TCTransitionType transitionType = transition.transitionType;

        // Creating one of the precondition tasks for temporal guidance

        Iterator it = transitionType.preconditionTasks.iterator();
        TCStateType state = (TCStateType)it.next();

        Variable v1 = gsc.createVariable_internal( "Temporal Reference" );
        gsc.setVariable_internal (v1 , transition.ile.getBeginning() );
        Variable v2 = gsc.createVariable_internal( "State" );
        gsc.setVariable_internal( v2 , state.state );
        ConventionalConstraint src =
            gsc.findConstraint( Registry.STATERESOURCE_CONSTRAINT ,
                Registry.RESOURCETYPE_EDGE , state.resourceType );
        Variable v3;
        synchronized( src.connections )
        {
            v3 = src.connections.getSpecificVariable( Registry.RESOURCETYPE_EDGE );
        }

        Embedding e = new Embedding( 3 );
        gsc.extendEmbedding_internal( e , v1 , Registry.TEMPORALREFERENCE_EDGE , Registry.INCOMING );
        gsc.extendEmbedding_internal( e , v2 , Registry.STATE_EDGE , Registry.INCOMING );
        gsc.extendEmbedding_internal( e , v3 , Registry.RESOURCETYPE_EDGE , Registry.INCOMING );
        ConventionalConstraint task = gsc.createConstraint_internal( Registry.PRECONDITIONTASK_OBJECTCONSTRAINT , e );

        // Creating the action object

        v1 = gsc.createVariable_internal( "Action Type" );
        gsc.setVariable_internal( v1 , transitionType.actionTypeID );

        e = new Embedding( 1 );
        gsc.extendEmbedding_internal( e , v1 , Registry.ACTIONTYPE_EDGE , Registry.INCOMING );
        ConventionalConstraint actionConstraint = gsc.createConstraint_internal( Registry.ACTION_OBJECTCONSTRAINT , e );

        // Connecting the task to the action object

        Embedding eTask = new Embedding( 1 );
        gsc.extendEmbedding_internal( eTask , task , Registry.PLAIN_EDGE , Registry.INCOMING );
        gsc.extendConstraint_internal( actionConstraint , eTask );

        // Connecting the action constraint to this constraint

        e = new Embedding( 1 );
        gsc.extendEmbedding_internal( e , actionConstraint , Registry.PLAIN_EDGE , Registry.INCOMING );
        gsc.extendConstraint_internal( this , e );

        // Connecting the task to the state resource constraint

        gsc.extendConstraint_internal( src , eTask );

        // Creating the task constraint

        e = new Embedding( 2 );
        gsc.extendEmbedding_internal( e , currentTimeConstraint , Registry.PLAIN_EDGE , Registry.INCOMING );
        gsc.extendEmbedding_internal( e , actionConstraint , Registry.PLAIN_EDGE , Registry.INCOMING );

        // TO BE IMPROVED: Cost function selection should be decided
        // by the user!

        gsc.selectCostFunction_internal( gsc.getStructuralCostFunction() );
        ConventionalConstraint taskConstraint = gsc.createConstraint_internal( Registry.TASK_CONSTRAINT , e );
        gsc.selectCostFunction_internal( gsc.getDefaultCostFunction() );
        gsc.addCosts( taskConstraint , TaskConstraint.TEMPORAL_INCONSISTENCY );

        // Updating internal structures

        transition.actionConstraint = actionConstraint;
        transitionAdditionSet.remove( transition );
        existingTransitions.put( actionConstraint , transition );

        if ( Registry.LOG_CHANGES )
        {
            System.out.println( "heuristic_AddTransition: adding transition " + transition );
        }

        // Updating costs

        costs[ TRANSITION_INCONSISTENCY ]-- ;
        costFunctionReference[ TRANSITION_INCONSISTENCY ].updateGlobalCosts();
        fireChange( actionConstraint );
        return true;

    }


    /**
     * Removes an invalid transition. Returns whether the heuristic was
     * able to succeed.
     */
    private boolean heuristic_RemoveTransition( TCTransition transition )
    {

        // Retrieving the task constraint

        ConventionalConstraint taskConstraint;
        synchronized( transition.actionConstraint )
        {
           taskConstraint = transition.actionConstraint.connections.getSpecificConstraint( Registry.TASK_CONSTRAINT , Registry.PLAIN_EDGE );
        }

        // Executing the removal

        if ( Registry.LOG_CHANGES )
        {
            System.out.println( "heuristic_RemoveTransition: removing transition " + transition );
        }

        transitionUnderRemoval = transition;
        transitionRemovalSet.remove( transition );
        ((TaskConstraint)taskConstraint).removeTaskConstraint();
        transitionUnderRemoval = null;
        fireChange( transition.actionConstraint );
        return true;

    }


    /**
     * Improves the current transition inconsistency by adding/removing
     * transitions.
     */
    private void improveTransitionInconsistency()
    {

        if ( Registry.LOG_CHANGES )
        {
            System.out.println( "improving transition inconsistency: " );
        }

        // Selecting transition to be repaired

        long chosenCosts = Math.abs( gsc.random.nextLong() )
            % costs[ TRANSITION_INCONSISTENCY ];

        int additions = transitionAdditionSet.size();

        Iterator it;
        TCTransition transition = null;
        boolean success;

        if ( chosenCosts >= additions )
        {
            chosenCosts -= additions;
            it = transitionRemovalSet.iterator();
            while ( chosenCosts-- >= 0 )
            {
                transition = (TCTransition)it.next();
            }
            success = heuristic_RemoveTransition( transition );
        }
        else
        {
            it = transitionAdditionSet.iterator();
            while ( chosenCosts-- >= 0 )
            {
                transition = (TCTransition)it.next();
            }
            success = heuristic_AddTransition( transition );
        }

    }


    public void improveCosts( int costID )
    {

        if ( Registry.LOG_CHANGES )
        {
            System.out.println( "Improvement by TransitionConstraint:" );
        }

        if ( costID == TRANSITION_INCONSISTENCY )
        {
            if ( gsc.structuralChangesEnabled )
            {
                improveTransitionInconsistency();
            }
        }
        else
        {
            throw new Error( "The cost ID \"" + costID +
                "\" is not supported by this constraint" );
        }

    }

}
