
package com.ai_center.dragonbreath.constraints.stateresourceconstraint;


import java.util.*;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.EventListenerList;

import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.constraints.PreconditionTaskObjectConstraint;
import com.ai_center.dragonbreath.constraints.StateResourceConstraint;
import com.ai_center.dragonbreath.constraints.stateresourceconstraint.numericalheapsrc.*;
import com.ai_center.dragonbreath.constraints.StateTaskObjectConstraint;
import com.ai_center.dragonbreath.GlobalSearchControl;
import com.ai_center.dragonbreath.Registry;
import com.ai_center.dragonbreath.utilities.OrderedShiftDLLElement;


/**
 * This constraint serves as a sub-class for a state resource constraint.
 * Only one numerical value is maintained, which is not projected over time
 * and influenced by all connected state tasks. ATTENTION: A connection
 * of precondition tasks is currently not yet supported. Only integer
 * contributions are supported by now.
 *
 * @author Alexander Nareyek
 * @version $Revision: 1.10 $
 *
 * @see StateResourceConstraint
 */
public class NumericalHeapSRC implements SRCInterface
{


    // **********************
    // ***** ATTRIBUTES *****
    // **********************


    /***** GENERAL INTERNAL STRUCTURES *****/

    private EventListenerList listenerList = new EventListenerList();

    /**
     * The identifier for the constraint's heap costs.
     */
    private static final int HEAP_COSTS
        = StateResourceConstraint.PRECONDITION_FAILURE;

    /**
     * The heap value, summing all contributions and the state value.
     */
    private double totalHeap = 0;

    /**
     * The sum of all positive contributions.
     */
    private double positiveContributions = 0;

    /**
     * The unspecific state resource constraint parent class.
     */
    private StateResourceConstraint src = null;

    /**
     * The GlobalSearchControl solver object.
     */
    private GlobalSearchControl gsc = null;


    /***** TASK MANAGEMENT *****/

    /**
     * The list of NumHSRCStateTasks ordered according to their heap
     * cost contribution.
     */
    private OrderedHeapCostsTaskList taskList = null;


    // *******************
    // ***** METHODS *****
    // *******************


    /***** USER'S REGISTRATION METHOD *****/


    /**
     * Reserves a resource type ID and returns it.
     */
    public static int registerNewResourceType( String name )
    {
        SRCResourceType resourceType = new SRCResourceType();
        resourceType.name = name;
        return StateResourceConstraint.registerNewResourceType( StateResourceConstraint.NUMERICALHEAP_SRC , resourceType );
    }


    /***** GENERAL SRC-INTERFACE-FUNCTIONS *****/


    public void initialize( StateResourceConstraint src , GlobalSearchControl gsc , SRCResourceType resourceType )
    {

        this.src = src;
        this.gsc = gsc;

        // Initializing cost structures

        taskList = new OrderedHeapCostsTaskList();

        totalHeap = src.state;

        long positiveHeap = Math.round( totalHeap );
        if ( positiveHeap < 0 )
        {
            positiveHeap = 0;
        }
        if ( positiveHeap != src.costs[ HEAP_COSTS ] )
        {
            src.costs[ HEAP_COSTS ] = positiveHeap;
            if ( src.costFunctionReference[ HEAP_COSTS ] != null )
            {
                src.costFunctionReference[ HEAP_COSTS ].updateGlobalCosts();
            }
        }

    }


    public void updateCurrentTime( long oldTime )
    {
        // No need for updates
    }


    public void updateCurrentState( double oldValue )
    {

        // Updating costs

        totalHeap = totalHeap - oldValue + src.state;

        long positiveHeap = Math.round( totalHeap );
        if ( positiveHeap < 0 )
        {
            positiveHeap = 0;
        }
        if ( positiveHeap != src.costs[ HEAP_COSTS ] )
        {
            src.costs[ HEAP_COSTS ] = positiveHeap;
            if ( src.costFunctionReference[ HEAP_COSTS ] != null )
            {
                src.costFunctionReference[ HEAP_COSTS ].updateGlobalCosts();
            }
        }
        fireChange( this );
    }


    public SRCStateTask includeStateTask( StateTaskObjectConstraint
        taskConstraint )
    {

        NumHSRCStateTask task = new NumHSRCStateTask();

        // Fetching state task's contribution variable

        Variable tcContributionVariable;

        synchronized( taskConstraint.connections )
        {
            tcContributionVariable = taskConstraint.connections.getSpecificVariable( Registry.CONTRIBUTION_EDGE );
        }
        if ( tcContributionVariable == null )
        {
            throw new Error( "The connected StateTaskObjectConstraint \"" + taskConstraint + "\" has no state variable" );
        }

        task.contributionVariable = tcContributionVariable;
        task.contribution = tcContributionVariable.value();

        // Updating cost structures

        task.taskLE = taskList.addListElement( task );

        if ( task.contribution > 0 )
        {
            positiveContributions = positiveContributions + task.contribution;
        }
        totalHeap = totalHeap + task.contribution;

        long positiveHeap = Math.round( totalHeap );
        if ( positiveHeap < 0 )
        {
            positiveHeap = 0;
        }
        if ( positiveHeap != src.costs[ HEAP_COSTS ] )
        {
            src.costs[ HEAP_COSTS ] = positiveHeap;
            if ( src.costFunctionReference[ HEAP_COSTS ] != null )
            {
                src.costFunctionReference[ HEAP_COSTS ].updateGlobalCosts();
            }
        }
        fireChange( this );
        fireChange( task );
        return task;

    }


    public SRCPreconditionTask includePreconditionTask( PreconditionTaskObjectConstraint taskConstraint )
    {
        throw new Error( "Precondition task handling is not yet supported" );
    }


    public void updateStateTask( SRCStateTask task , int edgeID )
    {

        NumHSRCStateTask specializedTask = (NumHSRCStateTask)task;

        if ( edgeID == Registry.CONTRIBUTION_EDGE )
        {

            if ( specializedTask.contribution > 0 )
            {
                positiveContributions
                    = positiveContributions - specializedTask.contribution;
            }
            totalHeap = totalHeap - specializedTask.contribution;

            // Updating contribution variable

            specializedTask.contribution
                = specializedTask.contributionVariable.value();

            // Updating cost structures

            taskList.updateListElementPosition( specializedTask.taskLE );

            if ( specializedTask.contribution > 0 )
            {
                positiveContributions
                    = positiveContributions + specializedTask.contribution;
            }
            totalHeap = totalHeap + specializedTask.contribution;

            long positiveHeap = Math.round( totalHeap );
            if ( positiveHeap < 0 )
            {
                positiveHeap = 0;
            }
            if ( positiveHeap != src.costs[ HEAP_COSTS ] )
            {
                src.costs[ HEAP_COSTS ] = positiveHeap;
                if ( src.costFunctionReference[ HEAP_COSTS ] != null )
                {
                    src.costFunctionReference[ HEAP_COSTS ].updateGlobalCosts();
                }
            }


        }
        fireChange( this );
        fireChange( specializedTask );
    }


    public void updatePreconditionTask( SRCPreconditionTask task , int edgeID )
    {
        throw new Error( "Precondition task handling is not supported yet" );
    }


    public void excludeStateTask( SRCStateTask task )
    {

        NumHSRCStateTask specializedTask = (NumHSRCStateTask)task;

        // Updating cost structures

        taskList.removeListElement( specializedTask.taskLE );

        if ( specializedTask.contribution > 0 )
        {
            positiveContributions
                = positiveContributions - specializedTask.contribution;
        }
        totalHeap = totalHeap - specializedTask.contribution;

        long positiveHeap = Math.round( totalHeap );
        if ( positiveHeap < 0 )
        {
            positiveHeap = 0;
        }
        if ( positiveHeap != src.costs[ HEAP_COSTS ] )
        {
            src.costs[ HEAP_COSTS ] = positiveHeap;
            if ( src.costFunctionReference[ HEAP_COSTS ] != null )
            {
                src.costFunctionReference[ HEAP_COSTS ].updateGlobalCosts();
            }
        }
        fireChange( this );
        fireChange( specializedTask );
    }


    public void excludePreconditionTask( SRCPreconditionTask task )
    {
        throw new Error( "Precondition task handling is not supported yet" );
    }


    public double getStateAtTime( long time )
    {
        return totalHeap;
    }


    public int getDifficultyForStateAtTime( double stateValue , long time )
    {
        return Math.abs( (int)( totalHeap - stateValue ) );
    }


    public int getDifficultyForContributionAtTime( double contributionValue ,
        long time )
    {
        // Precondition task handling is not yet supported
        return 0;
    }


    public void terminate()
    {

        // Removing costs induced by the state variable

        if ( src.costs[ HEAP_COSTS ] != 0 )
        {
            src.costs[ HEAP_COSTS ] = 0;
            if ( src.costFunctionReference[ HEAP_COSTS ] != null )
            {
                src.costFunctionReference[ HEAP_COSTS ].updateGlobalCosts();
            }
        }

    }


    /***** INFO *****/


    /**
     * Returns a description of the resource's total heap.
     */
    public String getHeapDescription()
    {
        return new String( "loading = " + totalHeap );
    }


    /***** IMPROVEMENT *****/


    /**
     * Removes a state task with a positive contribution. Returns whether
     * the heuristic was able to succeed.
     */
    private boolean heuristic_RemoveStateTask()
    {

        // Selecting a state task with positive costs

        double costs = (double)( Math.abs( gsc.random.nextLong() )
            % (long)Math.floor( positiveContributions ) );

        OrderedShiftDLLElement currentLE = taskList.first;
        NumHSRCStateTask task = null;
        boolean notFound = true;

        while ( notFound )
        {
            task = (NumHSRCStateTask)currentLE.element;
            costs = costs - task.contribution;
            if ( notFound = ( costs >= 0 ) )
            {
                currentLE = currentLE.next;
            }
        }

        // Deleting the state task

        boolean success = src.removeTask( task.taskConstraint , false );
        if ( Registry.LOG_CHANGES )
        {
            if ( success )
            {
                System.out.println( "heuristic_RemoveStateTask: " + task );
            }
            else {
                System.out.println( "heuristic_RemoveStateTask: improvement failed - task " + task + " cannot be removed ");
            }
        }

        return success;

    }


    public void improvePreconditionInconsistency()
    {

        if ( Registry.LOG_CHANGES )
        {
            System.out.println( "improving precondition inconsistency:" );
        }

        if ( positiveContributions == 0 )
        {
            throw new Error( "State task addition option is not supported yet" );
        }
        if ( gsc.structuralChangesEnabled )
        {
            heuristic_RemoveStateTask();
        }

    }


    public void improveGoalInconsistency()
    {
        if ( Registry.LOG_CHANGES )
        {
            System.out.println( "improving goal inconsistency:" );
        }
        throw new Error( "Improvement of static goals not yet implemented for NumericalHeapSRCs" );
    }


    public void addChangeListener( ChangeListener l )
    {
        listenerList.add( ChangeListener.class , (ChangeListener)l );
    }

    public void removeChangeListener( ChangeListener l )
    {
        listenerList.remove( ChangeListener.class , (ChangeListener)l );
    }

    /**
     * This method supports normal ChangeListeners.
     * @param changedObject The object which has changed.
     */
    protected void fireChange( Object changedObject )
    {
        ChangeEvent changeEvent = new ChangeEvent( changedObject );

        Object[] listeners = listenerList.getListenerList();
        ChangeListener listener = null;

        for ( int i = listeners.length - 2 ; i >= 0 ; i -= 2 )
        {
            if ( listeners[ i ] == ChangeListener.class )
            {
                ( (ChangeListener)listeners[ i + 1 ] ).stateChanged( changeEvent );
            }

        }
    }


    /**
     * Returns the taskList of the connected stateTasks.
     */
    public OrderedHeapCostsTaskList getTaskList()
    {
        return taskList;
    }


    /**
     * Returns the totalHeap of the NumericalHeapSRC.
     */
    public double getHeap()
    {
        return totalHeap;
    }


}
