
package com.ai_center.dragonbreath.constraints.stateresourceconstraint;


import com.ai_center.dragonbreath.constraints.PreconditionTaskObjectConstraint;
import com.ai_center.dragonbreath.constraints.StateResourceConstraint;
import com.ai_center.dragonbreath.constraints.StateTaskObjectConstraint;
import com.ai_center.dragonbreath.GlobalSearchControl;


/**
 * An interface that classes have to implement to qualify as a sub-class
 * for the state resource constraint.
 *
 * @author Alexander Nareyek
 * @version $Revision: 1.11 $
 *
 * @see StateResourceConstraint
 */
public interface SRCInterface
{


    // *******************
    // ***** METHODS *****
    // *******************


    /**
     * The following method has to be provided as a _static_ member
     * function in addition:
     *   public static int registerNewResourceType( String name );
     * It reserves a resource type ID and returns it.
     * @param name  The name to be given to the resource type.
     */

    /**
     * Initializes the sub-class.
     * @param src  The general state resource constraint parent instance.
     * @param gsc  The GlobalSearchControl solver object.
     * @param resourceType  The specific resource type definition.
     */
    void initialize( StateResourceConstraint src , GlobalSearchControl gsc , SRCResourceType resourceType );

    /**
     * Announces the change of the current time to the sub-class.
     * @param oldTime  The previous current time value.
     */
    void updateCurrentTime( long oldTime );

    /**
     * Announces the change of the current state to the sub-class.
     * @param oldValue  The previous state variable's value.
     */
    void updateCurrentState( double oldValue );

    /**
     * A connection of a state task is announced to the sub-class.
     * An internal representation is created and returned.
     * @param taskConstraint  The connected state task.
     */
    SRCStateTask includeStateTask( StateTaskObjectConstraint taskConstraint );

    /**
     * A connection of a precondition task is announced to the sub-class.
     * An internal representation is created and returned.
     * @param taskConstraint  The connected precondition task.
     */
    SRCPreconditionTask includePreconditionTask( PreconditionTaskObjectConstraint taskConstraint );

    /**
     * An update of a state task is announced to the sub-class.
     * @param task  The updated task's internal representation.
     * @param edgeID  The ID of the last edge to the changed variable.
     */
    void updateStateTask( SRCStateTask task , int edgeID );

    /**
     * An update of a precondition task is announced to the sub-class.
     * @param task  The updated task's internal representation.
     * @param edgeID  The ID of the last edge to the changed variable.
     */
    void updatePreconditionTask( SRCPreconditionTask task , int edgeID );

    /**
     * A disconnection of a state task is announced to the sub-class.
     * @param task  The disconnected task's internal representation.
     */
    void excludeStateTask( SRCStateTask task );

    /**
     * A disconnection of a precondition task is announced to the sub-class.
     * @param task  The disconnected task's internal representation.
     */
    void excludePreconditionTask( SRCPreconditionTask task );

    /**
     * Retrieves the projected state at a specific time.
     * @param time  The time at which the state is to be checked.
     * @return  The state variable's value at the given time.
     */
    double getStateAtTime( long time );

    /**
     * Returns a heuristic value for how difficult it would be to realize
     * a specific state at a specific time. The larger the value, the more
     * difficult would a realization be.
     * @param stateValue  The state that is to be realized.
     * @param time  The time at which the state is to be realized.
     * @return  The heuristic difficulty value.
     */
    int getDifficultyForStateAtTime( double stateValue , long time );

    /**
     * Returns a heuristic value for how difficult it would be to repair
     * a specific contribution's consequence at a specific time with
     * respect to following preconditions. The larger the value, the more
     * difficult would a repair be.
     * @param contributionValue  The contribution that would occur.
     * @param time  The time at which the contribution would occur.
     * @return  The heuristic difficulty value (can be negative for a
     *     positive influence).
     */
    int getDifficultyForContributionAtTime( double contributionValue ,
        long time );

    /**
     * The costs regarding non-matching preconditions are to be lowered.
     */
    void improvePreconditionInconsistency();

    /**
     * The costs regarding non-satisfied static goals are to be lowered.
     */
    void improveGoalInconsistency();

    /**
     * The sub-class is noticed that it will be deleted.
     */
    void terminate();


}
