
package com.ai_center.dragonbreath.constraints.stateresourceconstraint;


import java.math.BigInteger;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.EventListenerList;

import com.ai_center.dragonbreath.GlobalSearchControl;
import com.ai_center.dragonbreath.Registry;
import com.ai_center.dragonbreath.constraintgraph.ConventionalConstraint;
import com.ai_center.dragonbreath.constraintgraph.Embedding;
import com.ai_center.dragonbreath.constraintgraph.Node;
import com.ai_center.dragonbreath.constraintgraph.Variable;
import com.ai_center.dragonbreath.constraints.ActionObjectConstraint;
import com.ai_center.dragonbreath.constraints.PreconditionTaskObjectConstraint;
import com.ai_center.dragonbreath.constraints.StateResourceConstraint;
import com.ai_center.dragonbreath.constraints.StateTaskObjectConstraint;
import com.ai_center.dragonbreath.constraints.TaskConstraint;
import com.ai_center.dragonbreath.constraints.TransitionConstraint;
import com.ai_center.dragonbreath.constraints.stateresourceconstraint.symbolicsrc.OrderedPreconditionFailureList;
import com.ai_center.dragonbreath.constraints.stateresourceconstraint.symbolicsrc.SymSRCEvent;
import com.ai_center.dragonbreath.constraints.stateresourceconstraint.symbolicsrc.SymSRCInterval;
import com.ai_center.dragonbreath.constraints.stateresourceconstraint.symbolicsrc.SymSRCPreconditionTask;
import com.ai_center.dragonbreath.constraints.stateresourceconstraint.symbolicsrc.SymSRCResourceType;
import com.ai_center.dragonbreath.constraints.stateresourceconstraint.symbolicsrc.SymSRCState;
import com.ai_center.dragonbreath.constraints.stateresourceconstraint.symbolicsrc.SymSRCStateTask;
import com.ai_center.dragonbreath.constraints.stateresourceconstraint.symbolicsrc.SymSRCStateTransition;
import com.ai_center.dragonbreath.utilities.IntervalList;
import com.ai_center.dragonbreath.utilities.IntervalListElement;
import com.ai_center.dragonbreath.utilities.OrderedShiftDLLElement;


/**
 * This constraint serves as a sub-class with symbolic states for a
 * state resource constraint. ATTENTION: State addition after an instance
 * has been created is not yet supported.
 *
 * @author Alexander Nareyek
 * @version $Revision: 1.24 $
 *
 * @see StateResourceConstraint
 */
public class SymbolicSRC implements SRCInterface
{


    // **********************
    // ***** ATTRIBUTES *****
    // **********************


    /***** GENERAL INTERNAL STRUCTURES *****/

    /**
     * Constant representing an impossible state transition.
     */
    public final static int IMPOSSIBLE_TRANSITION = -1;

    /**
     * The unspecific state resource constraint parent class.
     */
    private StateResourceConstraint src = null;

    /**
     * The GlobalSearchControl solver object.
     */
    private GlobalSearchControl gsc = null;

    /**
     * Definition of the specific resource type.
     */
    private SymSRCResourceType resourceType = null;

    /**
     * List of precondition tasks ordered according to their subjective
     * precondition failure costs.
     */
    private OrderedPreconditionFailureList preconditionFailureList
        = new OrderedPreconditionFailureList();

    /**
     * List of goal-related precondition tasks ordered according to their
     * subjective precondition failure costs.
     */
    private OrderedPreconditionFailureList staticGoalFailureList = null;

    /**
     * A cost value for hard precondition inconsistencies, i.e.,
     * if there is no possible transition sequence from the interval
     * state to the checked state.
     */
    private long hardInconsistencyCosts;

    /**
     * The transition constraint which needs to be informed about the
     * occurrence of specific watched states.
     */
    private TransitionConstraint transitionConstraint = null;


    /***** INTERVAL MANAGEMENT *****/

    /**
     * Temporal interval list in which all connected tasks are included.
     */
    private IntervalList intervalList = null;

    /**
     * The interval including the current time.
     */
    private IntervalListElement currentTimeILE = null;


   /***** HEURISTICS *****/

    /**
     * The number of repair tries before giving up the improvement.
     * Pay attention with values that are similar or less than the
     * number of heuristics.
     */
    private final static int MAX_REPAIR_ATTEMPTS = 10;

    /**
     * The number of this constraint's different heuristics.
     */
    private final static int NUMBER_OF_HEURISTICS = 5;
    /**
     * Constants representing the single heuristics.
     */
    private final static int NO_HEURISTIC = -1;
    private final static int HEURISTIC_ADD_EVENT = 0;
    private final static int HEURISTIC_MOVE_EVENT = 1;
    private final static int HEURISTIC_MOVE_PRECONDITION_TASK = 2;
    private final static int HEURISTIC_DELETE_EVENT = 3;
    private final static int HEURISTIC_DELETE_PRECONDITION_TASK = 4;
    /**
     * Heuristic values expressing the preference for a heuristic.
     */
    private int[] choicePreference = new int[ NUMBER_OF_HEURISTICS ];
    /**
     * The initial value for the choice preference values (deletion-oriented
     * values generally start with 1).
     */
    private final static int INITIAL_PREFERENCE_VALUE = 10;
    /**
     * The maximal value for the choice preference values.
     */
    private final static int MAX_PREFERENCE_VALUE = 1000000;

    /**
     * The ID of last precondition improvement's chosen heuristic.
     */
    private int lastPreconditionImprovementChoice = NO_HEURISTIC;
    /**
     * The precondition-related costs at the last time the precondition
     * improvement was called.
     */
    private BigInteger lastPreconditionImprovementCosts = BigInteger.ZERO;

    /**
     * The ID of last goal improvement's chosen heuristic.
     */
    private int lastGoalImprovementChoice = NO_HEURISTIC;
    /**
     * The goal-related costs at the last time the goal improvement was
     * called.
     */
    private BigInteger lastGoalImprovementCosts = BigInteger.ZERO;

    private EventListenerList listenerList = new EventListenerList();


    // *******************
    // ***** METHODS *****
    // *******************


    /***** USER'S TYPE-RELATED REGISTRATION METHODS *****/


    /**
     * Reserves a resource type ID and returns it.
     * @param name  The name for the resource type.
     */
    public static int registerNewResourceType( String name )
    {
        SymSRCResourceType type = new SymSRCResourceType( name );
        int resourceTypeID = StateResourceConstraint.registerNewResourceType( StateResourceConstraint.SYMBOLIC_SRC , type );
        type.ID = resourceTypeID;
        return resourceTypeID;
    }


    /**
     * Registers a new possible state for the resource and returns a handle.
     * @param resourceTypeID  The ID of the resource type the state is for.
     * @param name  The name for the state.
     */
    public static int registerNewState( int resourceTypeID , String name )
    {
        SymSRCResourceType type = (SymSRCResourceType)StateResourceConstraint.getResourceTypeDefinition( resourceTypeID , StateResourceConstraint.SYMBOLIC_SRC );
        SymSRCState state = new SymSRCState( name );
        type.states.add( state );
        type.increaseTransitionTableSize();
        int stateID = type.states.size() - 1;
        state.ID = stateID;
        return stateID;
    }


    /**
     * Returns a specific state object.
     * @param resourceTypeID  The resource type ID of the state.
     * @param stateID  The ID of the state.
     */
    public static SymSRCState retrieveState( long resourceTypeID,
        long stateID )
    {
        SymSRCResourceType type = (SymSRCResourceType)StateResourceConstraint.getResourceTypeDefinition( (int)resourceTypeID , StateResourceConstraint.SYMBOLIC_SRC );
        int size = type.states.size();
        if ( ( stateID < 0 ) || ( stateID >= size ) )
        {
            throw new Error( "The given state handle \"" + stateID
                + "\" has not been registered for resource type \""
                + type.name + "\"" );
        }
        return (SymSRCState)type.states.get( (int)stateID );
    }


    /**
     * Declares a state to be the default state for a resource type.
     * @param resourceTypeID  The ID of the resource type.
     * @param stateID  The handle of the default state to be set.
     */
    public static void setDefaultState( int resourceTypeID , int stateID )
    {
        SymSRCResourceType type = (SymSRCResourceType)StateResourceConstraint.getResourceTypeDefinition( resourceTypeID , StateResourceConstraint.SYMBOLIC_SRC );
        int size = type.states.size();
        if ( ( stateID < 0 ) || ( stateID >= size ) )
        {
            throw new Error( "The given state handle \"" + stateID
                + "\" has not been registered for resource type \""
                + type.name + "\"" );
        }
        type.defaultState = (SymSRCState)type.states.get( stateID );
    }


    /**
     * Registers a new possible contribution for the resource and returns
     * a handle. This handle can also be used as value for contribution
     * variables for precondition tasks.
     * @param resourceTypeID  The ID of the resource type the event is for.
     * @param name  The name for the contribution.
     */
    public static int registerNewContribution( int resourceTypeID ,
        String name )
    {
        SymSRCResourceType type = (SymSRCResourceType)StateResourceConstraint.getResourceTypeDefinition( resourceTypeID , StateResourceConstraint.SYMBOLIC_SRC );
        SymSRCEvent event = new SymSRCEvent( name );
        type.events.add( event );
        int eventID = type.events.size() - 1;
        event.ID = eventID;
        return eventID;
    }


    /**
     * Registers a new possible contribution for the resource and returns
     * a handle. This handle can also be used as value for contribution
     * variables for precondition tasks.
     * @param resourceTypeID  The ID of the resource type the event is for.
     */
    public static int registerNewContribution( int resourceTypeID )
    {
        return registerNewContribution( resourceTypeID , null );
    }


    /**
     * Registers a contribution to be a sub-contribution for another
     * contribution. The other contribution can only happen if all its
     * sub-contributions happen at the same time.
     * @param resourceTypeID  The ID of the affected resource type.
     * @param subContributionID  The handle of the sub-contribution.
     * @param contributionID  The handle of the synergy contribution outcome.
     */
    public static void registerSynergyRelation( int resourceTypeID ,
        int subContributionID , int contributionID )
    {

        SymSRCResourceType type = (SymSRCResourceType)StateResourceConstraint.getResourceTypeDefinition( resourceTypeID , StateResourceConstraint.SYMBOLIC_SRC );

        // Consistency checking

        int size = type.events.size();
        if ( ( subContributionID < 0 ) || ( subContributionID >= size ) )
        {
            throw new Error( "The given contribution handle \""
                + subContributionID
                + "\" has not been registered for resource type \""
                + type.name + "\"" );
        }
        if ( ( contributionID < 0 ) || ( contributionID >= size ) )
        {
            throw new Error( "The given contribution handle \""
                + contributionID
                + "\" has not been registered for resource type \""
                + type.name + "\"" );
        }
        if ( subContributionID == contributionID )
        {
            throw new Error( "A contribution cannot be a sub-contribution of itself" );
        }

        // Adding the synergy information to the events

        SymSRCEvent subEvent
            = (SymSRCEvent)type.events.get( subContributionID );
        SymSRCEvent event
            = (SymSRCEvent)type.events.get( contributionID );

        if ( subEvent.synergyEvents == null )
        {
            subEvent.synergyEvents = new LinkedList();
        }
        subEvent.synergyEvents.add( event );

        if ( event.subEvents == null )
        {
            event.subEvents = new LinkedList();
        }
        event.subEvents.add( subEvent );

    }


    /**
     * Returns a specific contribution object.
     * @param resourceTypeID  The resource type ID of the contribution.
     * @param contributionID  The ID of the contribution.
     */
    public static SymSRCEvent retrieveContribution( long resourceTypeID ,
        long contributionID )
    {
        SymSRCResourceType type = (SymSRCResourceType)StateResourceConstraint.getResourceTypeDefinition( (int)resourceTypeID , StateResourceConstraint.SYMBOLIC_SRC );
        int size = type.events.size();
        if ( ( contributionID < 0 ) || ( contributionID >= size ) )
        {
            throw new Error( "The given contribution handle \""
                + contributionID
                + "\" has not been registered for resource type \""
                + type.name + "\"" );
        }
        return (SymSRCEvent)type.events.get( (int)contributionID );
    }


    /**
     * Registers a possible transition between states, caused by a
     * specific contribution.
     * @param resourceTypeID  The ID of the affected resource type.
     * @param predecessorStateID  The handle of the predecessor state.
     * @param contributionID  The handle of the contribution which causes
     *                        the transition.
     * @param successorStateID  The handle of the successor state.
     * @param actionTypeID  The ID of the action that causes this
     *                      transition. Pass -1 if multiple options.
     * @param duration  The total duration of the causing action.
     * @param difficulty  A heuristic value (> 0) that expresses the
     *                    expected difficulty to establish the conditions
     *                    for such a transition.
     */
    public static void registerStateTransition(int resourceTypeID ,
        int predecessorStateID , int contributionID , int successorStateID ,
        int actionTypeID , long duration , int difficulty )
    {

        SymSRCResourceType type = (SymSRCResourceType)StateResourceConstraint.getResourceTypeDefinition( resourceTypeID , StateResourceConstraint.SYMBOLIC_SRC );

        // Consistency checking

        int sizeEvents = type.events.size();
        if ( ( contributionID < 0 ) || ( contributionID >= sizeEvents ) )
        {
            throw new Error( "The given contribution handle \""
                + contributionID
                + "\" has not been registered for resource type \""
                + type.name + "\"" );
        }
        int sizeStates = type.states.size();
        if ( ( predecessorStateID < 0 )
            || ( predecessorStateID >= sizeStates ) )
        {
            throw new Error( "The given predecessor state handle \""
                + predecessorStateID
                + "\" has not been registered for resource type \""
                + type.name + "\"" );
        }
        if ( ( successorStateID < 0 ) || ( successorStateID >= sizeStates ) )
        {
            throw new Error( "The given successor state handle \""
                + successorStateID
                + "\" has not been registered for resource type \""
                + type.name + "\"" );
        }
        if ( duration <= 0 )
        {
            throw new Error( "Invalid duration value" );
        }
        if ( difficulty <= 0 )
        {
            throw new Error( "Invalid difficulty value" );
        }

        // Creating the transition objects and adding the transition
        // information to the states

        SymSRCState predecessorState
            = (SymSRCState)type.states.get( predecessorStateID );
        SymSRCState successorState
            = (SymSRCState)type.states.get( successorStateID );
        SymSRCEvent event
            = (SymSRCEvent)type.events.get( contributionID );

        if ( predecessorState.successorStates == null )
        {
            predecessorState.successorStates = new Hashtable();
            predecessorState.successorStatesCollection
                = predecessorState.successorStates.values();
        }
        else
        {
            if ( predecessorState.successorStates.get( event ) != null )
            {
                throw new Error( "A transition from state \""
                    + predecessorState
                    + "\" using this event has already been registered" );
            }
        }
        if ( successorState.predecessorStates == null )
        {
            successorState.predecessorStates = new Hashtable();
            successorState.predecessorStatesCollection
                = successorState.predecessorStates.values();
        }

        SymSRCStateTransition stateTransition = new SymSRCStateTransition( predecessorState , event , successorState , actionTypeID , duration , difficulty );
        predecessorState.successorStates.put( event , stateTransition );
        successorState.predecessorStates.put( event , stateTransition );

        // Propagating difficulty information

        type.difficultySum += difficulty;
        type.propagateDifficulty( stateTransition );

    }


    /***** USER'S INSTANCE-RELATED REGISTRATION METHODS *****/


    /**
     * Adds a goal to be considered for the static goal-related cost
     * function.
     * @param goalStateID  The handle of the state to be checked.
     * @param timePoint  The time point at which the state is to be checked.
     */
    public void addGoal( int goalStateID , long timePoint )
    {

        // Checking for a valid temporal reference value

        if ( timePoint < 0 )
        {
            throw new Error( "Invalid value \"" + timePoint
                + "\" for the goal's temporal reference" );
        }

        // Checking for a valid state value

        int sizeStates = resourceType.states.size();
        if ( ( goalStateID < 0 ) || ( goalStateID >= sizeStates ) )
        {
            throw new Error( "Invalid value \"" + goalStateID
                + "\" for the goal's state to be checked" );
        }

        // Creating precondition task; tempVariable == null indicates
        // that this is a goal-related precondition task

        SymSRCPreconditionTask task = new SymSRCPreconditionTask();

        task.tempVariable = null;
        task.tempValue = timePoint;
        task.state = goalStateID;
        task.taskConstraint = null;
        task.resourceTypeVariable = null;
        task.resourceType = resourceType.ID;

        // Including task in interval structures and updating costs

        addPreconditionTaskToProjection( task );
        fireChange( this );
    }


    /***** GENERAL SRC-INTERFACE-FUNCTIONS *****/


    public void initialize( StateResourceConstraint src , GlobalSearchControl gsc , SRCResourceType resourceType )
    {

        this.src = src;
        this.gsc = gsc;
        this.resourceType = (SymSRCResourceType)resourceType;
        if ( this.resourceType.defaultState == null )
        {
            throw new Error( "No default state has been set for resource type \"" + this.resourceType.name + "\"" );
        }
        hardInconsistencyCosts = this.resourceType.difficultySum + 1;

        // Requiring the state variable to be integer

        src.modificationModus = true;
        gsc.addIntegerConstraint( src.stateVariable );
        src.modificationModus = false;

        // Initializing interval structures

        SymSRCInterval firstInterval = new SymSRCInterval();
        firstInterval.event = null;
        firstInterval.state = this.resourceType.defaultState;
        intervalList = new IntervalList( 0 , Long.MAX_VALUE , firstInterval );
        currentTimeILE = intervalList.getFirst();

        if ( src.state != this.resourceType.defaultState.ID )
        {
            updateCurrentState( 0 );
        }

        // Initializing choice of heuristics

        for ( int i = 0 ; i < NUMBER_OF_HEURISTICS ; i++ )
        {
            choicePreference[ i ] = INITIAL_PREFERENCE_VALUE;
        }
        choicePreference[ HEURISTIC_DELETE_EVENT ] = 1;
        choicePreference[ HEURISTIC_DELETE_PRECONDITION_TASK ] = 1;

        // Retrieving transition constraint

        transitionConstraint = (TransitionConstraint)gsc.findConstraint( Registry.TRANSITION_CONSTRAINT );

        // Retrieving state watch information

        HashSet stateWatchList
            = TransitionConstraint.getStateWatchList( this.resourceType.ID );

        // Marking states to be watched

        Iterator it = stateWatchList.iterator();
        while ( it.hasNext() )
        {
            ( (SymSRCState)this.resourceType.states.get( ( (Long)it.next() ).intValue() ) ).watched = true;
        }

        // Checking state watch for initial state

        if ( firstInterval.state.watched && ( transitionConstraint != null ) )
        {
            transitionConstraint.registerStateOccurrence( src ,
                firstInterval.state , 0 , Long.MAX_VALUE );
        }

    }


    public void updateCurrentTime( long oldTime )
    {

        if ( oldTime < src.currentTime )
        {

            // Moving current time interval to correct position

            while ( currentTimeILE.getEnd() <= src.currentTime )
            {
                currentTimeILE = currentTimeILE.getNext();
                ( (SymSRCInterval)currentTimeILE.getElement() ).event = null;
            }

            // Updating state info at current time

            updateCurrentState( 0 );

        }
        else
        {
            throw new Error( "Backward moving time not supported for SymbolicSRC" );
        }

    }


    public void updateCurrentState( double oldValue )
    {

        int size = resourceType.states.size();
        if ( ( src.state >= 0 ) && ( src.state < size ) )
        {

            SymSRCState newState
                = (SymSRCState)resourceType.states.get( (int)src.state );
            SymSRCInterval interval
                = (SymSRCInterval)currentTimeILE.getElement();

            if ( interval.state != newState )
            {

                if ( src.currentTime != currentTimeILE.getBeginning() )
                {

                    IntervalListElement ile = intervalList.addBasePoint( src.currentTime , currentTimeILE );

                    // Adding a new interval

                    SymSRCInterval newInterval = new SymSRCInterval();
                    ile.setElement( newInterval );
                    newInterval.event = null;
                    newInterval.state = newState;

                    // Potentially shifting precondition tasks

                    SymSRCPreconditionTask preconditionTask;
                    Iterator it = interval.preconditionTasks.iterator();
                    while ( it.hasNext() )
                    {
                        preconditionTask = (SymSRCPreconditionTask)it.next();
                        if ( preconditionTask.tempValue >= ile.getBeginning() )
                        {
                            it.remove();
                            newInterval.preconditionTasks.add( preconditionTask );
                            preconditionTask.intervalReference = ile;
                        }
                    }

                    // Adjusting current time interval

                    currentTimeILE = ile;

                    if ( interval.state.watched
                        && ( transitionConstraint != null ) )
                    {
                        transitionConstraint.registerStateLoss( src , interval.state , currentTimeILE.getBeginning() , currentTimeILE.getEnd() );
                    }

                }
                else
                {

                    // Updating an existing interval

                    if ( interval.state.watched
                        && ( transitionConstraint != null ) )
                    {
                        transitionConstraint.registerStateLoss( src , interval.state , currentTimeILE.getBeginning() , currentTimeILE.getEnd() );
                    }

                    interval.state = newState;

                }

                if ( newState.watched && ( transitionConstraint != null ) )
                {
                    transitionConstraint.registerStateOccurrence( src , newState , currentTimeILE.getBeginning() , currentTimeILE.getEnd() );
                }

                // Re-projecting state

                projectState( currentTimeILE );

            }
        }

        fireChange( this );

    }


    public SRCStateTask includeStateTask( StateTaskObjectConstraint taskConstraint )
    {

        SymSRCStateTask task = new SymSRCStateTask();

        // Fetching state task's variables

        Variable tcTempVariable;
        Variable tcContributionVariable;

        synchronized( taskConstraint.connections )
        {
            tcTempVariable = taskConstraint.connections.getSpecificVariable( Registry.TEMPORALREFERENCE_EDGE );
            tcContributionVariable = taskConstraint.connections.getSpecificVariable( Registry.CONTRIBUTION_EDGE );
        }
        if ( tcTempVariable == null )
        {
            throw new Error( "The connected StateTaskObjectConstraint \""
                + taskConstraint + "\" has no temporal reference variable" );
        }
        if ( tcContributionVariable == null )
        {
            throw new Error( "The connected StateTaskObjectConstraint \""
                + taskConstraint + "\" has no state variable" );
        }

        // Processing state task's temporal reference variable

        src.modificationModus = true;
        gsc.addIntegerConstraint( tcTempVariable );
        src.modificationModus = false;
        task.tempVariable = tcTempVariable;
        task.tempValue = (long)tcTempVariable.value();

        // Checking for a valid temporal reference value

        if ( task.tempValue < 0 )
        {
            throw new Error( "Invalid value \"" + task.tempValue
                + "\" for the state task's temporal reference" );
        }

        // Processing state task's contribution variable

        src.modificationModus = true;
        gsc.addIntegerConstraint( tcContributionVariable );
        src.modificationModus = false;
        task.contributionVariable = tcContributionVariable;
        int contribution = (int)tcContributionVariable.value();

        // Checking for a valid contribution value

        int sizeEvents = resourceType.events.size();
        if ( ( contribution < 0 ) || ( contribution >= sizeEvents ) )
        {
            throw new Error( "Invalid value \"" + contribution
                + "\" for the state task's contribution" );
        }
        task.contribution
            = (SymSRCEvent)resourceType.events.get(contribution);

        // Checking if task is part of a world transition

        Node actionConstraint;
        synchronized( taskConstraint.connections )
        {
            actionConstraint = taskConstraint.connections.getSpecificConstraint( Registry.ACTION_OBJECTCONSTRAINT , Registry.PLAIN_EDGE );
        }
        if ( actionConstraint != null )
        {
            Object transitionConstraint;
            synchronized( actionConstraint.connections )
            {
                transitionConstraint = actionConstraint.connections.getSpecificNodes( Registry.TRANSITION_CONSTRAINT );
            }
            task.partOfWorldTransition = ( transitionConstraint != null );
        }

        // Updating state projection and costs

        addStateTaskToProjection( task );
        fireChange( this );
        return task;

    }


    public SRCPreconditionTask includePreconditionTask( PreconditionTaskObjectConstraint taskConstraint )
    {

        SymSRCPreconditionTask task = new SymSRCPreconditionTask();

        // Fetching precondition task's variables

        Variable tcTempVariable;
        Variable tcStateVariable;

        synchronized( taskConstraint.connections )
        {
            tcTempVariable = taskConstraint.connections.getSpecificVariable( Registry.TEMPORALREFERENCE_EDGE );
            tcStateVariable = taskConstraint.connections.getSpecificVariable( Registry.STATE_EDGE );
        }
        if ( tcTempVariable == null )
        {
            throw new Error( "The connected PreconditionTaskObjectConstraint \"" + taskConstraint + "\" has no temporal reference variable" );
        }
        if ( tcStateVariable == null )
        {
            throw new Error( "The connected PreconditionTaskObjectConstraint \"" + taskConstraint + "\" has no state variable ");
        }

        // Processing precondition task's temporal reference variable

        src.modificationModus = true;
        gsc.addIntegerConstraint( tcTempVariable );
        src.modificationModus = false;
        task.tempVariable = tcTempVariable;
        task.tempValue = (long)tcTempVariable.value();

        // Checking for a valid temporal reference value

        if ( task.tempValue < 0 )
        {
            throw new Error( "Invalid value \"" + task.tempValue
                + "\" for the precondition task's temporal reference" );
        }

        // Processing precondition task's state variable

        src.modificationModus = true;
        gsc.addIntegerConstraint( tcStateVariable );
        src.modificationModus = false;
        task.stateVariable = tcStateVariable;
        task.state = (int)tcStateVariable.value();

        // Checking for a valid state value

        int sizeStates = resourceType.states.size();
        if ( ( task.state < 0 ) || ( task.state >= sizeStates ) )
        {
            throw new Error( "Invalid value \"" + task.state
                + "\" for the precondition task's state" );
        }

        // Checking if task is part of a world transition

        Node actionConstraint;
        synchronized( taskConstraint.connections )
        {
            actionConstraint = taskConstraint.connections.getSpecificConstraint( Registry.ACTION_OBJECTCONSTRAINT , Registry.PLAIN_EDGE );
        }
        if ( actionConstraint != null )
        {
            Object transitionConstraint;
            synchronized( actionConstraint.connections )
            {
                transitionConstraint = actionConstraint.connections.getSpecificNodes( Registry.TRANSITION_CONSTRAINT );
            }
            task.partOfWorldTransition = ( transitionConstraint != null );
        }

        // Including task in interval structures and updating costs

        addPreconditionTaskToProjection( task );
        fireChange( this );
        return task;

    }


    public void updateStateTask( SRCStateTask task , int edgeID )
    {

        SymSRCStateTask specializedTask = (SymSRCStateTask)task;

        // Retrieving new value

        if ( edgeID == Registry.TEMPORALREFERENCE_EDGE )
        {

            // Updating state projection and costs

            removeStateTaskFromProjection( specializedTask );

            specializedTask.tempValue
                = (long)specializedTask.tempVariable.value();

            // Checking for a valid temporal reference value

            if ( specializedTask.tempValue < 0 )
            {
                throw new Error( "Invalid value \"" + specializedTask.tempValue + "\" for the state task's temporal reference" );
            }

            // Updating state projection and costs

            addStateTaskToProjection( specializedTask );

        }
        else
        {
            if ( edgeID == Registry.CONTRIBUTION_EDGE )
            {

                // Updating state projection and costs

                removeStateTaskFromProjection( specializedTask );

                int contribution
                    = (int)specializedTask.contributionVariable.value();

                // Checking for a valid contribution value

                int sizeEvents = resourceType.events.size();
                if ( ( contribution < 0 ) || ( contribution >= sizeEvents ) )
                {
                    throw new Error( "Invalid value \"" + contribution
                        + "\" for the state task's contribution" );
                }
                specializedTask.contribution
                    = (SymSRCEvent)resourceType.events.get( contribution );

                // Updating state projection and costs

                addStateTaskToProjection( specializedTask );

            }
        }
        fireChange( this );

    }


    public void updatePreconditionTask( SRCPreconditionTask task , int edgeID )
    {

        SymSRCPreconditionTask specializedTask = (SymSRCPreconditionTask)task;

        // Retrieving new value

        if ( edgeID == Registry.TEMPORALREFERENCE_EDGE )
        {

            // Excluding task from interval structures and updating costs

            removePreconditionTaskFromProjection( specializedTask );

            specializedTask.tempValue
                = (long)specializedTask.tempVariable.value();

            // Checking for a valid temporal reference value

            if ( specializedTask.tempValue < 0 )
            {
                throw new Error( "Invalid value \"" + specializedTask.tempValue + "\" for the precondition task's temporal reference" );
            }

            // Including task in interval structures and updating costs

            addPreconditionTaskToProjection( specializedTask );

        }
        else
        {
            if ( edgeID == Registry.STATE_EDGE )
            {

                specializedTask.state
                    = (int)specializedTask.stateVariable.value();

                // Checking for a valid state value

                int sizeStates = resourceType.states.size();
                if ( ( specializedTask.state < 0 )
                    || ( specializedTask.state >= sizeStates ) )
                {
                    throw new Error( "Invalid value \"" + specializedTask.state
                        + "\" for the precondition task's state" );
                }

                // Updating costs

                SymSRCState state = ( (SymSRCInterval)specializedTask.intervalReference.getElement() ).state;
                long costs = resourceType.transitionDifficulty[state.ID][specializedTask.state];
                if ( costs == IMPOSSIBLE_TRANSITION )
                {
                    costs = hardInconsistencyCosts;
                }
                if ( costs != specializedTask.preconditionCosts )
                {
                    src.costs[ src.PRECONDITION_FAILURE ]
                        += costs - specializedTask.preconditionCosts;
                    specializedTask.preconditionCosts = costs;
                    if ( src.costFunctionReference[ src.PRECONDITION_FAILURE ] != null )
                    {
                        src.costFunctionReference[ src.PRECONDITION_FAILURE ].updateGlobalCosts();
                    }
                }

            }
        }
        fireChange( this );

    }


    public void excludeStateTask( SRCStateTask task )
    {

        SymSRCStateTask specializedTask = (SymSRCStateTask)task;

        // Updating state projection and costs

        removeStateTaskFromProjection( specializedTask );

        // Removing IntegerConstraints from the variables

        src.modificationModus = true;
        gsc.removeIntegerConstraint( specializedTask.tempVariable );
        gsc.removeIntegerConstraint( specializedTask.contributionVariable );
        src.modificationModus = false;

        fireChange( this );

    }


    public void excludePreconditionTask( SRCPreconditionTask task )
    {

        SymSRCPreconditionTask specializedTask = (SymSRCPreconditionTask)task;

        // Excluding task from interval structures and updating costs

        removePreconditionTaskFromProjection( specializedTask );

        // Removing IntegerConstraints from the variables

        src.modificationModus = true;
        gsc.removeIntegerConstraint( specializedTask.tempVariable );
        gsc.removeIntegerConstraint( specializedTask.stateVariable );
        src.modificationModus = false;

        fireChange( this );

    }


    public double getStateAtTime( long time )
    {
        SymSRCInterval interval = (SymSRCInterval)intervalList.findInterval( time ).getElement();
        return interval.state.ID;
    }


    public int getDifficultyForStateAtTime( double stateValue , long time )
    {
        SymSRCInterval interval = (SymSRCInterval)intervalList.findInterval( time ).getElement();
        return (int)resourceType.transitionDifficulty[ interval.state.ID ][ (int)stateValue ];
    }


    public int getDifficultyForContributionAtTime( double contributionValue ,
        long time )
    {

        int difficulty = 0;
        SymSRCInterval interval = (SymSRCInterval)( intervalList.findInterval( time ).getElement() );

        if ( interval.preconditionTasks != null ) {

            SymSRCEvent event = (SymSRCEvent)( resourceType.events.get( (int)contributionValue ) );

            // Computing following state

            SymSRCState previousState = interval.state;
            SymSRCState nextState = previousState;
            if ( previousState.successorStates != null )
            {
                SymSRCStateTransition transition = (SymSRCStateTransition)previousState.successorStates.get( event );
                if ( transition != null )
                {
                    nextState = transition.successorState;
                }
            }

            // Computing repair difficulties

            if ( previousState != nextState )
            {

                SymSRCPreconditionTask preconditionTask;
                long costs = 0;
                Iterator it = interval.preconditionTasks.iterator();

                while ( it.hasNext() )
                {
                    preconditionTask = (SymSRCPreconditionTask)it.next();
                    if ( preconditionTask.tempValue > time )
                    {
                        costs += resourceType.transitionDifficulty[ nextState.ID ][ preconditionTask.state ] - preconditionTask.preconditionCosts;
                    }
                }

                if ( costs <= Integer.MAX_VALUE )
                {
                    difficulty = (int)costs;
                }
                else
                {
                    difficulty = Integer.MAX_VALUE;
                }

            }

        }

        return difficulty;

    }


    public void terminate()
    {

        // Removing IntegerConstraints from the state variable

        src.modificationModus = true;
        gsc.removeIntegerConstraint( src.stateVariable );
        src.modificationModus = false;

    }


    /***** INTERVAL LIST *****/


    /**
     * Returns a description of the projection including all tasks.
     */
    public StringBuffer getFullProjectionDescription()
    {

        StringBuffer s = new StringBuffer( resourceType.name + ":\n" );

        IntervalListElement currentILE = intervalList.getFirst();
        SymSRCInterval interval;

        while ( currentILE != null )
        {
            interval = (SymSRCInterval)currentILE.getElement();
            s.append( "  " + currentILE.getBeginning() + "..  Event: " + interval.event + "  CON: " + interval.stateTasks.toString() + "\n" );
            if ( currentILE == currentTimeILE )
            {
                s.append( "  (CT = " + src.currentTime + ")\n" );
            }
            s.append( "   " + interval.state + "  PRE: " + interval.preconditionTasks.toString() + "\n" );

            currentILE = currentILE.getNext();
        }
        s.append( "  " + intervalList.getLast().getEnd() + "\n" );

        return s;

    }


    /**
     * Returns a description of the state projection.
     */
    public StringBuffer getStateProjectionDescription()
    {

        StringBuffer s = new StringBuffer( resourceType.name + ": " );
        IntervalListElement currentILE = intervalList.getFirst();
        SymSRCInterval interval;

        if ( src.currentTime != currentILE.getBeginning() )
        {
            s.append( " [" + currentILE.getBeginning() + "]" );
        }
        else
        {
            s.append( " [CT=" + src.currentTime + "]" );
        }

        while ( currentILE != null )
        {
            if ( ( src.currentTime > currentILE.getBeginning() ) && ( src.currentTime <= currentILE.getEnd() ) )
            {
                s.append( " " + ( (SymSRCInterval)currentILE.getElement() ).state.toString() + " [CT=" + src.currentTime + "]" );
            }
            if ( src.currentTime != currentILE.getEnd() )
            {
                s.append( " " + ((SymSRCInterval) currentILE.getElement()).state.toString() + " [" + currentILE.getEnd() + "]" );
            }
            currentILE = currentILE.getNext();
        }

        return s;

    }


    /**
     * Projects the state of the resource from a given interval until the
     * last interval or until no change occurs with respect to the current
     * projection. The propagation continues in any case until the current
     * time is reached.
     * @param ile  The interval to start with.
     */
    private void projectState( IntervalListElement ile )
    {

        // Retrieving predecessor state

        SymSRCState previousState = null;
        if ( ile != intervalList.getFirst() )
        {
            previousState
                = ( (SymSRCInterval)ile.getPrevious().getElement() ).state;
        }

        // Propagating along intervals

        boolean change = true;
        boolean realChange;
        SymSRCInterval interval;
        SymSRCStateTransition transition;
        SymSRCState newState;

        while ( change && ( ile != null ) )
        {

            // Calculating new state

            interval = (SymSRCInterval)ile.getElement();

            if ( ile.getBeginning() > src.currentTime )
            {

                if ( ( interval.event != null )
                    && ( previousState.successorStates != null ) )
                {
                    transition = (SymSRCStateTransition)previousState.successorStates.get( interval.event );
                }
                else
                {
                    transition = null;
                }
                if ( transition == null )
                {
                    newState = previousState;
                }
                else
                {
                    newState = transition.successorState;
                }

                change = ( newState != interval.state );
                realChange = change;

            }
            else
            {
                newState = interval.state;
                change = true;
                realChange = false;
            }

            if ( change )
            {

                // Updating state watch information

                if ( realChange && ( transitionConstraint != null ) )
                {
                    if ( ( interval.state != null ) && interval.state.watched )
                    {
                        transitionConstraint.registerStateLoss( src , interval.state , ile.getBeginning() , ile.getEnd() );
                    }
                    if ( newState.watched )
                    {
                        transitionConstraint.registerStateOccurrence( src , newState , ile.getBeginning() , ile.getEnd() );
                    }
                }

                // Updating costs of precondition tasks

                interval.state = newState;

                SymSRCPreconditionTask preconditionTask;
                long costs;
                Iterator it = interval.preconditionTasks.iterator();
                while ( it.hasNext() )
                {
                    preconditionTask = (SymSRCPreconditionTask)it.next();
                    costs = resourceType.transitionDifficulty[ newState.ID ][ preconditionTask.state ];
                    if ( costs == IMPOSSIBLE_TRANSITION )
                    {
                        costs = hardInconsistencyCosts;
                    }
                    if ( costs != preconditionTask.preconditionCosts )
                    {
                        if ( preconditionTask.tempVariable != null )
                        {
                            src.costs[ src.PRECONDITION_FAILURE ]
                                += costs - preconditionTask.preconditionCosts;
                            preconditionTask.preconditionCosts = costs;
                            preconditionFailureList.updateListElementPosition( preconditionTask.preconditionFailureLE );
                        }
                        else
                        {
                            src.costs[ src.STATIC_GOAL_FAILURE ]
                                += costs - preconditionTask.preconditionCosts;
                            preconditionTask.preconditionCosts = costs;
                            staticGoalFailureList.updateListElementPosition( preconditionTask.preconditionFailureLE );
                        }
                    }
                }
                if ( src.costFunctionReference[ src.PRECONDITION_FAILURE ]
                    != null )
                {
                    src.costFunctionReference[ src.PRECONDITION_FAILURE ].updateGlobalCosts();
                }
                if ( src.costFunctionReference[ src.STATIC_GOAL_FAILURE ]
                    != null )
                {
                    src.costFunctionReference[ src.STATIC_GOAL_FAILURE ].updateGlobalCosts();
                }

            }
            previousState = newState;
            ile = ile.getNext();
        }

    }


    /**
     * Adds a state task to the interval structures and updates costs.
     * @param task  The state task to be added.
     */
    private void addStateTaskToProjection( SymSRCStateTask task )
    {

        // Splitting interval at task's temporal reference if necessary

        if ( task.tempValue != 0 )
        {

            IntervalListElement taskILE
                = intervalList.addBasePoint( task.tempValue );
            SymSRCInterval interval;

            if ( taskILE.getElement() == null )
            {

                SymSRCInterval prevInterval
                    = (SymSRCInterval)taskILE.getPrevious().getElement();

                // Adjusting current time interval

                if ( ( currentTimeILE == taskILE.getPrevious() )
                    && ( src.currentTime >= taskILE.getBeginning() ) )
                {
                    currentTimeILE = taskILE;
                }

                // Adding a new interval

                interval = new SymSRCInterval();
                taskILE.setElement( interval );
                interval.stateTasks.add( task );
                task.intervalReference = taskILE;

                // Potentially shifting precondition tasks

                SymSRCPreconditionTask preconditionTask;
                Iterator it = prevInterval.preconditionTasks.iterator();
                while ( it.hasNext() )
                {
                    preconditionTask = (SymSRCPreconditionTask)it.next();
                    if ( preconditionTask.tempValue >= taskILE.getBeginning() )
                    {
                        it.remove();
                        interval.preconditionTasks.add( preconditionTask );
                        preconditionTask.intervalReference = taskILE;
                    }
                }

                // Determining new event of the task's interval and
                // projecting state

                if ( taskILE.getBeginning() > src.currentTime )
                {

                    if ( prevInterval.state.watched
                        && ( transitionConstraint != null ) )
                    {
                        transitionConstraint.registerStateLoss( src , prevInterval.state , taskILE.getBeginning() , taskILE.getEnd() );
                    }

                    interval.event = task.contribution;
                    projectState( taskILE );

                }
                else
                {
                    interval.event = null;
                    interval.state = ( (SymSRCInterval)taskILE.getPrevious().getElement() ).state;
                }

            }
            else
            {

                // Interval already existent

                taskILE = taskILE.getNext();
                interval = (SymSRCInterval)taskILE.getElement();
                interval.stateTasks.add( task );
                task.intervalReference = taskILE;

                // Determining new event of the task's interval and
                // projecting state

                if ( taskILE.getBeginning() > src.currentTime )
                {
                    if ( interval.stateTasks.size() == 1 )
                    {
                        interval.event = task.contribution;
                    }
                    else
                    {
                        interval.event = resourceType.calculateSynergyOutcome( interval.stateTasks );
                    }
                    projectState( taskILE );
                }

            }

        }
        else
        {
            ( (SymSRCInterval)intervalList.getFirst().getElement() ).stateTasks.add( task );
            task.intervalReference = intervalList.getFirst();
        }


    }


    /**
     * Removes a state task from the interval structures and updates costs.
     * @param task  The state task to be removed.
     */
    private void removeStateTaskFromProjection( SymSRCStateTask task )
    {

        IntervalListElement ile = task.intervalReference;
        SymSRCInterval interval = (SymSRCInterval)ile.getElement();

        if ( ( interval.stateTasks.size() == 1 ) && ( ( ile.getBeginning() > src.currentTime ) || ( ( ile.getBeginning() != 0 ) && ( ( (SymSRCInterval)ile.getPrevious().getElement() ).state == interval.state ) ) ) )
        {

            // Adjusting current time interval

            if ( currentTimeILE == ile )
            {
                currentTimeILE = ile.getPrevious();
            }

            // Deleting interval

            IntervalListElement previousILE = ile.getPrevious();
            SymSRCInterval previousInterval
                = (SymSRCInterval)previousILE.getElement();
            boolean stateChange = ( interval.state != previousInterval.state );

            if ( stateChange && ( transitionConstraint != null ) )
            {
                if ( interval.state.watched )
                {
                    transitionConstraint.registerStateLoss( src , interval.state , ile.getBeginning() , ile.getEnd() );
                }
                if ( previousInterval.state.watched )
                {
                    transitionConstraint.registerStateOccurrence( src , previousInterval.state , ile.getBeginning() , ile.getEnd() );
                }
            }

            intervalList.removeBasePoint( ile.getBeginning() , previousILE );

            // Shifting precondition tasks

            SymSRCPreconditionTask preconditionTask;
            long costs;
            Iterator it = interval.preconditionTasks.iterator();
            while ( it.hasNext() )
            {

                preconditionTask = (SymSRCPreconditionTask)it.next();
                previousInterval.preconditionTasks.add( preconditionTask );
                preconditionTask.intervalReference = previousILE;

                if ( stateChange )
                {

                    // Updating costs for precondition task

                    costs = resourceType.transitionDifficulty[ previousInterval.state.ID ][ preconditionTask.state ];
                    if ( costs == IMPOSSIBLE_TRANSITION )
                    {
                        costs = hardInconsistencyCosts;
                    }
                    if ( costs != preconditionTask.preconditionCosts )
                    {
                        if ( preconditionTask.tempVariable != null )
                        {
                            src.costs[ src.PRECONDITION_FAILURE ]
                                += costs - preconditionTask.preconditionCosts;
                            preconditionTask.preconditionCosts = costs;
                            preconditionFailureList.updateListElementPosition( preconditionTask.preconditionFailureLE );
                        }
                        else
                        {
                            src.costs[ src.STATIC_GOAL_FAILURE ]
                                += costs - preconditionTask.preconditionCosts;
                            preconditionTask.preconditionCosts = costs;
                            staticGoalFailureList.updateListElementPosition( preconditionTask.preconditionFailureLE );
                        }
                    }

                }

            }

            if ( stateChange )
            {

                // Global cost propagation

                if ( src.costFunctionReference[ src.PRECONDITION_FAILURE ]
                    != null )
                {
                    src.costFunctionReference[ src.PRECONDITION_FAILURE ].updateGlobalCosts();
                }
                if ( src.costFunctionReference[ src.STATIC_GOAL_FAILURE ]
                    != null )
                {
                    src.costFunctionReference[ src.STATIC_GOAL_FAILURE ].updateGlobalCosts();
                }

                // Re-projecting state

                if ( previousILE.getNext() != null )
                {
                    projectState( previousILE.getNext() );
                }

            }

        }
        else
        {

            // Interval is not to be removed

            interval.stateTasks.remove( task );

            if ( ile.getBeginning() > src.currentTime )
            {

                // Determining new event

                if ( interval.stateTasks.size() == 1 )
                {
                    Iterator it = interval.stateTasks.iterator();
                    interval.event
                        = ( (SymSRCStateTask)it.next() ).contribution;
                }
                else
                {
                    interval.event = resourceType.calculateSynergyOutcome( interval.stateTasks );
                }

                // Re-projecting state

                projectState( ile );

            }

        }


    }


    /**
     * Adds a precondition task to the interval structures and updates the
     * task's costs.
     * @param task  The precondition task to be added.
     */
    private void addPreconditionTaskToProjection( SymSRCPreconditionTask task )
    {

        // Adding task to interval structures

        IntervalListElement ile = intervalList.findInterval( task.tempValue );
        if ( task.tempValue == ile.getEnd() )
        {
            ile = ile.getNext();
        }
        task.intervalReference = ile;
        ( (SymSRCInterval)ile.getElement() ).preconditionTasks.add( task );

        // Cost initialization

        SymSRCState state = ( (SymSRCInterval)ile.getElement() ).state;
        long costs
            = resourceType.transitionDifficulty[ state.ID ][ task.state ];
        if ( costs == IMPOSSIBLE_TRANSITION )
        {
            costs = hardInconsistencyCosts;
        }
        task.preconditionCosts = costs;

        if ( task.tempVariable != null )
        {
            if ( costs != 0 )
            {
                src.costs[ src.PRECONDITION_FAILURE ] += costs;
                if ( src.costFunctionReference[ src.PRECONDITION_FAILURE ]
                    != null )
                {
                    src.costFunctionReference[ src.PRECONDITION_FAILURE ].updateGlobalCosts();
                }
            }
            task.preconditionFailureLE
                = preconditionFailureList.addListElement( task );
        }
        else
        {
            if ( costs != 0 )
            {
                src.costs[ src.STATIC_GOAL_FAILURE ] += costs;
                if ( src.costFunctionReference[ src.STATIC_GOAL_FAILURE ]
                    != null )
                {
                    src.costFunctionReference[ src.STATIC_GOAL_FAILURE ].updateGlobalCosts();
                }
            }
            if ( staticGoalFailureList == null )
            {
                staticGoalFailureList = new OrderedPreconditionFailureList();
            }
            task.preconditionFailureLE
                = staticGoalFailureList.addListElement( task );
        }



    }


    /**
     * Removes a precondition task from the interval structures and updates
     * its costs.
     * @param task  The precondition task to be removed.
     */
    private void removePreconditionTaskFromProjection( SymSRCPreconditionTask
        task )
    {

        ( (SymSRCInterval)task.intervalReference.getElement() ).preconditionTasks.remove( task );
        preconditionFailureList.removeListElement( task.preconditionFailureLE );
        if ( task.preconditionCosts != 0 )
        {
            src.costs[ src.PRECONDITION_FAILURE ] -= task.preconditionCosts;
            if (src.costFunctionReference[ src.PRECONDITION_FAILURE ] != null )
            {
                src.costFunctionReference[ src.PRECONDITION_FAILURE ].updateGlobalCosts();
            }
        }

    }


    /***** IMPROVEMENT - SERVICE FUNCTIONS *****/


    /**
     * Realizes an event at a specific time by adding a new task constraint
     * with the required state task. Synergy event are not handled, i.e.,
     * the function must be called multiple times with the event's sub-events
     * for this.
     * @param event  The event to be realized.
     * @param actionTypeID  The ID of the action type to be used.
     * @param time  The time at which the event should take place.
     */
    private void realizeEvent( SymSRCEvent event , int actionTypeID ,
        long time )
    {

        // Creating the state task

        Variable v1 = gsc.createVariable_internal( "Temporal Reference" );
        gsc.setVariable_internal( v1 , time );
        Variable v2 = gsc.createVariable_internal( "Contribution" );
        gsc.setVariable_internal( v2 , event.ID );

        Embedding e = new Embedding( 3 );
        gsc.extendEmbedding_internal( e , v1 ,
            Registry.TEMPORALREFERENCE_EDGE , Registry.INCOMING );
        gsc.extendEmbedding_internal( e , v2 ,
            Registry.CONTRIBUTION_EDGE , Registry.INCOMING );
        gsc.extendEmbedding_internal( e , src.resourceTypeVariable ,
            Registry.RESOURCETYPE_EDGE , Registry.INCOMING );
        ConventionalConstraint task = gsc.createConstraint_internal( Registry.STATETASK_OBJECTCONSTRAINT , e );

        // Creating the action object

        v1 = gsc.createVariable_internal( "Action Type" );
        gsc.setVariable_internal( v1 , actionTypeID );

        e = new Embedding( 1 );
        gsc.extendEmbedding_internal( e , v1 ,
            Registry.ACTIONTYPE_EDGE , Registry.INCOMING );
        ConventionalConstraint actionConstraint = gsc.createConstraint_internal( Registry.ACTION_OBJECTCONSTRAINT , e );

        // Connecting the task to the action object and to this
        // state resource constraint

        e = new Embedding( 1 );
        gsc.extendEmbedding_internal( e , task ,
            Registry.PLAIN_EDGE , Registry.INCOMING );
        gsc.extendConstraint_internal( actionConstraint , e );
        gsc.extendConstraint_internal( src , e );

        // Creating the task constraint

        e = new Embedding( 2 );
        gsc.extendEmbedding_internal( e , src.currentTimeConstraint ,
            Registry.PLAIN_EDGE , Registry.INCOMING );
        gsc.extendEmbedding_internal(e , actionConstraint ,
            Registry.PLAIN_EDGE , Registry.INCOMING );

        // TO BE IMPROVED: Cost function selection should be decided
        // by the user!

        gsc.selectCostFunction_internal( gsc.getStructuralCostFunction() );
        ConventionalConstraint taskConstraint
            = gsc.createConstraint_internal( Registry.TASK_CONSTRAINT , e );
        gsc.addCosts( taskConstraint ,
            TaskConstraint.TRANSITION_TEMPORAL_INCONSISTENCY );
        gsc.selectCostFunction_internal( gsc.getDefaultCostFunction() );
        gsc.addCosts( taskConstraint , TaskConstraint.TEMPORAL_INCONSISTENCY );

    }


    /***** PRECONDITION FAILURE IMPROVEMENT *****/


    /**
     * An event that changes the state projection is introduced in the
     * most possible earlier or same interval of a selected inconsistent
     * precondition task to lower its costs. Returns whether the heuristic
     * was able to succeed.
     * @param costs  The costs to be lowered.
     * @param list  The list of precondition tasks to choose from.
     */
    private boolean heuristic_AddEvent( long costs ,
        OrderedPreconditionFailureList list )
    {

        // Selecting an inconsistent precondition task

        long selectedCosts = Math.abs( gsc.random.nextLong() ) % costs;

        OrderedShiftDLLElement currentLE = list.first;
        SymSRCPreconditionTask task = null;
        boolean notFound = true;

        while ( notFound )
        {
            task = (SymSRCPreconditionTask)currentLE.element;
            selectedCosts -= task.preconditionCosts;
            if ( notFound = ( selectedCosts >= 0 ) )
            {
                currentLE = currentLE.next;
            }
        }

        if ( task.tempValue <= src.currentTime )
        {
            return false;
        }
        IntervalListElement ile = task.intervalReference;
        if ( task.tempValue == ile.getBeginning() )
        {
            return false;
        }
        if ( ( (SymSRCState)resourceType.states.get( task.state ) ).predecessorStates == null )
        {
            return false;
        }

        // Selecting an interval for the event insertion

        SymSRCState state = ( (SymSRCInterval)ile.getElement() ).state;
        long availableTime;
        if ( src.currentTime <= ile.getBeginning() )
        {
            availableTime = task.tempValue - ile.getBeginning() - 1;
        }
        else
        {
            availableTime = task.tempValue - src.currentTime;
        }

        while ( ( resourceType.transitionDifficulty[ state.ID ][ task.state ] == IMPOSSIBLE_TRANSITION ) || ( availableTime < resourceType.transitionDuration[ state.ID ][ task.state ] ) )
        {
            ile = ile.getPrevious();
            if ( ( ile == null ) || ( ile.getEnd() <= src.currentTime + 1 ) )
            {
                return false;
            }
            state = ( (SymSRCInterval)ile.getElement() ).state;
            if ( src.currentTime <= ile.getBeginning() )
            {
                availableTime = task.tempValue - ile.getBeginning() - 1;
            }
            else
            {
                availableTime = task.tempValue - src.currentTime;
            }
        }

        // Selecting a transition to be added

        SymSRCStateTransition transition = null;
        SymSRCInterval interval = (SymSRCInterval)ile.getElement();
        selectedCosts = 0;
        long neededTime = 0;
        long t;

        Iterator it = interval.state.successorStatesCollection.iterator();
        while ( it.hasNext() )
        {
            transition = (SymSRCStateTransition)it.next();
            t = resourceType.transitionDifficulty[ transition.successorState.ID ][ task.state ] + transition.difficulty;
            neededTime = resourceType.transitionDuration[ transition.successorState.ID ][ task.state ] + transition.duration;
            // Not including dead-ends
            if ( ( t != IMPOSSIBLE_TRANSITION )
                && ( availableTime >= neededTime ) )
            {
                selectedCosts += (long)( 100000.0 / ( t * t * t ) );
            }
        }

        if ( selectedCosts == 0 )
        {
            return false;
        }

        selectedCosts = Math.abs( gsc.random.nextLong() ) % selectedCosts;

        it = interval.state.successorStatesCollection.iterator();
        while ( selectedCosts >= 0 )
        {
            transition = (SymSRCStateTransition)it.next();
            t = resourceType.transitionDifficulty[ transition.successorState.ID ][ task.state ] + transition.difficulty;
            neededTime = resourceType.transitionDuration[ transition.successorState.ID ][ task.state ] + transition.duration;
            if ( ( t != IMPOSSIBLE_TRANSITION )
                && ( availableTime >= neededTime ) )
            {
                selectedCosts -= (long)( 100000.0 / ( t * t * t ) );
            }
        }

        // Selecting a time point for event insertion; the time point will
        // be somewhere between t1 and t2

        long t1 = ile.getBeginning();
        long t2;

        if ( ile == task.intervalReference )
        {

            if ( interval.preconditionTasks.size() > 1 )
            {

                // Setting t1 to the last precondition task before the task

                SymSRCPreconditionTask otherTask;
                it = interval.preconditionTasks.iterator();

                while ( it.hasNext() )
                {
                    otherTask = (SymSRCPreconditionTask)it.next();
                    if ( task != otherTask )
                    {
                        t = otherTask.tempValue;
                        if ( ( t > t1 ) && ( t <= task.tempValue ) )
                        {
                            t1 = t;
                        }
                    }
                }

            }

            // Setting t2 to the last possible point

            t2 = task.tempValue - resourceType.transitionDuration[ transition.successorState.ID ][ task.state ];

        }
        else
        {

            // Setting t1 to the last precondition task of the interval

            it = interval.preconditionTasks.iterator();
            while ( it.hasNext() ) {
                t = ( (SymSRCPreconditionTask)it.next() ).tempValue;
                if ( t > t1 )
                {
                    t1 = t;
                }
            }

            // Setting t2 to the last possible time point of the interval

            t2 = ile.getEnd() - 1;
            if ( t2 > task.tempValue - resourceType.transitionDuration[ transition.successorState.ID ][ task.state ] )
            {
                t2 = task.tempValue - resourceType.transitionDuration[ transition.successorState.ID ][ task.state ];
            }

        }

        if ( t1 < src.currentTime )
        {
            t1 = src.currentTime;
        }

        // Calculating a balanced point between t1 and t2 involving
        // times to establish intermediate states based on the given
        // difficulty

        selectedCosts = transition.difficulty + resourceType.transitionDifficulty[ transition.successorState.ID ][ task.state ];
        t = t1 + Math.round( ( t2 - t1 ) / ( 1.0 + selectedCosts ) );
        if ( t == t1 )
        {
            t++;
            if ( t > t2 )
            {
                t = t2;
            }
        }

        // Adding the event

        if ( Registry.LOG_CHANGES )
        {
            System.out.println( "heuristic_AddEvent: "
                + transition.successorState + " at " + t );
        }

        if ( transition.event.subEvents == null )
        {
            realizeEvent( transition.event , transition.actionTypeID , t );
        }
        else
        {
            it = transition.event.subEvents.iterator();
            while ( it.hasNext() )
            {
                realizeEvent( (SymSRCEvent)it.next() , -1 , t );
            }
        }

        return true;

    }


    /**
     * An event is moved by one interval to a time point before an
     * unsatisfied precondition task to improve its costs. Returns
     * whether the heuristic was able to succeed.
     * @param costs  The costs to be lowered.
     * @param list  The list of tasks to choose from.
     */
    private boolean heuristic_MoveEvent( long costs ,
        OrderedPreconditionFailureList list )
    {

        // Selecting an inconsistent precondition task

        long selectedCosts = Math.abs( gsc.random.nextLong() ) % costs;

        OrderedShiftDLLElement currentLE = list.first;
        SymSRCPreconditionTask task = null;
        boolean notFound = true;

        while ( notFound )
        {
            task = (SymSRCPreconditionTask)currentLE.element;
            selectedCosts -= task.preconditionCosts;
            if ( notFound = ( selectedCosts >= 0 ) )
            {
                currentLE = currentLE.next;
            }
        }

        if ( task.tempValue <= src.currentTime )
        {
            return false;
        }
        IntervalListElement ile = task.intervalReference;
        if ( task.tempValue == ile.getBeginning() )
        {
            return false;
        }

        // Evaluating predecessor event movement option

        Iterator it;
        SymSRCState state;
        SymSRCStateTransition transition;
        SymSRCEvent event;
        long difficulty1 = 0;
        long difficulty2 = 0;
        long duration1insert = 0;
        long duration2insert = 0;
        long duration1left = 0;
        long duration2left = 0;
        long earliestBeginning = ile.getBeginning();
        if ( earliestBeginning < src.currentTime )
        {
            earliestBeginning = src.currentTime;
        }
        long latestEnd = task.tempValue;

        boolean previousIntervalOK = ( ( ile.getPrevious() != null ) && ( ile.getPrevious().getBeginning() > src.currentTime ) );

        if ( previousIntervalOK )
        {

            // Checking if state tasks do not involve a world transition

            it = ( (SymSRCInterval)ile.getPrevious().getElement() ).stateTasks.iterator();
            while ( it.hasNext() && previousIntervalOK )
            {
                previousIntervalOK
                    = !( (SymSRCStateTask)it.next() ).partOfWorldTransition;
            }

        }

        if ( previousIntervalOK )
        {

            // Determining new interval state in case of movement

            state = ( (SymSRCInterval)ile.getPrevious().getPrevious().getElement() ).state;

            event = ( (SymSRCInterval)ile.getElement() ).event;
            if ( event != null )
            {
                if ( state.successorStates != null )
                {
                    transition = (SymSRCStateTransition)state.successorStates.get( event );
                    if ( transition != null )
                    {
                        state = transition.successorState;
                    }
                }
            }

            event = ( (SymSRCInterval)ile.getPrevious().getElement() ).event;
            if ( event != null )
            {
                if ( state.successorStates != null )
                {
                    transition = (SymSRCStateTransition)state.successorStates.get( event );
                    if ( transition != null )
                    {
                        state = transition.successorState;
                        duration1insert = transition.duration;
                    }
                }
            }

            difficulty1
                = resourceType.transitionDifficulty[ state.ID ][ task.state ];
            if ( difficulty1 == IMPOSSIBLE_TRANSITION )
            {
                previousIntervalOK = false;
            }
            else
            {
                difficulty1 = task.preconditionCosts - difficulty1;
                if ( difficulty1 <= 0 )
                {
                    previousIntervalOK = false;
                }
                else
                {
                    duration1left = resourceType.transitionDuration[ state.ID ][ task.state ];
                    if ( latestEnd - earliestBeginning
                        < duration1insert + duration1left )
                    {
                        previousIntervalOK = false;
                    }
                }
            }

        }

        // Evaluating successor event movement options

        boolean nextIntervalOK = ( ile.getNext() != null );

        if ( nextIntervalOK )
        {

            // Checking if state tasks do not involve a world transition

            it = ( (SymSRCInterval)ile.getNext().getElement() ).stateTasks.iterator();
            while ( it.hasNext() && nextIntervalOK )
            {
                nextIntervalOK
                    = !( (SymSRCStateTask)it.next() ).partOfWorldTransition;
            }

        }

        if ( nextIntervalOK )
        {

            // Determining new interval state in case of movement

            state = ( (SymSRCInterval)ile.getElement() ).state;

            event = ( (SymSRCInterval)ile.getNext().getElement() ).event;
            if ( event != null )
            {
                if ( state.successorStates != null )
                {
                    transition = (SymSRCStateTransition)state.successorStates.get( event );
                    if ( transition != null )
                    {
                        state = transition.successorState;
                        duration2insert = transition.duration;
                    }
                }
            }

            difficulty2
                = resourceType.transitionDifficulty[ state.ID ][ task.state ];
            if ( difficulty2 == IMPOSSIBLE_TRANSITION )
            {
                nextIntervalOK = false;
            }
            else
            {
                difficulty2 = task.preconditionCosts - difficulty2;
                if ( difficulty2 <= 0 )
                {
                    nextIntervalOK = false;
                }
                else
                {
                    duration2left = resourceType.transitionDuration[ state.ID ][ task.state ];
                    if ( latestEnd - earliestBeginning
                        < duration2insert + duration2left )
                    {
                        nextIntervalOK = false;
                    }
                }
            }

        }

        if ( !previousIntervalOK && !nextIntervalOK )
        {
            return false;
        }

        // Selecting an event to be moved if both are possible

        if ( previousIntervalOK && nextIntervalOK )
        {

            difficulty1 = difficulty1 * difficulty1 * difficulty1;
            difficulty2 = difficulty2 * difficulty2 * difficulty2;

            selectedCosts = Math.abs( gsc.random.nextLong() )
                % ( difficulty1 + difficulty2 );
            if ( selectedCosts < difficulty1 )
            {
                nextIntervalOK = false;
            }
            else
            {
                previousIntervalOK = false;
            }

        }

        // Determining point to move involving times to establish
        // intermediate states based on the given difficulty

        SymSRCInterval interval;
        if ( previousIntervalOK )
        {
            interval = (SymSRCInterval)ile.getPrevious().getElement();
            selectedCosts = task.preconditionCosts - difficulty1;
            earliestBeginning += duration1insert;
            latestEnd -= duration1left;
        }
        else
        {
            interval = (SymSRCInterval)ile.getNext().getElement();
            selectedCosts = task.preconditionCosts - difficulty2;
            earliestBeginning += duration2insert;
            latestEnd -= duration2left;
        }

        long t = earliestBeginning + Math.round( ( latestEnd - earliestBeginning ) / ( 1.0 + selectedCosts ) );
        if ( t == ile.getBeginning() )
        {
            t++;
        }

        // Moving the selected event, i.e., the involved state tasks

        if ( Registry.LOG_CHANGES )
        {
            System.out.println( "heuristic_MoveEvent: "
                + interval.stateTasks + " to " + t );
        }

        it = interval.stateTasks.iterator();
        while ( it.hasNext() )
        {
            gsc.setVariable_internal( ( (SymSRCStateTask)it.next() ).tempVariable , t );
        }

        return true;

    }


    /**
     * A precondition task is moved to improve its costs. Only locations with
     * improved costs are considered.
     * @return  Returns whether the heuristic was able to succeed.
     */
    private boolean heuristic_MovePreconditionTask()
    {

        // Selecting an inconsistent precondition task

        long costs = Math.abs( gsc.random.nextLong() )
            % src.costs[ src.PRECONDITION_FAILURE ];

        OrderedShiftDLLElement currentLE = preconditionFailureList.first;
        SymSRCPreconditionTask task = null;
        boolean notFound = true;

        while ( notFound )
        {
            task = (SymSRCPreconditionTask)currentLE.element;
            costs -= task.preconditionCosts;
            if ( notFound = ( costs >= 0 ) )
            {
                currentLE = currentLE.next;
            }
        }

        // Analyzing predecessor interval

        int newState;
        long difficulty1 = 0;
        long difficulty2 = 0;
        long difficultyThis;
        IntervalListElement ile = task.intervalReference;

        IntervalListElement previousILE = ile;
        boolean previousIntervalOK = false;

        while ( !previousIntervalOK && ( previousILE.getPrevious() != null ) )
        {
            previousILE = previousILE.getPrevious();
            newState = ( (SymSRCInterval)previousILE.getElement() ).state.ID;
            difficulty1
                = resourceType.transitionDifficulty[ newState ][ task.state ];
            previousIntervalOK = ( ( newState != task.state )
                && ( difficulty1 != IMPOSSIBLE_TRANSITION ) );
        }

        // Analyzing successor interval

        IntervalListElement nextILE = ile;
        boolean nextIntervalOK = false;

        while ( !nextIntervalOK && ( nextILE.getNext() != null ) )
        {
            nextILE = nextILE.getNext();
            newState = ( (SymSRCInterval)nextILE.getElement() ).state.ID;
            difficulty2
                = resourceType.transitionDifficulty[ newState ][ task.state ];
            nextIntervalOK = ( ( newState != task.state )
                && ( difficulty2 != IMPOSSIBLE_TRANSITION ) );
        }

        // Choosing the interval for the movement

        if ( !previousIntervalOK && !nextIntervalOK )
        {
            return false;
        }

        costs = 0;
        if ( previousIntervalOK )
        {
            difficulty1++ ;
            difficulty1 = (long)( 100000.0 / ( difficulty1 * difficulty1 * difficulty1 ) );
            costs = difficulty1;
        }
        if ( nextIntervalOK )
        {
            difficulty2++ ;
            difficulty2 = (long)( 100000.0 / ( difficulty2 * difficulty2 * difficulty2 ) );
            costs += difficulty2;
        }

        costs = Math.abs( gsc.random.nextLong() ) % costs;
        if ( previousIntervalOK && ( costs < difficulty1 ) )
        {
            ile = previousILE;
        }
        else
        {
            ile = nextILE;
        }

        // Moving precondition task into the middle of the chosen interval

        if ( ile != intervalList.getLast() )
        {
            if ( Registry.LOG_CHANGES )
            {
                System.out.println( "heuristic_MovePreconditionTask: "
                    + task + " to " + ( ile.getBeginning()
                    + ( ile.getEnd() - ile.getBeginning() ) / 2 ) );
            }
            gsc.setVariable_internal( task.tempVariable , ile.getBeginning()
                + ( ile.getEnd() - ile.getBeginning() ) / 2 );
        }
        else
        {
            if ( Registry.LOG_CHANGES )
            {
                System.out.println( "heuristic_MovePreconditionTask: "
                    + task + " to " + ile.getBeginning() );
            }
            gsc.setVariable_internal( task.tempVariable , ile.getBeginning() );
        }

        return true;

    }


    /**
     * An event is deleted to improve a precondition task's costs. Returns
     * whether the heuristic was able to succeed.
     * @param costs  The costs to be lowered.
     * @param list  The list of tasks to choose from.
     */
    private boolean heuristic_DeleteEvent( long costs ,
        OrderedPreconditionFailureList list ) {

        // Selecting an inconsistent precondition task

        long selectedCosts = Math.abs( gsc.random.nextLong() ) % costs;

        OrderedShiftDLLElement currentLE = list.first;
        SymSRCPreconditionTask task = null;
        boolean notFound = true;

        while ( notFound )
        {
            task = (SymSRCPreconditionTask)currentLE.element;
            selectedCosts -= task.preconditionCosts;
            if ( notFound = ( selectedCosts >= 0 ) )
            {
                currentLE = currentLE.next;
            }
            else
            {
                if ( ( task.intervalReference.getBeginning()
                    <= src.currentTime ) && ( currentLE.next != null ) )
                {
                    notFound = true;
                    currentLE = currentLE.next;
                }
            }
        }

        if ( ( task.intervalReference.getBeginning() <= src.currentTime )
            || ( task.preconditionCosts == 0 ) )
        {
            return false;
        }

        // Removing event's state tasks (a random one if multiple)

        SymSRCStateTask stateTask = null;
        HashSet tasks = ( (SymSRCInterval)task.intervalReference.getElement() ).stateTasks;
        selectedCosts = gsc.random.nextInt( tasks.size() );
        Iterator it = tasks.iterator();
        while ( selectedCosts-- >= 0 )
        {
            stateTask = (SymSRCStateTask)it.next();
        }

        // If state task is part of a world transition, delegate deletion

        if ( stateTask.partOfWorldTransition )
        {

            if ( transitionConstraint == null )
            {
                transitionConstraint = (TransitionConstraint)gsc.findConstraint( Registry.TRANSITION_CONSTRAINT );
                if ( transitionConstraint == null )
                {
                    return false;
                }
            }

            // Retrieving the related action object constraint

            ActionObjectConstraint actionConstraint;
            synchronized( stateTask.taskConstraint )
            {
                actionConstraint = (ActionObjectConstraint)stateTask.taskConstraint.connections.getSpecificConstraint( Registry.ACTION_OBJECTCONSTRAINT , Registry.PLAIN_EDGE );
            }
            if ( actionConstraint == null ) {
                throw new Error( "The StateTaskObjectConstraint \"" + stateTask.taskConstraint + "\" has no connected ActionObjectConstraint" );
            }

            // Delegating the action's deletion to the transition constraint

            if ( Registry.LOG_CHANGES )
            {
                System.out.println( "heuristic_DeleteEvent: delegated deletion of " + stateTask.contribution + " at " + stateTask.tempValue + " to TransitionConstraint" );
            }
            return transitionConstraint.delegatedHeuristic_DeleteEvent( actionConstraint );

        }

        // Removing task

        boolean success = src.removeTask( stateTask.taskConstraint , false );
        if ( Registry.LOG_CHANGES && success )
        {
            System.out.println( "heuristic_DeleteEvent: " + stateTask );
        }

        return success;

    }


    /**
     * An inconsistent precondition task is deleted. Returns whether the
     * heuristic was able to succeed.
     */
    private boolean heuristic_DeletePreconditionTask()
    {

        // Selecting an inconsistent precondition task

        long costs = Math.abs( gsc.random.nextLong() )
            % src.costs[ src.PRECONDITION_FAILURE ];

        OrderedShiftDLLElement currentLE = preconditionFailureList.first;
        SymSRCPreconditionTask task = null;
        boolean notFound = true;

        while (notFound) {
            task = (SymSRCPreconditionTask)currentLE.element;
            costs -= task.preconditionCosts;
            if ( notFound = ( costs >= 0 ) )
            {
                currentLE = currentLE.next;
            }
        }

        // Checking if precondition task is not part of a world transition

        if ( task.partOfWorldTransition )
        {
            return false;
        }

        // Deleting the precondition task

        boolean success = src.removeTask( task.taskConstraint , false );
        if ( Registry.LOG_CHANGES && success )
        {
            System.out.println( "heuristic_DeletePreconditionTask: "
                + task );
        }

        return success;

    }


    public void improvePreconditionInconsistency()
    {

        if ( Registry.LOG_CHANGES )
        {
            System.out.println( "improving precondition inconsistency:" );
        }

        BigInteger newCosts = src.costFunctionReference[ src.PRECONDITION_FAILURE ].costFunction.objectiveCosts.add( BigInteger.ZERO );
        int i;

        if ( lastPreconditionImprovementChoice != NO_HEURISTIC )
        {

            // Updating preference values

            if ( lastPreconditionImprovementCosts.compareTo( newCosts ) == 1 )
            {
                choicePreference[ lastPreconditionImprovementChoice ]++ ;
                if ( choicePreference[ lastPreconditionImprovementChoice ]
                    > MAX_PREFERENCE_VALUE )
                {
                    choicePreference[ lastPreconditionImprovementChoice ]
                        = MAX_PREFERENCE_VALUE;
                }
            }
            else
            {

                // Decreasing last choice's preference value or increasing
                // all other preference values

                if ( choicePreference[ lastPreconditionImprovementChoice ]
                    == 1 )
                {
                    /*
                    for ( i = 0 ; i < NUMBER_OF_HEURISTICS ; i++ )
                    {
                        if ( i != lastPreconditionImprovementChoice ) {
                            choicePreference[ i ]++ ;
                            if ( choicePreference[ i ] > MAX_PREFERENCE_VALUE )
                            {
                                choicePreference[ i ] = MAX_PREFERENCE_VALUE;
                            }
                        }
                    }
                    */

                }
                else
                {
                    choicePreference[ lastPreconditionImprovementChoice ] = (int)Math.round( Math.sqrt( choicePreference[ lastPreconditionImprovementChoice ] ) );
                }

            }

        }

        lastPreconditionImprovementCosts = newCosts;

        // Deciding about applicability of the single heuristics and
        // making a copy of the preference values

        int[] currentPreference = new int[ NUMBER_OF_HEURISTICS ];
        if ( gsc.structuralChangesEnabled )
        {
            currentPreference[ HEURISTIC_ADD_EVENT ]
                = choicePreference[ HEURISTIC_ADD_EVENT ];
        }
        else
        {
            currentPreference[ HEURISTIC_ADD_EVENT ] = 0;
        }
        if ( src.stateTaskNumber > 0 )
        {
            currentPreference[ HEURISTIC_MOVE_EVENT ]
                = choicePreference[ HEURISTIC_MOVE_EVENT ];
        }
        else
        {
            currentPreference[ HEURISTIC_MOVE_EVENT ] = 0;
        }
        currentPreference[ HEURISTIC_MOVE_PRECONDITION_TASK ]
            = choicePreference[ HEURISTIC_MOVE_PRECONDITION_TASK ];
        if ( ( src.stateTaskNumber > 0 ) && ( gsc.structuralChangesEnabled ) )
        {
            currentPreference[ HEURISTIC_DELETE_EVENT ]
                = choicePreference[ HEURISTIC_DELETE_EVENT ];
        }
        else
        {
            currentPreference[ HEURISTIC_DELETE_EVENT ] = 0;
        }
        if ( gsc.structuralChangesEnabled )
        {
            currentPreference[ HEURISTIC_DELETE_PRECONDITION_TASK ]
                = choicePreference[ HEURISTIC_DELETE_PRECONDITION_TASK ];
        }
        else
        {
            currentPreference[ HEURISTIC_DELETE_PRECONDITION_TASK ] = 0;
        }

        // Selection iterations until a heuristic was successfully applied

        int max;
        int maxOccurance;
        int random;
        boolean notReady = true;
        int iterationCounter = MAX_REPAIR_ATTEMPTS;

        while ( notReady && ( iterationCounter-- != 0 ) )
        {

            // Finding out the maximal preference value

            max = 0;
            for ( i = 0 ; i < NUMBER_OF_HEURISTICS ; i++ )
            {
                if ( currentPreference[ i ] > max )
                {
                    max = currentPreference[ i ];
                }
            }

            // Counting how often the maximal value occurs

            maxOccurance = 0;
            for ( i = 0 ; i < NUMBER_OF_HEURISTICS ; i++ )
            {
                if ( currentPreference[ i ] == max )
                {
                    maxOccurance++ ;
                }
            }

            // Selecting a heuristic

            random = gsc.random.nextInt( maxOccurance );
            lastPreconditionImprovementChoice = -1;
            while ( notReady )
            {
                if ( currentPreference[ ++lastPreconditionImprovementChoice ]
                    == max ) {
                    notReady = ( --random >= 0 );
                }
            }

            switch ( lastPreconditionImprovementChoice )
            {
            case HEURISTIC_ADD_EVENT:
                notReady = !heuristic_AddEvent( src.costs[ src.PRECONDITION_FAILURE ] , preconditionFailureList );
                break;
            case HEURISTIC_MOVE_EVENT:
                notReady = !heuristic_MoveEvent( src.costs[ src.PRECONDITION_FAILURE] , preconditionFailureList );
                break;
            case HEURISTIC_MOVE_PRECONDITION_TASK:
                notReady = !heuristic_MovePreconditionTask();
                break;
            case HEURISTIC_DELETE_EVENT:
                notReady = !heuristic_DeleteEvent( src.costs[ src.PRECONDITION_FAILURE ] , preconditionFailureList );
                break;
            case HEURISTIC_DELETE_PRECONDITION_TASK:
                notReady = !heuristic_DeletePreconditionTask();
            }

            // Adapting temporary preference if case of failed application

            if ( notReady )
            {
                currentPreference[ lastPreconditionImprovementChoice ] /= 2;
                if ( currentPreference[ lastPreconditionImprovementChoice ]
                    == 0 )
                {
                    currentPreference[ lastPreconditionImprovementChoice ] = 1;
                }
                notReady = gsc.structuralChangesEnabled;
            }

        }

        // Clearing choice memory if no heuristic could be applied

        if ( notReady )
        {
            lastPreconditionImprovementChoice = NO_HEURISTIC;
            if ( Registry.LOG_CHANGES )
            {
                System.out.println( "improvement failed - maximal number of tries failed" );
            }
        }

        fireChange( this );

    }


    /***** STATIC GOAL FAILURE IMPROVEMENT *****/


    public void improveGoalInconsistency()
    {

        if ( Registry.LOG_CHANGES )
        {
            System.out.println( "improving goal inconsistency:" );
        }

        BigInteger newCosts = src.costFunctionReference[ src.STATIC_GOAL_FAILURE ].costFunction.objectiveCosts.add( BigInteger.ZERO );
        int i;

        if ( lastGoalImprovementChoice != NO_HEURISTIC )
        {

            // Updating preference values

            if ( lastGoalImprovementCosts.compareTo( newCosts ) == 1 )
            {
                choicePreference[ lastGoalImprovementChoice ]++ ;
                if ( choicePreference[ lastGoalImprovementChoice ]
                    > MAX_PREFERENCE_VALUE )
                {
                    choicePreference[ lastGoalImprovementChoice ] =
                        MAX_PREFERENCE_VALUE;
                }
            }
            else
            {

                // Decreasing last choice's preference value or increasing
                // all other preference values

                if ( choicePreference[ lastGoalImprovementChoice ] == 1 )
                {
                    /*
                    for ( i = 0 ; i < NUMBER_OF_HEURISTICS ; i++ )
                    {
                        if ( i != lastGoalImprovementChoice ) {
                            choicePreference[ i ]++ ;
                            if ( choicePreference[ i ] > MAX_PREFERENCE_VALUE )
                            {
                                choicePreference[ i ] = MAX_PREFERENCE_VALUE;
                            }
                        }
                    }
                    */

                }
                else
                {
                    choicePreference[ lastGoalImprovementChoice ] = (int)Math.round( Math.sqrt( choicePreference[ lastGoalImprovementChoice ] ) );
                }
            }

        }

        lastGoalImprovementCosts = newCosts;

        // Deciding about applicability of the single heuristics and
        // making a copy of the preference values

        int[] currentPreference = new int[ NUMBER_OF_HEURISTICS ];
        if ( gsc.structuralChangesEnabled )
        {
            currentPreference[ HEURISTIC_ADD_EVENT ]
                = choicePreference[ HEURISTIC_ADD_EVENT ];
        }
        else
        {
            currentPreference[ HEURISTIC_ADD_EVENT ] = 0;
        }
        if ( src.stateTaskNumber > 0 )
        {
            currentPreference[ HEURISTIC_MOVE_EVENT ]
                = choicePreference[ HEURISTIC_MOVE_EVENT ];
        }
        else
        {
            currentPreference[ HEURISTIC_MOVE_EVENT ] = 0;
        }
        currentPreference[ HEURISTIC_MOVE_PRECONDITION_TASK ] = 0;
        if ( ( src.stateTaskNumber > 0 ) && ( gsc.structuralChangesEnabled ) )
        {
            currentPreference[ HEURISTIC_DELETE_EVENT ]
                = choicePreference[HEURISTIC_DELETE_EVENT];
        }
        else
        {
            currentPreference[ HEURISTIC_DELETE_EVENT ] = 0;
        }
        currentPreference[ HEURISTIC_DELETE_PRECONDITION_TASK ] = 0;

        // Selection iterations until a heuristic was successfully applied

        int max;
        int maxOccurance;
        int random;
        boolean notReady = true;
        int iterationCounter = MAX_REPAIR_ATTEMPTS;

        while ( notReady && ( iterationCounter-- != 0 ) )
        {

            // Finding out the maximal preference value

            max = 0;
            for ( i = 0 ; i < NUMBER_OF_HEURISTICS ; i++ )
            {
                if ( currentPreference[ i ] > max )
                {
                    max = currentPreference[ i ];
                }
            }

            // Counting how often the maximal value occurs

            maxOccurance = 0;
            for ( i = 0 ; i < NUMBER_OF_HEURISTICS ; i++ )
            {
                if ( currentPreference[ i ] == max )
                {
                    maxOccurance++ ;
                }
            }

            // Selecting a heuristic

            random = gsc.random.nextInt( maxOccurance );
            lastGoalImprovementChoice = -1;
            while ( notReady )
            {
                if ( currentPreference[ ++lastGoalImprovementChoice ]
                    == max )
                {
                    notReady = ( --random >= 0 );
                }
            }

            switch ( lastGoalImprovementChoice )
            {
            case HEURISTIC_ADD_EVENT:
                notReady = !heuristic_AddEvent( src.costs[ src.STATIC_GOAL_FAILURE ] , staticGoalFailureList );
                break;
            case HEURISTIC_MOVE_EVENT:
                notReady = !heuristic_MoveEvent( src.costs[ src.STATIC_GOAL_FAILURE] , staticGoalFailureList );
                break;
            case HEURISTIC_DELETE_EVENT:
                notReady = !heuristic_DeleteEvent( src.costs[ src.STATIC_GOAL_FAILURE ] , staticGoalFailureList );
                break;
            }

            // Adapting temporary preference if case of failed application

            if ( notReady )
            {
                currentPreference[ lastGoalImprovementChoice ] /= 2;
                if ( currentPreference[ lastGoalImprovementChoice ]
                    == 0 )
                {
                    currentPreference[ lastGoalImprovementChoice ] = 1;
                }
                notReady = gsc.structuralChangesEnabled;
            }

        }

        // Clearing choice memory if no heuristic could be applied

        if ( notReady )
        {
            lastGoalImprovementChoice = NO_HEURISTIC;
            if ( Registry.LOG_CHANGES )
            {
                System.out.println( "improvement failed - maximal number of tries failed" );
            }
        }

        fireChange( this );

    }


    /* (non-Javadoc)
     * @see com.ai_center.dragonbreath.constraints.stateresourceconstraint.SRCInterface#getResourceType()
     */
    public SRCResourceType getResourceType()
    {
        return (SRCResourceType)resourceType;
    }


    public IntervalList getIntervalList()
    {
        return intervalList;
    }


    public void addChangeListener( ChangeListener l )
    {
        listenerList.add( ChangeListener.class , (ChangeListener)l );
    }


    public void removeChangeListener( ChangeListener l )
    {
        listenerList.remove( ChangeListener.class , (ChangeListener)l );
    }


    /**
     * This method supports normal ChangeListeners.
     * @param changedObject The object which has changed.
     */
    protected void fireChange( Object changedObject )
    {
        ChangeEvent changeEvent = new ChangeEvent( changedObject );

        Object[] listeners = listenerList.getListenerList();
        ChangeListener listener = null;

        for ( int i = listeners.length - 2 ; i >= 0 ; i -= 2 )
        {
            if ( listeners[ i ] == ChangeListener.class )
            {
                ( (ChangeListener)listeners[ i + 1 ] ).stateChanged( changeEvent );
            }
        }
    }


    /**
     * @param listener Adds the listener to the changeListeners of the intervalList
     */
    public void addIntervalListChangeListener( ChangeListener listener )
    {
        intervalList.addChangeListener( listener );
    }


    public void removeIntervalListChangeListener( ChangeListener listener )
    {
        intervalList.removeChangeListener( listener );
    }


    /**
     * Returns all states, tasks and preconditions in the interval between beginTick
     * and endTick.
     * @param beginTick
     * @param endTick
     * @return List with IntervalListElements between the beginTick and the endTick.
     */
    public List getStatesAndTasksBetween( long beginTick , long endTick )
    {
        IntervalList intervalListCopy = intervalList;
        List intervalLinkedList = new LinkedList();

        if ( intervalListCopy != null )
        {

            IntervalListElement currentILE = intervalListCopy.getFirst();
            IntervalListElement lastILE = null;
            SymSRCInterval interval;

            boolean firstElement = true;

            while ( currentILE != null )
            {
                synchronized ( currentILE )
                {

                    if ( ( ( currentILE.getBeginning() >= beginTick )
                        && ( currentILE.getBeginning() <= endTick ) )
                        || ( ( currentILE.getEnd() >= beginTick )
                        && ( currentILE.getEnd() <= endTick ) )
                        || ( ( currentILE.getBeginning() < beginTick )
                        && ( currentILE.getEnd() > endTick) ) )
                    {

                        if ( firstElement )
                        {
                            // first element was found now.
                            firstElement = false;

                            // check if there is an element, with beginning
                            // outside the interval, but ending inside.

                            if ( lastILE != null )
                            {
                                // add the predecessor element
                                intervalLinkedList.add( lastILE );
                            }
                        }

                        // this element begins in the interval
                        intervalLinkedList.add( currentILE );

                    }

                    lastILE = currentILE;
                    currentILE = currentILE.getNext();
                    // caution deadlock: two currentILE locked or
                    // mutualaccessexception
                } // end synchronized
            }
        }
        return intervalLinkedList;
    }


    /**
     *
     */
    public void fireIntervalListUpdate()
    {
        intervalList.fireIntervalListUpdate();

    }


}
