
package com.ai_center.dragonbreath.constraints.stateresourceconstraint.symbolicsrc;


import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;

import com.ai_center.dragonbreath.constraints.stateresourceconstraint.SRCResourceType;
import com.ai_center.dragonbreath.constraints.stateresourceconstraint.SymbolicSRC;


/**
 * A SymbolicSRC's internal representation of a user-registered
 * potential resource type.
 *
 * @author Alexander Nareyek
 * @version $Revision: 1.10 $
 */
public class SymSRCResourceType extends SRCResourceType
{


    // **********************
    // ***** ATTRIBUTES *****
    // **********************


    /**
     * The corresponding value of the resource type variable.
     */
    public int ID;

    /**
     * The possible states of the state resource.
     */
    public ArrayList states = new ArrayList();

    /**
     * The default state to be used for unknown states.
     */
    public SymSRCState defaultState = null;

    /**
     * The possible events that can change the state.
     */
    public ArrayList events = new ArrayList();

    /**
     * The table of (direct and indirect) transition difficulties between
     * states.
     */
    public long[][] transitionDifficulty;
    /**
     * The sum of all transition difficulties.
     */
    public long difficultySum = 0;

    /**
     * The table of (direct and indirect) transition durations between
     * states.
     */
    public long[][] transitionDuration;


    // ************************
    // ***** CONSTRUCTORS *****
    // ************************


    /**
     * Initializes the resource type.
     * @param name  The name for the resource type.
     */
    public SymSRCResourceType( String name )
    {
        this.name = name;
    }


    // *******************
    // ***** METHODS *****
    // *******************


    /**
     * Returns a SymSRCState object for a state ID.
     * @param stateID  The ID of the state to be retrieved.
     */
    public SymSRCState getState( int stateID )
    {
        return (SymSRCState)states.get( stateID );
    }


    /**
     * Returns a SymSRCEvent object for an event ID.
     * @param eventID  The ID of the event to be retrieved.
     */
    public SymSRCEvent getEvent( int eventID )
    {
        return (SymSRCEvent)events.get( eventID );
    }


    /**
     * Returns a string with the transition difficulty's data.
     */
    public String getTransitionDifficultyDescription()
    {

        StringBuffer s = new StringBuffer();
        int size = states.size();
        int from;
        int to;

        if ( transitionDifficulty != null )
        {

            s.append( "  " );
            for ( to = 0 ; to < size ; to++ )
            {
                if ( to < 10 )
                {
                    s.append( "  0" + to );
                }
                else
                {
                    s.append( "  " + to );
                }
            }
            s.append( "\n" );

            for ( from = 0 ; from < size ; from++ )
            {
                if ( from < 10 )
                {
                    s.append( "0" + from );
                }
                else
                {
                    s.append( from );
                }
                for ( to = 0 ; to < size ; to++ )
                {
                    s.append( " " );
                    if ( transitionDifficulty[ from ][ to ] < 100 )
                    {
                        s.append( " " );
                        if ( ( transitionDifficulty[ from ][ to ] < 10 )
                            && ( transitionDifficulty[ from ][ to ] != -1 ) )
                        {
                            s.append( " " );
                        }
                    }
                    s.append( transitionDifficulty[ from ][ to ] );
                }
                s.append( "\n" );
            }

        }
        else
        {
            s.append( "no difficulty table existent" );
        }

        return s.toString();

    }


    /**
     * Enlarges the transition-related tables by one state.
     */
    public void increaseTransitionTableSize()
    {

        int newSize = states.size();
        long[][] newTransitionDifficulty = new long[ newSize ][ newSize ];
        long[][] newTransitionDuration = new long[ newSize ][ newSize ];
        newSize-- ;

        newTransitionDifficulty[ newSize ][ newSize ] = 0;
        newTransitionDuration[ newSize ][ newSize ] = 0;
        if ( newSize > 0 )
        {
            for ( int from = 0 ; from < newSize ; from++ )
            {
                for ( int to = 0 ; to < newSize ; to++ )
                {
                    newTransitionDifficulty[ from ][ to ]
                        = transitionDifficulty[ from ][ to ];
                    newTransitionDuration[ from ][ to ]
                        = transitionDuration[ from ][ to ];
                }
            }
            for ( int fromto = 0 ; fromto < newSize ; fromto++ )
            {
                newTransitionDifficulty[ newSize ][ fromto ]
                    = SymbolicSRC.IMPOSSIBLE_TRANSITION;
                newTransitionDifficulty[ fromto ][ newSize ]
                    = SymbolicSRC.IMPOSSIBLE_TRANSITION;
                newTransitionDuration[ newSize ][ fromto ]
                    = SymbolicSRC.IMPOSSIBLE_TRANSITION;
                newTransitionDuration[ fromto ][ newSize ]
                    = SymbolicSRC.IMPOSSIBLE_TRANSITION;
            }
        }

        transitionDifficulty = newTransitionDifficulty;
        transitionDuration = newTransitionDuration;

    }


    /**
     * Updates the transitions' table information when a state
     * transition was added.
     * @param seedTransition  The transition that was added.
     */
    public void propagateDifficulty( SymSRCStateTransition seedTransition )
    {

        int predecessorStateID = seedTransition.predecessorState.ID;
        int successorStateID = seedTransition.successorState.ID;

        // Updating difficulty table

        long currentDifficulty
            = transitionDifficulty[ predecessorStateID ][ successorStateID ];

        if ( ( currentDifficulty > seedTransition.difficulty ) ||
            ( currentDifficulty == SymbolicSRC.IMPOSSIBLE_TRANSITION ) )
        {

            // Updating direct transition difficulty

            transitionDifficulty[ predecessorStateID ][ successorStateID ]
                = seedTransition.difficulty;

            SymSRCState fromState;
            SymSRCState intermediateState;
            SymSRCState toState;
            SymSRCStateTransition stateTransition;
            Collection transitionsToBeChecked;
            HashSet newTransitionsToBeChecked;
            HashSet cache;
            Iterator it;
            Iterator it2;
            long newDifficulty;
            int max = states.size();
            boolean[] verified = new boolean[max];

            for ( int from = 0 ; from < max ; from++ )
            {

                fromState = (SymSRCState)states.get( from );
                if ( fromState.successorStates != null )
                {

                    // Initializing tracking to which states routes
                    // were already verified

                    for ( int to = 0 ; to < max ; to++ )
                    {
                        verified[ to ] = false;
                    }
                    verified[ from ] = true;
                    transitionsToBeChecked
                        = fromState.successorStatesCollection;

                    // Chaining from base state along all possible
                    // forward routes

                    while ( transitionsToBeChecked.size() != 0 )
                    {
                        newTransitionsToBeChecked = new HashSet();
                        it = transitionsToBeChecked.iterator();
                        while ( it.hasNext() )
                        {
                            stateTransition
                                = (SymSRCStateTransition)it.next();
                            toState = stateTransition.successorState;

                            // Updating difficulty information from->to

                            intermediateState
                                = stateTransition.predecessorState;
                            newDifficulty = transitionDifficulty[ from ][ intermediateState.ID ] + stateTransition.difficulty;
                            currentDifficulty
                                = transitionDifficulty[ from ][ toState.ID ];
                            if ( (currentDifficulty == SymbolicSRC.IMPOSSIBLE_TRANSITION ) || ( newDifficulty < currentDifficulty ) )
                            {
                                transitionDifficulty[ from ][ toState.ID ]
                                    = newDifficulty;
                            }

                            if ( !verified[toState.ID] )
                            {

                                // Setting verification mark

                                verified[ toState.ID ] = true;

                                // Adding following transitions to
                                // verification cue

                                if ( toState.successorStates != null )
                                {
                                    newTransitionsToBeChecked.addAll( toState.successorStatesCollection );
                                }

                            }

                        }

                        // Switching to verification of next transition
                        // distance

                        transitionsToBeChecked
                            = (Collection)newTransitionsToBeChecked;
                    }
                }
            }

        }

        // Updating distance table

        long currentDuration
            = transitionDuration[ predecessorStateID ][ successorStateID ];

        if ( ( currentDuration > seedTransition.duration )
            || ( currentDuration == SymbolicSRC.IMPOSSIBLE_TRANSITION ) )
        {

            // Updating direct transition duration

            transitionDuration[ predecessorStateID ][ successorStateID ]
                = seedTransition.duration;

            SymSRCState fromState;
            SymSRCState intermediateState;
            SymSRCState toState;
            SymSRCStateTransition stateTransition;
            Collection transitionsToBeChecked;
            HashSet newTransitionsToBeChecked;
            HashSet cache;
            Iterator it;
            Iterator it2;
            long newDuration;
            int max = states.size();
            boolean[] verified = new boolean[ max ];

            for ( int from = 0 ; from < max ; from++ )
            {
                fromState = (SymSRCState)states.get( from );
                if ( fromState.successorStates != null )
                {

                    // Initializing tracking to which states routes
                    // were already verified

                    for ( int to = 0 ; to < max ; to++ )
                    {
                        verified[ to ] = false;
                    }
                    verified[ from ] = true;
                    transitionsToBeChecked
                        = fromState.successorStatesCollection;

                    // Chaining from base state along all possible
                    // forward routes

                    while ( transitionsToBeChecked.size() != 0 )
                    {
                        newTransitionsToBeChecked = new HashSet();
                        it = transitionsToBeChecked.iterator();
                        while ( it.hasNext() )
                        {
                            stateTransition
                                = (SymSRCStateTransition)it.next();
                            toState = stateTransition.successorState;

                            // Updating duration information from->to

                            intermediateState
                                = stateTransition.predecessorState;
                            newDuration = transitionDuration[ from ][ intermediateState.ID ] + stateTransition.duration;
                            currentDuration
                                = transitionDuration[ from ][ toState.ID ];
                            if ( ( currentDuration == SymbolicSRC.IMPOSSIBLE_TRANSITION ) || ( newDuration < currentDuration ) )
                            {
                                transitionDuration[ from ][ toState.ID ]
                                    = newDuration;
                            }

                            if ( !verified[ toState.ID ] )
                            {

                                // Setting verification mark

                                verified[ toState.ID ] = true;

                                // Adding following transitions to
                                // verification cue

                                if ( toState.successorStates != null )
                                {
                                    newTransitionsToBeChecked.addAll( toState.successorStatesCollection );
                                }

                            }

                        }

                        // Switching to verification of next transition
                        // distance

                        transitionsToBeChecked
                            = (Collection)newTransitionsToBeChecked;
                    }
                }
            }

        }

    }


    /**
     * Calculates the event outcome if multiple state tasks are placed
     * at the same time.
     * @param stateTasks  The set of the concurrent state tasks.
     */
    public SymSRCEvent calculateSynergyOutcome( HashSet stateTasks )
    {

        // Retrieving synergy events events of one task

        int stateTaskNumber = stateTasks.size();
        Iterator itTasks = stateTasks.iterator();
        SymSRCStateTask task = (SymSRCStateTask)itTasks.next();
        LinkedList potentialEvents = task.contribution.synergyEvents;
        LinkedList clonedSubEvents;

        // Checking if one of the events has all of the state tasks'
        // contributions as sub-events

        if ( potentialEvents != null )
        {

            Iterator itEvents = potentialEvents.iterator();
            boolean found = false;
            SymSRCEvent event = null;

            while ( !found && itEvents.hasNext() )
            {
                event = (SymSRCEvent)itEvents.next();
                if ( event.subEvents.size() == stateTaskNumber )
                {

                    // Checking for complete task match

                    clonedSubEvents = (LinkedList)event.subEvents.clone();
                    itTasks = stateTasks.iterator();
                    found = true;
                    while ( found && itTasks.hasNext() )
                    {
                        task = (SymSRCStateTask)itTasks.next();
                        found = clonedSubEvents.remove( task.contribution );
                    }

                }
            }

            if ( found )
            {
                return event;
            }
            else
            {
                return null;
            }

        } else {
            return null;
        }
    }


}
