
package com.ai_center.dragonbreath.constraints.taskconstraint;

/** 
 * A TaskConstraint's internal representation of a temporal relation.
 * The represented relation is "var1 comparator var2 + const".
 *
 * @author Alexander Nareyek
 * @version $Revision: 2.4 $
 */


import com.ai_center.dragonbreath.constraints.TaskConstraint;
import com.ai_center.dragonbreath.utilities.OrderedShiftDLLElement;


public class TCTemporalRelationInstance {


    // **********************
    // ***** ATTRIBUTES *****
    // **********************
    
    /**
     * The internal task instance of the first variable.
     */
    public Object task1;
    /**
     * The internal task instance of the second variable.
     */
    public Object task2;
 
    /**
     * Indicates the first task's variable that is involved.
     */
    public int variableID1;
    /**
     * Indicates the second task's variable that is involved.
     */
    public int variableID2;
 
    /**
     * The constant to be added to the second variable.
     */
    public long constant;
 
    /**
     * The relation's comparator.
     */
    public int comparator;

    /**
     * The relation's costs.
     */
    public long costs = 0;

    /** 
     * Indicating the most recent time at which one of the tasks' temporal 
     * variables were externally changed.
     */
    public int externalTaskMoveCount = 0;

    /**
     * Indicating if the relation has already been repaired during the
     * current improvement iteration.
     */
    public boolean repaired;

    /**
     * The back link to the TaskConstraint's temporalRelationList element.
     */
    public OrderedShiftDLLElement temporalRelationLE = null;
 

    // ************************
    // ***** CONSTRUCTORS *****
    // ************************


    /**
     * Initializes the new relation.
     * @param task1  The task instance of the first variable.
     * @param variableID1  Indicates the first task's variable to be involved.
     * @param comparator  The relation's comparator.
     * @param task2  The task instance of the second variable.
     * @param variableID2  Indicates the second task's variable to be involved.
     * @param constant The constant to be added to the second variable.
     */
    public TCTemporalRelationInstance(Object task1, int variableID1, 
				      int comparator, Object task2, 
				      int variableID2, long constant) {
	this.task1 = task1;
	this.variableID1 = variableID1;
	this.comparator = comparator;
	this.task2 = task2;
	this.variableID2 = variableID2;
	this.constant = constant;
   }


    // *******************
    // ***** METHODS *****
    // *******************

    
    /** 
     * Prints the relation's variables with their comparator and constant.
     */
    public String toString() {

	String desc1;
	if (task1 instanceof TCPreconditionTaskInstance) {
	    desc1 = new String(((TCPreconditionTaskInstance) task1).tempVariable.toString());
	} else {
	    if (task1 instanceof TCStateTaskInstance) {
		desc1 = new String(((TCStateTaskInstance) task1).tempVariable.toString());
	    } else {
		if (variableID1 == TaskConstraint.END) {
		    desc1 = new String("Task1End = " + ((TCActionTaskInstance) task1).tempValue[variableID1]);
		} else {
		    desc1 = new String(((TCActionTaskInstance) task1).tempVariable[variableID1].toString());
		}
	    }
	}

	String desc2;
        if (task2 instanceof TCPreconditionTaskInstance) {
            desc2 = new String(((TCPreconditionTaskInstance) task2).tempVariable.toString());
        } else {
            if (task2 instanceof TCStateTaskInstance) {
                desc2 = new String(((TCStateTaskInstance) task2).tempVariable.toString());
            } else {
                if (variableID2 == TaskConstraint.END) {
                    desc2 = new String("Task2End = " + ((TCActionTaskInstance) task2).tempValue[variableID2]);
                } else {
                    desc2 = new String(((TCActionTaskInstance) task2).tempVariable[variableID2].toString());
                }
            }
        }

        return new String("{" + desc1 + "} " + ((comparator == TaskConstraint.LESS_THAN)?"<":"=") + " {" + desc2 + "} " + ((constant < 0)?("- " + -constant):("+ " + constant)));

    }


    /**
     * Calculates the relation's costs and returns the cost change.
     */
    public long calculateCosts() {

	long oldCosts = costs;

        long rightSide;
        long leftSide;

        if (task2 instanceof TCPreconditionTaskInstance) {
	    rightSide = constant 
		+ ((TCPreconditionTaskInstance) task2).tempValue;
	} else {
           if (task2 instanceof TCStateTaskInstance) {
	       rightSide = constant 
		   + ((TCStateTaskInstance) task2).tempValue;
 	   } else {
	       rightSide = constant 
		   + ((TCActionTaskInstance) task2).tempValue[variableID2];
 	   }
	}

        if (task1 instanceof TCPreconditionTaskInstance) {
            leftSide = ((TCPreconditionTaskInstance) task1).tempValue;
        } else {
	    if (task1 instanceof TCStateTaskInstance) {
               leftSide = ((TCStateTaskInstance) task1).tempValue;
	    } else {
               leftSide = 
		   ((TCActionTaskInstance) task1).tempValue[variableID1];
	    }
        }

        if (comparator == TaskConstraint.LESS_THAN) {
	    if (leftSide < rightSide) {
		costs = 0;
	    } else {
		costs = leftSide - rightSide + 1;
	    }
	} else {
	    if (leftSide < rightSide) {
		costs = rightSide - leftSide;
	    } else {
		costs = leftSide - rightSide;
	    }
	}

	return costs - oldCosts;

    }

}
