
package com.ai_center.dragonbreath.costs;

/** 
 * This object maintains an ordered list of constraints with costs that
 * are to be at the same cost level.
 *
 * @author Alexander Nareyek
 * @version $Revision: 2.3 $
 * 
 * @see GlobalSearchControl
 * @see Constraint
 */


import java.math.BigInteger;
import java.util.*;

import com.ai_center.dragonbreath.Constraint;
import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.GlobalSearchControl;
import com.ai_center.dragonbreath.Registry;
import com.ai_center.dragonbreath.utilities.OrderedShiftDLLElement;


public class CostFunction {

    // **********************
    // ***** ATTRIBUTES *****
    // **********************

    /** 
     * The cost function's name. Is initialized by the constructor.
     */
    private String name = null; 

    /** 
     * The Global Search Control reference.
     */
    public final GlobalSearchControl gsc;

    /** 
     * The current total costs of the involved constraints.
     */
    public BigInteger objectiveCosts = new BigInteger("0");

    /** 
     * The cost function's level within the GlobalSearchControl.
     */
    public int level = -1;

    /**
     * Default object to perform the mapping from the constraint's subjective
     * costs to the cost functions' objective costs.
     */
    public CostMapping costMapping;

    /** 
     * List of cost function references to constraints that were added. The 
     * list is maintained such that there is a constraints' ordering of 
     * decreasing objective costs.
     */
    public OrderedCostFunctionReferenceList 
	costFunctionReferences = new OrderedCostFunctionReferenceList();

    /** 
     * Number of constraints that have non-zero costs.
     */
    private int numberOfNonZeroConstraintCosts = 0;

    /**
     * Constants for determining which constraint selection method is to be 
     * used for cost improvements. 
     */
    public final static int SELECTION_HIGHEST = 0;
    public final static int SELECTION_RANDOM = 1;
    public final static int SELECTION_PROBABILITY_DISTRIBUTION = 2;
    /**
     * Determines which constraint selection method is to be used for cost 
     * improvements. The default value is SELECTION_HIGHEST.
     */
    private int selectionMethod;


    // ************************
    // ***** CONSTRUCTORS *****
    // ************************

    /** 
     * Creates a new cost function.
     * @param name_ref  The name for the cost function.
     * @param gsc_ref  The related GlobalSearchControl object.
     * @param level_ref  The level of the cost function in the 
     *                   GlobalSearchControl.
     * @param costMapping_ref  The default CostMapping to be applied.
     */
    public CostFunction(String name_ref, GlobalSearchControl gsc_ref,
			int level_ref, CostMapping costMapping_ref) {
	setName(name_ref);
	gsc = gsc_ref;
	level = level_ref;
	costMapping = costMapping_ref;
	selectionMethod = SELECTION_HIGHEST;
    }


    // *******************
    // ***** METHODS *****
    // *******************

    /** 
     * Sets the cost function's name.
     * @param name  The name to be set.
     */
    public void setName(String name) {
	this.name = name;
    }

    /** 
     * Returns the cost function's name.
     */
    public String getName() {
	return name;
    }

    /** 
     * Returns the cost function's (objective) cost value.
     */
    public BigInteger getCosts() {
	return objectiveCosts;
    }

    /** 
     * Returns the cost function's name, default cost mapping and 
     * (objective) cost value.
     */
    public String toString() {
	return "costs[" + name + "]<" + costMapping + "> = " + objectiveCosts;
    }

     /** 
     * Returns the cost function's complete list information.
     */
    public String costListToString() {

	StringBuffer s = new StringBuffer();
	OrderedShiftDLLElement listElement = costFunctionReferences.first;
	CostFunctionReference costFunctionReference;
	Constraint c;

	while (listElement != null) {
	    costFunctionReference = 
		(CostFunctionReference) listElement.element;
	    c = costFunctionReference.constraint;
	    s.append(costFunctionReference.objectiveCosts + "  (" 
		     + c.getCosts(costFunctionReference.costID) + ")  " 
		     + c + "  " + costFunctionReference.costID + "  (" 
		     + costFunctionReference.costMapping + " <" 
		     + costFunctionReference.costMappingOptions + ">)\n");
	    listElement = listElement.next;
	}

	return s.toString();
    }

   /** 
     * Adds a constraint to the cost function.
     * @param constraint  The constraint object.
     * @param costID  The ID of the costs to be used.
     * @param costMapping  The CostMapping to be applied. Pass "null" for
     *                     using the cost function's default mapping.
     * @param costMappingOptions  Arbitrary object to pass information 
     *                            to the cost function's costMapping.
     */
    public void addConstraint(Constraint constraint, int costID, 
			      CostMapping costMapping,
			      Object costMappingOptions) {

	// Creating a new cost function reference for the constraint

	if (costMapping == null) {
	    costMapping = this.costMapping;
	}

	CostFunctionReference costFunctionReference = 
	    new CostFunctionReference(constraint, costID, this, costMapping, 
				      costMappingOptions);
	constraint.setCostFunctionReference(costID, costFunctionReference);

	// Adding the costFunctionReference to the cost function's 
	// costFunctionReference list

	OrderedShiftDLLElement costFunctionLE = costFunctionReferences.addListElement((Object) costFunctionReference);
	costFunctionReference.costFunctionLE = costFunctionLE;

	// Updating cost values

	constraint.initializeCostType(costID);
	costFunctionReference.updateGlobalCosts();

    }

    /** 
     * Removes a constraint from the cost function.
     * @param constraint  The constraint object.
     * @param costID  The ID of the costs to be used.
     */
    public void removeConstraint(Constraint constraint, int costID) {

	CostFunctionReference costFunctionReference = 
	    constraint.getCostFunctionReference(costID);

	// Removing the constraint from the cost function's 
	// costFunctionReference list

	costFunctionReferences.removeListElement(costFunctionReference.costFunctionLE);

	// Updating the cost function's costs

	objectiveCosts = 
	    objectiveCosts.subtract(costFunctionReference.objectiveCosts);

	if (!costFunctionReference.objectiveCosts.equals(BigInteger.ZERO)) {

	    numberOfNonZeroConstraintCosts--;

	    // Propagating cost change

	    gsc.registerCostFunctionUpdate(this);

	}

	// Removing the related cost function reference from the constraint

	constraint.setCostFunctionReference(costID, null);

    }

    /** 
     * Sets a specific method for constraint selection.
     * @param selectionMethod  Defines which method will be used for
     *                         constraint selection.
     */
    public void setConstraintSelectionMethod(int selectionMethod) {
	if ((selectionMethod < 0) || (selectionMethod > 2)) {
	    throw new Error("The requested constraint selection method is not available");
	}
	this.selectionMethod = selectionMethod;
    }

    /**
     * Returns the selectionMethod.
     */
    public int getConstraintSelectionMethod(){
        return selectionMethod;
    }
    
    /**
     * Registers that the inconsistency of a constraint has changed
     * and updates the related structures.
     * @param costFunctionReference  The affected costFunctionReference.
     * @param costChange  The change in costs.
     */
    public void updateGlobalCosts(CostFunctionReference costFunctionReference,
				  BigInteger costChange) {

	// Updating costFunctionReference list order

	costFunctionReferences.updateListElementPosition(costFunctionReference.costFunctionLE);

	// Updating the cost function's costs

	objectiveCosts = objectiveCosts.add(costChange);

	if (costFunctionReference.objectiveCosts.equals(BigInteger.ZERO)) {
	    numberOfNonZeroConstraintCosts--;
	} else {
	    if (costChange.equals(costFunctionReference.objectiveCosts)) {
		numberOfNonZeroConstraintCosts++;
	    }
	}

	// Propagating cost change

	gsc.registerCostFunctionUpdate(this);

    }

    /** 
     * Activates one of the constraints for cost improvement.
     */
    public void improveCosts() {

	OrderedShiftDLLElement listElement = costFunctionReferences.first;

	// Testing for wrong use

	if (Registry.SAFE_MODE) {
	    if (listElement == null) {
		throw new Error("No constraints were assigned to the cost function \"" + this);
	    }
	    if ((level > 0) && (objectiveCosts.equals(BigInteger.ZERO))) {
		throw new Error("The cost function \"" + this + 
				"\" has no costs to be improved");
	    }
	}

	// Selecting constraint for improvement

	CostFunctionReference costFunctionReference = 
	    (CostFunctionReference) listElement.element;

	switch (selectionMethod) {

	case SELECTION_HIGHEST: {
	    break;
	}

	case SELECTION_RANDOM: {
	    int selectedCostFunctionReference = 
		gsc.random.nextInt(numberOfNonZeroConstraintCosts);
	    while (selectedCostFunctionReference-- > 0) {
		listElement = listElement.next;
		costFunctionReference = 
		    (CostFunctionReference) listElement.element;
	    }
	}

	case SELECTION_PROBABILITY_DISTRIBUTION: {

	    // Computing random BigInteger

	    BigInteger selectedCosts;

	    int bl = objectiveCosts.bitLength();
	    do {
		selectedCosts = new BigInteger(bl, gsc.random);
	    } while (objectiveCosts.compareTo(selectedCosts) != 1);

	    // Retrieving related constraint

	    selectedCosts = 
		selectedCosts.subtract(costFunctionReference.objectiveCosts);

	    while (selectedCosts.compareTo(BigInteger.ZERO) != -1) {
		listElement = listElement.next;
		costFunctionReference = 
		    (CostFunctionReference) listElement.element;
		selectedCosts = selectedCosts.subtract(costFunctionReference.objectiveCosts);
	    }
	    break;
	}
	}

        // Executing improvement step

	costFunctionReference.constraint.improveCosts(costFunctionReference.costID);
    }

}
