
package com.ai_center.dragonbreath.costs;

/** 
 * A container for a constraint's cost function information. 
 * 
 * @author Alexander Nareyek
 * @version $Revision: 2.3 $ 
 */


import java.math.BigInteger;

import com.ai_center.dragonbreath.Constraint;
import com.ai_center.dragonbreath.constraintgraph.ConventionalConstraint;
import com.ai_center.dragonbreath.constraintgraph.ConventionalConstraintGraphics;
import com.ai_center.dragonbreath.constraintgraph.NodeGraphics;
import com.ai_center.dragonbreath.utilities.OrderedShiftDLLElement;


public class CostFunctionReference {
    
    // **********************
    // ***** ATTRIBUTES *****
    // **********************
    
    /** 
     * The constraint to which this CostFunctionReference is to belong to.
     */
    public Constraint constraint;

    /** 
     * The ID of the related costs in the related constraint.
     */
    public int costID;

    /** 
     * The CostFunction reference.
     */
    public CostFunction costFunction;

    /**
     * Object to perform the mapping from the constraint's subjective
     * costs to the cost functions' objective costs.
     */
    public CostMapping costMapping;

    /** 
     * The reference to the options object for the CostMapping.
     */
    public Object costMappingOptions;

    /** 
     * The related constraint's costs after being mapped for the cost function.
     */
    public BigInteger objectiveCosts = new BigInteger("0");

    /** 
     * The back link to the CostFunction's list element.
     */
    public OrderedShiftDLLElement costFunctionLE = null;


    // ************************
    // ***** CONSTRUCTORS *****
    // ************************

    /** 
     * Creates the reference object.
     * @param constraint  The constraint to which this CostFunctionReference
     *                    is to belong to.
     * @param costID  The ID of the costs in the related constraint.
     * @param costFunction  The reference's CostFunction.
     * @param costMapping  The CostMapping object to be used.
     * @param costMappingOptions  The options object for the CostMapping.
     * @param costFunctionLE  The back link to the CostFunction's list 
     *                        element.
     */
    public CostFunctionReference(Constraint constraint, int costID, 
				 CostFunction costFunction,
				 CostMapping costMapping,
				 Object costMappingOptions) {
	this.constraint = constraint;
	this.costID = costID;
	this.costFunction = costFunction;
	this.costMapping = costMapping;
	this.costMappingOptions = costMappingOptions;

	// Initializing cost mapping for the constraint

	costMapping.initialize(this);

    }


    // *******************
    // ***** METHODS *****
    // *******************

    /**
     * Registers that the costs of the constraint have changed
     * and updates the related structures.
     */
    public void updateGlobalCosts() {

	// Updating objective costs

	BigInteger newObjectiveCosts = 
	    costMapping.calculateObjectiveCosts(this);
	BigInteger costChange = newObjectiveCosts.subtract(objectiveCosts);
	objectiveCosts = newObjectiveCosts;

	// Propagating cost changes to the cost function

	if (!costChange.equals(BigInteger.ZERO)) {
	    costFunction.updateGlobalCosts(this, costChange);

	    // Updating constraint's cost graphics
            if (constraint instanceof ConventionalConstraint) {
                if (((ConventionalConstraint) constraint).graphObject instanceof NodeGraphics) {
                    ((ConventionalConstraintGraphics) ((ConventionalConstraint) constraint).graphObject).updateCosts();
                    ((ConventionalConstraintGraphics) ((ConventionalConstraint) constraint).graphObject).updateSize();
                }
            }
	}

   }


}
