
package com.ai_center.dragonbreath.costs;

/** 
 * Class for mapping subjective constraint costs to objective cost function
 * costs.
 *
 * @author Alexander Nareyek
 * @version $Revision: 2.0 $ 
 * 
 * @see Constraint
 * @see GlobalSearchControl
 */


import java.math.BigInteger;

import com.ai_center.dragonbreath.Constraint;


public abstract class CostMapping {

    // **********************
    // ***** ATTRIBUTES *****
    // **********************

    /** 
     * The cost mapping's name. Is initialized by the Registry.
     * @see Registry
     */
    private String name = null; 

    /**
     * The name of the type of the costmapping. Is initialized by the Registry.
     * @see Registry
     */
     private String typeName = null;
     
    // *******************
    // ***** METHODS *****
    // *******************

    /** 
     * Sets the cost mapping's name.
     * @param name  The name to be set.
     */
    public void setName(String name) {
	this.name = name;
    }

    /** 
     * Returns the cost mapping's name.
     */
    public String getName() {
	return name;
    }
     
    /** 
     * Returns the cost function's name for printing.
     */
    public String toString() {
	return name;
    }

    /** 
     * Sets the cost mapping's type's name.
     * @param name  The name to be set.
     */
    public void setTypeName(String name) {
	typeName = name;
    }
    
    /**
     * Returns the type of this class as a string.
     */
    public String getTypeName() {
        return typeName;
    }
    
    /** 
     * Returns an array of the cost mapping's option types.
     * @see Registry
     */
    public abstract int[] getOptionTypes();

    /** 
     * Returns a default value for an option.
     * @param optionID  The option's ID for which the default value is to be returned.
     * @see Registry
     */
    public abstract Object getOptionDefaultValue(int optionID);

    /** 
     * Returns a textual description of an option.
     * @param optionID  The option's ID for which the description is to be returned.
     * @see Registry
     */
    public abstract String getOptionDescription(int optionID);

    /** 
     * Removes all specific configurations that have been set.
     */
    public abstract void reset();

    /** 
     * May pre-compute information for the cost mapping to be stored in 
     * the constraint's CostFunctionReference.
     * @param costFunctionReference  The related costFunctionReference.
     */
    public abstract void initialize(CostFunctionReference 
				    costFunctionReference);

    /** 
     * Calculates the objective costs for a constraint.
     * @param costFunctionReference  The related costFunctionReference.
     */
    public abstract BigInteger calculateObjectiveCosts(CostFunctionReference 
						       costFunctionReference);

}


/* 
 * DO _NOT_ EDIT ANYTHING BELOW THIS POINT! 
 * 
 * $Log: CostMapping.java,v $
 * Revision 2.0  2002/09/17 00:33:06  nareyek
 * work on beta version began: resource type, state tasks and precondition tasks introduced
 *
 * Revision 1.8  2002/09/02 15:20:55  teutates
 * added typeName to store the type of an instance
 *
 * Revision 1.7  2002/08/21 05:01:50  nareyek
 * cost interface overhaul part 1 - completion tomorrow
 *
 * Revision 1.6  2002/05/30 20:40:11  nareyek
 * cost mapping options can now be entered also in the graphical mode
 *
 * Revision 1.5  2002/05/25 03:42:13  nareyek
 * cost functions and mappings can now be created/used also by the graphical interface
 *
 * Revision 1.4  2002/04/09 21:55:15  nareyek
 * costs display in graphics repaired; gsc's inconsistency variable eliminated
 *
 * Revision 1.3  2002/04/09 01:19:37  nareyek
 * Every cost item can now have a mapping; constraints can have multiple costs
 *
 * Revision 1.2  2002/04/04 09:25:36  nareyek
 * general constraint interface included; cost workflow improved
 *
 * Revision 1.1  2002/04/04 04:55:58  nareyek
 * directories costs and constraintgraph introduced
 *
 * Revision 1.1  2002/04/02 06:19:12  nareyek
 * cost mapping functionality added
 * 
 * 
 */
