
package com.ai_center.dragonbreath.costs;

/** 
 * This special CostMapping applies BigDecimal factors for computing the 
 * objective costs of a constraint. 
 *
 * @author Alexander Nareyek
 * @version $Revision: 2.1 $ 
 */


import java.math.BigInteger;
import java.math.BigDecimal;

import com.ai_center.dragonbreath.constraintgraph.ConventionalConstraint;
import com.ai_center.dragonbreath.Registry;


public class FactorCostMapping extends CostMapping {

    // **********************
    // ***** ATTRIBUTES *****
    // **********************

    /**
     * Default factor for a specific constraint class if no costMappingOption
     * is given.
     */
    private BigDecimal[] ccTypeDefaultFactor = 
	new BigDecimal[Registry.MAX_NODE_TYPE_NUMBER];

    /**
     * The constant "0".
     */
    protected final static BigDecimal ZERO = new BigDecimal(0.0);
    /**
     * The constant "1".
     */
    protected final static BigDecimal ONE = new BigDecimal(1.0);

    /**
     * Array of option types.
     */
    private final static int[] option = new int[1];
    static {
	option[0] = Registry.TYPE_POSITIVE_BIGDECIMAL;
    }
    /**
     * Array of option type defaults.
     */
    private final static Object[] optionDefault = new Object[1];
    static {
	optionDefault[0] = ONE;
    }
    /**
     * Array of option type descriptions.
     */
    private final static String[] optionDescription = new String[1];
    static {
	optionDescription[0] = new String("factor");
    }


    // ************************
    // ***** CONSTRUCTORS *****
    // ************************

    /** 
     * Initializes class-specific default factors.
     */
    public FactorCostMapping() {
	reset();
    }


    // *******************
    // ***** METHODS *****
    // *******************

    public int[] getOptionTypes() {
	return option;
    }

    public Object getOptionDefaultValue(int optionID) {
	if ((optionID < 0) || (optionID >= option.length)) {
	    throw new Error("Invalid option ID");
	}
	return optionDefault[optionID];
    }

    public String getOptionDescription(int optionID) {
	if ((optionID < 0) || (optionID >= option.length)) {
	    throw new Error("Invalid option ID");
	}
	return optionDescription[optionID];
    }

    private void checkIfValid(Object factor) {
	if (!(factor instanceof BigDecimal)) {
	    throw new Error("Valid values for a FactorCostMapping have to be of type BigDecimal");
	}
	if (((BigDecimal) factor).compareTo(ZERO) == -1) {
	    throw new Error("Valid values for a FactorCostMapping have to be greater than or equal to 0.0");
	}
    }

    /** 
     * Sets a default factor for a specific constraint type.
     * @param constraintID  The type of the constraint.
     * @param defaultFactor  The default factor to be set for this 
     *                       constraint type.
     * @see Registry
     */
    public void setDefaultFactor(int constraintID, BigDecimal defaultFactor) {
	if (!Registry.isConstraint(constraintID)) {
	    throw new Error("Invalid constraint ID");
	}
	checkIfValid(defaultFactor);
	ccTypeDefaultFactor[constraintID] = defaultFactor;
    }

    public void reset() {
	for (int i = 0; i < Registry.MAX_NODE_TYPE_NUMBER; i++) {
	    ccTypeDefaultFactor[i] = (BigDecimal) optionDefault[0];
	}
    }

    public void initialize(CostFunctionReference costFunctionReference) {
	if (costFunctionReference.costMappingOptions == null) {
	    if (costFunctionReference.constraint instanceof 
		ConventionalConstraint) {
		costFunctionReference.costMappingOptions = ccTypeDefaultFactor[costFunctionReference.constraint.getTypeID()];
	    } else {
                costFunctionReference.costMappingOptions = ccTypeDefaultFactor[costFunctionReference.constraint.getTypeID()];
		// TODO: for structural Constraints
		//throw new Error("Not implemented yet for structural constraints");
	    }
	} else {
	    if (costFunctionReference.costMappingOptions instanceof Object[]) {
		Object[] fArray = 
		    (Object[]) costFunctionReference.costMappingOptions;
		if (fArray.length > 1) {
		    throw new Error("Invalid option values for a FactorCostMapping");
		}
		costFunctionReference.costMappingOptions = fArray[0];
	    }
	    checkIfValid(costFunctionReference.costMappingOptions);
	}
    }

    public BigInteger calculateObjectiveCosts(CostFunctionReference 
					      costFunctionReference) {
	return BigDecimal.valueOf(costFunctionReference.constraint.getCosts(costFunctionReference.costID)).multiply((BigDecimal) costFunctionReference.costMappingOptions).toBigInteger();
    }

}


/* 
 * DO _NOT_ EDIT ANYTHING BELOW THIS POINT! 
 * 
 * $Log: FactorCostMapping.java,v $
 * Revision 2.1  2003/02/21 01:33:01  teutates
 * structural repair initial release
 *
 * Revision 2.0  2002/09/17 00:33:06  nareyek
 * work on beta version began: resource type, state tasks and precondition tasks introduced
 *
 * Revision 1.8  2002/05/30 20:40:11  nareyek
 * cost mapping options can now be entered also in the graphical mode
 *
 * Revision 1.7  2002/05/25 03:42:13  nareyek
 * cost functions and mappings can now be created/used also by the graphical interface
 *
 * Revision 1.6  2002/05/24 18:04:59  nareyek
 * cost-mapping classes can now be added dynamically
 *
 * Revision 1.5  2002/04/09 22:10:40  nareyek
 * introduced getTypeID() for Constraint interface
 *
 * Revision 1.4  2002/04/09 21:55:15  nareyek
 * costs display in graphics repaired; gsc's inconsistency variable eliminated
 *
 * Revision 1.3  2002/04/09 01:19:37  nareyek
 * Every cost item can now have a mapping; constraints can have multiple costs
 *
 * Revision 1.2  2002/04/04 09:25:36  nareyek
 * general constraint interface included; cost workflow improved
 *
 * Revision 1.1  2002/04/04 04:55:58  nareyek
 * directories costs and constraintgraph introduced
 *
 * Revision 1.1  2002/04/02 06:19:12  nareyek
 * cost mapping functionality added
 * 
 * 
 */
