
package com.ai_center.dragonbreath.examples.scheduling;


import java.math.BigInteger;
import java.util.*;

import com.ai_center.dragonbreath.*;
import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.constraints.*;
import com.ai_center.dragonbreath.constraints.taskconstraint.TCTask;
import com.ai_center.dragonbreath.costs.CostFunction;


/**
 * A scheduling sample application of the <a href="http://www.ai-center.com/projects/dragonbreath/"><font face="Times">DragonBreath</font> engine</a>.
 *
 * @author Alexander Nareyek
 * @version $Revision: 1.3 $
 */
public class Scheduling
{

    protected static class MyScheduling
    {

        public GlobalSearchControl gsc;

        final long testruns = 100;
        final int maxTimeInMS = 1000;

        final int horizon = 900;
        final int jobNumber = 10;
        final int taskNumber = 10;
        final int maxTaskDuration = 100;
        final int maxTemporalRelationConstant = 10;

        public Variable[][] beginVariable
            = new Variable[ jobNumber ][ taskNumber ];
        public int[][] duration = new int[ jobNumber ][ taskNumber ];
        public ConventionalConstraint[] actionConstraint
            = new ConventionalConstraint[ jobNumber ];

        public ConventionalConstraint currentTime;


        public void createActionTypes()
        {

            int actionID;
            TCTask task[] = new TCTask[ taskNumber ];
            long constant;
            int relation = 0;

            for ( int j = 0 ; j < jobNumber ; j++ )
            {

                actionID = TaskConstraint.registerNewActionType();

                for ( int t = 0 ; t < taskNumber ; t++ )
                {
                    task[ t ] = TaskConstraint.addActionTaskToActionType( actionID , t , t , gsc.random.nextInt(maxTaskDuration ) + 1 , 1 );
                    if ( t != 0 )
                    {
                        constant = gsc.random.nextInt( maxTemporalRelationConstant + 1 ) + 1;
                        switch ( gsc.random.nextInt( 5 ) )
                        {
                        case 0: relation = TaskConstraint.LESS_THAN; break;
                        case 1: relation
                                    = TaskConstraint.LESS_OR_EQUALS; break;
                        case 2: relation = TaskConstraint.EQUALS; break;
                        case 3: relation
                                    = TaskConstraint.GREATER_OR_EQUALS; break;
                        case 4: relation = TaskConstraint.GREATER_THAN; break;
                        }
                        TaskConstraint.addTemporalRelationToActionType( actionID , task[ t-1 ] , TaskConstraint.BEGIN , relation , task[ t ] , TaskConstraint.BEGIN , constant );
                    }
                }

            }

        }


        public void createSeedActions()
        {

            Embedding e;
            Variable v;
            ConventionalConstraint taskConstraint;

            for ( int j = 0 ; j < jobNumber ; j++ )
            {

                v = gsc.createVariable( "Action Type " + j );
                gsc.setVariable( v , j );

                e = new Embedding( 1 );
                gsc.extendEmbedding( e , v , Registry.ACTIONTYPE_EDGE ,
                    Registry.INCOMING );
                actionConstraint[ j ] = gsc.createConstraint( Registry.ACTION_OBJECTCONSTRAINT , e );

                e = new Embedding( 2 );
                gsc.extendEmbedding( e , currentTime , Registry.PLAIN_EDGE ,
                    Registry.INCOMING );
                gsc.extendEmbedding(e , actionConstraint[ j ] ,
                    Registry.PLAIN_EDGE , Registry.INCOMING );

                gsc.selectCostFunction( gsc.getStructuralCostFunction() );
                taskConstraint
                    = gsc.createConstraint( Registry.TASK_CONSTRAINT , e );
                gsc.selectCostFunction( gsc.getDefaultCostFunction() );
                gsc.addCosts( taskConstraint ,
                    TaskConstraint.TEMPORAL_INCONSISTENCY );

            }

        }


        public void preparingBeginVariables()
        {

            Variable variable = null;
            ConnectedNode connectedNode;
            Node actionTask;
            Embedding e;
            ConventionalConstraint c;
            int t;

            for ( int j = 0 ; j < jobNumber ; j++ )
            {

                synchronized( actionConstraint[ j ].connections )
                {
                    actionConstraint[ j ].connections.initializeEnumeration( false , true , false , false , null , Registry.nodeType[ Registry.ACTIONTASK_OBJECTCONSTRAINT ] , 0 , false );
                    for ( t = 0 ; t < taskNumber ; t++ )
                    {

                        connectedNode
                            = actionConstraint[ j ].connections.getNext();
                        actionTask = connectedNode.node;

                        synchronized( actionTask.connections )
                        {
                            beginVariable[ j ][ t ] = actionTask.connections.getSpecificVariable( Registry.BEGIN_EDGE );
                            duration[ j ][ t ] = (int)actionTask.connections.getSpecificVariable( Registry.DURATION_EDGE ).value();
                        }

                    }
                }
                for ( t = 0 ; t < taskNumber ; t++ )
                {
                    e = new Embedding( 1 );
                    gsc.extendEmbedding( e , beginVariable[ j ][ t ],
                        Registry.PLAIN_EDGE , Registry.INCOMING );
                    c = gsc.createConstraint( Registry.RANGE_CONSTRAINT , e );
                    gsc.setBounds( c , 0 , horizon - duration[ j ][ t ] );
                }

            }

        }


        public void randomizeStartingTimes()
        {
            for ( int j = 0 ; j < jobNumber ; j++ )
            {
                for ( int t = 0 ; t < taskNumber ; t++ )
                {
                    gsc.setVariable( beginVariable[ j ][ t ] , gsc.random.nextInt( horizon - duration[ j ][ t ] + 1 ) );
                }
            }
        }


        public long searchForSolution()
        {

            long startingTime;
            long currentTime;
            long maxTime;

            startingTime = System.currentTimeMillis();
            maxTime = startingTime + maxTimeInMS;

            while ( gsc.executeImprovementIteration() )
            {
                currentTime = System.currentTimeMillis();
                if ( currentTime > maxTime )
                {
                    return -1;
                }
            }
            currentTime = System.currentTimeMillis();
            return currentTime - startingTime;
        }


        public void run()
        {

            // Initialization

            gsc = new GlobalSearchControl( 0 );

            // Creating basic structures

            Variable ctVar = gsc.createVariable( "Current Time" );
            Embedding e = new Embedding( 1 );
            gsc.extendEmbedding( e , ctVar , Registry.PLAIN_EDGE ,
                Registry.INCOMING );
            currentTime = gsc.createConstraint( Registry.CURRENTTIME_OBJECTCONSTRAINT , e );

            createActionTypes();
            createSeedActions();

            // Let the graph be built

            CostFunction c = gsc.getStructuralCostFunction();
            while ( gsc.executeImprovementIteration( c ) );
            preparingBeginVariables();

            // Disabling further graph changes by improvement heuristics

            gsc.disableStructuralChanges();

            // Search

            long currentTime;
            long totalTime = 0;
            long timeMin = maxTimeInMS;
            long timeMax = 0;
            long notFound = 0;

            LinkedList times = new LinkedList();

            for ( int t = 1 ; t <= testruns ; t++ )
            {

                System.out.println( "Starting testrun " + t + "..." );

                randomizeStartingTimes();
                currentTime = searchForSolution();

                if ( currentTime != -1 )
                {
                    if ( currentTime < timeMin )
                    {
                        timeMin = currentTime;
                    }
                    if ( currentTime > timeMax )
                    {
                        timeMax = currentTime;
                    }
                    totalTime += currentTime;
                    times.add( new Long( currentTime ) );
                }
                else
                {
                    notFound++ ;
                }

            }

            long averageTime = 0;
            if ( testruns != notFound )
            {
                averageTime = totalTime / ( testruns - notFound );
            }

            System.out.println( "\nAverage time: " + averageTime + " ms" );
            System.out.println( "Unsuccessful test runs (more than "
                + maxTimeInMS + " ms): " + notFound );

            long totalVariance = 0;
            long currentVariance = 0;
            Iterator it = times.iterator();
            while ( it.hasNext() )
            {
                currentTime = ( (Long)it.next() ).longValue();
                currentVariance = Math.abs( averageTime - currentTime );
                totalVariance += currentVariance;
            }
            long averageVariance = totalVariance / testruns;;

            System.out.println( "Minimal time: " + timeMin + " ms" );
            System.out.println( "Maximal time: " + timeMax + " ms" );
            System.out.println( "Average variance: "
                + averageVariance + " ms" );

        }

    }


    public static void main( String[] args )
    {
        MyScheduling s = new MyScheduling();
        s.run();
    }


}
