package com.ai_center.dragonbreath.examples.structuralConstraintExample.fluffExampleConstraints;

import com.ai_center.dragonbreath.Registry;
import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.graphRepair.*;
import com.ai_center.dragonbreath.utilities.*;

/**
 * This constraint forces the sum over a set of variables  to equal another
 * variable. If the sum is constantly too big a FluffFactory is removed.
 * 
 * @see Variable
 */
import java.lang.Math;

import java.util.*;


public class CruftSumConstraint extends ConventionalConstraint {
    //~ Static variables/initializers ******************************************

    // **********************
    // ***** ATTRIBUTES *****
    // **********************

    /** The constraint type's name. */
    private static final String TYPE_NAME = new String("CruftSum");

    /** The constraint's base graph. */
    private static final Embedding BASE_GRAPH = new Embedding(1);

    /** The constraint's number of extension graphs. */
    private static final int EXTENSION_GRAPH_NUMBER = 1;

    /** The constraint's extension graphs. */
    private static final Embedding[] EXTENSION_GRAPH = 
            new Embedding[EXTENSION_GRAPH_NUMBER];
    private static final double      epsilon = 0.00001;

    //~ Instance variables *****************************************************

    /** Used to remove variables */
    GraphRepairer myGraphRepairer;
    private DebugLogger dl = new DebugLogger(true);

    /** The list of variables to be added. */
    private java.util.List addVariables = Collections.synchronizedList(
                                                  new LinkedList());

    /** The sum variable. */
    private Variable sumVariable = null;

    /** The value of the sum variable. */
    private double sum = 0;

    /**
     * When structural repair decides to add or remove fluff factories, it
     * does so on the average error of the sum since the last structural
     * repair. This is the sum of the errors.
     */
    private double sumErrorsSinceLastStructuralRepair = 0;

    /** The sum of the added variables. */
    private double sumVariables = 0;

    /** The number of variables to be added. */
    private int addVariableNumber = 0;

    /**
     * This is the number of repair attempts since the last structural repair.
     * sumErrorsSinceLastStructuralRepair / numRepairAttempts == average
     * error.
     */
    private long numRepairAttempts = 0;

    //~ Methods ****************************************************************

    public Embedding getBaseGraph() {
        return BASE_GRAPH;
    }


    public String getCostName(int costID) {
        if (costID == 0) {
            return new String("difference inconsistency");
        } else {
            throw new Error("Invalid cost ID");
        }
    }


    /**
     * Returns the actual Costs.
     */
    public long getCosts() {
        double abs = Math.abs(sum - sumVariables);

        if (abs < epsilon) {
            costs[0] = 0;
        } else {
            costs[0] = (long) Math.ceil(abs);
        }

        return costs[0];
    }


    public Embedding getExtensionGraph(int graphID) {
        return EXTENSION_GRAPH[graphID];
    }


    public int getExtensionGraphNumber() {
        return EXTENSION_GRAPH_NUMBER;
    }


    public int getMaximalConnectionNumber() {
        // There is no maximum for extensions
        return Integer.MAX_VALUE;
    }


    public int getNumberOfCosts() {
        return 1;
    }


    public String getTypeName() {
        return TYPE_NAME;
    }


    public void connectNode(ConnectionReference first, ConnectionReference last) {
        Variable variable = (Variable) first.node;

        if (first.edgeID == Registry.SUM_EDGE) {
            sumVariable = variable;
            sum         = sumVariable.value();

            boolean structureOK;
            synchronized (addVariables) {
                structureOK = (addVariables.size() >= 1);
            }

            if (structureOK) {
                updateCosts();
            }
        } else {
            synchronized (addVariables) {
                addVariables.add(variable);
                addVariableNumber++;
            }

            sumVariables = sumVariables + variable.value();

            if (sumVariable != null) {
                updateCosts();
            }
        }
    }


    // *******************
    // ***** METHODS *****
    // *******************
    public void create_graphs() {
        dl.switchOn(1);


        // Creating the base graph
        BASE_GRAPH.addConnectedNode(Registry.VARIABLE, Registry.SUM_EDGE, 
                                    Registry.INCOMING);


        // Creating the extension graph
        EXTENSION_GRAPH[0] = new Embedding(1);
        EXTENSION_GRAPH[0].addConnectedNode(Registry.VARIABLE, 
                                            Registry.ADD_EDGE, 
                                            Registry.INCOMING);
    }


    public void disconnectNode(ConnectionReference first, 
                               ConnectionReference last) {
        if (first.edgeID == Registry.SUM_EDGE) {
            sumVariable = null;
            sum         = 0;
        } else {
            Variable variable = (Variable) first.node;
            synchronized (addVariables) {
                addVariables.remove(variable);
                addVariableNumber--;
            }

            sumVariables = sumVariables - variable.value();
        }

        updateCosts();
    }


    public void improveCosts(int costID) {
        // Checking for valid cost ID
        if (costID != 0) {
            throw new Error("The cost ID  \"" + costID + 
                            "\" is not supported by this constraint");
        }

        sumErrorsSinceLastStructuralRepair += (sumVariables - sum);
        numRepairAttempts++;

        int       i;
        boolean[] possibleChange  = new boolean[addVariableNumber + 1];
        int       possibleChanges = 0;
        synchronized (addVariables) {
            for (i = 0; i < addVariableNumber; i++) {
                Variable addVar = (Variable) addVariables.get(i);

                if (! addVar.isConstant() && 
                        (((sumVariables > sum) && 
                                (addVar.value() > addVar.lowerBound)) || 
                            ((sumVariables < sum) && 
                                (addVar.value() < addVar.upperBound)))) {
                    possibleChanges++;
                    possibleChange[i] = true;
                } else {
                    possibleChange[i] = false;
                }
            }
        }

        boolean ready          = false;
        int     chosenVariable;

        if (! fluffFactoriesMissing() && 
                ((possibleChanges == 0) || 
                    (gsc.random.nextDouble() < RegistryConstants.getStructRepProp()))) { // / (addVariables.size() + 1)))) { 

            if (structuralRepair()) {
                return;
            }
        }

        while (! ready && (possibleChanges > 0)) {
            chosenVariable = gsc.random.nextInt(possibleChanges);

            i = -1;

            do {
                while (! possibleChange[++i])
                    ;
            } while (chosenVariable-- > 0);

            Variable var = (Variable) addVariables.get(i);
            double   v = var.value();

            if (sumVariables > sum) {
                if (var.lowerBound() < v) {
                    gsc.setVariable_internal(var, (v + sum) - sumVariables);
                    ready = true;
                }
            } else {
                if (var.upperBound() > v) {
                    gsc.setVariable_internal(var, (v + sum) - sumVariables);
                    ready = true;
                }
            }

            possibleChanges--;
            possibleChange[i] = false;
        }

        /* for (i = 0; i < addVariableNumber; i++) {
           if (possibleChanges == 0) {
                   return;
           }
           double neededChange = sum - sumVariables;
           double changePerAddVar = neededChange * gsc.random.nextDouble(); 
           if (possibleChange[i]) {
               Variable var = (Variable) addVariables.get(i);
               double v = var.value();
               if (sumVariables > sum) {
                   if (var.lowerBound() < v) {
                       gsc.setVariable_internal(var, v + changePerAddVar);
                       possibleChanges--;
                   }
               } else {
                   if (var.upperBound() > v) {
                       gsc.setVariable_internal(var, v +  changePerAddVar);
                       possibleChanges--;
                   }
               }
               
           }
           }*/
    }


    public void initializeCostType(int costID) {
        // Nothing to do
    }


    public void update(double oldValue, ConnectionReference first, 
                       ConnectionReference last) {
        Variable variable = (Variable) first.node;

        if (variable == sumVariable) {
            sum = sumVariable.value();
        } else {
            sumVariables = sumVariables - oldValue + variable.value();
        }

        updateCosts();
    }


    /**
     * Returns a Variable that enters negative in the sum.
     */
    protected Variable getChuteVar() {
        Iterator it = addVariables.iterator();

        while (it.hasNext()) {
            Variable var = (Variable) it.next();

            if (var.value() < 0) {
                return var;
            }
        } //while

        return null;
    }


    /**
     * Enters a Variable that enters positive in the sum.
     */
    protected Variable getFactoryVar() {
        Iterator it = addVariables.iterator();

        while (it.hasNext()) {
            Variable var = (Variable) it.next();

            if (var.value() > 0) {
                return var;
            }
        } //while

        return null;
    }


    /**
     * Are there still fluffFactories that can be used to change the value?
     * 
     * @return true if there are fluffFactories that are not yet connected to
     *         the sum, else false.
     */
    protected boolean fluffFactoriesMissing() {
        int numFactories = gsc.findConstraints(
                                   RegistryConstants.FLUFF_FACTORY_A_CONSTRAINT)
                              .size();
        numFactories += gsc.findConstraints(
                                RegistryConstants.FLUFF_FACTORY_AA_CONSTRAINT)
                           .size();
        numFactories += gsc.findConstraints(
                                RegistryConstants.FLUFF_FACTORY_AAA_CONSTRAINT)
                           .size();
        numFactories += gsc.findConstraints(
                                RegistryConstants.FLUFF_CHUTE_CONSTRAINT)
                           .size();

        return (numFactories > addVariables.size()); // || (numFactories == 0);
    }


    protected boolean structuralRepair() {
        dl.switchOn(1);

        if (myGraphRepairer == null) {
            myGraphRepairer = new GraphRepairer(gsc);
        }

        double error;

        if (numRepairAttempts == 0) {
            error = sumVariables - sum;
        } else {
            error = sumErrorsSinceLastStructuralRepair / numRepairAttempts;
        }

        sumErrorsSinceLastStructuralRepair = 0;
        numRepairAttempts                  = 0;

        if ((error == 0) || 
                (gsc.random.nextDouble() > java.lang.Math.abs(error))) { // if the error is 0 or near null 

            return false;
        }

        RepairPlan plan = null;
        double     val = gsc.random.nextDouble();

        if (error > 0) { //too much cruft

            int constrID;

            /* if (val < 0.03) {// remove a fluffFactory
               Variable factoryVar = getFactoryVar();
               if ( factoryVar != null) {
                    plan = myGraphRepairer.removeNode(factoryVar);
                    dl.print(1, "\ncruft Sum removed factory."); 
               }
               }*/
            if (plan == null) { // add a chute
                plan = myGraphRepairer.addNode(
                               RegistryConstants.FLUFF_CHUTE_CONSTRAINT);
                dl.print(1, "\ncruft Sum added chute.");
            }
        } else { //should almost never be the case, as goal is to have 0 cruft

            // if (val < 0.2) {
            Variable chuteVar = getChuteVar();

            if (chuteVar != null) {
                plan = myGraphRepairer.removeNode(chuteVar);
                dl.print(1, "\ncruft Sum removed chute.");
            }

            //}

            /*
               if (plan == null) {
                   int constrID;
                   val = gsc.random.nextDouble();
                   if (val < 0.3333) {
                       constrID = Registry.FLUFF_FACTORY_A_CONSTRAINT;
                       dl.print(1, "\ncruft Sum added A."); 
                   } else if (val < 0.66667) {
                       constrID = Registry.FLUFF_FACTORY_AA_CONSTRAINT;
                       dl.print(1, "\ncruft Sum added AA."); 
                   } else {
                       constrID = Registry.FLUFF_FACTORY_AAA_CONSTRAINT;
                       dl.print(1, "\ncruft Sum added AAA."); 
                   }
                   plan = myGraphRepairer.addNode(constrID);
               } */
        }

        if (plan != null) {
            plan.repair();
        }

        dl.print(1, "\nCruftSum error: " + error);

        return true;
    }


    /**
     * Calculates the constraint's costs.
     */
    private void updateCosts() {
        double abs = Math.abs(sum - sumVariables);

        if (abs < epsilon) {
            costs[0] = 0;
        } else {
            costs[0] = (long) Math.ceil(abs);
        }

        costFunctionReference[0].updateGlobalCosts();
    }
}