
package com.ai_center.dragonbreath.examples.taskconstraintdemo;


import java.math.BigDecimal;

import com.ai_center.dragonbreath.*;
import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.constraints.*;
import com.ai_center.dragonbreath.constraints.stateresourceconstraint.*;
import com.ai_center.dragonbreath.constraints.taskconstraint.TCTask;
import com.ai_center.dragonbreath.costs.CostFunction;
import com.ai_center.dragonbreath.costs.CostMapping;
import com.ai_center.dragonbreath.costs.FactorCostMapping;


/**
 * A task constraint demo of the <a href="http://www.ai-center.com/projects/dragonbreath/"><font face="Times">DragonBreath</font> engine</a>.
 *
 * @author Alexander Nareyek
 * @version $Revision: 1.3 $
 */
public class TaskConstraintDemo
{


    // **********************
    // ***** ATTRIBUTES *****
    // **********************


    // Action resource constraint definitions

    static final int AR_OwnActing = 0;
    static final int Operation_UNLOCK_DOOR = 0;
    static final int Operation_UNLOCK_DOOR_length = 10;
    static final int Operation_FORCE_DOOR = 1;
    static final int Operation_FORCE_DOOR_length = 60;


    // State resource constraint definitions

    static final int SR_DoorPassage
        = SymbolicSRC.registerNewResourceType( "Door.Passage" );
    static final int State_SR_DoorPassage_Locked
        = SymbolicSRC.registerNewState( SR_DoorPassage , "Locked" );
    static final int State_SR_DoorPassage_Possible
        = SymbolicSRC.registerNewState( SR_DoorPassage , "Possible" );
    static final int Contribution_SR_DoorPassage_ChangeToPossible
        = SymbolicSRC.registerNewContribution( SR_DoorPassage );

    static final int SR_KeyLocation
        = SymbolicSRC.registerNewResourceType( "Key.Location" );
    static final int State_SR_KeyLocation_Inventory
        = SymbolicSRC.registerNewState( SR_KeyLocation , "Inventory" );
    static final int State_SR_KeyLocation_Unknown
        = SymbolicSRC.registerNewState( SR_KeyLocation , "Unknown" );


    // Specifying possible action type configurations

    static final int Action_UNLOCK_DOOR
        = TaskConstraint.registerNewActionType();
    static final TCTask PTask1_Action_UNLOCK_DOOR = TaskConstraint.addPreconditionTaskToActionType( Action_UNLOCK_DOOR , SR_KeyLocation , State_SR_KeyLocation_Inventory );
    static final TCTask PTask2_Action_UNLOCK_DOOR = TaskConstraint.addPreconditionTaskToActionType( Action_UNLOCK_DOOR , SR_DoorPassage , State_SR_DoorPassage_Locked );
    static final TCTask ATask_Action_UNLOCK_DOOR = TaskConstraint.addActionTaskToActionType( Action_UNLOCK_DOOR, AR_OwnActing , Operation_UNLOCK_DOOR, Operation_UNLOCK_DOOR_length , 1 );
    static final TCTask STask_Action_UNLOCK_DOOR = TaskConstraint.addStateTaskToActionType( Action_UNLOCK_DOOR , SR_DoorPassage , Contribution_SR_DoorPassage_ChangeToPossible );

    static final int Action_FORCE_DOOR
        = TaskConstraint.registerNewActionType();
    static final TCTask PTask_Action_FORCE_DOOR = TaskConstraint.addPreconditionTaskToActionType( Action_FORCE_DOOR, SR_DoorPassage , State_SR_DoorPassage_Locked );
    static final TCTask ATask_Action_FORCE_DOOR = TaskConstraint.addActionTaskToActionType( Action_FORCE_DOOR, AR_OwnActing , Operation_FORCE_DOOR , Operation_FORCE_DOOR_length , 1 );
    static final TCTask STask_Action_FORCE_DOOR = TaskConstraint.addStateTaskToActionType( Action_FORCE_DOOR, SR_DoorPassage , Contribution_SR_DoorPassage_ChangeToPossible );


    // *******************
    // ***** METHODS *****
    // *******************


    public static void main( String[] args )
    {

        // Engine initialization

        GlobalSearchControl gsc = new GlobalSearchControl( 0 );


        // Configuring action resource constraints

        DiscreteResourceConstraint.setRepairTimeBuffer( 0 );


        // Configuring state resource constraints

        SymbolicSRC.setDefaultState( SR_DoorPassage ,
            State_SR_DoorPassage_Locked );
        SymbolicSRC.registerStateTransition( SR_DoorPassage , State_SR_DoorPassage_Locked , Contribution_SR_DoorPassage_ChangeToPossible , State_SR_DoorPassage_Possible , Action_UNLOCK_DOOR , Operation_UNLOCK_DOOR_length , 1 );


        SymbolicSRC.setDefaultState( SR_KeyLocation ,
            State_SR_KeyLocation_Unknown );


        // General configuration of the task constraint

        TaskConstraint.setRepairTimeBuffer( 0 );
        TaskConstraint.enableExecutionStateChanges( true );

        // Specifying temporal relations of the action configurations

        TaskConstraint.addTemporalRelationToActionType( Action_UNLOCK_DOOR , PTask1_Action_UNLOCK_DOOR , TaskConstraint.TEMPORAL_REFERENCE , TaskConstraint.EQUALS , ATask_Action_UNLOCK_DOOR , TaskConstraint.BEGIN , 0 );
        TaskConstraint.addTemporalRelationToActionType( Action_UNLOCK_DOOR , PTask2_Action_UNLOCK_DOOR , TaskConstraint.TEMPORAL_REFERENCE , TaskConstraint.EQUALS , ATask_Action_UNLOCK_DOOR , TaskConstraint.BEGIN , 0 );
        TaskConstraint.addTemporalRelationToActionType( Action_UNLOCK_DOOR , STask_Action_UNLOCK_DOOR , TaskConstraint.TEMPORAL_REFERENCE , TaskConstraint.EQUALS , ATask_Action_UNLOCK_DOOR , TaskConstraint.END , 0 );

        TaskConstraint.addTemporalRelationToActionType( Action_FORCE_DOOR , PTask_Action_FORCE_DOOR , TaskConstraint.TEMPORAL_REFERENCE , TaskConstraint.EQUALS , ATask_Action_FORCE_DOOR , TaskConstraint.BEGIN , 0 );
        TaskConstraint.addTemporalRelationToActionType( Action_FORCE_DOOR , STask_Action_FORCE_DOOR , TaskConstraint.TEMPORAL_REFERENCE , TaskConstraint.EQUALS , ATask_Action_FORCE_DOOR , TaskConstraint.END , 0 );


        // Creating current time structures

        Variable vCurrentTime = gsc.createVariable( "Time" , 75 , 40 );
        gsc.setVariable( vCurrentTime , 0 );

        Embedding e = new Embedding( 1 );
        gsc.extendEmbedding( e , vCurrentTime , Registry.PLAIN_EDGE ,
            Registry.INCOMING );
        ConventionalConstraint oCurrentTime
            = gsc.createConstraint( Registry.CURRENTTIME_OBJECTCONSTRAINT , e ,
            75 , 100 );


        // Creating action resource constraint

        Variable vRT_OwnActing
            = gsc.createVariable( "Resource Type" , 795 , 110 );
        gsc.setVariable( vRT_OwnActing , AR_OwnActing );
        Variable vC_OwnActing = gsc.createVariable( "Capacity" , 795 , 40 );
        gsc.setVariable( vC_OwnActing , 1 );

        e = new Embedding( 3 );
        gsc.extendEmbedding( e , oCurrentTime , Registry.PLAIN_EDGE ,
            Registry.INCOMING );
        gsc.extendEmbedding( e , vRT_OwnActing , Registry.RESOURCETYPE_EDGE ,
            Registry.INCOMING );
        gsc.extendEmbedding( e , vC_OwnActing , Registry.CAPACITY_EDGE ,
            Registry.INCOMING );

        gsc.selectCostFunction( gsc.getDefaultCostFunction() );
        DiscreteResourceConstraint cAR_OwnActing = (DiscreteResourceConstraint)gsc.createConstraint( Registry.DISCRETERESOURCE_CONSTRAINT , e , 795 , 190 );

        cAR_OwnActing.addDescription( Operation_UNLOCK_DOOR , "UNLOCK-DOOR" );
        cAR_OwnActing.addDescription( Operation_FORCE_DOOR , "FORCE-DOOR" );


        // Creating state resource constraint

        // Door's passage

        Variable vRT_DoorPassage
            = gsc.createVariable( "Resource Type" , 540 , 110 );
        gsc.setVariable( vRT_DoorPassage , SR_DoorPassage );
        Variable vS_DoorPassage = gsc.createVariable( "State" , 540 , 40 );
        gsc.setVariable( vS_DoorPassage , State_SR_DoorPassage_Locked );

        e = new Embedding( 3 );
        gsc.extendEmbedding( e , oCurrentTime , Registry.PLAIN_EDGE ,
            Registry.INCOMING );
        gsc.extendEmbedding( e , vRT_DoorPassage ,
            Registry.RESOURCETYPE_EDGE , Registry.INCOMING );
        gsc.extendEmbedding( e , vS_DoorPassage , Registry.STATE_EDGE ,
            Registry.INCOMING );

        gsc.selectCostFunction( gsc.getStructuralCostFunction() );
        ConventionalConstraint cSR_DoorPassage
            = gsc.createConstraint( Registry.STATERESOURCE_CONSTRAINT , e ,
            540 , 190 );

        gsc.selectCostFunction( gsc.getDefaultCostFunction() );
        gsc.addCosts( cSR_DoorPassage ,
            StateResourceConstraint.PRECONDITION_FAILURE );

        // Key's location

        Variable vRT_KeyLocation
            = gsc.createVariable( "Resource Type" , 260 , 110 );
        gsc.setVariable( vRT_KeyLocation , SR_KeyLocation );
        Variable vS_KeyLocation = gsc.createVariable( "State" , 260 , 40 );
        gsc.setVariable( vS_KeyLocation , State_SR_KeyLocation_Unknown );

        e = new Embedding( 3 );
        gsc.extendEmbedding( e , oCurrentTime , Registry.PLAIN_EDGE ,
            Registry.INCOMING );
        gsc.extendEmbedding( e , vRT_KeyLocation ,
            Registry.RESOURCETYPE_EDGE, Registry.INCOMING );
        gsc.extendEmbedding( e , vS_KeyLocation , Registry.STATE_EDGE ,
            Registry.INCOMING );

        gsc.selectCostFunction( gsc.getStructuralCostFunction() );
        ConventionalConstraint cSR_KeyLocation
            = gsc.createConstraint( Registry.STATERESOURCE_CONSTRAINT , e ,
            260 , 190 );

        gsc.selectCostFunction( gsc.getDefaultCostFunction() );
        gsc.addCosts( cSR_KeyLocation ,
            StateResourceConstraint.PRECONDITION_FAILURE );


        // Creating initial action structures

        // Creating the action object

        Variable vAT_Action
            = gsc.createVariable_internal( "Action Type" , 75 , 320 );
        gsc.setVariable_internal( vAT_Action , -1 );

        e = new Embedding( 1 );
        gsc.extendEmbedding_internal( e , vAT_Action ,
            Registry.ACTIONTYPE_EDGE , Registry.INCOMING );
        ConventionalConstraint oAction
            = gsc.createConstraint_internal( Registry.ACTION_OBJECTCONSTRAINT ,
            e , 75 , 260 );

        // Creating and connecting first precondition task

        Variable vTR_PT1 = gsc.createVariable_internal( "Time" , 450 , 320 );
        gsc.setVariable_internal( vTR_PT1 , 0 );
        Variable vS_PT1 = gsc.createVariable_internal( "State" , 450 , 390 );
        gsc.setVariable_internal( vS_PT1 , State_SR_DoorPassage_Locked );

        e = new Embedding( 3 );
        gsc.extendEmbedding_internal( e , vTR_PT1 ,
            Registry.TEMPORALREFERENCE_EDGE , Registry.INCOMING );
        gsc.extendEmbedding_internal( e , vS_PT1 , Registry.STATE_EDGE ,
            Registry.INCOMING );
        gsc.extendEmbedding_internal( e , vRT_DoorPassage ,
            Registry.RESOURCETYPE_EDGE , Registry.INCOMING );
        ConventionalConstraint oPreconditionTask1 = gsc.createConstraint_internal( Registry.PRECONDITIONTASK_OBJECTCONSTRAINT , e , 450 , 260 );

        e = new Embedding( 1 );
        gsc.extendEmbedding_internal( e , oPreconditionTask1 ,
            Registry.PLAIN_EDGE , Registry.INCOMING );
        gsc.extendConstraint_internal( oAction , e );
        gsc.extendConstraint_internal( cSR_DoorPassage , e );

        // Creating and connecting second precondition task

        Variable vTR_PT2 = gsc.createVariable_internal( "Time" , 260 , 320 );
        gsc.setVariable_internal( vTR_PT2 , 0 );
        Variable vS_PT2 = gsc.createVariable_internal( "State" , 260 , 390 );
        gsc.setVariable_internal( vS_PT2 , State_SR_KeyLocation_Inventory );

        e = new Embedding( 3 );
        gsc.extendEmbedding_internal( e , vTR_PT2 ,
            Registry.TEMPORALREFERENCE_EDGE , Registry.INCOMING );
        gsc.extendEmbedding_internal( e , vS_PT2 , Registry.STATE_EDGE ,
            Registry.INCOMING );
        gsc.extendEmbedding_internal( e , vRT_KeyLocation ,
            Registry.RESOURCETYPE_EDGE , Registry.INCOMING );
        ConventionalConstraint oPreconditionTask2 = gsc.createConstraint_internal( Registry.PRECONDITIONTASK_OBJECTCONSTRAINT , e , 260 , 260 );

        e = new Embedding( 1 );
        gsc.extendEmbedding_internal( e , oPreconditionTask2 ,
            Registry.PLAIN_EDGE , Registry.INCOMING );
        gsc.extendConstraint_internal( oAction , e );
        gsc.extendConstraint_internal( cSR_KeyLocation , e );

        // Creating and connecting state task

        Variable vTR_ST = gsc.createVariable_internal( "Time" , 630 , 320 );
        gsc.setVariable_internal( vTR_ST , 0 );
        Variable vC_ST
            = gsc.createVariable_internal( "Contribution" , 630 , 390 );
        gsc.setVariable_internal( vC_ST ,
            Contribution_SR_DoorPassage_ChangeToPossible );

        e = new Embedding( 3 );
        gsc.extendEmbedding_internal( e , vTR_ST ,
            Registry.TEMPORALREFERENCE_EDGE , Registry.INCOMING );
        gsc.extendEmbedding_internal( e , vC_ST , Registry.CONTRIBUTION_EDGE ,
            Registry.INCOMING );
        gsc.extendEmbedding_internal( e , vRT_DoorPassage ,
            Registry.RESOURCETYPE_EDGE , Registry.INCOMING );
        ConventionalConstraint oStateTask = gsc.createConstraint_internal( Registry.STATETASK_OBJECTCONSTRAINT , e , 630 , 260 );

        e = new Embedding( 1 );
        gsc.extendEmbedding_internal( e , oStateTask , Registry.PLAIN_EDGE ,
            Registry.INCOMING );
        gsc.extendConstraint_internal( oAction , e );
        gsc.extendConstraint_internal( cSR_DoorPassage , e );

        // Creating the task constraint

        e = new Embedding( 2 );
        gsc.extendEmbedding_internal( e , oCurrentTime ,
            Registry.PLAIN_EDGE , Registry.INCOMING );
        gsc.extendEmbedding_internal( e , oAction , Registry.PLAIN_EDGE ,
            Registry.INCOMING );

        gsc.selectCostFunction_internal( gsc.getStructuralCostFunction() );
        ConventionalConstraint cTask
            = gsc.createConstraint_internal( Registry.TASK_CONSTRAINT , e ,
                75 , 190 );
        gsc.selectCostFunction_internal( gsc.getDefaultCostFunction() );
        gsc.addCosts( cTask , TaskConstraint.TEMPORAL_INCONSISTENCY , new BigDecimal( "100" ) );


        // Displaying the graph

        gsc.createGraphics();

    }


}
