/*
 * GraphRepairer.java
 *
 * Created on 26. November 2002, 17:54
 */
package com.ai_center.dragonbreath.graphRepair;

import com.ai_center.dragonbreath.*;
import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.constraints.AllDifferentConstraint;
import com.ai_center.dragonbreath.graphRepair.RepairPlan.EdgeData;
import com.ai_center.dragonbreath.utilities.*;

import java.util.*;

import org.jdom.Element;


/**
 * Has funcitons to add and remove edges and nodes from the CSP-Graph. All
 * functions work properly with the extensions and basegraphs of the nodes.
 * That means that for a requested graphchange more operations might be
 * planned. All functions return RepairPlan objects.
 * 
 * @author tim baermann
 */
public class GraphRepairer {
    // Instance variables *****************************************************

    /** 
     * The GlobalSearchControl the GraphRepairer works on. 
     */
    protected GlobalSearchControl gsc;

    /** 
     * Edges that have been removed by the plan. 
     */
    protected HashSet removedEdges = new HashSet();

    /** 
     * Nodes that were removed 
     */
    protected HashSet removedNodes = new HashSet();

    /** 
     * Edges that were newly created during the planning process 
     */
    protected SetHashtable addedEdges = new SetHashtable();

    /** 
     * Nodes that were added by a plan, ordered by there nodeType. 
     */
    protected SetHashtable addedNodes = new SetHashtable();

    /**
     * Marker to know whether a function has been called by this or an external
     * evocation. If the call was external the addedEdges, addedNodes,
     * removedEdges and removedNodes containers are cleared after call.
     */
    protected boolean isInternalCall = false;

    /**
     * Where the GraphRepairer should ask for nodes that were not called to be
     * changed, but have to be for the demanded change to work.
     */
    NodeChooseInterface nodeChooser;

    /** 
     * The time the timing started, set with resetTime(). 
     */
    private long searchStartTime = 0;

    // Constructors ***********************************************************

    /**
     * Creates a new instance of GraphRepairer
     * 
     * @param _gsc Where the CSP-Graph can be found
     * @param _nodeChooser When a requested change requires further operations
     *        the nodechooser  is consulted which nodes should be changed.
     */
    public GraphRepairer(GlobalSearchControl _gsc, 
                         NodeChooseInterface _nodeChooser) {
        this.gsc = _gsc;

        if (_nodeChooser == null) {
            nodeChooser = new AllwaysCreateNodeChooser(gsc);
        } else {
            nodeChooser = _nodeChooser;
        }
    }


    /**
     * Creates a new instance of GraphRepairer A default nodechooser is used.
     * 
     * @param _gsc Where the CSP-Graph can be found
     */
    public GraphRepairer(GlobalSearchControl _gsc) {
        this.gsc    = _gsc;
        nodeChooser = new AllwaysCreateNodeChooser(gsc);
    }

    // Methods ****************************************************************

    /**
     * Returns if the given ConnectedNode is a necessary part of the basegraph
     * of one of its two nodes. I.e. the type is contained in the basegraph,
     * and there is no substitute present.
     * 
     * @param edge  The edge that might be contained in the Basegraph. Either a
     *        ConnectedNode or a EdgeData.
     * 
     * @return The node the edge is part of the basegraph. Null if of none.
     */
    public Node isEdgePartOfBaseGraph(Object edge) {
        Node from;
        Node to;
        int  edgeID;

        if (edge instanceof ConnectedNode) {
            ConnectedNode cn = (ConnectedNode) edge;

            if (cn.direction) {
                from = cn.node;
                to   = cn.other.node;
            } else {
                from = cn.other.node;
                to   = cn.node;
            }

            edgeID = cn.edgeID;
        } else {
            EdgeData ed = (EdgeData) edge;
            edgeID = ed.edgeID;
            from   = ed.from;
            to     = ed.to;
        }

        if (isEdgePartOfBaseGraph(from, to.typeID, edgeID, false)) {
            return from;
        } else if (isEdgePartOfBaseGraph(to, from.typeID, edgeID, true)) {
            return to;
        } else {
            return null;
        }
    }


    /**
     * Sets the used NodeChooser.
     * 
     * @param newNodeChooser The new nodeChooser.
     */
    public void setNodeChooser(NodeChooseInterface newNodeChooser) {
        nodeChooser = newNodeChooser;
    }


    /**
     * Returns the actual nodeChooser.
     * 
     * @return nodeChooser
     */
    public NodeChooseInterface getNodeChooser() {
        return nodeChooser;
    }


    /**
     * Adds an edge of the given type in the given direction from the given
     * node. the other node is determined by the function.
     * 
     * @param node  The node that should contain the edge.
     * @param edge  The edge that should be added
     * 
     * @return A plan of what to do to add the Edge.
     */
    public RepairPlan addEdge(Node node, ConnectedNode edge) {
        boolean cleanUpNecessary = false;

        if (! isInternalCall) {
            isInternalCall   = true;
            cleanUpNecessary = true;
        }

        RepairPlan plan = null;

        if (edge.nodeTypeID == Registry.ANY_NODE) {
            edge.nodeTypeID = getTypeForEdgeCreation(node, edge);
        }

        if (edgeOnlyInBasegraph(edge.nodeTypeID, node.typeID, edge.edgeID, 
                                ! edge.direction)) {
            // is the edge only addable by creating the other node?
            // the edge can only be added by the basegraph of the other node, so produce a new node 
            // of that type.
            ConnectedNode counterEdge = new ConnectedNode(node, node.typeID, 
                                                          edge.edgeID, 
                                                          ! edge.direction);

            if (edge.node == null) {
                plan = addNode_internal(edge.nodeTypeID, counterEdge);
            } else {
                nodeChooser.setNodeReplaced(edge.node);
                plan = addNode_internal(edge.nodeTypeID, counterEdge);
            }
        } else if (edgeOnlyInBasegraph(node.typeID, edge.nodeTypeID, 
                                       edge.edgeID, edge.direction)) {
            if (edge.node == null) {
                Node otherNode = nodeChooser.getFittingNodeForEdgeCreation(node, 
                                                                           edge);

                if ((otherNode == null) || (wasNodeRemoved(otherNode))) {
                    plan      = addNode(edge.nodeTypeID);
                    otherNode = (Node) plan.wildcard;
                }

                edge.node = otherNode;
            }

            nodeChooser.setNodeReplaced(node);

            if (plan == null) {
                plan = addNode_internal(node.typeID, edge);
            } else { //plan is the addition of the node
                plan.addSucceedingSubplan(addNode_internal(node.typeID, edge));
            }
        } else {
            boolean edgeAllreadyCreated = false;
            Node    otherNode = nodeChooser.getFittingNodeForEdgeCreation(node, 
                                                                          edge);

            if ((otherNode == null) || (wasNodeRemoved(otherNode))) {
                if (edgeInBaseGraph(edge.nodeTypeID, node.typeID, edge.edgeID, 
                                    ! edge.direction)) {
                    // the edge can be added by the basegraph of the other node
                    ConnectedNode counterEdge = new ConnectedNode(node, 
                                                                  node.typeID, 
                                                                  edge.edgeID, 
                                                                  ! edge.direction);
                    plan                = addNode_internal(edge.nodeTypeID, 
                                                           counterEdge);
                    edgeAllreadyCreated = true; // the requested edge was created by the basegraph of othernode.
                } else {
                    plan           = new RepairPlan(this);
                    plan.operation = plan.nothing;

                    RepairPlan subplan = addNode(edge.nodeTypeID);
                    plan.addSubplan(subplan);
                    otherNode = (Node) subplan.wildcard;
                }
            } else {
                plan           = new RepairPlan(this);
                plan.operation = plan.nothing;
            }

            if (! edgeAllreadyCreated) {
                Node from;
                Node to;

                if (edge.direction == Registry.OUTGOING) {
                    from = node;
                    to   = otherNode;
                } else {
                    from = otherNode;
                    to   = node;
                }

                RepairPlan addition = addEdge_internal(from, to, edge.edgeID);

                if (addition == null) { // the edge could not be added, that might happen, if the edge is only in the basegraph
                    plan.unregisterChanges();
                    plan = null;
                } else {
                    plan.addSubplan(addition);
                    plan.wildcard = addition.wildcard;
                }
            } // if edgeAllreadyCreated
        }

        if (cleanUpNecessary) {
            clearContainers();
            isInternalCall = false;
        } else {
            plan.registerChanges();
        }

        return plan;
    }


    /**
     * Finds a plan to create a node of the given type.
     * 
     * @param nodeTypeID The type of the new Node.
     * 
     * @return A plan of what to do to create the node
     */
    public RepairPlan addNode(int nodeTypeID) {
        boolean cleanUpNecessary = false;

        if (! isInternalCall) {
            isInternalCall   = true;
            cleanUpNecessary = true;
        }

        RepairPlan plan;

        if (nodeTypeID == Registry.ANY_NODE) {
            nodeTypeID = gsc.random.nextInt(Registry.nodeTypeNumber() - 1) + 1;

            // ANY_NODE is the first NodeID, so do not choose it.
        }

        plan              = new RepairPlan(this);
        plan.constraintID = nodeTypeID;

        Node node;
        node = nodeChooser.getFittingNode(nodeTypeID);

        if ((node == null) || (wasNodeRemoved(node))) { // a new node has to be created.
            node           = new Node();
            node.typeID    = nodeTypeID;
            plan.operation = plan.create;
            nodeChooser.registerNewNode(node);

            if (Registry.isConstraint(nodeTypeID)) {
                Embedding bg = ((ConventionalConstraint) Registry.nodeType[nodeTypeID]).getBaseGraph();
                plan.setRepairEmbeddingExample(bg);
                fillEmbeddingWithNewEdges(node, bg, plan);
            }

            plan.auxiliaryNodeData = nodeChooser.getAuxiliaryNodeData(node);
        } else { // node != null;
            plan.operation = plan.nothing;
        }

        plan.wildcard    = node;
        plan.changedNode = node;
        plan.registerChanges();

        if (cleanUpNecessary) {
            clearContainers();
            isInternalCall = false;
        }

        return plan;
    }


    /**
     * Finds a plan to create a node of the given type, which has allready the
     * given edge connected. Used to specify edges that can only be added via
     * the basegraph. Called by addEdge,  when that can only be done by this
     * function.
     * 
     * @param nodeTypeID The type of the new Node.
     * @param edge  an edge that should be contained in the node the node
     *        member of the edge has to be set.
     * 
     * @return A plan of what to do to create the Node, and the given edge.
     */
    public RepairPlan addNode_internal(int nodeTypeID, Object edge) {
        RepairPlan plan;
        Node       node;
        plan              = new RepairPlan(this);
        plan.operation    = plan.create;
        plan.constraintID = nodeTypeID;
        node              = new Node();
        node.typeID       = nodeTypeID;
        plan.wildcard     = node;
        plan.changedNode  = node;
        nodeChooser.registerNewNode(node);

        Embedding bg = ((ConventionalConstraint) Registry.nodeType[nodeTypeID]).getBaseGraph();
        setNodeAdded(node);
        plan.auxiliaryNodeData = nodeChooser.getAuxiliaryNodeData(node);
        plan.setRepairEmbeddingExample(bg);
        includeEdgeInEmbedding(edge, plan);
        fillEmbeddingWithNewEdges(node, bg, plan);
        plan.registerChanges();

        return plan;
    }


    /**
     * Finds a plan to remove the given Edge. Disconnects legally by reversing
     * extensions. That may include disconnecting more then only the two nodes
     * by using a bigger extension.
     * 
     * @param edge  The edge that should be removed. Either a ConnectedNode, or
     *        a EdgeData
     * 
     * @return The Plan. If null is Returned the edge can not be removed,  the
     *         edge belongs to one of the basegraphs, the caller has to know
     *         what to do.
     */
    public RepairPlan removeEdge(Object edge) {
        if (isEdgePartOfBaseGraph(edge) != null) {
            return null;
        }

        if (wasEdgeRemoved(edge)) {
            return new EmptyRepairPlan(this);
        }

        boolean cleanUpNecessary = false;

        if (! isInternalCall) {
            isInternalCall   = true;
            cleanUpNecessary = true;
            resetTime();
        }

        Vector possibleExtensions = getExtensions(edge);

        if (possibleExtensions.size() == 0) {
            if (cleanUpNecessary) {
                isInternalCall = false;
            }

            throw new Error("No Extension to remove edge" + edge + " found");
        }

        // first try if there is an extension that can be used without much work
        RepairPlan result = trySimpleEdgeRemovePlan(edge, possibleExtensions);

        if (result == null) {
            // try to find a more complicated removePlan, but not in a systematical way
            // this will allways result in one solution!
            result = greedyFindEdgeRemovePlan(edge, possibleExtensions);
        }

        if (result != null) {
            result.registerChanges();
        } else {
            throw new Error("No repair plan found!");
        }

        if (cleanUpNecessary) {
            clearContainers();
            isInternalCall = false;
        }

        return result;
    } //removeEdge()


    /**
     * Finds a plan to remove the given node, no matter if it is a Variable or
     * a Constraint. And also if the Constraint is not yet reduced to its
     * Basegraph, or the Variable still connected to Constraints.
     * 
     * @param n The node that should be deleted.
     * 
     * @return A plan to remove the node n.
     */
    public RepairPlan removeNode(Node n) {
        if (n == null) {
            return null;
        }

        boolean cleanUpNecessary = false;

        if (! isInternalCall) {
            isInternalCall   = true;
            cleanUpNecessary = true;
            resetTime();
        } else if (wasNodeRemoved(n)) {
            return new EmptyRepairPlan(this);
        }

        RepairPlan plan = new RepairPlan(this);
        plan.operation   = plan.delete;
        plan.changedNode = n;

        reduceToBaseGraph(n, plan);

        if (n instanceof ConventionalConstraint) {
            Embedding bg = ((ConventionalConstraint) Registry.nodeType[n.typeID]).getBaseGraph();
            plan.setRepairEmbeddingExample(bg);

            if (fillEmbeddingWithExistingEdges(plan)) {
                setAllEdgesRemoved(plan);
            } else {
                throw new Error("Could not reduce Node " + n.toString() + 
                                " to BaseGraph.");
            }
        }

        setNodeRemoved(n);

        if (cleanUpNecessary) {
            clearContainers();
            isInternalCall = false;
        }

        return plan;
    } //removeNode()


    /**
     * If the node was previousely marked as removed, this funktion undoes the
     * change.
     * 
     * @param n The node that should not be removed any more.
     */
    public void resetNodeRemoved(Node n) {
        if (! removedNodes.remove(n)) {
            addedNodes.add(Registry.getNodeType(n), n);
        }
    }


    /**
     * Returns all new edges for the Node key.
     * 
     * @param key  The node that got new edges.
     * 
     * @return Either null, a single EdgeData or a Set of EdgeData.
     */
    protected Object getAddedEdges(Node key) {
        return addedEdges.get(key);
    }


    /**
     * Returns all nodes of the given type that were added by a subplan until
     * now.
     * 
     * @param nodeType The searched type.
     * 
     * @return A single Node or a set of nodes or null if none exists.
     */
    protected Object getAddedNodes(Node nodeType) {
        return addedNodes.get(nodeType);
    }


    /**
     * Stores all edges in the given plan as added. Not considering subplans.
     * 
     * @param plan  The plan.
     */
    protected void setAllEdgesAdded(RepairPlan plan) {
        if (plan != null) {
            for (int eNr = 0; eNr < plan.getNumEdgesInEmbedding(); eNr++) {
                Object edge = plan.getEdgeAt(eNr);

                if (edge instanceof EdgeData) {
                    setEdgeAdded((EdgeData) edge);
                }
            }
        }
    }


    /**
     * Stores all edges in the given Plan's embedding as removed.
     * 
     * @param plan  The plan.
     */
    protected void setAllEdgesRemoved(RepairPlan plan) {
        if (plan != null) {
            for (int eNr = 0; eNr < plan.getNumEdgesInEmbedding(); eNr++) {
                Object edge = plan.getEdgeAt(eNr);
                setEdgeRemoved(edge);
            }
        }
    }


    /**
     * If an Edge is added by a subplan it is registered here, so it can be
     * used or removed by later subPlans.
     * 
     * @param edge The edge added.
     */
    protected void setEdgeAdded(EdgeData edge) {
        addedEdges.add(edge.from, edge);
        addedEdges.add(edge.to, edge);
    }


    /**
     * Returns true if the given ConnectedNode is a necessary part of the
     * basegraph of the constraint, i.e. the type is contained in the
     * basegraph, and there is no substitute present. False else.
     * 
     * @param n  The node whos Basegraph should be checked.
     * @param nodeTypeID The typeID as defined in Registry.
     * @param edgeID The edgeID as defined in Registry.
     * @param direction The direction as defined in Registry.
     * 
     * @return Returns whether cn is necessary for c to exist or not.
     */
    protected boolean isEdgePartOfBaseGraph(Node n, int nodeTypeID, int edgeID, 
                                            boolean direction) {
        if (Registry.isConstraint(n.typeID)) {
            ConventionalConstraint nodeType = (ConventionalConstraint) Registry.nodeType[n.typeID];
            Embedding              bg = nodeType.getBaseGraph();

            int                    edgesInBG = bg.allConnectedNodeTypeOccurences(
                                                       nodeTypeID, edgeID, 
                                                       direction).size();

            if (edgesInBG > 0) {
                boolean isCriterionNode      = false;
                boolean isCriterionNodeType  = true;
                boolean isCriterionEdgeID    = true;
                boolean isCriterionDirection = true;
                Node    node                 = null;
                int     nr                   = 0;

                if (n.connections != null) {
                    n.connections.initializeEnumeration(isCriterionNode, 
                                                        isCriterionNodeType, 
                                                        isCriterionEdgeID, 
                                                        isCriterionDirection, 
                                                        node, 
                                                        Registry.nodeType[nodeTypeID], 
                                                        edgeID, direction);

                    ConnectedNode next;

                    while ((next = n.connections.getNext()) != null) {
                        if (! wasEdgeRemoved(next)) {
                            nr++;
                        }

                        if (nr > edgesInBG) {
                            return false;
                        }
                    } //while
                }

                Object newConnections = addedEdges.get(n);

                if (newConnections != null) {
                    if (newConnections instanceof Set) {
                        Iterator it = ((Set) newConnections).iterator();

                        while (it.hasNext()) {
                            EdgeData newEdge           = (EdgeData) it.next();
                            int      newEdgeNodeTypeID = 
                                    newEdge.getOtherNode(n).typeID;
                            boolean  newEdgeDirection  = 
                                    ((n == newEdge.to)
                                         ? Registry.INCOMING : Registry.OUTGOING);
                            int      newEdgeID         = newEdge.edgeID;

                            if ((newEdgeNodeTypeID == nodeTypeID) && 
                                    (newEdgeID == edgeID) && 
                                    (newEdgeDirection == direction)) {
                                nr++;
                            } //if

                            if (nr > edgesInBG) {
                                return false;
                            }
                        } //while
                    } else {
                        EdgeData newEdge           = (EdgeData) newConnections;
                        int      newEdgeNodeTypeID = newEdge.getOtherNode(n).typeID;
                        boolean  newEdgeDirection  = ((n == newEdge.to)
                                                          ? Registry.INCOMING
                                                          : Registry.OUTGOING);
                        int      newEdgeID         = newEdge.edgeID;

                        if ((newEdgeNodeTypeID == nodeTypeID) && 
                                (newEdgeID == edgeID) && 
                                (newEdgeDirection == direction)) {
                            nr++;
                        } //if

                        if (nr > edgesInBG) {
                            return false;
                        }
                    }
                }

                return true;
            }
        }

        return false;
    } //isEdgePartOfBaseGraph()


    /**
     * Is the given Edge part of the basegraph of the given node?
     * 
     * @param n  The node Basegraph of which should be checked.
     * @param edge  The edge that might be contained in the Basegraph. Either a
     *        ConnectedNode or a EdgeData.
     * 
     * @return true if edge is part of the basegraph of node n, else false.
     */
    protected boolean isEdgePartOfBaseGraphOfNode(Node n, Object edge) {
        int     edgeID;
        int     nodeTypeID;
        boolean direction;

        if (edge instanceof ConnectedNode) {
            if (n == ((ConnectedNode) edge).node) {
                edge = ((ConnectedNode) edge).other;
            }

            ConnectedNode cn = (ConnectedNode) edge;
            nodeTypeID = cn.nodeTypeID;
            edgeID     = cn.edgeID;
            direction  = cn.direction;
        } else {
            EdgeData ed = (EdgeData) edge;
            edgeID     = ed.edgeID;
            nodeTypeID = ed.getOtherNode(n).typeID;
            direction  = ed.getDirection(n);
        }

        return isEdgePartOfBaseGraph(n, nodeTypeID, edgeID, direction);
    }


    /**
     * Stores the given edge as removed.
     * 
     * @param edge  The edge that is not removed any more.
     */
    protected void setEdgeRemoved(Object edge) {
        if (edge instanceof ConnectedNode) {
            ConnectedNode cn = (ConnectedNode) edge;

            if (cn.direction) {
                removedEdges.add(cn);
            } else {
                removedEdges.add(cn.other);
            }
        } else {
            addedEdges.remove(((EdgeData) edge).from, edge);
            addedEdges.remove(((EdgeData) edge).to, edge);
        }
    }


    /**
     * Returns the edges of node that have the same type as edgeExample. If the
     * edgeExample has the to-node set, the node is also used as criterion for
     * type comparison.
     * 
     * @param node  The node the edges of which should be collected.
     * @param edgeExample  The type of the searched edge.
     * 
     * @return A vector of edges. Either ConnectedNodes or EdgeData.
     */
    protected Vector getEdgesOfTypeInNode(Node node, Object edgeExample) {
        Vector result = new Vector();

        // count the edges in the csp-graph
        if (node.connections != null) {
            prepareIterationForEdgeType(node, edgeExample);

            ConnectedNode cn;

            while ((cn = node.connections.getNext()) != null) {
                if (! wasEdgeRemoved(cn)) {
                    result.add(cn);
                }
            }
        }

        //now count the new edges
        Object o = getAddedEdges(node);

        if (o != null) {
            if (o instanceof EdgeData) {
                EdgeData ed = (EdgeData) o;

                if (ed.isSameType(edgeExample)) {
                    result.add(ed);
                }
            } else {
                Iterator it = ((Set) o).iterator();

                while (it.hasNext()) {
                    EdgeData ed = (EdgeData) it.next();

                    if (ed.isSameType(edgeExample)) {
                        result.add(ed);
                    }
                } //while
            }
        }

        return result;
    }


    /**
     * Produces a random permutation of all extensions that contain an edge of
     * the same type as the given edge. Each extension is stored in a Pair,
     * together with the constraint it belongs to.
     * 
     * @param edge  The edge that should be included. Either a ConnectedNode or
     *        an EdgeData. Both nodes have to exist!
     * 
     * @return A Vector of all Pairs (c,ext),  with c the constraint that
     *         contained the extension, e the edge in that  extension and ext
     *         the extension itself.
     */
    protected Vector getExtensions(Object edge) {
        Vector result     = new Vector();
        Node   from;
        Node   to;
        int    edgeID;
        int    nodeTypeID;

        if (edge instanceof ConnectedNode) {
            ConnectedNode cn = (ConnectedNode) edge;

            if (cn.direction) {
                from = cn.node;
                to   = cn.other.node;
            } else {
                from = cn.other.node;
                to   = cn.node;
            }

            edgeID = cn.edgeID;
        } else {
            EdgeData ed = (EdgeData) edge;
            edgeID = ed.edgeID;
            from   = ed.from;
            to     = ed.to;
        }

        int  index;

        Node nodeType = Registry.getNodeType(to);

        if (nodeType instanceof ConventionalConstraint) {
            nodeTypeID = from.typeID;

            Iterator it = ((ConventionalConstraint) nodeType).getMatchingExtensions(
                                  nodeTypeID, edgeID, Registry.INCOMING)
                                                           .iterator();

            while (it.hasNext()) {
                Embedding ext        = (Embedding) it.next();
                int       resultSize = result.size();


                // put the extension somewhere in the result
                index = gsc.random.nextInt(resultSize + 1);
                index += gsc.random.nextInt(ext.connectedNodeNumber); // make bigger extensions less probable.

                if (index > resultSize) {
                    index = resultSize;
                }

                result.add(index, new Pair(to, ext));
            } //while
        }

        nodeType = Registry.getNodeType(from);

        if (nodeType instanceof ConventionalConstraint) {
            nodeTypeID = to.typeID;

            Iterator it = ((ConventionalConstraint) nodeType).getMatchingExtensions(
                                  nodeTypeID, edgeID, Registry.OUTGOING)
                                                           .iterator();

            while (it.hasNext()) {
                Embedding ext        = (Embedding) it.next();
                int       resultSize = result.size();


                // put the extension somewhere in the result
                index = gsc.random.nextInt(resultSize + 1);
                index += gsc.random.nextInt(ext.connectedNodeNumber); // make bigger extensions less probable.

                if (index > resultSize) {
                    index = resultSize;
                }

                result.add(index, new Pair(from, ext));
            } //while
        }

        return result;
    }


    /**
     * Adds the given node to the added nodes.
     * 
     * @param n The node that was added.
     */
    protected void setNodeAdded(Node n) {
        addedNodes.add(Registry.getNodeType(n), n);
    }


    /**
     * Stores the given Node as removed.
     * 
     * @param n  The Node.
     */
    protected void setNodeRemoved(Node n) {
        if (! addedNodes.remove(Registry.getNodeType(n), n)) {
            removedNodes.add(n);
        }
    }


    /**
     * When an arbitrary node should be connected to an other node via a given
     * edge this function chooses a nodeType for the created node, that can
     * have the given edge connected.
     * 
     * @param node The one node of the Edge.
     * @param edge The describtion of the requested edge.
     * 
     * @return A node type as defined in Registry that can be used to create
     *         the given edge.
     */
    protected int getTypeForEdgeCreation(Node node, ConnectedNode edge) {
        Vector typesWithEdge = new Vector();

        if (node instanceof ConventionalConstraint) {
            // first add all nodeTypes that could be connected via extensions of node
            Embedding bg                 = ((ConventionalConstraint) node).getBaseGraph();
            Vector    nodeTypeOccurences = bg.allConnectedNodeTypeOccurences(
                                                   edge.nodeTypeID, edge.edgeID, 
                                                   edge.direction);
            Iterator  it                 = nodeTypeOccurences.iterator();

            while (it.hasNext()) {
                Integer index = (Integer) it.next();
                typesWithEdge.add(
                        new Integer(
                                bg.connectedNode[index.intValue()].nodeTypeID));
            } //while

            int numExtensions = ((ConventionalConstraint) node).getExtensionGraphNumber();

            for (int i = 0; i < numExtensions; i++) {
                Embedding ext = ((ConventionalConstraint) node).getExtensionGraph(i);
                nodeTypeOccurences = ext.allConnectedNodeTypeOccurences(
                                             edge.nodeTypeID, edge.edgeID, 
                                             edge.direction);
                it                 = nodeTypeOccurences.iterator();

                while (it.hasNext()) {
                    Integer index = (Integer) it.next();
                    typesWithEdge.add(
                            new Integer(
                                    ext.connectedNode[index.intValue()].nodeTypeID));
                } //while
            }
        }

        // now add all nodeTypes that could be connected via the extension of bg of the other node
        for (int typeID = 1; typeID < Registry.nodeType.length; typeID++) {
            if (Registry.nodeType[typeID] instanceof ConventionalConstraint) {
                ConventionalConstraint cType = (ConventionalConstraint) Registry.nodeType[typeID];
                Embedding              bg = cType.getBaseGraph();

                if (bg.containsConnectedNodeType(node.typeID, edge.edgeID, 
                                                 ! edge.direction)) {
                    typesWithEdge.add(new Integer(typeID));
                } else {
                    int     numExtensions = cType.getExtensionGraphNumber();
                    int     i     = 0;
                    boolean found = false;

                    while ((i < numExtensions) && ! found) {
                        Embedding ext = cType.getExtensionGraph(i);

                        if (ext.containsConnectedNodeType(node.typeID, 
                                                          edge.edgeID, 
                                                          ! edge.direction)) {
                            typesWithEdge.add(new Integer(typeID));
                            found = true;
                        }

                        i++;
                    } //while i
                } //else
            } //if conventionalConstraint
        } //for

        if (typesWithEdge.size() == 0) {
            throw new Error("The edge " + edge.toString() + 
                            " can not be connected to Node " + 
                            node.toString());
        }

        int     chosenTypeIndex = gsc.random.nextInt(typesWithEdge.size());
        Integer chosenType = (Integer) typesWithEdge.elementAt(chosenTypeIndex);

        return chosenType.intValue();
    }


    /**
     * Finds a plan to add the requested Edge. If the edge can only be added
     * via the Basegraph it is not created!
     * 
     * @param from  The node the edge starts at
     * @param to  The node the edge goes to.
     * @param edgeID  The type of edge.
     * 
     * @return A plan of what to do to add the Edge.
     */
    protected RepairPlan addEdge_internal(Node from, Node to, int edgeID) {
        boolean cleanUpNecessary = false;

        if (! isInternalCall) {
            isInternalCall   = true;
            cleanUpNecessary = true;
        }

        // this is caught in the addEdge() method, which calls this method.
        // if(   edgeOnlyInBasegraph(from.typeID, to.typeID, edgeID, Registry.OUTGOING)
        //    edgeOnlyInBasegraph(to.typeID, from.typeID, edgeID, Registry.INCOMING)) {
        RepairPlan plan               = new RepairPlan(this);
        EdgeData   edge = new EdgeData(from, to, edgeID);
        plan.operation = plan.extend;
        plan.wildcard  = edge;

        Vector possibleExtensions = getExtensions(edge);

        if (possibleExtensions.size() > 0) {
            Pair extensionData = (Pair) possibleExtensions.firstElement();
            Node c = (Node) extensionData.first;
            plan.changedNode = c;

            Embedding extension = (Embedding) extensionData.second;
            plan.setRepairEmbeddingExample(extension);

            includeEdgeInEmbedding(edge, plan);
            fillEmbeddingWithNewEdges(c, extension, plan);

            if (cleanUpNecessary) {
                clearContainers();
                isInternalCall = false;
            } else {
                plan.registerChanges();
            }
        } else {
            throw new Error("No extension found for edge " + 
                            Registry.nodeTypeName[from.typeID] + " -- " + 
                            Registry.edgeTypeName[edgeID] + "--> " + 
                            Registry.nodeTypeName[to.typeID]);
        }

        return plan;
    }


    /**
     * Clears all the Containers addedEdges, addedNodes, removedEdges and
     * removedNodes.
     */
    protected void clearContainers() {
        addedEdges.clear();
        addedNodes.clear();
        removedEdges.clear();
        removedNodes.clear();
    }


    /**
     * If the given extension needs edges to be used, this function creates all
     * that are missing. (Registeres the changes needed)
     * 
     * @param plan  The plan that contains the extension and where subplans
     *        should be entered.
     */
    protected void createMissingEdgesForRemoval(RepairPlan plan) {
        Vector                 allreadyCheckedEdges = new Vector();
        ConventionalConstraint c       = (ConventionalConstraint) plan.changedNode;
        Embedding              pattern = plan.repairEmbeddingExample;

        for (int cnNr = 0; (cnNr < pattern.connectedNodeNumber); cnNr++) {
            if (! allreadyCheckedEdges.contains(new Integer(cnNr))) {
                ConnectedNode exampleEdge             = 
                        pattern.connectedNode[cnNr];
                Vector        sameEdgeTypeInExtension = 
                        pattern.allConnectedNodeTypeOccurences(exampleEdge);
                allreadyCheckedEdges.addAll(sameEdgeTypeInExtension);

                int numEdgesOfTypeInExtension = sameEdgeTypeInExtension.size();
                int numMissingEdges = numEdgesOfTypeInExtension - 
                                      numEdgesMoreThenInBaseGraph(c, 
                                                                  exampleEdge);

                if (numMissingEdges > 0) {
                    //edges have to be created.
                    for (int i = 0; i < numMissingEdges; i++) { // exactly overlap nodes are missing

                        ConnectedNode missingEdge = new ConnectedNode(
                                                            exampleEdge.nodeTypeID, 
                                                            exampleEdge.edgeID, 
                                                            exampleEdge.direction);
                        plan.addSubplan(addEdge(c, missingEdge));
                        numMissingEdges = numEdgesOfTypeInExtension - 
                                          numEdgesMoreThenInBaseGraph(c, 
                                                                      exampleEdge);
                    }
                }

                //if one of the removed edge is part of the basegraph of the other node it has to be replaced
                Vector   edgesOfTypeInNode = getEdgesOfTypeInNode(c, 
                                                                  exampleEdge);
                Iterator it = edgesOfTypeInNode.iterator();

                while (it.hasNext()) {
                    Object edge = it.next();
                    Node   node = isEdgePartOfBaseGraph(edge);

                    if ((node != null) && (node != c)) {
                        ConnectedNode missingEdge = new ConnectedNode(c.typeID, 
                                                                      exampleEdge.edgeID, 
                                                                      ! exampleEdge.direction);
                        plan.addSubplan(addEdge(c, exampleEdge));
                    }
                } //while
            }
        } //for
    } //createMissingEdgesForRemoval()


    /**
     * Compares two edges, no matter what types they have. If they have the
     * nodes defined these are also used for comparison.
     * 
     * @param node  The node that both edges could be connected to.
     * @param createdNode  The node that should be created
     * @param existingNode  The node that allready exists
     * 
     * @return true if has the same type, false else.
     */
    protected boolean edgeHasSameType(Node node, Object createdNode, 
                                      Object existingNode) {
        Node    otherNode1;
        Node    otherNode2;
        int     edgeID1;
        int     edgeID2;
        int     nodeTypeID1;
        int     nodeTypeID2;
        boolean direction1;
        boolean direction2;

        if (createdNode instanceof EdgeData) {
            EdgeData ed = (EdgeData) createdNode;
            edgeID1     = ed.edgeID;
            otherNode1  = ed.getOtherNode(node);
            nodeTypeID1 = otherNode1.typeID;
            direction1  = ed.getDirection(node);
        } else {
            ConnectedNode cn = (ConnectedNode) createdNode;
            otherNode1  = cn.node;
            edgeID1     = cn.edgeID;
            nodeTypeID1 = cn.nodeTypeID;
            direction1  = cn.direction;
        }

        if (existingNode instanceof EdgeData) {
            EdgeData ed = (EdgeData) existingNode;
            edgeID2     = ed.edgeID;
            otherNode2  = ed.getOtherNode(node);
            nodeTypeID2 = otherNode2.typeID;
            direction2  = ed.getDirection(node);
        } else {
            ConnectedNode cn = (ConnectedNode) existingNode;
            otherNode2  = cn.node;
            edgeID2     = cn.edgeID;
            nodeTypeID2 = cn.nodeTypeID;
            direction2  = cn.direction;
        }

        if (otherNode1 == null) {
            otherNode2 = null;
        }

        return (edgeID1 == edgeID2) && (nodeTypeID1 == nodeTypeID2) && 
               (direction1 == direction2) && (otherNode1 == otherNode2);
    }


    /**
     * Does the Basegraph of a node of the type testedNodeTypeID contain an
     * edge of the given type?
     * 
     * @param testedNodeTypeID The type of the node the basegraph of which
     *        should be tested as defined in Registry.
     * @param otherNodeTypeID The type of the node at the other end of the edge
     *        as defined in Registry.
     * @param edgeID The type of the edge as defined in Registry.
     * @param direction The direction of the edge as defined in Registry.
     * 
     * @return true if the basegraph contains the edge else false.
     */
    protected boolean edgeInBaseGraph(int testedNodeTypeID, int otherNodeTypeID, 
                                      int edgeID, boolean direction) {
        if (testedNodeTypeID != Registry.VARIABLE) {
            ConventionalConstraint testedNodeType = (ConventionalConstraint) Registry.nodeType[testedNodeTypeID];

            return testedNodeType.getBaseGraph()
                                 .containsConnectedNodeType(otherNodeTypeID, 
                                                            edgeID, direction);
        } else {
            return false;
        }
    }


    /**
     * Is the given node only found in the basegraph of the testedNode
     * 
     * @param testedNodeTypeID The type of the node the basegraph of which
     *        should be tested as defined in Registry.
     * @param otherNodeTypeID The type of the node at the other end of the edge
     *        as defined in Registry.
     * @param edgeID The type of the edge as defined in Registry.
     * @param direction The direction of the edge as defined in Registry.
     * 
     * @return true if basegraph contains the edge and no extension, else
     *         false.
     */
    protected boolean edgeOnlyInBasegraph(int testedNodeTypeID, 
                                          int otherNodeTypeID, int edgeID, 
                                          boolean direction) {
        if (testedNodeTypeID != Registry.VARIABLE) {
            if (edgeInBaseGraph(testedNodeTypeID, otherNodeTypeID, edgeID, 
                                direction)) {
                ConventionalConstraint testedNodeType = 
                        (ConventionalConstraint) Registry.nodeType[testedNodeTypeID];

                if (testedNodeType.getMatchingExtensions(otherNodeTypeID, 
                                                         edgeID, direction)
                                  .size() == 0) {
                    if (otherNodeTypeID != Registry.VARIABLE) {
                        ConventionalConstraint otherNodeType = 
                                (ConventionalConstraint) Registry.nodeType[otherNodeTypeID];

                        if (otherNodeType.getMatchingExtensions(
                                    testedNodeTypeID, edgeID, direction).size() == 0) {
                            return true;
                        }
                    } else {
                        return true;
                    }
                }
            }
        }

        return false;
    }


    /**
     * Fills the plans embedding with nodes from the csp-graph. e is the
     * pattern of how to fill the plan. If e is no basegraph this function
     * doesn`t assign edges that should belong to the basegraph.
     * 
     * @param plan  Where the embedding is filled. And where subplans have to
     *        be added, if necessary.
     * 
     * @return true if the embedding could be filled, else false.
     */
    protected boolean fillEmbeddingWithExistingEdges(RepairPlan plan) {
        boolean                found;
        ConventionalConstraint c           = (ConventionalConstraint) plan.changedNode;
        Embedding              pattern     = plan.repairEmbeddingExample;
        boolean                isBaseGraph = pattern == c.getBaseGraph();

        // step through embedding and search for each not interpreted ConnectedNode a counterpart in c
        for (int cnNr = 0; (cnNr < pattern.connectedNodeNumber); cnNr++) {
            Object edge = plan.getEdgeAt(cnNr);

            if (edge == null) { // this edge needs an interpretation

                ConnectedNode exampleEdge = pattern.connectedNode[cnNr];
                found = false;

                int numMissingEdges;

                if (isBaseGraph) {
                    numMissingEdges = 0;
                } else {
                    int numEdgesOfType = pattern.allConnectedNodeTypeOccurences(
                                                 exampleEdge.nodeTypeID, 
                                                 exampleEdge.edgeID, 
                                                 exampleEdge.direction).size();
                    numMissingEdges = numEdgesOfType - 
                                      numEdgesMoreThenInBaseGraph(c, 
                                                                  exampleEdge);
                }

                if (numMissingEdges <= 0) {
                    Vector   fittingEdges = getEdgesOfTypeInNode(c, exampleEdge);
                    Iterator it = fittingEdges.iterator();

                    while (it.hasNext() && ! found) {
                        Object interpretation = it.next();
                        Node   otherNode;

                        if (interpretation instanceof ConnectedNode) {
                            otherNode = ((ConnectedNode) interpretation).node;
                        } else {
                            otherNode = ((EdgeData) interpretation).getOtherNode(c);
                        }

                        if (! plan.embeddingContainsEdge(interpretation) && 
                                ! isEdgePartOfBaseGraphOfNode(otherNode, 
                                                              interpretation)) {
                            plan.addEmbeddingEdge(cnNr, interpretation);
                            found = true;
                        }
                    } //while
                }

                if (! found) {
                    return false;
                }
            } //if edge == null
        } //for

        return true;
    } //fillEmbeddingWithExistingEdges()


    /**
     * Given a Constraint c and an Embedding this function completes the
     * embedding with new ConnectedNodes. It decides whether to take existing
     * nodes for an extension or create new ones, via the nodeChooser.
     * 
     * @param c  The constraint the embedding e is Extension of. If Constraint
     *        not yet created a key to the later created one.
     * @param e  An Embedding that is an Extension of c.
     * @param plan  Where the embedding is filled. And where subplans have to
     *        be added, if necessary. Is initialized with the right embedding
     *        size.
     */
    protected void fillEmbeddingWithNewEdges(Node c, Embedding e, 
                                             RepairPlan plan) {
        boolean found;

        // step through embedding and search for each not interpreted ConnectedNode a counterpart in c
        for (int cnNr = 0; (cnNr < e.connectedNodeNumber); cnNr++) {
            Object edge = plan.getEdgeAt(cnNr);

            if (edge == null) { // this connectedNode needs an interpretation

                ConnectedNode cnPrototype = e.connectedNode[cnNr];
                Node          otherNode = nodeChooser.getFittingNodeForEdgeCreation(c, 
                                                                                    cnPrototype);

                if ((otherNode == null) || wasNodeRemoved(otherNode)) {
                    RepairPlan subplan = addNode(cnPrototype.nodeTypeID);
                    otherNode = (Node) subplan.wildcard;
                    plan.addSubplan(subplan);
                }

                plan.addEmbeddingEdge(cnNr, c, otherNode, cnPrototype.edgeID, 
                                      cnPrototype.direction);
            } //if
        } //for
    }


    /**
     * Adds or removes further nodes and edges to make one of the given
     * extensions fit.
     * 
     * @param edge  The edge that should be removed.
     * @param possibleExtensions  The extensionst that could be  used to remove
     *        the edge, as received by getExtensions(edge)
     * 
     * @return A plan to remove the edge, which might have subplans, or null,
     *         if none found.
     */
    protected RepairPlan greedyFindEdgeRemovePlan(Object edge, 
                                                  Vector possibleExtensions) {
        RepairPlan plan = new RepairPlan(this);
        plan.operation = RepairPlan.reduce;

        Iterator               it            = possibleExtensions.iterator();
        Pair                   extensionData = (Pair) it.next();
        ConventionalConstraint c             = (ConventionalConstraint) extensionData.first;
        Embedding              extension     = (Embedding) extensionData.second;
        plan.changedNode = c;
        plan.setRepairEmbeddingExample(extension);
        includeEdgeInEmbedding(edge, plan);
        createMissingEdgesForRemoval(plan);

        if (fillEmbeddingWithExistingEdges(plan)) {
            return plan;
        }

        throw new Error("Could not remove edge!");
    }


    /**
     * Puts the given edge into the embedding of the plan.  The edge has to be
     * one with an ANY_EDGE edgeID. Called from includeEdgeInEmbedding(). The
     * place where it can be added is determined by the pattern.
     * 
     * @param edge  The edge that should be included
     * @param plan  The plan the edge is added to.
     * 
     * @return Could the edge be added?
     */
    protected boolean includeAnyEdgeInEmbedding(EdgeData edge, RepairPlan plan) {
        int       nodeTypeID;
        int       edgeID;
        boolean   direction;
        Embedding pattern = plan.repairEmbeddingExample;
        nodeTypeID = (edge.from == plan.changedNode)
                         ? edge.to.typeID : edge.from.typeID;
        edgeID     = edge.edgeID;
        direction  = edge.getDirection(plan.changedNode);

        int   possibleEdgeLastIndex = 0;
        int[] possibleEdges = new int[pattern.connectedNodeNumber];

        for (int cnNr = 0; (cnNr < pattern.connectedNodeNumber); cnNr++) {
            ConnectedNode cnInEmbedding = pattern.connectedNode[cnNr];

            if ((cnInEmbedding.nodeTypeID == nodeTypeID) && 
                    (cnInEmbedding.direction == direction)) {
                Object existingEdge = plan.getEdgeAt(cnNr);

                if (existingEdge == null) { // this connectedNode needs an interpretation
                    possibleEdges[possibleEdgeLastIndex] = cnNr;
                    possibleEdgeLastIndex++;
                }
            }
        } //for

        if (possibleEdgeLastIndex > 0) {
            int           chosenEdgeIndex = possibleEdges[gsc.random.nextInt(
                                                                  possibleEdgeLastIndex)];
            ConnectedNode cnInEmbedding = pattern.connectedNode[chosenEdgeIndex];
            edge.edgeID = cnInEmbedding.edgeID;
            plan.addEmbeddingEdge(chosenEdgeIndex, edge);

            return true;
        }

        return false;
    } //includeAnyEdgeInEmbedding()


    /**
     * Puts the given edge into the embedding of the plan.  The place where it
     * can be added is determined by the pattern.
     * 
     * @param edge  The edge that should be included, either an EdgeData or a
     *        connectedNode.
     * @param plan  The plan the edge is added to.
     * 
     * @return Could the edge be added?
     */
    protected boolean includeEdgeInEmbedding(Object edge, RepairPlan plan) {
        int       nodeTypeID;
        int       edgeID;
        boolean   direction;
        Embedding pattern = plan.repairEmbeddingExample;

        if (edge instanceof ConnectedNode) {
            ConnectedNode cn = (ConnectedNode) edge;

            if (cn.other == null) { // this is no existing edge!
                edge = new EdgeData(cn, plan.changedNode);
            } else {
                if (((ConnectedNode) edge).node == plan.changedNode) {
                    cn = ((ConnectedNode) edge).other;
                } else {
                    cn = (ConnectedNode) edge;
                }
            }

            nodeTypeID = cn.nodeTypeID;
            edgeID     = cn.edgeID;
            direction  = cn.direction;
        } else {
            EdgeData ed = (EdgeData) edge;
            nodeTypeID = (ed.from == plan.changedNode)
                             ? ed.to.typeID : ed.from.typeID;
            edgeID     = ed.edgeID;
            direction  = ed.getDirection(plan.changedNode);
        }

        if (edgeID == Registry.ANY_EDGE) {
            // here it is clear that edge is instanceof EdgeData, an existing edge can not be ANY_Type
            return includeAnyEdgeInEmbedding((EdgeData) edge, plan);
        }

        for (int cnNr = 0; (cnNr < pattern.connectedNodeNumber); cnNr++) {
            ConnectedNode cnInEmbedding = pattern.connectedNode[cnNr];

            if ((cnInEmbedding.nodeTypeID == nodeTypeID) && 
                    (cnInEmbedding.edgeID == edgeID) && 
                    (cnInEmbedding.direction == direction)) {
                Object existingEdge = plan.getEdgeAt(cnNr);

                if (existingEdge == null) { // this connectedNode needs an interpretation
                    plan.addEmbeddingEdge(cnNr, edge);

                    return true;
                }
            }
        } //for

        return false;
    } //includeEdgeInEmbedding()


    /**
     * Returns the number of edges of the given type of c, that sober the edges
     * in the basegraph.
     * 
     * @param c  The node.
     * @param edgeExample  The type of the edge as can be found as member of an
     *        extension.
     * 
     * @return An integer number for the number of sobering edges.
     */
    protected int numEdgesMoreThenInBaseGraph(ConventionalConstraint c, 
                                              ConnectedNode edgeExample) {
        Embedding bg               = c.getBaseGraph();
        int       edgesInBaseGraph = bg.allConnectedNodeTypeOccurences(
                                             edgeExample.nodeTypeID, 
                                             edgeExample.edgeID, 
                                             edgeExample.direction).size();
        int       numEdgesInC      = getEdgesOfTypeInNode(c, edgeExample)
                                         .size();

        return numEdgesInC - edgesInBaseGraph;
    }


    /**
     * Removes all connections that are due to extensions of this or of other
     * constraints. When these were removed the constraint can be savely
     * deleted via deleteConstraint() If n is a Variable all edges are
     * removed.
     * 
     * @param n  The Constraint ,or the variable, that should be reduced.
     * @param plan The plan all actions to reduce the given node are added as
     *        subplans.
     */
    protected void reduceToBaseGraph(Node n, RepairPlan plan) {
        Vector subplans             = new Vector();
        Vector allreadyCheckedEdges = new Vector();
        Vector edges                = new Vector();
        int    bgSize               = 0;

        if (n instanceof ConventionalConstraint) {
            bgSize = ((ConventionalConstraint) n).getBaseGraph().connectedNodeNumber;
        }

        while (true) {
            edges.clear();

            Object edge;

            if (n == null) {
                throw new Error("Null Node should be reduced.");
            }

            n.connections.initializeEnumerationAll();

            while ((edge = n.connections.getNext()) != null) {
                if (! wasEdgeRemoved(edge)) {
                    edges.add(edge);
                }
            }

            Object o = getAddedEdges(n);

            if (o instanceof Set) {
                edges.addAll((Set) o);
            } else if (o instanceof EdgeData) {
                edges.add(o);
            }

            if (edges.size() == bgSize) {
                Iterator it = subplans.iterator();

                while (it.hasNext()) {
                    RepairPlan subplan = (RepairPlan) it.next();
                    plan.addSubplan(subplan);
                } //while

                return;
            }

            Iterator it = edges.iterator();

            while (it.hasNext()) {
                edge = it.next();

                if (! (isEdgePartOfBaseGraphOfNode(n, edge) || wasEdgeRemoved(
                                                                       edge))) {
                    Node    otherNode;
                    int     edgeID;
                    boolean direction; // the direction as seen from n

                    if (edge instanceof ConnectedNode) {
                        otherNode = ((ConnectedNode) edge).node;
                        direction = ((ConnectedNode) edge).direction;
                        edge      = ((ConnectedNode) edge).other;
                        edgeID    = ((ConnectedNode) edge).edgeID;
                    } else {
                        otherNode = ((EdgeData) edge).getOtherNode(n);
                        edgeID    = ((EdgeData) edge).edgeID;
                        direction = ((EdgeData) edge).getDirection(n);
                    }

                    if (isEdgePartOfBaseGraphOfNode(otherNode, edge)) {
                        if (nodeChooser.removeNodeOrElseAddEdge(otherNode, edge)) {
                            RepairPlan p = removeNode(otherNode);

                            if (p == null) {
                                throw new Error("Nullplan added");
                            }

                            subplans.add(p);
                        } else {
                            setNodeRemoved(n); // for the addEdge plan not to use this same node for the new edge.

                            ConnectedNode cn      = new ConnectedNode(n.typeID, 
                                                                      edgeID, 
                                                                      ! direction);
                            RepairPlan    subplan = addEdge(otherNode, cn);
                            resetNodeRemoved(n);

                            if (subplan == null) { //the edge could not be added, the edge is only available in Basegraph

                                RepairPlan p = removeNode(otherNode);

                                if (p == null) {
                                    throw new Error("Nullplan added");
                                }

                                subplans.add(p);
                            } else {
                                subplans.add(subplan); //order is important!

                                //now there is a substitute for the edge, remove it
                                RepairPlan removeSubplan = removeEdge(edge);

                                if (removeSubplan == null) { //  the substitution did not work, or added a new node, 
                                                             //  so that not only the edge was replaced, but also the othernode

                                    //  the substitution did not work, or added a new node, 
                                    //  so that not only the edge was replaced, but also the othernode
                                    RepairPlan p = removeNode(otherNode);

                                    if (p == null) {
                                        throw new Error("Nullplan added");
                                    }

                                    subplans.add(p);
                                } else {
                                    subplans.add(removeSubplan);
                                }
                            }
                        }
                    } else {
                        RepairPlan p = removeEdge(edge);

                        if (p == null) {
                            throw new Error("Nullplan added");
                        }

                        subplans.add(p);
                    }
                }
            }
        }
    }


    /**
     * Calls resetEdgeAdded for all edges of the plan.
     * 
     * @param plan  The plan.
     */
    protected void resetAllEdgesAdded(RepairPlan plan) {
        if (plan != null) {
            for (int eNr = 0; eNr < plan.getNumEdgesInEmbedding(); eNr++) {
                Object edge = plan.getEdgeAt(eNr);

                if (edge instanceof EdgeData) {
                    resetEdgeAdded((EdgeData) edge);
                }
            }
        }
    }


    /**
     * Undoes store of Edges of plan as removed.
     * 
     * @param plan  The plan.
     */
    protected void resetAllEdgesRemoved(RepairPlan plan) {
        if (plan != null) {
            for (int eNr = 0; eNr < plan.getNumEdgesInEmbedding(); eNr++) {
                Object edge = plan.getEdgeAt(eNr);
                resetEdgeRemoved(edge);
            }
        }
    }


    /**
     * Removes a node from the list of the added Edges.
     * 
     * @param edge The edge added.
     */
    protected void resetEdgeAdded(EdgeData edge) {
        addedEdges.remove(edge.from, edge);
        addedEdges.remove(edge.to, edge);
    }


    /**
     * Resets a marking of the edge as removed.
     * 
     * @param edge  The edge that is not removed any more.
     */
    protected void resetEdgeRemoved(Object edge) {
        if (edge instanceof ConnectedNode) {
            ConnectedNode cn = (ConnectedNode) edge;

            if (cn.direction) {
                removedEdges.remove(cn);
            } else {
                removedEdges.remove(cn.other);
            }
        } else {
            addedEdges.add(((EdgeData) edge).from, edge);
            addedEdges.add(((EdgeData) edge).to, edge);
        }
    }


    /**
     * Removes the given node from the added nodes.
     * 
     * @param n The node that is not added any more.
     */
    protected void resetNodeAdded(Node n) {
        addedNodes.remove(Registry.getNodeType(n), n);
    }


    /**
     * Starts the timer.
     */
    protected void resetTime() {
        searchStartTime = java.lang.System.currentTimeMillis();
    }


    /**
     * Tries if the given extensions can be used to remove the edge without
     * adding or removing much else.
     * 
     * @param edge  The edge that should be removed.
     * @param possibleExtensions  A vector as delivered by getExtensions(edge);
     * 
     * @return If the edge could be removed a repairplan, else null;
     */
    protected RepairPlan trySimpleEdgeRemovePlan(Object edge, 
                                                 Vector possibleExtensions) {
        RepairPlan plan = new RepairPlan(this);
        plan.operation = RepairPlan.reduce;

        Iterator it = possibleExtensions.iterator();

        while (it.hasNext()) {
            Pair                   extensionData = (Pair) it.next();
            ConventionalConstraint c         = (ConventionalConstraint) extensionData.first;
            Embedding              extension = (Embedding) extensionData.second;
            plan.changedNode = c;
            plan.setRepairEmbeddingExample(extension);

            if (includeEdgeInEmbedding(edge, plan) && 
                    fillEmbeddingWithExistingEdges(plan)) {
                return plan;
            }
        } //while

        return null;
    }


    /**
     * Was the given edge added?
     * 
     * @param edge the edge.
     * 
     * @return addedEdges contains edge.
     */
    protected boolean wasEdgeAdded(EdgeData edge) {
        return addedEdges.contains(edge.from, edge);
    }


    /**
     * Was the given connectedNode removed by any part of the plan so far? And
     * if edge is a ConnectedNode, it is also tested, if it is still in the
     * connections of its node in the CSP-graph.
     * 
     * @param edge The edge to be tested.
     * 
     * @return true if the edge or the node at its end was removed, else false.
     */
    protected boolean wasEdgeRemoved(Object edge) {
        if (edge instanceof ConnectedNode) {
            ConnectedNode cn = (ConnectedNode) edge;

            if (! cn.node.connections.contains(cn.other)) {
                return true;
            }

            if (cn.direction) {
                return removedEdges.contains(cn);
            } else {
                return removedEdges.contains(cn.other);
            }
        } else {
            return ! addedEdges.contains(((EdgeData) edge).from, edge);
        }
    }


    /**
     * Was the given Node removed by now?
     * 
     * @param n  The Node to be tested.
     * 
     * @return true if the node was removed by a part of the plan.
     */
    protected boolean wasNodeRemoved(Node n) {
        boolean removedNodesContains = removedNodes.contains(n);
        boolean addedNodesContains = addedNodes.contains(Registry.getNodeType(n), n);

        return removedNodesContains || 
               ! (gsc.containsNode(n) || addedNodesContains);
    }


    /**
     * Calls c.connections.initializeEnumeration() with the right parameters to
     * get all edges of the type cn. If the edge has the other node it
     * connects specified (ergo it is not null) the node is also used as
     * criterion for comparison.
     * 
     * @param c The node.
     * @param edge The edgetype that should be retrieved.
     */
    private void prepareIterationForEdgeType(Node c, Object edge) {
        Vector  result               = new Vector();
        boolean isCriterionNode;
        boolean isCriterionNodeType  = true;
        boolean isCriterionEdgeID    = true;
        boolean isCriterionDirection = true;
        Node    node;
        Node    nodeType;
        int     edgeId;
        boolean direction;

        if (edge instanceof ConnectedNode) {
            ConnectedNode cn = (ConnectedNode) edge;
            nodeType  = Registry.nodeType[cn.nodeTypeID];
            edgeId    = cn.edgeID;
            direction = cn.direction;

            if (cn.node == null) {
                node            = null;
                isCriterionNode = false;
            } else {
                node            = cn.node;
                isCriterionNode = true;
            }
        } else {
            EdgeData ed = (EdgeData) edge;
            node            = ed.getOtherNode(c);
            nodeType        = Registry.nodeType[node.typeID];
            edgeId          = ed.edgeID;
            direction       = ed.getDirection(c);
            isCriterionNode = true;
        }

        c.connections.initializeEnumeration(isCriterionNode, 
                                            isCriterionNodeType, 
                                            isCriterionEdgeID, 
                                            isCriterionDirection, node, 
                                            nodeType, edgeId, direction);
    }
}