/*
 * HeuristicNodeChooser.java
 *
 * Created on 17. Januar 2003, 13:14
 */
package com.ai_center.dragonbreath.graphRepair;

import com.ai_center.dragonbreath.*;
import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.testGraph.*;
import com.ai_center.dragonbreath.utilities.*;

import java.util.*;

import org.jdom.Element;


/**
 * Whenever the repair process is not straingth forward, the NodeCooser is
 * called, to resolve the problem. This is the standart NodeChooser that lets
 * the SCGraphRepairer decide what to do, so that the currently repaired
 * StructuralConstraintGraphPart is improved.
 * 
 * @version $Revision: 1.5 $
 * @author tim baermann
 */
public class HeuristicNodeChooser implements NodeChooseInterface {
    // Instance variables *****************************************************

    /** 
     * Link to the GlobalSearchControl of this application. 
     */
    GlobalSearchControl myGSC;

    /** 
     * The nodechooser that has the first and last word in choice. 
     */
    NodeChooseInterface mainNodeSource;

    /**
     * The StructuralConstraintManagement the nodeChooser gets its nodes from.
     */
    StructuralConstraintManagement mySCM;

    // Constructors ***********************************************************

    /**
     * Creates a new instance of HeuristicNodeChooser
     * 
     * @param scm  The StructuralConstraintManagement to be used.
     * @param gsc The GlobalSearchControl to be used.
     */
    public HeuristicNodeChooser(StructuralConstraintManagement scm, 
                                GlobalSearchControl gsc) {
        mySCM = scm;
        myGSC = gsc;
    }

    // Methods ****************************************************************

    /**
     * When a variable was created the GraphRepairer asks via getVariableValue
     * what value to set as default for the new Variable.
     * 
     * @param n The node that was created and should be initialized.
     * 
     * @return A jdom.Element with the node s initial data or null.
     */
    public Element getAuxiliaryNodeData(Node n) {
        if (mainNodeSource != null) {
            return mainNodeSource.getAuxiliaryNodeData(n);
        }

        return null;
    }


    /**
     * Returns either an allready existing Node, or null, if the GraphRepairer
     * should create its own new one. The node has as many of the given edges
     * as possible.
     * 
     * @param nodeTypeID The id of the node, as in Registry.
     * 
     * @return A node of the type nodeTypeID, or null, if none found.
     */
    public Node getFittingNode(int nodeTypeID) {
        if (mainNodeSource != null) {
            Node result = mainNodeSource.getFittingNode(nodeTypeID);

            if (result != null) {
                if (nodeStillExists(result)) {
                    return result;
                } else {
                    mainNodeSource.setNodeReplaced(result);

                    return null;
                }
            } else if (! Registry.isObjectConstraint(nodeTypeID)) {
                // only reuse ObjectConstraints
                return null;
            }
        }

        if (myGSC.random.nextDouble() < Registry.PROBABILITY_USE_EXISTING_NODE) {
            Object o = mySCM.getNodesForType(Registry.nodeType[nodeTypeID]);

            if (o == null) {
                return null;
            } else if (o instanceof Node) {
                if (isNodeOk((Node) o)) {
                    return (Node) o;
                } else {
                    return null;
                }
            }

            Set      nodes = (Set) o;
            Iterator it = nodes.iterator();

            while (it.hasNext()) {
                Node n = (Node) it.next();

                if (isNodeOk(n)) {
                    return n;
                }
            } //while
        }

        return null;
    }


    /**
     * Returns either an allready existing Node, that can be used to connect n
     * to, or null, if the GraphRepairer should create its own new one.
     * 
     * @param node  The node the edge comes from. May be null, if Node not yet
     *        created.
     * @param edge  The edge. The returned node will later be the node member
     *        of that edge.
     * 
     * @return Either a node from GlobalSearchControl or null.
     */
    public Node getFittingNodeForEdgeCreation(Node node, ConnectedNode edge) {
        if (mainNodeSource != null) {
            Node result = mainNodeSource.getFittingNodeForEdgeCreation(node, 
                                                                       edge);

            if (result != null) {
                if (nodeStillExists(result)) {
                    return result;
                } else {
                    mainNodeSource.setNodeReplaced(result);

                    return null;
                }
            } else if (! Registry.isObjectConstraint(edge.nodeTypeID)) {
                // only reuse ObjectConstraints
                return null;
            }
        }

        if (myGSC.random.nextDouble() < Registry.PROBABILITY_USE_EXISTING_NODE) {
            Vector potentialNodes = new Vector();

            // make a vector of the set, to be able to access via the index.
            Object nodes = mySCM.getNodesForType(
                                   Registry.nodeType[edge.nodeTypeID]);

            if (nodes != null) {
                if (nodes instanceof Set) {
                    // filter out nodes that are not accepted by the mainNodeSource.
                    Iterator it = ((Set) nodes).iterator();

                    while (it.hasNext()) {
                        Node n = (Node) it.next();

                        if (isNodeOk(n)) {
                            potentialNodes.add(n);
                        }
                    } //while
                } else {
                    if (isNodeOk((Node) nodes)) {
                        potentialNodes.add(nodes);
                    }
                }
            }

            int nodesSize = potentialNodes.size();

            if (nodesSize == 0) {
                return null;
            }

            int[]    votes             = new int[nodesSize];
            int      winnerVotes       = nodesSize + 1;
            int      winnerIndex       = 0;
            Vector   actualEdgesVector = null;
            boolean  useDifferentEdges;
            Iterator instanceIt        = mySCM.getSCInstances().iterator();

            while (instanceIt.hasNext()) {
                StructuralConstraintInstance instance = 
                        (StructuralConstraintInstance) instanceIt.next();

                for (int index = 0; index < nodesSize; index++) {
                    Node n = (Node) potentialNodes.elementAt(index);

                    if (instance.nacCanUseNewEdge(n, edge)) {
                        votes[index]++;
                    }
                } //for
            } //while

            // now find a winner
            for (int index = 0; index < nodesSize; index++) {
                if (votes[index] < winnerVotes) { //least votes wins
                    winnerVotes = votes[index];
                    winnerIndex = index;
                }
            }

            if (((winnerVotes == 0) || 
                        (myGSC.random.nextDouble() > Registry.PROBABILITY_USE_USED_NODE)) && 
                    (potentialNodes.size() > 0)) {
                Node result = (Node) potentialNodes.elementAt(winnerIndex);

                if (mainNodeSource != null) {
                    registerNewNode(result);
                }

                return result;
            }
        }

        return null;
    }


    /**
     * Before a StructuralConstraintInstance is repairing its TestGraph it
     * registered at the nodeChooser, so that when a new node is to be added
     * it is the first to be asked for a candidate.
     * 
     * @param source Where should this NodeChooser take the informations from?
     *        normally the SCGraphRepairer.
     */
    public void setMainNodeSource(NodeChooseInterface source) {
        mainNodeSource = source;
    }


    /**
     * When a node is chosen to be used it can be tested for compliance with
     * the nodeChooser. Especially used to not use nodes for repair of a
     * subgraph, that were allready assigned to an other subgraph. Allways
     * called by a nodechooser from its mainNodeSource.
     * 
     * @param n The tested node.
     * 
     * @return true if the node can be used, else false.
     */
    public boolean isNodeOk(Node n) {
        if (mainNodeSource != null) {
            return mainNodeSource.isNodeOk(n);
        } else {
            return nodeStillExists(n);
        }
    }


    /**
     * When a node has to be removed to add an edge that is only available via
     * the basegraph this function is called, so that the old interpretation
     * can be removed.
     * 
     * @param n The node that was replaced.
     */
    public void setNodeReplaced(Node n) {
        if (mainNodeSource != null) {
            mainNodeSource.setNodeReplaced(n);
        }
    }


    /**
     * Each time a NodeChooseInterface was called with
     * getFittingNodeForEdgeCreation() or getFittingNode() and returned null
     * it has to be called again as soon as the requested node was produced
     * with the new node.
     * 
     * @param node The new Node.
     */
    public void registerNewNode(Node node) {
        if (mainNodeSource != null) {
            mainNodeSource.registerNewNode(node);
        }
    }


    /**
     * Returns whether the node should be removed, or the edge added. Called if
     * an edge should be removed, but can not due to the fact that it is part
     * of the basegraph of the node. In such a case the addition of a new edge
     * of the same type could help, as it could then be used for the
     * basegraph.
     * 
     * @param node  The node.
     * @param edge  The edge that should be removed, but can not.
     * 
     * @return true if the node should be removed, false if an edge should be
     *         added.
     */
    public boolean removeNodeOrElseAddEdge(Node node, Object edge) {
        if (mainNodeSource != null) {
            return mainNodeSource.removeNodeOrElseAddEdge(node, edge);
        }

        return true;
    }


    /**
     * Is the given node still part of the gsc?
     * 
     * @param n The tested node.
     * 
     * @return true if the node is still in the gsc, else false.
     */
    protected boolean nodeStillExists(Node n) {
        return myGSC.containsNode(n);
    }
}