/*
 * NeverCreateNodeChooser.java
 *
 * Created on 17. Januar 2003, 12:27
 */
package com.ai_center.dragonbreath.graphRepair;

import com.ai_center.dragonbreath.*;
import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.*;
import com.ai_center.dragonbreath.utilities.*;

import java.util.*;

import org.jdom.Element;


/**
 * Used for testing purpose. Tries to use old nodes as often as possible.
 * 
 * @author tim baermann
 */
public class NeverCreateNodeChooser implements NodeChooseInterface {
    // Instance variables *****************************************************

    /** 
     * Link to the GlobalSearchControl of this application. 
     */
    GlobalSearchControl myGSC;

    /** 
     * The nodechooser that has the first and last word in choice. 
     */
    NodeChooseInterface mainNodeSource;

    /**
     * The StructuralConstraintManagement the nodeChooser gets its nodes from.
     */
    StructuralConstraintManagement mySCM;

    // Constructors ***********************************************************

    /**
     * Creates a new instance of NeverCreateNodeChooser
     * 
     * @param scm  The StructuralConstraintManagement to be used.
     * @param gsc The GlobalSearchControl to be used.
     */
    public NeverCreateNodeChooser(StructuralConstraintManagement scm, 
                                  GlobalSearchControl gsc) {
        mySCM = scm;
        myGSC = gsc;
    }

    // Methods ****************************************************************

    /**
     * When a Node was created the GraphRepairer asks via
     * getAuxiliaryNodeData() what values to set as default.
     * 
     * @param node The new node.
     * 
     * @return A jdom.Element with the relevant data.
     */
    public Element getAuxiliaryNodeData(Node node) {
        if (mainNodeSource != null) {
            return mainNodeSource.getAuxiliaryNodeData(node);
        } else {
            return null;
        }
    }


    /**
     * Returns either an allready existing Node, or null, if the GraphRepairer
     * should create its own new one.
     * 
     * @param nodeTypeID The id of the node, as in Registry.
     * 
     * @return A node that has the right typeID, or null.
     */
    public Node getFittingNode(int nodeTypeID) {
        Object o = mySCM.getNodesForType(Registry.nodeType[nodeTypeID]);

        if (mainNodeSource != null) {
            Node result = mainNodeSource.getFittingNode(nodeTypeID);

            if (result != null) {
                return result;
            }
        }

        if (o == null) {
            return null;
        } else if (o instanceof Node) {
            return (Node) o;
        } else {
            Object[] nodes = ((Set) o).toArray();
            int      index = myGSC.random.nextInt(nodes.length);

            return (Node) nodes[index];
        }
    }


    /**
     * Returns either an allready existing Node, that can be used to connect
     * node to, or null, if the GraphRepairer should create its own new one.
     * 
     * @param node  The node the edge comes from. May be null, if Node not yet
     *        created.
     * @param cn  The edge. The returned node will later be the node member of
     *        that edge.
     * 
     * @return Either a node from GlobalSearchControl or null.
     */
    public Node getFittingNodeForEdgeCreation(Node node, ConnectedNode cn) {
        // first try if the currently repaired testgraph has a special wish for the constraint.
        if (mainNodeSource != null) {
            Node result = mainNodeSource.getFittingNodeForEdgeCreation(node, cn);

            if (result != null) {
                return result;
            }
        }

        Object o = mySCM.getNodesForType(Registry.nodeType[cn.nodeTypeID]);

        if (o == null) {
            return null;
        } else if (o instanceof Node) {
            return (Node) o;
        }

        Set nodes = (Set) o;

        return (Node) nodes.iterator().next();
    }


    /**
     * Before a StructuralConstraintInstance is repairing its TestGraph it
     * registered at the nodeChooser, so that when a new node is to be added
     * it is the first to be asked for a candidate.
     * 
     * @param nodeChooser The new MainNodeSource.
     */
    public void setMainNodeSource(NodeChooseInterface nodeChooser) {
        mainNodeSource = nodeChooser;
    }


    /**
     * When a node is chosen to be used it can be tested for compliance with
     * the nodeChooser. Allways called by a nodechooser from its
     * mainNodeSource.
     * 
     * @param n The tested Node.
     * 
     * @return mainNodeSource.isNodeOk(n)
     */
    public boolean isNodeOk(Node n) {
        if (mainNodeSource != null) {
            return mainNodeSource.isNodeOk(n);
        }

        return true;
    }


    /**
     * When a node had to be replaced in order to create a requested edge this
     * method is called. That happens especially if an edge that is part of
     * the basegraph of the node was requested.
     * 
     * @param node The node that was replaced.
     */
    public void setNodeReplaced(Node node) {
        if (mainNodeSource != null) {
            mainNodeSource.setNodeReplaced(node);
        }
    }


    /**
     * Each time a NodeChooseInterface was called with
     * getFittingNodeForEdgeCreation() or getFittingNode() and returned null
     * it has to be called again as soon as the requested node was produced
     * with the new node.
     * 
     * @param node The new node.
     */
    public void registerNewNode(Node node) {
        if (mainNodeSource != null) {
            mainNodeSource.registerNewNode(node);
        }
    }


    /**
     * Returns whether the node should be removed, or the edge added. Called if
     * an edge should be removed, but can not due to the fact that it is part
     * of the basegraph of the node. In such a case the addition of a new edge
     * of the same type could help, as it could then be used for the
     * basegraph.
     * 
     * @param node  The node.
     * @param edge  The edge that should be removed, but can not.
     * 
     * @return true if the node should be removed, false if an edge should be
     *         added. Both answers have the same chance.
     */
    public boolean removeNodeOrElseAddEdge(Node node, Object edge) {
        if (node instanceof Variable) {
            return true;
        }

        if (myGSC.random.nextFloat() > 0.5) {
            return true;
        } else {
            return false;
        }
    }
}