/*
 * RepairPlan.java
 *
 * Created on 26. November 2002, 18:34
 */
package com.ai_center.dragonbreath.graphRepair;

import com.ai_center.dragonbreath.*;
import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.constraints.*;
import com.ai_center.dragonbreath.utilities.*;

import java.util.*;

import org.jdom.Element;


/**
 * A plan of how to repair the CSP-Graph.
 * 
 * @author tim baermann
 */
public class RepairPlan {
    // Static variables/initializers ******************************************

    /** 
     * When creating variables this is a counter to give them distinct names. 
     */
    static int variableNr = 0;

    // Values that can be set to the member operation.

    /** 
     * Add edges to the changedNode 
     */
    public static final int extend = 2;

    /** 
     * Create a new Node 
     */
    public static final int create = 1;

    /** 
     * This plan is only used as container for its subplans 
     */
    public static final int nothing = 0;

    /** 
     * Remove edges from changedNode 
     */
    public static final int reduce = -1;

    /** 
     * Remove changedNodes. 
     */
    public static final int delete = -2;

    // Instance variables *****************************************************

    /**
     * Holds an Embedding that determins the form of the extension that is used
     * for repair.
     */
    public Embedding repairEmbeddingExample;

    /** 
     * The node that should be extendet or reduced by this plan. 
     */
    public Node changedNode = null;

    /**
     * If this plan creates a certain Node or edge this is a placeholder for
     * the real object until it was created.
     */
    public Object wildcard = null;

    /**
     * If this Plan exists to create a node this is the constraintId of the
     * node.
     */
    public int constraintID;

    /** 
     * Should the node be extended , or reduced, or removed? 
     */
    public int operation = 0;

    /** 
     * Data used to initialize a newly created node 
     */
    Element auxiliaryNodeData = null;

    /** 
     * The GlobalSearchControl the GraphRepairer works on. 
     */
    GlobalSearchControl gsc;

    /** 
     * The GraphRepairer, that created this Plan. 
     */
    GraphRepairer graphrepairer;

    /** 
     * If this plan is a subplan this is the superPlan.. 
     */
    RepairPlan superPlan = null;

    /** 
     * The results of the subplans. 
     */
    private HashMap subplanResults = new HashMap();

    /**
     * If there are further steps to be done before this plan can be executed
     * here they are.
     */
    private Vector subplans = null;

    /** 
     * Subplans that are handled after this plan was done. 
     */
    private Vector succeedingSubplans;

    /** 
     * The edges that should be part of the embedding 
     */
    private Object[] repairEmbedding = null;

    /** 
     * The number of edges in repairEmbedding. 
     */
    private int edgesInEmbedding = 0;

    // Constructors ***********************************************************

    /**
     * Creates a new instance of RepairPlan
     * 
     * @param repairer The GraphRepairer that created this plan.
     */
    public RepairPlan(GraphRepairer repairer) {
        graphrepairer = repairer;
        gsc           = repairer.gsc;
    }


    /**
     * Creates a new instance of RepairPlan  Uses the same data as the given
     * example.
     * 
     * @param example The example RepairPlan that is copied.
     */
    public RepairPlan(RepairPlan example) {
        this.graphrepairer     = example.graphrepairer;
        this.gsc               = example.gsc;
        this.auxiliaryNodeData = example.auxiliaryNodeData;
        this.constraintID      = example.constraintID;
        this.operation         = example.operation;
        this.changedNode       = example.changedNode;
        this.superPlan         = example.superPlan;
        this.wildcard          = example.wildcard;
        this.setRepairEmbeddingExample(example.repairEmbeddingExample);

        for (int i = 0; i < repairEmbedding.length; i++) {
            repairEmbedding[i] = example.getEdgeAt(i);
        }
    }

    // Methods ****************************************************************

    /**
     * Returns the actual repairEmbedding entry at the given number.
     * 
     * @param index The number of the edge in the repairEmedding that should be
     *        retrieved.
     * 
     * @return One edge of the repairEmbedding.
     */
    public Object getEdgeAt(int index) {
        if (index >= edgesInEmbedding) {
            throw new Error("Not so many edges in RepairEmbedding");
        }

        return repairEmbedding[index];
    }


    /**
     * Returns all edges in the repairEmbedding with the given type.
     * 
     * @param other  The node that is not part of cn.
     * @param cn  The example edge. As it normally comes from an Extension it
     *        has no valid other-member
     * 
     * @return A Vector of all edges that have the given edgeType.
     */
    public Vector getFittingEdges(Node other, ConnectedNode cn) {
        Vector result = new Vector();

        for (int i = 0; i < repairEmbedding.length; i++) {
            if (repairEmbedding[i] instanceof EdgeData) {
                if (((EdgeData) repairEmbedding[i]).isSameType(cn)) {
                    result.add(repairEmbedding[i]);
                }
            } else {
                ConnectedNode embEdge = (ConnectedNode) repairEmbedding[i];

                if (cn.isSameType(embEdge)) {
                    result.add(embEdge);
                } else {
                    if ((embEdge.other.direction == cn.direction) && 
                            (embEdge.node == other) && 
                            (embEdge.edgeID == cn.edgeID) && 
                            (changedNode.typeID == cn.nodeTypeID)) {
                        result.add(embEdge.other);
                    }
                }
            }
        } //for

        return result;
    }


    /**
     * Returns the number of edges in repairEmbedding.
     * 
     * @return A positive number.
     */
    public int getNumEdgesInEmbedding() {
        return edgesInEmbedding;
    }


    /**
     * Sets the repairEmbeddingExample member. Prepares the Plan to receive
     * edges for the embedding via addEmbeddingEdge().
     * 
     * @param e  The embedding.
     */
    public void setRepairEmbeddingExample(Embedding e) {
        repairEmbeddingExample = e;

        int size = e.connectedNodeNumber;
        repairEmbedding  = new Object[size];
        edgesInEmbedding = size;
    }


    /**
     * Returns a Vector with all subplans (only the ones comming before this
     * plan)
     * 
     * @return A Vector of subplans. An empty Vector if none.
     */
    public Vector getSubplans() {
        if (subplans == null) {
            return new Vector();
        }

        return subplans;
    }


    /**
     * Sets the superPlan.
     * 
     * @param plan The plan this plan is subplan of.
     */
    public void setSuperPlan(RepairPlan plan) {
        superPlan = plan;
    }


    /**
     * Returns the superPlan member.
     * 
     * @return The superPlan.
     */
    public RepairPlan getSuperPlan() {
        return superPlan;
    }


    /**
     * B Adds an edge to the Embedding used for this repairPlan.
     * 
     * @param pos  The position of the edge.
     * @param edge  The edge. Either EdgeData, or if allready exists in GSC a
     *        ConnectedNode.
     */
    public void addEmbeddingEdge(int pos, Object edge) {
        if (edge instanceof ConnectedNode) {
            if (((ConnectedNode) edge).node == null) {
                throw new Error("Edge without node added to plan");
            }

            if (((ConnectedNode) edge).other == null) {
                throw new Error(
                        "ConnectedNode without other member added to plan. That should be an EdgeData.");
            }

            if (changedNode == ((ConnectedNode) edge).node) {
                repairEmbedding[pos] = ((ConnectedNode) edge).other;

                return;
            }
        }

        repairEmbedding[pos] = edge;
    }


    /**
     * Adds an edge to the Embedding used for this repairPlan.
     * 
     * @param pos  The position of the edge.
     * @param a  One of the two connected Nodes.
     * @param b  The other Node.
     * @param edgeID  The type of edge.
     * @param direction  If true the edge goes from b to a, else reverse.
     */
    public void addEmbeddingEdge(int pos, Node a, Node b, int edgeID, 
                                 boolean direction) {
        if ((a == null) || (b == null)) {
            throw new Error("Null Node added to repairPlan");
        }

        Node to;
        Node from;

        if (direction == Registry.INCOMING) {
            from = b;
            to   = a;
        } else {
            from = a;
            to   = b;
        }

        repairEmbedding[pos] = new EdgeData(from, to, edgeID);
    }


    /**
     * Adds a subplan to this plan.
     * 
     * @param subplan  The subplan..
     */
    public void addSubplan(RepairPlan subplan) {
        if (subplans == null) {
            subplans = new Vector();
        }

        subplans.add(subplan);

        if (superPlan != null) {
            subplan.superPlan = superPlan;
        } else {
            subplan.superPlan = this;
        }
    }


    /**
     * Adds a subplan that should be executed after the execution of this plan.
     * 
     * @param subplan The new subplan.
     */
    public void addSucceedingSubplan(RepairPlan subplan) {
        if (succeedingSubplans == null) {
            succeedingSubplans = new Vector();
        }

        succeedingSubplans.add(subplan);

        if (superPlan != null) {
            subplan.superPlan = superPlan;
        } else {
            subplan.superPlan = this;
        }
    }


    /**
     * Is the given edge contained in the edges?
     * 
     * @param edge
     * 
     * @return The tested edge.
     */
    public boolean embeddingContainsEdge(Object edge) {
        if (edge instanceof ConnectedNode) {
            if (changedNode == ((ConnectedNode) edge).node) {
                edge = ((ConnectedNode) edge).other;
            }
        }

        for (int i = 0; i < repairEmbedding.length; i++) {
            if (edge == repairEmbedding[i]) {
                return true;
            }
        }

        return false;
    }


    /**
     * For other Plans to be able to take the edges and nodes of this plan into
     * consideration  this function registeres all changes (also of subtasks!)
     * with the graphrepairer. E.g. if two
     */
    public void registerChanges() {
        //first register the subplanchanges
        registerSubplanChanges();

        if (operation == extend) {
            graphrepairer.setAllEdgesAdded(this);
        } else if (operation == create) {
            graphrepairer.setNodeAdded((Node) wildcard);
            graphrepairer.setAllEdgesAdded(this);
        } else if (operation == reduce) {
            graphrepairer.setAllEdgesRemoved(this);
        } else if (operation == delete) {
            graphrepairer.setAllEdgesRemoved(this);
            graphrepairer.setNodeRemoved(changedNode);
        }
    }


    /**
     * Takes the given subplan out of the subplans.
     * 
     * @param subplan  The plan.
     */
    public void removeSubplan(RepairPlan subplan) {
        if (subplans != null) {
            subplans.remove(subplan);
        }
    }


    /**
     * Repairs the Graph acording to the repairPlan.
     * 
     * @return If this.operation is create the newly created Node is returned,
     *         if extend the ConnectedNode else null.
     */
    public Object repair() {
        executeSubplans(subplans);

        return executeOperation(createEmbedding());
    } //repair()


    /**
     * Unregisteres all the edges and nodes changed by this RepairPlan, and
     * registered as changed via registerChanges().
     */
    public void unregisterChanges() {
        unregisterSubplanChanges();

        if (operation == extend) {
            graphrepairer.resetAllEdgesAdded(this);
        } else if (operation == create) {
            graphrepairer.resetNodeAdded((Node) wildcard);
        } else if (operation == reduce) {
            graphrepairer.resetAllEdgesRemoved(this);
        } else if (operation == delete) {
            graphrepairer.resetAllEdgesRemoved(this);
            graphrepairer.resetNodeRemoved((Node) wildcard);
        }
    }


    /**
     * Either gets the edge from the stored allready created edges, or creates
     * it and enters it there.
     * 
     * @param edge  The searched edge.
     * @param viewedFrom  The Node that determines the viewpoint on the
     *        ConnectedNode.
     * 
     * @return An ConnectedNode with the same data as edge, seen as from the
     *         node of this plan.
     */
    protected ConnectedNode getEdge(Object edge, Node viewedFrom) {
        ConnectedNode cn;

        if (superPlan != null) {
            cn = superPlan.getEdge(edge, viewedFrom);
        } else {
            cn = (ConnectedNode) subplanResults.get(edge);

            if (cn == null) {
                boolean direction;
                Node    otherNode;
                int     edgeID;

                if (edge instanceof EdgeData) {
                    direction = ((EdgeData) edge).getDirection(viewedFrom);
                    otherNode = getRealNode(((EdgeData) edge).getOtherNode(
                                                    viewedFrom));
                    edgeID    = ((EdgeData) edge).edgeID;
                } else {
                    direction = ((ConnectedNode) edge).direction;
                    otherNode = getRealNode(((ConnectedNode) edge).node);
                    edgeID    = ((ConnectedNode) edge).edgeID;
                }

                cn = new ConnectedNode(otherNode, otherNode.typeID, edgeID, 
                                       direction);
            } else {
                if (cn.node == viewedFrom) {
                    cn = cn.other;
                }
            }
        }

        return cn;
    }


    /**
     * Returns the node that is entered for the given key in the
     * superPlan.subplanResults or null if none.
     * 
     * @param key A placeholder node, used as the key to the real node.
     * 
     * @return A CSP Graph Node
     */
    protected Node getNode(Node key) {
        if (superPlan == null) {
            Node result = (Node) subplanResults.get(key);

            if (result == null) { // the key was allready the searched node.
                result = key;
            }

            return result;
        } else {
            return superPlan.getNode(key);
        }
    }


    /**
     * Returns the node of the edge that is not changedNode. If the node was a
     * placeholder for a real node that one is returned.
     * 
     * @param n Either a placeholder or a real node.
     * 
     * @return If n is a real node n, else getNode(n)
     */
    protected Node getRealNode(Node n) {
        Node realNode = getNode(n);

        if (realNode == null) {
            return n; // other is allready the realnode..
        } else {
            return realNode;
        }
    }


    /**
     * Creates the embedding of the operation.
     * 
     * @return A valid embedding, or null if none needed for the operation.
     */
    protected Embedding createEmbedding() {
        if (repairEmbedding == null) {
            return null;
        } else {
            Embedding result = new Embedding(repairEmbedding.length);

            for (int i = 0; i < repairEmbedding.length; i++) {
                Object        o                = repairEmbedding[i];
                ConnectedNode newConnectedNode = getEdge(o, changedNode);
                result.addConnectedNode(newConnectedNode);
                subplanResultsAddNewEdge(o, newConnectedNode);
            } //for

            return result;
        }
    } //createEmbedding()


    /**
     * Executes the operation of this plan. If the plan was a node creation the
     * new node is returned.
     * 
     * @param embedding  The embedding needed for the execution. null if none
     *        needed.
     * 
     * @return A new node in the CSP-Graph or null.
     */
    protected Object executeOperation(Embedding embedding) {
        Object result = null;

        if (operation == extend) {
            changedNode = getNode(changedNode);
            gsc.extendConstraintUseEdgeObjects_internal(
                    (ConventionalConstraint) changedNode, embedding);
        } else if (operation == create) {
            Node newNode;

            if (Registry.isVariable(constraintID)) {
                newNode = gsc.createVariable_internal("var" + variableNr++, 
                                                      gsc.random.nextInt(1000), 
                                                      gsc.random.nextInt(800));
                gsc.setVariable_internal((Variable) newNode, 
                                         gsc.random.nextDouble() + 1);
            } else {
                newNode = gsc.createConstraintUseEdgeObjects_internal(
                                  constraintID, embedding, 
                                  gsc.random.nextInt(1000), 
                                  gsc.random.nextInt(800));
            }

            subplanResultsAddNewNode(wildcard, newNode);
            newNode.setAuxiliaryNodeData(auxiliaryNodeData);
            result = newNode;
        } else if (operation == reduce) {
            changedNode = getNode(changedNode);
            gsc.reduceConstraint_internal((ConventionalConstraint) changedNode, 
                                          embedding);
        } else if (operation == delete) {
            if (gsc.containsNode(changedNode)) {

                if (changedNode instanceof ConventionalConstraint) {
                    gsc.deleteConstraint_internal(
                            (ConventionalConstraint) changedNode);
                } else {
                    gsc.deleteVariable_internal((Variable) changedNode);
                }
            }
        } else if (operation == nothing) {
            result = wildcard;
        }

        executeSubplans(succeedingSubplans);

        return result;
    }


    /**
     * Executes all subplans in given collection.
     * 
     * @param executedSubplans The subplans that should be executed.
     */
    protected void executeSubplans(Vector executedSubplans) {
        if (executedSubplans != null) {
            Iterator it = executedSubplans.iterator();

            while (it.hasNext()) {
                RepairPlan plan = (RepairPlan) it.next();
                plan.repair();
            } //while  
        }
    } //executesubplans()


    /**
     * Calls registerChanges for all Subplans.
     */
    protected void registerSubplanChanges() {
        if (subplans != null) {
            Iterator it = subplans.iterator();

            while (it.hasNext()) {
                RepairPlan plan = (RepairPlan) it.next();
                plan.registerChanges();
            } //while  
        }
    }


    /**
     * Puts a new edge in the subplanResults map.
     * 
     * @param key  The key the node is stored at.
     * @param newEdge  The node to be added.
     */
    protected void subplanResultsAddNewEdge(Object key, ConnectedNode newEdge) {
        if (superPlan == null) {
            subplanResults.put(key, newEdge);
        } else {
            superPlan.subplanResultsAddNewEdge(key, newEdge);
        }
    }


    /**
     * Puts a new Node in the subplanResults map.
     * 
     * @param key  The key the node is stored at.
     * @param node  The node to be added.
     */
    protected void subplanResultsAddNewNode(Object key, Node node) {
        if (superPlan == null) {
            subplanResults.put(key, node);
        } else {
            superPlan.subplanResultsAddNewNode(key, node);
        }
    }


    /**
     * Calls unregisterChanges for all Subplans.
     */
    protected void unregisterSubplanChanges() {
        if (subplans != null) {
            int i = subplans.size();

            //go in reverse order
            while (i > 0) {
                i--;

                RepairPlan plan = (RepairPlan) subplans.elementAt(i);
                plan.unregisterChanges();
            } //while  
        }
    }

    // Classes ****************************************************************

    /**
     * Represents an edge
     */
    public static class EdgeData {
        public Node from;
        public Node to;
        public int  edgeID;

        public EdgeData(Node _from, Node _to, int _edgeID) {
            from   = _from;
            to     = _to;
            edgeID = _edgeID;
        }


        public EdgeData(ConnectedNode cn, Node other) {
            if (cn.direction == Registry.INCOMING) {
                from = cn.node;
                to   = other;
            } else {
                from = other;
                to   = cn.node;
            }

            edgeID = cn.edgeID;
        }


        public EdgeData(ConnectedNode cn, Node cnNode, Node other) {
            if (cn.direction == Registry.INCOMING) {
                from = cnNode;
                to   = other;
            } else {
                from = other;
                to   = cnNode;
            }

            edgeID = cn.edgeID;

            if (cn.nodeTypeID != cnNode.typeID) {
                throw new Error("Creation Data inconsistent.");
            }
        }

        /**
         * Returns the direction as seen from a
         * 
         * @param a The node.
         * 
         * @return a == to.
         */
        public boolean getDirection(Node a) {
            return a == to;
        }


        /**
         * Each edge contains two nodes, from and to. This method returns the
         * node that is not a.
         * 
         * @param a The node.
         */
        public Node getOtherNode(Node a) {
            return (a == from) ? to : from;
        }


        /**
         * Does the given edge, either a ConnectedNode or an EdgeData have the
         * same type as this edge?
         * 
         * @param o Either a ConnectedNode or an EdgeData.
         * 
         * @return a represents the same edge as this.
         */
        public boolean isSameType(Object o) {
            if (o instanceof ConnectedNode) {
                ConnectedNode cn = (ConnectedNode) o;

                if (edgeID == cn.edgeID) {
                    return ((cn.direction == Registry.INCOMING) && 
                           (cn.nodeTypeID == from.typeID)) || 
                           ((cn.direction == Registry.OUTGOING) && 
                           (cn.nodeTypeID == to.typeID));
                }
            } else {
                EdgeData ed = (EdgeData) o;

                if (edgeID == ed.edgeID) {
                    return (ed.from.typeID == from.typeID) && 
                           (ed.to.typeID == to.typeID);
                }
            }

            return false;
        }
    }
}