/*
 * TestGraphRepairer.java
 *
 * Created on 25. Januar 2003, 12:51
 */
package com.ai_center.dragonbreath.graphRepair;

import com.ai_center.dragonbreath.*;
import com.ai_center.dragonbreath.Constraint;
import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.costs.CostFunctionReference;
import com.ai_center.dragonbreath.structuralConstraintHandling.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.applicationCondition.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.dockingGraph.SCIDockingGraph;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.movableSubgraph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.testGraph.*;
import com.ai_center.dragonbreath.utilities.*;

import java.util.*;

import org.jdom.Element;


/**
 * Provides methods to repair or destroy StructuralConstraintGraphs
 */
public class SCGraphRepairer implements NodeChooseInterface {
    // Static variables/initializers ******************************************

    /**
     * A SCNode that is used in the nodesWaitingForInterpretation stack, to
     * mark  that there will come an interpretation that is not needed.
     */
    static Object NULL_SCNode = new Object();

    // Instance variables *****************************************************

    /**
     * Each time this class is asked for an interpretation via
     * getFittingNodeForEdgeCreation()  or getFittingNode() but doesnt have an
     * interpretation at hand, it returns null, and enters the node A that
     * would need the questioned interpretation in the stack. When the
     * graphRepairer has decided for an other node, or a new one it calls
     * registerNewNode(). A is poped from the stack and gets the new
     * interpretation. If no node is interested in the interpretation that was
     * requested via getFittingNodeForEdgeCreation() a NULL_SCNode was pushed
     * on the stack as a marker.
     */
    protected Stack nodesWaitingForInterpretation = new Stack();

    /**
     * An edgeInterpretation that is used to mark edges that got repaired as
     * interpreted, until the real interpretation is available.
     */
    ConnectedNode edgeInterpretedMarker;

    /** 
     * Used for debug output. 
     */
    DebugLogger dl = new DebugLogger(true);

    /** 
     * The GlobalSearchControl that holds the CSP. 
     */
    GlobalSearchControl myGSC;

    /** 
     * Class that is used to change the CSP-Graph 
     */
    GraphRepairer myGraphRepairer;

    /**
     * Where the GraphRepairer should ask for nodes that were not called to be
     * changed, but have to be for the demanded change to work.
     */
    NodeChooseInterface myNodeChooser;

    /** 
     * The edge that is actually beeing repaired. 
     */
    SCEdge actualCreatedEdge;

    /**
     * The node that should be created. When the graphRepairer calls this class
     * to know which  nodes to use for repair, the created node has to be
     * known.
     */
    SCNode actualCreatedNode;

    /** 
     * The Graph currently beeing repaired. 
     */
    StructuralConstraintGraph repairedGraph;

    /** 
     * The Part of the repairedGraph that is actually beeing repaired. 
     */
    StructuralConstraintGraphPart repairedPart;

    /** 
     * .. 
     */
    StructuralConstraintManagement mySCM;

    /** 
     * When did the repair start? 
     */
    long startRepairTime;

    // Constructors ***********************************************************

    /**
     * Creates a new instance of TestGraphRepairer
     * 
     * @param scm  The StructuralConstraintManagement to be used.
     * @param gsc The GlobalSearchControl to be used.
     */
    public SCGraphRepairer(StructuralConstraintManagement scm, 
                           GlobalSearchControl gsc) {
        mySCM                             = scm;
        myGSC                             = gsc;
        myGraphRepairer                   = new GraphRepairer(gsc, 
                                                              myNodeChooser);
        edgeInterpretedMarker             = new ConnectedNode(0, 
                                                              Registry.ANY_EDGE, 
                                                              true);
        edgeInterpretedMarker.other       = new ConnectedNode(0, 
                                                              Registry.ANY_EDGE, 
                                                              false);
        edgeInterpretedMarker.other.other = edgeInterpretedMarker;

        //dl.switchOn(1);// uncomment for debug output
    }

    // Methods ****************************************************************

    /**
     * When a Node was created the GraphRepairer asks via
     * getAuxiliaryNodeData() what values to set as default.
     * 
     * @param node The new node.
     * 
     * @return A jdom.Element with the relevant data.
     */
    public Element getAuxiliaryNodeData(Node node) {
        SCNode nodeWithInterpretation = null;

        if (repairedPart != null) {
            Map interpretedNodes = repairedPart.getInterpretedNodes();
            nodeWithInterpretation = (SCNode) interpretedNodes.get(node);
        }

        if (nodeWithInterpretation != null) {
            return nodeWithInterpretation.getAuxiliaryNodeData();
        } else {
            return null;
        }
    }


    /**
     * Returns either an allready existing Node, or null, if the GraphRepairer
     * should create its own new one. The node should contain all the edges in
     * edges. Allways returns null. For a single node without an edge
     * connected the SCGraphRepairer doesnt know which node to use.
     * 
     * @param nodeTypeID The id of the node, as in Registry.
     * 
     * @return null
     */
    public Node getFittingNode(int nodeTypeID) {
        return null;
    }


    /**
     * Is there a specific node that should be used to complete the given edge?
     * Called by the actualNodeChooser.
     * 
     * @param node The node that should contain the edge.
     * @param edge The definition of the produced edge.
     * 
     * @return A node for the other end of the edge, or null.
     */
    public Node getFittingNodeForEdgeCreation(Node node, ConnectedNode edge) {
        if (repairedPart == null) {
            return null;
        }

        Map    interpretedNodes       = repairedPart.getInterpretedNodes();
        SCNode nodeWithInterpretation = (SCNode) interpretedNodes.get(node);

        if (nodeWithInterpretation != null) {
            SCEdge createdEdge = getUninterpretedEdgeOfSameType(
                                         nodeWithInterpretation, edge);

            if (createdEdge == null) {
                // the next returned nodeInterpretation doesn't matter for the testGraph
                enterNodeWaitingForInterpretation(NULL_SCNode);

                return null;
            } else {
                // mark edge as interpreted
                createdEdge.setInterpretation(edgeInterpretedMarker);

                SCNode resultNode     = createdEdge.getOtherNode(
                                                nodeWithInterpretation);
                Node   interpretation = resultNode.peekInterpretation();

                if (interpretation == null) {
                    if (resultNode != actualCreatedNode) { // else the node is allready on the stack

                        // the next returned nodeInterpretation is for the resultNode
                        enterNodeWaitingForInterpretation(resultNode);
                    } else {
                        if (nodesWaitingForInterpretation.peek() != resultNode) {
                            throw new Error("Wrong order on stack.");
                        }
                    }
                }

                return interpretation;
            }
        } else if (actualCreatedEdge != null) {
            if (actualCreatedEdge.isSameType(edge)) {
                if (edge.direction != actualCreatedEdge.getDirection(
                                              actualCreatedNode)) {
                    // the edge had to be created via a basegraph,  so it is now pointing in the opposite direction
                    actualCreatedEdge = null; // the actualCreatedEdge has its intepretation.

                    Node interpretation = actualCreatedNode.peekInterpretation();

                    return interpretation;
                }
            }
        }


        // the next returned nodeInterpretation doesn't matter for the testGraph
        enterNodeWaitingForInterpretation(NULL_SCNode);

        return null;
    }


    /**
     * Inherited from NodeChooseInterface. Here is nothing to do as this is the
     * mainNodesSource.
     * 
     * @param nodeSource
     */
    public void setMainNodeSource(NodeChooseInterface nodeSource) {
        //nothing to do, this is the mainNodesSource.
    }


    /**
     * Sets the used NodeChooser.
     * 
     * @param newNodeChooser The nodeChooser that should be used.
     */
    public void setNodeChooser(NodeChooseInterface newNodeChooser) {
        myNodeChooser = newNodeChooser;
        myGraphRepairer.setNodeChooser(newNodeChooser);
    }


    /**
     * When a node is chosen to be used it can be tested for compliance with
     * the nodeChooser.  Especially used to not use nodes for repair of a
     * subgraph, that were allready assigned to an other subgraph. Allways
     * called by a nodechooser from its mainNodeSource.
     * 
     * @param n The tested node.
     * 
     * @return ! repairedGraph.mainPartsContainInterpretation(n)
     */
    public boolean isNodeOk(Node n) {
        return ! repairedGraph.mainPartsContainInterpretation(n);
    }


    /**
     * Resets the interpretation of the node that has currently the given
     * interpretation and pushes it on the nodesWaitingForInterpretation
     * stack.
     * 
     * @param node The node that was replaced.
     */
    public void setNodeReplaced(Node node) {
        SCNode nodeWithInterpretation = null;

        if (repairedPart != null) {
            Map interpretedNodes = repairedPart.getInterpretedNodes();
            nodeWithInterpretation = (SCNode) interpretedNodes.get(node);
        }

        if (nodeWithInterpretation != null) {
            nodeWithInterpretation.resetInterpretationWithEdges();
            enterNodeWaitingForInterpretation(nodeWithInterpretation);
        } else {
            enterNodeWaitingForInterpretation(NULL_SCNode);
        }
    }


    /**
     * Adds all the edges given in the Vector to the given node. Maybe more
     * then the given edges are added.
     * 
     * @param node  The node that gets the given edges.
     * @param edge  The edge that should be added. If a ConnectedNode here
     *        allready holds a node that is taken.
     */
    public void addEdge(Node node, ConnectedNode edge) {
        RepairPlan plan = myGraphRepairer.addEdge(node, edge);
        plan.repair();
    }


    /**
     * Creates a node of the given type.
     * 
     * @param nodeTypeID The type of the new Node.
     * 
     * @return The node.
     */
    public Node addNode(int nodeTypeID) {
        RepairPlan plan = myGraphRepairer.addNode(nodeTypeID);

        return (Node) plan.repair();
    }


    /**
     * Tries to repair the given subgraph. That can be either the
     * TestGraphMainParts  from repair() or a dockingGraph NAC from
     * destroyDockingGraph()
     * 
     * @param subgraph Subgraph to be repaired.
     * @param millisForRepair Time in milliseconds a repair may take.
     */
    public void completeSubgraph(MovableSubgraph subgraph, long millisForRepair) {
        Vector parts = subgraph.getAllParts();
        myNodeChooser.setMainNodeSource(this);
        startTimer(millisForRepair);

        Iterator it = parts.iterator();

        while (it.hasNext()) {
            MovableSubgraphPart part = (MovableSubgraphPart) it.next();

            if (! part.isInterpreted()) {
                repairPart(part);

                if (timeUp()) {
                    return;
                } // if timeup        
            } //if ! part.isInterpreted()
        }
    } //completeSubgraph()


    /**
     * Removes one Edge or Node of the part. After call all interpretations of
     * the part are reset.
     * 
     * @param part The part that should be destroyed.
     * @param removeNodeProbability The probabillity that a node and not an
     *        edge of the part should be removed.
     */
    public void destroyGraphPart(StructuralConstraintGraphPart part, 
                                 double removeNodeProbability) {
        myNodeChooser.setMainNodeSource(this);
        repairedGraph = part.mySCGraph;

        boolean elementRemoved = false;

        if (myGSC.random.nextDouble() > removeNodeProbability) {
            elementRemoved = removeOneNode(part, false);
        }

        if (! elementRemoved) {
            removeOneEdge(part);
        }
    }


    /**
     * Called to indicate which new node was used for the last requested node
     * (via getFittingNodeForEdgeCreation() that was returned as null. That
     * node is then used as interpretation of the first node in
     * nodesWaitingForInterpretation.
     * 
     * @param interpretation  The new interpretation
     */
    public void registerNewNode(Node interpretation) {
        SCNode node = getNodeWaitingForInterpretation();

        if ((node != null) && 
                ! node.setInterpretationForRepair(interpretation)) {
            throw new Error("Interpretation used twice.");
        }
    }


    /**
     * Removes the given node, no matter if it is a Variable or a Constraint.
     * And also if the Constraint is not yet reduced to its Basegraph, or the
     * Variable still connected to Constraints.
     * 
     * @param edge  The edge that should be removed.
     * 
     * @return The edge could be removed.
     */
    public boolean removeEdge(ConnectedNode edge) {
        RepairPlan plan = myGraphRepairer.removeEdge(edge);

        if (plan != null) {
            plan.repair();

            return true;
        } else {
            return false;
        }
    }


    /**
     * Removes the given node, no matter if it is a Variable or a Constraint.
     * And also if the Constraint is not yet reduced to its Basegraph, or the
     * Variable still connected to Constraints.
     * 
     * @param node The node that should be deleted.
     */
    public void removeNode(Node node) {
        RepairPlan plan = myGraphRepairer.removeNode(node);
        plan.repair();
    }


    /**
     * Returns whether the node should be removed, or the edge added. Called if
     * an edge should be removed, but can not due to the fact that it is part
     * of the basegraph of the node. In such a case the addition of a new edge
     * of the same type could help, as it could then be used for the
     * basegraph.
     * 
     * @param node  The node.
     * @param edge  The edge that should be removed, but can not.
     * 
     * @return true if the node should be removed, false if an edge should be
     *         added.
     */
    public boolean removeNodeOrElseAddEdge(Node node, Object edge) {
        return ! repairedGraph.containsInterpretation(node);
    }


    /**
     * Sets the actualCreatedEdge member for the same reason as
     * setActualCreatedNode. Also calls edge.setInterpretedForRepair()
     * 
     * @param edge  After this call it is set as interpreted, although it has
     *        no real interpretation attached.
     */
    protected void setActualCreatedEdge(SCEdge edge) {
        actualCreatedEdge = edge;
    }


    /**
     * Iterates through the edges of the given node, and creates for each one
     * that doesn't have an Interpretation a ConnectedNode of the same type.
     * First edges are returned whose other node allready has an
     * interpretation
     * 
     * @param node A node of the repaired graph part.
     * 
     * @return An edge that doesnt yet have an interpretation.
     */
    protected ConnectedNode getMissingEdge(SCNode node) {
        ConnectedNode missingEdge;
        Vector        edgesWithoutInterpretedNodes = new Vector();
        Iterator      it                           = node.getEdges().iterator();

        while (it.hasNext()) {
            SCEdge edge = (SCEdge) it.next();

            if (! edge.hasInterpretation()) {
                SCNode otherNode = edge.getOtherNode(node);

                if (otherNode.hasInterpretation()) {
                    int     nodeTypeID = otherNode.typeID;
                    int     edgeID    = edge.getEdgeID();
                    boolean direction = edge.getDirection(node);
                    missingEdge = new ConnectedNode(
                                          otherNode.getInterpretation(), 
                                          nodeTypeID, edgeID, direction);
                    setActualCreatedEdge(edge);

                    return missingEdge;
                } else {
                    edgesWithoutInterpretedNodes.add(edge);
                }
            }
        } //while


        // there was no edge with an interpreted other node, so use one without.
        it = edgesWithoutInterpretedNodes.iterator();

        while (it.hasNext()) {
            SCEdge  edge       = (SCEdge) it.next();
            SCNode  otherNode  = edge.getOtherNode(node);
            int     nodeTypeID = otherNode.typeID;
            int     edgeID     = edge.getEdgeID();
            boolean direction  = edge.getDirection(node);
            missingEdge = new ConnectedNode(nodeTypeID, edgeID, direction);
            setActualCreatedNode(otherNode);
            setActualCreatedEdge(edge);

            return missingEdge;
        } //while

        return null;
    }


    /**
     * Returns one edge of the given node that has the same type as the
     * ConnectedNode, but has no interpretation yet.
     * 
     * @param node The tested Node
     * @param edge  The searched edge.
     * 
     * @return Either an edge of node with the same type, but without
     *         interpretation, or null, if none found.
     */
    protected SCEdge getUninterpretedEdgeOfSameType(SCNode node, 
                                                    ConnectedNode edge) {
        SCEdge   scEdge;
        Iterator it = node.getEdges().iterator();

        while (it.hasNext()) {
            scEdge = (SCEdge) it.next();

            if (! scEdge.hasInterpretation() && scEdge.isSameType(edge)) {
                return scEdge;
            }
        } //while

        return null;
    }


    /**
     * Tries to assign an interpretation to the given SCNode and its edges.
     * 
     * @param node  The interpreted SCNode the edge is connected to.
     * @param missingEdge The edge that should be added.
     */
    protected void findInterpretationForEdge(SCNode node, 
                                             ConnectedNode missingEdge) {
        Node interpretation = node.peekInterpretation();
        addEdge(interpretation, missingEdge);
    } //findInterpretationForNode()


    /**
     * Removes the given edges interpretation
     * 
     * @param edge The edge with an interpretation.
     * 
     * @return true if the edge was removed, if false the edge was part of the
     *         basegraph of one of the two nodes.
     */
    protected boolean removeEdgeInterpretation(ConnectedNode edge) {
        RepairPlan plan = myGraphRepairer.removeEdge(edge);

        if (plan == null) {
            return false;
        }

        plan.repair();

        return true;
    }


    /**
     * Tries to remove one edge of the given part. If none can be removed
     * without removing a node a node is removed. After call all
     * interpretations of part are reset.
     * 
     * @param part The part that should get one less edge
     */
    protected void removeOneEdge(StructuralConstraintGraphPart part) {
        boolean removed = false;
        Vector  edges = part.getEdges();

        if (edges.size() > 0) {
            int    randomEdgeIndex = myGSC.random.nextInt(edges.size());
            SCEdge edge = (SCEdge) edges.elementAt(randomEdgeIndex);
            Node   from = edge.getFrom().peekInterpretation();

            if (removeEdge(edge.getConnectedNodePointingFrom(from))) {
                removed = true;
            } else {
                // the random edge could not be removed
                //try all the edges, until the first one is removed, or none could be removed
                Iterator it = edges.iterator();

                while (it.hasNext() && ! removed) {
                    edge    = (SCEdge) it.next();
                    from    = edge.getFrom().peekInterpretation();
                    removed = removeEdge(edge.getConnectedNodePointingFrom(from));
                } //while
            }
        }

        if (! removed) { //no edge could be removed
            removeOneNode(part, true);
        } else {
            part.resetInterpretations();
        }
    }


    /**
     * Tries to remove a Node of the given part.
     * 
     * @param part The part that should have one interpretation less
     * @param forceRemove If all Nodes are linked, should an interpretation be
     *        removed anyway? when removed == true all interpretations of part
     *        are reset
     * 
     * @return forceRemove || (one node removed)
     */
    protected boolean removeOneNode(StructuralConstraintGraphPart part, 
                                    boolean forceRemove) {
        boolean removed         = false;
        Vector  nodes           = part.getNodes();
        int     randomNodeIndex = myGSC.random.nextInt(nodes.size());
        SCNode  node            = (SCNode) nodes.elementAt(randomNodeIndex);

        if (! node.hasLink()) {
            removeNode(node.peekInterpretation());
            removed = true;
        } else {
            // the random node could not be removed
            //try all the nodes, until the first one is removed, or none could be removed
            Iterator it = nodes.iterator();

            while (it.hasNext() && ! removed) {
                node = (SCNode) it.next();

                if (! node.hasLink()) {
                    removeNode(node.peekInterpretation());
                    removed = true;
                }
            } //while
        }

        if (! removed) {
            if (forceRemove) { //no node could be removed
                node = (SCNode) nodes.elementAt(randomNodeIndex);

                SCNode linkedNode = node.getLinkedNode();

                if (linkedNode != null) {
                    repairedPart = linkedNode.myGraphPart;
                    linkedNode.resetInterpretation();
                    linkedNode.resetInterpretationExtern();

                    //enterNodeWaitingForInterpretation(linkedNode);
                    // the removed node is also a node of this part. so let it decide 
                    // how to handle any additional removal of nodes that might be necessary due to this remove.
                }

                removeNode(node.peekInterpretation());
                part.resetInterpretations();

                if (linkedNode != null) {
                    linkedNode.myGraphPart.resetInterpretations();
                }

                removed = true;
            }
        } else {
            part.resetInterpretations();
        }

        return removed;
    }


    /**
     * Starts the timing, called at the beginning of repair.
     * 
     * @param millisForRepair The time a repair may take.
     */
    protected void startTimer(long millisForRepair) {
        startRepairTime = java.lang.System.currentTimeMillis() + 
                          millisForRepair;
    }


    /**
     * True if since startTimer more then
     * actualconstraintinstance.millisForRepair have passed.
     * 
     * @return true if time is up
     */
    protected boolean timeUp() {
        return java.lang.System.currentTimeMillis() > startRepairTime;
    }


    /**
     * Sets the actualCreatedNode member that is used to determine in
     * getFittingNodeForEdgeCreation which node is actually extended.
     * 
     * @param node The created node.
     */
    private void setActualCreatedNode(SCNode node) {
        enterNodeWaitingForInterpretation(node);
        actualCreatedNode = node;
    }


    /**
     * Returns the last node pushed on the nodesWaitingForInterpretation stack
     * or null if it is a NULL_SCNode.
     * 
     * @return The SCNode that gets the next created interpretation.
     */
    private SCNode getNodeWaitingForInterpretation() {
        dl.print(1, "\nnodesWaitingForInterpretation.pop ");

        if (! nodesWaitingForInterpretation.empty()) {
            Object o = nodesWaitingForInterpretation.pop();

            if (o != NULL_SCNode) {
                SCNode node = (SCNode) o;
                dl.print(1, 
                         Registry.nodeType[((SCNode) node).typeID].getTypeName() + 
                         "\n");

                return node;
            } else {
                dl.print(1, " null \n");
            }
        } else {
            dl.print(1, " null \n");
        }

        return null;
    }


    /**
     * Returns a SCNode that is used as seed for the repair. If no node is
     * interpreted the seed of the subgraph is taken and an interpretation
     * created and assigned. Else the interpreted node is returned, that has
     * least possible assignments for its edges.
     * 
     * @return A node of the repaired GraphPart that has an interpretation.
     */
    private SCNode getSearchSeed() {
        SCNode actualNode     = repairedPart.getSeed();
        Node   interpretation = actualNode.peekInterpretation(); // the interpretation is provided via a link

        if (interpretation != null) {
            actualNode.setInterpretationBiggestEmbedding(interpretation);

            return actualNode;
        }

        Collection interpretedNodes = repairedPart.getInterpretedNodes()
                                                  .values();
        actualNode = null;

        ConnectedNode missingEdge = null;

        if (interpretedNodes.isEmpty()) { //not one single node of the part has an interpretation.
            actualNode = repairedPart.getSeed();
            setActualCreatedNode(actualNode);
            interpretation = addNode(actualNode.typeID);
            actualNode.setInterpretationForRepair(interpretation);

            return actualNode;
        } else {
            SCNode   winner               = null;
            int      numAssignmentsWinner = java.lang.Integer.MAX_VALUE;
            Iterator it                   = interpretedNodes.iterator();

            while (it.hasNext()) {
                actualNode = (SCNode) it.next();

                int numAssignments = actualNode.numberOfConAssignments();

                if (numAssignments > numAssignmentsWinner) {
                    numAssignmentsWinner = numAssignments;
                    winner               = actualNode;
                }
            } //while

            return winner;
        }
    }


    /**
     * Sets the interpretation for one node and all its neighborhood. One (any)
     * node with an interpretation is returned. Adds only one edge, as whole
     * new subgraphs may get connected by the edge,  and without a
     * findBiggestEmbedding() the edges and nodes of this graph are not
     * interpreted correctly.
     */
    private void addOneEdgeToPart() {
        nodesWaitingForInterpretation.clear();
        dl.print(1, "\nnodesWaitingForInterpretation.clear \n");

        Collection    interpretedNodes = repairedPart.getInterpretedNodes()
                                                     .values();
        SCNode        actualNode  = null;
        ConnectedNode missingEdge = null;
        Iterator      it          = interpretedNodes.iterator();

        while ((missingEdge == null) && (it.hasNext())) {
            actualNode  = (SCNode) it.next();
            missingEdge = getMissingEdge(actualNode);

            if (timeUp()) {
                return;
            }
        }

        if (missingEdge != null) {
            findInterpretationForEdge(actualNode, missingEdge);
        }
    }


    /**
     * Pushes the given node to the nodesWaitingForInterpretation stack.
     * 
     * @param node  Either one SCNode of the graphpart, or NULL_SCNode
     */
    private void enterNodeWaitingForInterpretation(Object node) {
        nodesWaitingForInterpretation.push(node);
        dl.print(1, "\nnodesWaitingForInterpretation.push ");

        if (node == NULL_SCNode) {
            dl.print(1, "null \n");
        } else {
            dl.print(1, 
                     Registry.nodeType[((SCNode) node).typeID].getTypeName() + 
                     "\n");
        }
    }


    /**
     * Repairs a MovableSubgraphPart.
     * 
     * @param part  The part that should be repaired.
     * 
     * @return true if part could be interpreted
     */
    private boolean repairPart(StructuralConstraintGraphPart part) {
        repairedGraph = part.mySCGraph;
        repairedPart  = part;

        SCNode seed   = null;
        Node   seedInterpretation = null;

        while (! repairedPart.isInterpreted()) {
            if (seed == null) {
                seed               = getSearchSeed();
                seedInterpretation = seed.peekInterpretation();
            } else {
                addOneEdgeToPart();
            }

            if (timeUp()) {
                repairedPart.resetInterpretations();

                return false;
            }

            repairedPart.resetInterpretations();
            seed.setInterpretationBiggestEmbedding(seedInterpretation);

            if (repairedPart.findBiggestEmbedding(seed)) {
                if (repairedPart instanceof UnconnectedMovableSubgraphPart) {
                    ((UnconnectedMovableSubgraphPart) part).registerSeedInterpretation();
                }

                return true;
            }
        }

        return true;
    } //repairPart()
}