/*
 * TestGraphRepairer.java
 *
 * Created on 12. Februar 2003, 15:20
 */
package com.ai_center.dragonbreath.graphRepair;

import com.ai_center.dragonbreath.*;
import com.ai_center.dragonbreath.Constraint;
import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.costs.CostFunctionReference;
import com.ai_center.dragonbreath.structuralConstraintHandling.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.applicationCondition.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.dockingGraph.SCIDockingGraph;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.movableSubgraph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.testGraph.*;
import com.ai_center.dragonbreath.utilities.*;

import java.util.*;


/**
 * This class is used to repair a TestGraph of a StructuralConstraintInstance.
 * Missing MainParts are created and NACs destroyed.
 * 
 * @version $Revision: 1.4 $^
 * @author tim baermann
 */
public class TestGraphRepairer {
    // Instance variables *****************************************************

    /** 
     * Used for debug output. 
     */
    DebugLogger dl = new DebugLogger(true);

    /** 
     * The GlobalSearchControl that controls the CSP. 
     */
    GlobalSearchControl myGSC;

    /** 
     * The Part of the repairedTestGraph that is actually beeing repaired. 
     */
    MovableSubgraphPart repairedPart;

    /**
     * Where the GraphRepairer should ask for nodes that were not called to be
     * changed, but have to be for the demanded change to work.
     */
    NodeChooseInterface myNodeChooser;

    /** 
     * Used to change the graphs 
     */
    SCGraphRepairer mySCGraphRepairer;

    /** 
     * The instance the repairedTestGraph is part of. 
     */
    StructuralConstraintInstance repairedInstance;

    /** 
     * The StructuralConstraintManagement used. 
     */
    StructuralConstraintManagement mySCM;

    /** 
     * The testGraph currently beeing repaired. 
     */
    TestGraph repairedTestGraph;

    /** 
     * When did the repair start? 
     */
    long startRepairTime;

    // Methods ****************************************************************

    /**
     * Creates a new instance of TestGraphRepairer
     * 
     * @param scm  The StructuralConstraintManagement to be used.!
     * @param gsc The GlobalSearchControl to be used.
     * @param scGraphRepairer
     */
    public void setConstructionParameters(StructuralConstraintManagement scm, 
                                          GlobalSearchControl gsc, 
                                          SCGraphRepairer scGraphRepairer) {
        mySCM             = scm;
        myGSC             = gsc;
        mySCGraphRepairer = scGraphRepairer;
    }


    /**
     * Adds missing Nodes/Edges, or removes NACs from the given testGraph. If
     * after a certain period of time no solution is found it returns. The
     * structural error remains.
     * 
     * @param repairedTestGraph  The testgraph that should be repaired.
     * @param instance The StructuralConstraintInstance the testGraph belongs
     *        to.
     */
    public void repair(TestGraph repairedTestGraph, 
                       StructuralConstraintInstance instance) {
        repairedInstance = instance;
        startTimer();

        if (repairedTestGraph.getNACError() == 0) {
            mySCGraphRepairer.completeSubgraph(repairedTestGraph.getMainParts(), 
                                               repairedInstance.millisForRepair);
            repairedTestGraph.resetInterpretations();

            if (repairedTestGraph.testTestGraph()) {
                instance.createNewTestGraphInstance(repairedTestGraph);
            } // else a nac might have been created, repair in next iteration
        } else {
            destroyNACs(repairedTestGraph);

            if (repairedTestGraph.findOneEmbedding()) {
                instance.createNewTestGraphInstance(repairedTestGraph);
            }
        }
    }


    /**
     * Destroys the given NAC
     * 
     * @param nac The NAC to be destroyed.
     */
    protected void destroyNAC(ApplicationCondition nac) {
        while (true) {
            int                           numParts = nac.allParts.size();
            int                           partNr = myGSC.random.nextInt(
                                                           numParts);
            StructuralConstraintGraphPart part   = (StructuralConstraintGraphPart) nac.allParts.elementAt(
                                                           partNr);
            mySCGraphRepairer.destroyGraphPart(part, 
                                               repairedInstance.removeNodeProbability);

            if (timeUp() || ! nac.findAllParts()) {
                return;
            } // if timeup    
        }
    }


    /**
     * Destroys all not found NACs of the repairedTestGraph.
     * 
     * @param repairedTestGraph The testGraph that contains the NACs.
     */
    protected void destroyNACs(TestGraph repairedTestGraph) {
        Iterator it = repairedTestGraph.NACs.iterator();

        while (it.hasNext()) {
            ApplicationCondition nac = (ApplicationCondition) it.next();

            if (nac.getMinMatchError() > 0) {
                destroyNAC(nac);
            }

            if (timeUp()) {
                return;
            } // if timeup    
        } //while
    }


    /**
     * Starts the timing, called at the beginning of repair.
     */
    protected void startTimer() {
        startRepairTime = java.lang.System.currentTimeMillis();
    }


    /**
     * True if since startTimer more then millisForRepair have passed.
     * 
     * @return true if since startTimer more then millisForRepair have passed,
     *         else false.
     */
    protected boolean timeUp() {
        return java.lang.System.currentTimeMillis() > (startRepairTime + 
               repairedInstance.millisForRepair);
    }
}