/*
 * CostMappingLoader.java
 *
 * Created on 30. August 2002, 12:19
 */
package com.ai_center.dragonbreath.loader;

import com.ai_center.dragonbreath.*;
import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.costs.*;

import java.io.File;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;

import java.util.*;

import org.jdom.Attribute;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.Namespace;
import org.jdom.input.SAXBuilder;
import org.jdom.output.XMLOutputter;


/**
 * Loading and saving of CostMappings.
 * 
 * @version $Revision: 2.3 $
 * @author tim baermann
 */
public class CostDataLoader {
    // Static variables/initializers ******************************************

    /** 
     * Default SAX Driver class to use. 
     */
    private static final String DEFAULT_SAX_DRIVER_CLASS = 
            "org.apache.xerces.parsers.SAXParser";

    // Instance variables *****************************************************

    /** 
     * Reference to the global search control. 
     */
    protected GlobalSearchControl myGSC;

    /** 
     * The namespace of doc. 
     */
    private Namespace ns;

    // Constructors ***********************************************************

    /**
     * Creates a new instance of CostMappingLoader
     * 
     * @param g The GlobalSearchControl used.
     */
    public CostDataLoader(GlobalSearchControl gsc) {
        myGSC = gsc;
    }

    // Methods ****************************************************************

    /**
     * Returns an Element that contains all the costmappings and functions in
     * the engine.
     * 
     * @return An jdom.Element with the costData.
     * 
     * @throws IOException
     * @throws JDOMException
     */
    public Element getCostData() throws IOException, JDOMException {
        this.ns = ns;

        Element costData       = new Element("CostData", ns);
        Element defaultMapping = new Element("DefaultCostMapping", ns);
        defaultMapping.setAttribute("name", 
                                    myGSC.getDefaultCostMapping().getName());
        costData.addContent(defaultMapping);

        Element mappingTypes = getCostMappingTypes();

        if (mappingTypes != null) {
            costData.addContent(mappingTypes);
        }

        Element mappingInstances = getCostMappingInstances();

        if (mappingInstances != null) {
            costData.addContent(mappingInstances);
        }

        Element costFunctions = getCostFunctions();

        if (costFunctions != null) {
            costData.addContent(costFunctions);
        }

        return costData;
    }


    /**
     * Creates a Constraints CostData for storage.
     * 
     * @param c The constraint.
     * 
     * @return A jdom.Element containing
     */
    public Element getCostDataForConstraint(ConventionalConstraint c) {
        Element costRefs      = null;
        int     numberOfCosts = c.getNumberOfCosts();

        if (numberOfCosts > 0) {
            for (int i = 0; i < numberOfCosts; i++) {
                if (c.costFunctionReference[i] != null) {
                    if (costRefs == null) {
                        costRefs = new Element("CostFunctionReferences", ns);
                    }

                    Element cr = new Element("Reference", ns);
                    cr.setAttribute("costFunction", 
                                    c.costFunctionReference[i].costFunction.getName());
                    cr.setAttribute("costMapping", 
                                    c.costFunctionReference[i].costMapping.getName());
                    cr.setAttribute("costID", String.valueOf(i));
                    costRefs.addContent(cr);
                }
            }
        }

        return costRefs;
    }


    /**
     * Returns an Element with all the costMappingTypes in Registry.
     * 
     * @return A jdom.Element with serialized costMappings.
     * 
     * @throws IOException
     * @throws JDOMException
     */
    public Element getCostMappingTypes() throws IOException, JDOMException {
        Element mappingTypes = null;
        int     numberTypes = Registry.getCostMappingNumber();

        if (numberTypes > 2) { // only save mappings if there are more then the standart ones.
            mappingTypes = new Element("CostMappingTypes", ns);

            for (int nr = 2; nr < numberTypes; nr++) {
                mappingTypes.addContent(getMappingElement(
                                                (Class) Registry.costMappingClass[nr], 
                                                (String) Registry.costMappingName[nr]));
            }
        }

        return mappingTypes;
    }


    /**
     * Sets the namespace.
     * 
     * @param ns The new Namespace.
     */
    public void setNamespace(Namespace ns) {
        this.ns = ns;
    }


    /**
     * Loads all the CostMappings and CostFunctions in the element.
     * 
     * @param root The Element that contains the costData
     * 
     * @throws IOException
     * @throws JDOMException
     */
    public void loadCostData(Element root) throws IOException, JDOMException {
        Element costData = root.getChild("CostData", ns);

        if (costData != null) {
            loadCostMappingTypes(costData);
            loadCostMappingInstances(costData);
            loadCostFunctions(costData);

            // load default mapping after having loaded the mappings
            Element defaultMappingElement = costData.getChild(
                                                    "DefaultCostMapping", ns);

            if (defaultMappingElement != null) {
                CostMapping defaultMapping = myGSC.getCostMapping(
                                                     defaultMappingElement.getAttributeValue(
                                                             "name"));
                myGSC.setDefaultCostMapping(defaultMapping);
            }
        }
    }


    /**
     * Loads the CostFunctions stored with the given constraint.
     * 
     * @param c The Constraint.
     * @param constraintData The Element that contains all the data for the
     *        constraint.
     * 
     * @throws IOException
     * @throws JDOMException
     */
    public void loadCostDataForConstraint(ConventionalConstraint c, 
                                          Element constraintData)
                                   throws IOException, JDOMException {
        Element costFunctionRefs = constraintData.getChild(
                                           "CostFunctionReferences", ns);

        if (costFunctionRefs != null) {
            List     costFunctionRefsList = costFunctionRefs.getChildren(
                                                    "Reference", ns);
            Iterator it = costFunctionRefsList.iterator();

            if (it.hasNext()) {
                it.next(); // Do not load the first costfunction again, it was assigned to the constraint
                           // when the constraint was created, via loadFirstCostReferenceID().

                // Do not load the first costfunction again, it was assigned to the constraint
                // when the constraint was created, via loadFirstCostReferenceID().
            }

            while (it.hasNext()) {
                Element     cr           = (Element) it.next();
                String      functionName = cr.getAttributeValue("costFunction");
                String      mappingName  = cr.getAttributeValue("costMapping");
                int         costID       = cr.getAttribute("costID")
                                             .getIntValue();
                CostMapping m            = myGSC.getCostMapping(mappingName);
                myGSC.selectCostMapping(m);

                CostFunction cf = myGSC.getCostFunction(functionName);
                myGSC.selectCostFunction(cf);
                myGSC.addCosts(c, costID);
            }
        }
    }


    /**
     * Before a constraint can be created in GraphLoader the right costFunction
     * has to be chosen in myGSC.
     * 
     * @param constraintData An Element with all the Data of a Constraint.
     * 
     * @return The id of the first costfunction, normally 0.
     * 
     * @throws IOException
     * @throws JDOMException
     */
    public int loadFirstCostReferenceID(Element constraintData)
                                 throws IOException, JDOMException {
        int     costID           = -1;
        Element costFunctionRefs = constraintData.getChild(
                                           "CostFunctionReferences", ns);

        if (costFunctionRefs != null) {
            List     costFunctionRefsList = costFunctionRefs.getChildren(
                                                    "Reference", ns);
            Iterator it = costFunctionRefsList.iterator();

            if (it.hasNext()) {
                Element cr           = (Element) it.next();
                String  functionName = cr.getAttributeValue("costFunction");
                String  mappingName  = cr.getAttributeValue("costMapping");
                costID = cr.getAttribute("costID").getIntValue();

                CostMapping m = myGSC.getCostMapping(mappingName);
                myGSC.selectCostMapping(m);

                CostFunction cf = myGSC.getCostFunction(functionName);
                myGSC.selectCostFunction(cf);
            }
        }

        return costID;
    }


    /**
     * Returns an Element with all the registered costFunctionInstances in
     * GlobalSearchControl.
     * 
     * @return An Element with all the registered costFunctionInstances in
     *         GlobalSearchControl.
     */
    private Element getCostFunctions() {
        Element        costFunctions     = null;
        java.util.List costFunctionsList = myGSC.getAdditionalCostFunctions();

        if (costFunctionsList.size() > 0) {
            costFunctions = new Element("CostFunctions", ns);

            Iterator it = costFunctionsList.iterator();

            while (it.hasNext()) {
                CostFunction cf     = (CostFunction) it.next();
                Element      cmData = new Element("CostFunction", ns);
                cmData.setAttribute("name", cf.getName());
                cmData.setAttribute("selectionMethod", 
                                    java.lang.Integer.toString(
                                            cf.getConstraintSelectionMethod()));
                cmData.setAttribute("costMapping", cf.costMapping.getName());
                costFunctions.addContent(cmData);
            }
        }

        return costFunctions;
    }


    /**
     * Returns an Element with all the registered costMappingInstances in
     * GlobalSearchControl.
     * 
     * @return An Element with all the registered costMappingInstances in
     *         GlobalSearchControl.
     * 
     * @throws IOException
     * @throws JDOMException
     */
    private Element getCostMappingInstances() throws IOException, JDOMException {
        Element instances    = null;
        List    costMappings = myGSC.getCostMappings();

        if (costMappings.size() > 0) {
            instances = new Element("CostMappingInstances", ns);

            Iterator it = costMappings.iterator();

            while (it.hasNext()) {
                CostMapping cm     = (CostMapping) it.next();
                Element     cmData = new Element("CMInstance", ns);
                cmData.setAttribute("name", cm.getName());
                cmData.setAttribute("type", cm.getTypeName());
                instances.addContent(cmData);
            }
        }

        return instances;
    }


    /**
     * Returns an Element with a costMappingType describtion.
     * 
     * @param c  The class of the costMappingType
     * @param name The name of the costMapping
     * 
     * @return An Element with a describtion for the given costMapping.
     * 
     * @throws IOException
     * @throws JDOMException
     */
    private Element getMappingElement(Class c, String name)
                               throws IOException, JDOMException {
        Element newMap = new Element("CMType", ns);
        newMap.setAttribute("classname", c.getName());
        newMap.setAttribute("name", name);

        return newMap;
    }


    /**
     * Loads all the CostFunctions in costData into Registry.
     * loadCostMappingInstances() has to be called first.
     * 
     * @param costData The "CostData" section of a CSP description.
     * 
     * @throws IOException
     * @throws JDOMException
     */
    private void loadCostFunctions(Element costData) throws IOException, 
                                                            JDOMException {
        Element costFunctions = costData.getChild("CostFunctions", ns);

        if (costFunctions != null) {
            List     functions = costFunctions.getChildren("CostFunction", ns);
            Iterator it = functions.iterator();

            while (it.hasNext()) {
                Element f       = (Element) it.next();
                String  name    = f.getAttributeValue("name");
                int     selMeth = f.getAttribute("selectionMethod")
                                   .getIntValue();
                String  map     = f.getAttributeValue("costMapping");
                myGSC.selectCostMapping(myGSC.getCostMapping(map));

                CostFunction cf = myGSC.createCostFunction(name);
                cf.setConstraintSelectionMethod(selMeth);
            }
        }
    }


    /**
     * Loads all the CostMappingInstances in costData into GlobalSearchControl.
     * loadCostMappingTypes() has to be called first.
     * 
     * @param costData The "CostData" section of a CSP description.
     */
    private void loadCostMappingInstances(Element costData) {
        Element costMappingInstances = costData.getChild("CostMappingInstances", 
                                                         ns);

        if (costMappingInstances != null) {
            List     mappings = costMappingInstances.getChildren("CMInstance", 
                                                                 ns);
            Iterator it = mappings.iterator();

            while (it.hasNext()) {
                Element map      = (Element) it.next();
                String  name     = map.getAttributeValue("name");
                String  typeName = map.getAttributeValue("type");
                int     id       = Registry.getCostMappingID(typeName);

                if (id != -1) {
                    if (myGSC.getCostMapping(name) == null) { // only if there exists no mapping with the given name
                        myGSC.createCostMapping(id, name);
                    }
                } else {
                    throw new Error("For mapping " + name + 
                                    " in file exists no type " + typeName);
                }
            }
        }
    }


    /**
     * Loads all the CostMappingTypes given in costData into Registry.
     * 
     * @param costData The "CostData" section of a CSP description.
     */
    private void loadCostMappingTypes(Element costData) {
        Element costMappings = costData.getChild("CostMappingTypes", ns);

        if (costMappings != null) {
            List     mappings = costMappings.getChildren("CMType", ns);
            Iterator it = mappings.iterator();

            while (it.hasNext()) {
                Element map  = (Element) it.next();
                String  name = map.getAttributeValue("name");

                if (Registry.getCostMappingID(name) == -1) {
                    String className = map.getAttributeValue("classname");
                    Registry.registerCostMapping(className, name);
                }
            }
        }
    }
}