package com.ai_center.dragonbreath.loader;

import com.ai_center.dragonbreath.*;
import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.constraints.*;
import com.ai_center.dragonbreath.utilities.*;

import java.io.File;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;

import java.util.*;

import org.jdom.Attribute;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.Namespace;
import org.jdom.input.SAXBuilder;
import org.jdom.output.XMLOutputter;


/**
 * Loading and saving of constraint graphs.
 * 
 * @version $Revision: 2.5 $
 * @author tim baermann
 */
public class GraphLoader {
    // Static variables/initializers ******************************************

    /** 
     * Default SAX Driver class to use. 
     */
    private static final String DEFAULT_SAX_DRIVER_CLASS = 
            "org.apache.xerces.parsers.SAXParser";

    // Instance variables *****************************************************

    /** 
     * To save Costmappings 
     */
    protected CostDataLoader costDataLoader;

    /** 
     * Reference to the global search control. 
     */
    protected GlobalSearchControl gsc;

    /**
     * For the case that two nodes of the same kind are connected to an other
     * node here are all allready loaded nodes saved.
     */
    java.util.TreeSet connectedNodes = new java.util.TreeSet();

    /** 
     * For debug output 
     */
    private DebugLogger dl;

    /** 
     * The xml document. 
     */
    private Document doc;

    /** 
     * The root element of doc. 
     */
    private Element root;

    /**
     * Contains the newly created nodes , hashed with their uniqueIDs from file
     * used to make embeddings.
     */
    private HashMap nodeMap = new HashMap();

    /** 
     * All the edges in the XML graph description. 
     */
    private List edges;

    /** 
     * All the nodes in the XML graph description. 
     */
    private List nodes;

    /** 
     * The namespace of doc. 
     */
    private Namespace ns;

    /** 
     * XML reader class. 
     */
    private SAXBuilder builder;

    /** 
     * XML writer class. 
     */
    private XMLOutputter outputter;

    // Constructors ***********************************************************

    /**
     * Creates new GraphLoader
     * 
     * @param g The search control the graph should be loaded into.
     */
    public GraphLoader(GlobalSearchControl g) {
        gsc            = g;
        costDataLoader = new CostDataLoader(g);
        builder        = new SAXBuilder(DEFAULT_SAX_DRIVER_CLASS);


        // Create an outputter with default formatting
        outputter = new XMLOutputter("     ", true);

        dl = new DebugLogger(true);

        //dl.switchOn(1);// uncomment for debug output
    }

    // Methods ****************************************************************

    /**
     * Loads the graph at filename into GlobalSearchControl.
     * 
     * @param filename name of graph to be loaded
     * 
     * @throws IOException file io errors
     * @throws JDOMException XML io errors
     */
    public void load(java.lang.String filename) throws IOException, 
                                                       JDOMException {
        load(new File(filename));
    }


    /**
     * Loads the graph in file into gsc.
     * 
     * @param file The name of the graph to be loaded.
     * 
     * @throws IOException File io errors.
     * @throws JDOMException XML io errors.
     */
    public void load(File file) throws IOException, JDOMException {
        doc = builder.build(file);

        Element el = doc.getRootElement();
        load(el);
    } //public void load( File file)


    /**
     * Loads the graph specified by element into gsc.
     * 
     * @param graphRoot the root of the graph to be loaded
     * 
     * @throws IOException file io errors
     * @throws JDOMException XML io errors
     */
    public void load(Element graphRoot) throws IOException, JDOMException {
        root = graphRoot;

        Namespace ns = root.getNamespace();
        costDataLoader.setNamespace(ns);

        costDataLoader.loadCostData(root);

        Element el = root.getChild("Nodes", ns);


        //make shure there is nothing left from last call
        nodeMap.clear();

        if (el != null) {
            nodes = el.getChildren("Node", ns);
        } else {
            return;
        }

        el = root.getChild("Edges", ns);

        if (el != null) {
            edges = el.getChildren("Edge", ns);
        }

        // Get one node after the other, then load it ( if node needs other
        // nodes to exist first, the loadNode funciton will do that)
        Iterator itr = nodes.iterator();

        while (itr.hasNext()) {
            Element xmlNode = (Element) itr.next();
            Node    n = loadNode(xmlNode);
        }


        // Load all edges that are not part of basegraph
        itr = edges.iterator();

        while (itr.hasNext()) {
            Element xmlEdge = (Element) itr.next();
            loadEdge(xmlEdge);
        }
    } //public void load(Element graphRoot)


    /**
     * Save the actual constraintnet in filename.
     * 
     * @param name filename
     * 
     * @throws IOException file io error
     * @throws JDOMException XLM io error
     */
    public void save(java.lang.String name) throws IOException, JDOMException {
        save(new File(name));
    }


    /**
     * Save the actual constraintnet in file
     * 
     * @param file File reference.
     * 
     * @throws IOException file io error
     * @throws JDOMException XML io error
     */
    public void save(File file) throws IOException, JDOMException {
        //init jdom
        ns = Namespace.getNamespace(
                     "http://www.ai-center.com/projects/dragonbreath");

        Namespace w3cNs = Namespace.getNamespace("xsi", 
                                                 "http://www.w3.org/2001/XMLSchema-instance");
        root = new Element("Graph", ns);
        root.addNamespaceDeclaration(w3cNs);
        root.setAttribute("schemaLocation", 
                          "http://www.ai-center.com/projects/dragonbreath graph.xsd", 
                          w3cNs);
        doc = new Document(root);
        costDataLoader.setNamespace(ns);

        Element nodesXml = new Element("Nodes", ns);

        //save all variables
        Set      variables = gsc.getVariables();
        Iterator itr = variables.iterator();

        while (itr.hasNext()) {
            Variable n = (Variable) itr.next();
            nodesXml.addContent(createVariableElement(n));
        } //while

        //save all constraints 
        Set constraints = gsc.getConstraints();
        itr = constraints.iterator();

        while (itr.hasNext()) {
            Node n = (Node) itr.next();
            nodesXml.addContent(createConstraintElement(n));
        } //while

        root.addContent(nodesXml);

        //save all connections
        java.util.Vector allNodes = new java.util.Vector();
        allNodes.addAll(variables);
        allNodes.addAll(constraints);
        itr = allNodes.iterator();

        Element edgesXml = new Element("Edges", ns);

        while (itr.hasNext()) {
            Node n = (Node) itr.next();
            n.connections.initializeEnumerationAll();

            ConnectedNode cn = n.connections.getNext();

            while (cn != null) {
                if (cn.direction == Registry.OUTGOING) { //only handle outgoing edges, so every edge is handled only once
                    edgesXml.addContent(createEdgeElement(cn));
                } //if

                cn = n.connections.getNext();
            } //while cn
        } //while itr

        root.addContent(edgesXml);

        root.addContent(costDataLoader.getCostData());

        OutputStream out = new java.io.FileOutputStream(file);
        outputter.output(doc, out);
    } //save()


    /**
     * Which node is to be added to wich one?
     * 
     * @param nFrom the first Node of the connection
     * @param nTo the second Node of the connection
     * @param edgeType the type of the coonnection
     * 
     * @return true if to is to be added to from, false else
     */
    boolean findExtendible(Node nFrom, Node nTo, int edgeType) {
        int     i;
        boolean correct = false;

        if (nFrom instanceof Variable) {
            return false;
        }

        int extensionGraphSize = ((ConventionalConstraint) nFrom).getExtensionGraphNumber();

        for (i = 0; (i < extensionGraphSize) && ! correct; i++) {
            Embedding extGraph = ((ConventionalConstraint) nFrom).getExtensionGraph(i);

            if (extGraph.connectedNodeNumber > 1) {
                throw new Error(
                        "The constraint's extension has more than 1 edges, please report this to author of GraphLoader.java!");
            }

            if ((extGraph.connectedNode[0].direction == Registry.OUTGOING) && 
                    (extGraph.connectedNode[0].nodeTypeID == nTo.typeID) && 
                    (extGraph.connectedNode[0].edgeID == edgeType)) {
                return true;
            }
        } //for 

        return false;
    }


    /**
     * Find the xmlNode with the data n, that is connected to xmlNode nodeID
     * Call resetNextNodeConnectedWith() before first call.
     * 
     * @param nodeID The id of the searched XML node.
     * @param cn attributes of the searched node
     * 
     * @return null if none found else the XML node.
     * 
     * @throws IOException file io error
     * @throws JDOMException XML io error
     */
    Element findNextNodeConnectedWith(int nodeID, ConnectedNode cn)
                               throws IOException, JDOMException {
        Iterator itr = edges.iterator();

        while (itr.hasNext()) {
            Element   e      = (Element) itr.next();
            Attribute from   = e.getAttribute("From");
            boolean   isFrom = from.getIntValue() == nodeID;
            boolean   isTo   = e.getAttribute("To").getIntValue() == nodeID;

            if (isFrom || isTo) { // vertex forms part of edge e

                if (isTo == cn.direction) { //vertex is on right side of arrow

                    java.lang.String eType    = e.getAttributeValue("EType");
                    int              edgeType = Registry.getEdgeTypeID(eType);

                    if (edgeType == cn.edgeID) { // ok, candidate found, edgeID is right, too

                        int connectNodeID = -1;

                        if (isTo) { // read the id of the other xmlNode
                            connectNodeID = e.getAttribute("From")
                                             .getIntValue();
                        } else if (isFrom) {
                            connectNodeID = e.getAttribute("To").getIntValue();
                        }

                        if (getNodeTypeID(connectNodeID) == cn.nodeTypeID) { // the node has the right type

                            if (! connectedNodes.contains(computeUniqueIDFrom(
                                                                  connectNodeID, 
                                                                  edgeType))) { // add the new node to the allready connected nodes
                                connectedNodes.add(computeUniqueIDFrom(
                                                           connectNodeID, 
                                                           edgeType));

                                return findNode(connectNodeID);
                            }
                        }
                    } //edgeID
                } //if isTo == n.direction
            } //isFrom/isTo
        } //while itr.hasNext

        return null; // if the algorithm came here, no xmlNode was found.
    } //findNodeConnectedWith()


    /**
     * Finds the xmlNode with the given id.
     * 
     * @param id The id of the searched node.
     * 
     * @return The found element, null if doesnt exist.
     * 
     * @throws IOException file io error
     * @throws JDOMException XML io error
     */
    Element findNode(int id) throws IOException, JDOMException {
        Iterator itr = nodes.iterator();

        while (itr.hasNext()) {
            Element xmlNode = (Element) itr.next();
            int     nodeID = xmlNode.getAttribute("Id").getIntValue();

            if (nodeID == id) { // heureca, found it

                return xmlNode;
            }
        } //while

        return null;
    } //findNode()


    /**
     * Load the edge given in Element.
     * 
     * @param xmlEdge The edge to be loaded.
     * 
     * @throws IOException file io errors
     * @throws JDOMException XML io errors
     */
    void loadEdge(Element xmlEdge) throws IOException, JDOMException {
        int from = xmlEdge.getAttribute("From").getIntValue();
        int to = xmlEdge.getAttribute("To").getIntValue();

        if (nodeMap.containsKey(new Integer(from)) && 
                nodeMap.containsKey(new Integer(to))) {
            Node nFrom = (Node) nodeMap.get(new Integer(from));
            Node nTo = (Node) nodeMap.get(new Integer(to));


            //Is the edge allready established?
            nFrom.connections.initializeEnumerationAll();

            ConnectedNode cn = nFrom.connections.getNext();

            while ((cn != null) && (cn.node != nTo)) {
                cn = nFrom.connections.getNext();
            } //while cn

            if (cn != null) {
                return; // the edge is present
            } else { // create edge

                Embedding        emb      = new Embedding(1);
                java.lang.String eType    = xmlEdge.getAttributeValue("EType");
                int              edgeType = Registry.getEdgeTypeID(eType);

                boolean          direction = findExtendible(nFrom, nTo, 
                                                            edgeType);

                if (direction) {
                    emb.addConnectedNode(nTo, nTo.typeID, edgeType, 
                                         Registry.OUTGOING);
                    gsc.extendConstraint((ConventionalConstraint) nFrom, emb);
                } else {
                    emb.addConnectedNode(nFrom, nFrom.typeID, edgeType, 
                                         Registry.INCOMING);
                    gsc.extendConstraint((ConventionalConstraint) nTo, emb);
                }
            } //else
        } // if nodemap contains from && to
    } //loadedge()


    /**
     * Load the node given in Element.
     * 
     * @param xmlNode The node to be loaded
     * 
     * @return The newly created node
     * 
     * @throws IOException File io errors
     * @throws JDOMException XML io errors
     */
    Node loadNode(Element xmlNode) throws IOException, JDOMException {
        Node newNode = null;
        int  nodeID = xmlNode.getAttribute("Id").getIntValue();

        if (nodeMap.containsKey(new Integer(nodeID))) { //xmlNode allready loaded..

            Node n = (Node) nodeMap.get(new Integer(nodeID));
            dl.print(1, " allready loaded: " + n.toString() + "\n");

            return n;
        } //containtsKey

        java.lang.String nodeTypeName = xmlNode.getAttributeValue("NType");

        dl.print(1, nodeTypeName + "\n");

        int nodeType = Registry.getNodeTypeID(nodeTypeName);

        if (Registry.isVariable(nodeType)) {
            newNode = loadVariable(xmlNode);

            //nothing more to do..
        } else if (Registry.isConstraint(nodeType)) {
            newNode = loadConstraint(xmlNode, nodeType, nodeID);
        } else {
            throw new Error("Node neither variable nor constraint.");
        }

        newNode.setAuxiliaryNodeData(xmlNode);

        nodeMap.put(new Integer(nodeID), newNode);

        return newNode;
    } //loadNode()


    /**
     * Resets connectedNodes.
     */
    void resetNodeConnectedWith() {
        connectedNodes.clear();
    }


    /**
     * Returns the TypeID of the Node with the given id.
     * 
     * @param nodeID The nodeID as stored in the xml-file
     * 
     * @return The typeID as in Registry.
     * 
     * @throws IOException
     * @throws JDOMException
     */
    private int getNodeTypeID(int nodeID) throws IOException, JDOMException {
        Integer key = new Integer(nodeID);

        if (nodeMap.containsKey(key)) { //xmlNode allready loaded..

            return ((Node) nodeMap.get(key)).typeID;
        } else {
            Element xmlNode = findNode(nodeID);

            if (xmlNode == null) {
                throw new Error(
                        "There is an Edge in the file to a not existing Node.");
            }

            String nodeTypeName = xmlNode.getAttributeValue("NType");

            return Registry.getNodeTypeID(nodeTypeName);
        }
    }


    /**
     * Adds the Coordinates of the given Node as attributes to the element el.
     * 
     * @param el The Element that takes the data.
     * @param n The node that is stored in el.
     */
    private void addGraphicData(Element el, Node n) {
        if (n.graphObject != null) {
            int x = 0;
            int y = 0;

            if (n.graphObject instanceof NodeGraphics) {
                x = ((NodeGraphics) n.graphObject).x;
                y = ((NodeGraphics) n.graphObject).y;
            } else {
                x = ((NodeGraphicsCache) n.graphObject).x;
                y = ((NodeGraphicsCache) n.graphObject).y;
            }

            el.setAttribute("X", java.lang.String.valueOf(x));
            el.setAttribute("Y", java.lang.String.valueOf(y));
        }
    }


    /**
     * Computes a unique id from the given integers, used as keys into
     * connectedNodes[].
     * 
     * @param connectedNodeID The id of a node as stored in xml-file.
     * @param edgeType The edgeTypeID as in Registry.
     * 
     * @return An unique Long value, that can be used as key in
     *         connectedNodes[].
     */
    private Long computeUniqueIDFrom(int connectedNodeID, int edgeType) {
        String asString = new String();
        asString = asString.concat(Integer.toString(connectedNodeID));
        asString = asString.concat(Integer.toString(edgeType));

        return new Long(asString);
    }


    /**
     * Creates an element with all the data of the Constraint.
     * 
     * @param n  The constraint that should be stored.
     * 
     * @return An Element with the serialized Constraint.
     */
    private Element createConstraintElement(Node n) {
        Element newConstr = new Element("Node", ns);
        newConstr.setAttribute("Id", java.lang.String.valueOf(n.hashCode())); // i hope this is unique for this program-run..
        addGraphicData(newConstr, n);
        newConstr.setAttribute("NType", 
                               Registry.nodeTypeName[((Node) n).typeID]);
        n.getAuxiliaryNodeData(newConstr);

        if (n instanceof ConventionalConstraint) {
            Element costFunctionReferences = costDataLoader.getCostDataForConstraint(
                                                     (ConventionalConstraint) n);

            if (costFunctionReferences != null) {
                newConstr.addContent(costFunctionReferences);
            }
        }

        return newConstr;
    }


    /**
     * Creates an element with all the data of the Edge.
     * 
     * @param e  The edge that should be stored.
     * 
     * @return A jdom.Element that contains the data for e.
     */
    private Element createEdgeElement(ConnectedNode e) {
        Element newEdge = new Element("Edge", ns);
        newEdge.setAttribute("EType", Registry.edgeTypeName[e.edgeID]);
        newEdge.setAttribute("From", 
                             java.lang.String.valueOf(e.other.node.hashCode()));
        newEdge.setAttribute("To", java.lang.String.valueOf(e.node.hashCode()));

        return newEdge;
    }


    /**
     * Creates an element with all the data of the variable.
     * 
     * @param v  The variable that should be stored.
     * 
     * @return An jdom.Element that represents the given v.
     */
    private Element createVariableElement(Variable v) {
        Element newVar = new Element("Node", ns);
        newVar.setAttribute("Id", java.lang.String.valueOf(v.hashCode())); // i hope this id is unique for this program-run..
        addGraphicData(newVar, v);
        newVar.setAttribute("NType", "Variable");
        v.getAuxiliaryNodeData(newVar);

        return newVar;
    }


    /**
     * Loads the constraint at xmlNode. Therefore also loads all other nodes
     * necessary to create the basegraph of the constraint.
     * 
     * @param xmlNode The loaded nodes data.
     * @param nodeType The type of the node loaded, as in Registry.
     * @param nodeID The id value, as in the xml-File.
     * 
     * @return A node that is registered in the gsc.
     * 
     * @throws IOException
     * @throws JDOMException
     */
    private Node loadConstraint(Element xmlNode, int nodeType, int nodeID)
                         throws IOException, JDOMException {
        Node      newNode;
        int       x        = xmlNode.getAttribute("X").getIntValue();
        int       y        = xmlNode.getAttribute("Y").getIntValue();
        Embedding bg       = ((ConventionalConstraint) Registry.nodeType[nodeType]).getBaseGraph();
        Embedding emb      = new Embedding(bg.connectedNodeNumber);
        Node      bgNode_i = null;
        resetNodeConnectedWith();
        dl.print(1, "\ncreating embedding \n");

        for (int i = 0; i < bg.connectedNodeNumber; i++) {
            // add all the nodes that comprise the embedding of this one
            // find an xmlNode in xml file that represents the given node in baseGraph
            ConnectedNode c = bg.connectedNode[i];
            dl.print(1, "edgeID: " + c.edgeID + " nodeTypeID: " + 
                     c.nodeTypeID);

            Element connectedNode = findNextNodeConnectedWith(nodeID, 
                                                              bg.connectedNode[i]);

            if (connectedNode != null) {
                bgNode_i = loadNode(connectedNode);
                gsc.extendEmbedding(emb, bgNode_i, bg.connectedNode[i].edgeID, 
                                    bg.connectedNode[i].direction);
            } //if connenctednode!=0
        } //for int i

        int costID = costDataLoader.loadFirstCostReferenceID(xmlNode);

        if (costID != -1) {
            newNode = gsc.createConstraint(nodeType, emb, costID, x, y);
        } else {
            newNode = gsc.createConstraint(nodeType, emb, x, y);
        }

        costDataLoader.loadCostDataForConstraint(
                (ConventionalConstraint) newNode, xmlNode);

        return newNode;
    }


    /**
     * Loads the variable at xmlNode.
     * 
     * @param xmlNode The variables data from the xml-file.
     * 
     * @return A variable with the given data, that is registered in the gsc.
     * 
     * @throws IOException
     * @throws JDOMException
     */
    private Node loadVariable(Element xmlNode) throws IOException, 
                                                      JDOMException {
        Node newNode;
        int  x  = xmlNode.getAttribute("X").getIntValue();
        int  y  = xmlNode.getAttribute("Y").getIntValue();
        newNode = gsc.createVariable("", x, y);

        return newNode;
    }
} //class
