/*
 * ScLoader.java
 *
 * Created on 3. November 2001, 19:34
 */
package com.ai_center.dragonbreath.loader;

import com.ai_center.dragonbreath.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.SCGraphFactory;
import com.ai_center.dragonbreath.utilities.*;

import java.io.*;

import java.util.*;

import org.jdom.*;
import org.jdom.input.SAXBuilder;
import org.jdom.output.XMLOutputter;


/**
 * Loading and saving of constraint graphs.
 * 
 * @version $Revision: 2.5 $
 * @author tim baermann
 */
public class ScLoader {
    // Static variables/initializers ******************************************

    /** 
     * Default SAX Driver class to use. 
     */
    private static final String DEFAULT_SAX_DRIVER_CLASS = 
            "org.apache.xerces.parsers.SAXParser";

    // Instance variables *****************************************************

    /** 
     * To load Costmappings 
     */
    protected CostDataLoader costDataLoader;

    /** 
     * For output of debug messages. 
     */
    protected PrintStream out;

    /** 
     * The associated StructuralConstraintSearchControl. 
     */
    StructuralConstraintManagement mySCM;

    /** 
     * The xml document. 
     */
    private Document doc;

    /**
     * Contains the newly created nodes , hashed with their uniqueIDs from
     * file. Used to make embeddings.
     */
    private HashMap nodeMap = new HashMap();

    /** 
     * The documents namespace. 
     */
    private Namespace ns;

    /** 
     * The XML input class. 
     */
    private SAXBuilder builder;

    /** 
     * The XML output class. 
     */
    private XMLOutputter outputter;

    // Constructors ***********************************************************

    /**
     * Creates new ScLoader
     * 
     * @param scm
     * @param gsc The GlobalSearchControl to be used.
     */
    public ScLoader(StructuralConstraintManagement scm, GlobalSearchControl gsc) {
        mySCM          = scm;
        costDataLoader = new CostDataLoader(gsc);
        builder        = new SAXBuilder(DEFAULT_SAX_DRIVER_CLASS);


        // Create an outputter with default formatting
        outputter = new XMLOutputter(" ", true);
    }

    // Methods ****************************************************************

    /**
     * Loads the graph at filename into gsc.
     * 
     * @param filename name of graph to be loaded
     * 
     * @throws IOException file errors
     * @throws JDOMException XML load errors
     */
    public void load(java.lang.String filename) throws IOException, 
                                                       JDOMException {
        System.out.print("\nLoading " + filename + "\n");
        load(new File(filename));
    }


    /**
     * loads the graph in file into gsc
     * 
     * @param file graph to be loaded
     * 
     * @throws IOException file errors
     * @throws JDOMException XML load errors
     */
    public void load(File file) throws IOException, JDOMException {
        doc = builder.build(file);

        // Get the root element
        Element el = doc.getRootElement();

        load(el);
    } //public void load(File file)


    /**
     * Loads a Structural Constraint description (as defined in sc.xsd) into
     * SCSC.
     * 
     * @param graphRoot the root of the graph
     * 
     * @throws IOException file io errors
     * @throws JDOMException XML load errors
     */
    public void load(Element graphRoot) throws IOException, JDOMException {
        ns = graphRoot.getNamespace();
        costDataLoader.setNamespace(ns);

        Iterator       it;
        Element        subgraph;

        SCGraphFactory graphFactory            = new SCGraphFactory();
        String         customInstanceClassName = graphRoot.getAttributeValue(
                                                         "customInstanceClassName", 
                                                         ns);

        long           millisForRepair                = 
                Registry.MILLISECONDS_FOR_REPAIR;
        double         removeNodeProbability          = 
                Registry.PROBABILITY_REMOVE_NODE;
        double         removeDockingProbability       = 
                Registry.PROBABILITY_REMOVE_DOCKING;
        double         keepOnRepairingLastProbability = 
                Registry.PROBABILITY_KEEP_ON_REPAIRING_LAST;

        Attribute      aMillisForRepair = graphRoot.getAttribute(
                                                  "millisForRepair", ns);

        if (aMillisForRepair != null) {
            millisForRepair = aMillisForRepair.getLongValue();
        }

        Attribute aRemoveNodeProbability = graphRoot.getAttribute(
                                                   "removeNodeProbability");

        if (aRemoveNodeProbability != null) {
            removeNodeProbability = aRemoveNodeProbability.getDoubleValue();
        }

        Attribute aRemoveDockingProbability = graphRoot.getAttribute(
                                                      "removeDockingProbability");

        if (aRemoveDockingProbability != null) {
            removeDockingProbability = aRemoveDockingProbability.getDoubleValue();
        }

        Attribute aKeepOnRepairingLastProbability = graphRoot.getAttribute(
                                                            "keepOnRepairingLastProbability");

        if (aKeepOnRepairingLastProbability != null) {
            keepOnRepairingLastProbability = aKeepOnRepairingLastProbability.getDoubleValue();
        }

        String testGraphRepairerClassName = graphRoot.getAttributeValue(
                                                    "TestGraphRepairer");

        String dockingGraphRepairerClassName = graphRoot.getAttributeValue(
                                                       "DockingGraphRepairer");

        graphFactory.setCustomInstanceParameters(customInstanceClassName, 
                                                 dockingGraphRepairerClassName, 
                                                 testGraphRepairerClassName, 
                                                 millisForRepair, 
                                                 removeNodeProbability, 
                                                 removeDockingProbability, 
                                                 keepOnRepairingLastProbability);

        Element docking = graphRoot.getChild("Docking", ns);

        if (docking != null) {
            graphFactory.beginDocking();
            loadGraph(docking, graphFactory);
        } //if maingraph

        List testGraphs = graphRoot.getChildren("Test", ns);

        //load all test graphs in file
        Element test;

        if (testGraphs != null) {
            it = testGraphs.iterator();

            while (it.hasNext()) {
                test = (Element) it.next();
                graphFactory.beginTest();
                loadGraph(test, graphFactory);
            } //while
        } //if

        Element links = graphRoot.getChild("Links", ns);

        if (links != null) {
            List linkList = links.getChildren("Link", ns);
            it = linkList.iterator();

            while (it.hasNext()) { // add all links

                Element xmlLink = (Element) it.next();
                int     from = xmlLink.getAttribute("From").getIntValue();
                int     to   = xmlLink.getAttribute("To").getIntValue();

                if (nodeMap.containsKey(new Integer(from)) && 
                        nodeMap.containsKey(new Integer(to))) {
                    Object nFrom = nodeMap.get(new Integer(from));
                    Object nTo = nodeMap.get(new Integer(to));
                    graphFactory.addLink(nFrom, nTo);
                }
            } //while
        }

        Element          nameEl = graphRoot.getChild("Name", ns);

        java.lang.String name;

        if (nameEl != null) {
            name = nameEl.getText();
        } else {
            name = "unnamed";
        }

        int costID = costDataLoader.loadFirstCostReferenceID(graphRoot);

        if (mySCM.createConstraintType(name, graphFactory)) {
            System.out.print("\nLoaded Structural Constraint: " + name + 
                             "\n");
        }
    } //public void load(Element graphRoot)


    /**
     * Loads the graph specified by element into a SCGraphgraphFactory.
     * 
     * @param gRoot the root of the graph
     * @param graphFactory the SCGraphgraphFactory this graph should be loaded
     *        into
     * 
     * @throws IOException file io errors
     * @throws JDOMException XML load errors
     */
    public void loadGraph(Element gRoot, SCGraphFactory graphFactory)
                   throws IOException, JDOMException {
        Iterator graphIt;
        Iterator linkIt;
        Element  main = gRoot.getChild("Main", ns); // every graph has one main subgraph

        if (main != null) {
            graphFactory.beginSubgraph(SCGraphFactory.Main);
            loadSubgraph(main, graphFactory);
        }

        //load all the pacs
        List pacs = gRoot.getChildren("PAC", ns); //and every graph has 0-n pac/nac subgraphs

        if (pacs != null) {
            graphIt = pacs.iterator();

            while (graphIt.hasNext()) { //for every pac

                Element pac = (Element) graphIt.next();
                graphFactory.beginSubgraph(SCGraphFactory.PAC);
                loadSubgraph(pac, graphFactory);
            } //while graphIt
        }

        List nacs = gRoot.getChildren("NAC", ns);

        if (nacs != null) {
            graphIt = nacs.iterator();

            while (graphIt.hasNext()) {
                Element nac = (Element) graphIt.next();
                graphFactory.beginSubgraph(SCGraphFactory.NAC);
                loadSubgraph(nac, graphFactory);
            }
        }

        Element links = gRoot.getChild("Links", ns);

        if (links != null) {
            List linkList = links.getChildren("Link", ns);
            linkIt = linkList.iterator();

            while (linkIt.hasNext()) {
                Element xmlLink = (Element) linkIt.next();
                int     from = xmlLink.getAttribute("From").getIntValue();
                int     to   = xmlLink.getAttribute("To").getIntValue();

                if (nodeMap.containsKey(new Integer(from)) && 
                        nodeMap.containsKey(new Integer(to))) {
                    Object nFrom = nodeMap.get(new Integer(from));
                    Object nTo = nodeMap.get(new Integer(to));

                    graphFactory.addLink(nFrom, nTo);
                }
            } //while
        } //if (links!=null)
    } //loadGraph()


    /**
     * Loads the subgraph specified by element into a SCGraphgraphFactory.
     * 
     * @param graph The XML subgraph describtion that should be loaded.
     * @param graphFactory the SCGraphgraphFactory this graph should be loaded
     *        into
     * 
     * @throws IOException file io errors
     * @throws JDOMException XML load errors
     */
    public void loadSubgraph(Element graph, SCGraphFactory graphFactory)
                      throws IOException, JDOMException {
        List     nodes      = null;
        List     edges      = null;
        Iterator itr;
        Iterator subgraphIt;
        List     parts      = graph.getChildren("Graph", ns);

        if (parts != null) {
            subgraphIt = parts.iterator();

            while (subgraphIt.hasNext()) {
                Element subgraph = (Element) subgraphIt.next();
                graphFactory.beginSubgraphpart();

                Element el = subgraph.getChild("Nodes", ns);

                if (el != null) {
                    nodes = el.getChildren("Node", ns);
                } else {
                    nodes = null;
                }

                el = subgraph.getChild("Edges", ns);

                if (el != null) {
                    edges = el.getChildren("Edge", ns);
                } else {
                    edges = null;
                }

                if (nodes != null) {
                    // get one node after the other, then load it
                    itr = nodes.iterator();

                    while (itr.hasNext()) {
                        Element xmlNode         = (Element) itr.next();
                        int     nodeID          = xmlNode.getAttribute("Id")
                                                         .getIntValue();
                        String  isLeafNodeValue = xmlNode.getAttributeValue(
                                                          "isLeafNode");
                        boolean isLeafNode      = false;

                        if (isLeafNodeValue != null) {
                            isLeafNode = isLeafNodeValue.compareToIgnoreCase(
                                                 "true") == 0;
                        }

                        java.lang.String nodeTypeName = 
                                xmlNode.getAttributeValue("NType");
                        int              nodeType = Registry.getNodeTypeID(
                                                            nodeTypeName);

                        if (nodeType == -1) {
                            throw new Error("Wrong NodeType name " + 
                                            nodeTypeName + 
                                            " in Structural Constraint Describtion.");
                        }

                        Object newNode = graphFactory.addNode(nodeType, 
                                                              isLeafNode, 
                                                              xmlNode);
                        nodeMap.put(new Integer(nodeID), newNode);
                    } //while
                } //if

                if (edges != null) {
                    //load all edges
                    itr = edges.iterator();

                    while (itr.hasNext()) {
                        Element xmlEdge = (Element) itr.next();

                        int     from = xmlEdge.getAttribute("From")
                                              .getIntValue();
                        int     to   = xmlEdge.getAttribute("To").getIntValue();

                        if (nodeMap.containsKey(new Integer(from)) && 
                                nodeMap.containsKey(new Integer(to))) {
                            Object           nFrom    = 
                                    nodeMap.get(new Integer(from));
                            Object           nTo      = 
                                    nodeMap.get(new Integer(to));
                            java.lang.String eType    = 
                                    xmlEdge.getAttributeValue("EType");
                            int              edgeType = 
                                    Registry.getEdgeTypeID(eType);

                            if (edgeType == -1) {
                                throw new Error(
                                        "Wrong EdgeType name in Structural Constraint Describtion.");
                            }

                            graphFactory.addEdge(nFrom, nTo, edgeType);
                        }
                    } //while
                } // if edges!=null
            } //while subgraphIt.hasNext()
        } //parts!=null
    } //public void loadSubgraph
} //class
/* 
 * DO _NOT_ EDIT ANYTHING BELOW THIS POINT!
 *
 * $Log: ScLoader.java,v $
 * Revision 2.5  2003/05/07 23:38:53  teutates
 * documentation for structural constraint handling updated
 *
 * Revision 2.4  2003/02/24 21:42:03  teutates
 * no message
 *
 * Revision 2.3  2003/02/21 16:55:23  teutates
 * improved auxiliary data load for nodes
 *
 * Revision 2.2  2003/02/21 01:33:02  teutates
 * structural repair initial release
 *
 * Revision 2.1  2002/10/22 21:45:47  teutates
 * testGraph initial.
 * dockingGraph reorganized.
 *
 * Revision 2.0  2002/09/17 00:33:06  nareyek
 * work on beta version began: resource type, state tasks and precondition tasks introduced
 *
 * Revision 1.2  2002/09/02 15:22:22  teutates
 * cost data is saved and loaded now
 *
 * Revision 1.1  2002/08/30 09:48:53  teutates
 * moved all load functionality into own package
 *
 * Revision 1.2  2002/08/07 16:19:29  teutates
 * added structural constraint access from GSCGraphics
 *
 * Revision 1.1  2002/07/25 01:23:10  teutates
 * at last..
 * a new version of the dockingGraph matching algorithm.
 *
 * Revision 1.2  2002/03/27 11:17:49  teutates
 * not connected graphs added
 *
 * Revision 1.1  2002/01/30 00:44:37  teutates
 * initial release
 *
 *
 *
 */
