/*
 * ScpLoader.java
 *
 * Created on 3. November 2001, 19:34
 */
package com.ai_center.dragonbreath.loader;

import com.ai_center.dragonbreath.GlobalSearchControl;
import com.ai_center.dragonbreath.loader.GraphLoader;
import com.ai_center.dragonbreath.structuralConstraintHandling.StructuralConstraintManagement;

import java.io.File;
import java.io.IOException;

import java.util.Iterator;
import java.util.List;

import org.jdom.*;
import org.jdom.input.SAXBuilder;
import org.jdom.output.XMLOutputter;


/**
 * Loading and saving of <B>S</B>tructural <B>C</B>onstraint <B>P</B>roblems.
 * 
 * @version $Revision: 2.2 $
 * @author tim baermann
 */
public class ScpLoader {
    // Static variables/initializers ******************************************

    /** 
     * Default SAX Driver class to use 
     */
    private static final String DEFAULT_SAX_DRIVER_CLASS = 
            "org.apache.xerces.parsers.SAXParser";

    // Instance variables *****************************************************

    /**
     * Reference to the GlobalSearchControl to which the problem should be
     * loaded.
     */
    protected GlobalSearchControl gsc;

    /** 
     * For loading of allready existing problem graphs. 
     */
    GraphLoader graphLoader;

    /** 
     * To load <B>S</B>tructural <B>C</B>onstraints. 
     */
    ScLoader scLoader;

    /**
     * The StructuralConstraintSearchControl to which the structural
     * constraints should be loaded.
     */
    StructuralConstraintManagement mySCM;

    /** 
     * The XML document. 
     */
    private Document doc;

    /** 
     * The root element of doc. 
     */
    private Element root;

    /** 
     * XML namespace of the document. 
     */
    private Namespace ns;

    /** 
     * XML-Input interface 
     */
    private SAXBuilder builder;

    /** 
     * XML output interface 
     */
    private XMLOutputter outputter;

    // Constructors ***********************************************************

    /**
     * Creates new ScpLoader.
     * 
     * @param scm The StructuralConstraintManagement responsible for the loaded
     *        Constraints.
     * @param gsc The GlobalSearchControl to be used.
     */
    public ScpLoader(StructuralConstraintManagement scm, 
                     GlobalSearchControl gsc) {
        builder = new SAXBuilder(DEFAULT_SAX_DRIVER_CLASS);


        // Create an outputter with default formatting
        outputter   = new XMLOutputter(" ", true);
        graphLoader = new GraphLoader(gsc);

        mySCM = scm;

        scLoader = new ScLoader(mySCM, gsc);
    } //public ScpLoader()

    // Methods ****************************************************************

    /**
     * Loads the structural constraint satisfaction problem at filename into
     * gsc.
     * 
     * @param filename name of graph to be loaded
     * 
     * @throws IOException Errors in file handling
     * @throws JDOMException errors in the XML file format
     */
    public void load(java.lang.String filename) throws IOException, 
                                                       JDOMException {
        load(new File(filename));
    } //public void load(java.lang.String filename)


    /**
     * Loads the structural constraint satisfaction problem in file into gsc.
     * 
     * @param file graph to be loaded
     * 
     * @throws IOException Errors in file handling
     * @throws JDOMException errors in the XML file format
     */
    public void load(File file) throws IOException, JDOMException {
        // first load all the structural constraints for this problem
        doc = builder.build(file);


        // Get the root element
        root = doc.getRootElement();

        Namespace ns     = root.getNamespace();
        List      SCs;
        List      SCrefs;

        Element   el = root.getChild("StructuralConstraints", ns);

        if (el != null) { // there are Structural Constraints in the problem
            SCs    = el.getChildren("SC", ns);
            SCrefs = el.getChildren("SCFile", ns);

            Iterator itr = SCs.iterator();

            while (itr.hasNext()) { // load all defined structural constraints

                Element xmlSC = (Element) itr.next();
                scLoader.load(xmlSC);
            }

            itr = SCrefs.iterator();

            while (itr.hasNext()) { // load all referenced (a reference is a filename) structural constraints

                Element xmlSCref = (Element) itr.next();
                scLoader.load(xmlSCref.getText());
            }
        } //if el!=null

        el = root.getChild("CSP", ns);

        if (el != null) { // a problem graph has allready been specified
            el = el.getChild("Graph", ns);

            if (el != null) {
                graphLoader.load(el);
            }
        } //if
    } //public void load(File file)
} //class
/* 
 * DO _NOT_ EDIT ANYTHING BELOW THIS POINT!
 *
 * $Log: ScpLoader.java,v $
 * Revision 2.2  2003/05/07 23:38:53  teutates
 * documentation for structural constraint handling updated
 *
 * Revision 2.1  2003/02/21 01:33:02  teutates
 * structural repair initial release
 *
 * Revision 2.0  2002/09/17 00:33:06  nareyek
 * work on beta version began: resource type, state tasks and precondition tasks introduced
 *
 * Revision 1.2  2002/09/02 15:22:22  teutates
 * cost data is saved and loaded now
 *
 * Revision 1.1  2002/08/30 09:48:53  teutates
 * moved all load functionality into own package
 *
 * Revision 1.1  2002/07/25 01:23:10  teutates
 * at last..
 * a new version of the dockingGraph matching algorithm.
 *
 * Revision 1.1  2002/01/30 00:44:37  teutates
 * initial release
 *
       
 *
 */
