/*
 * StructuralConstraintInstance.java
 *
 * Created on 7. April 2002, 16:51
 */
package com.ai_center.dragonbreath.structuralConstraintHandling;

import com.ai_center.dragonbreath.*;
import com.ai_center.dragonbreath.Constraint;
import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.costs.CostFunctionReference;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.dockingGraph.SCIDockingGraph;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.movableSubgraph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.testGraph.*;
import com.ai_center.dragonbreath.utilities.*;

import java.util.*;


/**
 * Represents one Structural constraint instance, with its docking and its test
 * Graph parts and with the repair strategies.
 * Whenever a new DockingGraph match is found a new StructuralConstraintInstance is
 * created.
 * @version $Revision: 2.8 $
 * @author tim baermann
 */
public class StructuralConstraintInstance implements NodeChangedInterface {
    // Instance variables *****************************************************

    /**
     * When the dockingGraph is destroyed and has NACs this is the probabillity
     * that a nac is created and not a mainpart removed.
     */
    public double createDockingNACProbability;

    /**
     * If the testGraph could not be repaired in one step this number
     * determines the probability of choosing the same testGraph again in the
     * next repair step.
     */
    public double keepOnRepairingLastProbability;

    /**
     * The probability that a StructuralConstraintInstance destroys its docking
     * graph instead of repairing the testGraph.
     */
    public double removeDockingProbability;

    /**
     * The probability, that when a graphpart is destroyed during repair, a
     * node is removed and not an edge. normally it is better to remove only
     * edges, as that changes less of the graph, but we do not want to exclude
     * any possible graph, so include a small probability that a node is
     * removed instead of an edge.
     */
    public double removeNodeProbability;

    /** 
     * How much time may a repair step take? 
     */
    public long millisForRepair;

    /**
     * When edges are created the edgeChooser has to know the node that is
     * currently created. the getFittingNodeForEdgeCreation() is called
     * therefore. and there the currently changed node has to be known, to
     * decide which other nodes to use. This is it. It is set before a
     * createNode call is placed.
     */
    protected SCNode actualCreatedNode = null;

    /** 
     * For debug output. 
     */
    DebugLogger dl;

    /** 
     * The Instances dockingGraph. 
     */
    SCIDockingGraph dockingGraph;

    /**
     * If there is no valid testGraph this Vector stores the order of
     * correctness of the testgraphs. The first element is the most correct
     * testGraph, the last the least correct.
     */
    SortedMap invalidTestGraphs;

    /** 
     * The StructuralConstraintProblemControl. 
     */
    StructuralConstraintManagement mySCM;

    /** 
     * The type of this StructuralConstraintInstance 
     */
    StructuralConstraintType myType;

    /** 
     * The testGraph used for this instance. 
     */
    TestGraph actualTestGraph;

    /** 
     * The testGraph that was chosen for repair in the last repairstep. 
     */
    TestGraph lastRepairedTestGraph = null;

    /** 
     * All the testGraphs of this StructuralConstraint 
     */
    TestGraphTypesContainer testGraphs;

    /** 
     * The structural Costs. 
     */
    int structuralCosts;

    // Constructors ***********************************************************

    /**
     * Creates a new instance of StructuralConstraintInstance
     */
    public StructuralConstraintInstance() {
    }

    // Methods ****************************************************************

    /**
     * Has to be called straight after ObjectCreation. Is seperate from
     * constructor, as there should be the possibility of constructing a class
     * via java.lang.Class.forName(customInstanceClassName).newInstance().
     * 
     * @param type The type of this instance.
     * @param docking The dockingGraph with the match for this instance.
     * @param tGs  The testGraphs.
     * @param millisForRepair The time a structural repair of one instance may
     *        take.
     * @param removeNodeProbability The probability that, when a graph part is
     *        destroyed for repair, a node is removed and not an edge.
     * @param removeDockingProbability The probability that the DockingGraph is
     *        destroyed and not a TestGraph repaired.
     * @param keepOnRepairingLastProbability When a TestGraph could not be
     *        repaired in one iteration     this is the probability that next
     *        time the same testgraph is repaired again.
     */
    public void setConstructionParameters(StructuralConstraintType type, 
                                          SCIDockingGraph docking, 
                                          TestGraphTypesContainer tGs, 
                                          long millisForRepair, 
                                          double removeNodeProbability, 
                                          double removeDockingProbability, 
                                          double keepOnRepairingLastProbability) {
        mySCM        = type.mySCM;
        myType       = type;
        dockingGraph = docking;
        dockingGraph.setInstance(this);
        testGraphs = tGs;
        dl         = new DebugLogger(true);

        this.millisForRepair                = millisForRepair;
        this.removeNodeProbability          = removeNodeProbability;
        this.removeDockingProbability       = removeDockingProbability;
        this.keepOnRepairingLastProbability = keepOnRepairingLastProbability;
        dl.switchOn(7);
    }


    /**
     * Queries costs of the constraintInstance.
     * 
     * @return The Costs.
     */
    public long getCosts() {
        return structuralCosts;
    }


    /**
     * Queries the number of different costs available for the constraint.
     * 
     * @return An unique number.
     */
    public int getTypeID() {
        return 3;
    }


    /**
     * Checks nodes that were newly created for creation or destruction of
     * StructuralConstraintInstances.
     * 
     * @param newNodes Nodes that were recently created.
     * @param deletedNodes The nodes deleted.
     * @param expandedNodes Expanded Nodes
     * @param reducedNodes Nodes that lost connections.
     */
    public void checkChangedNodes(Set newNodes, Set deletedNodes, 
                                  Set expandedNodes, Set reducedNodes) {
        dockingGraph.checkChangedNodes(newNodes, deletedNodes, expandedNodes, 
                                       reducedNodes);
        
        // the docking graph has not become invalid
        if (! dockingGraph.isValid()) {
           myType.markInstanceForRemoval(this);
        } else {
            if (actualTestGraph != null) {
                actualTestGraph.checkChangedNodes(newNodes, deletedNodes, 
                                                  expandedNodes, reducedNodes);

                if (! actualTestGraph.isValid()) {
                    actualTestGraph = null;
                }
            }

            updateCosts();
        }
    }


    /**
     * Sets the actualTestGraph variable. Therefore has to be called with one
     * of the testgraph Types of this constarint.
     * 
     * @param interpretedExampleGraph A testgraph
     */
    public void createNewTestGraphInstance(TestGraph interpretedExampleGraph) {
        actualTestGraph = testGraphs.getTestGraphInstance(
                                  interpretedExampleGraph);
    }


    /**
     * If a unconnected NAC was created for the dockingGraph this is checked by
     * the SCType, this function is called if the Docking graph got invalid
     * due to a new unconnected NAC, or if the repair removed the
     * dockinggraph.
     */     
    public void dockingGraphInvalid() {
        myType.markInstanceForRemoval(this);
        dockingGraph = null;
    }


    /**
     * Repair the graph, so that the costs are lower. That could be by
     * completing one testGraph to be found or destroying the dockinggraph.
     * Overwrite this function to introduce custom repair function.
     */
    public void improveCosts() {
        if (structuralCosts != 0) {
            if (mySCM.myGSC.random.nextDouble() < removeDockingProbability) {
                if (! (dockingGraph.mainParts.elementAt(0) instanceof NullMainPart)) { // if the mainpart has no nodes it can not be destroyed..
                    dl.print(2, 
                             "\n" + myType.constraintName + 
                             " DockingGraph destroy.");
                    destroyDockingGraph();

                    return;
                }
            }

            PartFrameContainer dockingFrames = dockingGraph.getTypeFrames();
            dockingFrames.setAllFrames();

            TestGraph tg = chooseTestGraph();
            tg.setBiggestEmbedding();
            dl.print(2, "\n" + myType.constraintName + " TestGraph repair.");
            myType.myTestGraphRepairer.repair(tg, this);
            dockingFrames.resetAllFrames();

            //here no updateCosts!  the dockingGraph might have been destroyed
        }
    }


    /**
     * Marks all the nodes used in the docking graph
     */
    public void mark() {
        dockingGraph.mark();
    }


    /**
     * If the given node is one of a NAC, and one of the given edges would help
     * to fill the NAC this method returns true, else false. Used by
     * nodeChooser to determine whether to use the given node or not.
     * 
     * @param interpretation The tested node.
     * @param edge The edge that should be created.
     * 
     * @return true if edge helps create a NAC, else false.
     */
    public boolean nacCanUseNewEdge(Node interpretation, ConnectedNode edge) {
        return nacCanUseNewEdge(dockingGraph, interpretation, edge) || 
               ((actualTestGraph != null) && 
               nacCanUseNewEdge(actualTestGraph, interpretation, edge));
    }


    /**
     * Returns whether the destruction of the given node would disturb the
     * Constraint. Used by StructuralConstraintManagement to determine which
     * nodes to chage and which ones not. Does not guarantee, that the node is
     * not destroyed.
     * 
     * @param interpretation The tested node.
     * 
     * @return true if interpretations should rather not be removed, else
     *         false.
     */
    public boolean needsInterpretation(Node interpretation) {
        if (dockingGraph.containsInterpretation(interpretation) || 
                ((actualTestGraph != null) && 
                    actualTestGraph.containsInterpretation(interpretation))) {
            return false;
        } else {
            return true;
        }
    }


    /**
     * When the current testGraph has become invalid this function is called.
     */
    public void testGraphInvalid() {
        actualTestGraph = null;
    }


    /**
     * Unmarks all the nodes used in the docking graph.
     */
    public void unmark() {
        dockingGraph.unmark();
    }


    /**
     * Has one edge of the given node the same type as the given edge?
     * 
     * @param node  The node the edges of which should be compared with edges.
     * @param edge  The tested edge.
     * 
     * @return true if
     */
    boolean nodeContainsEdgeOfType(SCNode node, ConnectedNode edge) {
        Node     nodeInterpretation = node.peekInterpretation();
        Vector   nodeEdges   = node.getEdges();
        Iterator nodeEdgesIt = nodeEdges.iterator();

        while (nodeEdgesIt.hasNext()) {
            SCEdge scEdge = (SCEdge) nodeEdgesIt.next();

            if (scEdge.isSameType(edge)) {
                return true;
            }
        } //while

        return false;
    } //nodeContainsOneEdgeOfType(


    /**
     * Chooses one of the testGraphs, for repair. Choice is done by random with
     * a higher probability for less broken testGraphs. If no Testgraph was
     * invalid null is returned.
     * 
     * @return One TestGraph.
     */
    private TestGraph chooseTestGraph() {
        TestGraph result = null;

        if ((lastRepairedTestGraph != null) && 
                (mySCM.myGSC.random.nextDouble() < keepOnRepairingLastProbability)) {
            return lastRepairedTestGraph;
        }

        Collection col = invalidTestGraphs.values();
        Iterator   it = col.iterator();

        while (it.hasNext()) {
            result = (TestGraph) it.next();
        } //while

        lastRepairedTestGraph = result;

        return lastRepairedTestGraph;
    }


    /**
     * Destroys the DockingGraph of this instance.
     */
    private void destroyDockingGraph() {
        myType.myDockingGraphRepairer.destroy(dockingGraph, this);
    }


    /**
     * Does one of the nacs of the given graph have the given node as
     * interpretation, and does one of the edges create the NAC?
     * 
     * @param graph A TestGraph or a DockingGraph.
     * @param interpretation The tested node.
     * @param edge The edge that should be created.
     * 
     * @return true if the edge would create a new NAC match else false.
     */
    private boolean nacCanUseNewEdge(StructuralConstraintGraph graph, 
                                     Node interpretation, ConnectedNode edge) {
        Iterator nacIt = graph.NACs.iterator();

        while (nacIt.hasNext()) {
            StructuralConstraintGraphPart nac = (StructuralConstraintGraphPart) nacIt.next();
            Collection                    linkedNodes = 
                    nac.getInnerGraphLinkNodes();
            Iterator                      it = linkedNodes.iterator();

            while (it.hasNext()) {
                SCNode linkedNode = (SCNode) it.next();

                if (linkedNode.peekInterpretation() == interpretation) {
                    if (nodeContainsEdgeOfType(linkedNode, edge)) {
                        return true;
                    }
                }
            } //while
        } // while nacIt

        return false;
    }


    /**
     * Calculates the constraint's costs. Fills the invalidTestGraphs and
     * structuralCosts members with values. make shure that after a call to
     * updateCosts() the interpretations of all testGraphs are reset again!
     */
    private void updateCosts() {
        if (actualTestGraph == null) {
            actualTestGraph = testGraphs.getOneFittingTestGraph(
                                      dockingGraph.getTypeFrames());

            if (actualTestGraph == null) {
                invalidTestGraphs = testGraphs.getInvalidTestGraphs();
                structuralCosts   = testGraphs.getStructuralCosts();
            } else {
                structuralCosts = 0;
            }
        } else {
            structuralCosts = 0;
        }

        String n = this.toString();
        n = n.substring(50);
        dl.print(1, "\n" + n + "\n   SI Costs " + structuralCosts);
    }
}
