/*

 * StructuralConstraintManagement.java

 *

 * Created on 7. April 2002, 13:33

 */

package com.ai_center.dragonbreath.structuralConstraintHandling;



import com.ai_center.dragonbreath.*;

import com.ai_center.dragonbreath.GlobalSearchControl;

import com.ai_center.dragonbreath.constraintgraph.*;

import com.ai_center.dragonbreath.costs.*;

import com.ai_center.dragonbreath.graphRepair.*;

import com.ai_center.dragonbreath.loader.ScpLoader;

import com.ai_center.dragonbreath.structuralConstraintHandling.graph.*;

import com.ai_center.dragonbreath.structuralConstraintHandling.graph.testGraph.*;

import com.ai_center.dragonbreath.utilities.*;



import java.io.File;



import java.util.*;



import org.jdom.JDOMException;





/**

 * Manages the handling of Structural Constraints.

 * Structural Constraints are loaded and created here.

 * The GlobalSearchControl registers all structural changes, like addition and

 * removal of nodes to the StructuralConstraintManagement. With a call of

 * checkStructuralConstraints() all these changes are propagated to all structural constraints 

 * and if necessary a repair of the CSP-Graph is performed.

 * 

 * @version $Revision: 2.7 $

 * @author tim baermann

 */

public class StructuralConstraintManagement {

    // Static variables/initializers ******************************************



    /**

     * Should in the GraphSearch process for the Structural Constraints the

     * queue be sorted to use the fail-First principle?

     */

    public static boolean graphSearchSortQueueOn = false;



    // Instance variables *****************************************************



    /**

     * Standard DockingGraphRepairer, used by all structural constraints that

     * do not define a special one.

     */

    protected DockingGraphRepairer myDockingGraphRepairer;



    /**

     * Standard SCGraphRepairer, used by myDockingGraphRepairer and

     * myTestGraphRepairer

     */

    protected SCGraphRepairer mySCGraphRepairer;



    /**

     * Standard TestGraphRepairer, used by all structural constraints that do

     * not define a special one.

     */

    protected TestGraphRepairer myTestGraphRepairer;



    /** 

     * To write debug information. 

     */

    DebugLogger dl = new DebugLogger(true);



    /** 

     * Link to the GlobalSearchControl of this application. 

     */

    GlobalSearchControl myGSC;



    /**

     * Set by calling setNodeChooseMode(). Should the repair be done:

     * ALLWAYS_CREATE:  via allways creating new nodes,  NEVER_CREATE: via

     * never creating new nodes (only if no node of the given type exists)

     * HEURISTIC_CREATE:  via the heuristic of letting the ConstraintInstances

     * decide which nodes they do not want to be changed.

     */

    NodeChooseInterface myNodeChooser;



    /** 

     * Object that loads structural constraints. 

     */

    ScpLoader scpLoader;



    /**

     * The probability that addRandomGraphElement() adds a node and not an

     * edge. A value between 0 and 1 the higher the higher the connectivity of

     * the graph.

     */

    double randomCreateDensity = 0.5;



    /**

     * Used when adding random elements via addRandomGraphElement(). Determines

     * the number of different node-Types that can be added.

     */

    int numRandomNodeTypes = 8;



    /**

     * Used to time the graphsearch process. Increased by

     * checkStructuralConstraints(), can be retrieved and is reset by

     * getTiming().

     */

    long time = 0;



    /**

     * When during a repair nodes are changed, deleted nodes are stored with

     * the ConstraintType that caused the change in this map. Each entry is a

     * HashSet with the nodes and the type as key.

     */

    private HashMap deletedNodesForType = new HashMap();



    /**

     * When during a repair nodes are changed, expanded nodes are stored with

     * the ConstraintType that caused the change in this map. Each entry is a

     * HashSet with the nodes and the StructuralConstraintType as key.

     */

    private HashMap expandedNodesForType = new HashMap();



    /**

     * When during a repair nodes are changed, new nodes are stored with the

     * ConstraintType that caused the change in this map. Each entry is a

     * HashSet with the new nodes and the StructuralConstraintType as key.

     */

    private HashMap newNodesForType = new HashMap();



    /**

     * When during a repair nodes are changed, reduced nodes are stored with

     * the ConstraintType that caused the change in this map. Each entry is a

     * HashSet with the nodes and the StructuralConstraintType as key.

     */

    private HashMap reducedNodesForType = new HashMap();



    /**

     * Nodes of the problem graph that were deleted, and are waiting for

     * synchronization with structural constraints. To add Node call

     * registerDeletedNode().

     */

    private HashSet deletedNodes = new HashSet();



    /**

     * Nodes of the problem graph that got new edges. (which edges is saved in

     * expansions)

     */

    private HashSet expandedNodes = new HashSet();



    /** 

     * Nodes of the problem graph that were created. 

     */

    private HashSet newNodes = new HashSet();



    /**

     * Nodes of the problem graph that were reduced. To add Node call

     * addReducedNode().

     */

    private HashSet reducedNodes = new HashSet();



    /** 

     * All the structural constraint instances. 

     */

    private HashSet scInstances = new HashSet();



    /**

     * All the nodes currently managed by GlobalSearchControl, ordered by there

     * types.

     */

    private SetHashtable allNodes = new SetHashtable();



    /**

     * The Changes of the nodes in expandedNodes. Nodes of the problem graph

     * that have got new edges, as key and the change as values. To add Node

     * call addExpandedNode().

     */

    private SetHashtable expansions = new SetHashtable();



    /**

     * The reductions of the reducedNodes. The reductions of node n is a set

     * retrieved by reductions.get(n).

     */

    private SetHashtable reductions = new SetHashtable();



    /** 

     * All the StructuralConstraintTypes loaded. 

     */

    private Vector scTypes = new Vector();



    /**

     * Should the structuralConstraintInstances repair existing structural

     * errors?

     */

    private boolean repairOn = true;



    // Constructors ***********************************************************



    /**

     * Creates a new instance of StructuralConstraintManagement

     * 

     * @param gsc The GlobalSearchControl responsible for this

     *        StructuralConstraintManagement.

     */

    public StructuralConstraintManagement(GlobalSearchControl gsc) {

        myGSC               = gsc;

        scpLoader           = new ScpLoader(this, gsc);

        mySCGraphRepairer   = new SCGraphRepairer(this, gsc);

        myTestGraphRepairer = new TestGraphRepairer();

        myTestGraphRepairer.setConstructionParameters(this, gsc, 

                                                      mySCGraphRepairer);

        myDockingGraphRepairer = new DockingGraphRepairer();

        myDockingGraphRepairer.setConstructionParameters(this, gsc, 

                                                         mySCGraphRepairer);

        setNodeChooseMode(Registry.HEURISTIC_CREATE);



        //dl.switchOn(55);// uncomment for debug output

        //dl.switchOn(5);// uncomment for debug output

    }



    // Methods ****************************************************************



    /**

     * Returns myDockingGraphRepairer. Used by all Structural Constraints, that

     * do not implement special Repair mechanisms

     * 

     * @return myDockingGraphRepairer

     */

    public DockingGraphRepairer getDefaultDockingGraphRepairer() {

        return myDockingGraphRepairer;

    }





    /**

     * Returns myTestGraphRepairer. Used by all Structural Constraints, that do

     * not implement special Repair mechanisms

     * 

     * @return myTestGraphRepairer;

     */

    public TestGraphRepairer getDefaultTestGraphRepairer() {

        return myTestGraphRepairer;

    }





    /**

     * Returns the SCGraphRepairer.

     * 

     * @return mySCGraphRepairer

     */

    public SCGraphRepairer getGraphRepairer() {

        return mySCGraphRepairer;

    }





    /**

     * Sets the nodeChooser. Should the repair be done: <BR> ALLWAYS_CREATE:

     * via allways creating new nodes, <BR> NEVER_CREATE:  via never creating

     * new nodes (only if no node of the given type exists) <BR>

     * HEURISTIC_CREATE:  via the heuristic of letting the ConstraintInstances

     * decide which nodes they dont want to be changed.

     * 

     * @param nodeChooseMode  One of ALLWAYS_CREATE NEVER_CREATE

     *        HEURISTIC_CREATE

     */

    public void setNodeChooseMode(int nodeChooseMode) {

        switch (nodeChooseMode) {

            case Registry.ALLWAYS_CREATE:

                myNodeChooser = new AllwaysCreateNodeChooser(myGSC);



                break;



            case Registry.NEVER_CREATE:

                myNodeChooser = new NeverCreateNodeChooser(this, myGSC);



                break;



            case Registry.HEURISTIC_CREATE:

                myNodeChooser = new HeuristicNodeChooser(this, myGSC);



                break;



            default:

                throw new Error(

                        "Mode not found, choose one of: ALLWAYS_CREATE, NEVER_CREATE, HEURISTIC_CREATE");

        }



        mySCGraphRepairer.setNodeChooser(myNodeChooser);

    }





    /**

     * Returns all Nodes of the given type. Used especially by the

     * nodeChoosers.

     * 

     * @param type The nodeType as in Registry.nodeType.

     * 

     * @return Null, if none found, a Node if one found, or a Set of nodes if

     *         more then one.

     */

    public Object getNodesForType(Node type) {

        return allNodes.get(type);

    }





    /**

     * Sets the number of different nodeTypes the addRandomGraphElement()

     * function should add.

     * 

     * @param numNodeTypes A number between 1 and 12.

     */

    public void setNumberRandomNodeTypes(int numNodeTypes) {

        numRandomNodeTypes = numNodeTypes;



        if (numRandomNodeTypes < 12) {

            numRandomNodeTypes = 8;

        }

    }





    /**

     * Sets the probability that addRandomGraphElement() adds a new node

     * instead of a new edge between existing nodes.

     * 

     * @param dens  A value between 0 and 1 the higher the higher the

     *        connectivity of the graph.

     */

    public void setRandomCreateDensity(double dens) {

        randomCreateDensity = dens;

    }





    /**

     * Returns scInstances. Carefull! It returns the original data. Do not

     * change.

     * 

     * @return scInstances

     */

    public Set getSCInstances() {

        return scInstances;

    }





    /**

     * Returns the number of instances.

     * 

     * @return scInstances.size()

     */

    public int getSCInstancesSize() {

        return scInstances.size();

    }





    /**

     * Returns a list of all the Structural Constraints currently loaded.

     * 

     * @return scTypes; not a copy!

     */

    public Collection getStructuralConstraints() {

        return scTypes;

    }





    /**

     * Returns the total Structural Costs.

     * 

     * @return a positive integer, representing the structural costs of all

     *         structural constraints.

     */

    public int getStructuralCosts() {

        int      structuralCosts = 0;

        Iterator it = scTypes.iterator();



        while (it.hasNext()) {

            StructuralConstraintType t      = (StructuralConstraintType) it.next();

            long                     tCosts = t.getCosts(0);

            structuralCosts += tCosts;

        }



        return structuralCosts;

    }





    /**

     * Returns the time spent in checkStructuralConstraints() since last

     * getTiming(). Measured via java.lang.System.currentTimeMillis().

     * 

     * @return A positive number.

     */

    public long getTiming() {

        long result = time;

        time = 0;



        return result;

    }





    /**

     * Used for performance tests, adds any node or edge to the graph. Not at

     * all optimized. Use only for testing! Adds only nodes of type: Range,

     * AllDifferent, Variable, Sum, Even, LessEqual

     */

    public void addRandomGraphElement() {

        if (myGSC.random.nextDouble() < randomCreateDensity) {

            setNodeChooseMode(Registry.ALLWAYS_CREATE);

        } else {

            setNodeChooseMode(Registry.NEVER_CREATE);

        }



        int           nodeType      = chooseRandomNodeType();

        RepairPlan    plan;

        GraphRepairer graphRepairer = new GraphRepairer(myGSC, myNodeChooser);

        Vector        nodes         = new Vector(allNodes.values());



        if ((myGSC.random.nextDouble() > 0.5) && (nodes.size() > 0)) {

            int           nodeNr = myGSC.random.nextInt(nodes.size());

            Node          node = (Node) nodes.elementAt(nodeNr);

            ConnectedNode edge;



            if (node instanceof Variable) {

                while (nodeType == Registry.VARIABLE) {

                    nodeType = chooseRandomNodeType();

                }



                edge = new ConnectedNode(nodeType, Registry.ANY_EDGE, false);

            } else {

                edge = new ConnectedNode(Registry.VARIABLE, Registry.ANY_EDGE, 

                                         true);

            }



            plan = graphRepairer.addEdge(node, edge);

        } else {

            plan = graphRepairer.addNode(nodeType);

        }



        try { //some constraints still throw errors when connected with the wrong extension

            plan.repair();

        } catch (Error e) {

            e.printStackTrace();

        }

    }





    /**

     * Searches for new instances of structural constraints in the problem

     * graph and removes instances that have got invalid, and updates the

     * testGraphs. Before calling this function changes of the problem graph

     * have to be submitted via the registerNewNode(),registerDeletedNode()...

     * functions. Runs until Structural Error is solved!

     */

    public void checkStructuralConstraints() {

        dl.print(55, " \n checkStructuralConstraints");



        // take the timing of the structural constraint handling

        long    startTime = java.lang.System.currentTimeMillis();



        HashSet newNodesToCheck      = new HashSet(newNodes);

        HashSet deletedNodesToCheck  = new HashSet(deletedNodes);

        HashSet expandedNodesToCheck = new HashSet(expandedNodes);

        HashSet reducedNodesToCheck  = new HashSet(reducedNodes);

        Object  dummySCType          = new Object();



        // the nodes were not really changed by the last type.. 

        // but this way they are preserved the whole first iteration of checks

        if (scTypes.size() != 0) {

            registerNodesChangedByType(scTypes.lastElement());



            /*

             * Each type has to test the changed nodes, and moreover the nodes changed by repair of all types.

             * After each test there might again be changed nodes.

             * if there are 3 structural constraint types T1, T2, T3  and the initial changed nodes

             * are a and the changed nodes by Ti are ti then the tests should be:

             *

             *  T1  a                           < type T1 tests only the nodes changed at method call.

             *  T2  a, t1                       < type T2 tests the initial nodes + the nodes changed by T1

             *  T3  a, t1, t2                   < type T3 tests the initial nodes + the nodes changed by T1 + T2

             *

             *  T1     t1, t2, t3

             *  T2         t2, t3, t1

             *  T3             t3, t1, t2

             * ...

             * Until no more changed nodes exist.

             */

            boolean ready = allTypesCheckedWithAllNodes();



            while (! ready) {

                Iterator it = scTypes.iterator();



                while (it.hasNext()) {

                    StructuralConstraintType type = (StructuralConstraintType) it.next();

                    fillNodeContainers(newNodesToCheck, deletedNodesToCheck, 

                                       expandedNodesToCheck, 

                                       reducedNodesToCheck);

                    type.checkChangedNodes(newNodesToCheck, deletedNodesToCheck, 

                                           expandedNodesToCheck, 

                                           reducedNodesToCheck);



                    registerNodesChangedByType(type);

                } //while it



                PartFrameContainer.allContainerCheckDeletedNodes(

                        deletedNodesToCheck); //clean up.



                updateStructuralCosts();

                ready = allTypesCheckedWithAllNodes();

            }

        }



        time += (java.lang.System.currentTimeMillis() - startTime);

    } //checkStructuralConstraints()





    /**

     * Clears all the containers of created or changed or deleted nodes.

     */

    public void clearNodeContainers() {

        expandedNodes.clear();

        expansions.clear();

        newNodes.clear();

        reducedNodes.clear();

        reductions.clear();

        deletedNodes.clear();

    }





    /**

     * Removes all StructuralConstraint instances.

     */

    public void clearSCInstances() {

        Iterator it = scTypes.iterator();



        while (it.hasNext()) {

            StructuralConstraintType t = (StructuralConstraintType) it.next();

            t.clearInstances();

        } //while it         



        scInstances.clear();

    }





    /**

     * Resets all the Containers of the StructuralConstraintManagement.

     */

    public void clearSCM() {

        clearSCTypes();

        clearNodeContainers();

        allNodes.clear();

    }





    /**

     * Removes all Structural Constraint Types.

     */

    public void clearSCTypes() {

        scTypes.clear();

        scInstances.clear();

    }





    /**

     * Creates a structural constraint type.

     * 

     * @param scGraphFactory A Factory that can produce the given type.

     * @param name the constraints name

     * 

     * @return the new StructuralConstraint type.

     */

    public boolean createConstraintType(java.lang.String name, 

                                        SCGraphFactory scGraphFactory) {

        if (! existsConstraintWithName(name)) {

            int                      costID        = 0;

            StructuralConstraintType newConstraint = new StructuralConstraintType(

                                                             name, 

                                                             scGraphFactory, 

                                                             this);

            CostMapping costMapping = myGSC.getSelectedCostMapping();

            myGSC.getStructuralCostFunction()

                 .addConstraint(newConstraint, costID, costMapping, null);

            scTypes.add(newConstraint);



            HashSet      dummyHashSet      = new HashSet();

            SetHashtable dummySetHashtable = new SetHashtable();

            newConstraint.checkChangedNodes(allNodes.values(), dummyHashSet, 

                                            dummyHashSet, dummyHashSet);



            return true;

        } else {

            java.lang.System.out.print("Constraint allready loaded.");



            return false;

        }

    }





    /**

     * Should in the GraphSearch process for the Structural Constraints the

     * queue be sorted to use the fail-First principle? Changes the Registry

     * Value.

     * 

     * @param val

     */

    public void graphSearchSortQueueOn(boolean val) {

        graphSearchSortQueueOn = val;

    }





    /**

     * Loads the structural contstraints in the given file.

     * 

     * @param filename  The name of the file.

     * 

     * @return true if the file was loaded.

     */

    public boolean loadStructralConstraints(String filename) {

        return loadStructralConstraints(new File(filename));

    }





    /**

     * Loads the structural contstraints in the given file.

     * 

     * @param file  The file that should be loaded.

     * 

     * @return true if the file was loaded.

     */

    public boolean loadStructralConstraints(File file) {

        try {

            scpLoader.load(file);

        } catch (JDOMException ex) {

            if (ex.getCause() != null) {

                ex.getCause().printStackTrace();

            } else {

                ex.printStackTrace();

            }



            return false;

        } catch (Exception exx) {

            exx.printStackTrace();



            return false;

        }



        return true;

    }





    /**

     * Prints how many docking graph instances were found.

     * 

     * @param numberOfInstances How many instances should have been found?

     * 

     * @throws Error If the numberOfInstances != number of found instances.

     */

    public void printDockingStatus(int numberOfInstances) {

        int      matchesFound = 0;

        Iterator it = scTypes.iterator();



        while (it.hasNext()) {

            StructuralConstraintType t = (StructuralConstraintType) it.next();

            matchesFound += t.constraintInstances.size();

            System.out.print("Number of Instances   : " + 

                             t.constraintInstances.size() + " / " + 

                             numberOfInstances + "\n");

        }



        if (matchesFound < numberOfInstances) {

            throw new Error("Less instances found then anticipated.");

        } else if (matchesFound > numberOfInstances) {

            throw new Error("More instances found then anticipated.");

        }

    }





    /**

     * Prints the structural Costs.

     * 

     * @param expectedCosts  The expected costs.

     * 

     * @throws Error If the real Costs are 0 and the expectedCosts not, or

     *         reverse.

     */

    public void printTestStatus(int expectedCosts) {

        int      matchesFound = 0;



        long     structuralCosts = 0;

        Iterator it              = scTypes.iterator();



        while (it.hasNext()) {

            StructuralConstraintType t      = (StructuralConstraintType) it.next();

            long                     tCosts = t.getCosts(0);

            structuralCosts += tCosts;

            System.out.print("Structural Costs: " + tCosts + " / " + 

                             expectedCosts);

        }



        if ((expectedCosts > 0) && (structuralCosts == 0)) {

            throw new Error("Costs too low.");

        } else if ((expectedCosts == 0) && (structuralCosts > 0)) {

            throw new Error("Costs too high.");

        }

    }





    /**

     * Adds a deleted node for check on structural consistency.

     * 

     * @param deletedNode The node deleted.

     */

    public void registerDeletedNode(Node deletedNode) {

        //remove node from allnodes

        allNodes.remove(Registry.getNodeType(deletedNode), deletedNode);



        if (! newNodes.contains(deletedNode)) {

            deletedNodes.add(deletedNode);



            //else the node would have been added this round, the deletion would not affect anything.

        }



        if (deletedNode instanceof Variable) {

            dl.print(5, "\n     " + deletedNode + " removed");

        }



        newNodes.remove(deletedNode);



        reducedNodes.remove(deletedNode);

        reductions.removeKey(deletedNode);



        expandedNodes.remove(deletedNode);

        expansions.removeKey(deletedNode);

    }





    /**

     * A node that got added connections, for check on structural consistency.

     * 

     * @param cn  The connection that was added to the node.

     * @param expandedNode  A node that got a new connection.

     */

    public void registerExpandedNode(Node expandedNode, ConnectedNode cn) {

        if (! newNodes.contains(expandedNode) && 

                ! newNodes.contains(cn.node)) {

            expandedNodes.add(expandedNode);

        }



        //else the node is already registered as new node it is checked there and doesnt have to be added her 

    }





    /**

     * Adds nodes that were newly created for check on structural consistency.

     * 

     * @param newNode  A node that was recently created.

     */

    public void registerNewNode(Node newNode) {

        newNodes.add(newNode);



        if (newNode instanceof Variable) {

            dl.print(5, "\n     " + newNode + " added");

        }





        // add node to allNodes

        allNodes.add(Registry.getNodeType(newNode), newNode);

    }





    /**

     * Adds a node that lost connections for check on structural consistency.

     * 

     * @param reducedNode The node that was reduced.

     * @param cn  The connection that was removed from the node.

     */

    public void registerReducedNode(Node reducedNode, ConnectedNode cn) {

        if (! newNodes.contains(reducedNode)) {

            reducedNodes.add(reducedNode);

        }

    }





    /**

     * Used for performance tests, removes any node from the graph. Not at all

     * optimized. Use only for testing!

     */

    public void removeRandomGraphElement() {

        if (myGSC.random.nextDouble() < randomCreateDensity) {

            setNodeChooseMode(Registry.ALLWAYS_CREATE);

        } else {

            setNodeChooseMode(Registry.NEVER_CREATE);

        }



        RepairPlan    plan;

        GraphRepairer graphRepairer = new GraphRepairer(myGSC, myNodeChooser);

        Vector        v             = new Vector(allNodes.values());

        int           chosen        = myGSC.random.nextInt(v.size());

        plan = graphRepairer.removeNode((Node) v.elementAt(chosen));

        plan.repair();

    }





    /**

     * Removes the given Structural Constraint.

     * 

     * @param sc  The Constraint that should be removed.

     */

    public void removeStructuralConstraint(StructuralConstraintType sc) {

        scTypes.remove(sc);

        sc.clearInstances();

    }





    /**

     * Returns true if structural errors are repaired, or else false.

     * 

     * @return repairOn

     */

    public boolean repairOn() {

        return repairOn;

    }





    /**

     * Switches the repair of the graph off.

     */

    public void switchOffRepair() {

        repairOn = false;

    }





    /**

     * Switches the repair of the graph on.

     */

    public void switchOnRepair() {

        repairOn = true;

    }





    /**

     * Are there still nodes to be checked for some StructuralConstraintTypes?

     * 

     * @return false if there are nodes unchecked, else true.

     */

    protected boolean allTypesCheckedWithAllNodes() {

        if (newNodesForType.size() > 0) {

            return false;

        } else if (deletedNodesForType.size() > 0) {

            return false;

        } else if (expandedNodesForType.size() > 0) {

            return false;

        } else if (reducedNodesForType.size() > 0) {

            return false;

        } else {

            return true;

        }

    }





    /**

     * After each types checkChangedNodes() there might exist new unchecked

     * nodes. This function updates the newNodes, deletedNodes, expandedNodes

     * and reducedNodes containers for the next StructuralConstraintType.

     * 

     * @param newN Container where new nodes that were not yet checked are

     *        added.

     * @param delN Container where deleted nodes are added.

     * @param expN Container where expanded nodes are added.

     * @param redN Container where reduced nodes are added.

     */

    protected void fillNodeContainers(HashSet newN, HashSet delN, HashSet expN, 

                                      HashSet redN) {

        Iterator it = scTypes.iterator();



        while (it.hasNext()) {

            Object type = it.next();

            Set    s = (HashSet) newNodesForType.get(type);



            if (s != null) {

                newN.addAll(s);

            }



            s = (HashSet) deletedNodesForType.get(type);



            if (s != null) {

                delN.addAll(s);

            }



            s = (HashSet) expandedNodesForType.get(type);



            if (s != null) {

                expN.addAll(s);

            }



            s = (HashSet) reducedNodesForType.get(type);



            if (s != null) {

                redN.addAll(s);

            }

        }

    }





    /**

     * Takes the changed nodes from the newNodes, deletedNodes... containers,

     * and registers them as added by the given type. Resets the former nodes

     * containers.

     * 

     * @param type The constraintType that caused the actual nodes changes.

     */

    protected void registerNodesChangedByType(Object type) {

        if (newNodes.isEmpty()) {

            newNodesForType.remove(type);

        } else {

            newNodesForType.put(type, newNodes);

            newNodes = new HashSet();

        }



        if (deletedNodes.isEmpty()) {

            deletedNodesForType.remove(type);

        } else {

            deletedNodesForType.put(type, deletedNodes);

            deletedNodes = new HashSet();

        }



        if (expandedNodes.isEmpty()) {

            expandedNodesForType.remove(type);

        } else {

            expandedNodesForType.put(type, expandedNodes);

            expandedNodes = new HashSet();

        }



        if (reducedNodes.isEmpty()) {

            reducedNodesForType.remove(type);

        } else {

            reducedNodesForType.put(type, reducedNodes);

            reducedNodes = new HashSet();

        }

    }





    /**

     * Return one of Range, AllDifferent, Variable, Sum, Even, LessEqual

     * SUM_CONSTRAINT least probable.

     * 

     * @return A number that stands for a Registry nodeTypeID.

     */

    int chooseRandomNodeType() {

        /*if (myGSC.random.nextDouble() > 0.05) {

           return Registry.SUM_CONSTRAINT;    

           }*/

        int r = myGSC.random.nextInt(numRandomNodeTypes);



        switch (r) {

            case 0:

                return Registry.VARIABLE;



            case 1:

                return Registry.INTEGER_CONSTRAINT;



            case 2:

                return Registry.RANGE_CONSTRAINT;



            case 3:

                return Registry.EVEN_CONSTRAINT;



            case 4:

                return Registry.ALLDIFFERENT_CONSTRAINT;



            case 5:

                return Registry.LESSEQUAL_CONSTRAINT;



            case 6:

                return Registry.PROPORTION_CONSTRAINT;



            case 7:

                return Registry.SUM_CONSTRAINT;

        }



        return 1;

    }





    static void pause(long millis) {

        long a = java.lang.System.currentTimeMillis();

        long c = a;



        while (c < (a + millis)) {

            c = java.lang.System.currentTimeMillis();

        }

    }





    /**

     * Adds a new StructuralConstraintInstance to the list of instances. Called

     * by the type of the instance.

     * 

     * @param instance  The new instance.

     */

    void registerSCInstance(StructuralConstraintInstance instance) {

        scInstances.add(instance);

    }





    /**

     * Removes a StructuralConstraintInstance from the list of instances.

     * Called by the type of the instance.

     * 

     * @param instance  The deleted instance.

     */

    void removeSCInstance(StructuralConstraintInstance instance) {

        scInstances.remove(instance);

    }





    /**

     * Creates a Collection with all the broken structural constraint

     * instances. The bigger the error of an instance the higher the

     * probability that it will come first.

     * 

     * @return A collection of the broken instances.

     */

    private Collection createBrokenInstancesPermutation() {

        SortedMap result = new TreeMap();

        Random    random = new Random();



        Iterator  it = scInstances.iterator();



        while (it.hasNext()) {

            StructuralConstraintInstance inst  = (StructuralConstraintInstance) it.next();

            long                         costs = inst.getCosts();



            if (costs > 0) {

                int index = random.nextInt(

                                    (int) (java.lang.Integer.MAX_VALUE - 

                                        (costs * 10))); // * 10 to spread the data more  





                // 10000 - x   to reverse order.. 10000 is a big number, that should never be exceeded by an error..

                result.put(new IntValue(index), inst);

            }

        }



        return result.values();

    }





    /**

     * Does one of the StructuralConstraints have the given name?

     * 

     * @param name name of the Constraint.

     * 

     * @return Name found.

     */

    private boolean existsConstraintWithName(java.lang.String name) {

        Iterator it = scTypes.iterator();



        while (it.hasNext()) {

            StructuralConstraintType t = (StructuralConstraintType) it.next();



            if (t.constraintName.equals(name)) {

                return true;

            }

        }



        return false;

    }





    /**

     * Updates the costs of all StructuralConstraints.

     */

    private void updateStructuralCosts() {

        Iterator it = scTypes.iterator();



        while (it.hasNext()) {

            StructuralConstraintType type = (StructuralConstraintType) it.next();

            type.updateCosts();

        }

    }

} //class

