/*
 * StructuralConstraintType.java
 *
 * Created on 7. April 2002, 15:23
 */
package com.ai_center.dragonbreath.structuralConstraintHandling;

import com.ai_center.dragonbreath.*;
import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.costs.*;
import com.ai_center.dragonbreath.graphRepair.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.dockingGraph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.testGraph.*;
import com.ai_center.dragonbreath.utilities.*;

import java.lang.String;

import java.util.*;


/**
 * Represents one Structural constraint type, with its docking- and its testGraphs. 
 * A new StructuralConstraintInstance is created when a new match for the 
 * dockingGraph is found.
 * 
 * @version $Revision: 2.6 $
 * @author tim baermann
 */
public class StructuralConstraintType implements Constraint, NodeChangedInterface {
    // Instance variables *****************************************************

    /** 
     * CostFunction for this Constraint. 
     */
    CostFunctionReference myCostFunctionReference;

    /** 
     * Used to output debug information 
     */
    DebugLogger dl = new DebugLogger(true);

    /** 
     * Called by instances for destroying the dockingGraph. 
     */
    DockingGraphRepairer myDockingGraphRepairer;

    /**
     * Instances that got invalid are first entered here via
     * markInstanceForRemoval() and later removed.
     */
    HashSet deletedInstances = new HashSet();

    /** 
     * The factory, that produces the docking and the testgraphs. 
     */
    SCGraphFactory myGraphFactory;

    /** 
     * The dockingGraph of this type 
     */
    SCTDockingGraph dockingGraph;

    /** 
     * The constraints name; 
     */
    String constraintName;

    /** 
     * The management of this structural constraint. 
     */
    StructuralConstraintManagement mySCM;

    /** 
     * Called by instances for repairing a broken testgraph. 
     */
    TestGraphRepairer myTestGraphRepairer;

    /** 
     * The testGraph of this Structural Constraint type. 
     */
    TestGraphTypesContainer testGraphs;

    /** 
     * The instances of this structural constraint. 
     */
    Vector constraintInstances = new Vector();

    /** 
     * The unique id of this constraint type. 
     */
    int typeID;

    /** 
     * The momentary costs. They are computed by updateCosts(). 
     */
    long structuralCosts;

    // Constructors ***********************************************************

    /**
     * Creates a new instance of StructuralConstraintType
     * 
     * @param name The name of the constraintType
     * @param graphFactory The factory that produced this type and can produce
     *        instances.
     * @param scm The StructuralConstraintManagement that is responsible for
     *        this type
     */
    public StructuralConstraintType(String name, SCGraphFactory graphFactory, 
                                    StructuralConstraintManagement scm) {
        mySCM          = scm;
        myGraphFactory = graphFactory;
        constraintName = name;
        dockingGraph   = myGraphFactory.getTypeDockingGraph();
        testGraphs     = myGraphFactory.getTypeTestGraphs();
        dockingGraph.setSCManagementAndSCType(mySCM, this);
        testGraphs.setSCManagement(mySCM);

        if (graphFactory.dockingGraphRepairerClassName != null) {
            try {
                myDockingGraphRepairer = (DockingGraphRepairer) java.lang.Class.forName(
                                                                        graphFactory.dockingGraphRepairerClassName)
                                                                               .newInstance();
                myDockingGraphRepairer.setConstructionParameters(mySCM, 
                                                                 mySCM.myGSC, 
                                                                 scm.getGraphRepairer());
            } catch (java.lang.Exception e) {
                if (e instanceof java.lang.ClassNotFoundException) {
                    throw new Error("Class not found: " + e.getMessage());
                } else {
                    throw new Error("Instance not created: " + 
                                    e.getMessage());
                }
            }
        } else {
            myDockingGraphRepairer = mySCM.getDefaultDockingGraphRepairer();
        }

        if (graphFactory.testGraphRepairerClassName != null) {
            try {
                myTestGraphRepairer = (TestGraphRepairer) java.lang.Class.forName(
                                                                  graphFactory.testGraphRepairerClassName)
                                                                         .newInstance();
                myTestGraphRepairer.setConstructionParameters(mySCM, 
                                                              mySCM.myGSC, 
                                                              scm.getGraphRepairer());
            } catch (java.lang.Exception e) {
                if (e instanceof java.lang.ClassNotFoundException) {
                    throw new Error("Class not found: " + e.getMessage());
                } else {
                    throw new Error("Instance not created: " + 
                                    e.getMessage());
                }
            }
        } else {
            myTestGraphRepairer = mySCM.getDefaultTestGraphRepairer();
        }

        //dl.switchOn(1);// uncomment for debug output
    }

    // Methods ****************************************************************

    /**
     * Sets a cost function reference for a constraint.
     * 
     * @param costID  The ID of the costs.
     * @param costFunctionReference  The reference to be set for the
     *        constraint.
     */
    public void setCostFunctionReference(int costID, 
                                         CostFunctionReference costFunctionReference) {
        myCostFunctionReference = costFunctionReference;
    }


    /**
     * Queries a cost function reference of a StructuralConstraintType.
     * 
     * @param costID  The ID of the related costs.
     */
    public CostFunctionReference getCostFunctionReference(int costID) {
        return myCostFunctionReference;
    }


    /**
     * Queries the name of a specific cost type of the constraint.
     * 
     * @param costID  The ID of the costs for which the name is queried.
     */
    public String getCostName(int i) {
        return "a";
    }


    /**
     * Queries specific costs of the constraint.
     * 
     * @param costID  The ID of the costs to be retrieved.
     */
    public long getCosts(int costID) {
        return structuralCosts;
    } //getCosts()


    /**
     * Returns the frames of the current embedding of the dockingGraph. Called
     * by constructor of StructuralConstraintInstance.
     * 
     * @return Frame of current DockingGraph embedding.
     */
    public PartFrameContainer getDockingGraphFrames() {
        return dockingGraph.createInterGraphFrames();
    } // getDockingGraphFrame


    /**
     * Returns the name of the constraint.
     * 
     * @return A string with the name.
     */
    public String getName() {
        return constraintName;
    }


    /**
     * Queries the number of different costs available for the constraint.
     * 
     * @return how many different types of costs are there?
     */
    public int getNumberOfCosts() {
        return 1;
    }


    /**
     * Returns the number of instances of the StructuralConstraintType.
     * 
     * @return constraintInstances.size()
     */
    public int getNumberOfInstances() {
        return constraintInstances.size();
    }


    /**
     * Returns a constraint's type ID. This must be a unique number for each
     * constraint type. However, different kinds of constraint types (like
     * conventional and structural constraint types) may use the same numbers.
     * 
     * @return A unique number.
     */
    public int getTypeID() {
        return typeID;
    }


    /**
     * Checks nodes that were newly created for creation or destruction of
     * StructuralConstraintInstances.
     * 
     * @param deletedNodes Nodes that were deleted.
     * @param reducedNodes Nodes that lost connections.
     * @param newNodes Nodes that were recently created.
     * @param expandedNodes Expanded Nodes
     */
    public void checkChangedNodes(Set newNodes, Set deletedNodes, 
                                  Set expandedNodes, Set reducedNodes) {
        dockingGraph.checkChangedNodes(newNodes, deletedNodes, expandedNodes, 
                                       reducedNodes);

        Collection newDockingGraphInstances = dockingGraph.getNewGraphInstances();
        Iterator   it = newDockingGraphInstances.iterator();

        while (it.hasNext()) {
            SCIDockingGraph d = (SCIDockingGraph) it.next();
            createInstance(d);
        } //while

        testGraphs.checkUnanchoredParts(newNodes, deletedNodes, expandedNodes, 
                                        reducedNodes);

        StructuralConstraintInstance instance;
        it = constraintInstances.iterator();

        while (it.hasNext()) {
            instance = (StructuralConstraintInstance) it.next();
            instance.checkChangedNodes(newNodes, deletedNodes, expandedNodes, 
                                       reducedNodes);

            if (mySCM.repairOn() && ! deletedInstances.contains(instance)) {
                instance.improveCosts();
            }

            if (deletedInstances.contains(instance)) {
                it.remove();
                mySCM.removeSCInstance(instance);
            }
        } //while it

        deletedInstances.clear();
    } //checkChangedNodes()


    /**
     * Removes all StructuralConstraint instances.
     */
    public void clearInstances() {
        Iterator it = constraintInstances.iterator();

        while (it.hasNext()) {
            StructuralConstraintInstance inst = (StructuralConstraintInstance) it.next();
            removeSCInstance(inst);
        }

        dockingGraph.clear();
    }


    /**
     * Creates a new StructuralConstraintInstance with the given docking graph
     * and registeres it at the SCM.
     * 
     * @param docking  The dockingGraph of the new instance.
     */
    public void createInstance(SCIDockingGraph docking) {
        dl.print(1, "\nnew Constraint instance:" + constraintName);

        StructuralConstraintInstance instance;
        String                       customInstanceClassName = 
                myGraphFactory.getCustomInstanceClassName();

        if (customInstanceClassName != null) {
            try {
                instance = (StructuralConstraintInstance) java.lang.Class.forName(
                                                                  customInstanceClassName)
                                                                         .newInstance();
            } catch (java.lang.Exception e) {
                throw new Error(e.getMessage());
            }
        } else {
            instance = new StructuralConstraintInstance();
        }

        instance.setConstructionParameters(this, docking, testGraphs, 
                                           myGraphFactory.millisForRepair, 
                                           myGraphFactory.removeNodeProbability, 
                                           myGraphFactory.removeDockingProbability, 
                                           myGraphFactory.keepOnRepairingLastProbability);
        constraintInstances.add(instance);
        mySCM.registerSCInstance(instance);
    }


    /**
     * Does nothing, the structural constraints are repaired via
     * StructuralConstraintManagement The StructuralConstraintType is only
     * registered with the CostFunction to update the structural costs.
     * 
     * @param costId The cost type that should be corrected
     */
    public void improveCosts(int costId) {
    }


    /**
     * Performs changes of the CSP in order to improve the constraint's cost.
     * 
     * @param costID  The ID of the costs to be improved.
     */
    public void initializeCostType(int costID) {
    }


    /**
     * Does nothing.
     */
    public void initializeInternalStructures() {
    }


    /**
     * When during checkChangedNodes(), or during improveCosts() an instance is
     * destroyed, it registers itself via this function as removed. It is not
     * removed straight away from the constraintInstances list, as that would
     * result in a ConcurrentModificationException. The deletedInstances are
     * removed at  the end of checkChangedNodes().
     * 
     * @param instance The removed instance.
     */
    public void markInstanceForRemoval(StructuralConstraintInstance instance) {
        deletedInstances.add(instance);
    }


    /**
     * Marks all nodes used in instances of this structuralConstraint.
     */
    public void markInstances() {
        Iterator it = constraintInstances.iterator();

        while (it.hasNext()) {
            StructuralConstraintInstance inst = (StructuralConstraintInstance) it.next();
            inst.mark();
        }
    }


    /**
     * Returns the number of instances found for this Structural Constraint
     * Type.
     * 
     * @return constraintInstances.size()
     */
    public int numInstances() {
        return constraintInstances.size();
    }


    /**
     * Removes a StructuralConstraintInstance from the list of instances, and
     * from the SCM.
     * 
     * @param instance  The deleted instance.
     */
    public void removeSCInstance(StructuralConstraintInstance instance) {
        constraintInstances.remove(instance);
        mySCM.removeSCInstance(instance);
    }


    /**
     * Unmarks all nodes used in instances of this structuralConstraint.
     */
    public void unmarkInstances() {
        Iterator it = constraintInstances.iterator();

        while (it.hasNext()) {
            StructuralConstraintInstance inst = (StructuralConstraintInstance) it.next();
            inst.unmark();
        }
    }


    /**
     * Calculates the constraint's costs.
     */
    public void updateCosts() {
        structuralCosts = 0;

        Iterator it = constraintInstances.iterator();

        while (it.hasNext()) {
            StructuralConstraintInstance i = (StructuralConstraintInstance) it.next();
            structuralCosts += i.getCosts();
        }

        myCostFunctionReference.updateGlobalCosts();
    }
}