/*
 * GraphPartFrame.java
 *
 * Created on 22. April 2002, 21:10
 */
package com.ai_center.dragonbreath.structuralConstraintHandling.graph;

import com.ai_center.dragonbreath.constraintgraph.Node;
import com.ai_center.dragonbreath.constraintgraph.Variable;

import java.math.BigInteger;

import java.util.*;
import java.util.Map.Entry;


/**
 * When a subgraph was found, but can not be stored (eg. not all parts are
 * found, or not all ACs are fullfilled), its location is stored in this
 * structure. 
 * 
 * @version $Revision: 1.3 $
 * @author tim baermann
 */
public class GraphPartFrame {
    // Instance variables *****************************************************

    /** 
     * SCNodes as keys to Nodes 
     */
    public HashMap interpretation2node = new HashMap();

    /** 
     * The part the interpretations are for. 
     */
    public StructuralConstraintGraphPart myGraphPart;

    /**
     * The hashcode is a hash over the nodes in the object. It is used to find
     * doublettes.
     */
    BigInteger hashCode = null;

    /** 
     * The number of nodes that were used to create the hashcode. 
     */
    int numberOfNodesInHashCode;

    /** 
     * Nodes as keys to SCNodes 
     */
    private HashMap node2interpretation = new HashMap();

    // Constructors ***********************************************************

    /**
     * Constructor.
     * 
     * @param part  The part the interpretations are for.
     */
    GraphPartFrame(StructuralConstraintGraphPart part) {
        myGraphPart = part;
    }

    // Methods ****************************************************************

    /**
     * Sets all the interpretations stored in node2interpretation to their
     * nodes.
     * 
     * @return Could the interpretations be set?
     */
    public boolean setInterpretations() {
        boolean  result = true;
        Set      keys = interpretation2node.keySet();
        Node     key  = null;
        SCNode   val;
        Iterator it   = keys.iterator();

        while (it.hasNext() && result) {
            key    = (Node) it.next();
            val    = (SCNode) interpretation2node.get(key);
            result = val.setInterpretationExtern(key); // set the extern interpretation for the graph match algorithm not to think he allready visited the interpretated nodes.
        } //while

        myGraphPart.usedGraphPartFrame = this;

        if (! result) {
            resetInterpretations();
        }

        return result;
    } //setInterpretations()


    /**
     * Checks if the given interpretations for the nodes coincide with the
     * interpretations allready registered for the nodes.
     * 
     * @return Is the Frame compatible withe the current interpretation?
     */
    public boolean checkInterpretations() {
        boolean  result = true;
        Set      keys = interpretation2node.keySet();
        Node     n;
        SCNode   scn;
        Iterator it   = keys.iterator();

        while (it.hasNext() && result) {
            n   = (Node) it.next();
            scn = (SCNode) interpretation2node.get(n);

            if (! scn.checkInterpretation(n)) {
                return false;
            }
        } //while

        return true;
    } //checkInterpretations()


    /**
     * Adds a node to the stored interpretations.
     * 
     * @param node  The node of the StructuralConstraintGraphPart.
     * @param interpretation  The problem graph node that serves as
     *        interpretation for node.
     */
    public void enterNode(SCNode node, Node interpretation) {
        interpretation2node.put(interpretation, node);
        node2interpretation.put(node, interpretation);
    }


    /**
     * Resets the interpretations for all nodes stored in this GraphPartFrame.
     */
    public void resetInterpretations() {
        Collection scnodes = interpretation2node.values();
        SCNode     scn;
        Iterator   it = scnodes.iterator();

        while (it.hasNext()) {
            scn = (SCNode) it.next();
            scn.resetInterpretationExtern();
        } //while
    }


    /**
     * How many nodes are stored in the frame.
     * 
     * @return number of nodes.
     */
    public int size() {
        return interpretation2node.size();
    }


    /**
     * Prints the interpretations stored in the frame.
     * 
     * @return The Describtion.
     */
    public String toString() {
        String   result = new String();
        Iterator it = interpretation2node.entrySet().iterator();

        while (it.hasNext()) {
            Entry  entry = (Entry) it.next();
            Node   n   = (Node) entry.getKey();
            SCNode scn = (SCNode) entry.getValue();
            result = result + "Node ";
            result = result.concat(java.lang.String.valueOf(scn.hashCode())) + 
                     ": ";
            result = result.concat(java.lang.String.valueOf(n.getTypeName())) + 
                     "  ";

            if (n instanceof Variable) {
                result = result + " = " + ((Variable) n).getName() + "  ";
            }
        } //while it

        result = result + " \n";

        return result;
    }


    /**
     * Returns a hashcode over the nodes stored. Uses both, the nodes stored
     * and the interpretations stored with the nodes.
     * 
     * @return A unique hashcode representing the graphInterpretation.
     */
    BigInteger getHashCode() {
        if (((hashCode != null) && 
                    (numberOfNodesInHashCode != interpretation2node.size())) || 
                (hashCode == null)) { //nodes added since hashCode was created last time
            computeHashCode();
        }

        return hashCode;
    }


    /**
     * Computes the hashcode from the nodes stored.
     */
    void computeHashCode() {
        String hashAsString = new String();
        numberOfNodesInHashCode = interpretation2node.size();
        hashAsString            = hashAsString.concat(java.lang.String.valueOf(
                                                              myGraphPart.hashCode()));

        Iterator it             = interpretation2node.entrySet().iterator();

        while (it.hasNext()) {
            Entry  entry = (Entry) it.next();
            Node   n   = (Node) entry.getKey();
            SCNode scn = (SCNode) entry.getValue();

            if (n == null) {
                throw new Error("Variable without interpretation found");
            }

            hashAsString = hashAsString.concat(java.lang.String.valueOf(
                                                       n.hashCode()));
            hashAsString = hashAsString.concat(java.lang.String.valueOf(
                                                       scn.hashCode()));
        } //while it

        hashCode = new BigInteger(hashAsString);
    } //computeHashcode


    /**
     * Does the frame contain the given interpretation for any node?
     * 
     * @param interpretation  The searched interpretation.
     * 
     * @return Is the interpretation contained?
     */
    boolean containsInterpretation(Node interpretation) {
        return interpretation2node.containsKey(interpretation);
    }


    /**
     * Does the frame contain the nodes with the same interpretation? If a node
     * in nodes has no interpretation assigned it is not taken into account.
     * 
     * @param nodes  The searched SCNodes.
     * 
     * @return Does the frame contain the nodes with the same interpretation?
     */
    boolean containsNodes(Set nodes) {
        boolean  result           = true;
        boolean  foundOneGoodNode = false;
        Iterator it               = nodes.iterator();

        while (it.hasNext() && result) {
            SCNode scn = (SCNode) it.next();

            if (scn.interpretation != null) {
                SCNode n = (SCNode) interpretation2node.get(scn.interpretation);

                if (n == null) {
                    Node i = (Node) node2interpretation.get(scn); // check if the node is not stored with a different interpretation
                    result = (i == null); // if i != null it is not scn.interpretation, or else we would not be in this branch
                } else {
                    result = (n == scn) || (n.link == scn) || 
                             (scn.link == n);
                    foundOneGoodNode = result;
                }
            } // else the interpretation doesnt matter
        }

        return result && foundOneGoodNode;
    }
} // GraphPartInterpretations
