/*
 * GraphPartStack.java
 *
 * Created on 8. Juli 2002, 16:01
 */
package com.ai_center.dragonbreath.structuralConstraintHandling.graph;

import com.ai_center.dragonbreath.utilities.Pair;

import java.util.Collection;
import java.util.Stack;


/**
 * A Stack for the Search for Graphs, in SCTDockingGraph, to find a graphMatch
 * per backtracking. Stores positions in the search for a solution.
 * 
 * @version $Revision: 1.2 $
 * @author tim baermann
 */
public class GraphPartStack {
    // Instance variables *****************************************************

    /** 
     * If for a state the queue is empty this object is stored. 
     */
    Object nullObject = new Object();

    /** 
     * The missing interpretations for the current part. 
     */
    Stack interpretations = new Stack();

    /** 
     * And the actual part itself. 
     */
    Stack partsInStack = new Stack();

    /**
     * Each state in the search for a GraphMatch consists in the queue at the
     * given moment.
     */
    Stack queues = new Stack();

    // Constructors ***********************************************************

    /**
     * Creates a new instance of GraphPartStack
     */
    public GraphPartStack() {
    }

    // Methods ****************************************************************

    /**
     * Clears the whole stack
     */
    public void clear() {
        queues.clear();
        interpretations.clear();
        partsInStack.clear();
    }


    /**
     * Looks into the stack to test if a given part is allready in it. Used as
     * the search process may go in circles, and the same part should not be
     * pushed twice.
     * 
     * @param part The searched part.
     * 
     * @return Is the part on the stack?
     */
    public boolean containsPart(StructuralConstraintGraphPart part) {
        return partsInStack.contains(part);
    }


    /**
     * Is the Stack empty?
     * 
     * @return empty?
     */
    public boolean empty() {
        return interpretations.empty();
    }


    /**
     * Returns a Pair with the queue as first and a Stack of interpretations as
     * second element. If there are no alternative interpretations, first is
     * null, and second one single interpretation.
     * 
     * @return The Pair on top of the Stack.
     */
    public Pair pop() {
        Pair   result = new Pair();
        Object queue = queues.pop();

        if (queue == nullObject) {
            result.first = null;
        } else {
            result.first = queue;
        }

        result.second = interpretations.pop();
        partsInStack.pop();

        return result;
    }


    /**
     * Pushes the actual queue and a stack full of interpretations for the
     * actual part on the Stack. Used if there is only one possible
     * interpretation for the part.
     * 
     * @param f The frame that should be added next.
     */
    public void push(GraphPartFrame f) {
        queues.push(nullObject);
        interpretations.push(f);
        partsInStack.push(f.myGraphPart);
    }


    /**
     * Pushes the actual queue and a stack full of interpretations for the
     * actual part on the Stack. Used if there were several possible
     * interpretations for the part.
     * 
     * @param partInterpretations All the interpretations not yet tried for the
     *        actual part.
     * @param queue The searchqueue at the moment the interpretations were
     *        added, may be null
     */
    public void push(Collection queue, Stack partInterpretations) {
        queues.push(queue);
        interpretations.push(partInterpretations);

        GraphPartFrame f = (GraphPartFrame) partInterpretations.peek();
        partsInStack.push(f.myGraphPart);
    }
}