/*
 * NodeContainer.java
 *
 * Created on 12. April 2002, 14:22
 */
package com.ai_center.dragonbreath.structuralConstraintHandling.graph;

import com.ai_center.dragonbreath.*;
import com.ai_center.dragonbreath.constraintgraph.NodeType;
import com.ai_center.dragonbreath.utilities.SetHashtable;

import java.util.*;


/**
 * Container for Nodes / NodeTypes. Nodes can be requested with a certain Node
 * as an example, and all the Nodes whose attributes are a superset of the
 * given one are returned.
 * 
 * @version $Revision: 1.4 $
 * @author tim baermann
 */
public class NodeContainer {
    // Instance variables *****************************************************

    /**
     * Arbitrary Nodes are stored seperately, as they can be accessed by any
     * key.
     */
    HashSet arbitraryNodes;

    /** 
     * The real Container 
     */
    SetHashtable container = new SetHashtable();

    // Constructors ***********************************************************

    /**
     * Creates a new NodeContainer object.
     */
    public NodeContainer() {
        arbitraryNodes = new HashSet();
    }

    // Methods ****************************************************************

    /**
     * Returns all the nodes in the container that have the same type as the
     * given node.
     * 
     * @param n The Type that is searched.
     * 
     * @return A Set of all the Nodes in this container that fit to the given
     *         node.
     */
    public Set getNodes(NodeType n) {
        Set result = new HashSet();

        if (n == null) {
            return result;
        }

        result.addAll(arbitraryNodes);

        Integer key = new Integer(n.typeID);
        Object  o = container.get(key);

        if (o instanceof Set) {
            result.addAll((Collection) o);
        } else {
            if (o != null) {
                result.add(o);
            }
        }

        return result;
    } //getNodes()


    /**
     * Adds a Node to the Container.
     * 
     * @param n The node, normally an SCNode, but may also be a Node.
     */
    public void addNode(NodeType n) {
        if (n.typeID == Registry.ANY_NODE) {
            arbitraryNodes.add(n);
        }

        Integer key = new Integer(n.typeID);
        container.add(key, n);
    }


    /**
     * addNode() for all Nodes in the Container.
     * 
     * @param nodes A Container full of nodes.
     */
    public void addNodes(Collection nodes) {
        NodeType n;
        Iterator it = nodes.iterator();

        while (it.hasNext()) {
            n = (NodeType) it.next();
            addNode(n);
        }
    }


    /**
     * Deleted all the nodes stored.
     */
    public void clear() {
        container.clear();
    }


    /**
     * Is any of the given nodes in the container?
     * 
     * @param nodes Nodes.
     * 
     * @return True if at least one found.
     */
    public boolean containsAnyNode(Collection nodes) {
        boolean  found = false;
        Iterator it = nodes.iterator();

        while (it.hasNext() && ! found) {
            NodeType n   = (NodeType) it.next();
            Integer  key = new Integer(n.typeID);
            found = container.containsKey(n);
        }

        return found;
    }


    /**
     * Is a node of the given type stored in the nodesContainer? Or if there is
     * an arbitrary node allways true.
     * 
     * @param n The searched nodeType.
     * 
     * @return true if there is a fitting node.
     */
    public boolean containsNode(NodeType n) {
        Integer key = new Integer(n.typeID);

        if (arbitraryNodes.size() > 0) {
            return true;
        }

        return container.containsKey(key);
    }


    /**
     * Remove all the Nodes with the given type.
     * 
     * @param n The type that should be removed.
     */
    public void removeNodes(NodeType n) {
        container.removeKey(new Integer(n.typeID));
    }


    /**
     * The number of Nodes in this Container.
     * 
     * @return Size of Container.
     */
    public int size() {
        return container.size();
    }


    /**
     * Given a Collection of Nodes it sorts the nodes in the collection after
     * the frequency  of the given type in the container. ( Used eg. in
     * ApplicationCondition.checkChangedNodes() )
     * Not yet implemented!
     * 
     * @param nodes  The nodes to be sorted first the less frequent then the
     *        frequent ones
     */
    public void sortNodesByFrequency(Collection nodes) {
        
    }


    /**
     * How often is the given type in the set?
     * 
     * @param n The searched nodeType
     * 
     * @return The number of times the nodeType is in the Container.
     */
    public int typeFrequency(NodeType n) {
        Integer key = new Integer(n.typeID);
        Object  o = container.get(key);

        if (o instanceof Set) {
            return ((Set) o).size();
        } else {
            if (o == null) {
                return 0;
            } else {
                return 1;
            }
        }
    }
} //class
