/*
 * NullMainPart.java
 *
 * Created on 14. Juli 2002, 01:11
 */
package com.ai_center.dragonbreath.structuralConstraintHandling.graph;

import com.ai_center.dragonbreath.constraintgraph.Node;
import com.ai_center.dragonbreath.structuralConstraintHandling.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.applicationCondition.ApplicationCondition;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.dockingGraph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.dockingGraph.applicationCondition.type.TypeDockingConnectingPACPart;
import com.ai_center.dragonbreath.utilities.SetHashtable;

import java.util.*;


/**
 * If a dockingGraph has no mainPart this one is entered.
 * 
 * @version $Revision: 1.5 $
 * @author tim baermann
 */
public class NullMainPart extends StructuralConstraintGraphPart
    implements MainPartInterface, NodeChangedInterface, ConnectingPartInterface {
    // Instance variables *****************************************************

    /** 
     * Used as return value for getFittingFrames(). 
     */
    Stack framesStack = null;

    /** 
     * A List, containing only this. used in checkChangedNodes(). 
     */
    Vector thisList = new Vector();

    /**
     * There can be only one instance of the NullMainPart. This variable
     * registeres if it was allready created.
     */
    boolean instanceCreated = false;

    // Constructors ***********************************************************

    /**
     * Creates a new instance of NullMainPart
     * 
     * @param scg The DockingGraph this mainPart is part of.
     */
    public NullMainPart(StructuralConstraintGraph scg) {
        super(scg);
        thisList.add(this);
    }

    // Methods ****************************************************************

    /**
     * Returns all the interpretations used in this part.
     * 
     * @return All the interpretations in this Part. Do not change the
     *         Collection, as these changes are reflected in the part!
     */
    public Collection getAllInterpretations() {
        return new HashSet();
    }


    /**
     * Returns an Iterator over the connectedGraphParts.
     * 
     * @return connectedGraphPartsNodes.iterator()
     */
    public Iterator getConnectedGraphPartsIterator() {
        return new Vector().iterator();
    }


    /**
     * Is the StructuralConstraintGraphPart a main part? Should never be
     * called.
     * 
     * @return true.
     */
    public boolean isConnectingPart() {
        return true;
    }


    /**
     * Returns all the frames of the part that fit to the actual interpretation
     * of  the linked nodes of this part i.e. have the same interpretaiton.
     * 
     * @return All the found frames.
     */
    public Stack getFittingFrames() {
        if (framesStack == null) {
            framesStack = new Stack();
            framesStack.push(getPartFrame());
        }

        return framesStack;
    }


    /**
     * If the Graph has an interpretation this function returns a hash for the
     * graph.
     * 
     * @return Hash of the Interpretations. It is circa a concatenation of all
     *         the hashes of all the nodes and edges.
     */
    public java.lang.String getHashCodeString() {
        return "0";
    }


    /**
     * Returns usedGraphPartFrame or if not set calls createPartFrame(). Should
     * not be called if the Part has no interpretation.
     * 
     * @return A frame of the current interpretation.
     */
    public GraphPartFrame getInnerGraphFrame() {
        if (usedGraphPartFrame == null) {
            usedGraphPartFrame = createPartFrame();
        }

        return usedGraphPartFrame;
    }


    /**
     * Returns all nodes that have a link assigned
     * 
     * @return The Nodes of linked nodes in this graph
     */
    public Vector getLinkedNodes() {
        return new Vector();
    }


    /**
     * Returns the node of the StructuralConstraintGraphPart that has the given
     * Node as interpretation, null if none.
     * 
     * @param interpretation  The searched interpretation.
     * 
     * @return The SCNode that has interpretation as interpretation or null if
     *         none exists.
     */
    public SCNode getNodeWithInterpretation(Node interpretation) {
        return null;
    }


    /**
     * Returns all the nodes of the StructuralConstraintGraphPart. Returns the
     * Vector used in this part, so dont mess with it!
     * 
     * @return The nodes.
     */
    public Vector getNodes() {
        return nodes;
    }


    /**
     * Returns one of the parts in connectedGraphPartsNodes.
     * 
     * @return A graph part registered as connected to this one, or null if
     *         none exists.
     */
    public StructuralConstraintGraphPart getOneConnectedGraphPart() {
        return null;
    }


    /**
     * Is this StructuralConstraintGraphPart connected to any other part?
     * 
     * @return False.
     */
    public boolean isPartConnected() {
        return false;
    }


    /**
     * Returns usedGraphPartFrame or if not set calls createPartFrame(). Should
     * not be called if the Part has no interpretation.
     * 
     * @return A frame of the current interpretation.
     */
    public GraphPartFrame getPartFrame() {
        if (usedGraphPartFrame == null) {
            usedGraphPartFrame = new GraphPartFrame(this);
        }

        return usedGraphPartFrame;
    }


    /**
     * Adds one ApplicationCondition to the list of PACs or NACs.
     * 
     * @param ac An ApplicationCondition.
     */
    public void addAC(ApplicationCondition ac) {
    }


    /**
     * Adds one PACPart to the list of the connectingPACParts. Should never be
     * called for NullMainPart.
     * 
     * @param pacPart The PACPart that connects this mainPart wiht an other
     *        one.
     */
    public void addConnectingPACPart(TypeDockingConnectingPACPart pacPart) {
    }


    /**
     * Adds an edge to the list of edges. Called by SCNode.connect()
     * 
     * @param newEdge The edge that should be added.
     */
    public void addEdge(SCEdge newEdge) {
    }


    /**
     * Enters an interpreatation of a node as allready used. The mapping from
     * structural constraint graph g to problem graph p must be injective,
     * i.e. no two different nodes from p may have the same interpretation in
     * p. (a link between two nodes means in this sense that they are the same
     * node, ergo they have to have the same interpretation)
     * 
     * @param interpretation The interpretation.
     * @param node  The node that gets the interpretation.
     */
    public void addInterpretation(Node interpretation, SCNode node) {
    }


    /**
     * Registeres a part and a node to be linked by an other part. Called by
     * SCNode.setLink()
     * 
     * @param part  The part that is linked.
     * @param node  A node from this part, that is either linked by a node from
     *        part, or links to a node from part.
     */
    public void addLinkedPart(StructuralConstraintGraphPart part, SCNode node) {
    }


    /**
     * Produces a new node that can be connected via connectNodes to others in
     * this graph Also adds the node to the searched ones.
     * 
     * @param id the nodeID as defined in Registry
     * 
     * @return the newly created SCNode
     */
    public SCNode addNode(int id) {
        return null;
    }


    /**
     * Checks wheter with the given nodes a new mainPart can be found.
     * 
     * @param newNodes
     * @param deletedNodes
     * @param expandedNodes
     * @param reducedNodes
     */
    public void checkChangedNodes(Set newNodes, Set deletedNodes, 
                                  Set expandedNodes, Set reducedNodes) {
        tryCreateInstance();
    }


    /**
     * Sets the interpretations as in the frame, then tries all the embeddings
     * possible with the frame, until an embedding is found with wich the  all
     * PACs are valid, or no more embeddings are possible, then saves the
     * frame with partlyFoundMatches or foundMatches. Is called when a PAC has
     * got valid, and the frame for the MainPart match  can be tried again, if
     * it is valid now.
     * 
     * @param frame  A GraphPartFrame for this MainPart. it should be contained
     *        in partlyFoundParts.
     */
    public void checkFrame(GraphPartFrame frame) {
        tryCreateInstance();
    }


    /**
     * Same as tryInterpretation() but doesnt register intersection.
     * 
     * @param interpretation The checked interpretation.
     * 
     * @return Allways false
     */
    public boolean containsInterpretation(Node interpretation) {
        return false;
    }


    /**
     * Copies the interpretations of the other graph to this one. If mainPart
     * copies the PACs and NACs too. The two graphs must be exactly identical
     * in structure.
     * 
     * @param source the graph with the interpretations (has to be also a
     *        NullMainPart)
     * 
     * @throws Error If source is not a NullMainPart.
     */
    public void copyInterpretations(StructuralConstraintGraphPart source) {
        if (! (source instanceof NullMainPart)) {
            throw new Error(
                    "Tried to create NullMainPart from not NullMainPart.");
        }
    }


    /**
     * Stores all the nodes of the given mainPart that connect to an other
     * mainPart via an application condition. Or if part is an
     * ApplicationConditionPart all the nodes that connect to mainParts. Sets
     * usedGraphPartFrame to the new created frame.
     * 
     * @return The created frame.
     */
    public GraphPartFrame createPartFrame() {
        if (usedGraphPartFrame == null) {
            usedGraphPartFrame = new GraphPartFrame(this);
        }

        return usedGraphPartFrame;
    } //createPartFrame()


    /**
     * If there exists a frame that fits to the actual interpretation of  the
     * linked nodes of part i.e. have the same interpretaiton, this function
     * returns true, else false.
     * 
     * @param part  The other part connected to this one.
     * 
     * @return There is at least one frame fitting to part.
     */
    public boolean existFittingFrames(StructuralConstraintGraphPart part) {
        return true;
    }


    /**
     * Is there at least one frame for the part?
     * 
     * @return allways true.
     */
    public boolean existsFrame() {
        return true;
    }


    /**
     * Tests whether the ConstraintGraph can be embedded at this node with its
     * interpretation into the problem graph. Finds the next valid
     * isomorphism. <b>For the first embedding findOneEmbedding() has to be
     * called!!</b> Every time this function returns with true all the nodes
     * of the graph contain a new valid interpretation. If returns false the
     * interpretations of all nodes are set null, but all extern
     * interpretations stay valid. Returns false if no more embedding can be
     * found.
     * 
     * @return True if one more embedding was found, else false
     */
    public boolean findNextEmbedding() {
        return true;
    } //findNextEmbedding()


    /**
     * Tests whether the ConstraintGraph can be embedded at the node seed into
     * the problem graph. Finds only one isomorphism, starting from a start
     * seed.  Seed can be either the dedicated seed, or if that doesnt have an
     * interpretation any linked Node with an interpretation via the link. If
     * none found the root seed interpretation is reset.
     * 
     * @return was one Embedding found?
     */
    public boolean findOneEmbedding() {
        return true;
    }


    /**
     * Tests whether the ConstraintGraph can be embedded at the given node into
     * the problem graph. Finds only one isomorphism. If none found the root
     * seed interpretation is reset. But all extern interpretations stay
     * valid.
     * 
     * @param searchSeed  The seed the search starts from
     * 
     * @return was one Embedding found?
     */
    public boolean findOneEmbedding(SCNode searchSeed) {
        return true;
    }


    /**
     * When a new match for the MainPart is found during this function checks
     * whether the mainParts ApplicationConditions are fullfilled, and
     * initiates the search for whole graphs with this mainPart.
     * 
     * @return Were the ACs valid for this frame?
     */
    public boolean handleMatch() {
        return true;

        /* dl.print(3, "\n mainPart.handleMatch() ");
                                                                                         
                                          boolean found = checkPartPACs();
                                                                                         
                                          if (found) {
                                              dl.print(3, "\n PACs found");
                                              checkingNewFrame = true;
                                              ((SCTDockingGraph) mySCGraph).findAllMatchesWithPart(this);
                                              checkingNewFrame = false;
                                          } else {
                                              dl.print(3, "\n PACs not found");
                                              found = checkPartPACs();
                                          }
                                                                                         
                                          return found;*/
    }


    /**
     * Is a further call of findNextEmbedding() usefull?
     * 
     * @return numberOfBranches > 0
     */
    public boolean hasMorePossibleEmbeddings() {
        return false;
    }


    /**
     * How many other parts connect to this part?
     * 
     * @return Number of parts connected to this part.
     */
    public int numberOfConnectedParts() {
        return 0;
    }


    /**
     * Is the given interpretation used in this graph part?
     * 
     * @param interpretation The interpretation.
     * 
     * @return Allways false.
     */
    public boolean partContainsInterpretation(Node interpretation) {
        return false;
    }


    /**
     * Prints that there are no nodes.
     */
    public void printNodes() {
        dl.print(0, "\nNullMainPart has no nodes");
    }


    /**
     * Never Called.
     * 
     * @param n ..
     */
    public void registerLinkedNode(SCNode n) {
    }


    /**
     * If an interpretation is not used any more it should be removed from list
     * of used interpretations by this function. The mapping from structural
     * constraint graph g to problem graph p must be injective, i.e. no two
     * different nodes from p may have the same interpretation in p.  Does
     * nothing as this is the NullMainPart.
     * 
     * @param interpretation The interpretation.
     * @param node The node that takes the interpretation.
     */
    public void removeInterpretation(Node interpretation, SCNode node) {
    }


    /**
     * Resets the interpretations of all nodes and edges in the graph.
     */
    public void resetInterpretations() {
    } //void resetInterpretations()


    /**
     * Resets the extern interpretations of all nodes and edges in the graph.
     */
    public void resetInterpretationsExtern() {
    }


    /**
     * No intersection is registered after this call. Called before the first
     * node of this graph gets an assignment in
     * SCNode.computeAllPossibleConAssignments() .
     */
    public void resetIntersection() {
    }


    /**
     * Given a Collection of Nodes it sorts the nodes in the collection after
     * the frequency  of the given type in the searchedNodes container. ( Used
     * eg. in ApplicationCondition.checkChangedNodes() )
     * 
     * @param nodes  The nodes to be sorted first the less frequent then the
     *        frequent ones
     */
    public void sortNodesByFrequency(Collection nodes) {
    }


    /**
     * Is the given interpretation allready used in this graph? The mapping
     * from structural constraint graph g to problem graph p must be
     * injective, i.e. no two different nodes from g may have the same
     * interpretation in p. Registeres an Intersection if true.
     * 
     * @param interpretation The interpretation.
     * @param node  The node with the interpretation.
     * 
     * @return true if the interpretation is used in the graph?
     */
    public boolean tryInterpretation(Node interpretation, SCNode node) {
        return false;
    }


    /**
     * Inherited, does nothing.
     * 
     * @return null
     */
    protected SCNode findSeed() {
        return null;
    }


    /**
     * Does nothing, as a NullMainPart is allways found.
     * 
     * @param searchSeed  Not used.
     * 
     * @return true.
     */
    private boolean _findOneEmbedding(SCNode searchSeed) {
        return true;
    } //findOneEmbedding()


    /**
     * If there doesnt yet exist an instance tries to create one.
     */
    private void tryCreateInstance() {
        if (mySCGraph instanceof SCTDockingGraph) {
            SCTDockingGraph dg = (SCTDockingGraph) mySCGraph;

            if (! dg.hasInstances()) {
                dg.findAllMatchesWithPart(this);
            }
        }
    }
} //class
