/*
 * foundPartFramesContainer.java
 *
 * Created on 12. Mai 2002, 16:21
 */
package com.ai_center.dragonbreath.structuralConstraintHandling.graph;

import com.ai_center.dragonbreath.constraintgraph.Node;

import java.util.*;


/**
 * A container to hold GraphPartFrames it recognizes equal frames and stores
 * them only once.  The class also offers a mechanism to check all its
 * instances for frames that have lost nodes, so that they have become
 * invalid.
 * 
 * @version $Revision: 1.3 $
 * @author tim baermann
 */
public class PartFrameContainer {
    // Static variables/initializers ******************************************

    /**
     * All the instances of PartFrameContainer, it is used to cycle through all
     * containers to remove frames that contain deleted nodes. (done by
     * allContainerCheckDeletedNodes())
     */
    static Vector allInstances = new Vector();

    // Instance variables *****************************************************

    /** 
     * A Set with all the hashes of the frames in this container. 
     */
    HashSet frameHashes = new HashSet();

    /** 
     * The backing HashSet. 
     */
    HashSet frames = new HashSet();

    // Constructors ***********************************************************

    /**
     * Constructor, adds instance to allInstances.
     */
    public PartFrameContainer() {
        allInstances.add(this);
    }

    // Methods ****************************************************************

    /**
     * Checks for all existing PartFrameContainer whether there is a frame
     * stored with the given nodes. If yes the frame is removed from that
     * container. Used to clean up obsolete links.
     * 
     * @param deletedNodes  The nodes that were deleted.
     */
    public static void allContainerCheckDeletedNodes(Collection deletedNodes) {
        if (deletedNodes.size() == 0) {
            return;
        }

        Iterator it = allInstances.iterator();

        while (it.hasNext()) {
            PartFrameContainer c = (PartFrameContainer) it.next();
            c.checkDeletedNodes(deletedNodes);
        }
    } //allContainerCheckDeletedNodes( Container deletedNodes )


    /**
     * Returns the total number of graphPartFrames stored.
     * 
     * @return The number of graphPartFrames of all existing container
     *         instances.
     */
    public static long allContainersSize() {
        long     result = 0;
        Iterator it = allInstances.iterator();

        while (it.hasNext()) {
            PartFrameContainer c = (PartFrameContainer) it.next();
            result += c.size();
        }

        return result;
    }


    /**
     * Stores the given frame under the given key.
     * 
     * @param f The frame to be added.
     * 
     * @return true if the frame was not present before.
     */
    public final boolean add(GraphPartFrame f) {
        if (f != null) {
            if (! frameHashes.add(f.getHashCode())) {
                return false; // frame allready saved.
            }

            frames.add(f);
        }

        return true;
    }


    /**
     * Returns all the frames stored. The original HashSet, so dont change it!
     * 
     * @return A HashSet with the frames
     */
    public HashSet getAll() {
        return frames;
    }


    /**
     * Calls setInterpretations() for all frames stored in this container.
     */
    public void setAllFrames() {
        Iterator it = iterator();

        while (it.hasNext()) {
            GraphPartFrame f = (GraphPartFrame) it.next();
            f.setInterpretations();
        }
    }


    /**
     * Searches for frames that have all the nodes given in connections.
     * 
     * @param nodes     SCNodes that either connect via link to the part of a
     *        frame, or are connected to.
     * 
     * @return A Collection of the GraphPartFrames that connect to the given
     *         connections.
     */
    public Collection getFittingParts(Set nodes) {
        HashSet result;

        if (nodes.size() == 0) { // not connected to any part
            result = frames; // so all parts fit
            clear();
        } else {
            result = new HashSet();

            Iterator it = frames.iterator();

            while (it.hasNext()) {
                GraphPartFrame actualFrame = (GraphPartFrame) it.next();

                if (actualFrame.containsNodes(nodes)) {
                    result.add(actualFrame);
                }
            }
        }

        return result;
    } //getFittingParts 


    /**
     * Adds all the given GraphPartFrames to the container.
     * 
     * @param all All the Frames that should be added.
     */
    public void addAll(Collection all) {
        Iterator it = all.iterator();

        while (it.hasNext()) {
            add((GraphPartFrame) it.next());
        }
    }


    /**
     * Clear the Container e.g. remove all objects.
     */
    public void clear() {
        frames.clear();
        frameHashes.clear();
    }


    /**
     * Is the given frame stored somewhere in the container?
     * 
     * @param f The frame.
     * 
     * @return Container.contains( f )
     */
    public boolean containsFrame(GraphPartFrame f) {
        if (f != null) {
            return frameHashes.contains(f.getHashCode());
        } else {
            return false;
        }
    }


    /**
     * Tests wheter there exists at least one frame that is compatible with the
     * given nodes.
     * 
     * @param nodes     SCNodes that either connect via link to the part of a
     *        frame, or are connected to.
     * 
     * @return Was at least one match for each set of SCNodes found?
     */
    public boolean existFittingParts(Set nodes) {
        boolean  found = false;
        Iterator it = frames.iterator();

        while (it.hasNext() && ! found) { // iterate through all stored frames

            GraphPartFrame actualFrame = (GraphPartFrame) it.next();

            if (actualFrame.containsNodes(nodes)) {
                found = true;
            }
        } //while it

        return found;
    }


    /**
     * Gets all the Frames that contain all the nodes in nodes with the same
     * interpretation.
     * 
     * @param nodes A set of SCNodes, wich are all interpreted.
     * 
     * @return All the Parts that are compatible with the given nodes, and
     *         contain at least one of the nodes.
     */
    public Collection extractFittingParts(Set nodes) {
        HashSet result;

        if (nodes.size() == 0) { // no node in set
            result = frames; // so all parts fit
            frames = new HashSet();
            clear();
        } else {
            result = new HashSet();

            Iterator it = frames.iterator();

            while (it.hasNext()) {
                GraphPartFrame actualFrame = (GraphPartFrame) it.next();

                if (actualFrame.containsNodes(nodes)) {
                    result.add(actualFrame);
                    it.remove();
                    frameHashes.remove(actualFrame.getHashCode());
                }
            }
        }

        return result;
    } //extractFittingParts


    /**
     * Destructor, removes instance from allInstances.
     */
    public void finalize() {
        allInstances.remove(this);
    }


    /**
     * Returns an Iterator over all frames stored.
     * 
     * @return An Iterator.
     */
    public Iterator iterator() {
        return frames.iterator();
    }


    /**
     * Removes a frame from the container, using its part as a key.
     * 
     * @param f The frame that should be removed.
     */
    public void remove(GraphPartFrame f) {
        if (f != null) {
            frames.remove(f);
            frameHashes.remove(f.getHashCode());
        }
    }


    /**
     * Removes all the given frames from the container. Only if they are stored
     * with their own graphPart as keys! Calls for all frames in all remove(
     * frame.myGraphPart, frame ).
     * 
     * @param all  The frames that should be removed.
     */
    public void removeAll(Collection all) {
        Iterator it = all.iterator();

        while (it.hasNext()) {
            GraphPartFrame f = (GraphPartFrame) it.next();
            remove(f);
        }
    }


    /**
     * Calls resetInterpretations() for all frames stored in this container.
     */
    public void resetAllFrames() {
        Iterator it = iterator();

        while (it.hasNext()) {
            GraphPartFrame f = (GraphPartFrame) it.next();
            f.resetInterpretations();
        }
    }


    /**
     * The number of parts for that frames are stored.
     * 
     * @return The number of parts for that frames are stored.
     */
    public int size() {
        return frames.size();
    }


    /**
     * Checks whether any one of the Frames in this container contains one of
     * the deletedNodes, and if that is true removes it. Is called from
     * allContainerCheckDeletedNodes().
     * 
     * @param deletedNodes  The nodes that were deleted.
     */
    private void checkDeletedNodes(Collection deletedNodes) {
        Iterator frameIt = frames.iterator();

        while (frameIt.hasNext()) {
            GraphPartFrame actualFrame = (GraphPartFrame) frameIt.next();
            Iterator       nodeIt   = deletedNodes.iterator();
            boolean        contains = false;

            while (nodeIt.hasNext() && ! contains) {
                Node interpretation = (Node) nodeIt.next();
                contains = actualFrame.containsInterpretation(interpretation);
            } //while nodeIt

            if (contains) { // remove each frame that contains any of the deleted nodes
                frameIt.remove();
                frameHashes.remove(actualFrame.getHashCode());
            } //if
        } //while frameIt
    } //checkDeletedNodes()
} //class
