/*
 * SimpleEdge.java
 *
 * Created on 5. Dezember 2001, 23:28
 */
package com.ai_center.dragonbreath.structuralConstraintHandling.graph;

import com.ai_center.dragonbreath.Registry;
import com.ai_center.dragonbreath.constraintgraph.ConnectedNode;
import com.ai_center.dragonbreath.constraintgraph.Node;


/**
 * The connection between two SCNodes. Used in StructuralConstraints to
 * represent Docking- and TestGraphs.
 * 
 * @version $Revision: 1.3 $
 * @author tim baermann
 */
public class SCEdge {
    // Instance variables *****************************************************

    /**
     * The first edge in the problem graph this connection is associated with.
     */
    ConnectedNode interpretation1;

    /**
     * The second edge in the problem graph this connection is associated with.
     */
    ConnectedNode interpretation2;

    /** 
     * The node this connection comes from. 
     */
    SCNode nodeFrom;

    /** 
     * The node this connection points to. 
     */
    SCNode nodeTo;

    /** 
     * The name of the edgeType 
     */
    String edgeTypeName;

    /** 
     * The root of the graph, this edge is part of. 
     */
    StructuralConstraintGraphPart myGraphPart;

    /** 
     * The edge Id as given in the Registry. 
     */
    int edgeID;

    // Constructors ***********************************************************

    /**
     * Creates new SCEdge. The direction is From To.
     * 
     * @param from first node
     * @param to second node
     * @param id The edgeType id see
     * 
     * @see Registry
     */
    public SCEdge(SCNode from, SCNode to, int id) {
        edgeID       = id;
        nodeFrom     = from;
        nodeTo       = to;
        myGraphPart  = nodeFrom.myGraphPart;
        edgeTypeName = Registry.edgeTypeName[id];
    }

    // Methods ****************************************************************

    /**
     * Get the interpretation of this edge, viewed from the node n.
     * 
     * @param n node that has the return value as ConnectedNode in its
     *        connections table
     * 
     * @return the other ConnectedNode
     */
    public ConnectedNode getConnectedNodePointingFrom(Node n) {
        if (interpretation1 != null) {
            if (interpretation1.node == n) {
                return interpretation2;
            }
        }

        if (interpretation2 != null) {
            if (interpretation2.node == n) {
                return interpretation1;
            }
        }

        return null; //if there was only one interpretation, and none fitted.
    }


    /**
     * Returns the direction as given in Registry
     * 
     * @param node the viewpoint of the query
     * 
     * @return direction as seen from from
     */
    public boolean getDirection(SCNode node) {
        if (node != nodeFrom) {
            return Registry.INCOMING;
        } else {
            return Registry.OUTGOING;
        }
    }


    /**
     * Returns the typeID of this edge.
     * 
     * @return The edgeID as defined in Registry.
     */
    public int getEdgeID() {
        return edgeID;
    }


    /**
     * Returns the nodeFrom member
     * 
     * @return The SCNode this edge comes from.
     */
    public SCNode getFrom() {
        return nodeFrom;
    }


    /**
     * Set an edge in the problem graph this connection is associated with.
     * Call resetInterpretation first to change interpretation.
     * 
     * @param i a ConnectedNode from the problemgraph
     * 
     * @throws Error If the edge is incompatible to the ConnectedNode.
     */
    public void setInterpretation(ConnectedNode i) {
        if (i == null) {
            return;
        }

        if ((i.edgeID != edgeID) && (edgeID != Registry.ANY_EDGE) && 
                (i.edgeID != Registry.ANY_EDGE)) {
            throw new Error("The edge can not have the given interpretation.");
        }

        interpretation1 = i;
        interpretation2 = i.other;
        myGraphPart.addEdgeInterpretation();
    } //public void setConnection(ConnectedNode con)


    /**
     * Returns the current interpretation.
     * 
     * @return interpretation1
     */
    public ConnectedNode getInterpretation() {
        return interpretation1;
    }


    /**
     * An edge is allways connected with two nodes. This call returns the other
     * node.
     * 
     * @param n One Node
     * 
     * @return The other node
     * 
     * @throws Error If the given node is not part of the edge.
     */
    public SCNode getOtherNode(SCNode n) {
        if (n == nodeFrom) {
            return nodeTo;
        } else if (n == nodeTo) {
            return nodeFrom;
        } else {
            throw new Error("Node is not part of edge.");
        }
    }


    /**
     * Is the given SCNode part of this edge?
     * 
     * @param node The tested node.
     * 
     * @return (node == nodeFrom) || (node == nodeTo)
     */
    public boolean isSCNodeConnected(SCNode node) {
        return (node == nodeFrom) || (node == nodeTo);
    }


    /**
     * Does the given ConnectedNode have the same type as this SCEdge?
     * 
     * @param cn  The other edge.
     * 
     * @return true if both edges have the same type, else false.
     */
    public boolean isSameType(ConnectedNode cn) {
        if ((cn.edgeID == edgeID) || (edgeID == Registry.ANY_EDGE)) {
            if (cn.direction == Registry.INCOMING) {
                if ((cn.nodeTypeID == nodeFrom.typeID) || 
                        (nodeFrom.typeID == Registry.ANY_NODE)) {
                    return true;
                }
            } else if ((cn.nodeTypeID == nodeTo.typeID) || 
                           (nodeTo.typeID == Registry.ANY_NODE)) {
                return true;
            }
        }

        return false;
    }


    /**
     * Returns the nodeTo member.
     * 
     * @return nodeTo.
     */
    public SCNode getTo() {
        return nodeTo;
    }


    /**
     * Is the node n the interpretation of an SCNode on one side of this edge?
     * 
     * @param n the node
     * 
     * @return Is the node n the interpretation of an SCNode on one side of
     *         this edge?
     */
    public boolean connectsNode(Node n) {
        if (interpretation2 != null) {
            if (interpretation2.node == n) {
                return true;
            }
        }

        if (interpretation1 != null) {
            if (interpretation1.node == n) {
                return true;
            }
        }

        return false;
    }


    /**
     * Is there an interpretation assigned to this edge?
     * 
     * @return interpretation exists
     */
    public boolean hasInterpretation() {
        return (interpretation1 != null) || (interpretation2 != null);
    }


    /**
     * Set interpretation null.
     */
    public void resetInterpretation() {
        if (interpretation1 != null) {
            myGraphPart.removeEdgeInterpretation();
        }

        interpretation1 = null;
        interpretation2 = null;
    }


    /**
     * Is the connectedNode one of the interpretations?
     * 
     * @param cn The connected node to be checked
     * 
     * @return is n one interpretation?
     */
    boolean isInterpretation(ConnectedNode cn) {
        if ((interpretation1 == cn) || (interpretation2 == cn)) {
            return true;
        }

        return false;
    }
} //class
