/*
 *
 *
 * Created on 10. Januar 2002, 10:21
 */
package com.ai_center.dragonbreath.structuralConstraintHandling.graph;

import com.ai_center.dragonbreath.*;
import com.ai_center.dragonbreath.graphRepair.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.applicationCondition.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.dockingGraph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.dockingGraph.applicationCondition.instance.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.dockingGraph.applicationCondition.type.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.movableSubgraph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.testGraph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.testGraph.applicationCondition.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.testGraph.applicationCondition.instance.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.testGraph.applicationCondition.type.*;
import com.ai_center.dragonbreath.utilities.Pair;

import java.util.*;

import org.jdom.Element;


/**
 * Holds all the data for the graphs of one structural constraint. It is stored
 * in StructuralConstraint Type. If a new instance of this constraint is
 * spawned, the BuilderKit produces the docking/testing graphs needed.
 * 
 * @version $Revision: 1.6 $
 * @author teutates
 */
public class SCGraphFactory {
    // Static variables/initializers ******************************************

    /** 
     * Constant that qualifies a GraphPart as a MainPart. 
     */
    public static final int Main = 0;

    /** 
     * Constant that qualifies a GraphPart as a PAC. 
     */
    public static final int PAC = 1;

    /** 
     * Constant that qualifies a GraphPart as a NAC. 
     */
    public static final int NAC = 2;

    // Instance variables *****************************************************

    /** 
     * Used to modify the way all instances repair their DockingGraph 
     */
    public String dockingGraphRepairerClassName;

    /** 
     * Used to modify the way all instances repair their TestGraphs 
     */
    public String testGraphRepairerClassName;

    /**
     * If the testGraph could not be repaired in one step this number
     * determines the probability of choosing the same testGraph again in the
     * next repair step.
     */
    public double keepOnRepairingLastProbability;

    /**
     * The probability that a StructuralConstraintInstance destroys its docking
     * graph instead of repairing the testGraph.
     */
    public double removeDockingProbability;

    /**
     * The probability, that when a graphpart is destroyed during repair, a
     * node is removed and not an edge. normally it is better to remove only
     * edges, as that changes less of the graph, but we do not want to exclude
     * any possible graph, so include a small probability that a node is
     * removed instead of an edge.
     */
    public double removeNodeProbability;

    /** 
     * How much time may a repair step take? 
     */
    public long millisForRepair;

    /**
     * Default values of loaded Nodes. A mapping from the XML-Node that is used
     * to call for each new node node.setAuxiliaryNodeData().
     */
    HashMap nodeAuxiliaryData = new HashMap();

    /** 
     * Stores which nodes are symmetrical twins to other nodes. 
     */
    HashSet symmetricalTwins = new HashSet();

    /**
     * If the Structural Constraint should use other functions for cost
     * calculation and repair then the class with this name is created as
     * constraintInstances, and there the new functions can be implemented. It
     * has to be a subclass of StructuralConstraintInstance.
     */
    String customInstanceClassName;

    /** 
     * The subgraphpart that the next nodes and edges and links are added to. 
     */
    SubgraphpartDescription actualSubgraphpart;

    /** 
     * The graph that the next subgraph is added to. 
     */
    Vector actualGraph;

    /** 
     * The subgraph that the next subgraphparts are added to. 
     */
    Vector actualSubgraph;

    /**
     * All the SubgraphDescriptions that form together the docking graph. The
     * parts that form together one Main graph, one NAC or one PAC are stored
     * in one vector of Subgraphdescribtions. That means dockingGraph is a
     * vector of vectors.
     */
    Vector dockingGraph = new Vector();

    /** 
     * A vector with all the test graphs as vectors of SubgraphDescriptions. 
     */
    Vector testGraphs = new Vector();

    /** 
     * The type of all the added subgraphparts 
     */
    int actualType;

    // Methods ****************************************************************

    /**
     * The customInstanceClassName.
     * 
     * @return customInstanceClassName
     */
    public String getCustomInstanceClassName() {
        return customInstanceClassName;
    }


    /**
     * If a Structural Constraint should be Customized that is done via the
     * Structural Constraint Instance. This is the function that informs the
     * SCGraphFactory of the new class name.
     * 
     * @param instanceClassName  The name of the new
     *        StructuralConstraintInstance class. With full qualification.
     * @param dockingGraphRepairerClassName  The name of a custom
     *        dockingGraphRepairerClass.
     * @param testGraphRepairerClassName  The name of a custom
     *        testGraphRepairerClass.
     * @param millisForRepair The time a structural repair of one instance may
     *        take.
     * @param removeNodeProbability The probability that, when a graph part is
     *        destroyed for repair, a node is removed and not an edge.
     * @param removeDockingProbability The probability that the DockingGraph is
     *        destroyed and not a TestGraph repaired.
     * @param keepOnRepairingLastProbability When a TestGraph could not be
     *        repaired in one iteration     this is the probability that next
     *        time the same testgraph is repaired again.
     */
    public void setCustomInstanceParameters(String instanceClassName, 
                                            String dockingGraphRepairerClassName, 
                                            String testGraphRepairerClassName, 
                                            long millisForRepair, 
                                            double removeNodeProbability, 
                                            double removeDockingProbability, 
                                            double keepOnRepairingLastProbability) {
        customInstanceClassName = instanceClassName;

        this.dockingGraphRepairerClassName = dockingGraphRepairerClassName;
        this.testGraphRepairerClassName    = testGraphRepairerClassName;

        this.millisForRepair                = millisForRepair;
        this.removeNodeProbability          = removeNodeProbability;
        this.removeDockingProbability       = removeDockingProbability;
        this.keepOnRepairingLastProbability = keepOnRepairingLastProbability;
    }


    /**
     * Creates a SCIDockingGraph, to use for an structural constraint instance.
     * 
     * @param typeGraph The dockingGraph of the type.
     * 
     * @return The Instance DockingGraph
     */
    public SCIDockingGraph getInstanceDockingGraph(SCTDockingGraph typeGraph) {
        SCIDockingGraph result = new SCIDockingGraph(typeGraph);
        addDockingMainParts(result);
        addAllDockingApplicationConditions(result);

        return result;
    }


    /**
     * Creates one TestGraph with the key stored in the TypeTestGraph as
     * keyToFactoryData. Returns an instance of the TestGraph.
     * 
     * @param key  The key.
     * 
     * @return A new TestGraph that coresponds to the given key.
     */
    public TestGraph getTestGraphForKey(Object key) {
        return getTestGraph((Vector) key, false);
    }


    /**
     * Returns the docking graph, with all its application conditions and main
     * parts. For use in StructuralConstraintTypes
     * 
     * @return The docking graph contained in this factory.
     */
    public SCTDockingGraph getTypeDockingGraph() {
        SCTDockingGraph result = new SCTDockingGraph(this);
        addDockingMainParts(result);
        addAllDockingApplicationConditions(result);

        return result;
    } //getDocking


    /**
     * Returns all testgraphs for usage in the Structural constraint type.
     * 
     * @return A testGraph that has only the parts that are relevant in the
     *         type.
     */
    public TestGraphTypesContainer getTypeTestGraphs() {
        Iterator                it     = testGraphs.iterator();
        TestGraphTypesContainer result = new TestGraphTypesContainer(this);

        while (it.hasNext()) {
            Vector actualTestGraphData = (Vector) it.next();
            result.addTestGraph(getTestGraph(actualTestGraphData, true));
        }

        return result;
    }


    /**
     * Adds the dockingGraph mainParts to newDocking.
     * 
     * @param newDocking The DockingGraph the mainParts should be added to.
     */
    public void addDockingMainParts(DockingGraph newDocking) {
        Vector main = null;

        if (dockingGraph.size() > 0) {
            main = (Vector) dockingGraph.elementAt(0);
        }

        boolean found = false;

        if (main != null) {
            boolean  hasMore = true;
            Iterator it = main.iterator();

            while (it.hasNext() && hasMore) { //load the main parts.

                SubgraphpartDescription sgp = (SubgraphpartDescription) it.next();

                if (sgp.type == Main) {
                    StructuralConstraintGraphPart newPart = 
                            getDockingMainPart(sgp, newDocking);
                    newDocking.addMainPart(newPart);
                    found = true;
                } else {
                    hasMore = false;
                }
            } //while
        } //if main!=0

        if (! found) { // there is no mainPart
            newDocking.addMainPart(new NullMainPart(newDocking));
        }
    } //getDockingMainParts()


    /**
     * Adds an edge between node a and node b.
     * 
     * @param a reference to from node received by addNode()
     * @param b reference to to node received by addNode()
     * @param edgeType the type of the Edge as defined in Registry
     */
    public void addEdge(Object a, Object b, int edgeType) {
        actualSubgraphpart.edges.add(new EdgeDef((Pair) a, (Pair) b, edgeType));
    }


    /**
     * Adds a link between node a and node b.
     * 
     * @param a reference to from node received by addNode()
     * @param b reference to to node received by addNode()
     */
    public void addLink(Object a, Object b) {
        LinkDef                 link             = new LinkDef();
        boolean                 isInterGraphLink = false;
        SubgraphpartDescription ga               = findDockingSubgraph(a);

        if (ga == null) {
            ga = findTestSubgraph(a);

            if (findDockingSubgraph(b) != null) {
                isInterGraphLink    = true;
                ga.hasLinkToDocking = true;
            }
        }

        link.from             = (Pair) a;
        link.to               = (Pair) b;
        link.isInterGraphLink = isInterGraphLink;
        ga.links.add(link);
        ga.partsLinkedTo.add(b);
    } //addLink()


    /**
     * Adds one node into actual graphpart. Returns the added node for use as
     * reference when adding edges and links.
     * 
     * @param typeId the type of the Node as defined in {@link
     *        com.ai_center.dragonbreath.Registry }
     * @param isLeafNode  Is the given node the Symmetrical Twin of an other
     *        node? (only one of the two is marked)
     * @param auxiliaryData  Data of other parameters of the node, like the
     *        value of a variable, or the range of a RangeConstraint.
     * 
     * @return a reference to the node created
     */
    public Object addNode(int typeId, boolean isLeafNode, Element auxiliaryData) {
        Pair result = new Pair(new Integer(typeId), null);
        actualSubgraphpart.nodes.add(result);
        nodeAuxiliaryData.put(result, auxiliaryData);

        if (isLeafNode) {
            symmetricalTwins.add(result);
        }

        return result;
    }


    /**
     * Adds all the Application Conditions that have no link to the
     * dockingGraph or a link to a mainPart with a link to the dockingGraph to
     * the given testGraph.
     * 
     * @param testGraph The TestGraph that gets the ACs.
     * @param tGraphData The describtions of the ACs.
     */
    public void addTestApplicationConditions(TestGraph testGraph, 
                                             Vector tGraphData) {
        boolean  main;
        Iterator subgraphIt = tGraphData.iterator();

        while (subgraphIt.hasNext()) {
            main = false;

            Vector                  subgraph  = (Vector) subgraphIt.next();
            Iterator                partIt    = subgraph.iterator();
            SubgraphpartDescription sgp       = null;
            Vector                  allParts  = new Vector(); // list of all parts of a subgraph
            ApplicationCondition    ac        = null;
            boolean                 firstPart = true;

            while (partIt.hasNext() && ! main) { //load the parts.
                sgp = (SubgraphpartDescription) partIt.next();

                if (sgp.type != Main) {
                    if (firstPart) {
                        ac        = new TestNAC(testGraph);
                        firstPart = false;
                    }

                    ApplicationConditionPart acp = getApplicationCondition(sgp, 
                                                                           testGraph, 
                                                                           ac);

                    //addACToConnectedMainParts(acp);
                } else { //this subgraph is a main graph, dont add to the list of acs
                    main = true;
                } //else
            } //while partIt
        } //while subgraphIt  
    } //addTestApplicationConditions()


    /**
     * Adds all mainParts of the given testGraph.
     * 
     * @param testGraph  The testGraph.
     * @param tGraphData  All the subgraphs of one TestGraph.
     */
    public void addTestMainParts(TestGraph testGraph, Vector tGraphData) {
        Vector  main  = (Vector) tGraphData.elementAt(0);
        boolean found = false;

        if (main != null) {
            boolean  hasMore = true;
            Iterator it = main.iterator();

            while (it.hasNext() && hasMore) { //load the main parts.

                SubgraphpartDescription sgp = (SubgraphpartDescription) it.next();

                if (sgp.type == Main) {
                    StructuralConstraintGraphPart newPart = 
                            getTestMainPart(sgp, testGraph.getMainParts());
                    getPartLinks(sgp, newPart);
                    found = true;
                } else {
                    hasMore = false;
                }
            } //while
        } //if main!=0

        if (! found) { // there is no mainPart
            testGraph.addMainPart(new NullMainPart(testGraph));
        }
    }


    /**
     * All following subgraphs are added to the docking part. Call
     * <B>before</B> calling beginSubgraph.
     */
    public void beginDocking() {
        actualGraph = dockingGraph;
    }


    /**
     * Every graph may contain multiple main, nac and pac subgraphs. Call this
     * <B>before</B> adding subgraphparts.
     * 
     * @param type StructuralConstraintType.NAC or StructuralConstraintType.PAC
     *        or StructuralConstraintType.Main
     */
    public void beginSubgraph(int type) {
        actualType     = type;
        actualSubgraph = new Vector();
        actualGraph.add(actualSubgraph);
    } //void beginSubgraph(int type)


    /**
     * Every subgraph may contain multiple not connected parts. Call this
     * <B>before</B> adding nodes.
     */
    public void beginSubgraphpart() {
        actualSubgraphpart      = new SubgraphpartDescription();
        actualSubgraphpart.type = actualType;
        actualSubgraph.add(actualSubgraphpart);
    } //beginsubgraphpart()


    /**
     * All following subgraphs are added to a test part. If there are multiple
     * testparts call this before <B>every</B> single one. Call <B>before</B>
     * calling beginSubgraph.
     */
    public void beginTest() {
        actualGraph = new Vector();
        testGraphs.add(actualGraph);
    }


    /**
     * loads all the nodes in subg, into result
     * 
     * @param subg The describtion of the AC, with the nodes
     * @param ac  The AC that gets the nodes.
     */
    private void getACNodes(final SubgraphpartDescription subg, 
                            final ApplicationConditionPart ac) {
        Iterator                      it     = subg.nodes.iterator();
        StructuralConstraintGraphPart acPart = (StructuralConstraintGraphPart) ac;

        if (it.hasNext()) { // use first node as first seed.

            Pair   actualNode = (Pair) it.next();
            int    nodeType  = ((Integer) actualNode.first).intValue();
            SCNode addedNode = acPart.addNode(nodeType, 
                                              symmetricalTwins.contains(
                                                      actualNode));
            actualNode.second = addedNode; // set reference
            acPart.setSeed(addedNode);
        } //if

        while (it.hasNext()) { //add all nodes

            Pair   actualNode = (Pair) it.next();
            int    nodeType  = ((Integer) actualNode.first).intValue();
            SCNode addedNode = acPart.addNode(nodeType, 
                                              symmetricalTwins.contains(
                                                      actualNode));
            actualNode.second = addedNode; // set reference
        } //while
    }


    /**
     * Constructs a Subgraph, with links to the given existing Graphs, enters
     * the subgraph as Application Condition.
     * 
     * @param subg  The describtion of the subgraph to be created
     * @param scg  The StructuralConstraintType the graph should belong to.
     * @param ac The ApplicationCondition the loaded part belongs to.
     * 
     * @return A graph ready to use.
     */
    private ApplicationConditionPart getApplicationCondition(SubgraphpartDescription   subg, 
                                                             StructuralConstraintGraph scg, 
                                                             ApplicationCondition      ac) {
        ApplicationConditionPart result = createSubgraphType(subg, scg, ac);
        getPartNodes(subg, (StructuralConstraintGraphPart) result);
        getPartEdges(subg);
        getPartLinks(subg, (StructuralConstraintGraphPart) result);

        return result;
    }


    /**
     * Constructs a Subgraph, without any links.
     * 
     * @param subg The describtion of the subgraph to be created
     * @param scg  The StructuralConstraintGraph the graph part should belong
     *        to.
     * 
     * @return A graph ready to use.
     */
    private DockingMainPart getDockingMainPart(SubgraphpartDescription subg, 
                                               StructuralConstraintGraph scg) {
        DockingMainPart result = new DockingMainPart(scg);
        getPartNodes(subg, result);
        getPartEdges(subg);

        return result;
    } //private StructuralConstraintGraph getGraph()


    /**
     * Call getPartNodes() first!
     * 
     * @param subg The data that contains the edges.
     */
    private void getPartEdges(final SubgraphpartDescription subg) {
        Iterator it;
        it = subg.edges.iterator();

        while (it.hasNext()) { //add all edges

            EdgeDef actualEdge = (EdgeDef) it.next();
            SCNode  from     = (SCNode) actualEdge.from.second;
            SCNode  to       = (SCNode) actualEdge.to.second;
            int     edgeType = actualEdge.edgeType;

            if (from != null) {
                from.connectTo(to, edgeType);
            }
        } //while
    }


    /**
     * Loads all the links given in the SubgraphpartDescription.
     * 
     * @param subg The subgraph, whose links should be loaded.
     * @param result
     */
    private void getPartLinks(final SubgraphpartDescription subg, 
                              final StructuralConstraintGraphPart result) {
        Iterator it = subg.links.iterator();

        if (it.hasNext()) { //this is necessary..

            while (it.hasNext()) { //add all links

                LinkDef actualLink = (LinkDef) it.next();
                SCNode  from = (SCNode) actualLink.from.second;
                SCNode  to   = (SCNode) actualLink.to.second;

                if (from != null) {
                    from.setLink(to); //the linked node is set by setLink as the seed of the graph

                    if (actualLink.isInterGraphLink) {
                        from.myGraphPart.addInterGraphLink(from);
                        to.myGraphPart.addInterGraphLink(to);
                    } else {
                        result.addInnerGraphLink(to.myGraphPart, from);
                        to.myGraphPart.addInnerGraphLink(result, to);
                    }
                } //if from != null
            } //while
        }
    }


    /**
     * Adds the nodes in subg to the part result.
     * 
     * @param subg The data of the subgraphPart that contains the nodes.
     * @param result The StructuralConstraintGraphPart that gets the nodes
     *        added.
     */
    private void getPartNodes(final SubgraphpartDescription subg, 
                              final StructuralConstraintGraphPart result) {
        Iterator it = subg.nodes.iterator();

        if (it.hasNext()) { // use first node as first seed.

            Pair   actualNode = (Pair) it.next();
            int    nodeType  = ((Integer) actualNode.first).intValue();
            SCNode addedNode = result.addNode(nodeType, 
                                              symmetricalTwins.contains(
                                                      actualNode));
            addedNode.setAuxiliaryNodeData(
                    (Element) nodeAuxiliaryData.get(actualNode));
            actualNode.second = addedNode; // set reference
            result.setSeed(addedNode);
        } //if

        while (it.hasNext()) { //add all nodes

            Pair   actualNode = (Pair) it.next();
            int    nodeType  = ((Integer) actualNode.first).intValue();
            SCNode addedNode = result.addNode(nodeType, 
                                              symmetricalTwins.contains(
                                                      actualNode));
            addedNode.setAuxiliaryNodeData(
                    (Element) nodeAuxiliaryData.get(actualNode));
            actualNode.second = addedNode; // set reference
        } //while
    }


    /**
     * Creates one TestGraph.
     * 
     * @param actualTestGraphData  The describtion of the TestGraph.
     * @param type  If the TestGraph should be for a SC type(true) or an
     *        instance.
     * 
     * @return A new TestGraph, with the structure given in
     *         actualTestGraphData.
     */
    private TestGraph getTestGraph(Vector actualTestGraphData, boolean type) {
        TestGraph testGraph = new TestGraph(actualTestGraphData, type);
        addTestMainParts(testGraph, actualTestGraphData);
        addTestApplicationConditions(testGraph, actualTestGraphData);

        return testGraph;
    }


    /**
     * Constructs a Subgraph, without any links.
     * 
     * @param subg The describtion of the subgraph to be created
     * @param mainGraph  The StructuralConstraintGraph the graph part should
     *        belong to.
     * 
     * @return A graph ready to use.
     */
    private StructuralConstraintGraphPart getTestMainPart(SubgraphpartDescription subg, 
                                                          MovableSubgraph         mainGraph) {
        StructuralConstraintGraphPart result;

        if (subg.links.size() == 0) {
            result = new TestUnconnectedMainPart(mainGraph);
        } else {
            result = new MovableSubgraphPart(mainGraph);
        }

        getPartNodes(subg, result);
        getPartEdges(subg);

        return result;
    } //private StructuralConstraintGraph getGraph()


    /**
     * In an ApplicationConditionPart there may be some links to MainParts.
     * These MainParts are notified by this function that they are linked by
     * the AC.
     * 
     * @param acp The Part that is to be introduced.
     */
    private void addACToConnectedMainParts(final ApplicationConditionPart acp) {
        Iterator connectedMainPartsIt = ((StructuralConstraintGraphPart) acp).getInnerGraphLinkParts()
                                                                           .iterator();

        while (connectedMainPartsIt.hasNext()) {
            MainPartInterface mp = (MainPartInterface) connectedMainPartsIt.next();
            mp.addAC(acp.getApplicationCondition());
        }
    }


    /**
     * This function returns all the PACs and NACs of the docking graph, with
     * the links set correctly.
     * 
     * @param scg The docking graph the application conditions should be added
     *        to.
     */
    private void addAllDockingApplicationConditions(DockingGraph scg) {
        boolean  main;
        Iterator subgraphIt = dockingGraph.iterator();

        while (subgraphIt.hasNext()) {
            main = false;

            Vector                  subgraph       = (Vector) subgraphIt.next();
            Iterator                partIt         = subgraph.iterator();
            SubgraphpartDescription sgp            = null;
            Vector                  allParts       = new Vector(); // list of all parts of a subgraph
            HashSet                 graphsLinkedTo = new HashSet();
            ApplicationCondition    ac             = null;
            boolean                 firstPart      = true;

            while (partIt.hasNext() && ! main) { //load the parts.
                sgp = (SubgraphpartDescription) partIt.next();

                if (sgp.type != Main) {
                    if (firstPart) {
                        ac        = createDockingACPartInstance(scg, sgp);
                        firstPart = false;
                    }

                    ApplicationConditionPart acp = getApplicationCondition(sgp, 
                                                                           scg, 
                                                                           ac);
                    addACToConnectedMainParts(acp);
                } else { //this subgraph is a main graph, dont add to the list of acs
                    main = true;
                } //else
            } //while partIt
        } //while subgraphIt
    } //getAllDockingApplicationConditions()


    /**
     * Computes the number of _different_ Mainparts a PAC is linked to.
     * 
     * @param subg The describtion of the tested PAC.
     * 
     * @return A positive number.
     */
    private int computeMainPartsLinkedTo(SubgraphpartDescription subg) {
        if (subg.numMainPartsLinkedTo == -1) {
            HashSet  allMPs = new HashSet();
            Iterator it = subg.partsLinkedTo.iterator();

            while (it.hasNext()) {
                Pair   linkedNode = (Pair) it.next();
                SCNode otherNode = (SCNode) linkedNode.second;

                if (otherNode.myGraphPart instanceof MainPartInterface) {
                    allMPs.add(otherNode.myGraphPart);
                }
            } //while

            subg.numMainPartsLinkedTo = allMPs.size();
        }

        return subg.numMainPartsLinkedTo;
    }


    /**
     * Creates a new Object of the right type, dependent on the dockingGraph,
     * and the  ApplicationCondition that should be produced.
     * 
     * @param scg The DockingGraph the AC should belong to.
     * @param sgp  The describtion of the AC.
     * 
     * @return A new ApplicationCondition Object
     */
    private ApplicationCondition createDockingACPartInstance(final DockingGraph            scg, 
                                                             final SubgraphpartDescription sgp) {
        ApplicationCondition ac;

        if (sgp.type == PAC) {
            if (scg instanceof SCTDockingGraph) {
                ac = new TypeDockingPAC((SCTDockingGraph) scg); //create the ApplicationCondition
            } else {
                ac = new InstanceDockingPAC(scg);
            }
        } else { //nac

            if (scg instanceof SCTDockingGraph) {
                ac = new TypeDockingNAC((SCTDockingGraph) scg); //create the ApplicationCondition
            } else {
                ac = new InstanceDockingNAC(scg);
            }
        }

        return ac;
    }


    /**
     * Creates an instance of the right class, according to the data given in
     * subg and scg.
     * 
     * @param subg A describtion of the part.
     * @param scg The StructuralConstraintGraph the ac belongs to.
     * @param ac The ApplicationConditionPart the part should belong to.
     * 
     * @return The new Object.
     * 
     * @throws Error  Thrown if subg is not a valid AC describtion.
     */
    private ApplicationConditionPart createSubgraphType(SubgraphpartDescription   subg, 
                                                        StructuralConstraintGraph scg, 
                                                        ApplicationCondition      ac) {
        ApplicationConditionPart result = null;

        if (scg instanceof SCTDockingGraph) {
            switch (subg.type) {
                case Main:
                    break;

                case PAC:

                    int mainpartsLinkedTo = computeMainPartsLinkedTo(subg);

                    if (mainpartsLinkedTo == 0) {
                        result = new TypeDockingUnconnectedPACPart(ac);
                    } else if (mainpartsLinkedTo == 1) {
                        result = new TypeDockingPartPACPart(ac);
                    } else {
                        result = new TypeDockingConnectingPACPart(ac);
                    }

                    break;

                case NAC:

                    if (subg.partsLinkedTo.size() == 0) {
                        result = new TypeDockingUnconnectedNACPart(ac);
                    } else { //if (subg.partsLinkedTo.size() == 1 ) {
                        result = new TypeDockingNACPart(ac);
                    }/*else {
                       result = new ConnectingCSTNACPart();
                       }*/

                    break;

                default:
                    throw new Error("Wrong ApplicationCondition deffinition.");
            }
        } else if (scg instanceof SCIDockingGraph) {
            switch (subg.type) {
                case PAC:

                    if (subg.partsLinkedTo.size() == 0) {
                        result = new InstanceDockingUnconnectedPACPart(ac);
                    } else {
                        result = new InstanceDockingPACPart(ac);
                    }

                    break;

                case NAC:

                    if (subg.partsLinkedTo.size() == 0) {
                        result = new InstanceDockingUnconnectedNACPart(ac);
                    } else {
                        result = new InstanceDockingNACPart(ac);
                    }

                    break;
            }
        } else if (scg instanceof TestGraph) {
            if (((TestGraph) scg).isType()) {
                if (subg.partsLinkedTo.size() == 0) {
                    result = new TypeTestUnconnectedNACPart(ac);
                } else {
                    result = new TypeTestNACPart(ac);
                }
            } else {
                if (subg.partsLinkedTo.size() == 0) {
                    result = new InstanceTestUnconnectedNACPart(ac);
                } else {
                    result = new InstanceTestNACPart(ac);
                }
            }
        }

        return result;
    }


    /**
     * Returns the subgraph that contains the given node. Or null if not found
     * in dockinggraph.
     * 
     * @param node The node that should be searched.
     * 
     * @return The subgraph that contained the given node, or null if none
     *         found.
     */
    private SubgraphpartDescription findDockingSubgraph(Object node) {
        SubgraphpartDescription result    = null;
        boolean                 found     = false;
        int                     nodeIndex;

        //search all docking graphs for the given node
        Iterator subgraphIt = dockingGraph.iterator();
        Iterator partIt;

        while (subgraphIt.hasNext() && (! found)) {
            Vector sg = (Vector) subgraphIt.next();
            partIt = sg.iterator();

            while (partIt.hasNext() && (! found)) {
                SubgraphpartDescription sgp = (SubgraphpartDescription) partIt.next();
                nodeIndex = sgp.nodes.indexOf(node);

                if (nodeIndex != -1) { //the node was found
                    result = sgp;
                    found  = true;
                }
            } //while partIt
        } //while subgraphIt

        return result;
    }


    /**
     * Returns the subgraph that contains the given node. Or null if not found
     * in dockinggraph.
     * 
     * @param node The node that should be searched.
     * 
     * @return The subgraph that contained the given node, or null if none
     *         found.
     */
    private SubgraphpartDescription findTestSubgraph(Object node) {
        SubgraphpartDescription result     = null;
        boolean                 found      = false;
        int                     nodeIndex;
        Iterator                partIt;
        Iterator                subgraphIt;

        //search all the testGraphs for the node
        if (! found) {
            found = false;

            Iterator graphIt = testGraphs.iterator();

            while (graphIt.hasNext() && (! found)) {
                Vector g = (Vector) graphIt.next();
                subgraphIt = g.iterator();

                while (subgraphIt.hasNext() && (! found)) {
                    Vector sg = (Vector) subgraphIt.next();
                    partIt = sg.iterator();

                    while (partIt.hasNext()) {
                        SubgraphpartDescription sgp = (SubgraphpartDescription) partIt.next();
                        nodeIndex = sgp.nodes.indexOf(node);

                        if (nodeIndex != -1) { //the node was found
                            result = sgp;
                            found  = true;
                        }
                    } //while partIt
                } //while subgraphIt
            } //while graphit
        }

        return result;
    } //findSubgraph()

    // Classes ****************************************************************

    /**
     * The description of an link between two nodes in SubgraphDescriptions.
     */
    class LinkDef {
        /** 
         * The from node in the same subgraph. 
         */
        Pair from;

        /** 
         * This is the node the link goes to. 
         */
        Pair to;

        /** 
         * Is this link between a testGraph and a DockingGraph? 
         */
        boolean isInterGraphLink;
    }

    /**
     * Contains the description of a constraintgraph with the nodes, edges and
     * links.
     */
    class SubgraphpartDescription {
        HashSet partsLinkedTo = new HashSet();

        /** 
         * A vector with EdgeDef elements. 
         */
        Vector edges = new Vector();

        /** 
         * A vector of LinkDescriptions. 
         */
        Vector links = new Vector();

        /**
         * A vector of the nodes of the subgraph. each node is represented by a
         * pair with first as the nodeType(as defined in Registry), and second
         * as a link to the SCNode that was produced for it.
         */
        Vector nodes = new Vector();

        /**
         * If this describtion describes a testGraphPart this variable shows,
         * whether it has a link to a docking graph or not.
         */
        boolean hasLinkToDocking     = false;
        int     numMainPartsLinkedTo = -1;

        /** 
         * Is this description a main graph, a nac or a pac? 
         */
        int type;
    } //class 

    /**
     * Every Edge in the testGraphEdges class is described with one object of
     * this type.
     */
    private class EdgeDef {
        /** 
         * A link to a pair in GraphsNodeVectors, that represents one node. 
         */
        Pair from;

        /** 
         * A link to a pair in GraphsNodeVectors, that represents one node. 
         */
        Pair to;

        /** 
         * The edge's type as defined in Registry 
         */
        int edgeType;

        /**
         * constructor
         * 
         * @param f from
         * @param t to
         * @param et  edge type
         */
        public EdgeDef(Pair f, Pair t, int et) {
            from     = f;
            to       = t;
            edgeType = et;
        }
    }
} //class
