/*
 * SCNode.java
 *
 * Created on 5. Dezember 2001, 15:30
 */
package com.ai_center.dragonbreath.structuralConstraintHandling.graph;

import com.ai_center.dragonbreath.Registry;
import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.applicationCondition.UnconnectedACPart;
import com.ai_center.dragonbreath.utilities.*;

import java.util.*;

import org.jdom.Element;


/**
 * A node of a
 * 
 * @version $Revision: 1.7 $
 * @author tim baermann
 * @see StructuralConstraintGraphPart. Implements search for matches with
 *      problem graph.
 */
public class SCNode extends NodeType {
    // Static variables/initializers ******************************************

    /**
     * To compute all combinations/permutations of possible
     * connection-assignments
     * 
     * @see allPossibleConAssignments.
     */
    static AllCombinations combiner = new AllCombinations();

    /**
     * Computes all possible assignments of the connections of this SCNode to
     * the ones of its interpretation. The result is stored in
     * allPossibleConAssignments.
     */
    public static long calls = 0;

    // Instance variables *****************************************************

    /** 
     * The SCNodes type. 
     */
    public String typeName;

    /** 
     * The graph, this node is part of. 
     */
    public StructuralConstraintGraphPart myGraphPart;

    /**
     * Is this SCNode a leaf node? If that is the case it doesnt have to be
     * searched.
     */
    public boolean isLeafNode;

    /**
     * When this node is missing for a testgraph, may an existing node be used
     * for it?
     */
    public boolean reuseNode;

    /** 
     * For debugging output. 
     */
    DebugLogger dl;

    /**
     * When this node is created by graphrepair it uses these values to
     * initialize the new node.
     */
    Element auxiliaryNodeData;

    /**
     * If an interpretation is set from outside and not from within search
     * algorithm, this variable stores it. (e.g. if some nodes should have a
     * certain interpretation)
     */
    Node externInterpretation;

    /** 
     * The node in the problem graph that is associated with tis one. 
     */
    Node interpretation;

    /**
     * This node should allways have the same interpretation as the linked one.
     * Used for the unification of docking and test part nodes, or to mark the
     * connection between pac, nac and the rest of the graph.
     */
    SCNode link = null;

    /**
     * If myGraphPart is a UnconnectedACPart this is set to myGraphpart, else
     * to null. This saves time, as there doesnt have to be a classcast in
     * every setInterpretation call.
     */
    UnconnectedACPart uPart = null;

    /**
     * The actual set of assignments from
     * 
     * @see allPossibleConAssignments for the edges,  the SCEdge at position n
     *      in this.edges is assigned to the ConnectedNode n in
     *      actualAssignments.
     */
    Vector actualConAssignments;

    /**
     * A Vector containing vectors with assignments for the edges. Once the
     * interpretation of the SCNode was set the edges of the Node can be
     * assigned in various ways to the edges of this SCNode. Here all such
     * permutations are stored.
     */
    Vector allPossibleConAssignments = new Vector();

    /** 
     * All edges connected to this SCNode. 
     */
    Vector edges = new Vector();

    /** 
     * All the nodes that link to this one. 
     */
    Vector linkedFrom = null;

    /**
     * The position in allPossibleConAssignments that contains the next
     * actualConAssignments after the call from
     * nextSetOfPossibleConAssignments()
     */
    int nextConAssignmentIndex;

    // Constructors ***********************************************************

    /**
     * Constructor
     * 
     * @param graph The StructuralConstraintGraphPart this node belongs to
     * @param id The type of this node as defined in Registry.
     * @param isLeaf Is the node a leaf node?
     * 
     * @see Registry
     */
    public SCNode(StructuralConstraintGraphPart graph, int id, boolean isLeaf) {
        myGraphPart = graph;
        isLeafNode  = isLeaf;

        if (myGraphPart instanceof UnconnectedACPart) {
            uPart = (UnconnectedACPart) myGraphPart;
        }

        typeID   = id;
        typeName = Registry.nodeTypeName[id];
        dl       = new DebugLogger(true);

        //dl.switchOn(1);// uncomment for debug output
    }

    // Methods ****************************************************************

    /**
     * Once a SCNode has an interpreation this function returns all the SCNodes
     * in the adjacency of this, and assigns to each of them an
     * interpretation, that is compatible with the current interpretation of
     * this. If there is no interpretation that fits all the edges null is
     * returned, if node has no adjacency to be returned an empty Vector is
     * returned. If one of the SCNodes in the adjacency was visited before it
     * is not included in list.
     * 
     * @return A Vector with all the nodes in the adjacency, all interpreted,
     *         or null if node can not be properly expanded.
     */
    public Vector getAdjacency() {
        Vector result = new Vector();

        if (interpretation == null) {

            if ((link != null) || (externInterpretation != null)) { //this node may have an interpretation via its link, but is not yet assigned

                if (getInterpretation() == null) {
                    return null;
                }
            } else { // there is no interpretation, ergo there is no match

                return null;
            }
        }

        boolean matchFound = false;

        // repeat until all edges match, or no more possible assignments exist
        // an assignment might not match, if nodes do have allready an interpretation that is not compatible
        while (hasMoreConAssignments() && ! matchFound) {
            boolean neighborFound = true; // is a neighbor for the actual edge found?
            int     edgeNr = 0;

            nextSetOfPossibleConAssignments(); //next permutation of assignments from SCNodes to ConnectedNodes

            //test all edges that this SCNode has, all edges have to fit!
            while ((edgeNr < edges.size()) && neighborFound) {
                SCEdge        edge          = (SCEdge) edges.elementAt(edgeNr);
                SCNode        nextNode      = edge.getOtherNode(this);
                ConnectedNode connectedNode = getNextConnection(edgeNr);
                edgeNr++; //next round next edge

                if (! edge.hasInterpretation()) { //this edge isnt allready used 

                    if (nextNode.interpretation == null) { //the node was not yet visited

                        if (! nextNode.setInterpretation(connectedNode.node)) {
                            neighborFound = false;
                        } else {
                            edge.setInterpretation(connectedNode);
                            neighborFound = true;
                            result.add(nextNode);
                        } //else
                    } //if (nextNode.interpretation == NULL
                    else if (nextNode.interpretation == connectedNode.node) {
                        //the node was allready visited, but not the edge
                        edge.setInterpretation(connectedNode);
                        neighborFound = true;
                    } else { //the node was visited, but has allready a different interpretation assigned
                        neighborFound = false;
                    } //else 
                } //if (!edge.hasInterpretation())
                else if (edge.isInterpretation(connectedNode)) { //the edge has a valid interpretation

                    //this is the edge we came from.. so there are no other interpretations for it
                    neighborFound = true;
                } else {
                    neighborFound = false;
                } //else 
            } //while ( edgeNr < edges.size() && neighborFound)

            matchFound = neighborFound;

            if (! neighborFound) { //reset all the nodes allready added until this moment

                Iterator it = result.iterator();

                while (it.hasNext()) {
                    SCNode addedNode = (SCNode) it.next();
                    addedNode.resetInterpretationWithEdges();
                } //while

                result.clear();
            }
        } //while (hasMoreConAssignments() && ! matchFound)

        if (! matchFound && (edges.size() > 0)) { // not all the edges could be extended

            return null;
        }

        return result;
    } // getAdjacency()


    /**
     * Once a SCNode has an interpreation this function returns all the SCNodes
     * in the adjacency of this, and assigns to each of them an
     * interpretation, that is compatible with the current interpretation of
     * this.  If there is no interpretation that fits all the edges: only the
     * fitting nodes are  returned, if this node has no adjacency to be
     * returned an empty Vector is returned. If one of the SCNodes in the
     * adjacency was visited before it is not included in list.
     * 
     * @return A Vector with all the nodes in the adjacency, all interpreted,
     *         or null if node can not be properly expanded.
     */
    public Vector getAdjacencyBiggestEmbedding() {
        Vector result = new Vector();

        if (interpretation == null) { 

            if ((link != null) || (externInterpretation != null)) { //this node may have an interpretation via its link, but is not yet assigned

                if (getInterpretation() == null) {
                    return null;
                }
            } else { // there is no interpretation, ergo there is no match

                return null;
            }
        }

        boolean matchFound = false;

        // repeat until all edges match, or no more possible assignments exist
        // an assignment might not match, if nodes do have allready an interpretation that is not compatible
        while (hasMoreConAssignments() && ! matchFound) {
            boolean neighborFound = true; // is a neighbor for the actual edge found?
            int     edgeNr = 0;
            nextSetOfPossibleConAssignments(); //next permutation of assignments from SCNodes to ConnectedNodes

            //test all edges that this SCNode has, all edges have to fit!
            while ((edgeNr < edges.size()) && neighborFound) {
                SCEdge        edge          = (SCEdge) edges.elementAt(edgeNr);
                SCNode        nextNode      = edge.getOtherNode(this);
                ConnectedNode connectedNode = getNextConnection(edgeNr);
                edgeNr++; //next round next edge

                if (! edge.hasInterpretation()) { //this edge isnt allready used 

                    if (connectedNode.edgeID != Registry.NULL_EDGE) {
                        // there was no interpretation for the given edge, the node will not be added to the result
                        // go on with the next edge.
                        if (nextNode.interpretation == null) { //the node was not yet visited

                            if (! nextNode.setInterpretationBiggestEmbedding(
                                          connectedNode.node)) {
                                neighborFound = false;
                            } else {
                                edge.setInterpretation(connectedNode);
                                neighborFound = true;
                                result.add(nextNode);
                            } //else
                        } else if (nextNode.interpretation == connectedNode.node) {
                            //the node was allready visited, but not the edge
                            edge.setInterpretation(connectedNode);
                            neighborFound = true;
                        } else { //the node was visited, but has allready a different interpretation assigned
                            neighborFound = false;
                        } //else 
                    } //if (connectedNode != NO_EDGE_FOUND)
                    else {
                        neighborFound = true; // it is not really true, but as we only search for the biggest 

                        //embedding we accept that there is not an intepretation for each neighbor.
                    }
                } else if (edge.isInterpretation(connectedNode)) { //the edge has a valid interpretation

                    //this is the edge we came from.. so there are no other interpretations for it
                    neighborFound = true;
                } else {
                    neighborFound = false;
                } //else 
            } //while ( edgeNr < edges.size() && neighborFound)

            matchFound = neighborFound;

            if (! neighborFound) { //reset all the nodes allready added until this moment

                Iterator it = result.iterator();

                while (it.hasNext()) {
                    SCNode addedNode = (SCNode) it.next();
                    addedNode.resetInterpretationWithEdges();
                } //while

                result.clear();
            }
        } //while (hasMoreConAssignments() && ! matchFound)

        if (! matchFound && (edges.size() > 0)) { // not all the edges could be extended

            return null;
        }

        return result;
    } // getAdjacencyBiggestEmbedding()


    /**
     * Sets the auxiliaryNodeData which is used to initialize nodes created
     * after this SCNode.
     * 
     * @param nodeData A jdom.Element as stored in file with the nodes default
     *        values.
     */
    public void setAuxiliaryNodeData(Element nodeData) {
        auxiliaryNodeData = nodeData;

        String reuse = auxiliaryNodeData.getAttributeValue("reuse");

        if ((reuse != null) && reuse.equals("true")) {
            reuseNode = true;
        } else {
            reuseNode = false;
        }
    }


    /**
     * Returns the stored auxiliaryNodeData.
     * 
     * @return auxiliaryNodeData
     */
    public Element getAuxiliaryNodeData() {
        return auxiliaryNodeData;
    }


    /**
     * Set the interpretation in the problem graph, for a match every node has
     * to be interpreted in one way  SCNodes with a link can not directly get
     * an interpretation.
     * 
     * @param n the new interpretation
     * 
     * @return Could the interpretation sucessfully be assigned?
     */
    public final boolean setInterpretation(Node n) {
        if (uPart != null) {
            if (uPart.findNewSeedsOn()) {
                if (uPart.getSeed() == this) {
                    if (uPart.isSeedInterpretation(n)) {
                        return false; // the node n was allready stored as an seed interpretation, so no new match can be found when assigning it. 
                    }
                }
            }
        } // else do nothing, myGraphPart was simply no UnconnectedACPart.

        if (interpretation == n) {
            return true;
        }

        if ((externInterpretation != null) && (externInterpretation != n)) {
            return false;
        }

        if (link != null) {
            Node linkInterpretation = link.peekInterpretation();

            if (linkInterpretation == null) {
                if (myGraphPart.tryInterpretation(n, this)) { //this interpretation is allready used

                    return false;
                }
            } else if (linkInterpretation != n) {
                return false;
            }
        } else {
            if (interpretation == n) { //the actual interpretation is the same

                return true;
            } else if (myGraphPart.tryInterpretation(n, this)) { //this interpretation is allready used

                return false;
            }
        }

        if (interpretation != null) { //there was an interpretation before

            return false; //myGraphPart.removeInterpretation(interpretation, this);
        }

        interpretation = n;

        if (interpretation != null) {
            if (! computeAllPossibleConAssignments()) {
                interpretation = null;

                return false;
            }

            interpretation.mark();
            myGraphPart.addInterpretation(interpretation, this);
        }

        return true;
    } //private boolean setInterpretation(Node interpretation) {


    /**
     * Returns the edges member. Be carefull, returns the original!
     * 
     * @return The edges (SCEdge) of the node.
     */
    public Vector getEdges() {
        return edges;
    }


    /**
     * During Repair process and during search for biggest embedding it may
     * happen that the SCNode should have a certain Interpretation, although
     * the interpretation has not yet all the necessary edges connected. This
     * function sets the interpretation and calls
     * computeAllPossibleConAssignmentsFindBiggestEmbeddig() Only sets the
     * edges to other SCNodes that do allready have an interpretation.
     * 
     * @param n The new Interpretation
     * 
     * @return true if the interpretation could be set.
     */
    public boolean setInterpretationBiggestEmbedding(Node n) {
        if (link != null) {
            Node linkInterpretation = link.peekInterpretation();

            if (! ((linkInterpretation == n) || 
                        (linkInterpretation == null))) {
                return false;
            }
        } else {
            if (interpretation == n) { //the actual interpretation is the same

                return true;
            } else if (myGraphPart.tryInterpretation(n, this)) { //this interpretation is allready used

                return false;
            }
        }

        if (interpretation != null) { //there was an interpretation before
            myGraphPart.removeInterpretation(interpretation, this);
        }

        interpretation = n;

        if (interpretation != null) {
            if (computeAllPossibleConAssignmentsBiggestEmbedding()) {
                myGraphPart.addInterpretation(interpretation, this);
                interpretation.mark();
            } else {
                interpretation = null;

                return false;
            }
        }

        return true;
    }


    /**
     * If a node should get a certain interpretation, but not via a link. As
     * the search for a match does not travel a node twice i.e. doesnt visit
     * a node with an interpretation in such a case setInterpretation can not
     * be used.
     * 
     * @param n The interpretation
     * 
     * @return could it be set?
     * 
     * @throws Error If external interpretation contradicts interpretation.
     */
    public final boolean setInterpretationExtern(Node n) {
        resetInterpretationExtern();

        if (! checkLinkingNodesInterpretations(n)) {
            return false;
        }

        if (((interpretation != null) && (interpretation != n)) || 
                ((link != null) && (link.peekInterpretation() != n) && 
                    (link.peekInterpretation() != null))) {
            throw new Error(
                    "External interpretation contradicts interpretation!");
        }

        if (! myGraphPart.tryInterpretation(n, this)) {
            if (externInterpretation == null) { //only intcrease numberExternInterpretations if realy a new interpretation is added
                myGraphPart.numberExternInterpretations++;
                myGraphPart.setSeed(this);
            }

            externInterpretation = n;

            return true;
        }

        return false;
    } //setInterpretationExtern()


    /**
     * Returns the StructuralConstraintGraphPart this node belongs to.
     * 
     * @return The part.
     */
    public StructuralConstraintGraphPart getGraphPart() {
        return myGraphPart;
    }


    /**
     * Returns the interpretation of this SCNode. If it has a link the
     * interpretation is the interpretation of the link.
     * 
     * @return the interpretation of this node.
     */
    public Node getInterpretation() {
        if (link != null) {
            Node linkInterpretation = link.peekInterpretation();

            if ((linkInterpretation != null) && 
                    (interpretation != linkInterpretation)) {
                if (interpretation != null) {
                    myGraphPart.removeInterpretation(interpretation, this); //first remove the current interpretation, as it will be replaced
                } //^^throw Error("Link interpretation is not same as interpretation");

                interpretation = linkInterpretation; // a new interpretation for this was defined in link

                if (computeAllPossibleConAssignments()) {
                    myGraphPart.addInterpretation(interpretation, this);

                    return interpretation;
                } else {
                    dl.print(1, "computeAllPossibleConAssignments wrong");
                    interpretation = null;

                    return null;
                }
            }
        }

        if (externInterpretation != null) {
            if (interpretation != null) {
                if (interpretation != externInterpretation) {
                    throw new Error(
                            "Node has differnt interpretation then extern Interpretation.");
                }
            } else if (! myGraphPart.tryInterpretation(externInterpretation, 
                                                       this)) {
                interpretation = externInterpretation;
                if (! computeAllPossibleConAssignments()) {
                    interpretation = null;

                    return null;
                }

                myGraphPart.addInterpretation(interpretation, this);
            } else {
                return null;
            }
        }

        return interpretation;
    }


    /**
     * Circumvents most correctness tests. May only be called from
     * TestGraphRepairer. Sets the interpretation as given, also registers the
     * interpretation at the graph. But does not compute the edges.
     * 
     * @param n The new interpretation.
     * 
     * @return true if the interpretation could be set.
     */
    public boolean setInterpretationForRepair(Node n) {
        if (myGraphPart.tryInterpretation(n, this)) { //this interpretation is allready used

            return false;
        }

        if (interpretation != null) { //there was an interpretation before
            myGraphPart.removeInterpretation(interpretation, this);
        }

        myGraphPart.addInterpretation(n, this);
        interpretation = n;

        return true;
    }


    /**
     * Links this node with an other node the other node thereby allways takes
     * the same interpretation as this one. Is used for the links between
     * docking and test part of a structural constraint <b>Never use links
     * circular!</b> When a node is linked it is automatically set as the seed
     * of its graph.
     * 
     * @param l the SCNode this node should be linked to
     */
    public void setLink(SCNode l) {
        link = l;
        l.addLinkedFrom(this);
        myGraphPart.setSeed(this);
        myGraphPart.registerLinkedNode(this);
    }


    /**
     * Returns the linked node.
     * 
     * @return link
     */
    public SCNode getLinkedNode() {
        return link;
    }


    /**
     * Adds a node that is linked to this one.
     * 
     * @param n The linking node.
     */
    public void addLinkedFrom(SCNode n) {
        if (linkedFrom == null) {
            linkedFrom = new Vector();
        }

        linkedFrom.add(n);
    }


    /**
     * For a graphPart without interpretations this functions checks whether
     * the given interpretation for this node would be consistent with links
     * and externInterpretations. Called by e.g. by
     * GraphPartFrame.checkInterpretations().
     * 
     * @param i The interpretation to be tested.
     * 
     * @return true if the interpretation is compatible?
     */
    public boolean checkInterpretation(Node i) {
        if (myGraphPart.tryInterpretation(i, this)) { //the interpretation is allready used

            return false;
        }

        if (link == null) {
            if (externInterpretation == null) {
                if (checkLinkingNodesInterpretations(i)) {
                    return true;
                } else {
                    return false;
                }
            } else if (externInterpretation == i) {
                return true;
            } else {
                return false;
            }
        } else {
            Node linkInterpretation = link.peekInterpretation();

            if ((linkInterpretation == i) || (linkInterpretation == null) || 
                    (i == null)) {
                return true;
            } else {
                return false;
            }
        } // else
    } //checkInterpretation()


    /**
     * Computes all possible assignments of the connections of this SCNode to
     * the ones of its interpretation. The result is stored in
     * allPossibleConAssignments. In contrast to
     * computeAllPossibleConAssignments() this function does not require that
     * all edges can be assigned. It also creates partial assignments.
     * 
     * @return Are all currently assigned interpretations compatible with the
     *         nodes interpretation?
     */
    public boolean computeAllPossibleConAssignmentsBiggestEmbedding() {
        Vector        connections;
        ConnectedNode cnode;
        ArrayList     allLists = new ArrayList();
        nextConAssignmentIndex = 0;

        int currentConnection = 0;

        //read for all edges leaving this SCNode the possible assignments from interpretation
        while (currentConnection < edges.size()) {
            SCEdge nextConnection = (SCEdge) edges.elementAt(currentConnection);
            connections = new Vector();

            if (nextConnection.hasInterpretation()) {
                cnode = nextConnection.getConnectedNodePointingFrom(
                                this.interpretation);

                if (cnode == null) { // the edge has an other interpretation that is not compatible with
                                     // the interpretation of this node

                    // the edge has an other interpretation that is not compatible with
                    // the interpretation of this node
                    return false;
                }

                connections.add(cnode);
            } else {
                // enter for each edge the possibillity that it gets no interpretation.
                cnode = new ConnectedNode(Registry.ANY_NODE, Registry.NULL_EDGE, 
                                          true);
                connections.add(cnode);

                SCNode  searchedNode         = nextConnection.getOtherNode(this);
                boolean isCriterionNode      = false;
                boolean isCriterionNodeType  = true;
                boolean isCriterionEdgeID    = true;
                boolean isCriterionDirection = true;
                Node    node                 = null;
                Node    nodeType             = Registry.nodeType[searchedNode.typeID];
                int     edgeID               = nextConnection.edgeID;
                boolean direction            = nextConnection.getDirection(this);

                if (edgeID == Registry.ANY_EDGE) {
                    isCriterionEdgeID    = false;
                    isCriterionDirection = false;
                }

                if (nodeType.typeID == Registry.ANY_NODE) {
                    isCriterionNodeType = false;
                }
                synchronized (interpretation.connections) {
                    interpretation.connections.initializeEnumeration(
                            isCriterionNode, isCriterionNodeType, 
                            isCriterionEdgeID, isCriterionDirection, null, 
                            nodeType, edgeID, direction);
                    cnode = (ConnectedNode) interpretation.connections.getNext();

                    while (cnode != null) {
                        boolean newInterpretationUsable = 
                                ! myGraphPart.tryInterpretation(cnode.node, 
                                                                searchedNode);
                        Node    oldInterpretation = searchedNode.peekInterpretation();
                        boolean OldInterpretationCompatible = 
                                (oldInterpretation == null) || 
                                (cnode.node == oldInterpretation);

                        if (newInterpretationUsable && 
                                OldInterpretationCompatible) {
                            //if the node has allready an interpretation and it is the one now found
                            connections.add(cnode); //enter all the possible connections for this connectiontype
                        }

                        cnode = (ConnectedNode) interpretation.connections.getNext();
                    } //while
                }
            }

            if (connections.isEmpty()) { // should never be the case as allways a NULL_EDGE is added.
                throw new Error("Not connections for all edges found!");
            }

            allLists.add(connections);
            currentConnection++;
        } //while

        if (allLists.size() > 0) {
            allPossibleConAssignments = combiner.findAllCombinations(allLists);
        }


        //now remove the conAssignments that have more null_edges then necessary.
        removeSuperfluousConAssignments();

        return true;
    }


    /**
     * Creates a connection FROM this node TO the given node, adds the
     * connection to <b>both</b> nodes.
     * 
     * @param sn the SCNode this SCNode should be connected to.
     * @param typeId The type of the edge between the two nodes
     */
    public void connectTo(SCNode sn, int typeId) {
        SCEdge newEdge = new SCEdge(this, sn, typeId);
        sn.addConnection(newEdge);
        addConnection(newEdge);
        myGraphPart.addEdge(newEdge);
    }


    /**
     * Does the node contain an edge with the given interpretation?
     * 
     * @param cnode The tested edge interpretation.
     * 
     * @return true if one of the nodes edges is interpreted with cnode.
     */
    public boolean edgeInterpretationUsed(ConnectedNode cnode) {
        Iterator it = edges.iterator();

        while (it.hasNext()) {
            SCEdge edge = (SCEdge) it.next();

            if (edge.isInterpretation(cnode)) {
                return true;
            }
        }

        return false;
    }


    /**
     * Tests whether the ConstraintGraph can be embedded at this node with its
     * interpretation into the problem graph. Finds the next valid
     * isomorphism. <b>For the first embedding findOneEmbedding() has to be
     * called!!</b> Every time this function returns with true all the nodes
     * of the graph contain a new valid interpretation. Returns false if no
     * more embedding can be found.
     * 
     * @return true if one more embedding was found, else false.
     */
    public boolean findNextEmbedding() {
        return myGraphPart.findNextEmbedding();
    }


    /**
     * Tests whether the ConstraintGraph can be embedded at this into the
     * problem graph. Finds only one isomorphism.
     * 
     * @return true if an Embedding was found, else false.
     */
    public boolean findOneEmbedding() {
        return myGraphPart.findOneEmbedding(this);
    }


    /**
     * Does this SCNode have an interpretation allready?
     * 
     * @return interpretation != null;
     */
    public final boolean hasInterpretation() {
        return interpretation != null;
    }


    /**
     * Has this Node a Link to an other node in a different Graph?
     * 
     * @return true if link != null, else false.
     */
    public boolean hasLink() {
        return link != null;
    }


    /**
     * Returns the number of possible interpretations for the edges.
     * 
     * @return allPossibleConAssignments.size()
     */
    public int numberOfConAssignments() {
        return allPossibleConAssignments.size();
    }


    /**
     * Returns the same as getInterpretation() but doesnt set the
     * interpretation, if only an externInterpretation was set, or a link had
     * an interpretation.
     * 
     * @return the interpretation.
     */
    public Node peekInterpretation() {
        if (interpretation != null) {
            return interpretation;
        }

        if (link != null) {
            Node linkInterpretation = link.peekInterpretation();

            if (linkInterpretation != null) {
                return linkInterpretation;
            } else {
                if (externInterpretation != null) {
                    return externInterpretation;
                } else {
                    return null;
                }
            }
        }

        if (externInterpretation != null) {
            return externInterpretation;
        }

        return null;
    }


    /**
     * Reset the interpretation of the node.
     */
    public void resetInterpretation() {
        if (interpretation != null) {
            interpretation.unmark();
            myGraphPart.removeInterpretation(interpretation, this);
        }

        interpretation = null;
        cleanPossibleConAssignments();
    } //resetInterpretation()


    /**
     * Sets the externInterpretation to null
     */
    public void resetInterpretationExtern() {
        if (externInterpretation != null) {
            externInterpretation = null;
            myGraphPart.numberExternInterpretations--;
        }
    }


    /**
     * Set interpretation and interpretations of all edges null.
     */
    public void resetInterpretationWithEdges() {
        if (interpretation != null) {
            interpretation.unmark();
            myGraphPart.removeInterpretation(interpretation, this);
        }

        interpretation = null;
        cleanPossibleConAssignments();

        SCEdge   se;
        Iterator it = edges.iterator();

        while (it.hasNext()) {
            se = (SCEdge) it.next();
            se.resetInterpretation();
        } //while
    } //resetInterpretationWithEdges()


    /**
     * Tests whether the current interpretation is still valid, tests only
     * locally for this one SCNode!
     * 
     * @return true if the interpretation is valid, else false.
     */
    public boolean testInterpretation() {
        ConnectedNode connectedNode;

        Connections   cons              = interpretation.connections;
        boolean       connectionFound   = true;
        int           currentConnection = 0;

        //if all connections are still connected the interpretation is good
        while ((currentConnection < edges.size()) && connectionFound) {
            SCEdge nextConnection = (SCEdge) edges.elementAt(currentConnection);

            if (! cons.contains(nextConnection.getConnectedNodePointingFrom(
                                        getInterpretation()))) {
                connectionFound = false;
            } //if

            currentConnection++;
        } //while

        return connectionFound;
    } //testInterpretation


    /**
     * Set a pointer to the problem graph, for a match every node has to be
     * interpreted in one way SCNodes with a link can not directly get an
     * interpretation, as it gets the one of its link.
     * 
     * @param se The edge that should get the interpretation.
     * @param cn the interpretation of se, and its node is the interpretation
     *        of <B>this</B>.
     * 
     * @return true if the interpretation could sucessfully be assigned, else false.
     */
    final boolean setInterpretation(SCEdge se, ConnectedNode cn) {
        if (cn != null) {
            boolean result = setInterpretation(cn.node);

            if (interpretation != null) { //if the interpretation could be set.

                if (se != null) {
                    se.setInterpretation(cn);
                }
            }

            return result;
        }

        throw new Error("Tried to assign null edge.");
    } //boolean setInterpretation(SCEdge se, ConnectedNode cn)


    /**
     * Connects a node to this.
     * 
     * @param c The edge to be added
     */
    void addConnection(SCEdge c) {
        edges.addElement(c);
    } //public void addConnection(boolean dir, int edgeType, SCNode node)


    /**
     * Is there a set of assignments not yet tried?
     * 
     * @return true if there is a permutation (a set of assignments) left.
     */
    final boolean hasMoreConAssignments() {
        return nextConAssignmentIndex < allPossibleConAssignments.size();
    }


    /**
     * Next set of assignments.
     */
    void nextSetOfPossibleConAssignments() {
        actualConAssignments = (Vector) allPossibleConAssignments.get(
                                       nextConAssignmentIndex);
        nextConAssignmentIndex++;
    }


    /**
     * When the node has an interpretation, how many conAssignments ==
     * different assignments for the connections, are still pending?
     * 
     * @return allPossibleConAssignments.size() - nextConAssignmentIndex;
     */
    final int numberOfMissingConAssignments() {
        return allPossibleConAssignments.size() - nextConAssignmentIndex;
    }


    /**
     * After this call nextSetOfPossibleConAssignments starts again with the
     * first set.
     */
    void resetConAssignments() {
        nextConAssignmentIndex = 0;
    }


    /**
     * Returns the a ConnectedNode of the Node interpretation that could be
     * assigned to the edge edgeNr, and is part of the actual permutation.
     * 
     * @param edgeNr the position of the edge in the edges Vector
     * 
     * @return a ConnectedNode that could be assigned
     */
    private ConnectedNode getNextConnection(int edgeNr) {
        ConnectedNode cnode = (ConnectedNode) actualConAssignments.elementAt(
                                      edgeNr);

        return cnode;
    } //connectedNode getNextConnection()


    /**
     * Are the nodes linking to this one compatible with the given
     * interpretation? I.e. do they have the same interpretation, or none?
     * Called by checkInterpretation()
     * 
     * @param i  The interpretation to be tested.
     * 
     * @return Are the interpretations compatible.
     */
    private boolean checkLinkingNodesInterpretations(Node i) {
        if (linkedFrom != null) {
            boolean  allGood = true;
            Iterator it = linkedFrom.iterator();

            while (it.hasNext() && allGood) {
                SCNode n  = (SCNode) it.next();
                Node   ni = n.peekInterpretation();

                if ((ni != i) && (ni != null)) {
                    allGood = false; // if one of the nodes linked to this node have allready an interpretation but a different one 
                }
            } //while

            return allGood;
        } else {
            return true;
        }
    }


    /**
     * Removes all empty stacks from possibleConnections
     */
    private void cleanPossibleConAssignments() {
        allPossibleConAssignments.clear();
    } //clean


    /**
     * When this SCNode has an interpretation assigned, this function computes
     * all possible assignments of the edges of the interpretation to the
     * edges of this node.
     * 
     * @return Was there an assignment that gives each edge of this node an
     *         interpretation?
     */
    private boolean computeAllPossibleConAssignments() {
        calls++;

        Vector        connections;
        ConnectedNode cnode;
        ArrayList     allLists = new ArrayList();
        nextConAssignmentIndex = 0;

        int currentConnection = 0;

        //read for all edges leaving this SCNode the possible assignments from interpretation
        while (currentConnection < edges.size()) {
            SCEdge nextConnection = (SCEdge) edges.elementAt(currentConnection);
            connections = new Vector();

            if (nextConnection.hasInterpretation()) {
                cnode = nextConnection.getConnectedNodePointingFrom(
                                this.interpretation);
                connections.add(cnode);
            } else {
                SCNode  searchedNode         = nextConnection.getOtherNode(this);
                boolean isCriterionNode      = false;
                boolean isCriterionNodeType  = true;
                boolean isCriterionEdgeID    = true;
                boolean isCriterionDirection = true;
                Node    node                 = null;
                Node    nodeType             = Registry.nodeType[searchedNode.typeID];
                int     edgeID               = nextConnection.edgeID;
                boolean direction            = nextConnection.getDirection(this);

                if (edgeID == Registry.ANY_EDGE) {
                    isCriterionEdgeID    = false;
                    isCriterionDirection = false;
                }

                if (nodeType.typeID == Registry.ANY_NODE) {
                    isCriterionNodeType = false;
                }
                synchronized (interpretation.connections) {
                    interpretation.connections.initializeEnumeration(
                            isCriterionNode, isCriterionNodeType, 
                            isCriterionEdgeID, isCriterionDirection, null, 
                            nodeType, edgeID, direction);

                    connections = new Vector();
                    cnode       = (ConnectedNode) interpretation.connections.getNext();

                    while (cnode != null) {
                        boolean newInterpretationUsable = 
                                ! myGraphPart.tryInterpretation(cnode.node, 
                                                                searchedNode);
                        Node    oldInterpretation = searchedNode.peekInterpretation();
                        boolean OldInterpretationCompatible = 
                                (oldInterpretation == null) || 
                                (cnode.node == oldInterpretation);

                        if (newInterpretationUsable && 
                                OldInterpretationCompatible) {
                            //if the node has allready an interpretation and it is the one now found
                            connections.add(cnode); //enter all the possible connections for this connectiontype
                        }

                        cnode = (ConnectedNode) interpretation.connections.getNext();
                    } //while
                }
            }

            if (! connections.isEmpty()) {
                allLists.add(connections);
            } else { // there is an edge in this SCNode that has no counterpart in the Node, there is no possible assignment.
                allPossibleConAssignments.clear();

                return false;
            }

            currentConnection++;
        } //while

        if (allLists.size() > 0) {
            allPossibleConAssignments = combiner.findAllCombinations(allLists);

            if (allPossibleConAssignments.size() == 0) {
                return false;
            }

            return true;
        } else if (edges.size() == 0) {
            return true;
        }

        return false;
    } //boolean computeAllPossibleConnections()


    /**
     * Called from computePossibleEdgeAssignments Removes all possible
     * Connections Assignments that are smaller then the biggest one. This is
     * done, as we want to find the biggest embedding, so all smaller
     * assingments will never deliver the  biggest embedding.
     */
    private void removeSuperfluousConAssignments() {
        if (allPossibleConAssignments.size() == 1) {
            return;
        }

        int      numEdgesInBiggestAssignment = 0;
        Iterator it = allPossibleConAssignments.iterator();

        while (it.hasNext()) {
            Vector   assignment          = (Vector) it.next();
            int      numInterpretedEdges = 0;
            Iterator edgeIt              = assignment.iterator();

            while (edgeIt.hasNext()) {
                ConnectedNode edge = (ConnectedNode) edgeIt.next();

                if (edge.edgeID != Registry.NULL_EDGE) {
                    numInterpretedEdges++;
                }
            } //while

            if ((numInterpretedEdges >= numEdgesInBiggestAssignment) && 
                    (numInterpretedEdges > 0)) {
                numEdgesInBiggestAssignment = numInterpretedEdges;
            } else {
                it.remove();
            }
        } //while


        // now remove all assignments that were tested before the winning biggestAssignment was found
        // and have not yet been removed.
        it = allPossibleConAssignments.iterator();

        while (it.hasNext()) {
            Vector   assignment          = (Vector) it.next();
            int      numInterpretedEdges = 0;
            Iterator edgeIt              = assignment.iterator();

            while (edgeIt.hasNext()) {
                ConnectedNode edge = (ConnectedNode) edgeIt.next();

                if (edge.edgeID != Registry.NULL_EDGE) {
                    numInterpretedEdges++;
                }
            } //while

            if (numInterpretedEdges < numEdgesInBiggestAssignment) {
                it.remove();
            }
        } //while
    }
} //class
