/*
 * StructuralConstraintGraph.java
 *
 * Created on 10. April 2002, 15:22
 */
package com.ai_center.dragonbreath.structuralConstraintHandling.graph;

import com.ai_center.dragonbreath.constraintgraph.Node;
import com.ai_center.dragonbreath.structuralConstraintHandling.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.applicationCondition.*;
import com.ai_center.dragonbreath.utilities.SetHashtable;

import java.math.BigInteger;

import java.util.*;


/**
 * A Graph describing a Docking- or a TestGraph of a Structural Constraint.
 * 
 * @version $Revision: 1.5 $
 * @author tim baermann
 */
public abstract class StructuralConstraintGraph implements NodeChangedInterface {
    // Instance variables *****************************************************

    /**
     * The mapping from graph g to problem graph p must be injective, i.e. no
     * two different nodes from g may have the same interpretation in p. This
     * hashmap contains all interpretations (Nodes) used in this graph. The
     * nodes of all subgraphs with its interpretations as keys
     */
    public HashMap interpretedNodes = new HashMap();

    /** 
     * All the NACs. 
     */
    public Vector NACs = new Vector();

    /** 
     * All the PACs of the graph 
     */
    public Vector PACs = new Vector();

    /** 
     * Parts of the StructuralConstraintGraph that are obligatory 
     */
    public Vector mainParts = new Vector();

    /**
     * The HashCode of a StructuralConstraintGraph is the concatenation of the
     * HashCodes of its mainParts. This is unique for every possible
     * interpretation of the nodes.
     */
    protected BigInteger hashCode;

    /** 
     * The management of this graph. 
     */
    protected StructuralConstraintManagement mySCM = null;

    // Constructors ***********************************************************

    /**
     * Creates a new instance of StructuralConstraintGraph
     */
    public StructuralConstraintGraph() {
    }

    // Methods ****************************************************************

    /**
     * Checks nodes that were newly created for creation or destruction of
     * StructuralConstraintInstances.
     * 
     * @param newNodes Nodes that were recently created.
     * @param deletedNodes Nodes that were deleted.
     * @param expandedNodes Expanded Nodes
     * @param reducedNodes Connected Nodes describing the expansions,
     *        assoziated to the nodes from expandedNodes.
     */
    public abstract void checkChangedNodes(Set newNodes, Set deletedNodes, 
                                           Set expandedNodes, Set reducedNodes);


    /**
     * Returns one of the mainParts. Called by
     * SCTPAC.getAnyConnectedMainPartFrame(). If the Graph has no MainPart the
     * NullMainPart is returned.
     * 
     * @return A mainPart.
     */
    public final StructuralConstraintGraphPart getOneMainPart() {
        return (StructuralConstraintGraphPart) mainParts.firstElement();
    }


    /**
     * Returns all interpretedNodes.
     * 
     * @return All the interpretations in this Part. Do not change the
     *         Collection, as these changes are reflected in the part!
     */
    public Map getInterpretedNodes() {
        return interpretedNodes;
    }


    /**
     * Enters an interpreatation of a node as allready used. The mapping from
     * structural constraint graph g to the CSP graph p must be injective,
     * i.e. no two different nodes from p may have the same interpretation in
     * p.
     * 
     * @param interpretation The interpretation.
     * @param node The node that gets the interpretation.
     */
    public void addInterpretation(Node interpretation, SCNode node) {
        SCNode nodeWithInterpretation = (SCNode) interpretedNodes.get(
                                                interpretation);

        if (nodeWithInterpretation == null) {
            interpretedNodes.put(interpretation, node);
        } else if (nodeWithInterpretation == node) {
            throw new Error(
                    "Interpretation was allready appointed for the node");
        } else if (node.getLinkedNode() != nodeWithInterpretation) {
            throw new Error("Interpretation allready used by other node");
        }
    }


    /**
     * Adds a new mainPart to the graph.
     * 
     * @param newPart The added Part.
     */
    public void addMainPart(StructuralConstraintGraphPart newPart) {
        mainParts.add(newPart);
    }


    /**
     * Adds one negative ApplicationCondition to the Graph.
     * 
     * @param nac A Negative Application Condition.
     */
    public void addNAC(ApplicationCondition nac) {
        NACs.add(nac);
    }


    /**
     * Adds one positive ApplicationCondition to the Graph.
     * 
     * @param pac A Positive Application Condition.
     */
    public void addPAC(ApplicationCondition pac) {
        PACs.add(pac);
    }


    /**
     * Is the given interpretation allready used in one of the parts of the
     * Graph?
     * 
     * @param interpretation The searched interpretation.
     * 
     * @return true if the interpretation was allready used, else false.
     */
    public boolean containsInterpretation(Node interpretation) {
        return interpretedNodes.get(interpretation) != null;
    }


    /**
     * Copies the interpretations of the other graph to this one. The two
     * graphs must be exactly identical in structure.
     * 
     * @param srcG the graph with the interpretations
     */
    public void copyInterpretations(StructuralConstraintGraph srcG) {
        Iterator srcIt;
        Iterator dstIt;

        if (mainParts != null) {
            srcIt = srcG.mainParts.iterator();
            dstIt = this.mainParts.iterator();

            while (srcIt.hasNext()) {
                StructuralConstraintGraphPart src = (StructuralConstraintGraphPart) srcIt.next();
                StructuralConstraintGraphPart dst = (StructuralConstraintGraphPart) dstIt.next();
                dst.copyInterpretations(src);
            }
        }

        if (PACs != null) {
            //same with PACs
            srcIt = srcG.PACs.iterator();
            dstIt = this.PACs.iterator();

            while (srcIt.hasNext()) {
                ApplicationCondition src = (ApplicationCondition) srcIt.next();
                ApplicationCondition dst = (ApplicationCondition) dstIt.next();
                dst.copyInterpretations(src);
            }
        }


        //copy seed interpretations of nacs
        srcIt = srcG.NACs.iterator();
        dstIt = this.NACs.iterator();

        while (srcIt.hasNext()) {
            ApplicationCondition src = (ApplicationCondition) srcIt.next();
            ApplicationCondition dst = (ApplicationCondition) dstIt.next();
            dst.copyInterpretations(src);
        }
    }


    /**
     * Is the given interpretation allready used in one of the MainParts of the
     * Graph?
     * 
     * @param interpretation The searched interpretation.
     * 
     * @return true if the interpretation was allready used, else false.
     */
    public boolean mainPartsContainInterpretation(Node interpretation) {
        Iterator it = mainParts.iterator();

        while (it.hasNext()) {
            StructuralConstraintGraphPart part = (StructuralConstraintGraphPart) it.next();

            if (part.containsInterpretation(interpretation)) {
                return true;
            }
        } //while

        return false;
    }


    /**
     * If an interpretation is not used any more it should be removed from list
     * of used interpretations by this function. The mapping from structural
     * constraint graph g to problem graph p must be injective, i.e. no two
     * different nodes from p may have the same interpretation in p.
     * 
     * @param interpretation The interpretation.
     * @param node The node that gets the interpretation.
     */
    public void removeInterpretation(Node interpretation, SCNode node) {
        if (interpretation != null) {
            SCNode nodeWithInterpretation = (SCNode) interpretedNodes.get(
                                                    interpretation);

            if (nodeWithInterpretation == node) {
                interpretedNodes.remove(interpretation);
            } else if (nodeWithInterpretation == null) {
                if (! node.hasLink()) { // if it has a link it might have happened that the linked node has removed the

                    // interpretation and therefore it can not be found here any more.
                    throw new Error("interpretation not part of subgraph!");
                }
            } else if ((nodeWithInterpretation.getLinkedNode() == node) || 
                           (node.getLinkedNode() == nodeWithInterpretation)) {
                // do not remove the interpretation.
            } else {
                throw new Error("Interpretation was not registered for node.");
            }
        }
    }


    /**
     * Resets the interpretations of all Main and PAC subgraphs
     */
    public void resetInterpretations() {
        Iterator it;

        if (mainParts != null) {
            it = this.mainParts.iterator();

            while (it.hasNext()) {
                StructuralConstraintGraphPart p = (StructuralConstraintGraphPart) it.next();
                p.resetInterpretations();
            }
        }

        if (PACs != null) {
            it = this.PACs.iterator();

            while (it.hasNext()) {
                StructuralConstraintGraphPart p = (StructuralConstraintGraphPart) it.next();
                p.resetInterpretations();
            }
        }
    }


    /**
     * Is the given interpretation allready used in one of the parts of the
     * Graph?
     * 
     * @param interpretation The searched interpretation.
     * @param node The node that gets the interpretation.
     * 
     * @return true if the interpretation was allready used, else false.
     */
    public boolean tryInterpretation(Node interpretation, SCNode node) {
        SCNode nodeWithInterpretation = (SCNode) interpretedNodes.get(
                                                interpretation);

        if (nodeWithInterpretation == null) {
            return false;
        } else if ((node != null) && 
                       ((node == nodeWithInterpretation) || 
                           (node.getLinkedNode() == nodeWithInterpretation))) {
            return false;
        }

        return true;
    }


    /**
     * Register the StructuralConstraintManagement;
     * 
     * @param scm  The new StructuralConstraintManagement.
     */
    void setSCManagement(StructuralConstraintManagement scm) {
        mySCM = scm;
    }
}