/*
 * StructuralConstraintGraphPart.java
 *
 * Created on 5. Dezember 2001, 15:53
 */
package com.ai_center.dragonbreath.structuralConstraintHandling.graph;

import com.ai_center.dragonbreath.*;
import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.applicationCondition.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.dockingGraph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.dockingGraph.applicationCondition.type.*;
import com.ai_center.dragonbreath.utilities.*;

import java.math.*;

import java.util.*;


/**
 * The representation of a graph of a StructuralConstraint.
 * 
 * @version $Revision: 1.7 $
 * @author tim baermann
 */
public abstract class StructuralConstraintGraphPart {
    // Instance variables *****************************************************

    /** 
     * To output debug informations. 
     */
    public DebugLogger dl;

    /** 
     * The nodes of the part with its interpretations as keys 
     */
    public HashMap interpretedNodes = new HashMap();

    /** 
     * The graph describtion this GraphPart belongs to. 
     */
    public StructuralConstraintGraph mySCGraph;

    /**
     * How many nodes of the graph have externInterpretation set? Set and reset
     * in SCNode.setInterpretationExtern() and
     * SCNode.resetInterpretationExtern()
     */
    public int numberExternInterpretations = 0;

    /**
     * The hashCode of this graph, computed by a concatenation of the hashcodes
     * of all interpretations of the graph
     */
    protected BigInteger hashCode;

    /** 
     * If the part was build from a GraphPartFrame this is it. 
     */
    protected GraphPartFrame usedGraphPartFrame = null;

    /**
     * All the SCNodes of this part that link to/ are linked by other Graphs.
     * (these are links between docking- and testgraph) With the Parts as
     * keys. GraphParts that are linked via a node to this part. (Can be
     * either a link from other parts SCNode to this ones nodes or reverse)
     * Filled via addInterGraphLink().
     */
    protected HashSet interGraphLinks = new HashSet();

    /**
     * The nodes of this ApplicationConditionPart, that are missing for a
     * match.
     */
    protected NodeContainer searchedNodes = new NodeContainer();

    /**
     * For use in getCopyOfInterpretations() when a node has no interpretation.
     */
    protected Object dummyInterpretation = new Object();

    /**
     * A dedicated node that serves as the comparison seed, from here the
     * search recursion can be started. Used especially for NACs and PACs as
     * they are often connected via a SCNode to their main graph, and such a
     * node is the seed.
     */
    protected SCNode seed;

    /**
     * All the SCNodes of this part that link to/ are linked by other parts.
     * With the Parts as keys. GraphParts that are linked via a node to this
     * part. (Can be either a link from other parts SCNode to this ones nodes
     * or reverse) Filled via addInnerGraphLink().
     */
    protected SetHashtable innerGraphLinks = new SetHashtable();

    /**
     * While searching for all mappings this is the stack of the actual partial
     * solution.
     */
    protected Stack graphSearchStack = new Stack();

    /** 
     * All the nodes of the Graph, that do not have a link. 
     */

    //protected Vector unlinkedNodes = new Vector();

    /** 
     * All the edges of this graph. 
     */
    protected Vector edges = new Vector();

    /**
     * All the nodes of this part that have themself a link. Not as in
     * interGraphLinks and innerGraphLinks where also nodes are stored that
     * are beeing linked.
     */
    protected Vector linkedNodes = new Vector();

    /** 
     * All the nodes of this graph. 
     */
    protected Vector nodes = new Vector();

    /**
     * After each call of findOneEmbedding / findNextEmbedding  this variable
     * holds whether during search it was tried to use an  interpretation,
     * that was allready used by an other graph part.
     */
    protected boolean intersectionWithOtherGraphPart = false;

    /**
     * When resetInterpretations() is called the method iterates through all
     * nodes, and calls resetInterpretation(). Normally that removes the
     * intepretation from the interpretedNodes list. But that would mean a
     * concurrent modification of the list. To prevent this the
     * resetInterpretations flag signalizes that the resetInterpretations is
     * executed, and removeInterpretation should behave differently.
     */
    protected boolean resetInterpretations = false;

    /** 
     * the number of edges that have an interpretation 
     */
    protected int interpretedEdgesSize = 0;

    /**
     * Is after a findOneEmbedding() or findNextEmbedding() a next trie
     * usefull? Are there branches to be explored?
     */
    protected int numberOfBranches;

    /**
     * Used in tryNextBranch. Membervariable, to not allways create new object.
     */
    SCNodesComparator nodesComparator = new SCNodesComparator();

    // Constructors ***********************************************************

    /**
     * Creates new StructuralConstraintGraphPart
     * 
     * @param scg The DockingGraph this Part should belong to.
     */
    public StructuralConstraintGraphPart(StructuralConstraintGraph scg) {
        mySCGraph = scg;
        dl        = new DebugLogger(true);
		
				
				// uncomment for debug output
				//dl.switchOn(12);
        //dl.switchOn(0);
        //dl.switchOn(1);
        //dl.switchOn(4);
    }

    // Methods ****************************************************************

    /**
     * When a new match for the Part is found during checkNodes() this function
     * performs the actions that have to be done.
     * 
     * @return true if further search is necessary. eg. if no more mainParts are
     *         missing.
     */
    public abstract boolean handleMatch();


    /**
     * Returns the number of actually interpreted edges + the number of
     * actually interpreted nodes.
     * 
     * @return interpretedNodes.size() + interpretedEdgesSize
     */
    public int getActualMatchSize() {
        return interpretedNodes.size() + interpretedEdgesSize;
    }


    /**
     * Returns all the interpretations used in this part.
     * 
     * @return All the interpretations in this Part. Do not change the
     *         Collection, as these changes are reflected in the part!
     * 
     * @throws Error If not all nodes are interpreted.
     */
    public Collection getAllInterpretations() {
        if (isInterpreted()) {
            return interpretedNodes.keySet();
        } else {
            throw new Error(
                    "getAllInterpretations() called although not all nodes have an interpretation assigned.");
        }
    }


    /**
     * Is the StructuralConstraintGraphPart a main part, or if it is an AC is
     * it connecting two mainParts?
     * 
     * @return true if this graphpart implements the ConnectingPartInterface.
     */
    public boolean isConnectingPart() {
        return (this instanceof ConnectingPartInterface);
    }


    /**
     * Sets the Interpretations as in the given Collection, the collection has
     * to be obtained from the same subgraph via getCopyOfInterpretations().
     * 
     * @param ints Interpretations that should be copied into this part.
     */
    public void setCopyOfInterpretations(Collection ints) {
        resetInterpretations();

        Iterator it     = nodes.iterator();
        Iterator intsIt = ints.iterator();

        while (it.hasNext()) {
            SCNode n = (SCNode) it.next();
            Object i = intsIt.next();

            if (i != dummyInterpretation) {
                if (! n.setInterpretationBiggestEmbedding((Node) i)) {
                    throw new Error("interpretationsCopy could not be set.");
                }
            }
        }

        it = edges.iterator();

        while (it.hasNext()) {
            SCEdge e = (SCEdge) it.next();
            Object i = intsIt.next();

            if (i != dummyInterpretation) {
                e.setInterpretation((ConnectedNode) i);
            } else {
                e.resetInterpretation();
            }
        }
    }


    /**
     * Returns a Collection with all the Interpretations of the Subgraph.
     * 
     * @return A Collection of Interpretations of this.
     */
    public Collection getCopyOfInterpretations() {
        Vector   result = new Vector();
        Iterator it = nodes.iterator();

        while (it.hasNext()) {
            SCNode n = (SCNode) it.next();
            Object i = n.peekInterpretation();

            if (i != null) {
                result.add(i);
            } else {
                result.add(dummyInterpretation);
            }
        }

        it = edges.iterator();

        while (it.hasNext()) {
            SCEdge e = (SCEdge) it.next();
            Object i = e.getInterpretation();

            if (i != null) {
                result.add(i);
            } else {
                result.add(dummyInterpretation);
            }
        }

        return result;
    }


    /**
     * Returns the edges Vector. Carefull, this is the original Vector of the
     * GraphPart.
     * 
     * @return edges
     */
    public Vector getEdges() {
        return edges;
    }


    /**
     * If the Graph has an interpretation this function returns a hash for the
     * graph.
     * 
     * @return Hash of the Interpretations. It is circa a concatenation of all
     *         the hashes of all the nodes and edges.
     * 
     * @throws Error If there was a node without interpretation.
     */
    public java.lang.String getHashCodeString() {
        java.lang.String hash = new java.lang.String();

        //  hash for nodes
        Iterator nodeIt         = nodes.iterator();
        Node     interpretation;

        while (nodeIt.hasNext()) {
            SCNode node = (SCNode) nodeIt.next();
            interpretation = node.interpretation;

            if (interpretation == null) {
                throw new Error(
                        "The graph contained a node without interpretation!");
            }

            hash = hash.concat(java.lang.String.valueOf(
                                       interpretation.hashCode()));
        }

        //hash for edges
        SCEdge   edge;
        long     h1; //HashCode edgeinterpretation1
        long     h2; //HashCode edgeinterpretation2
        Iterator edgeIt = edges.iterator();

        while (edgeIt.hasNext()) {
            edge = (SCEdge) edgeIt.next();

            if ((edge.interpretation1 == null) || 
                    (edge.interpretation2 == null)) {
                throw new Error(
                        "The graph contained an edge without interpretation!");
            }

            h1 = edge.interpretation1.hashCode();
            h2 = edge.interpretation2.hashCode();

            if (h1 < h2) { // dependend on the order of search the interpretations would else be set differently
                hash = hash.concat(java.lang.String.valueOf(h1));
            } else {
                hash = hash.concat(java.lang.String.valueOf(h2));
            }
        }

        hashCode = new BigInteger(hash);

        return hash;
    }


    /**
     * Returns usedGraphPartFrame or if not set calls createPartFrame(). Should
     * not be called if the Part has no interpretation.
     * 
     * @return A frame of the current interpretation.
     */
    public GraphPartFrame getInnerGraphFrame() {
        if (usedGraphPartFrame == null) {
            return createInnerGraphFrame();
        } else {
            return usedGraphPartFrame;
        }
    }


    /**
     * Returns all the nodes of this part connected with this part, or vice
     * versa.
     * 
     * @param part The part that should be connected.
     * 
     * @return All the nodes connected to the given part.
     */
    public Set getInnerGraphLinkNodes(StructuralConstraintGraphPart part) {
        Set    result;
        Object o = innerGraphLinks.get(part);

        try {
            result = (Set) o;
        } catch (java.lang.ClassCastException e) {
            result = new HashSet();
            result.add(o);
        }

        return result;
    } //getinnerGraphLinkNodes()


    /**
     * Returns all the nodes that connect to other parts of the same Graph
     * 
     * @return A Set of Nodes that are linked to or link to an other node.
     */
    public Set getInnerGraphLinkNodes() {
        return innerGraphLinks.values();
    }


    /**
     * Returns an Iterator over all the Nodes that have an connection to an
     * other part of this same Graph
     * 
     * @return innerGraphLinks.iterator()
     */
    public Iterator getInnerGraphLinkNodesIterator() {
        return innerGraphLinks.valuesIterator();
    }


    /**
     * Returns all the parts that are linked from or to this part, and are in
     * the same Graph
     * 
     * @return A Set of StructuralConstraintGraphParts.
     */
    public Set getInnerGraphLinkParts() {
        return innerGraphLinks.keySet();
    }


    /**
     * Do ALL the nodes have interpretations?
     * 
     * @return true if ALL the nodes and edges have interpretations.
     */
    public final boolean isInterpreted() {
        if (interpretedNodes.size() == nodes.size()) {
            Iterator edgeIt = edges.iterator();

            while (edgeIt.hasNext()) {
                SCEdge edge = (SCEdge) edgeIt.next();

                if ((edge.interpretation1 == null) || 
                        (edge.interpretation2 == null)) {
                    return false;
                }
            }

            return true;
        } else {
            return false;
        }
    }


    /**
     * Returns all interpretedNodes.
     * 
     * @return All the interpretations in this Part. Do not change the
     *         Collection, as these changes are reflected in the part!
     */
    public Map getInterpretedNodes() {
        return interpretedNodes;
    }


    /**
     * Returns an Iterator over the connectedGraphParts.
     * 
     * @return innerGraphLinks.iterator()
     */
    public Iterator getLinkedGraphPartsIterator() {
        return innerGraphLinks.iterator();
    }


    /**
     * After findBiggestEmbedding() this function returns the number of
     * elements (edges and nodes) that have an interpretation. (For TestGraph
     * NACs it is implemented differently =>  If a match was found the result
     * is the size of the subgraphPart, else 0.)
     * 
     * @return nodes.size() + edges.size() - getActualMatchSize().
     */
    public int getMinMatchError() {
        return (nodes.size() + edges.size()) - getActualMatchSize();
    }


    /**
     * Returns the node of the StructuralConstraintGraphPart that has the given
     * Node as interpretation, null if none.
     * 
     * @param interpretation  The searched interpretation.
     * 
     * @return The SCNode that has interpretation as interpretation or null if
     *         none exists.
     */
    public SCNode getNodeWithInterpretation(Node interpretation) {
        return (SCNode) interpretedNodes.get(interpretation);
    }


    /**
     * Returns all the nodes of the StructuralConstraintGraphPart. Returns the
     * Vector used in this part, so dont mess with it!
     * 
     * @return The nodes.
     */
    public Vector getNodes() {
        return nodes;
    }


    /**
     * Returns one of the parts in innerGraphLinks.
     * 
     * @return A graph part registered as connected to this one, or null if
     *         none exists.
     */
    public StructuralConstraintGraphPart getOneConnectedGraphPart() {
        Iterator it = innerGraphLinks.iterator();

        if (it.hasNext()) {
            return (StructuralConstraintGraphPart) it.next();
        } else {
            return null;
        }
    }


    /**
     * Is this StructuralConstraintGraphPart connected to any other part?
     * 
     * @return true if connected.
     */
    public boolean isPartConnected() {
        return innerGraphLinks.numberOfKeys() > 0;
    }


    /**
     * Returns all the nodes of this Part that have the same type as NodeType.
     * 
     * @param n The type.
     * 
     * @return All the nodes with the type n.
     */
    public Set getSearchedNodes(NodeType n) {
        return searchedNodes.getNodes(n);
    }


    /**
     * Sets the seed of this Graph, that is the node from which starts the
     * comparison. If a node of this graph gets a link to an other graph that
     * node is set automatically as the seed! The last linked node remains the
     * seed. If an other configuration is wished the seed has to be set after
     * all linking.
     * 
     * @param newSeed the node to be the seed.
     */
    public void setSeed(SCNode newSeed) {
        seed = newSeed;
    }


    /**
     * Returns the seed node.
     * 
     * @return this.seed
     */
    public SCNode getSeed() {
        return seed;
    }


    /**
     * Returns all the nodes of the GraphPart that are not yet interpreted. The
     * result is a copy. It can be freely modified.
     * 
     * @return All the nodes of this part that do not have an interpretation.
     */
    public Vector getUninterpretedNodes() {
        Vector result = new Vector(nodes);
        result.removeAll(interpretedNodes.values());

        return result;
    }


    /**
     * Adds an edge to the list of edges. Called by SCNode.connect()
     * 
     * @param newEdge The edge that should be added.
     */
    public void addEdge(SCEdge newEdge) {
        edges.add(newEdge);
    }


    /**
     * Registeres that one edge interpretation was added. increases
     * interpretedEdgesSize.
     */
    public void addEdgeInterpretation() {
        interpretedEdgesSize++;
    }


    /**
     * Registeres a part and a node to be linked by an other part. Called by
     * SCGraphFactory.getPartLinks()
     * 
     * @param part  The part that is linked.
     * @param node  A node from this part, that is either linked by a node from
     *        part, or links to a node from part.
     */
    public void addInnerGraphLink(StructuralConstraintGraphPart part, 
                                  SCNode node) {
        innerGraphLinks.add(part, node);
    }


    /**
     * Registeres a node that is linked by an other Graph, or links to an other
     * Graph. Called by SCGraphFactory.getPartLinks()
     * 
     * @param node  A node from this part, that is either linked by a node from
     *        part, or links to a node from an other
     *        StructuralConstraintGraph.
     */
    public void addInterGraphLink(SCNode node) {
        interGraphLinks.add(node);
    }


    /**
     * Enters an interpreatation of a node as allready used. The mapping from
     * structural constraint graph g to problem graph p must be injective,
     * i.e. no two different nodes from p may have the same interpretation in
     * p. (a link between two nodes means in this sense that they are the same
     * node, ergo they have to have the same interpretation)
     * 
     * @param interpretation The interpretation.
     * @param node  The node that gets the interpretation.
     */
    public void addInterpretation(Node interpretation, SCNode node) {
        SCNode nodeWithInterpretation = (SCNode) interpretedNodes.get(
                                                interpretation);

        if (nodeWithInterpretation == null) {
            interpretedNodes.put(interpretation, node);

            if (interpretedNodes.size() > nodes.size()) {
                throw new Error("too many interpretations");
            }

            mySCGraph.addInterpretation(interpretation, node);
        } else if (nodeWithInterpretation == node) {
            throw new Error(
                    "Interpretation was allready appointed for the node");
        } else {
            throw new Error("Interpretation allready used by other node");
        }
    }


    /**
     * Produces a new node that can be connected via connectNodes to others in
     * this graph
     * 
     * @param id the nodeID as defined in Registry
     * @param isLeafNode is this node the twin of an other node? Then it can be
     *        used to spare us a lot of work
     * 
     * @return the newly created SCNode
     */
    public SCNode addNode(int id, boolean isLeafNode) {
        SCNode newNode = new SCNode(this, id, isLeafNode);
        nodes.add(newNode);

        //unlinkedNodes.add(newNode);
        return newNode;
    }


    /**
     * Same as addNode, but sets the variables value.
     * 
     * @param value The Value of the variable.
     * @param isLeafNode Is the variable a leaf node?
     * 
     * @return The created node.
     */
    public SCNode addVariable(double value, boolean isLeafNode) {
        SCNode newNode = new SCNode(this, Registry.VARIABLE, isLeafNode);

        nodes.add(newNode);

        //unlinkedNodes.add(newNode);
        return newNode;
    }


    /**
     * Does the part contain any node of the given types?
     * 
     * @param nodes Nodes.
     * 
     * @return true if at least one found.
     */
    public boolean containsAnyNode(Collection nodes) {
        return searchedNodes.containsAnyNode(nodes);
    }


    /**
     * Same as tryInterpretation() but doesnt register intersection.
     * 
     * @param interpretation The searched interpretation.
     * 
     * @return true if interpretation was allready used, else false.
     */
    public boolean containsInterpretation(Node interpretation) {
        //overwritten in ApplicationCondition
        if (interpretedNodes.containsKey(interpretation)) {
            return true; // the node is used allready in same part.
        }

        // else check whether it it used by other parts 
        boolean result = mySCGraph.containsInterpretation(interpretation);

        return result;
    }


    /**
     * Copies the interpretations of the other graph to this one. If mainPart
     * copies the PACs and NACs too. The two graphs must be exactly identical
     * in structure.
     * 
     * @param source the graph with the interpretations
     * 
     * @throws Error If the graph are of different structure.
     */
    public void copyInterpretations(StructuralConstraintGraphPart source) {
        this.usedGraphPartFrame = source.usedGraphPartFrame;

        Iterator nodeIt         = nodes.iterator();
        Iterator sourceNodeIt   = source.nodes.iterator();
        Node     interpretation;

        while (nodeIt.hasNext()) { //copy node interpretations

            if (! sourceNodeIt.hasNext()) {
                throw new Error(
                        " tried copyInterpretations() for two graphs with different nodes");
            }

            SCNode node       = (SCNode) nodeIt.next();
            SCNode sourceNode = (SCNode) sourceNodeIt.next();


            //if (! node.hasLink()) { // no interpretation has to be copied if link exists
            interpretation = sourceNode.interpretation;
            node.setInterpretation(interpretation);

            if (interpretation == null) {
                throw new Error(
                        "The source graph contained a node without interpretation!");
            }

            //}
        } //while nodeIt

        Iterator      edgeIt             = edges.iterator();
        Iterator      sourceEdgeIt       = source.edges.iterator();
        SCEdge        edge;
        SCEdge        sourceEdge;
        ConnectedNode edgeInterpretation;

        while (edgeIt.hasNext()) { //copy edge interpretations

            if (! sourceEdgeIt.hasNext()) {
                throw new Error(
                        " tried copyInterpretations() for two graphs with different edges");
            }

            edge       = (SCEdge) edgeIt.next();
            sourceEdge = (SCEdge) sourceEdgeIt.next();

            if ((sourceEdge.interpretation1 == null) || 
                    (sourceEdge.interpretation2 == null)) {
                throw new Error(
                        "The source graph contained an edge without interpretation!");
            }

            edge.setInterpretation(sourceEdge.interpretation1);
            edge.setInterpretation(sourceEdge.interpretation2);
        } //while edgeIt
    } // public void copyInterpretations(StructuralConstraintGraphPart source)


    /**
     * Stores all the nodes of the given mainPart that connect to an other
     * mainPart via an application condition. Or if part is an
     * ApplicationConditionPart all the nodes that connect to mainParts. Sets
     * usedGraphPartFrame to the new created frame.
     * 
     * @return the created frame.
     */
    public GraphPartFrame createInnerGraphFrame() {
        GraphPartFrame result        = new GraphPartFrame(this);
        SCNode         ln;
        Iterator       linkedNodesIt = getInnerGraphLinkNodesIterator();

        while (linkedNodesIt.hasNext()) {
            ln = (SCNode) linkedNodesIt.next();

            if (ln.interpretation == null) {
                throw new Error("Node without Interpretation!");
            }

            result.enterNode(ln, ln.interpretation);
        }

        if (result.size() == 0) {
            result.enterNode(seed, seed.interpretation);
        }

        usedGraphPartFrame = result;

        return result;
    } //createInnerGraphFrame()


    /**
     * Stores all the nodes of the given mainPart that connect to an other
     * StructuralConstraintGraph. Normally in that case this is a part of a
     * DockingGraph, and the other Graphs are TestGraphs.
     * 
     * @return the created frame.
     */
    public GraphPartFrame createInterGraphFrame() {
        GraphPartFrame result        = new GraphPartFrame(this);
        SCNode         ln;
        Iterator       linkedNodesIt = interGraphLinks.iterator();

        while (linkedNodesIt.hasNext()) {
            ln = (SCNode) linkedNodesIt.next();
            result.enterNode(ln, ln.interpretation);
        }

        return result;
    } //createInterPartFrame()


    /**
     * Does the StructuralConstraintGraphPart have any interpretation?
     * 
     * @return true if the StructuralConstraintGraphPart has a single interpretation, else false.
     */
    public final boolean hasInterpretations() {
        return (interpretedNodes.size() + numberExternInterpretations) > 0;
    }


    /**
     * Occured during last call of findOneEmbedding() or findNextEmbedding() a
     * intersection with an other graph part? Especially interesting if no
     * embedding could be found. If no embedding could be found, and no
     * intersection occured and the embedding was searched from a
     * GraphPartFrame, the frame is not valid any more.
     * 
     * @return true if there was an intersection.
     */
    public final boolean intersectionWithOtherGraphPartOcurred() {
        return intersectionWithOtherGraphPart;
    }


    /**
     * Calls findBiggestEmbedding(findSeed())
     * 
     * @return true if a whole embedding could be found else false.
     */
    public boolean findBiggestEmbedding() {
        return findBiggestEmbedding(findSeed());
    }


    /**
     * Same as findOneEmbedding, but if no embedding found the biggest found
     * partly embedding is left as interpretation.
     * 
     * @param searchSeed  The beginning of the search.
     * 
     * @return true if a whole embedding could be found else false.
     */
    public boolean findBiggestEmbedding(SCNode searchSeed) {
        dl.print(1, 
                 "\nFindBiggestEmbedding " + this.hashCode() + 
                 " interpretedNodes: " + interpretedNodes.size() + "\n");

        if (searchSeed == null) {
            return false;
        }

        Node seedInterpretation = searchSeed.peekInterpretation();

        if (seedInterpretation == null) {
            dl.print(1, "no interpretation for searchSeed");

            return false;
        }

        searchSeed.setInterpretationBiggestEmbedding(seedInterpretation);

        Vector initialQueue = new Vector();
        initialQueue.add(new Pair(searchSeed, seedInterpretation)); //initiate the search from given node
        numberOfBranches = 0;
        graphSearchStack.clear();

        addAllLinkedNodesToQueue(initialQueue, searchSeed);

        tryNextBranchFindBiggestEmbedding(initialQueue);

        boolean found = isInterpreted();

        dl.print(1, "\nbranches: " + numberOfBranches);

        if (! found) {
            if (numberOfBranches == 0) {
                return false;
            }

            Collection biggestEmbeddingFound     = getCopyOfInterpretations();
            int        biggestEmbeddingFoundSize = getActualMatchSize();
            boolean    biggestEmbeddingStillSet  = true;

            while (! graphSearchStack.empty() && ! found) {
                if (numberOfBranches == 0) {
                    if (! biggestEmbeddingStillSet) {
                        setCopyOfInterpretations(biggestEmbeddingFound);
                    }

                    return found;
                }

                Pair actualQueuePair = (Pair) graphSearchStack.pop();

                if (actualQueuePair.first == null) {
                    // the node didnt have alternative interpretations, 
                    // reset the current interpretation and proceed to the previous search step.
                    if (actualQueuePair.second != null) {
                        Pair nodePair = (Pair) actualQueuePair.second;
                        ((SCNode) nodePair.first).resetInterpretationWithEdges();
                    }
                } else { // try a different branch

                    Vector actualQueue    = new Vector();
                    Pair   actualNodePair = (Pair) actualQueuePair.second;
                    SCNode actualNode     = (SCNode) actualNodePair.first;
                    Vector queuenodes     = (Vector) actualQueuePair.first;

                    // bring all the other nodes in the queue to their initial state and remove the actual node from queue.
                    // the actual node is anywhere in the queue, it has to be at the beginning
                    Iterator it = queuenodes.iterator();

                    while (it.hasNext()) {
                        SCNode n = (SCNode) ((Pair) it.next()).first;

                        if (n == actualNode) {
                            it.remove();
                        } else {
                            n.resetConAssignments();
                        }
                    }

                    actualQueue.add(actualNodePair); // make the branch point the first node to be checked from queue
                    actualQueue.addAll(queuenodes);
                    tryNextBranchFindBiggestEmbedding(actualQueue);

                    int actualMatchSize = getActualMatchSize();

                    if (actualMatchSize > biggestEmbeddingFoundSize) {
                        biggestEmbeddingFound     = getCopyOfInterpretations();
                        found                     = isInterpreted();
                        biggestEmbeddingStillSet  = true;
                        biggestEmbeddingFoundSize = actualMatchSize;
                    } else {
                        found                    = false;
                        biggestEmbeddingStillSet = false;
                    }

                    numberOfBranches--; //one branch used
                } // else try a different branche
            } //if     

            if (found) {
                dl.print(1, "\n Embedding found ");
            }
        }

        return found;
    } //findBiggestEmbedding()


    /**
     * Tests whether the ConstraintGraph can be embedded at this node with its
     * interpretation into the problem graph. Finds the next valid
     * isomorphism. <b>For the first embedding findOneEmbedding() has to be
     * called!!</b> Every time this function returns with true all the nodes
     * of the graph contain a new valid interpretation. If returns false the
     * interpretations of all nodes are set null, but all extern
     * interpretations stay valid. Returns false if no more embedding can be
     * found.
     * 
     * @return True if one more embedding was found, else false
     */
    public boolean findNextEmbedding() {
        if (interpretedNodes.size() > nodes.size()) {
            throw new Error("too many interpretations");
        }

        boolean found = false;
        dl.print(1, 
                 "\n\n findNextEmbedding " + this.hashCode() + 
                 " interpretedNodes: " + interpretedNodes.size() + "\n");

        if (graphSearchStack.empty()) {
            resetInterpretations();

            return false;
        }

        //Vector copyOfSearchStack = new Vector(graphSearchStack);
        while (! graphSearchStack.empty() && ! found) {
            if (numberOfBranches == 0) {
                graphSearchStack.clear();
                resetInterpretations();

                return false;
            }

            Pair actualQueuePair = (Pair) graphSearchStack.pop();

            if (actualQueuePair.first == null) {
                // the node didnt have alternative interpretations, 
                // reset the current interpretation and proceed to the previous search step.
                if (actualQueuePair.second != null) {
                    Pair nodePair = (Pair) actualQueuePair.second;
                    dl.print(1, 
                             "\nresetInterpretation " + 
                             ((SCNode) nodePair.first).peekInterpretation());
                    ((SCNode) nodePair.first).resetInterpretationWithEdges();
                }
            } else { // try a different branch

                Vector actualQueue    = new Vector();
                Pair   actualNodePair = (Pair) actualQueuePair.second;
                SCNode actualNode     = (SCNode) actualNodePair.first;
                Vector queuenodes     = (Vector) actualQueuePair.first;

                // bring all the other nodes in the queue to their initial state and remove the actual node from queue.
                // the actual node is anywhere in the queue, it has to be at the beginning
                Iterator it = queuenodes.iterator();

                while (it.hasNext()) {
                    SCNode n = (SCNode) ((Pair) it.next()).first;

                    if (n == actualNode) {
                        it.remove();
                    } else {
                        n.resetConAssignments();
                    }
                }


                //pushToSearchStack(actualNode, queuenodes);
                actualQueue.add(actualNodePair); // make the branch point the first node to be checked from queue
                actualQueue.addAll(queuenodes);
                found = tryNextBranch(actualQueue);

                numberOfBranches--; //one branch used
            } // else try a different branche
        } //while

        if (! found) {
            if (interpretedNodes.size() > 0) {
                //throw new Error("No match found, but interpreted Nodes left.");
                resetInterpretations();
            }

            graphSearchStack.clear();
            dl.print(1, "\n No Embedding found ");
        }

        if (found) {
            dl.print(1, "\n Embedding found(");
            dl.printNodeList(1, interpretedNodes);
            dl.print(1, ")");
        }

        return found;
    } //findNextEmbedding()


    /**
     * Tests whether the ConstraintGraph can be embedded at the node seed into
     * the problem graph. Finds only one isomorphism, starting from a start
     * seed.  Seed can be either the dedicated seed, or if that doesnt have
     * an interpretation any linked Node with an interpretation via the link.
     * If none found the root seed interpretation is reset.
     * 
     * @return true if one Embedding was found.
     */
    public boolean findOneEmbedding() {
        return _findOneEmbedding(findSeed());
    }


    /**
     * Tests whether the ConstraintGraph can be embedded at the given node into
     * the problem graph. Finds only one isomorphism. If none found the root
     * seed interpretation is reset. But all extern interpretations stay
     * valid.
     * 
     * @param searchSeed  The seed the search starts from
     * 
     * @return true if one Embedding was found.
     */
    public boolean findOneEmbedding(SCNode searchSeed) {
        return _findOneEmbedding(searchSeed);
    }


    /**
     * Does the part have any links?
     * 
     * @return linkedNodes.size() > 0.
     */
    public boolean hasInnerGraphLinks() {
        return innerGraphLinks.size() > 0;
    }


    /**
     * Are there links to other Graphs?
     * 
     * @return interGraphLinks.size() > 0
     */
    public boolean hasInterGraphLinks() {
        return interGraphLinks.size() > 0;
    }


    /**
     * Is a further call of findNextEmbedding() usefull?
     * 
     * @return numberOfBranches > 0
     */
    public boolean hasMorePossibleEmbeddings() {
        return numberOfBranches > 0;
    }


    /**
     * How many other parts connect to this part?
     * 
     * @return Number of parts connected to this part.
     */
    public int numberOfConnectedParts() {
        int result = innerGraphLinks.numberOfKeys();

        return result;
    }


    /**
     * Is the given interpretation used in this graph part?
     * 
     * @param interpretation The interpretation.
     * 
     
     * @return was the interpretation found?
     */
    public boolean partContainsInterpretation(Node interpretation) {
        return interpretedNodes.containsKey(interpretation);
    }


    /**
     * Prints the nodes and their interpretation.
     */
    public void printNodes() {
        dl.printNodeList(2, interpretedNodes);
    }


    /**
     * Stores a node as linked, called in SCNode.setLink()
     * 
     * @param linkedNode A node that has a link.
     */
    public void registerLinkedNode(SCNode linkedNode) {
        linkedNodes.add(linkedNode);
    }


    /**
     * Decreses interpretedEdgesSize.
     */
    public void removeEdgeInterpretation() {
        interpretedEdgesSize--;

        if (interpretedEdgesSize < 0) {
            interpretedEdgesSize = 0;
        }
    }


    /**
     * If an interpretation is not used any more it should be removed from list
     * of used interpretations by this function. The mapping from structural
     * constraint graph g to problem graph p must be injective, i.e. no two
     * different nodes from p may have the same interpretation in p.
     * 
     * @param interpretation The interpretation.
     * @param node The node that should get the interpretation.
     */
    public void removeInterpretation(Node interpretation, SCNode node) {
        if (interpretation != null) {
            SCNode nodeWithInterpretation = (SCNode) interpretedNodes.get(
                                                    interpretation);

            if (nodeWithInterpretation == node) {
                if (! resetInterpretations) {
                    interpretedNodes.remove(interpretation);
                }

                mySCGraph.removeInterpretation(interpretation, node);
            } else if (nodeWithInterpretation == null) {
                throw new Error("interpretation not part of subgraph!");
            }
        }
    }


    /**
     * Resets the interpretations of all nodes and edges in the graph.
     */
    public void resetInterpretations() {
        if (interpretedNodes.size() > 0) {
            //  make copy, as the node.resetInterpretation() functions also access interpretedNodes, 
            // and that will result in a ConcurrentModificationException
            resetInterpretations = true;

            Iterator nodeIt = interpretedNodes.values().iterator();

            while (nodeIt.hasNext()) {
                SCNode node = (SCNode) nodeIt.next();
                node.resetInterpretation();
                nodeIt.remove(); // <- important, remove called after node.resetInterpretation()
            }

            Iterator edgeIt = edges.iterator();

            while (edgeIt.hasNext()) {
                SCEdge edge = (SCEdge) edgeIt.next();
                edge.resetInterpretation();
            }
        }

        resetInterpretations = false;
        interpretedEdgesSize = 0;
    } //void resetInterpretations()


    /**
     * Resets the extern interpretations of all nodes and edges in the graph.
     */
    public void resetInterpretationsExtern() {
        Iterator nodeIt = nodes.iterator();

        while (nodeIt.hasNext()) {
            SCNode node = (SCNode) nodeIt.next();
            node.resetInterpretationExtern();
        }
    }


    /**
     * No intersection is registered after this call. Called before the first
     * node of this graph gets an assignment in
     * SCNode.computeAllPossibleConAssignments() .
     */
    public void resetIntersection() {
        intersectionWithOtherGraphPart = false;
    }


    /**
     * Returns the number of nodes and edges of the part.
     * 
     * @return nodes.size()
     */
    public int size() {
        return nodes.size() + edges.size();
    }


    /**
     * Given a Collection of Nodes it sorts the nodes in the collection after
     * the frequency  of the given type in the searchedNodes container. ( Used
     * eg. in ApplicationCondition.checkChangedNodes() )
     * 
     * @param nodes  The nodes to be sorted first the less frequent then the
     *        frequent ones
     */
    public void sortNodesByFrequency(Collection nodes) {
        searchedNodes.sortNodesByFrequency(nodes);
    }


    /**
     * Is the given interpretation allready used in this graph? (If the graph
     * contains multiple parts all are tested.) The mapping from structural
     * constraint graph g to problem graph p must be injective, i.e. no two
     * different nodes from g may have the same interpretation in p.
     * Registeres an Intersection if true.
     * 
     * @param interpretation The interpretation.
     * @param node The node that should get the interpretation.
     * 
     * @return Is the interpretation used in graph?
     */
    public boolean tryInterpretation(Node interpretation, SCNode node) {
        //overwritten in ApplicationCondition
        SCNode nodeWithInterpretation = (SCNode) interpretedNodes.get(
                                                interpretation);

        if ((nodeWithInterpretation != null) && 
                (nodeWithInterpretation != node)) {
            return true; // the node is used allready in same part.
        }

        // else check whether it it used by other parts 
        boolean result = mySCGraph.tryInterpretation(interpretation, node);

        if (result == true) {
            intersectionWithOtherGraphPart = true;
        }

        return result;
    } //boolean tryInterpretation(Node i)


    /**
     * Adds the linked nodes to the given queue called in
     * findBiggestEmbedding().
     * 
     * @param initialQueue A Vector with one Pair(actualSeed, interpretation)
     *        entrance.
     * @param actualSeed The seed of the GraphPart.
     */
    protected void addAllLinkedNodesToQueue(Vector initialQueue, 
                                            SCNode actualSeed) {
        Iterator it = linkedNodes.iterator();

        while (it.hasNext()) {
            SCNode node = (SCNode) it.next();

            if (node != actualSeed) {
                Node interpretation = node.peekInterpretation();

                if (interpretation != null) {
                    initialQueue.add(new Pair(node, interpretation));
                }
            }
        }
    }


    /**
     * Finds a valid seed to start the findOneEmbedding search from. Either the
     * dedicated seed, or if that has no interpretation any node with a link
     * that gets an interpretation via the link.
     * 
     * @return A seed node.
     */
    protected SCNode findSeed() {
        SCNode result = null;

        if ((seed == null) || (seed.peekInterpretation() == null)) {
            Iterator it = this.getInnerGraphLinkNodesIterator();

            while (it.hasNext()) {
                result = (SCNode) it.next();

                if (result.peekInterpretation() != null) {
                    return result;
                }
            }

            it = this.interGraphLinks.iterator();

            while (it.hasNext()) {
                result = (SCNode) it.next();

                if (result.peekInterpretation() != null) {
                    return result;
                }
            }
        } else {
            return seed;
        }

        return null;
    }


    /**
     * Pushes a given node to the graphSearchStack.
     * 
     * @param n The node for the stack.
     * @param queue The searchqueue at the moment the node n was interpreted,
     *        and pushed on the stack. it is saved with the node.
     */
    protected final void pushToSearchStack(SCNode n, Vector queue) {
        // are there other ways the edges can be interpreted, so save the node and the actual queue
        if ((n.numberOfMissingConAssignments() > 0) && (queue != null)) {
            // further search in findNextEmbedding()
            Pair nextBranch = new Pair(new Vector(queue), 
                                       new Pair(n, n.interpretation));
            graphSearchStack.push(nextBranch);
            numberOfBranches++;

            if (n.interpretation == null) {
                throw new Error("no interpretation for node");
            }
        } else { // enter node to stack ( only to know when to reset it again )
            graphSearchStack.push(new Pair(null, new Pair(n, n.interpretation)));
        }
    }


    /**
     * Takes a queue as used in tryNextBranch and resets the interpretations of
     * all entries.
     * 
     * @param queue A Vector with Pair( SCNode, Node) entries.
     */
    protected void resetInterpretationsInQueue(Vector queue) {
        Iterator it = queue.iterator();

        while (it.hasNext()) {
            Pair   nodePair = (Pair) it.next();
            SCNode node = (SCNode) nodePair.first;
            dl.print(1, 
                     "\nresetInterpretation " + 
                     ((SCNode) nodePair.first).peekInterpretation());
            node.resetInterpretationWithEdges();
        } //while    
    }


    /**
     * Travels the graph with "branch and bound" and assigns interpretations
     * for all graph parts. If for the actual part there are multiple ways of
     * interpreting them the other interpretations are pushed on the
     * graphSearchStack, and explored later in findMatchesWithPart via
     * backtracking.
     * 
     * @param queue The actual search queue.
     * 
     * @return Was an interpretation found?
     */
    protected boolean tryNextBranch(Vector queue) {
        dl.print(1, "\n TryNextBranch");

        HashSet  nodesInQueue = new HashSet();
        Iterator qit = queue.iterator();

        while (qit.hasNext()) {
            Pair o = (Pair) qit.next();
            nodesInQueue.add(o.first);
        } //while

        while (queue.size() > 0) {
            // actualNode = queue.head()
            Pair actualElement = (Pair) queue.firstElement();
            queue.remove(0);

            SCNode actualNode = (SCNode) actualElement.first;

            if (actualNode.interpretation == null) { // the node was still in the queue from last call

                if (! actualNode.setInterpretation((Node) actualElement.second)) {
                    dl.print(1, 
                             "\n interpretation could not be set" + 
                             ((Node) actualElement.second).toString() + 
                             " ( " + interpretedNodes.size() + " )");


                    //now all nodes that still are in the queue have to be reset
                    resetInterpretationsInQueue(queue);
                    printGraphStack();

                    return false;
                }
            }

            if ((actualNode.interpretation != null) && 
                    (interpretedNodes != null)) {
                dl.print(1, 
                         "\n" + actualNode.interpretation.toString() + " ( " + 
                         interpretedNodes.size() + " )" + "q(" + 
                         queue.size() + ")cons:" + 
                         actualNode.numberOfMissingConAssignments());
            }

            Vector nextNodes = actualNode.getAdjacency();

            if (nextNodes == null) { // there was no valid Adjacency found
                dl.print(1, 
                         "\nresetInterpretation " + 
                         actualNode.peekInterpretation());
                actualNode.resetInterpretationWithEdges();


                //now all nodes that still are in the queue have to be reset
                resetInterpretationsInQueue(queue);

                dl.print(1, "\n NO valid Adjacency found.");
                printGraphStack();

                return false;
            } else { // there exists an Adjacency
                pushToSearchStack(actualNode, queue);

                Iterator nodesIt = nextNodes.iterator();

                // enter all the nodes in the search queue
                if (nodesIt.hasNext()) {
                    while (nodesIt.hasNext()) {
                        SCNode newNode = (SCNode) nodesIt.next();

                        if (! nodesInQueue.contains(newNode)) {
                            queue.add(new Pair(newNode, newNode.interpretation));
                            dl.print(1, 
                                     "\n          " + 
                                     newNode.interpretation.toString());
                            nodesInQueue.add(newNode);
                        }
                    }

                    if (StructuralConstraintManagement.graphSearchSortQueueOn) {
                        Collections.sort(queue, nodesComparator); //for the least probable nodes to be tested first. do not enter at beginning of loop, as the first node in queue might be branch point.
                    }
                }
            } //else there exists an Adjacency
        } //while queue

        if (! isInterpreted()) {
            throw new Error(
                    "Not all Nodes found, probably graph not contiguous. ");
        }

        printGraphStack();

        return true;
    } //tryNextBranch()


    /**
     * Travels the graph with "branch and bound" and assigns interpretations
     * for all graph parts. If for the actual part there are multiple ways of
     * interpreting them the other interpretations are pushed on the
     * graphSearchStack, and explored later in findMatchesWithPart via
     * backtracking. Called from findBiggestEmbedding(). In contrast to
     * tryNextBranch() this function also develops further, if a node is not
     * totally compatible, i.e. if not all edges exist.
     * 
     * @param queue The actual search queue.
     */
    protected void tryNextBranchFindBiggestEmbedding(Vector queue) {
        dl.print(1, "\n TryNextBranchBiggestEmbedding");

        HashSet  nodesInQueue = new HashSet();
        Iterator qit = queue.iterator();

        while (qit.hasNext()) {
            Pair o = (Pair) qit.next();
            nodesInQueue.add(o.first);
        } //while

        while (queue.size() > 0) {
            Pair actualElement = (Pair) queue.firstElement();
            queue.remove(0);

            SCNode actualNode = (SCNode) actualElement.first;

            if (actualNode.interpretation == null) {
                if (! actualNode.setInterpretationBiggestEmbedding(
                              (Node) actualElement.second)) {
                    resetInterpretationsInQueue(queue);
                    dl.print(1, 
                             "\n interpretation could not be set" + 
                             ((Node) actualElement.second).toString() + 
                             " ( " + interpretedNodes.size() + " )");

                    return;
                }
            }

            if ((actualNode.interpretation != null) && 
                    (interpretedNodes != null)) {
                dl.print(1, 
                         "\n" + actualNode.interpretation.toString() + " ( " + 
                         interpretedNodes.size() + " )");
            }

            Vector nextNodes = actualNode.getAdjacencyBiggestEmbedding();
            pushToSearchStack(actualNode, queue);

            if (nextNodes == null) { // there was no valid Adjacency found
                dl.print(1, 
                         "\nresetInterpretation " + 
                         actualNode.peekInterpretation());
                actualNode.resetInterpretationWithEdges();
                resetInterpretationsInQueue(queue);
                dl.print(1, "\n NO valid Adjacency found.");

                return;
            } else { // there exists an Adjacency

                Iterator nodesIt = nextNodes.iterator();

                // enter all the nodes in the search queue
                if (nodesIt.hasNext()) {
                    while (nodesIt.hasNext()) {
                        SCNode newNode = (SCNode) nodesIt.next();

                        if (! nodesInQueue.contains(newNode)) {
                            queue.add(new Pair(newNode, newNode.interpretation));
                            dl.print(1, 
                                     "\n          " + 
                                     newNode.interpretation.toString());
                            nodesInQueue.add(newNode);
                        }
                    }

                    if (StructuralConstraintManagement.graphSearchSortQueueOn) {
                        Collections.sort(queue, nodesComparator); //for the least probable nodes to be tested first. do not enter at beginning of loop, as the first node in queue might be branch point.
                    }
                }
            } //else there exists an Adjacency

            //do not enter at beginning of loop, as the first node in queue might be branch point.
        } //while queue

        printGraphStack();
    } //tryNextBranchFindBiggestEmbedding()


    /**
     * Prints the contents of the graphstack
     */
    void printGraphStack() {
        dl.print(1, "\n GraphStack:");

        Iterator it = graphSearchStack.iterator();

        while (it.hasNext()) {
            Pair   actualQueuePair = (Pair) it.next();
            Pair   actualNodePair = (Pair) actualQueuePair.second;
            Node   actualNode     = (Node) actualNodePair.second;
            SCNode actualSCNode   = (SCNode) actualNodePair.first;
            dl.print(1, "\n\t" + actualNode);

            dl.print(1, 
                     " cons:" + actualSCNode.numberOfConAssignments() + 
                     " q(");

            //print the queue
            if (actualQueuePair.first != null) {
                Vector   queue      = (Vector) actualQueuePair.first;
                Vector   queuenodes = new Vector();
                Iterator qit        = queue.iterator();

                while (qit.hasNext()) {
                    Node n = (Node) ((Pair) qit.next()).second;
                    queuenodes.add(n);
                }

                dl.printNodeList(1, queuenodes);
            }

            dl.print(1, ")");
        } //while

        dl.print(1, "\n");
    }


    /**
     * Same as findOneEmbedding. Method used to prevent overwriting in
     * childclasses.
     * 
     * @param searchSeed This is where the search starts.
     * 
     * @return Could one Embedding be found?
     */
    private boolean _findOneEmbedding(SCNode searchSeed) {
        dl.print(1, 
                 "\nFindOneEmbedding " + this.hashCode() + 
                 " interpretedNodes: " + interpretedNodes.size() + "\n");

        if (searchSeed == null) {
            return false;
        }

        if (isInterpreted()) {
            return true;
        }

        Node seedInterpretation = searchSeed.getInterpretation();

        if (seedInterpretation == null) {
            dl.print(1, "no interpretation for searchSeed");

            return false;
        }

        Vector initialQueue = new Vector();
        initialQueue.add(new Pair(searchSeed, seedInterpretation)); //initiate the search from given node
        numberOfBranches = 0;
        graphSearchStack.clear();

        boolean found = tryNextBranch(initialQueue);

        dl.print(1, "\nbranches: " + numberOfBranches);

        if (! found) {
            found = findNextEmbedding();
        }

        if (found) {
            dl.print(1, " Embedding found ");
        }

        return found;
    } //findOneEmbedding()
}