package com.ai_center.dragonbreath.structuralConstraintHandling.graph.applicationCondition;

import com.ai_center.dragonbreath.constraintgraph.Node;
import com.ai_center.dragonbreath.structuralConstraintHandling.StructuralConstraintInstance;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.dockingGraph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.movableSubgraph.*;
import com.ai_center.dragonbreath.utilities.SetHashtable;

import java.util.*;


/**
 * Basic functionality for ApplicationConditions. Each PAC or NAC is derived
 * from this class.
 * 
 * @version $Revision: 1.3 $
 * @author tim baermann
 */
public class ApplicationCondition extends MovableSubgraph {
    // Instance variables *****************************************************

    /**
     * The number of MainParts connected to all the parts of this
     * ApplicationCondition. Set during build in SCGraphFactory.
     */
    public int numberOfConnectedMainParts; // not simply computed from all the numberOfConnectedMainParts of all the parts, as they might connect to the same parts.

    /**
     * All the MainParts that leck one of the parts of this
     * ApplicationCondition.
     */
    PartFrameContainer mainPartsMissingAnACPart = new PartFrameContainer();

    /**
     * The StructuralConstraintInstance this Application Condition belongs to.
     */
    StructuralConstraintInstance myInstance;

    /**
     * One of the MainParts connected to this AC, or if none any mainPart in
     * Graph
     */
    private StructuralConstraintGraphPart aConnectedMainPart;

    // Constructors ***********************************************************

    /**
     * Creates a new instance of ApplicationCondition.
     * 
     * @param scg The StructuralConstraintGraph, a DockingGraph, this part
     *        belongs to.
     */
    public ApplicationCondition(StructuralConstraintGraph scg) {
        super(scg);
    }

    // Methods ****************************************************************

    /**
     * Returns an MainPartFrame connected to any Part of the AC. If no MainPart
     * is connected to the PAC any MainPart is chosen, and if no
     * MainPartExists  null is returned.
     * 
     * @return A Frame of a MainPart that is connected to the AC.
     */
    public GraphPartFrame getAnyConnectedMainPartFrame() {
        if (aConnectedMainPart == null) {
            StructuralConstraintGraphPart mainPart = null;
            Iterator                      it = allParts.iterator();

            while (it.hasNext() && (mainPart == null)) {
                MovableSubgraphPart part = (MovableSubgraphPart) it.next();
                mainPart = part.getOneConnectedGraphPart();
            }

            if (mainPart == null) { // no part is connected to a mainPart
                mainPart = mySCGraph.getOneMainPart();
            }

            aConnectedMainPart = mainPart;
        }

        return aConnectedMainPart.getInnerGraphFrame();
    }


    /**
     * If the Graph has an interpretation this function returns a hash for the
     * Part.  Movable Subgraphs are either Application Condition- or
     * TestgraphParts. Tes
     * 
     * @return Hash of the Interpretations. It is circa a concatenation of all
     *         the hashes of all the linked nodes.
     * 
     * @throws Error If there was a node without interpretation.
     */
    public java.lang.String getHashCodeString() {
        java.lang.String hash           = new java.lang.String();
        Node             interpretation;

        // for all parts..
        //  hash, only uses linked nodes
        Iterator nodeIt = getInnerGraphLinkNodesIterator();

        while (nodeIt.hasNext()) {
            SCNode node = (SCNode) nodeIt.next();
            interpretation = node.getInterpretation();

            if (interpretation == null) {
                throw new Error(
                        "The graph contained a node without interpretation!");
            }

            hash = hash.concat(java.lang.String.valueOf(
                                       interpretation.hashCode()));
        }

        nodeIt = interGraphLinks.iterator();

        while (nodeIt.hasNext()) {
            SCNode node = (SCNode) nodeIt.next();
            interpretation = node.getInterpretation();

            if (interpretation == null) {
                throw new Error(
                        "The graph contained a node without interpretation!");
            }

            hash = hash.concat(java.lang.String.valueOf(
                                       interpretation.hashCode()));
        }

        return hash;
    } //getHashCodeString()


    /**
     * Returns all nodes that have a link assigned
     * 
     * @return The Nodes of linked nodes in this graph
     */
    public Set getInnerGraphLinkNodes() {
        return null;
    }


    /**
     * Returns allways true, is never called.
     * 
     * @return true
     * 
     * @throws Error if called..
     */
    public boolean handleMatch() {
        throw new Error(
                "handleMatch should be implemented in concrete ApplicationCondition");
    }


    /**
     * Used in type ApplicationConditions. When all the MainParts have their
     * nodes assigned this method tests whether the whole AC can be embedded
     * in the graph. If false: the missing part is registered for search. uses
     * any part connected to the wrong AC.
     * 
     * @return All parts could be found.
     */
    public boolean testAC() {
        return testAC(getAnyConnectedMainPartFrame());
    }


    /**
     * Tests whether this part can be embedded. Is called, when at least one
     * MainPart has allready an embedding. The given
     * 
     * @param mainPartFrame  A valid mainPartFrame.
     * 
     * @return is the AC valid.
     */
    public boolean testAC(GraphPartFrame mainPartFrame) {
        boolean found = findAllParts();
        resetInterpretations();

        return found;
    }


    /**
     * Dummy, is implemented in the Subclasses.  Is called to check whether
     * this AC is fullfilled with the given interpretation for the0 MainPart
     * in mainPartFrame.
     * 
     * @param mainPartFrame  The interpretation for one MainPart, normally one
     *        connected to this AC.
     * 
     * @return true.
     */
    public boolean testPartAC(GraphPartFrame mainPartFrame) {
        return true;
    }


    /**
     * The number of mainParts all the parts of this application condition are
     * linked to.
     * 
     * @return Number of mainParts this application condition is linked to.
     */
    int numberOfConnectedMainParts() {
        return numberOfConnectedMainParts;
    }
} //class
