/*
 * MainPart.java
 *
 * Created on 25. Juni 2002, 23:27
 */
package com.ai_center.dragonbreath.structuralConstraintHandling.graph.dockingGraph;

import com.ai_center.dragonbreath.constraintgraph.Node;
import com.ai_center.dragonbreath.structuralConstraintHandling.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.applicationCondition.ApplicationCondition;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.dockingGraph.applicationCondition.instance.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.dockingGraph.applicationCondition.type.*;
import com.ai_center.dragonbreath.utilities.SetHashtable;

import java.util.*;


/**
 * A mainPart of a DockingGraph. It has to be found for each match.
 * 
 * @version $Revision: 1.5 $
 * @author tim baermann
 */
public class DockingMainPart extends StructuralConstraintGraphPart
    implements ConnectingPartInterface, MainPartInterface, NodeChangedInterface {
    // Instance variables *****************************************************

    /** 
     * The NACs of this graph 
     */
    protected HashSet NACs = new HashSet();

    /** 
     * The PACs of this graphpart 
     */
    protected HashSet PACs = new HashSet();

    /** 
     * All the PACParts that are connected to this mainPart and to others. 
     */
    protected HashSet connectingPACParts = new HashSet();

    /** 
     * All the embeddings found for this part. 
     */
    protected PartFrameContainer foundMatches = new PartFrameContainer();

    /**
     * Used in checkChangedNodes() to garantie that nodes allready checked are
     * not used again in embeddings.
     */
    HashSet allreadyCheckedNewNodes = new HashSet();

    /**
     * GraphPartFrames that were found but are still missing an
     * ApplicationCondition
     */
    PartFrameContainer partlyFoundMatches = new PartFrameContainer();

    /**
     * While a new Frame is tested, and matches for the whole graph are
     * searched in handleMatch() the DockingGraph calls existsFrame() but it
     * may be that the tested frame is the first one, and so there exists no
     * frame yet in foundMatches, then this variable signals that there is a
     * frame but not yet entered.
     */
    boolean checkingNewFrame = false;

    /**
     * While search for new part match in checkNewNodes() this variable is set
     * true when the part is searched, and false when it was found, and the
     * rest of the graph is searched. All the matches with the nodes in
     * allreadyCheckedNewNodes are known to be allready found. So a match with
     * these nodes can not lead to a new match  => stop search by forbidding
     * the interpretation in containsInterpretation() / tryInterpretation().
     */
    boolean searchWithoutChecked = false;

    // Constructors ***********************************************************

    /**
     * Creates a new instance of MainPart
     * 
     * @param scg The graph the MainPart belongs to.
     */
    public DockingMainPart(StructuralConstraintGraph scg) {
        super(scg);

        //dl.switchOn(3);// uncomment for debug output
        //dl.switchOn(4);// uncomment for debug output
    }

    // Methods ****************************************************************

    /**
     * Returns all the frames of the part that fit to the actual interpretation
     * of the linked nodes of this part i.e. have the same interpretaiton.
     * 
     * @return All the found frames.
     */
    public Stack getFittingFrames() {
        Stack    result             = new Stack();
        Set      allInterpretations = foundMatches.getAll();
        Iterator interpretIt        = allInterpretations.iterator();

        while (interpretIt.hasNext()) {
            GraphPartFrame newInterpretations = (GraphPartFrame) interpretIt.next();

            if (newInterpretations.checkInterpretations()) {
                result.push(newInterpretations);
            }
        } //while

        return result;
    }


    /**
     * Adds one ApplicationCondition to the list of PACs or NACs.
     * 
     * @param ac An ApplicationCondition.
     */
    public void addAC(ApplicationCondition ac) {
        if ((ac instanceof TypeDockingPAC) || 
                (ac instanceof InstanceDockingPAC)) {
            PACs.add(ac);
        } else { // either this is docking nac, or for testGraphs there are only nacs
            NACs.add(ac);
        }
    }


    /**
     * Adds one PACPart to the list of the connectingPACParts.
     * 
     * @param pacPart A PAC that connects this MainPart with an other one.
     */
    public void addConnectingPACPart(TypeDockingConnectingPACPart pacPart) {
        connectingPACParts.add(pacPart);
    }


    /**
     * Produces a new node that can be connected via connectNodes to others in
     * this graph Also adds the node to the searched ones.
     * 
     * @param id the nodeID as defined in Registry
     * @param isLeafNode
     * 
     * @return the newly created SCNode
     */
    public SCNode addNode(int id, boolean isLeafNode) {
        SCNode result = super.addNode(id, isLeafNode);

        if (! isLeafNode) {
            searchedNodes.addNode(result);
        }

        return result;
    }


    /**
     * Checks wheter with the given nodes a new mainPart can be found.
     * 
     * @param newNodes  Nodes that were recently created.
     * @param deletedNodes  The nodes deleted.
     * @param expandedNodes  Expanded Nodes.
     * @param reducedNodes  The reduced Nodes.
     */
    public void checkChangedNodes(Set newNodes, Set deletedNodes, 
                                  Set expandedNodes, Set reducedNodes) {
        dl.print(3, "\n nodes: " + nodes.size() + " edges: " + edges.size());
        checkNewNodes(newNodes);
        checkNewNodes(expandedNodes);
    }


    /**
     * Sets the interpretations as in the frame, then tries all the embeddings
     * possible with the frame, until an embedding is found with wich the  all
     * PACs are valid, or no more embeddings are possible, then saves the
     * frame with partlyFoundMatches or foundMatches. Is called when a PAC has
     * got valid, and the frame for the MainPart match  can be tried again, if
     * it is valid now.
     * 
     * @param frame  A GraphPartFrame for this MainPart. it should be contained
     *        in partlyFoundParts.
     */
    public void checkFrame(GraphPartFrame frame) {
        boolean found = false;

        if (hasInterpretations()) {
            resetInterpretations();
        }

        if (frame.setInterpretations()) {
            if (findOneEmbedding()) {
                found = handleMatch();

                while (findNextEmbedding()) {
                    // either check all the embeddings, and create whole graphs with them, or find one match with the ACs fullfilled.
                    found |= handleMatch();
                } // while
            } // if findOneEmbedding

            frame.resetInterpretations();
        }

        if (found) {
            // is this ok?? there may still be matches missing with the given frame..
            partlyFoundMatches.remove(frame);
            foundMatches.add(frame);
        }
    } // checkFrame()


    /**
     * Checks for this mainPart if all the ApplicationConditions are
     * fullfilled,  Is called either from checkFrame() or from checkNodes
     * 
     * @return true if all PACs are valid, else false.
     */
    public boolean checkPartPACs() {
        GraphPartFrame frame = createInnerGraphFrame();
        boolean        valid = true;
        Iterator       it    = PACs.iterator();

        while (it.hasNext() && valid) {
            ApplicationCondition ac = (ApplicationCondition) it.next();
            valid &= ac.testPartAC(frame);
        }

        return valid;
    } //checkPartPACs


    /**
     * If there exists a frame that fit to the actual interpretation of the
     * linked nodes of the part i.e. have the same interpretaiton, this
     * function returns true, else false. The frame may be one of foundMatches
     * or partlyFoundMatches.
     * 
     * @param part  The other part connected to this one.
     * 
     * @return There is at least one frame fitting to part.
     */
    public boolean existFittingFrames(StructuralConstraintGraphPart part) {
        Set linkedNodes = part.getInnerGraphLinkNodes(this);

        return foundMatches.existFittingParts(linkedNodes) || 
               partlyFoundMatches.existFittingParts(linkedNodes);
    }


    /**
     * Is there at least one frame for the part?
     * 
     * @return true if there is a frame stored for this part, else false.
     */
    public boolean existsFrame() {
        return (foundMatches.size() > 0) || checkingNewFrame;
    }


    /**
     * Tests whether the MainPart can be embedded at the node seed into the
     * problem graph. Called when the whole graph should be build Finds only
     * one isomorphism, starting from seed. Seed needs an interpretation
     * before calling this function If none found the root seed interpretation
     * is reset. If the part can not be found and no intersection with other
     * parts occured the frame is removed from the list foundMatches.
     * 
     * @return true if one Embedding was found, else false.
     */
    public boolean findOneEmbedding() {
        boolean result = super.findOneEmbedding();

        if (! result) { // the part could not be found, maybe the used frame is outdated?

            if (! intersectionWithOtherGraphPartOcurred()) {
                // if an intersection occured nothing can be said about the validity of the frame.
                foundMatches.remove(usedGraphPartFrame); // else the frame is not valid no more, remove it.
                partlyFoundMatches.remove(usedGraphPartFrame);
            }
        }

        return result;
    }


    /**
     * When a new match for the MainPart is found during this function checks
     * whether the mainParts ApplicationConditions are fullfilled, and
     * initiates the search for whole graphs with this mainPart.
     * 
     * @return true if the ACs were valid for this frame.
     */
    public boolean handleMatch() {
        dl.print(3, "\n mainPart.handleMatch() ");

        boolean found = checkPartPACs();

        if (found) {
            dl.print(3, "\n PACs found");
            checkingNewFrame = true;
            ((SCTDockingGraph) mySCGraph).findAllMatchesWithPart(this);
            checkingNewFrame = false;
        } else {
            dl.print(3, "\n PACs not found");
        }

        return found;
    }


    /**
     * Is the given interpretation allready used in this graph? The mapping
     * from structural constraint graph g to problem graph p must be
     * injective, i.e. no two different nodes from g may have the same
     * interpretation in p. If the interpretation is used in an other Part the
     * function tries to find other interpretations for the other parts in
     * order to free this interpretation.
     * 
     * @param interpretation The interpretation.
     * @param node The node that should get the interpretation.
     * 
     * @return true if it contains the interpretation.
     */
    public boolean tryInterpretation(Node interpretation, SCNode node) {
        SCNode nodeWithInterpretation = (SCNode) interpretedNodes.get(
                                                interpretation);

        if ((nodeWithInterpretation != null) && 
                (nodeWithInterpretation != node)) {
            return true; // the node is used allready in same part.
        } else if (nodeWithInterpretation == node) {
            return false;
        }

        boolean result = allreadyCheckedNewNodes.contains(interpretation);


        //check whether it it used by other parts 
        result |= mySCGraph.tryInterpretation(interpretation, node);

        if (result == true) {
            intersectionWithOtherGraphPart = true;
        }

        return result;
    }


    /**
     * Creates all the missing ConnectingSCTPACParts that were in need of this
     * embedding of this mainPart.
     */
    void createConnectingPACParts() {
        Iterator it = connectingPACParts.iterator();

        while (it.hasNext()) {
            TypeDockingConnectingPACPart part = (TypeDockingConnectingPACPart) it.next();

            if (part.hasInterpretations()) {
                throw new Error("part has allready interpretations");
            }

            part.findNewMatchesUsingMainPart();
        }
    }


    /**
     * Tries to find a new MainPart with the new nodes
     * 
     * @param newNodes  Nodes that were recently created.
     */
    private void checkNewNodes(Set newNodes) {
        searchedNodes.sortNodesByFrequency(newNodes);

        Iterator it = newNodes.iterator();

        while (it.hasNext()) { // check all newNodes, whether one of them leads to a new match

            Node     newNode         = (Node) it.next();
            Set      s               = searchedNodes.getNodes(newNode);
            Iterator searchedNodesIt = s.iterator();

            while (searchedNodesIt.hasNext()) {
                SCNode n = (SCNode) searchedNodesIt.next();

                if (n.setInterpretation(newNode)) {
                    if (findOneEmbeddingWithoutCheckedNodes(n)) {
                        GraphPartFrame frame = createInnerGraphFrame();
                        partlyFoundMatches.add(frame);
                        printNodes();
                        createConnectingPACParts(); //try to create all ConnectingPACParts that are connected to this part

                        if (handleMatch()) {
                            foundMatches.add(frame);
                            dl.print(3, "\n MP added foundMatch");
                            partlyFoundMatches.remove(frame);
                        } else {
                            dl.print(3, "\n MP added partlyFoundMatch");
                        }

                        while (findNextEmbeddingWithoutCheckedNodes()) {
                            frame = createInnerGraphFrame();
                            partlyFoundMatches.add(frame);
                            printNodes();
                            createConnectingPACParts();

                            if (handleMatch()) {
                                foundMatches.add(frame);
                                partlyFoundMatches.remove(frame);
                            }
                        }
                    } // if findOneEmbedding
                } //if setInterpretation
            } // while missingNodesIt

            allreadyCheckedNewNodes.add(newNode);
        } // while it

        allreadyCheckedNewNodes.clear();
    }


    /**
     * Calls ac.findNextPartEmbedding(n) but sets the variable
     * searchWithoutChecked = true so that the search doesnt find matches with
     * allready checked nodes.
     * 
     * @return true if an embedding was found, else false.
     */
    private boolean findNextEmbeddingWithoutCheckedNodes() {
        boolean result;
        searchWithoutChecked = true; // switch on that the new nodes allready checked can not be part of the interpretations found now.
        result               = findNextEmbedding();
        searchWithoutChecked = false;

        return result;
    }


    /**
     * Calls findOneEmbedding(n) but sets the variable searchWithoutChecked =
     * true so that the search doesnt find matches with allready checked
     * nodes.
     * 
     * @param n  The node that should be embedded.
     * 
     * @return true if an embedding was found, else false.
     */
    private boolean findOneEmbeddingWithoutCheckedNodes(SCNode n) {
        boolean result;
        searchWithoutChecked = true; // switch on that the new nodes allready checked can not be part of the interpretations found now.
        result               = findOneEmbedding(n);
        searchWithoutChecked = false;

        return result;
    }
}