/*
 * SCIDockingGraph.java
 *
 * Created on 25. Mai 2002, 14:21
 */
package com.ai_center.dragonbreath.structuralConstraintHandling.graph.dockingGraph;

import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.StructuralConstraintInstance;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.applicationCondition.ApplicationConditionPart;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.dockingGraph.applicationCondition.instance.InstanceDockingNAC;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.dockingGraph.applicationCondition.instance.InstanceDockingPAC;
import com.ai_center.dragonbreath.utilities.SetHashtable;

import java.math.BigInteger;

import java.util.*;


/**
 * The docking graph used in the StructuralConstraintInstance, Checks its
 * consistency.
 * 
 * @version $Revision: 1.5 $
 * @author tim baermann
 */
public class SCIDockingGraph extends DockingGraph {
    // Instance variables *****************************************************

    /** 
     * The hashCode of this DockingGraphInstance 
     */
    BigInteger hashCode;

    /** 
     * Frames that could reconstruct the TypeDockingGraph. 
     */
    PartFrameContainer typeFrames;

    /**
     * The DockingGraph of the type of the StructuralConstraintInstance this
     * docking graph belongs to.
     */
    SCTDockingGraph typeDockingGraph;

    /** 
     * The StructuralConstraintInstance this docking graph belongs to. 
     */
    StructuralConstraintInstance myStructuralConstraintInstance;

    /**
     * Is the docking graph valid? I.e. all PACs and main parts exist and no
     * NAC exists.  Is set by checkChangedNodes and retrieved by valid();
     */
    boolean valid = true;

    // Constructors ***********************************************************

    /**
     * Creates a new instance of SCIDockingGraph.
     * 
     * @param typeDocking The docking graph of the type of the constraint.
     */
    public SCIDockingGraph(SCTDockingGraph typeDocking) {
        typeDockingGraph = typeDocking;
    }

    // Methods ****************************************************************

    /**
     * Sets the HashCode of the instance, as it is computed allready by the
     * typeDockingGraph it only has to be transferred.
     * 
     * @param hash A unique number (Same that would be computed by
     *        getHashCode() )
     */
    public void setHashCode(BigInteger hash) {
        hashCode = hash;
    }


    /**
     * Returns a Hashcode of the graph with its actual mainParts
     * interpretations. (Therefore the graph needs a valid interpretation
     * first!)
     * 
     * @return a unique number.
     */
    public BigInteger getHashCode() {
        return hashCode;
    }


    /**
     * Sets the instance the DockingGraph belongs to
     * 
     * @param constraint The Constraint instance of this DockingGraphInstance.
     */
    public void setInstance(StructuralConstraintInstance constraint) {
        myStructuralConstraintInstance = constraint;
    }


    /**
     * Sets the Frames of the SCTDockingGraph that reconstruct the type docking
     * Graph with the same match this instance represents. Is called in
     * SCTDockingGraph.createInstance().
     * 
     * @param tfc A Container with the said frames.
     */
    public void setTypeFrames(PartFrameContainer tfc) {
        typeFrames = tfc;
    }


    /**
     * Returns the typeFrames.
     * 
     * @return typeFrames.
     */
    public PartFrameContainer getTypeFrames() {
        return typeFrames;
    }


    /**
     * Checks nodes that were newly changed for creation of NACs or destruction
     * of MainParts or PACs. If NAC was found the graph is registered at the
     * SCTDockingGraph as partly found. If PAC is destroyed the function tries
     * to find a substitute, and if none found this is also registered at the
     * SCTDockingGraph. If a main part is destroyed the part is removed from
     * the list of the parts in the SCTDockingGraph. In all cases the valid()
     * function returns false at the next call.
     * 
     * @param newNodes  Nodes that were recently created.
     * @param expandedNodes  Expanded Nodes.
     * @param reducedNodes  The reduced Nodes.
     * @param deletedNodes  The nodes deleted.
     */
    public void checkChangedNodes(Set newNodes, Set deletedNodes, 
                                  Set expandedNodes, Set reducedNodes) {
        this.valid = true;

        if ((deletedNodes.size() > 0) || (reducedNodes.size() > 0)) {
            if (! (checkMainParts(deletedNodes, reducedNodes) && checkPACs(deletedNodes, 
                                                                           reducedNodes))) {
                setInvalid();
            }
        }

        if (valid && ((newNodes.size() > 0) || (expandedNodes.size() > 0))) {
            if (! checkNACs(newNodes, expandedNodes)) {
                setInvalid();
            }
        }
        
    } //checkChangedNodes()


    /**
     * Adds an ApplicationConditionPart to the graph.
     * 
     * @param part  The new ApplicationCondition.
     */
    protected void addACPart(ApplicationConditionPart part) {
    }


    /**
     * Is the docking graph valid? I.e. all PACs and main parts exist and no
     * NAC exists.  called after checkChangedNodes() to retrieve the result.
     * 
     * @return this.valid
     */
    public boolean isValid() {
        return this.valid;
    } //valid()

    
    
    /**
     * Marks the DockingGraph as invalid.
     */
    public void setInvalid() {
        typeDockingGraph.removeGraphInstanceHash(this);
        valid = false;
    }
    
    
    
    /**
     * Checks whether a new NAC was created via the given nodes.
     * 
     * @param newNodes  Nodes that were recently created.
     * @param expandedNodes  Expanded Nodes.
     * 
     * @return Returns whether the was created.
     */
    boolean checkNACs(Set newNodes, Set expandedNodes) {
        if ((newNodes.size() > 0) || (expandedNodes.size() > 0)) {
            Iterator nacIt = this.NACs.iterator();

            while (nacIt.hasNext()) {
                InstanceDockingNAC nac = (InstanceDockingNAC) nacIt.next();

                if (! nac.testAC()) {
                    return false;
                }
            } // while nacIt
        }

        return true;
    } // checkNACs()


    /**
     * Checks whether the main parts still exist.
     * 
     * @param reducedNodes  The reduced Nodes.
     * @param deletedNodes  The nodes deleted.
     * 
     * @return true if a mainPart was destroyed, else false.
     */
    private boolean checkMainParts(Set deletedNodes, Set reducedNodes) {
        Iterator mainIt = mainParts.iterator();
        boolean  valid = true;

        while (mainIt.hasNext() && valid) {
            StructuralConstraintGraphPart mp = (StructuralConstraintGraphPart) mainIt.next();
            valid = mainPartCheckDeletedNodes(deletedNodes, mp) && 
                    mainPartCheckReducedNodes(reducedNodes, mp);
        } // while mainIt

        return valid;
    } //checkMainParts()


    /**
     * Checks whether the PACs parts still exist. If not tries to find a
     * substitute
     * 
     * @param deletedNodes  The nodes deleted.
     * @param reducedNodes  The reduced Nodes.
     * 
     * @return true if a PAC was destroyed, and no substitute found, else false.
     */
    private boolean checkPACs(Set deletedNodes, Set reducedNodes) {
        Iterator pacIt = this.PACs.iterator();

        while (pacIt.hasNext()) {
            InstanceDockingPAC pac = (InstanceDockingPAC) pacIt.next();

            if (! pac.checkAC(deletedNodes, reducedNodes)) {
                return false;
            }
        } // while pacIt

        return true;
    } //checkPACs()


    /**
     * Checks whether the mainPart contains one of the given deletedNodes.
     * 
     * @param deletedNodes  The nodes that were deleted.
     * @param mp  The main Part that should be tested.
     * 
     * @return Returns whether a mainpart contains one of the deleted nodes or
     *         not.
     */
    private boolean mainPartCheckDeletedNodes(final Set deletedNodes, 
                                              final StructuralConstraintGraphPart mp) {
        Iterator it = deletedNodes.iterator();

        while (it.hasNext()) {
            Node n = (Node) it.next();

            if (mp.partContainsInterpretation(n)) {
                return false;
            }
        } //while it

        return true;
    }


    /**
     * Checks whether the mainPart was destroyed due to one of the given
     * reducedNodes.
     * 
     * @param reducedNodes  The nodes that lost connections.
     * @param mp  The main Part that should be tested.
     * 
     * @return Whether one of the mainparts was destroyed.
     */
    private boolean mainPartCheckReducedNodes(final Set reducedNodes, 
                                              final StructuralConstraintGraphPart mp) {
        Iterator it = reducedNodes.iterator();

        while (it.hasNext()) {
            Node   n   = (Node) it.next();
            SCNode scn = mp.getNodeWithInterpretation(n);

            if (scn != null) {
                if (! scn.testInterpretation()) {
                    return false;
                } // if scn testInterpretation
            } // if scn != null
        } //while it

        return true;
    }
} //class
