/*
 * SCDockingGraph.java
 *
 * Created on 7. April 2002, 18:40
 */
package com.ai_center.dragonbreath.structuralConstraintHandling.graph.dockingGraph;

import com.ai_center.dragonbreath.*;
import com.ai_center.dragonbreath.constraintgraph.Node;
import com.ai_center.dragonbreath.structuralConstraintHandling.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.SCGraphFactory;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.applicationCondition.ApplicationCondition;
import com.ai_center.dragonbreath.utilities.DebugLogger;
import com.ai_center.dragonbreath.utilities.Pair;
import com.ai_center.dragonbreath.utilities.SetHashtable;

import java.math.BigInteger;

import java.util.*;


/**
 * The docking graph used in the StructuralConstraintType, Finds new matches,
 * creates copies of these matches and  gives them to the Constraint type.
 * 
 * @version $Revision: 1.8 $
 * @author tim baermann
 */
public class SCTDockingGraph extends DockingGraph {
     
    // Instance variables *****************************************************

    /** 
     * Output for debugging purpose 
     */
    DebugLogger dl = new DebugLogger(true);

    /**
     * During the search for the whole docking graph all the
     * StructuralConstraintGraphParts added are stored in the stack, in order
     * to create via backtracking all solutions.
     */
    GraphPartStack graphSearchStack = new GraphPartStack();

    /**
     * For every found graph instance a hashcode is stored in this container.
     * So if a new instance is found a search in it can indicate whether the
     * same interpretation was allready found.
     */
    HashSet graphInstancesHashes = new HashSet();

    /**
     * MainParts (a frame with an embedding) that have improved due to newly
     * found ApplicationConditionParts. They are added by the parts via
     * registerImprovedMainParts() and rechecked at the end of
     * checkChangedNodes() via checkImprovedMainParts():
     */
    PartFrameContainer improvedMainPartMatches = new PartFrameContainer();

    /** 
     * To order StructuralConstraintGraphParts by their conectivity. 
     */
    SCGPComparator partComparator = new SCGPComparator(true);

    /** 
     * The factory that produced this graph. 
     */
    SCGraphFactory myFactory;

    /** 
     * The Type of the StructuralConstraint this DockingGraph belongs to. 
     */
    StructuralConstraintType mySCType;

    /** 
     * All the SCIDockingGraph instances of this docking graph 
     */
    Vector myInstances = new Vector();

    /**
     * If checkReducedNodes(), checkNewNodes() etc. finds new graph instances
     * they are stored in this Container, and can be retrieved by
     * getNewGraphInstances().
     */
    Vector newGraphInstances = new Vector();

    // Constructors ***********************************************************

    /**
     * Creates a new instance of SCDockingGraph
     * 
     * @param graphFactory  The factory that produced this graph.
     */
    public SCTDockingGraph(SCGraphFactory graphFactory) {
        myFactory = graphFactory;

        //dl.switchOn(2);// uncomment for debug output
        //dl.switchOn(1);// uncomment for debug output
    }

    // Methods ****************************************************************

    /**
     * This function should be called after checkChangedNodes() to retrieve all
     * the graphInstances created during the calls. The call resets the list.
     * 
     * @return All the new Instances.
     */
    public Collection getNewGraphInstances() {
        Vector result = newGraphInstances;
        newGraphInstances = new Vector();

        return result;
    }


    /**
     * Register the StructuralConstraintManagement;
     * 
     * @param scm  The new StructuralConstraintManagement.
     * @param type The StructuralConstraintType of this graph.
     */
    public void setSCManagementAndSCType(StructuralConstraintManagement scm, 
                                         StructuralConstraintType type) {
        mySCM    = scm;
        mySCType = type;
    }


    /**
     * Checks nodes that were newly created for creation or destruction of
     * StructuralConstraintInstances.
     * 
     * @param newNodes  Nodes that were recently created.
     * @param deletedNodes   The nodes deleted.
     * @param expandedNodes  Expanded Nodes
     * @param reducedNodes   The reduced Nodes.
     */
    public void checkChangedNodes(Set newNodes, Set deletedNodes, 
                                  Set expandedNodes, Set reducedNodes) {
        findNACs(newNodes, deletedNodes, expandedNodes, reducedNodes); 
        // has to be done before mainParts are searched! because if a new match of the mainparts 
        // is found the not connected nacs have to be allready known, or else the match is created.
        findNewMainParts(newNodes, deletedNodes, expandedNodes, reducedNodes); 
        // has to be done before pacs are searched, else connecting pacs are not found, as the mainparts are still missing!
        findNewPACs(newNodes, deletedNodes, expandedNodes, reducedNodes);

        checkImprovedMainPartMatches();

        // FIXME: check from time to time the partlyFoundGraphParts to get rid of zombies.
    } //public void checkChangedNodes() 


    
    /**
     * Resets the docking graph to its intitial state.
     */
    public void clear() {
        myInstances.clear();
        newGraphInstances.clear();
        resetAllACs();
    }


    /**
     * Returns a PartFrameContainer with all the frames of all MainParts.
     * 
     * @return A Container with Frames.
     */
    public PartFrameContainer createInterGraphFrames() {
        PartFrameContainer result = new PartFrameContainer();
        Iterator           it = mainParts.iterator();

        while (it.hasNext()) {
            StructuralConstraintGraphPart mp = (StructuralConstraintGraphPart) it.next();
            result.add(mp.createInterGraphFrame());
        }

        return result;
    }


    /**
     * Given a interpreted part, tries to find all the graphmatches that
     * contain the given part. Calls findMatchesWithPartFrame().
     * 
     * @param part  An interpreted StructuralConstraintGraphPart.
     */
    public final void findAllMatchesWithPart(StructuralConstraintGraphPart part) {
        GraphPartFrame frame = part.getInnerGraphFrame();
        findAllMatchesWithPartFrame(frame);
    }


    /**
     * Tries to find the the docking graph by combining the allready found
     * parts, using the given partFrame, and the current interpretation of its
     * GraphPart in every match. Stores the fond matches in newGraphInstances
     * 
     * @param partFrame The frame that should be part of every found match.
     */
    public final void findAllMatchesWithPartFrame(GraphPartFrame partFrame) {
        dl.print(1, "\n findAllMatchesWithPartFrame");

        //is at least one exemplar of every part found?
        if (allRelevantPartsFound()) {
            if (mainParts.size() == 1) {
                if (checkAllACs()) {
                    createInstanceGraph();
                    resetAllACs();
                }
            } else {
                boolean result;
                Vector  initialQueue                = new Vector();
                Stack   initialInterpretationsStack = new Stack();
                initialInterpretationsStack.push(partFrame);
                graphSearchStack.clear();
                graphSearchStack.push(initialQueue, initialInterpretationsStack);

                while (! graphSearchStack.empty()) {
                    Pair actualQueuePair = (Pair) graphSearchStack.pop();

                    if (actualQueuePair.first == null) { //first is allways the searchqueue at the point in branch and bound search 
                        ((GraphPartFrame) actualQueuePair.second).resetInterpretations();
                    } else { // try a different branch

                        Vector         actualQueue = new Vector();
                        Collection     oldQueue = (Collection) actualQueuePair.first; // the queue when the part was added
                        Stack          interpretationsStack = 
                                (Stack) actualQueuePair.second;
                        GraphPartFrame newInterpretations = 
                                (GraphPartFrame) interpretationsStack.pop();

                        if (! interpretationsStack.empty()) {
                            graphSearchStack.push(oldQueue, 
                                                  interpretationsStack); // there are still other possible interpretations, so add the part again.
                        }

                        actualQueue.add(newInterpretations.myGraphPart); // add the actual part as the first in the queue
                        actualQueue.addAll(oldQueue); //reconstruct the rest of the queue
                        
                        if (! newInterpretations.setInterpretations()) { // take the next branch
                            throw new Error(
                                    "Could not set frame interpretations.");
                        }

                        result = tryNextBranch(actualQueue, newInterpretations);

                        if (result) { // a valid interpretation was found
                            dl.print(1, "\n checkAllGraphParts \n");
                            checkAllGraphParts(0, partFrame.myGraphPart);
                        }
                    } // else try a different branche
                } //while ! graphSearchStack.empty()
            } //if graph more complicated then only one part
        } //if one occurence of every part found
        else {
            dl.print(2, "\n not all parts found");
        }
    } //findMatchesWithPartFrame()


    /**
     * Returns a Hashcode of the graph with its actual mainParts
     * interpretations. (Therefore the graph needs a valid interpretation
     * first!)
     * 
     * @return A unique number for the intepretation.
     */
    public BigInteger getHashCode() {
        String   stringHash = new String("0");
        Iterator it = mainParts.iterator();

        while (it.hasNext()) {
            StructuralConstraintGraphPart part = (StructuralConstraintGraphPart) it.next();
            stringHash = stringHash.concat(part.getHashCodeString());
        } //while 

        return new BigInteger(stringHash);
    }


    /**
     * Does the dockingGraph have any instances?
     * 
     * @return True if there are instances.
     */
    public boolean hasInstances() {
        return graphInstancesHashes.size() > 0;
    }


    /**
     * Whenever a ApplicationConditionPart is found and there are mainParts
     * that lacked that part all these MainParts are added to
     * improvedMainParts. At the end of each checkChangedNodes cycle the
     * improved mainParts are checked again, to find new matches with them.
     * 
     * @param newImprovedMainPartMatches The matches that were improved.
     */
    public void registerImprovedMainPartMatches(Collection newImprovedMainPartMatches) {
        improvedMainPartMatches.addAll(newImprovedMainPartMatches);
    }


    /**
     * When the queue is extended with extendQueueWithSuccessors() this
     * function checks whether the intended extension was not allready done
     * before.
     * 
     * @param part The part to be checked.
     * 
     * @return Was part checked before?
     */
    boolean allreadyCheckedPart(StructuralConstraintGraphPart part) {
        return graphSearchStack.containsPart(part);
    }


    /**
     * After all the mainParts have their interpretation assigned this function
     * checks whether all ApplicationConditions are fullfilled.
     * 
     * @return Are ACs valid?
     */
    boolean checkAllACs() {
        boolean  valid = true;
        Iterator it = PACs.iterator();

        while (it.hasNext() && valid) {
            ApplicationCondition pac = (ApplicationCondition) it.next();
            valid = pac.testAC();
        }

        it = NACs.iterator();

        while (it.hasNext() && valid) {
            ApplicationCondition nac = (ApplicationCondition) it.next();
            valid = nac.testAC();
        }

        if (! valid) {
            resetAllACs();  
            dl.print(2, "\n not all ACs valid");
        }

        return valid;
    }


    /**
     * After assigning interpretations to all linked nodes of all mainParts and
     * PACs this function tests if all the parts are still valid and all their
     * NACs still exist.
     * 
     * @param partNr  The number of the part currently tested.
     * @param foundPart  The part that was found last, and does not have to be
     *        found again. (if last found part was an application condition it
     *        will be tested again.)
     */
    final void checkAllGraphParts(int partNr, 
                                  StructuralConstraintGraphPart foundPart) {
        StructuralConstraintGraphPart nextPart;

        if (partNr == mainParts.size()) { // all parts assigned?

            //check the application conditions
            printAllMainParts();

            if (checkAllACs()) {
                createInstanceGraph();


                // clean up again
                resetAllACs();
            }
        } else {
            nextPart = (StructuralConstraintGraphPart) mainParts.elementAt(
                               partNr);

            if (nextPart == foundPart) { //the last found part is to be tested again
                checkAllGraphParts(partNr + 1, foundPart); // then try the next part until all checked
            } else if (nextPart.findOneEmbedding()) { // can the rest of the nodes be assigned?
                checkAllGraphParts(partNr + 1, foundPart); // then try the next part until all checked

                while (nextPart.findNextEmbedding()) {
                    checkAllGraphParts(partNr + 1, foundPart); // then try the next part until all checked
                } //while
            }
        } // else not all parts found
    } // checkAllGraphParts


    /**
     * Resets the interpretations of all PAC parts.
     */
    final void resetAllACs() {
        Iterator it = PACs.iterator();

        while (it.hasNext()) {
            ApplicationCondition pac = (ApplicationCondition) it.next();
            pac.resetInterpretationsExtern();
            pac.resetInterpretations();
        }

        it = NACs.iterator();

        while (it.hasNext()) {
            ApplicationCondition nac = (ApplicationCondition) it.next();
            nac.resetInterpretationsExtern();
            nac.resetInterpretations();
        }
    }


    /**
     * Resets the Interpretations of all mainParts.
     */
    final void resetAllMainParts() {
        Iterator it = mainParts.iterator();

        while (it.hasNext()) {
            StructuralConstraintGraphPart p = (StructuralConstraintGraphPart) it.next();
            p.resetInterpretationsExtern();
            p.resetInterpretations();
        }
    }


    /**
     * Travels the graph with "branch and bound" and assigns interpretations
     * for all graph parts. If for the actual part there are multiple ways of
     * interpreting them the other interpretations are pushed on the
     * graphSearchStack, and explored later in findMatchesWithPartFrame via
     * backtracking.
     * 
     * @param queue the queue at the start of the branch and bound.
     * @param firstInterpretation The interpretation of the first part, if the
     *        queue is empty.
     * 
     * @return Interpretation for all parts found.
     */
    final boolean tryNextBranch(Vector queue, 
                                GraphPartFrame firstInterpretation) {
        // to iterate over the main parts of the graph, and add them one by one to the 
        // queue, if they are not added automatically by the algorithm because they are not connected to the rest.
        Iterator                mainIt              = mainParts.iterator();
        boolean                 result              = true;
        boolean                 interpretationFound = true;
        boolean                 firstPart           = true; // when called the first part has allready an interpretation assigned by findMatchesWithPartFrame(), so tread it specially
        ConnectingPartInterface actualPart;

        // if for any of the parts in the queue no interpretation is found the search cant find a valid embedding any more. backtracking has to occure.
        while ((queue.size() > 0) && interpretationFound) {
            // actualPart = queue.head()
            actualPart = (ConnectingPartInterface) queue.firstElement();
            queue.remove(0);

            // if the interpretation was not allready set in calling method findMatchesWithPartFrame()
            if (! firstPart) {
                //check whether there are valid interpretations for the part
                // if there are more then one interpretations: save the rest and the actual queue in graphSearchStack
                // get all the possible interpretations for this part.
                Stack allPartInterpretations = actualPart.getFittingFrames();

                if (allPartInterpretations.empty()) { // no interpretation fitted! 

                    return false; //dont search on
                }

                // now use the first of the found interpretations
                GraphPartFrame actualInterpretation = (GraphPartFrame) allPartInterpretations.pop();
                boolean        interpretationGood = actualInterpretation.setInterpretations();

                while (! interpretationGood && 
                       ! allPartInterpretations.empty()) {
                    actualInterpretation = (GraphPartFrame) allPartInterpretations.pop();
                    interpretationGood   = actualInterpretation.setInterpretations();
                }

                if (! interpretationGood) {
                    return false; // there was no good interpretation for the part.
                }

                if (! allPartInterpretations.empty()) { //there are more then the one used possibilities

                    Vector queueCopy = new Vector(queue);
                    graphSearchStack.push(queueCopy, allPartInterpretations);
                } else { //only one interpretation possible
                    graphSearchStack.push(actualInterpretation); //store the only valid interpretation to reset it later..
                }
            } else { //if ! firstPart => else the part already has an interpretation.
                graphSearchStack.push(firstInterpretation); //store the only valid interpretation to reset it later..
                firstPart = false;
            }


            // extend the queue with all the successors of the actual part.
            extendQueueWithSuccessors(queue, actualPart);


            //add further not connected mainParts to the queue, in order to search them.
            addMissingMainPart(queue, mainIt);

            // sort the queue, so that parts that are less probable to be found are tested first.
            if (StructuralConstraintManagement.graphSearchSortQueueOn) {
                Collections.sort(queue, partComparator);
            }
        } //while

        return result;
    } //tryNextBranch()


    /**
     * Tests whether for all mainParts and for all connectingPACs exists at
     * least one embedding in the graph, so that a whole match can be found.
     * 
     * @return A match is possible.
     */
    private final boolean allRelevantPartsFound() {
        boolean  result = true;
        Iterator it = mainParts.iterator();

        while (it.hasNext() && result) {
            ConnectingPartInterface mp = (ConnectingPartInterface) it.next();
            result = mp.existsFrame();
        }

        // here could be tested too if all the connectingPACParts exist,
        // but that would be very costly, so dont do it.
        return result;
    }


    /**
     * Extends the queue with the Successors of the given part. Called in
     * tryNextBranch()
     * 
     * @param queue  The queue where the StructuralConstraintGraphParts
     *        connected to actualPart are added.
     * @param actualPart  The part with the successors.
     */
    private final void extendQueueWithSuccessors(final Vector queue, 
                                                 final ConnectingPartInterface actualPart) {
        Iterator it = actualPart.getInnerGraphLinkParts().iterator();

        while (it.hasNext()) {
            StructuralConstraintGraphPart nextPart = (StructuralConstraintGraphPart) it.next();

            if (! allreadyCheckedPart(nextPart) && 
                    nextPart.isConnectingPart()) {
                queue.add(nextPart);
            }
        } //while 
    } // extendQueueWithSuccessors()


    /**
     * Searches for new embeddings of NACs.
     * 
     * @param newNodes  Nodes that were recently created.
     * @param expandedNodes  Expanded Nodes.
     * @param reducedNodes  The reduced Nodes.
     * @param deletedNodes  The nodes deleted.
     */
    private final void findNACs(Set newNodes, Set deletedNodes, 
                                Set expandedNodes, Set reducedNodes) {
        Iterator it = NACs.iterator();

        while (it.hasNext()) {
            ApplicationCondition nac = (ApplicationCondition) it.next();
            nac.checkChangedNodes(newNodes, deletedNodes, expandedNodes, 
                                  reducedNodes);
        }
    }


    /**
     * Finds all the main parts that contain the given nodes.
     * 
     * @param newNodes  Nodes that were recently created.
     * @param expandedNodes  Expanded Nodes.
     * @param reducedNodes  The reduced Nodes.
     * @param deletedNodes  The nodes deleted.
     */
    private final void findNewMainParts(Set newNodes, Set deletedNodes, 
                                        Set expandedNodes, Set reducedNodes) {
        Iterator it = mainParts.iterator();

        while (it.hasNext()) {
            NodeChangedInterface mp = (NodeChangedInterface) it.next();
            dl.print(2, "\n Searching for MainPart");
            mp.checkChangedNodes(newNodes, deletedNodes, expandedNodes, 
                                 reducedNodes);
        }
    } //void findNewMainParts(  )


    /**
     * Searches for new embeddings of PACs.
     * 
     * @param newNodes  Nodes that were recently created.
     * @param expandedNodes  Expanded Nodes.
     * @param reducedNodes  The reduced Nodes.
     * @param deletedNodes  The nodes deleted.
     */
    private final void findNewPACs(Set newNodes, Set deletedNodes, 
                                   Set expandedNodes, Set reducedNodes) {
        Iterator it = PACs.iterator();

        while (it.hasNext()) {
            ApplicationCondition pac = (ApplicationCondition) it.next();
            dl.print(2, "\n Searching for PAC ");
            pac.checkChangedNodes(newNodes, deletedNodes, expandedNodes, 
                                  reducedNodes);
        }
    }


    /**
     * If the queue is empty it is filled with the next not interpreted
     * mainPart delivered by mainIt. called in tryNextBranch()
     * 
     * @param queue  The queue that is expanded.
     * @param mainIt  An iterator over all the mainParts of the dockingGraph.
     */
    private void addMissingMainPart(final Vector queue, Iterator mainIt) {
        if (queue.isEmpty()) {
            boolean nextMainAdded = false;

            while (mainIt.hasNext() && ! nextMainAdded) {
                StructuralConstraintGraphPart nextMain = 
                        (StructuralConstraintGraphPart) mainIt.next();

                if (! allreadyCheckedPart(nextMain)) {
                    queue.add(nextMain);
                    nextMainAdded = true;
                }
            }
        } //if queue empty
    }


    /**
     * This function tests for all improvedMainPartMatches that were added  via
     * registerImprovedMainParts() if a whole match can be found with it.
     */
    private void checkImprovedMainPartMatches() {
        dl.print(2, "\n\n oooooooooocheckImprovedMainPartoooooooooo");
        dl.print(2, 
                 "\nimprovedMainPartMatches.size = " + 
                 improvedMainPartMatches.size());

        Iterator it = improvedMainPartMatches.iterator();

        while (it.hasNext()) {
            GraphPartFrame    frame = (GraphPartFrame) it.next();
            MainPartInterface mp = (MainPartInterface) frame.myGraphPart;
            dl.print(2, "\n checkFrame");
            mp.checkFrame(frame);
        }

        improvedMainPartMatches.clear();
    }


    /**
     * Given an interpretation for the docking graph this function creates an
     * Instance of this SCTDockingGraph  and adds it to myInstances and
     * newGraphInstances. (But only if the same interpretation doesnt allready
     * exist as an instance.)
     */
    private final void createInstanceGraph() {
        BigInteger newHash = getHashCode();

        if (graphInstancesHashes.add(newHash) == true) { //the newHash could be added
            dl.print(2, "\n instance created ++++++++++++++++++++\n");

            
            SCIDockingGraph newDocking = myFactory.getInstanceDockingGraph(this);
            newDocking.setHashCode(newHash);


            //copy the interpretations from this to newDocking.
            newDocking.setTypeFrames(this.createInterGraphFrames());

            newDocking.copyInterpretations(this);

            myInstances.add(newDocking);
            newGraphInstances.add(newDocking);
        } //if graphInstancesHashes.add()
        else {
            dl.print(2, "\n double instance =========== \n");
        }
    } // createInstanceGraph()


    /**
     * Prints the interpretations of all mainParts.
     */
    private void printAllMainParts() {
        Iterator it = mainParts.iterator();

        while (it.hasNext()) {
            DockingMainPart mp = (DockingMainPart) it.next();
            mp.printNodes();
        }
    }


    /**
     * Removes the instance hashcode from the list of hashes.
     * 
     * @param instanceGraph  The instance to be removed.
     */
    public void removeGraphInstanceHash(SCIDockingGraph instanceGraph) {
        graphInstancesHashes.remove(instanceGraph.getHashCode());
        dl.print(2, "\n instance deleted --------------\n");
    }
}