/*
 * ConnectingSCTPACPart.java
 *
 * Created on 23. Juni 2002, 23:48
 */
package com.ai_center.dragonbreath.structuralConstraintHandling.graph.dockingGraph.applicationCondition.type;

import com.ai_center.dragonbreath.structuralConstraintHandling.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.applicationCondition.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.dockingGraph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.movableSubgraph.*;

import java.util.*;


/**
 * A part of a PAC of a Docking Graph Type that is connected to more then one
 * MainPart.
 * 
 * @version $Revision: 1.3 $
 * @author tim baermann
 */
public class TypeDockingConnectingPACPart extends MovableSubgraphPart
    implements TypeDockingACPart, ConnectingPartInterface {
    // Instance variables *****************************************************

    /** 
     * All the embeddings found for this part. 
     */
    protected PartFrameContainer foundMatches = new PartFrameContainer();

    /**
     * All the mainParts that were not created due to this
     * ApplicationConditionPart.
     */
    protected PartFrameContainer mainPartsMissingThisACPart = 
            new PartFrameContainer();

    // Constructors ***********************************************************

    /**
     * Creates a new instance of ConnectingSCTPACPart
     * 
     * @param ac The PAC this part belongs to.
     */
    public TypeDockingConnectingPACPart(ApplicationCondition ac) {
        super(ac);

        //dl.switchOn(6);// uncomment for debug output
    }

    // Methods ****************************************************************

    /**
     * Returns the ApplicationCondition this part belongs to.
     * 
     * @return The mySubgraph member.
     */
    public ApplicationCondition getApplicationCondition() {
        return (ApplicationCondition) mySubgraph;
    }


    /**
     * Returns all the frames of the part that fit to the actual interpretation
     * of the linked nodes of this part i.e. have the same interpretaiton.
     * 
     * @return All the found frames.
     */
    public Stack getFittingFrames() {
        Stack    result             = new Stack();
        Set      allInterpretations = foundMatches.getAll();
        Iterator interpretIt        = allInterpretations.iterator();

        while (interpretIt.hasNext()) {
            GraphPartFrame newInterpretations = (GraphPartFrame) interpretIt.next();

            if (newInterpretations.checkInterpretations()) {
                result.push(newInterpretations);
            }
        } //while

        return result;
    }


    /**
     * Registeres a part and a node to be linked by an other part. Called by
     * SCNode.setLink()
     * 
     * @param part  The part that is linked.
     * @param node  A node from this part, that is either linked by a node from
     *        part, or links to a node from part.
     */
    public void addInnerGraphLink(StructuralConstraintGraphPart part, 
                                  SCNode node) {
        super.addInnerGraphLink(part, node);
        ((DockingMainPart) part).addConnectingPACPart(this);
    }


    /**
     * Produces a new node that can be connected via connectNodes to others in
     * this graph Also adds the node to the searched ones.
     * 
     * @param id the nodeID as defined in Registry
     * @param isLeafNode does the node have only one edge connected?
     * 
     * @return the newly created SCNode
     */
    public SCNode addNode(int id, boolean isLeafNode) {
        SCNode result = super.addNode(id, isLeafNode);

        if (! isLeafNode) {
            searchedNodes.addNode(result);
        }

        return result;
    }


    /**
     * Used to restore all the mainParts in mainPartsMissingThisACPart that
     * have now the missing part assigned. They are registered with the
     * DockingGraph, and there tested after all new nodes were tested.
     */
    public void checkWrongMainParts() {
        Collection fittingMainParts = mainPartsMissingThisACPart.extractFittingParts(
                                              this.getInnerGraphLinkNodes());

        if (mySCGraph instanceof SCTDockingGraph) {
            ((SCTDockingGraph) mySCGraph).registerImprovedMainPartMatches(
                    fittingMainParts);
        }
    } // checkWrongMainParts()


    /**
     * If there exists a frame that fit to the actual interpretation of the
     * linked nodes of part i.e. have the same interpretaiton, this function
     * returns true, else false.
     * 
     * @param part The checked Part.
     * 
     * @return There is at least one frame fitting to part.
     */
    public boolean existFittingFrames(StructuralConstraintGraphPart part) {
        Set connectedGraphPartsNodes = part.getInnerGraphLinkNodes(this);

        return foundMatches.existFittingParts(connectedGraphPartsNodes);
    }


    /**
     * Are there mainParts that fit this parts interpretation?
     * 
     * @return mainPartsFound
     */
    public boolean existFittingMainParts() {
        boolean  result = true;
        Iterator it = getInnerGraphLinkParts().iterator();

        while (it.hasNext() && result) {
            DockingMainPart mainPart = (DockingMainPart) it.next();
            result = mainPart.existFittingFrames(this);
        }

        return result;
    }


    /**
     * Is there at least one frame for the part?
     * 
     * @return true if a Frame was found, else false.
     */
    public boolean existsFrame() {
        return foundMatches.size() > 0;
    }


    /**
     * Whenever a new MainPart is found this function is called for all
     * ConnectingSCTPACParts  of that mainPart, so that matches of the pac
     * part that were in need of the new mainPart can be found.
     */
    public void findNewMatchesUsingMainPart() {
        boolean found = super.findOneEmbedding();

        if (found) {
            checkMatchValidity();

            while (super.findNextEmbedding()) {
                checkMatchValidity();
            }
        }
    }


    /**
     * Tests whether the ConnectingSCTPACPart can be embedded at the node seed
     * into the problem graph. Called when the whole graph should be build
     * Finds only one isomorphism, starting from seed. Seed needs an
     * interpretation before calling this function If none found the root seed
     * interpretation is reset. If the part can not be found and no
     * intersection with other parts occured the frame is removed from the
     * list foundMatches.
     * 
     * @return true if one Embedding was found, else false.
     */
    public boolean findOneEmbedding() {
        dl.print(6, "\n ConnectingPACPart foe");

        boolean result = super.findOneEmbedding();

        if (! result) { // the part could not be found, maybe the used frame is outdated?

            if (! intersectionWithOtherGraphPartOcurred()) {
                // if an intersection occured nothing can be said about the validity of the frame.
                foundMatches.remove(usedGraphPartFrame); // else the frame is not valid no more, remove it.
            }
        }

        return result;
    }


    /**
     * When a new match for the AC is found during
     * ApplicationCondition.checkNodes() if connecting MainParts exist: store
     * frame for this part, if part was missing for mainPart: check main part
     * again, else create whole graph with this part.
     * 
     * @return true
     */
    public boolean handleMatch() {
        if (checkMatchValidity()) {
            ((SCTDockingGraph) mySCGraph).findAllMatchesWithPart(this);
        }

        return true; // ConnectingSCTPACParts allways have to be searched.
    }


    /**
     * When no match was found for the part this funciton is called to prepare
     * the search of the part, and the recheck of the mainPart, once the part
     * is found. Called by ApplicationCondition.testPartAC()
     * 
     * @param mainPartFrame  A frame of the mainPart that is in need of this
     *        part.
     */
    public void registerMainPartFrame(GraphPartFrame mainPartFrame) {
        if (mainPartFrame != null) {
            mainPartsMissingThisACPart.add(mainPartFrame);
        }
    }


    /**
     * Called for every part of AC in ApplicationCondition.testAC(). Searches
     * for an embedding, and if none found initiates search for the ACPart to
     * create the mainPart when the ACPart is created.
     * 
     * @param mainPartFrame  A frame of the mainPart
     * 
     * @return true if the ACPart could be found, else false.
     */
    public boolean testAC(GraphPartFrame mainPartFrame) {
        boolean found = findOneEmbeddingAllParts();

        if (! found) {
            registerMainPartFrame(mainPartFrame);

            Iterator intersectingPartsIt = getIntersectingPartsIterator();

            while (intersectingPartsIt.hasNext()) {
                TypeDockingACPart missingPart = (TypeDockingACPart) intersectingPartsIt.next();
                missingPart.registerMainPartFrame(mainPartFrame);
            }
        }

        return found;
    }


    /**
     * Called for every part of AC in ApplicationCondition.testPartAC().
     * Searches for an embedding, and if none found initiates search for the
     * ACPart to create the mainPart when the ACPart is created.
     * 
     * @param mainPartFrame  A frame of the mainPart
     * 
     * @return true if the ACPart could be found, else false.
     */
    public boolean testPartAC(GraphPartFrame mainPartFrame) {
        boolean found         = false;
        Stack   fittingFrames = getFittingFrames(); // get all the found matches for this partFrame

        if (fittingFrames.empty()) {
            registerMainPartFrame(mainPartFrame);

            return false;
        }

        GraphPartFrame lastUsedFrame = null;

        while (! fittingFrames.empty() && ! found) {
            GraphPartFrame fittingFrame = (GraphPartFrame) fittingFrames.pop();

            if (fittingFrame.setInterpretations()) {
                found         = findOneEmbeddingAllParts();
                lastUsedFrame = fittingFrame;
            }
        }

        if (! found) {
            registerMainPartFrame(mainPartFrame);

            Iterator intersectingPartsIt = getIntersectingPartsIterator();

            while (intersectingPartsIt.hasNext()) {
                ApplicationConditionPart missingPart = 
                        (ApplicationConditionPart) intersectingPartsIt.next();
                ((TypeDockingACPart) missingPart).registerMainPartFrame(
                        mainPartFrame);
            }
        }

        if (lastUsedFrame != null) {
            lastUsedFrame.resetInterpretations();
        }

        return found;
    }


    /**
     * Checks whether the current Embedding is a valid one i.e. the mainParts
     * directly connected exist, and if that is the case tries to rebuild
     * mainParts that were in need of this  ConnectingSCTPACPart.
     * 
     * @return true if the match is good.
     */
    boolean checkMatchValidity() {
        if (existFittingMainParts()) {
            foundMatches.add(createInnerGraphFrame());
            dl.print(6, "\nconnecting pac part frame added");
            checkWrongMainParts();

            return true;
        }

        return false;
    }
} //class
