/*
 * SCTNAC.java
 *
 * Created on 13. Juli 2002, 22:06
 */
package com.ai_center.dragonbreath.structuralConstraintHandling.graph.dockingGraph.applicationCondition.type;

import com.ai_center.dragonbreath.constraintgraph.Node;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.GraphPartFrame;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.applicationCondition.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.dockingGraph.SCTDockingGraph;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.movableSubgraph.*;
import com.ai_center.dragonbreath.utilities.Pair;
import com.ai_center.dragonbreath.utilities.SetHashtable;

import java.util.*;


/**
 * A Negative Application Condition of a DockingGraph Type.
 * 
 * @version $Revision: 1.3 $
 * @author tim baermann
 */
public class TypeDockingNAC extends ApplicationCondition {
    // Instance variables *****************************************************

    /**
     * Fullfilled matches of this NAC that are waiting to be destroyed to
     * create a complete match. Consists in Pairs, each containing as first
     * member a fitting mainPartFrame, to create when the match is destroyed,
     * and second a list of all the nodes in the match
     */
    SetHashtable searchedMatches = new SetHashtable();

    // Constructors ***********************************************************

    /**
     * Creates a new instance of SCTNAC
     * 
     * @param scg The DockingGraph the NAC belongs to.
     */
    public TypeDockingNAC(SCTDockingGraph scg) {
        super(scg);
        scg.addNAC(this);
    }

    // Methods ****************************************************************

    /**
     * When a NAC was found it is stored for search, together with the nodes
     * used in the match.
     * 
     * @param mainPartFrame  A mainPart that can be used to restore the match,
     *        when the NAC was destroyed.
     * @param nodes  All the intepretations used for this match.
     */
    public void registerMatch(GraphPartFrame mainPartFrame, Collection nodes) {
        Pair     match = new Pair(mainPartFrame, nodes);
        Iterator it = nodes.iterator();

        while (it.hasNext()) {
            Node n = (Node) it.next();
            searchedMatches.add(n, match);
        }
    }


    /**
     * Tests wheter the NAC can be embedded, if that is the case the
     * mainPartFrame is stored, together with the found match, so that when
     * the match is destroyed the graph can be restored.
     * 
     * @param mainPartFrame  the frame of the mainPart that should be tested.
     * 
     * @return true if all parts could be found.
     */
    public boolean testAC(GraphPartFrame mainPartFrame) {
        boolean found = findAllParts();

        if (found) {
            Set      nodesInMatch = new HashSet();
            Iterator it = allParts.iterator();

            while (it.hasNext()) {
                MovableSubgraphPart g = (MovableSubgraphPart) it.next();
                nodesInMatch.addAll(g.getInterpretedNodes().keySet());
            } //while

            registerMatch(mainPartFrame, nodesInMatch);
        }

        resetInterpretations();

        return ! found;
    } //testAC()


    /**
     * A nac is never tested for a single mainPart, so the function is never
     * called.
     * 
     * @param mainPartFrame  the frame of the mainPart that should be tested.
     * 
     * @return true if all PACParts could be found, else false.
     */
    public boolean testPartAC(GraphPartFrame mainPartFrame) {
        return false; // a nac is never tested for a single part
    } // testPartAC()


    /**
     * Checks whether the deletion of the nodes deleted one of the stored
     * matches. If a match was destroyed the docking graph is rebuild. It
     * might be that the nac was destroyed, but can be found at the same place
     * with a slightly different match. Then this is done later, when the
     * improved mainPart is tested the next time.(It can not be done now, as
     * it is not known which nodes are used by the mainPart)
     * 
     * @param deletedNodes nodes deleted.
     */
    protected void checkDeletedNodes(Set deletedNodes) {
        try {
            Vector          improvedMainPartMatches = new Vector();
            SCTDockingGraph dg = (SCTDockingGraph) mySCGraph;
            Iterator        it = deletedNodes.iterator();

            while (it.hasNext()) {
                Node   deletedNode = (Node) it.next();
                Object o = searchedMatches.get(deletedNode);

                if (o != null) {
                    searchedMatches.removeKey(deletedNode);

                    if (o instanceof Set) {
                        Iterator matchesIt = ((Set) o).iterator();

                        while (matchesIt.hasNext()) {
                            Pair match = (Pair) matchesIt.next();
                            removeMatch(match);
                            improvedMainPartMatches.add(match.first);
                        }
                    } else {
                        Pair match = (Pair) o;
                        removeMatch(match);
                        improvedMainPartMatches.add(match.first);
                    }
                } //if o!= null
            } //while it

            dg.registerImprovedMainPartMatches(improvedMainPartMatches);
        } //try 
        catch (java.lang.ClassCastException e) {
        }

        Iterator partsIt = allParts.iterator();

        while (partsIt.hasNext()) {
            try {
                UnconnectedACPart part = (UnconnectedACPart) partsIt.next();
                part.testForDeletedSeeds(deletedNodes);
            } catch (ClassCastException e) {
            }
        }
    } //checkDeletedNodes()


    /**
     * Tries to find all the ApplicationConditionParts that were improved due
     * to new nodes.
     * 
     * @param newNodes Nodes added.
     */
    protected void checkNewNodes(Set newNodes) {
        Iterator partsIt = allParts.iterator();

        while (partsIt.hasNext()) {
            Object o = partsIt.next();

            if (o instanceof TypeDockingUnconnectedNACPart) {
                TypeDockingUnconnectedNACPart part = (TypeDockingUnconnectedNACPart) o;
                findNewMatchesForPart(part, newNodes);
            }
        }
    }


    /**
     * Checks whether the deletion of the nodes deleted one of the stored
     * matches. If a match was destroyed the docking graph is rebuild. Calls
     * checkDeletedNodes().
     * 
     * @param reducedNodes
     * @param reductions
     */
    private void checkReducedNodes(Set reducedNodes, SetHashtable reductions) {
        checkDeletedNodes(reducedNodes); // this is cheating, but it works..
    } // checkReducedNodes()   


    /**
     * Removes the given match from the list of searched matches.
     * 
     * @param match  First is a mainPartFrame , second a Map of Nodes and
     *        SCNodes.
     */
    private void removeMatch(Pair match) {
        Set      nodes = (Set) match.second;
        Iterator it = nodes.iterator();

        while (it.hasNext()) {
            Node n = (Node) it.next();
            searchedMatches.remove(n, match);
        }
    } // removeMatch()
} //class
