/*
 * AlternativeSeeds.java
 *
 * Created on 28. Januar 2003, 19:52
 */
package com.ai_center.dragonbreath.structuralConstraintHandling.graph.movableSubgraph;

import com.ai_center.dragonbreath.constraintgraph.Node;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.*;
import com.ai_center.dragonbreath.utilities.*;

import java.util.*;


/**
 * Contains seeds for partial matches.  When partial matches of TestGraphs are
 * searched all the not connected graph parts need a seed, even if yet no
 * total match was found.
 * 
 * @version $Revision: 1.3 $
 * @author tim baermann
 */
public class AlternativeSeeds {
    // Instance variables *****************************************************

    /**
     * The seeds ordered, as Pairs. The size of the match determines in wich
     * Vector the seed is stored.
     */
    Vector[] sortedSeeds;

    /**
     * When iterating through the seeds, this is the size (and the index in
     * sortedSeeds) of the  solutions regarded now.
     */
    int actualSeedSizeIndex;

    /**
     * Each time setFirstSeedInterpretation() is called all the seeds and their
     * interpretations are  added to the sortedSeeds. But not any number of
     * interpretations is allowed. After adding we add up the number of
     * interpretations in sortedSeeds, begining from the biggest to the
     * smallest, when the number allowed is reached the given index determines
     * the size of the smallest match that fits in the number of
     * interpretations we allow. Next call of setFirstSeedInterpretation() all
     * the interpretations are removed that have less matches then
     * smallestMatchThatFitSize.
     */
    int smallestMatchThatFitSize = 0;

    /** 
     * Where the seedInterpretations are stored, with the SCNodes as keys. 
     */
    private HashMap container = new HashMap();

    /** 
     * Here the same seedInterpretaions are stored, but all at once. 
     */
    /**
     * When all the Interpretations for the seeds were tried, but no match
     * found, this Iterator iterates through the nodes in alternativeSeeds via
     * setNextSeed().
     */
    private Iterator actualSeedSizeVectorIt;

    /** 
     * The size of the biggestMatch found. 
     */
    private int biggestMatchFoundSize = 0;

    /** 
     * The size the alternativeSeeds map should maximally have. 
     */
    private int numStoredPartlyFoundMatches = 0;

    // Constructors ***********************************************************

    /**
     * Creates a new instance of AlternativeSeeds
     */
    public AlternativeSeeds() {
    }

    // Methods ****************************************************************

    /**
     * Resets the iteration through the seeds.
     * 
     * @param allowedNumberOfSeedInterpretations The number of seed
     *        interpretations that are stored maximally.
     */
    public void setFirstSeedInterpretation(int allowedNumberOfSeedInterpretations) {
        if (sortedSeeds == null) {
            sortedSeeds = new Vector[allowedNumberOfSeedInterpretations];
        } else if (sortedSeeds.length < allowedNumberOfSeedInterpretations) {
            sortedSeeds = new Vector[allowedNumberOfSeedInterpretations];
        }

        for (int index = 0; index < allowedNumberOfSeedInterpretations; index++) {
            if (sortedSeeds[index] != null) {
                sortedSeeds[index].clear();
            }
        }

        Iterator it = container.keySet().iterator();

        while (it.hasNext()) {
            SCNode   seed            = (SCNode) it.next();
            HashMap  interpretations = (HashMap) container.get(seed);
            Iterator interpretIt     = interpretations.keySet().iterator();

            while (interpretIt.hasNext()) {
                Node    interpretation = (Node) interpretIt.next();

                Integer matchSize = (Integer) interpretations.get(
                                            interpretation);
                int     iMatchSize = matchSize.intValue();

                // if the index is smaller then the size of the last seedInterpretation that was used last
                // time it will probally not be used this time too. Remove it.
                if (iMatchSize < smallestMatchThatFitSize) {
                    interpretIt.remove();
                } else {
                    double quotient = (double) iMatchSize / (double) biggestMatchFoundSize;
                    double mult  = quotient * allowedNumberOfSeedInterpretations;
                    int    index = (int) java.lang.Math.ceil(mult);

                    if (index > allowedNumberOfSeedInterpretations) {
                        index = allowedNumberOfSeedInterpretations; //should never happen
                    } else if (index < 1) {
                        index = 1; // neither
                    }

                    if (sortedSeeds[index - 1] == null) { // - 1, as the index begins at 0, and the sizes at 1 
                        sortedSeeds[index - 1] = new Vector();
                    }

                    sortedSeeds[index - 1].add(new Pair(seed, interpretation));
                }
            } //while

            if (interpretations.size() == 0) {
                it.remove();
            }
        } //while

        actualSeedSizeVectorIt = null;

        int actualSize = 0;
        int index;

        // now compute the smallest index that fits 
        for (index = allowedNumberOfSeedInterpretations - 1;
             (index >= 0) && (actualSize < allowedNumberOfSeedInterpretations);
             index--) {
            if (sortedSeeds[index] != null) {
                if (actualSeedSizeVectorIt == null) {
                    actualSeedSizeVectorIt = sortedSeeds[index].iterator();
                    actualSeedSizeIndex    = index;
                }

                actualSize += sortedSeeds[index].size();
            }
        }

        smallestMatchThatFitSize = index;
    }


    /**
     * Returns the next seed in order. The seed has allready an interpretation
     * set. The method setFirstSeedInterpretation() must be called befor
     * calling setNextSeed().
     * 
     * @param biggestEmbeddingSizeSoFar The size of the biggest embedding that
     *        was found with the seeds returned before.
     * 
     * @return A SCNode that is the next seed, with an interpretation set.  If
     *         the embeddingSize of the next node is smaller then the
     *         biggestEmbeddingSize null is returned.
     */
    public SCNode setNextSeed(int biggestEmbeddingSizeSoFar) {
        if ((actualSeedSizeVectorIt != null) && 
                (actualSeedSizeIndex > biggestEmbeddingSizeSoFar)) {
            while (actualSeedSizeVectorIt.hasNext()) {
                Pair   nextSeedInterpretation = (Pair) actualSeedSizeVectorIt.next();
                SCNode seed           = (SCNode) nextSeedInterpretation.first;
                Node   interpretation = (Node) nextSeedInterpretation.second;

                if (seed.setInterpretationBiggestEmbedding(interpretation)) {
                    return seed;
                }

                if (! actualSeedSizeVectorIt.hasNext()) {
                    // there is no mor partly found match of the actual size, decrese size
                    boolean found = false;

                    while (! found) {
                        actualSeedSizeIndex--;

                        if (actualSeedSizeIndex < 0) {
                            return null;
                        } else if (sortedSeeds[actualSeedSizeIndex] != null) {
                            actualSeedSizeVectorIt = sortedSeeds[actualSeedSizeIndex].iterator();
                            found                  = true;
                        }
                    } //while
                } //if
            } //while  
        } //if 

        return null;
    }


    /**
     * Updates the seed s data. When a seed was used it can allways happen,
     * that a bigger embedding was found then before (e.g. due to repair) then
     * this function updates the seeds data.
     * 
     * @param seed The seed that was changed.
     * @param interpretation The interpretation for the seed.
     * @param matchSize The new size for the match found with seed.
     */
    public void actualizeAlternativeSeed(SCNode seed, Node interpretation, 
                                         int matchSize) {
        addAlternativeSeed(seed, interpretation, matchSize);
    }


    /**
     * Adds a new alternative seed.
     * 
     * @param seed The new seed node.
     * @param interpretation The interpretation for the seed.
     * @param matchSize The size of the match that could be found using the
     *        given seed.
     */
    public void addAlternativeSeed(SCNode seed, Node interpretation, 
                                   int matchSize) {
        HashMap existingSeedInterpretations = (HashMap) container.get(seed);

        if (existingSeedInterpretations != null) {
            Integer size = (Integer) existingSeedInterpretations.get(
                                   interpretation);

            if (size != null) {
                if (size.intValue() < matchSize) {
                    existingSeedInterpretations.put(interpretation, 
                                                    new Integer(matchSize));

                    if (matchSize > biggestMatchFoundSize) {
                        biggestMatchFoundSize = matchSize;
                    }
                }
            } else {
                existingSeedInterpretations.put(interpretation, 
                                                new Integer(matchSize));

                if (matchSize > biggestMatchFoundSize) {
                    biggestMatchFoundSize = matchSize;
                }
            }
        } else {
            existingSeedInterpretations = new HashMap();
            existingSeedInterpretations.put(interpretation, 
                                            new Integer(matchSize));
            container.put(seed, existingSeedInterpretations);

            if (matchSize > biggestMatchFoundSize) {
                biggestMatchFoundSize = matchSize;
            }
        }
    }


    /**
     * Removes all the seedInterpretations that were deleted.
     * 
     * @param deletedNodes Nodes that were deleted.
     */
    public void testForDeletedSeeds(Collection deletedNodes) {
        Iterator delNodesIt = deletedNodes.iterator();

        while (delNodesIt.hasNext()) {
            Node     interpretation = (Node) delNodesIt.next();
            Iterator it = container.keySet().iterator();

            while (it.hasNext()) {
                SCNode seed = (SCNode) it.next();

                if (seed.typeID == interpretation.typeID) {
                    HashMap interpretationsForSeed = (HashMap) container.get(
                                                             seed);
                    interpretationsForSeed.remove(interpretation);

                    if (interpretationsForSeed.isEmpty()) {
                        it.remove();
                    }
                }
            } //while
        }
    } //testForDeletedSeeds()
}