/*
 * GraphPartIntersection.java
 *
 * Created on 6. Juni 2002 22:22
 */
package com.ai_center.dragonbreath.structuralConstraintHandling.graph.movableSubgraph;

import com.ai_center.dragonbreath.constraintgraph.Node;
import com.ai_center.dragonbreath.utilities.*;

import java.util.*;


/**
 * When a MovableSubgraph consists of multiple parts, and while searching for
 * the parts there appear intersections they are stored here.
 * 
 * @version $Revision: 1.3 $
 * @author tim baermann
 */
public class GraphPartIntersections {
    // Instance variables *****************************************************

    /**
     * A mapping from the MovableSubgraphParts to the number of nodes in the
     * intersection.
     */
    HashMap intersectingParts;

    /**
     * A mapping from the intersecting nodes to the MovableSubgraphParts of the
     * intersection.
     */
    SetHashtable nodesInIntersections;

    // Constructors ***********************************************************

    /**
     * Normal Constructor.
     */
    public GraphPartIntersections() {
        intersectingParts    = new HashMap();
        nodesInIntersections = new SetHashtable();
    }

    // Methods ****************************************************************

    /**
     * When it was tried to give the intersecting parts a new interpretation to
     * resolve the intersections it might happen, that after that try a part
     * has no interpretation any more, as no interpretation without the nodes
     * in the intersection could be found.
     * 
     * @return All the parts in the intersection that have no more
     *         interpretation assigned.
     */
    Set getUninterpretedParts() {
        HashSet  result = new HashSet();
        Iterator it = intersectingParts.keySet().iterator();

        while (it.hasNext()) {
            MovableSubgraphPart part = (MovableSubgraphPart) it.next();

            if (! part.isInterpreted()) {
                result.add(part);
            }
        }

        return result;
    }


    /**
     * The Part part uses the same node (interpretation) as the calling Part.
     * 
     * @param part  The part that intersected.
     * @param interpretation  A node in the intersection.
     */
    void addIntersection(MovableSubgraphPart part, Node interpretation) {
        if (! nodesInIntersections.contains(interpretation, part)) {
            IntValue number = (IntValue) intersectingParts.get(part);

            if (number == null) {
                intersectingParts.put(part, new IntValue(1));
            } else {
                number.value++;
            }

            nodesInIntersections.add(interpretation, part);
        }
    }


    /**
     * Removes all stored intersections.
     */
    void clear() {
        nodesInIntersections.clear();
        intersectingParts.clear();
    }


    /**
     * Is there an intersection stored?
     * 
     * @return true if an intersection exists, else false.
     */
    boolean hasIntersection() {
        return intersectingParts.size() > 0;
    }


    /**
     * Returns an Iterator that iterates over all the Parts stored.
     * 
     * @return The Iterator.
     */
    Iterator intersectingPartsIterator() {
        return intersectingParts.keySet().iterator();
    }


    /**
     * When the given interpretation is no longer used by the calling
     * MovableSubgraphPart it can be removed. When all the intersections of
     * one part were removed the intersecting part is removed too.
     * 
     * @param interpretation  An interpretation no longer intersecting.
     */
    void removeIntersection(Node interpretation) {
        Object o = nodesInIntersections.get(interpretation);

        if (o instanceof Set) {
            Set      s  = (Set) o;
            Iterator it = s.iterator();

            while (it.hasNext()) {
                IntValue i = (IntValue) intersectingParts.get(it.next());

                if (i != null) {
                    i.value--;

                    if (i.value == 0) {
                        it.remove(); // when the last node from the intersection with the part was removed there is no intersection with the part any more.
                    }
                }
            }
        } else {
            IntValue i = (IntValue) intersectingParts.get(o);

            if (i != null) {
                i.value--;

                if (i.value == 0) { // when the last node from the intersection with the part was removed there is no intersection with the part any more.
                    intersectingParts.remove(o);
                }
            }
        }

        nodesInIntersections.removeKey(interpretation);
    }


    /**
     * How many intersections are stored?
     * 
     * @return number of intersections stored.
     */
    int size() {
        return intersectingParts.keySet().size();
    }
} // GraphPartIntersections
