/*
 * OneEmbeddingSubgraph.java
 *
 * Created on 12. Oktober 2002, 16:39
 */
package com.ai_center.dragonbreath.structuralConstraintHandling.graph.movableSubgraph;

import com.ai_center.dragonbreath.constraintgraph.Node;
import com.ai_center.dragonbreath.structuralConstraintHandling.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.applicationCondition.*;
import com.ai_center.dragonbreath.utilities.SetHashtable;

import java.util.*;


/**
 * If there has to be found only one embedding for a Subgraph the parts can be
 * shifted around while searching for an embedding, so that in the end there
 * is at least one valid embedding for each part. That is the case with
 * ApplicationConditions, and the TestGraph.
 * 
 * @version $Revision: 1.7 $
 * @author tim baermann
 */
public class MovableSubgraph extends StructuralConstraintGraphPart
    implements NodeChangedInterface {
    // Instance variables *****************************************************

    /** 
     * Used for debug output, to hold the actual tab width. 
     */
    public String TAB = new String("");

    /**
     * A nac or pac or a testgraph can have several not connected clusters this
     * vector links at all parts.
     */
    public Vector allParts = new Vector();

    /**
     * After searching for an embedding of allParts and not finding it, this is
     * the first part that could not be found.
     */
    protected MovableSubgraphPart firstMissingPart = null;

    /**
     * Nodes that should not be used when searching for an interpretation, used
     * when multiple parts exist, and one part intersects with an other one
     * Used in checkChangedNodes() to garantie that nodes allready checked are
     * not used again in embeddings.
     */
    HashSet allreadyCheckedNewNodes = new HashSet();

    /**
     * The mapping from structural constraint graph g to problem graph p must
     * be injective, i.e. no two different nodes from g may have the same
     * interpretation in p. This hashmap contains all interpretations used in
     * this graph as keys and the nodes that have the mapping as values.
     */
    HashSet usedInterpretations = new HashSet();

    /**
     * When while search for an embedding two parts tried to use the same node
     * they overlapped. If no embedding could be found, and overlap is true an
     * embedding might exist when the overlap is avoided.
     */
    boolean nodeIntersection;

    /**
     * While search for new part match in checkNewNodes() this variable is set
     * true when the part is searched, and false when it was found, and the
     * rest of the graph is searched. All the matches with the nodes in
     * allreadyCheckedNewNodes are known to be allready found. So a match with
     * these nodes can not lead to a new match  => stop search by forbidding
     * the interpretation in tryInterpretation().
     */
    boolean searchWithoutChecked = false;

    /** 
     * Number of nodes in all Parts. 
     */
    private int subgraphSize = -1;

    // Constructors ***********************************************************

    /**
     * Creates a new instance of OneEmbeddingSubgraph
     * 
     * @param scg The Graph the movable part belongs to.
     */
    public MovableSubgraph(StructuralConstraintGraph scg) {
        super(scg);

        //dl.switchOn(4);// uncomment for debug output
        //dl.switchOn(12);// uncomment for debug output
    }

    // Methods ****************************************************************

    /**
     * Returns the number of actually interpreted edges + the number of
     * actually interpreted nodes of all parts.
     * 
     * @return The size of the current match.
     */
    public int getActualMatchSize() {
        int      size = 0;
        Iterator it = allParts.iterator();

        while (it.hasNext()) { //check if any part contains the given interpretation

            MovableSubgraphPart part = (MovableSubgraphPart) it.next();
            size += part.getActualMatchSize();
        }

        return size;
    }


    /**
     * Returns all the interpretations used in this MovableSubgraph.
     * 
     * @return all the interpretations used in the MovableSubgraph
     */
    public Collection getAllInterpretations() {
        HashSet             result = new HashSet();
        MovableSubgraphPart part;
        Iterator            partIt = allParts.iterator();

        while (partIt.hasNext()) {
            part = (MovableSubgraphPart) partIt.next();
            result.addAll(part.getAllInterpretations());
        } //while

        return result;
    }


    /**
     * Returns allParts.  Carefull, returns the original Vector, do not change.
     * 
     * @return allParts
     */
    public Vector getAllParts() {
        return allParts;
    }


    /**
     * If the Graph has an interpretation this function returns a hash for the
     * graph. As two MovableSubgraphs are the same if only the links to main
     * parts are the same only the linked nodes are used for the hash.
     * 
     * @return Hash of the Interpretations. It is circa a concatenation of all
     *         the hashes of all the linked nodes.
     */
    public java.lang.String getHashCodeString() {
        String   result = new String();
        Iterator it = allParts.iterator();

        while (it.hasNext()) {
            MovableSubgraphPart g = (MovableSubgraphPart) it.next();
            result = result.concat(g.getHashCodeString());
        }

        return result;
    }


    /**
     * Returns all nodes of all Parts that have a link assigned to an other
     * part of the same Graph.
     * 
     * @return The Nodes of linked nodes in this graph.
     */
    public Set getInnerGraphLinks() {
        HashSet  result = new HashSet();

        Iterator it = allParts.iterator();

        while (it.hasNext()) {
            MovableSubgraphPart g = (MovableSubgraphPart) it.next();
            result.addAll(g.getInnerGraphLinkNodes());
        }

        return result;
    }


    /**
     * Returns the number of nodes of all graphs minus the number of nodes
     * found for embedding.
     * 
     * @return The distance of this Subgraph and the largest found Embedding.
     */
    public int getMinMatchError() {
        if (subgraphSize == -1) {
            computeSubgraphSize();
        }

        return subgraphSize - getActualMatchSize();
    }


    /**
     * Returns the node of the StructuralConstraintGraphPart that has the given
     * Node as interpretation, null if none.
     * 
     * @param interpretation  The searched interpretation.
     * 
     * @return The SCNode that has interpretation as interpretation or null if
     *         none exists.
     */
    public SCNode getNodeWithInterpretation(Node interpretation) {
        SCNode   result = null;
        Iterator it = allParts.iterator();

        while (it.hasNext() && (result == null)) { //check if any part contains the given interpretation

            MovableSubgraphPart part = (MovableSubgraphPart) it.next();
            result = part.getNodeWithInterpretation(interpretation);
        } //while

        return result;
    }


    /**
     * Returns all the parts that have no link.
     * 
     * @return A Collection with MovableSubgraphParts that are not linked.
     */
    public Collection getUnconnectedParts() {
        Vector   result = new Vector();
        Iterator it = allParts.iterator();

        while (it.hasNext()) {
            MovableSubgraphPart g = (MovableSubgraphPart) it.next();

            if (! g.hasInnerGraphLinks() && ! g.hasInterGraphLinks()) {
                result.add(g);
            }
        }

        return result;
    }


    /**
     * Enters an interpreatation of a node as allready used. The mapping from
     * structural constraint graph g to problem graph p must be injective,
     * i.e. no two different nodes from p may have the same interpretation in
     * p. (a link between two nodes means in this sense that they are the same
     * node, ergo they have to have the same interpretation)
     * 
     * @param interpretation The interpretation.
     * @param node  The node that gets the interpretation.
     */
    public void addInterpretation(Node interpretation, SCNode node) {
        SCNode nodeWithInterpretation = (SCNode) interpretedNodes.get(
                                                interpretation);

        if (nodeWithInterpretation == null) {
            interpretedNodes.put(interpretation, node);
            mySCGraph.addInterpretation(interpretation, node);
        } else if (nodeWithInterpretation == node) {
            throw new Error(
                    "Interpretation was allready appointed for the node");
        } else {
            throw new Error("Interpretation allready used by other node");
        }
    }


    /**
     * Adds a part. Called by constructor of MovableSubgraphPart
     * 
     * @param part  The new MovableSubgraphPart.
     */
    public void addPart(MovableSubgraphPart part) {
        allParts.add(part);
    } //addPart()


    /**
     * Was the Interpretation allready checked in this turn? If it was, it is
     * known that all embeddings using the node were allready found before.
     * 
     * @param interpretation The interpretation that should be checked.
     * 
     * @return true if the interpretation was allready checked, else false.
     */
    public boolean allreadyCheckedInterpretation(Node interpretation) {
        if (searchWithoutChecked) {
            if (allreadyCheckedNewNodes.contains(interpretation)) {
                return true;

                // all the matches with the nodes in allreadyCheckedNewNodes are known to be found
                // before. So a match with this node can not lead to a new match
                // => stop search by forbidding the interpretation.
            } //if
        } //if

        return false;
    }


    /**
     * Checks if the changed nodes modify the matching of this GraphPart.
     * 
     * @param newNodes Nodes that were recently created.
     * @param deletedNodes The nodes deleted.
     * @param expandedNodes Expanded Nodes
     * @param reducedNodes Nodes that lost connections.
     */
    public void checkChangedNodes(Set newNodes, Set deletedNodes, 
                                  Set expandedNodes, Set reducedNodes) {
        checkNewNodes(newNodes);
        checkDeletedNodes(deletedNodes);
        checkReducedNodes(reducedNodes);
        checkNewNodes(expandedNodes);
    }


    /**
     * Is the given Interpretation used in one of the Parts? Or is it used by a
     * mainPart? An intersection is not registered.
     * 
     * @param interpretation The checked interpretation.
     * 
     * @return true if the interpretation was allready used, else false.
     */
    public boolean containsInterpretation(Node interpretation) {
        if (mySCGraph.containsInterpretation(interpretation)) { // first check if a main graph allready uses the interpretation

            return true;
        }

        Iterator it = allParts.iterator();

        while (it.hasNext()) { //check if any part contains the given interpretation

            MovableSubgraphPart part = (MovableSubgraphPart) it.next();

            if (part.interpretedNodesContains(interpretation)) {
                return true;
            }
        } //while

        return false;
    }


    /**
     * Calls for all parts copyInterpretations() with the respective part of
     * the source. Both have to have the same structure!
     * 
     * @param source  The Subgraph with the interpretations.
     * 
     * @throws Error If the structure is differently.
     */
    public void copyInterpretations(StructuralConstraintGraphPart source) {
        Iterator destIt = allParts.iterator();
        Iterator srcIt = ((MovableSubgraph) source).allParts.iterator();

        while (destIt.hasNext()) {
            if (! srcIt.hasNext()) {
                throw new Error("Subgraphs should have the same stucture.");
            }

            MovableSubgraphPart g = (MovableSubgraphPart) destIt.next();
            g.copyInterpretations((MovableSubgraphPart) srcIt.next());
        }
    } //copyInterpretations()


    /**
     * Embedds all parts of the MovableSubgraph.
     * 
     * @return true if all parts could be found.
     */
    public boolean findAllParts() {
        boolean found = true;
        firstMissingPart = null;

        Iterator it = allParts.iterator();

        while (it.hasNext() && found) { // while all parts are found

            MovableSubgraphPart part = (MovableSubgraphPart) it.next();
            found = part.findOneEmbeddingAllParts();

            if (! found) {
                firstMissingPart = part;
            }
        }

        return found;
    }


    /**
     * Does nothing, overwritten by subclasses.
     * 
     * @return true
     */
    public boolean handleMatch() {
        return true;
    }


    /**
     * How many parts has the MovableSubgraph?
     * 
     * @return The number of parts.
     */
    public int numberOfParts() {
        return allParts.size();
    }


    /**
     * If an interpretation is not used any more it should be removed from list
     * of used interpretations by this function. The mapping from structural
     * constraint graph g to problem graph p must be injective, i.e. no two
     * different nodes from p may have the same interpretation in p.
     * 
     * @param interpretation The interpretation.
     * @param node The node that has the given interpretation.
     */
    public void removeInterpretation(Node interpretation, SCNode node) {
        if (interpretation != null) {
            SCNode nodeWithInterpretation = (SCNode) interpretedNodes.get(
                                                    interpretation);

            if (nodeWithInterpretation == node) {
                interpretedNodes.remove(interpretation);
                mySCGraph.removeInterpretation(interpretation, node);
            } else if (nodeWithInterpretation == null) {
                throw new Error("interpretation not part of subgraph!");
            }
        }
    }


    /**
     * Resets the interpretations of this part, or if parameter true the
     * interpretations of allParts
     */
    public void resetInterpretations() {
        Iterator it = allParts.iterator();

        while (it.hasNext()) {
            MovableSubgraphPart g = (MovableSubgraphPart) it.next();
            g.resetInterpretations();
        }
    }


    /**
     * Resets the extern interpretations of all nodes and edges in the graph.
     */
    public void resetInterpretationsExtern() {
        Iterator it = allParts.iterator();

        while (it.hasNext()) {
            MovableSubgraphPart g = (MovableSubgraphPart) it.next();
            g.resetInterpretationsExtern();
        }
    }


    /**
     * Called  from checkChangedNodes().
     * 
     * @param deletedNodes The nodes that were deleted.
     */
    protected void checkDeletedNodes(Set deletedNodes) {
        Iterator partsIt = allParts.iterator();

        while (partsIt.hasNext()) {
            Object o = partsIt.next();

            if (o instanceof UnconnectedGraphPartInterface) {
                UnconnectedGraphPartInterface part = (UnconnectedGraphPartInterface) o;
                part.testForDeletedSeeds(deletedNodes);
            }
        }
    }


    /**
     * Tries to find all the parts that were improved due to new nodes.
     * 
     * @param newNodes New nodes in the SCGraph.
     */
    protected void checkNewNodes(Set newNodes) {
        Iterator partsIt = allParts.iterator();

        while (partsIt.hasNext()) {
            MovableSubgraphPart part = (MovableSubgraphPart) partsIt.next();
            findNewMatchesForPart(part, newNodes);
        }
    }


    /**
     * Tries to find matches of the given part with the new nodes. Calls
     * part.handleMatch() if it could be found.
     * 
     * @param part The tested Part.
     * @param newNodes Nodes that should be contained in the found matches.
     */
    protected void findNewMatchesForPart(MovableSubgraphPart part, Set newNodes) {
        part.sortNodesByFrequency(newNodes);

        // sort the new nodes ascending with the frequency in the searchedNodesContainer,
        // first the new Nodes which are less frequent in searchedNodes
        // are tested, and later the more frequent ones.
        // thereby the ramification is more at the end of the backtracking, and not at the root.
        Iterator it = newNodes.iterator();

        while (it.hasNext()) { // check all newNodes, whether one of them leads to a new match

            Node     newNode        = (Node) it.next();
            Set      s              = part.getSearchedNodes(newNode);
            Iterator missingNodesIt = s.iterator();

            while (missingNodesIt.hasNext()) {
                SCNode n = (SCNode) missingNodesIt.next();

                if (n.setInterpretation(newNode)) {
                    if (findOnePartEmbeddingWithoutCheckedNodes(n) && 
                            part.handleMatch()) {
                        while (findNextPartEmbeddingWithoutCheckedNodes(n) && 
                               part.handleMatch()) {
                            // do nothing..
                        }
                    } // if findOnePartEmbeddingWithoutCheckedNodes
                } // if setInterpretation
            } // while n.missingNodesIt

            allreadyCheckedNewNodes.add(newNode);
        } // while it

        allreadyCheckedNewNodes.clear();
    }


    /**
     * Calls part.findOnePartEmbedding(n) but sets the variable
     * searchWithoutChecked = true so that the search doesnt find matches with
     * allready checked nodes.
     * 
     * @param n  The node that should be embedded.
     * 
     * @return true if an embedding was found.
     */
    protected boolean findOnePartEmbeddingWithoutCheckedNodes(SCNode n) {
        boolean             result;
        MovableSubgraphPart part = (MovableSubgraphPart) n.myGraphPart;
        searchWithoutChecked = true; // switch on that the new nodes allready checked can not be part of the interpretations found now.
        result               = part.findOneEmbedding(n);
        searchWithoutChecked = false;

        return result;
    }


    /**
     * Is the given interpretation allready used in this graph? The mapping
     * from structural constraint graph g to problem graph p must be
     * injective, i.e. no two different nodes from g may have the same
     * interpretation in p. If the interpretation is used in an other Part the
     * function tries to find other interpretations  for the other parts in
     * order to free this interpretation.
     * 
     * @param interpretation  The interpretation.
     * @param node The node that should get the interpretation.
     * @param useOtherPartNodes  Return false if the interpretation is allready
     *        used by other part? or just register the intersection?
     * @param caller  The MovableSubgraph that called the function, used to
     *        return a possible intersection.
     * 
     * @return true if the subgraph contains the interpretation, else false.
     */
    boolean tryInterpretation(Node interpretation, SCNode node, 
                              boolean useOtherPartNodes, 
                              MovableSubgraphPart caller) {
        if (allreadyCheckedInterpretation(interpretation)) { //or was the interpretation allready checked?

            return true;
        }

        boolean  interpretationContainedInOtherPart = false;
        boolean  contains = false;

        Iterator it = allParts.iterator();

        while (it.hasNext() && ! contains) { //check if any part contains the given interpretation

            MovableSubgraphPart part = (MovableSubgraphPart) it.next();

            if (part != caller) {
                contains = part.interpretedNodesContains(interpretation);

                if (contains) {
                    if (useOtherPartNodes) { // if the embedding may contain nodes allready used by other parts contains is set to false
                        contains                           = false;
                        interpretationContainedInOtherPart = true;
                    }

                    caller.registerIntersection(part, interpretation, 
                                                useOtherPartNodes);
                    intersectionWithOtherGraphPart = true;
                }
            }
        } //while

        if (! contains && ! interpretationContainedInOtherPart) {
            contains = mySCGraph.tryInterpretation(interpretation, node);
        }

        return contains;
    } //boolean tryInterpretation(Node interpretation)


    /**
     * Called  from checkChangedNodes().
     * 
     * @param reducedNode Nodes that lost connections.
     */
    private void checkReducedNodes(Set reducedNode) {
        //implemented for nacs
    }


    /**
     * Adds up the number of nodes of all parts, and stores it in allNodesSize.
     */
    private void computeSubgraphSize() {
        subgraphSize = 0;

        Iterator it = allParts.iterator();

        while (it.hasNext()) {
            MovableSubgraphPart g = (MovableSubgraphPart) it.next();
            subgraphSize += g.size();
        }
    }


    /**
     * Calls part.findNextPartEmbedding(n) but sets the variable
     * searchWithoutChecked = true so that the search doesnt find matches with
     * allready checked nodes.
     * 
     * @param n  The node that should be embedded.
     * 
     * @return true if an embedding was found, else false.
     */
    private boolean findNextPartEmbeddingWithoutCheckedNodes(SCNode n) {
        boolean             result;
        MovableSubgraphPart part = (MovableSubgraphPart) n.myGraphPart;
        searchWithoutChecked = true; // switch on that the new nodes allready checked can not be part of the interpretations found now.
        result               = part.findNextEmbedding();
        searchWithoutChecked = false;

        return result;
    }
}