/*
 * MovableSubgraphPart.java
 *
 * Created on 13. Oktober 2002, 15:42
 */
package com.ai_center.dragonbreath.structuralConstraintHandling.graph.movableSubgraph;

import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.constraintgraph.Node;
import com.ai_center.dragonbreath.structuralConstraintHandling.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.*;
import com.ai_center.dragonbreath.utilities.*;

import java.util.*;


/**
 * A part of a MovableSubgraph. It provides functions to search for a match by
 * shifting all the parts of the MovableSubgraph around.
 * 
 * @version $Revision: 1.5 $
 * @author tim baermann
 */
public class MovableSubgraphPart extends StructuralConstraintGraphPart {
    // Static variables/initializers ******************************************

    static String TAB = new String();

    // Instance variables *****************************************************

    /** 
     * The the MovableSubgraph this part belongs to. 
     */
    public MovableSubgraph mySubgraph;

    /**
     * Parts that were intersecting with the embedding found in
     * findOneEmbeddingWithOtherPartNodes()
     */
    GraphPartIntersections intersectingParts = new GraphPartIntersections();

    /**
     * Nodes that should not be used when searching for an interpretation, used
     * when multiple parts exist, and one part intersects with an other one.
     */
    HashSet noUseInterpretations = new HashSet();

    /**
     * When during search the interpretation is copied, a new one tried, but
     * doesnt work and the old one reset by setOriginalInterpretations() this
     * is true, it modifies the behaviour of tryInterpretation().
     */
    boolean settingOriginalInterpretations = false;

    /**
     * Should the containsInterpretation() function return true if the
     * interpretation is used by an other part of the Subgraph?
     */
    boolean useOtherPartNodes = false;

    // Constructors ***********************************************************

    /**
     * Creates a new instance of MovableSubgraphPart
     * 
     * @param ms  The MovableSubgraph this part belongs to.
     */
    public MovableSubgraphPart(MovableSubgraph ms) {
        super(ms.mySCGraph);
        mySubgraph = ms;
        mySubgraph.addPart(this);

				// uncomment for debug output
        //dl.switchOn(3);
        //dl.switchOn(5); 
        // dl.switchOn(7); 
    }

    // Methods ****************************************************************

    /**
     * Returns an Iterator over all the parts that intersected during match
     * search.
     * 
     * @return An Iterator over all the parts in the intersection.
     */
    public Iterator getIntersectingPartsIterator() {
        return intersectingParts.intersectingPartsIterator();
    }


    /**
     * Enters an interpreatation of a node as allready used. The mapping from
     * structural constraint graph g to problem graph p must be injective,
     * i.e. no two different nodes from p may have the same interpretation in
     * p. (a link between two nodes means in this sense that they are the same
     * node, ergo they have to have the same interpretation)
     * 
     * @param interpretation The interpretation.
     * @param node  The node that gets the interpretation.
     */
    public void addInterpretation(Node interpretation, SCNode node) {
        SCNode nodeWithInterpretation = (SCNode) interpretedNodes.get(
                                                interpretation);

        if (nodeWithInterpretation == null) {
            interpretedNodes.put(interpretation, node);

            if (! useOtherPartNodes) {
                mySubgraph.addInterpretation(interpretation, node);
            }
        } else if (nodeWithInterpretation == node) {
            throw new Error(
                    "Interpretation was allready appointed for the node");
        } else {
            throw new Error("Interpretation allready used by other node");
        }
    }


    /**
     * Same as tryInterpretation, but doesnt register an intersection And also
     * doesn`t check the interpretation with the SCGraph
     * 
     * @param interpretation The checked interpretation.
     * 
     * @return true if the interpretation was allready used, else false.
     */
    public boolean containsInterpretation(Node interpretation) {
        if (interpretedNodes.containsKey(interpretation)) {
            return true; // the interpretation is used in this part allready
        }

        if (noUseInterpretations.contains(interpretation) || 
                mySubgraph.containsInterpretation(interpretation)) {
            intersectionWithOtherGraphPart = true;

            return true; // if the interpretation should not be used in the part (set for example in findOneEmbeddingWithoutNodes() )
        }

        return false;
    }


    /**
     * Tests wether the Part can be embedded at the node seed into the problem
     * graph. If other parts of the MovableSubgraph are in the way they are
     * moved to find an embedding.
     * 
     * @return true if an embedding was found, else false.
     */
    public boolean findOneEmbeddingAllParts() {
        return findOneEmbeddingWithoutNodes(new HashSet());
    } // findOneEmbeddingAllParts()


    /**
     * Tries to find one embedding for the part that doesnt contain the given
     * nodes. If necessary moves other parts to find an embedding.
     * 
     * @param noUseNodes The nodes that should not be contained in the
     *        embedding.
     * 
     * @return true if an embedding was found, else false.
     */
    public boolean findOneEmbeddingWithoutNodes(HashSet noUseNodes) {
        if (this.isInterpreted() && 
                (! containsOneOfInterpretations(noUseNodes))) { // if there is allready an interpretation without any of the noUseNodes

            return true;
        }

        String TABold    = new String(TAB);
        int    tablength = TAB.length();
        TAB = TAB.concat("    ");
        dl.print(3, 
                 "\n" + TAB + "[" + mySubgraph.allParts.indexOf(this) + 
                 "]findOneEmbeddingWithoutNodes");

        dl.print(3, " /{");
        dl.printNodeList(3, noUseNodes);
        dl.print(3, "}");

        if (this instanceof UnconnectedGraphPartInterface) {
            dl.print(3, 
                     " #seeds:" + 
                     ((UnconnectedGraphPartInterface) this).nrSeedInterpretations());
        }

        dl.print(3, "   (");
        dl.printNodeList(3, mySubgraph.getInterpretedNodes());
        dl.print(3, ")");

        boolean    wasInterpretedBeforeCall = false;
        Collection interpretationsCopy = null;

        if (isInterpreted()) {
            interpretationsCopy      = getCopyOfInterpretations();
            wasInterpretedBeforeCall = true;
            dl.print(3, "\n" + TAB + "interpretation before(");
            dl.printNodeList(3, interpretedNodes);
            dl.print(3, ")");
        }

        resetInterpretations();
        resetIntersection();

        int     maxMatchSizeWithoutOtherPartNodes = 0;
        boolean found = false;
        this.noUseInterpretations = (HashSet) noUseNodes.clone();

        if (findOneEmbeddingWithoutOtherPartNodes()) { //first try to find one simple embedding 
            found = true;
        } else { // the embedding could not be found

            if (intersectionWithOtherGraphPart) { // if that is due to an intersection

                if (findOneEmbeddingWithOtherPartNodes()) {
                    dl.print(3, 
                             "\n" + TAB + "[" + 
                             mySubgraph.allParts.indexOf(this) + "]1 uses: ");
                    dl.printNodeList(3, interpretedNodes);

                    if (! findOtherEmbeddingForIntersectingParts()) {
                        while (! found && 
                               findNextEmbeddingWithOtherPartNodes()) { // while new embeddings can still be found
                            dl.print(3, 
                                     "\n" + TAB + "[" + 
                                     mySubgraph.allParts.indexOf(this) + 
                                     "]2 uses: ");
                            dl.printNodeList(3, interpretedNodes);
                            found = findOtherEmbeddingForIntersectingParts();
                        } // while
                    } else { //if ! findOtherEmbeddingForIntersectingParts
                        found = true;
                    }

                    if (found) {
                        flushUsedInterpretations();
                    }
                } //if findOneEmbeddingWithOtherPartNodes
            } //if nodeIntersection 
        } //else findOneEmbeddingWithoutOtherPartNodes  

        this.noUseInterpretations.clear();

        if (! found) {
            dl.print(3, 
                     "\n" + TAB + "[" + mySubgraph.allParts.indexOf(this) + 
                     "]not found ");

            if (wasInterpretedBeforeCall) {
                setOriginalInterpretations(interpretationsCopy);
            }

            dl.print(3, "\n" + TAB + "interpretation after (");
            dl.printNodeList(3, interpretedNodes);
            dl.print(3, ")");
            dl.print(3, "   (");
            dl.printNodeList(3, mySubgraph.getInterpretedNodes());
            dl.print(3, ")");
        } else {
            dl.print(3, 
                     "\n" + TAB + "[" + mySubgraph.allParts.indexOf(this) + 
                     "]found: ");
            dl.printNodeList(3, interpretedNodes);
            dl.print(3, "   (");
            dl.printNodeList(3, mySubgraph.getInterpretedNodes());
            dl.print(3, ")");
        }

        TAB = TABold;

        return found;
    } // findOneEmbeddingWithoutNodes()


    /**
     * Same as findOneEmbedding(), in contrast to
     * findOneEmbeddingWithOtherPartNodes() this function doesnt use nodes
     * allready used in other parts of the MovableSubgraph.
     * 
     * @return true if an embedding was found, else false.
     */
    public boolean findOneEmbeddingWithoutOtherPartNodes() {
        useOtherPartNodes = false;

        boolean result = findOneEmbedding();

        return result;
    }


    /**
     * Implemented in subclasses. When a new match for the Part is found during
     * checkNodes() this function performs the actions that have to be done.
     * 
     * @return true
     */
    public boolean handleMatch() {
        return true;
    }


    /**
     * Returns interpretedNodes.containsKey( interpretation ). Called from
     * MovableSubgraphPart.containsInterpretation  as
     * this.containsInterpretation() would result in a recursion.
     * 
     * @param interpretation  The interpretation to be checked.
     * 
     * @return interpretedNodes.containsKey( interpretation )
     */
    public boolean interpretedNodesContains(Node interpretation) {
        return interpretedNodes.containsKey(interpretation);
    }


    /**
     * If an interpretation is not used any more it should be removed from list
     * of used interpretations by this function. The mapping from structural
     * constraint graph g to problem graph p must be injective, i.e. no two
     * different nodes from p may have the same interpretation in p.
     * 
     * @param interpretation The interpretation.
     * @param node The node that should get the interpretation.
     */
    public void removeInterpretation(Node interpretation, SCNode node) {
        if (interpretation != null) {
            SCNode nodeWithInterpretation = (SCNode) interpretedNodes.get(
                                                    interpretation);

            if (nodeWithInterpretation == node) {
                if (! resetInterpretations) {
                    interpretedNodes.remove(interpretation);
                }

                intersectingParts.removeIntersection(interpretation);

                if (! useOtherPartNodes) {
                    mySubgraph.removeInterpretation(interpretation, node);
                }
            } else if (nodeWithInterpretation == null) {
                throw new Error("interpretation not part of subgraph!");
            }
        }
    }


    /**
     * No intersection is registered after this call.
     */
    public void resetIntersection() {
        intersectingParts.clear();
        intersectionWithOtherGraphPart = false;
        super.resetIntersection();
    }


    /**
     * Is the given interpretation allready used in this graph? The mapping
     * from structural constraint graph g to problem graph p must be
     * injective, i.e. no two different nodes from g may have the same
     * interpretation in p. Registers an intersection if the tried
     * interpretation was allready used.
     * 
     * @param interpretation The interpretation.
     * @param node The node that should get the interpretation.
     * 
     * @return Does it contain the interpretation?
     */
    public boolean tryInterpretation(Node interpretation, SCNode node) {
        if (settingOriginalInterpretations) { // the tests were made when the original interpretation was set.

            return false;
        }

        SCNode nodeWithInterpretation = (SCNode) interpretedNodes.get(
                                                interpretation);

        if (nodeWithInterpretation != null) {
            if (nodeWithInterpretation == node) {
                return false;
            } else {
                return true;
            }
        }

        if (noUseInterpretations.contains(interpretation)) {
            intersectionWithOtherGraphPart = true;

            return true; // if the interpretation should not be used in the part (set for example in findOneEmbeddingWithoutNodes() )
        } else if (mySubgraph.tryInterpretation(interpretation, node, 
                                                useOtherPartNodes, this)) { // check if an other part allready uses the interpretation, order matters! do this last.
            intersectionWithOtherGraphPart = true;

            return true;
        }

        return false;
    } //boolean tryInterpretation(Node interpretation)


    /**
     * Is one of the given interpretations in the list of the interpretedNodes?
     * 
     * @param interpretations A set of interpretations to be tested.
     * 
     * @return true if one of the interpretations was allready used, else false.
     */
    protected boolean containsOneOfInterpretations(Set interpretations) {
        Iterator it = interpretations.iterator();

        while (it.hasNext()) {
            Node interpretation = (Node) it.next();

            if (interpretedNodes.containsKey(interpretation)) { // if one of the given interpretations is found

                return true;
            }
        }

        return false;
    }


    /**
     * After findOneEmbeddingWithOtherPartNodes() or
     * findNextEmbeddingWithOtherPartNodes() The used interpretations are
     * stored for this part, but not for the whole graph. When all
     * intersections are cleared call this method to register the
     * interpretedNodes with the Subgraph.
     */
    protected void flushUsedInterpretations() {
        Iterator it = interpretedNodes.entrySet().iterator();

        while (it.hasNext()) {
            Map.Entry entry          = (Map.Entry) it.next();
            SCNode    node           = (SCNode) entry.getValue();
            Node      interpretation = (Node) entry.getKey();
            mySubgraph.addInterpretation(interpretation, node);
        } //while
    }


    /**
     * Sets the interpretations as they were before call.
     * 
     * @param interpretations  The interpretations to be set.
     */
    void setOriginalInterpretations(Collection interpretations) {
        settingOriginalInterpretations = true;
        setCopyOfInterpretations(interpretations);
        settingOriginalInterpretations = false;
    }


    /**
     * Finds the next embedding for the part, using also interpretations
     * allready used by other parts
     * 
     * @return true if an embedding was found, else false.
     */
    boolean findNextEmbeddingWithOtherPartNodes() {
        useOtherPartNodes = true;

        boolean result = findNextEmbedding();
        useOtherPartNodes = false;

        return result;
    }


    /**
     * Finds one embedding for the part, using also interpretations allready
     * used by other parts
     * 
     * @return true if an embedding was found, else false.
     */
    boolean findOneEmbeddingWithOtherPartNodes() {
        useOtherPartNodes = true;
        resetIntersection();

        boolean result = findOneEmbedding();
        useOtherPartNodes = false;

        return result;
    } //findOneEmbeddingWithOtherPartNodes()


    /**
     * Tries to find embeddings for all the parts in this intersectingParts and
     * in missingParts, without using the nodes used in  this.interpretedNodes
     * and this.noUseInterpretations.
     * 
     * @return true if embeddings for all the parts were found, else false.
     */
    boolean findOtherEmbeddingForIntersectingParts() {
        boolean foundOthers = true;

        dl.print(3, 
                 "\n" + TAB + "[" + mySubgraph.allParts.indexOf(this) + 
                 "]1 findOther intersecting Parts ");

        HashSet noUseNodes = (HashSet) this.noUseInterpretations.clone();
        noUseNodes.addAll(interpretedNodes.keySet());

        Iterator it = intersectingParts.intersectingPartsIterator();

        while (it.hasNext() && foundOthers) { // move all the parts that were in the way.

            MovableSubgraphPart otherPart = (MovableSubgraphPart) it.next();
            foundOthers = otherPart.findOneEmbeddingWithoutNodes(noUseNodes);
        } //while

        if (foundOthers) {
            dl.print(3, 
                     "\n" + TAB + "[" + mySubgraph.allParts.indexOf(this) + 
                     "]others found, #intersecPts:");
            dl.print(3, intersectingParts.size());
        } else {
            dl.print(3, 
                     "\n" + TAB + "[" + mySubgraph.allParts.indexOf(this) + 
                     "]others not found, #intersecPts:");
            dl.print(3, intersectingParts.size());
        }

        return foundOthers; // if intersectingParts was empty foundOthers is true*/
    } // findOtherEmbeddingForIntersectingParts()


    /**
     * When in the call mySubgraph.tryInterpretation( ) an other part contains
     * the given interpretation this function is called, to register the
     * intersection.
     * 
     * @param useOtherPartNodes Only if true the intesection is stored in
     *        detail. intersectionWithOtherGraphPart is set anyway.
     * @param otherPart The part that was intersecting.
     * @param interpretation The interpretation that was used by this and
     *        otherPart.
     */
    void registerIntersection(MovableSubgraphPart otherPart, 
                              Node interpretation, boolean useOtherPartNodes) {
        if (useOtherPartNodes) { //store the nodes of the intersection
            intersectingParts.addIntersection(otherPart, interpretation);
        }

        intersectionWithOtherGraphPart = true;
    }
}