/*
 * UnconnectedMovableSubgraphPart.java
 *
 * Created on 23. Juni 2002, 22:01
 */
package com.ai_center.dragonbreath.structuralConstraintHandling.graph.movableSubgraph;

import com.ai_center.dragonbreath.constraintgraph.Node;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.*;
import com.ai_center.dragonbreath.utilities.*;

import java.util.*;


/**
 * A GraphPart that is not connected to any other part. Needs a special
 * treatment, as seeds have to be stored to refind the parts.
 * 
 * @version $Revision: 1.4 $
 * @author tim baermann
 */
public class UnconnectedMovableSubgraphPart extends MovableSubgraphPart
    implements UnconnectedGraphPartInterface {
    // Instance variables *****************************************************

    /** 
     * If this object is an instance this is the type part it belongs to. 
     */
    public UnconnectedMovableSubgraphPart myPartType = null;

    /**
     * Not only seeds for found embeddings for the part are stored, but also
     * seeds of not totally found parts. These are used to create the biggest
     * match possible, if the testGraph can not be found as a whole. Here not
     * only the dedicated seed, but any node of the part can serve as the
     * seed. The SCNode used as seed is the key, the interpretations are
     * stored as AlternativeSeedData As there is no
     */
    protected AlternativeSeeds alternativeSeeds = new AlternativeSeeds();

    /**
     * The interpretation actually set last time setNextSeedInterpretation()
     * was called.
     */
    protected Node lastUsedSeedInterpretation = null;

    /**
     * This HashSet contains possible interpretations for the seed node. The
     * seeds are stored, to be able to refind a once found match easily.
     */
    HashSet seedInterpretations = new HashSet();

    /**
     * The iterator to the seedInterpratation that was used the last time
     * findOtherEmbedding was called.
     */
    Iterator lastUsedSeedInterpretationIt = null;

    /**
     * Causes that the search process does not proceed if no new seed
     * interpretation could be found.
     */
    boolean findNewSeeds = false;

    /**
     * Set by setNextAlternativeSeedInterpretation(), if the seed was an
     * alternative one.
     */
    boolean usedAlternativeSeed = false;

    /**
     * When searching for the biggest possible match the function
     * findBiggestEmbedding() not only tries the interpretations for the Seed,
     * but also the nodes and interpretations stored in alternativeSeeds,
     * which might be other nodes then the seed. In this variable the
     * setNextSeedInterpretation() function stores the SCNode that functions
     * next as seed.
     */
    private SCNode actualSeed = null;

    /** 
     * The size of the smallest partly found match 
     */
    private int smallestPartlyFoundMatch = 0;

    // Constructors ***********************************************************

    /**
     * Creates a new instance of UnconnectedACPart
     * 
     * @param ms The subgraph this part belongs to.
     */
    public UnconnectedMovableSubgraphPart(MovableSubgraph ms) {
        super(ms);

        //dl.switchOn(10);// uncomment for debug output
    }

    // Methods ****************************************************************

    /**
     * Sets the seedInterpretation to be the first one available.
     */
    public void setFirstSeedInterpretation() {
        lastUsedSeedInterpretation = null;

        if (myPartType == null) { // see if this is the partType, that stores the seedInterpretations or an instance.
            lastUsedSeedInterpretationIt = seedInterpretations.iterator();
        } else {
            alternativeSeeds             = myPartType.alternativeSeeds;
            lastUsedSeedInterpretationIt = myPartType.seedInterpretations.iterator();
        }

        alternativeSeeds.setFirstSeedInterpretation(nodes.size());
    }


    /**
     * First this function calls setNextSeedInterpretation(), if there are no
     * more seed interpretations the function interates through the nodes in
     * alternativeSeeds.  It then sets the variable actualSeed to be the next
     * SCNode stored in alternativeSeeds.
     * 
     * @param biggestEmbeddingSizeSoFar the seed has to deliver an embedding
     *        that is bigger then this number, else it is not used.
     * 
     * @return true if there are still untested seed interpretations.
     */
    public boolean setNextAlternativeSeedInterpretation(int biggestEmbeddingSizeSoFar) {
        boolean result = false;

        if (setNextSeedInterpretation()) {
            result = true;
        } else {
            if (actualSeed != null) {
                actualSeed.resetInterpretation();
            }

            actualSeed = alternativeSeeds.setNextSeed(biggestEmbeddingSizeSoFar);

            if (actualSeed != null) {
                usedAlternativeSeed = true;
                result              = true;
            } else {
                actualSeed          = seed;
                result              = false;
                usedAlternativeSeed = false;
            }
        }

        return result;
    } // setNextAlternativeSeedInterpretation()


    /**
     * If the AC is not connected to a mainPart there might be several possible
     * seed interpretation. This function steps through them. If the AC is
     * connected to a mainPart the seed interpretation is copied from the
     * seeds linked node.
     * 
     * @return true if there are still untested seed interpretations.
     */
    public boolean setNextSeedInterpretation() {
        while (lastUsedSeedInterpretationIt.hasNext()) {
            if (seed.setInterpretation(
                        (Node) lastUsedSeedInterpretationIt.next())) {
                return true;
            }
        } //while

        if (! lastUsedSeedInterpretationIt.hasNext()) {
            lastUsedSeedInterpretation = null;

            return false;
        }


        // never called  vv
        lastUsedSeedInterpretation = null;

        return false;
    } // setNextSeedInterpretation()


    /**
     * Is the given interpretation allready an seedInterpretations?
     * 
     * @param n The checked node.
     * 
     * @return true if n is in seedInterpretations, else false.
     */
    public boolean isSeedInterpretation(Node n) {
        return seedInterpretations.contains(n);
    }


    /**
     * Returns the seedInterpretations of the part.
     * 
     * @return The seedInterpretations.
     */
    public HashSet getSeedInterpretations() {
        return seedInterpretations;
    }


    /**
     * Produces a new node that can be connected via connectNodes to others in
     * this graph Also adds the node to the searched ones.
     * 
     * @param id the nodeID as defined in Registry
     * @param isLeafNode Is the created node a leafNode? If true it is added to
     *        the searchedNodes, else not.
     * 
     * @return the newly created SCNode
     */
    public SCNode addNode(int id, boolean isLeafNode) {
        SCNode result = super.addNode(id, isLeafNode);

        if (! isLeafNode) {
            searchedNodes.addNode(result);
        }

        return result;
    }


    /**
     * When with the actual seedInterpretaion no embedding could be found the
     * interpretation is invalid, so delete it with this function. Deletes the
     * actual element of lastUsedSeedInterpretationIt.
     */
    public void deleteActualSeedInterpretation() {
        if (lastUsedSeedInterpretationIt != null) {
            lastUsedSeedInterpretationIt.remove();
        }
    }


    /**
     * Searches for a whole match, if  Uses also the Seeds in alternativeSeeds
     * that do not result in whole matches.
     * 
     * @return true if a whole embedding could be found, false else.
     */
    public boolean findBiggestEmbedding() {
        boolean found = false;
        setFirstSeedInterpretation();

        if (setNextAlternativeSeedInterpretation(0)) {
            Object     firstSeedInterpretation   = actualSeed.getInterpretation();
            Object     firstUsedSeed             = actualSeed;
            Collection biggestEmbeddingFound     = null;
            int        biggestEmbeddingFoundSize = 0;

            while (! found) {
                found = findBiggestEmbedding(actualSeed);

                int actualMatchSize = getActualMatchSize();

                if (actualMatchSize > biggestEmbeddingFoundSize) {
                    biggestEmbeddingFound     = getCopyOfInterpretations();
                    biggestEmbeddingFoundSize = actualMatchSize;
                }

                alternativeSeeds.actualizeAlternativeSeed(actualSeed, 
                                                          actualSeed.peekInterpretation(), 
                                                          actualMatchSize);

                if (! found) {
                    if (! intersectionWithOtherGraphPart && 
                            (actualSeed != seed) && ! usedAlternativeSeed) { // if no intersection took place the embedding could not be found at the

                        // given seed interpretation. So it is invalid.
                        deleteActualSeedInterpretation();
                    }


                    // this is allready the biggest embedding, no findNextBiggest() or so necessary
                    resetInterpretations();

                    if (! setNextAlternativeSeedInterpretation(
                                  biggestEmbeddingFoundSize)) {
                        setCopyOfInterpretations(biggestEmbeddingFound);

                        return false; // there was no embedding found, and there is no other seed interpretation
                    } else if ((firstUsedSeed == actualSeed) && 
                                   (actualSeed.peekInterpretation() == firstSeedInterpretation)) {
                        return false; // we circled once through all seed interpretations, and no embedding was found
                    } // if ! setNextSeedInterpretation else
                } // if !found
            } //while
        } //if setNextSeedInterpretation

        return found;
    } // findBiggestEmbedding()


    /**
     * Calls first setNextSeedInterpretation() and then findOneEmbedding(). If
     * a seed interpretation didnt result in a match and that was not due to
     * an intersection it is deleted.
     * 
     * @return true if an embedding was found, else false.
     */
    public boolean findEmbeddingWithNextSeedInterpretation() {
        boolean found = false;

        if (setNextSeedInterpretation()) {
            Object firstSeedInterpretation = seed.getInterpretation();

            while (! found) {
                found = findOneEmbedding(seed);

                if (! found) {
                    if (! intersectionWithOtherGraphPart) { // if no intersection took place the embedding could not be found at the

                        // given seed interpretation. So it is invalid.
                        deleteActualSeedInterpretation();
                    }

                    if (! setNextSeedInterpretation()) {
                        return false; // there was no embedding found, and there is no other seed interpretation
                    } else {
                        if (seed.peekInterpretation() == firstSeedInterpretation) {
                            seed.resetInterpretation();

                            return false; // we circled once through all seed interpretations, and no embedding was found
                        } // if seed.interpretations
                    } //else
                } // if !found
            } //while
        } //if setNextSeedInterpretation

        return found;
    } // findOneEmbeddingWithSeedInterpretations()


    /**
     * Called before findNextEmbedding(). Causes that the search process does
     * not proceed if no new seed interpretation could be found. Speeds up a
     * little bit the search.
     * 
     * @param on  Switch on or off.
     */
    public void findNewSeedsOn(boolean on) {
        findNewSeeds = on;
    }


    /**
     * Is the current search searching for new seeds?
     * 
     * @return findNewSeeds
     */
    public boolean findNewSeedsOn() {
        return findNewSeeds;
    }


    /**
     * Calls super findNextEmbedding( searchSeed ) But beforehands switches
     * findNewSeeds on. That switch saves quite some computation time for
     * symetrical graphParts.
     * 
     * @return true if a new Embedding was found, else false.
     */
    public boolean findNextEmbedding() {
        boolean result;
        findNewSeedsOn(true);
        result = super.findNextEmbedding();
        findNewSeedsOn(false);

        return result;
    }


    /**
     * Finds next valid embedding for the part using next seedInterpretation if
     * with current seed no more embeddings can be found. Uses also
     * interpretations allready used by other parts.
     * 
     * @return true if an embedding was found, else false.
     */
    public boolean findNextEmbeddingWithSeedInterpretations() {
        if (! super.findNextEmbedding()) { // with the current seed no new embedding could be found,try new seed

            return findEmbeddingWithNextSeedInterpretation();
        } // if ! super.super.findNextEmbedding

        return true;
    } // findNextEmbeddingWithSeedInterpretations()


    /**
     * finds one embedding trying all the seed interpretations.
     * 
     * @return found.
     */
    public boolean findOneEmbeddingWithSeedInterpretations() {
        setFirstSeedInterpretation();

        return findEmbeddingWithNextSeedInterpretation();
    }


    /**
     * Same as findOneEmbedding(), in contrast to
     * findOneEmbeddingWithOtherPartNodes() this function doesnt use nodes
     * allready used in other parts. Calls
     * findOneEmbeddingWithSeedInterpretations()
     * 
     * @return true if an embedding was found, else false.
     */
    public boolean findOneEmbeddingWithoutOtherPartNodes() {
        useOtherPartNodes = false;

        return findOneEmbeddingWithSeedInterpretations();
    }


    /**
     * Returns the number of real seed Interpretations.
     * 
     * @return seedInterpretations.size()
     */
    public int nrSeedInterpretations() {
        return seedInterpretations.size();
    }


    /**
     * Unconnected AC parts store all the possible interpretations for their
     * seed, that resulted in a complete match. So when searching for an
     * embedding for all AC parts the part can easily find a valid embedding.
     */
    public void registerSeedInterpretation() {
        Node actualSeedInterpretation = seed.getInterpretation();

        if (actualSeedInterpretation != null) {
            registerSeedInterpretation(actualSeedInterpretation);
            alternativeSeeds.addAlternativeSeed(seed, actualSeedInterpretation, 
                                                getActualMatchSize());
        }
    } //registerSeedInterpretation()


    /**
     * Registeres the given interpretation as seedInterpretation. Does not
     * check wheter seedInterpretation is null
     * 
     * @param seedInterpretation  The new seedInterpretation.
     */
    public void registerSeedInterpretation(Node seedInterpretation) {
        seedInterpretations.add(seedInterpretation);
    }


    /**
     * Remove an interpretation from the list of the possible interpretations.
     * 
     * @param n A node that should no longer be an seed interpretation.
     */
    public void removeSeedInterpretation(Node n) {
        seedInterpretations.remove(n);
    }


    /**
     * Removes all the seedInterpretations that were deleted.
     * 
     * @param deletedNodes Nodes that were deleted.
     */
    public void testForDeletedSeeds(Collection deletedNodes) {
        seedInterpretations.removeAll(deletedNodes);
        alternativeSeeds.testForDeletedSeeds(deletedNodes);
    }


    /**
     * Finds the next embedding for the AC part, using also interpretations
     * allready used by other parts
     * 
     * @return true if an embedding was found, else false.
     */
    boolean findNextEmbeddingWithOtherPartNodes() {
        useOtherPartNodes = true;

        boolean result = findNextEmbeddingWithSeedInterpretations();
        useOtherPartNodes = false;

        return result;
    }


    /**
     * Finds one embedding for the AC part, using also interpretations allready
     * used by other parts
     * 
     * @return true if an embedding was found, else false.
     */
    boolean findOneEmbeddingWithOtherPartNodes() {
        useOtherPartNodes = true;

        boolean result = findOneEmbeddingWithSeedInterpretations();
        useOtherPartNodes = false;

        return result;
    }
}