/*
 * SCTestGraph.java
 *
 * Created on 27. Maerz 2002, 17:21
 */
package com.ai_center.dragonbreath.structuralConstraintHandling.graph.testGraph;

import com.ai_center.dragonbreath.constraintgraph.Node;
import com.ai_center.dragonbreath.structuralConstraintHandling.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.applicationCondition.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.dockingGraph.SCTDockingGraph;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.movableSubgraph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.testGraph.applicationCondition.*;
import com.ai_center.dragonbreath.utilities.SetHashtable;

import java.math.BigInteger;

import java.util.*;


/**
 * The testGraph of a Structural Constraint.
 * 
 * @version $Revision: 1.5 $
 * @author tim baermann
 */
public class TestGraph extends StructuralConstraintGraph {
    // Instance variables *****************************************************

    /**
     * When this testGraph should be reproduced for an instance this is the key
     * to the testGraphs data.
     */
    Object keyToFactoryData;

    /**
     * Is this TestGraph the TestGraph of a StructuralConstraintType or of  an
     * Instance?
     */
    boolean isType;

    /**
     * Is the testGraph valid? I.e. all main parts exist and no NAC exists.  Is
     * set by checkChangedNodes and retrieved by valid();
     */
    boolean valid = true;

    /** 
     * If a NAC was found its size is added to the nacError. 
     */
    int nacError = 0;

    /**
     * For the TestGraph only one match has to be found, so all the
     * mainPart-Embeddings can be shifted around  in order to find a match.
     */
    private TestGraphMainParts allMainParts = new TestGraphMainParts(this);

    // Constructors ***********************************************************

    /**
     * Creates a new instance of StructuralConstraintTestGraph
     * 
     * @param key  When this testGraph should be reproduced for an instance
     *        this is the key to the testGraphs data.
     * @param type  If the TestGraph should be for a SC type(true) or an
     *        instance.
     */
    public TestGraph(Object key, boolean type) {
        keyToFactoryData = key;
        isType           = type;
    }

    // Methods ****************************************************************

    /**
     * Finds the biggest embedding of the testgraph. Call if it is known that
     * there is no valid embedding, but the testgraph is not interpreted. (in
     * StructuralConstraintInstance.improveCosts()) Also tests the NACs.
     * 
     * @return true if nevertheless a complete embedding could be found; false
     *         else.
     */
    public boolean setBiggestEmbedding() {
        if (NACs.size() == 0) {
            // faster search method, bud can find only one embedding. So it can not be used with NACs.
            // also it doesnt find the best match, but only any match
            return allMainParts.findAllParts();
        } else {
            resetNACErrors();

            boolean result = allMainParts.secureFindAllParts();

            if (result) {
                if (! checkNACs()) {
                    computeNACError();
                } else {
                    return true;
                }
            }
        }

        return false;
    }


    /**
     * Returns all the nodes of the Testgraph, that have allready an
     * interpretation. The result can be changed, it is not connected with the
     * Collections of the Graphs.
     * 
     * @return The interpretations as keys to the nodes that are interpreted.
     */
    public Map getInterpretedNodes() {
        return allMainParts.getInterpretedNodes();
    }


    /**
     * Returns the mainParts Variable.
     * 
     * @return mainParts.
     */
    public MovableSubgraph getMainParts() {
        return allMainParts;
    }


    /**
     * Returns the sum of the Errors of all its subgraphs. This number is not
     * deterministic, dependent on the search process!  But the size is more
     * or less a acurate guess of the Structural Error.  And it is never 0 if
     * there is an Error.
     * 
     * @return The Structural Error for this TestGraph.
     */
    public int getMinMatchError() {
        if (nacError != 0) {
            return nacError;
        } else {
            return allMainParts.getMinMatchError();
        }
    }


    /**
     * Returns nacError.  Make shure that beforehands the getMinMatchError()
     * was called, where the nacError is computed.
     * 
     * @return The structural error that is caused by found NACs.
     */
    public int getNACError() {
        return nacError;
    }


    /**
     * Register the StructuralConstraintManagement;
     * 
     * @param scm  The new StructuralConstraintManagement.
     */
    public void setSCManagement(StructuralConstraintManagement scm) {
        mySCM = scm;
    }


    /**
     * Returns isType.
     * 
     * @return isType
     */
    public boolean isType() {
        return isType;
    }


    /**
     * Returns all the MainParts of this TestGraph that are not anchored to the
     * DockingGraph, and all the NACs that are not connected at all.
     * 
     * @return A Collection of StructuralConstraintGraphParts.
     */
    public Collection getUnanchoredParts() {
        Vector   unanchoredParts = new Vector();

        Iterator it = allMainParts.getAllParts().iterator();

        while (it.hasNext()) {
            StructuralConstraintGraphPart part = (StructuralConstraintGraphPart) it.next();

            if (! part.hasInterGraphLinks()) {
                unanchoredParts.add(part);
            }
        } //while

        it = NACs.iterator();

        while (it.hasNext()) {
            ApplicationCondition nac = (ApplicationCondition) it.next();
            unanchoredParts.addAll(nac.getUnconnectedParts());
        }

        return unanchoredParts;
    }


    /**
     * Returns all the nodes of the Structural Constraint that are not yet
     * interpreted.
     * 
     * @return Nodes of MainParts that have no interpretation set.
     */
    public Vector getUninterpretedNodes() {
        return allMainParts.getUninterpretedNodes();
    }


    /**
     * Is the TestGraph valid? I.e. all main parts exist and no NAC exists.
     * called after checkChangedNodes() to retrieve the result.
     * 
     * @return Is the Embedding of this TestGraph still valid?
     */
    public boolean isValid() {
        return this.valid;
    } //valid()


    /**
     * Adds one negative ApplicationCondition to the Graph.
     * 
     * @param nac A Negative Application Condition.
     */
    public void addNAC(ApplicationCondition nac) {
        NACs.add(nac);
    }


    /**
     * Called for instances of TestGraphs.
     * 
     * @param newNodes Nodes that were recently created.
     * @param deletedNodes Nodes that were deleted.
     * @param expandedNodes Expanded Nodes
     * @param reducedNodes The reduced Node.
     */
    public void checkChangedNodes(Set newNodes, Set deletedNodes, 
                                  Set expandedNodes, Set reducedNodes) {
        if (checkMainParts(deletedNodes, reducedNodes)) { // only if the main parts were not destroyed the rest is necessary.
            checkNACs(newNodes, deletedNodes, expandedNodes, reducedNodes);
        }
    } //checkChangedNodes()


    /**
     * Searches for an Embedding for the NAC, and if found sets this.valid =
     * false Called by NACPart if instance was found.
     * 
     * @param nac  The nac that has a newly found part.
     */
    public void checkNAC(ApplicationCondition nac) {
        if (! nac.testAC()) {
            this.valid = false;
        }
    }


    /**
     * Copies the interpretations of the other graph to this one. The two
     * graphs must be exactly identical in structure.
     * 
     * @param srcG the graph with the interpretations
     * 
     * @throws Error If the graphs are not compatible.
     */
    public void copyInterpretations(StructuralConstraintGraph srcG) {
        if (srcG instanceof TestGraph) {
            allMainParts.copyInterpretations(((TestGraph) srcG).allMainParts);

            //copy seed interpretations of nacs
            Iterator srcIt = srcG.NACs.iterator();
            Iterator dstIt = this.NACs.iterator();

            while (srcIt.hasNext()) {
                ApplicationCondition src = (ApplicationCondition) srcIt.next();
                ApplicationCondition dst = (ApplicationCondition) dstIt.next();
                dst.copyInterpretations(src);
            }
        } else {
            throw new Error(
                    "Tried to copy interpretations between incompatible graph.");
        }
    }


    /**
     * Not used.
     * 
     * @return false
     */
    public boolean findOneEmbedding() {
        return false;
    }


    /**
     * Is the given interpretation allready used in one of the MainParts of the
     * Graph?
     * 
     * @param interpretation The searched interpretation.
     * 
     * @return true if the interpretation was allready used, else false.
     */
    public boolean mainPartsContainInterpretation(Node interpretation) {
        return allMainParts.containsInterpretation(interpretation);
    }


    /**
     * Resets the interpretations of all Main subgraphs
     */
    public void resetInterpretations() {
        allMainParts.resetInterpretations();
    }


    /**
     * Tries to find embedding for the whole testGraph. Including the NACs.
     * Sets the Structural Errors of the testGraph. If the testGraph can not
     * be embedded afterwards getMinMatchError() returns the number of missing
     * edges and nodes. After a call to this method resetInterpretations have
     * to be called before using the testgraph again.
     * 
     * @return true if an embedding was found, else false.
     */
    public boolean testTestGraph() {
        boolean result = true;
        nacError = 0;

        if (NACs.size() == 0) {
            // faster search method, bud can find only one embedding. So it can not be used with NACs.
            // also it doesnt find the best match, but only any match
            result = allMainParts.findAllParts();
        } else {
            resetNACErrors();
            result = allMainParts.secureFindAllParts();

            while (result) {
                if (! checkNACs()) {
                    computeNACError();
                    resetNACs();
                    result = allMainParts.nextSecureFindAllParts();
                } else {
                    nacError = 0;

                    return true;
                }
            }
        }

        return result;
    }


    /**
     * Computes the number of NACs currently fullfilled. Stores the result in
     * nacError.
     */
    protected void computeNACError() {
        nacError = 0;

        Iterator it = NACs.iterator();

        while (it.hasNext()) {
            ApplicationCondition nac = (ApplicationCondition) it.next();
            nacError += nac.getMinMatchError();
        }
    }


    /**
     * Checks whether the main parts still exist.
     * 
     * @param reducedNodes  The reduced Nodes.
     * @param deletedNodes  The nodes deleted.
     * 
     * @return true if a mainPart was destroyed.
     */
    boolean checkMainParts(Set deletedNodes, Set reducedNodes) {
        Iterator it = deletedNodes.iterator();

        while (it.hasNext() && this.valid) {
            Node n = (Node) it.next();

            if (allMainParts.containsInterpretation(n)) {
                this.valid = false;
            }
        } //while it


        it = reducedNodes.iterator();

        while (it.hasNext() && this.valid) {
            Node   n   = (Node) it.next();
            SCNode scn = allMainParts.getNodeWithInterpretation(n);

            if (scn != null) {
                if (! scn.testInterpretation()) {
                    this.valid = false;
                } // if scn testInterpretation
            } // if scn != null
        } //while it

        return this.valid;
    }


    /**
     * Are all NACs valid?
     * 
     * @return False if at least one NAC was invalid, true else.
     */
    boolean checkNACs() {
        boolean  result = true;
        Iterator it = NACs.iterator();

        while (it.hasNext() && result) {
            ApplicationCondition nac = (ApplicationCondition) it.next();
            result = nac.testAC();
        }

        return result;
    }


    /**
     * Checks whether a new NAC could be created.
     * 
     * @param newNodes  Nodes that were recently created.
     * @param deletedNodes  The nodes deleted.
     * @param expandedNodes  Expanded Nodes.
     * @param reducedNodes  The reduced Nodes.
     */
    void checkNACs(Set newNodes, Set deletedNodes, Set expandedNodes, 
                   Set reducedNodes) {
        if ((newNodes.size() > 0) || (expandedNodes.size() > 0)) {
            Iterator nacIt = this.NACs.iterator();

            while (nacIt.hasNext() && this.valid) {
                ApplicationCondition nac = (ApplicationCondition) nacIt.next();
                nac.checkChangedNodes(newNodes, deletedNodes, expandedNodes, 
                                      reducedNodes);
            } // while nacIt
        }
    } // checkNACs()


    /**
     * Resets the stored Error of the NACs.
     */
    void resetNACErrors() {
        Iterator it = NACs.iterator();

        while (it.hasNext()) {
            TestNAC nac = (TestNAC) it.next();
            nac.resetNACError();
        }
    }


    /**
     * Resets the interpretations of all NACs
     */
    void resetNACs() {
        Iterator it = NACs.iterator();

        while (it.hasNext()) {
            TestNAC nac = (TestNAC) it.next();
            nac.resetInterpretations();
            nac.resetNACError();
        }
    }
} //class
