/*
 * TestGraphMainParts.java
 *
 * Created on 7. November 2002, 15:07
 */
package com.ai_center.dragonbreath.structuralConstraintHandling.graph.testGraph;

import com.ai_center.dragonbreath.constraintgraph.Node;
import com.ai_center.dragonbreath.structuralConstraintHandling.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.applicationCondition.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.movableSubgraph.*;
import com.ai_center.dragonbreath.utilities.SetHashtable;

import java.util.*;


/**
 * All the MainParts of one TestGraph.
 * @author tim baermann
 * @version $Revision: 1.5 $
 */
public class TestGraphMainParts extends MovableSubgraph {
    // Instance variables *****************************************************

    /**
     * The same as findAllParts() but slower.  Use this function if
     * findNextEmbeddingAllParts should also be used, as it guarantees that
     * all solutions are found.
     */
    int cnt = 0;

    // Constructors ***********************************************************

    /**
     * Creates a new instance of TestGraphMainParts
     * 
     * @param scg The Graph the movable part belongs to.
     */
    public TestGraphMainParts(StructuralConstraintGraph scg) {
        super(scg);

        //dl.switchOn(4);// uncomment for debug output
    }

    // Methods ****************************************************************

    /**
     * Returns all the nodes of the Structural Constraint that are not yet
     * interpreted. The result is a copy. It can be freely modified.
     */
    public Vector getUninterpretedNodes() {
        Vector result = new Vector(nodes);
        result.removeAll(interpretedNodes.values());

        return result;
    }


    /**
     * Embedds all parts of the mainParts. The Parts keep their Interpretations
     * 
     * @return true if all parts could be found.
     */
    public boolean findAllParts() {
        boolean found = true;
        firstMissingPart = null;

        Iterator it = allParts.iterator();

        while (it.hasNext()) { // while all parts are found

            MovableSubgraphPart part = (MovableSubgraphPart) it.next();
            found &= part.findOneEmbeddingAllParts();

            if (! found && (firstMissingPart == null)) {
                firstMissingPart = part;
            }
        }

        if (! found) {
            findBiggestEmbeddingAllMissingParts();
        }

        return found;
    }


    /**
     * Finds the biggest embedding for all parts, that do not have an
     * intepretation yet.
     */
    public void findBiggestEmbeddingAllMissingParts() {
        // assign interpretations to all parts that could not be found.
        Iterator it = allParts.iterator();

        while (it.hasNext()) {
            MovableSubgraphPart part = (MovableSubgraphPart) it.next();

            if (! part.isInterpreted()) { // if a part is uninterpreted by now it could only partly be found. so try again, but be contend with a partiall match
                part.findBiggestEmbedding();
            }
        }
    }


    /**
     * Finds the next valid embedding.
     * 
     * @return An embedding could be found.
     */
    public boolean nextSecureFindAllParts() {
        cnt++;
        dl.print(4, cnt + " findNextEmbeddingAllParts\n");

        return findNextEmbeddingForPart(allParts.size() - 1);
    }


    public boolean secureFindAllParts() {
        cnt = 0;
        dl.print(4, "findOneEmbeddingAllPart\n");

        boolean result = findOneEmbeddingForPart(0);

        if (! result) {
            findBiggestEmbeddingAllMissingParts();
        }

        return result;
    }


    /**
     * Recursively travels the parts to create a new embedding.
     * 
     * @param index The numer of the part in allParts.
     * 
     * @return Was an embedding found?
     */
    private boolean findNextEmbeddingForPart(int index) {
        String tabOld = TAB;
        TAB = TAB.concat("  ");

        boolean result = false;

        if (index > -1) {
            MovableSubgraphPart p = (MovableSubgraphPart) allParts.elementAt(
                                            index);
            dl.print(4, TAB + "findNextEmbeddingForPart:" + index);

            if (p instanceof UnconnectedMovableSubgraphPart) {
                dl.print(4, 
                         " numSeeds: " + 
                         ((UnconnectedMovableSubgraphPart) p).nrSeedInterpretations() + 
                         "\n");
            } else {
                dl.print(4, "\n");
            }

            while (! result && 
                   (((p instanceof UnconnectedMovableSubgraphPart) && ((UnconnectedMovableSubgraphPart) p).findNextEmbeddingWithSeedInterpretations()) || 
                       (! (p instanceof UnconnectedMovableSubgraphPart) && 
                           p.findNextEmbedding()))) {
                result = findOneEmbeddingForPart(index + 1);
            }

            if (! result) {
                result = findNextEmbeddingForPart(index - 1);
            }

            if (result) {
                dl.print(4, TAB + "found ");
                dl.printNodeList(4, p.interpretedNodes);
                dl.print(4, " for part nr:" + index + "\n");
            }
        }

        if (! result) {
            dl.print(4, TAB + "not found\n");
        }

        TAB = tabOld;

        return result;
    }


    /**
     * Recursively finds an embedding for allParts.
     * 
     * @param index  The index of the actually searched part
     * 
     * @return true if an embedding was found, else false.
     */
    private boolean findOneEmbeddingForPart(int index) {
        String tabOld = TAB;
        TAB = TAB.concat("  ");

        if (index == allParts.size()) {
            TAB = tabOld;

            return true;
        }

        MovableSubgraphPart p = (MovableSubgraphPart) allParts.elementAt(index);
        dl.print(4, TAB + "findOneEmbeddingForParts:" + index);

        if (p instanceof UnconnectedMovableSubgraphPart) {
            dl.print(4, 
                     " numSeeds: " + 
                     ((UnconnectedMovableSubgraphPart) p).nrSeedInterpretations() + 
                     "\n");
        } else {
            dl.print(4, "\n");
        }

        if (p.findOneEmbeddingWithoutOtherPartNodes()) {
            if (! findOneEmbeddingForPart(index + 1)) {
                while (p.findNextEmbedding()) {
                    dl.print(4, TAB + "findNextEmbedding:" + index + "\n");

                    if (findOneEmbeddingForPart(index + 1)) {
                        dl.print(4, TAB + "found ");
                        dl.printNodeList(4, p.interpretedNodes);
                        dl.print(4, " for part nr:" + index + "\n");
                        TAB = tabOld;

                        return true;
                    }
                } //while findnext
            } else {
                dl.print(4, TAB + "found ");
                dl.printNodeList(4, p.interpretedNodes);
                dl.print(4, " for part nr:" + index + "\n");
                TAB = tabOld;

                return true;
            }
        } else {
            findOneEmbeddingForPart(index + 1);
        }

        dl.print(4, TAB + "not found\n");
        TAB = tabOld;

        return false;
    } // findOneEmbeddingForPart()
}