/*
 * SCTTestGraph.java
 *
 * Created on 13. September 2002, 14:41
 */
package com.ai_center.dragonbreath.structuralConstraintHandling.graph.testGraph;

import com.ai_center.dragonbreath.structuralConstraintHandling.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.StructuralConstraintManagement;
import com.ai_center.dragonbreath.structuralConstraintHandling.StructuralConstraintType;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.*;
import com.ai_center.dragonbreath.utilities.*;

import java.util.*;


/**
 * A collection of all the testgraphs of a constraint.
 */
public class TestGraphTypesContainer {
    // Instance variables *****************************************************

    /** 
     * For debug output. 
     */
    DebugLogger dl;

    /** 
     * The Factory to produce TestGraph instances. 
     */
    SCGraphFactory myGraphFactory;

    /**
     * When searching for one matching testGraph this Vector is used to keep
     * track of the degree in which the testgraphs could be embedded. Each
     * entry is the number of nodes that matched, and the TestGraph.
     */
    SortedMap matchSizes = new TreeMap();

    /** 
     * All the TestGraphs 
     */
    Vector testGraphs = new Vector();

    /**
     * All Main Parts of all testGraphs that do not have a link to a mainPart,
     * or are NACs that do not have a link at all.
     */
    Vector unanchoredTestGraphParts = new Vector();

    // Constructors ***********************************************************

    /**
     * Creates a new instance of SCTTestGraph
     * 
     * @param graphFactory  The Factory that constructs new instances of the
     *        TestGraphs in this Container.
     */
    public TestGraphTypesContainer(SCGraphFactory graphFactory) {
        myGraphFactory = graphFactory;
        dl             = new DebugLogger(true);

        //dl.switchOn(1);// uncomment for debug output
    }

    // Methods ****************************************************************

    /**
     * If getOneFitting() returned null this function returns all the
     * testGraphs in the order of fitting. The best fitting first, the least
     * fitting last.
     * 
     * @return All the testGraphs associated with their error.
     */
    public SortedMap getInvalidTestGraphs() {
        return new TreeMap(matchSizes);
    }


    /**
     * Sets the interpretations given in dockingPartFrames to the docking
     * graph, and tries afterwards to find a good embedding.
     * 
     * @param dockingPartFrames The interpretation of the DockingGraph.
     * 
     * @return A new testGraphInstance, with a good interpretation, or null if
     *         none found.
     */
    public TestGraph getOneFittingTestGraph(PartFrameContainer dockingPartFrames) {
        TestGraph actualTG = null;
        matchSizes.clear();
        dockingPartFrames.setAllFrames();
        permuteTestGraphs();

        Iterator it = testGraphs.iterator();

        while (it.hasNext()) {
            actualTG = (TestGraph) it.next();

            if (actualTG.testTestGraph()) {
                TestGraph tg = getTestGraphInstance(actualTG);
                dockingPartFrames.resetAllFrames();
                actualTG.resetInterpretations();

                return tg;
            } else {
                int costs = actualTG.getMinMatchError();
                actualTG.resetInterpretations();

                String n = actualTG.toString();
                n = n.substring(50);
                dl.print(1, "\n" + n + "\n   Costs " + costs);
                matchSizes.put(new IntValue(costs), actualTG);
            }
        }

        dockingPartFrames.resetAllFrames();

        return null;
    }


    /**
     * Register the StructuralConstraintManagement at all Testgraphs in the
     * container
     * 
     * @param scm  The new StructuralConstraintManagement.
     */
    public void setSCManagement(StructuralConstraintManagement scm) {
        Iterator it = testGraphs.iterator();

        while (it.hasNext()) {
            TestGraph actualTG = (TestGraph) it.next();
            actualTG.setSCManagement(scm);
        }
    }


    /**
     * If getOneFitting() returned null this function returns the costs,
     * calculated with the best fitting testGraph. Overwrite this function if
     * you want to have a special structuralcost calculation.
     * 
     * @return The costs.
     */
    public int getStructuralCosts() {
        if (matchSizes.size() > 0) {
            return ((IntValue) matchSizes.firstKey()).value;
        } else {
            return 0;
        }
    }


    /**
     * Produces a new testGraphInstance of the given type, and copies the
     * interpretations of the type into the new instance. Resets the
     * interpretations of the key.
     * 
     * @param key  The testGraphType, with valid interpretations.
     * 
     * @return The new TestGraph instance.
     */
    public TestGraph getTestGraphInstance(TestGraph key) {
        TestGraph tg = myGraphFactory.getTestGraphForKey(key.keyToFactoryData);
        tg.copyInterpretations(key);
        key.resetInterpretations();

        return tg;
    }


    /**
     * Adds a new TestGraph to the collection.
     * 
     * @param testGraph The graph to be added!
     */
    public void addTestGraph(TestGraph testGraph) {
        testGraphs.add(testGraph);
        unanchoredTestGraphParts.addAll(testGraph.getUnanchoredParts());
    }


    /**
     * Checks all the parts of all the testGraphs that are not linked to any
     * dockingPart.
     * 
     * @param newNodes  Nodes that were recently created.
     * @param deletedNodes   The nodes deleted.
     * @param expandedNodes  Expanded Nodes
     * @param reducedNodes   The reduced Nodes.
     */
    public void checkUnanchoredParts(Set newNodes, Set deletedNodes, 
                                     Set expandedNodes, Set reducedNodes) {
        Iterator it = unanchoredTestGraphParts.iterator();

        while (it.hasNext()) {
            Object               o    = it.next();
            NodeChangedInterface part = (NodeChangedInterface) o;
            part.checkChangedNodes(newNodes, deletedNodes, expandedNodes, 
                                   reducedNodes);
        }
    }


    /**
     * After a call of this function the testGraphs are in a different, random
     * order. And the Vector testGraphs is a different object.
     */
    private void permuteTestGraphs() {
        if (testGraphs.size() == 1) {
            return;
        }

        Vector testGraphsPermutation = new Vector();
        Random random = new Random();
        int    tgSize = testGraphs.size();

        for (int i = 0; i < tgSize; i++) {
            int index = random.nextInt(testGraphs.size());
            testGraphsPermutation.add(i, testGraphs.elementAt(index));
            testGraphs.remove(index);
        }

        testGraphs = testGraphsPermutation;
    }
}