/*
 * TestUnconnectedMainPart.java
 *
 * Created on 16. Oktober 2002, 23:19
 */
package com.ai_center.dragonbreath.structuralConstraintHandling.graph.testGraph;

import com.ai_center.dragonbreath.constraintgraph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.applicationCondition.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.movableSubgraph.*;
import com.ai_center.dragonbreath.utilities.*;

import java.util.*;


/**
 * A mainPart of a TestGraph, that is not linked to no dockingGraph.
 * 
 * @version $Revision: 1.6 $
 * @author tim baermann
 */
public class TestUnconnectedMainPart extends UnconnectedMovableSubgraphPart
    implements NodeChangedInterface, MainPartInterface {
    // Instance variables *****************************************************

    /** 
     * The NACs of this graph 
     */
    protected HashSet NACs = new HashSet();

    /**
     * Nodes that should not be used when searching for an interpretation, used
     * when multiple parts exist, and one part intersects with an other one
     * Used in checkChangedNodes() to garantie that nodes allready checked are
     * not used again in embeddings.
     */
    HashSet allreadyCheckedNewNodes = new HashSet();

    /**
     * While search for new part match in checkNewNodes() this variable is set
     * true when the part is searched, and false when it was found, and the
     * rest of the graph is searched. All the matches with the nodes in
     * allreadyCheckedNewNodes are known to be allready found. So a match with
     * these nodes can not lead to a new match  => stop search by forbidding
     * the interpretation in tryInterpretation().
     */
    boolean searchWithoutChecked = false;

    // Constructors ***********************************************************

    /**
     * Creates a new instance of TestUnconnectedMainPart
     * 
     * @param ms The Subgraph this part belongs to.
     */
    public TestUnconnectedMainPart(MovableSubgraph ms) {
        super(ms);
    }

    // Methods ****************************************************************

    /**
     * Adds one ApplicationCondition to the list of NACs.
     * 
     * @param ac An ApplicationCondition for testGraphs there are only NACs.
     */
    public void addAC(ApplicationCondition ac) {
        //NACs.add(ac);
    }


    /**
     * Checks nodes that were newly changed for creation or destruction of
     * StructuralConstraintInstances.
     * 
     * @param newNodes  Nodes that were recently created.
     * @param expandedNodes  Expanded Nodes.
     * @param reducedNodes  The reduced Nodes.
     * @param deletedNodes  The nodes deleted.
     */
    public void checkChangedNodes(Set newNodes, Set deletedNodes, 
                                  Set expandedNodes, Set reducedNodes) {
        checkNewNodes(newNodes);
        testForDeletedSeeds(deletedNodes);
        checkNewNodes(expandedNodes); 
        allreadyCheckedNewNodes.clear();
    }


    /**
     * Does nothing. Inherited.
     * 
     * @param frame  dummy
     */
    public void checkFrame(GraphPartFrame frame) {
    }


    /**
     * Tries to find all the parts that were improved due to new nodes.
     * 
     * @param newNodes New nodes in the SCGraph.
     */
    protected void checkNewNodes(Set newNodes) {
        sortNodesByFrequency(newNodes);

        // sort the new nodes ascending with the frequency in the searchedNodesContainer, 
        // first the new Nodes which are less frequent in searchedNodes
        // are tested, and later the more frequent ones.
        // thereby the ramification is more at the end of the backtracking, and not at the root.
        Iterator it = newNodes.iterator();

        while (it.hasNext()) { // check all newNodes, whether one of them leads to a new match

            Node     newNode        = (Node) it.next();
            Set      s              = getSearchedNodes(newNode);
            Iterator missingNodesIt = s.iterator();

            while (missingNodesIt.hasNext()) {
                SCNode n = (SCNode) missingNodesIt.next();

                if (n.setInterpretationBiggestEmbedding(newNode)) {
                    if (findOnePartEmbeddingWithoutCheckedNodes(n)) {
                        registerSeedInterpretation();

                        while (findNextPartEmbeddingWithoutCheckedNodes(n)) {
                            registerSeedInterpretation();
                        }
                    } else {
                        registerAlternativeSeed();
                    }
                } // if setInterpretation

                if (interpretedNodes.size() > 0) {
                    resetInterpretations();
                }
            } // while n.missingNodesIt

            allreadyCheckedNewNodes.add(newNode);
        } // while it
    }


    /**
     * Calls part.findBiggestEmbedding(n) but sets the variable
     * searchWithoutChecked = true so that the search doesnt find matches
     * with allready checked nodes.
     * 
     * @param n  The node that should be embedded.
     * 
     * @return true if an embedding was found, else false.
     */
    protected boolean findOnePartEmbeddingWithoutCheckedNodes(SCNode n) {
        boolean             result;
        MovableSubgraphPart part = (MovableSubgraphPart) n.myGraphPart;
        searchWithoutChecked = true; // switch on that the new nodes allready checked can not be part of the interpretations found now.
        result               = part.findBiggestEmbedding(n);
        searchWithoutChecked = false;

        return result;
    }


    /**
     * If there is one ObjectConstraint in the found partial embedding it is
     * registered as alternative Seed.
     */
    protected void registerAlternativeSeed() {
        Iterator it = interpretedNodes.keySet().iterator();

        while (it.hasNext()) {
            Node interpretation = (Node) it.next();

            if (interpretation instanceof ObjectConstraint) {
                alternativeSeeds.addAlternativeSeed(
                        (SCNode) interpretedNodes.get(interpretation), 
                        interpretation, getActualMatchSize());

                return;
            }
        } //while
    }


    /**
     * Calls part.findNextPartEmbedding(n) but sets the variable
     * searchWithoutChecked = true so that the search doesnt find matches
     * with allready checked nodes.
     * 
     * @param n  The node that should be embedded.
     * 
     * @return true if an embedding was found, else false.
     */
    private boolean findNextPartEmbeddingWithoutCheckedNodes(SCNode n) {
        boolean             result;
        MovableSubgraphPart part = (MovableSubgraphPart) n.myGraphPart;
        searchWithoutChecked = true; // switch on that the new nodes allready checked can not be part of the interpretations found now.
        result               = part.findNextEmbedding();
        searchWithoutChecked = false;

        return result;
    }
}