/*
 * TypeTestUnconnectedNACPart.java
 *
 * Created on 6. Oktober 2002, 19:33
 */
package com.ai_center.dragonbreath.structuralConstraintHandling.graph.testGraph.applicationCondition.type;

import com.ai_center.dragonbreath.constraintgraph.Node;
import com.ai_center.dragonbreath.structuralConstraintHandling.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.applicationCondition.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.movableSubgraph.*;
import com.ai_center.dragonbreath.structuralConstraintHandling.graph.testGraph.*;
import com.ai_center.dragonbreath.utilities.*;

import java.util.*;


/**
 * A part of a NAC that is not connected to any MainPart. Therefore it has to
 * be searched all the time to allways know where the instances are.
 * 
 * @version $Revision: 1.5 $
 * @author tim baermann
 */
public class TypeTestUnconnectedNACPart extends UnconnectedACPart
    implements NodeChangedInterface {
    // Instance variables *****************************************************

    /**
     * Nodes that should not be used when searching for an interpretation, used
     * when multiple parts exist, and one part intersects with an other one
     * Used in checkChangedNodes() to garantie that nodes allready checked are
     * not used again in embeddings.
     */
    HashSet allreadyCheckedNewNodes = new HashSet();

    /**
     * While search for new part match in checkNewNodes() this variable is set
     * true when the part is searched, and false when it was found, and the
     * rest of the graph is searched. All the matches with the nodes in
     * allreadyCheckedNewNodes are known to be allready found. So a match with
     * these nodes can not lead to a new match  => stop search by forbidding
     * the interpretation in tryInterpretation().
     */
    boolean searchWithoutChecked = false;

    // Constructors ***********************************************************

    /**
     * Creates a new instance of TypeTestUnconnectedNACPart
     * 
     * @param ac The ApplicationCondition of this part.
     */
    public TypeTestUnconnectedNACPart(ApplicationCondition ac) {
        super(ac);
    }

    // Methods ****************************************************************

    /**
     * Produces a new node that can be connected via connectNodes to others in
     * this graph Also adds the node to the searched ones.
     * 
     * @param id the nodeID as defined in Registry
     * @param isLeafNode Is the created node a leaf node? If false it is added
     *        to the searchedNodes.
     * 
     * @return the newly created SCNode
     */
    public SCNode addNode(int id, boolean isLeafNode) {
        SCNode result = super.addNode(id, isLeafNode);

        if (! isLeafNode) {
            searchedNodes.addNode(result);
        }

        return result;
    }


    /**
     * Checks nodes that were newly changed for creation or destruction of
     * StructuralConstraintInstances.
     * 
     * @param newNodes  Nodes that were recently created.
     * @param deletedNodes  The nodes deleted.
     * @param expandedNodes  Expanded Nodes.
     * @param reducedNodes  The reduced Nodes.
     */
    public void checkChangedNodes(Set newNodes, Set deletedNodes, 
                                  Set expandedNodes, Set reducedNodes) {
        checkNewNodes(newNodes);
        testForDeletedSeeds(deletedNodes);
        checkNewNodes(expandedNodes); 
    }


    /**
     * Called whenever a match was found for the part. Does nothing.
     * 
     * @return false
     */
    public boolean handleMatch() {
        return false;
    }


    /**
     * Tries to find all the parts that were improved due to new nodes.
     * 
     * @param newNodes New nodes in the SCGraph.
     */
    protected void checkNewNodes(Set newNodes) {
        sortNodesByFrequency(newNodes);

        // sort the new nodes ascending with the frequency in the searchedNodesContainer, 
        // first the new Nodes which are less frequent in searchedNodes
        // are tested, and later the more frequent ones.
        // thereby the ramification is more at the end of the backtracking, and not at the root.
        Iterator it = newNodes.iterator();

        while (it.hasNext()) { // check all newNodes, whether one of them leads to a new match

            Node     newNode        = (Node) it.next();
            Set      s              = getSearchedNodes(newNode);
            Iterator missingNodesIt = s.iterator();

            while (missingNodesIt.hasNext()) {
                SCNode n = (SCNode) missingNodesIt.next();

                if (n.setInterpretation(newNode)) {
                    if (findOnePartEmbeddingWithoutCheckedNodes(n)) {
                        registerSeedInterpretation();

                        while (findNextPartEmbeddingWithoutCheckedNodes(n)) {
                            registerSeedInterpretation();
                        }
                    } // if findOnePartEmbeddingWithoutCheckedNodes
                } // if setInterpretation
            } // while n.missingNodesIt

            allreadyCheckedNewNodes.add(newNode);
        } // while it

        allreadyCheckedNewNodes.clear();
    }


    /**
     * Calls part.findOnePartEmbedding(n) but sets the variable
     * searchWithoutChecked = true so that the search doesnt find matches with
     * allready checked nodes.
     * 
     * @param n  The node that should be embedded.
     * 
     * @return true if an embedding was found, else false.
     */
    protected boolean findOnePartEmbeddingWithoutCheckedNodes(SCNode n) {
        boolean             result;
        MovableSubgraphPart part = (MovableSubgraphPart) n.myGraphPart;
        searchWithoutChecked = true; // switch on that the new nodes allready checked can not be part of the interpretations found now.
        result               = part.findOneEmbedding(n);
        searchWithoutChecked = false;

        return result;
    }


    /**
     * Calls part.findNextPartEmbedding(n) but sets the variable
     * searchWithoutChecked = true so that the search doesnt find matches with
     * allready checked nodes.
     * 
     * @param n  The node that should be embedded.
     * 
     * @return true if an embedding was found, else false.
     */
    private boolean findNextPartEmbeddingWithoutCheckedNodes(SCNode n) {
        boolean             result;
        MovableSubgraphPart part = (MovableSubgraphPart) n.myGraphPart;
        searchWithoutChecked = true; // switch on that the new nodes allready checked can not be part of the interpretations found now.
        result               = part.findNextEmbedding();
        searchWithoutChecked = false;

        return result;
    }
}