
package com.ai_center.dragonbreath.utilities;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.EventListenerList;

/** 
 * This list maintains an ordered list of intervals. Time points are 
 * instantaneous events, i.e., beginning + duration = end.
 *
 * @author Alexander Nareyek
 * @version $Revision: 2.3 $
 */


public class IntervalList {

    // **********************
    // ***** ATTRIBUTES *****
    // **********************

    /** 
     * The beginning of the covered interval.
     */
    private long beginning;

    /** 
     * The end of the covered interval.
     */
    private long end;

    /** 
     * The DLL's first element.
     */
    private IntervalListElement first = null;

    /** 
     * The DLL's last element.
     */
    private IntervalListElement last = null;
    
    private EventListenerList listenerList = new EventListenerList();


    // ************************
    // ***** CONSTRUCTORS *****
    // ************************

    /** 
     * Initializes the IntervalList with an interval that described the 
     * maximal coverage of the list.
     * @param beginning  The beginning of the covered interval.
     * @param end  The end of the covered interval.
     * @param element  The element object for the first interval list element.
     */
    public IntervalList(long beginning, long end, Object element) {
	this.beginning = beginning;
	this.end = end;
	first = new IntervalListElement(beginning, end, element);
	last = first;
    }


    // *******************
    // ***** METHODS *****
    // *******************

    /** 
     * Prints a list of all list elements.
     */
    public synchronized String toString() {
	StringBuffer s = new StringBuffer("[ ");
	IntervalListElement listElement = first;
	while (listElement != null) {
	    s.append(listElement);
	    listElement = listElement.getNext();
	    if (listElement != null) {
		s.append(", ");
	    }
	}
	s.append(" ]");
	return s.toString();
    }

    /** 
     * Searches from the beginning of the list for an interval list element 
     * that contains a given point and returns it.
     * @param x  The point for which the interval list element is searched.
     */
    public synchronized IntervalListElement findInterval(long x) {

	if (x < beginning) {
	    throw new Error("The given point \"" + x + "\" is outside the interval list's cover interval [" + beginning + ".." + end + "]");
	}

	// Searching from beginning for an interval that contains x
	
	IntervalListElement listElement = first;
	
	while (x > listElement.getEnd()) {
	    listElement = listElement.getNext();
	    if (listElement == null) {
		throw new Error("The given point \"" + x + "\" is outside the interval list's cover interval [" + beginning + ".." + end + "]");
	    }
	}

	return(listElement);

    }

    /** 
     * Searches, starting from a given beginning interval list element, for 
     * an interval list element that contains a given point and returns it.
     * @param x  The point for which the interval list element is searched.
     * @param listElement  The interval list element to start the search from.
     */
    public synchronized IntervalListElement findIntervalWithOffset(long x,
						      IntervalListElement 
						      listElement) {

	if (x < listElement.getBeginning()) {
	    throw new Error("The given point \"" + x + "\" is before the given list element's interval  [" + listElement.getBeginning() + ".." + listElement.getEnd() + "]");
	}

	// Searching for an interval that contains x
	
	while (x > listElement.getEnd()) {
	    listElement = listElement.getNext();
	    if (listElement == null) {
		throw new Error("The given point \"" + x + "\" is outside the interval list's cover interval [" + beginning + ".." + end + "]");
	    }
	}

	return(listElement);

    }

    /** 
     * Inserts a base point that might split a given interval list element. 
     * Returns the new interval list element if one is created, and the old
     * interval list element otherwise. If a new interval list element is 
     * created, this will be following the old interval list element (whose 
     * end is adapted accordingly).
     * @param x  The point at which there is to be a split.
     * @param listElement  The interval list element to be split.
     */
    public synchronized IntervalListElement addBasePoint(long x, 
					    IntervalListElement listElement) {

	if ((listElement.getBeginning() < x) && (listElement.getEnd() > x)) {

	    // Creating new interval (listElement -> newListElement)
	    
	    IntervalListElement newListElement = 
		new IntervalListElement(x, listElement.getEnd());
	    listElement.setEnd( x );
	    listElement.setDuration(x - listElement.getBeginning());
	    
	    // Adapting list pointers
	    
	    newListElement.setPrevious( listElement );
	    newListElement.setNext(listElement.getNext());
	    listElement.setNext( newListElement );
	    if (newListElement.getNext() == null) {
		last = newListElement;
	    } else {
		newListElement.getNext().setPrevious( newListElement );
	    }
	    
	    fireChange(newListElement);
	    return newListElement;

	} else {

	    // Base point already exists
		fireChange(listElement);
	    return listElement;
	}

    }

    /** 
     * Inserts a base point that might split an existing interval. Returns
     * the new interval list element if one is created, and the existing
     * interval list element in which the point is contained otherwise. 
     * If a new interval list element is created, this will be following 
     * the old interval list element (whose end is adapted accordingly).
     * @param x  The point at which there is to be a split.
     */
    public synchronized IntervalListElement addBasePoint(long x) {
	return addBasePoint(x, findInterval(x));
    }

    /** 
     * Inserts a base point that might split an existing interval that is 
     * equal to or after a given interval element. Returns the new interval 
     * list element if one is created, and the existing interval list element 
     * in which the point is contained otherwise. If a new interval list 
     * element is created, this will be following the old interval list 
     * element (whose end is adapted accordingly).
     * @param x  The point at which there is to be a split.
     * @param listElement  The interval list element to start the search from.
     */
    public synchronized IntervalListElement addBasePointWithOffset(long x, 
						      IntervalListElement 
						      listElement) {
	return addBasePoint(x, findIntervalWithOffset(x, listElement));
    }

    /** 
     * Removes a base point that might split a given interval list element
     * and its successor. The interval list element before the base point is 
     * extended and the following interval list element is removed. Returns 
     * the extended interval list element if successful. 
     * @param x  The point at which there is to be no split.
     * @param listElement  The interval list element to be merged with its
     *                     successor.
     */
    public synchronized IntervalListElement removeBasePoint(long x, IntervalListElement 
					       listElement) {

	if ((x != beginning) && (x != end) && (x == listElement.getEnd())) {

	    // Merging interval with following interval

	    IntervalListElement followingListElement = listElement.getNext();
	    listElement.setEnd( followingListElement.getEnd() );
	    listElement.setDuration( listElement.getEnd() - listElement.getBeginning());
	    listElement.setNext( followingListElement.getNext() );
	    if (followingListElement.getNext() == null) {
		last = listElement;
	    } else {
		followingListElement.getNext().setPrevious( listElement );
	    }
		fireChange(listElement);
	    return listElement;

	} else {

	    // Base point is no base point or part of the interval list's
	    // cover interval
		
		// TODO fireChange() necessary?
	    return null;
	}

    }

    /** 
     * Removes a base point that might split an existing interval. The 
     * interval list element before the base point is extended and the 
     * following interval list element is removed. Returns the extended 
     * interval list element if successful. 
     * @param x  The point at which there is to be no split.
     */
    public synchronized IntervalListElement removeBasePoint(long x) {
	return removeBasePoint(x, findInterval(x));
    }

    /** 
     * Removes a base point that might split an existing interval that is 
     * equal to or after a given interval element. The interval list element 
     * before the base point is extended and the following interval list 
     * element is removed. Returns the extended interval list element if 
     * successful. 
     * @param x  The point at which there is to be no split.
     * @param listElement  The interval list element to start the search from.
     */
    public synchronized IntervalListElement removeBasePointWithOffset(long x,
							 IntervalListElement 
							 listElement) {
	return removeBasePointWithOffset(x, findIntervalWithOffset(x, listElement));
    }

    /**
     * @return
     */
    public synchronized long getBeginning() {
        return beginning;
    }

    /**
     * @return
     */
	public synchronized long getEnd() {
        return end;
    }

    /**
     * @return
     */
	public synchronized IntervalListElement getFirst() {
        return first;
    }

    /**
     * @return
     */
	public synchronized IntervalListElement getLast() {
        return last;
    }

    /**
     * @param l
     */
	public synchronized void setBeginning(long l) {
        beginning = l;
    }

    /**
     * @param l
     */
	public synchronized void setEnd(long l) {
        end = l;
    }

    /**
     * @param element
     */
	public synchronized void setFirst(IntervalListElement element) {
        first = element;
    }

    /**
     * @param element
     */
	public synchronized void setLast(IntervalListElement element) {
        last = element;
    }
    
	public void addChangeListener(ChangeListener l) {
    	 listenerList.add(ChangeListener.class, l);
	 }

	 public void removeChangeListener(ChangeListener l) {
		 listenerList.remove(ChangeListener.class,  l);
	 }

	 /**
	  * This method supports normal ChangeListeners
	  */
	 protected void fireChange(Object changedObject) {

		 
		 ChangeEvent simpleChangeEvent = new ChangeEvent(changedObject);

		 
		 Object[] listeners = listenerList.getListenerList();
		 ChangeListener listener = null;

		 for (int i = listeners.length - 2; i >= 0; i -= 2) {
			 if (listeners[i] == ChangeListener.class) {

				 ((ChangeListener) listeners[i + 1]).stateChanged( simpleChangeEvent );

			 }

		 }
	 }


    /**
     * 
     */
    public void fireIntervalListUpdate() {
       fireChange(this);
        
    }


}
