
package com.ai_center.dragonbreath.utilities;


import com.ai_center.dragonbreath.GlobalSearchControl;


/**
 * This class realizes a simple object storage and selection, e.g., for
 * a random selection with a probability distribution according to the
 * objects' integer values. Objects are stored in a list but NOT ordered
 * according to an integer value.
 *
 * @author Alexander Nareyek
 * @version $Revision: 2.4 $
 */
public class ObjectIntSelection
{


    // **********************
    // ***** ATTRIBUTES *****
    // **********************


    /**
     * The global search control to be used for random number generation.
     */
    private GlobalSearchControl gsc;

    /**
     * The number of objects that were added.
     */
    private int numberOfElements = 0;

    /**
     * The number of currently existing containers. This can be larger than
     * the number of elements because containers will not be deleted and
     * are left for re-use.
     */
    private int numberOfContainers = 0;

    /**
     * The list's first container.
     */
    private ObjectIntContainer first = null;

    /**
     * The first container in the list that is available to hold an
     * element.
     */
    private ObjectIntContainer nextAvailableContainer = null;

    /**
     * The total of all object's integer values.
     */
    private long total = 0;


    // ************************
    // ***** CONSTRUCTORS *****
    // ************************


    /**
     * Initializes the ObjectIntSelection with the gloabl search control
     * object.
     * @param gsc  The global search control to be used for random number
     *     generation.
     */
    public ObjectIntSelection( GlobalSearchControl gsc )
    {
        this.gsc = gsc;
    }


    // *******************
    // ***** METHODS *****
    // *******************


    /**
     * Prints a list of all list elements.
     */
    public String toString()
    {
        StringBuffer s = new StringBuffer( "[ total=" + total + " " );
        if ( numberOfElements > 0 )
        {
            s.append( "| " );
        }

        ObjectIntContainer container = first;
        int i = 0;
        while ( i++ < numberOfElements )
        {
            s.append( container.element.toString() + "("
                + container.intValue + ")" );
            container = container.next;
            if ( i < numberOfElements )
            {
                s.append( ", " );
            }
        }

        if ( numberOfElements > 0 )
        {
            s.append( " ]" );
        }
        else
        {
            s.append( "]" );
        }

       return s.toString();

    }


    /**
     * Removes all elements from the list. The actual container elements
     * will not be deleted.
     */
    public void clear()
    {
        ObjectIntContainer container = first;
        int i = 0;
        while ( i++ < numberOfElements )
        {
            container.element = null;
            container = container.next;
        }
        numberOfElements = 0;
        nextAvailableContainer = first;
        total = 0;
    }


    /**
     * Adds a content element to the selection.
     * @param element  The content element to be added.
     * @param intValue  The element's integer value. This value must be
     *     greater or equal to 0.
     */
    public void addElement( Object element , int intValue )
    {

        if ( intValue >= 0 )
        {

            ObjectIntContainer container;

            if ( numberOfContainers != numberOfElements )
            {
                // Using an existing available container object
                nextAvailableContainer.element = element;
                nextAvailableContainer.intValue = intValue;
                nextAvailableContainer = nextAvailableContainer.next;
            }
            else
            {
                // Adding a new container object
                container = new ObjectIntContainer( element , intValue );
                container.next = first;
                first = container;
                numberOfContainers++ ;
            }

            numberOfElements++ ;
            total += intValue;

        }
        else
        {
            throw new Error( "The integer value \"" + intValue
                + "\" is less than 0" );
        }

    }


    /**
     * Adds a content element to the selection such that its integer value
     * will be considered anti-proportionally, i.e., the higher the value,
     * the lesser the chance for selection. If this type of addition is used,
     * it should be used for all objects of the selection.
     * @param element  The content element to be added.
     * @param intValue  The element's integer value. This value must be
     *     greater or equal to 0.
     */
    public void addElementAntiproportionally( Object element , int intValue )
    {

        if ( intValue >= 0 )
        {

            ObjectIntContainer container;

            double d = 1.0 / ( intValue + 1 );
            intValue = (int)( d * Integer.MAX_VALUE );

            if ( numberOfContainers != numberOfElements )
            {
                // Using an existing available container object
                nextAvailableContainer.element = element;
                nextAvailableContainer.intValue = intValue;
                nextAvailableContainer = nextAvailableContainer.next;
            }
            else
            {
                // Adding a new container object
                container = new ObjectIntContainer( element , intValue );
                container.next = first;
                first = container;
                numberOfContainers++ ;
            }

            numberOfElements++ ;
            total += intValue;

        }
        else
        {
            throw new Error( "The integer value \"" + intValue
                + "\" is less than 0" );
        }

    }


    /**
     * Randomly chooses one object with a probability distribution
     * according to the objects' integer values. The higher the integer
     * value, the more likely it will get chosen.
     * @param element  The content element to be added.
     * @param intValue  The element's integer value.
     * @return  The selected element, or null if no element with a
     *     positive integer value was added.
     */
    public Object chooseElement()
    {

        if ( total != 0 ) {

            ObjectIntContainer container = first;
            long chosenValue = Math.abs( gsc.random.nextLong() ) % total;

            chosenValue -= container.intValue;
            while ( chosenValue >= 0 )
            {
                container = container.next;
                chosenValue -= container.intValue;
            }

            return container.element;

        }
        else {
            return null;
        }

    }


}
