
package com.ai_center.dragonbreath.utilities;

/**
 * This object maintains an ordered list of elements. The order is kept
 * by simple shift operations after each change of an element. The content
 * elements should have a variable which points back to the DLLElement.
 *
 * @author Alexander Nareyek
 * @version $Revision: 2.3 $
 */
public abstract class OrderedShiftDLL
{

    // **********************
    // ***** ATTRIBUTES *****
    // **********************

    /**
     * The DLL's first element.
     */
    public OrderedShiftDLLElement first = null;

    /**
     * The DLL's last element.
     */
    public OrderedShiftDLLElement last = null;

    /**
     * The number of list elements.
     */
    public int numberOfElements = 0;


    // *******************
    // ***** METHODS *****
    // *******************

    /**
     * Prints a list of all list elements.
     */
    public String toString()
    {
        StringBuffer s = new StringBuffer( "[" );
        OrderedShiftDLLElement listElement = first;
        while ( listElement != null )
        {
            s.append( listElement.element.toString() );
            listElement = listElement.next;
            if ( listElement != null )
            {
                s.append( ", " );
            }
        }
        s.append( "]" );
        return s.toString();
    }


    /**
     * Removes all elements from the list.
     * ATTENTION: Don't forget to remove the content elements' links to
     * the DLLElements.
     */
    public void clear()
    {
        first = null;
        last = null;
        numberOfElements = 0;
    }


    /**
     * Inserts a list element into the list.
     * @param listElement  The list element to be inserted.
     * @param listElementNext  A list element before which the list
     *                         element is to be inserted.
     */
    private void insertListElement( OrderedShiftDLLElement listElement ,
        OrderedShiftDLLElement listElementNext )
    {

        if ( listElementNext == null )
        {
            // -X-|
            if ( first == null )
            {
                // |-X-|
                first = listElement;
            }
            else
            {
                // O-X-|
                last.next = listElement;
                listElement.previous = last;
                listElement.next = null;
            }
            last = listElement;
        }
        else
        {
            // -X-O
            if ( first == listElementNext )
            {
                // |-X-O
                first = listElement;
                listElement.previous = null;
            }
            else
            {
                // O-X-O
                listElementNext.previous.next = listElement;
                listElement.previous = listElementNext.previous;
            }
            listElement.next = listElementNext;
            listElementNext.previous = listElement;
        }

    }


    /**
     * Determines an element's position in the list. The list's element
     * before which the element should be inserted is returned.
     * @param element  The content element for which the position is
     *                 to be determined.
     */
    private OrderedShiftDLLElement determinePosition( Object element )
    {

        // Allowing for quick check if it should be the last element

        if ( ( last != null )
            && ( compareElements( last.element , element ) != -1 ) )
        {
            return null;
        }
        else
        {

            // Search from the beginning otherwise

            OrderedShiftDLLElement listElement = first;
            while ( ( listElement != null )
                && ( compareElements( listElement.element , element ) == 1 ) )
            {
                listElement = listElement.next;
            }
            return listElement;
        }
    }


    /**
     * Adds a content element via a new list element to the list.
     * Returns a DLLElement to which the element should establish a pointer.
     * @param element  The content element to be added.
     */
    public OrderedShiftDLLElement addListElement( Object element )
    {
        OrderedShiftDLLElement listElement = determinePosition( element );
        OrderedShiftDLLElement newListElement
            = new OrderedShiftDLLElement( element );
        insertListElement( newListElement , listElement );
        numberOfElements++ ;
        return newListElement;
    }


    /**
     * Unlinks the list element from the list.
     * @param listElement  The list element to be unlinked.
     */
    private void unlinkListElement( OrderedShiftDLLElement listElement )
    {

        if ( listElement.previous != null )
        {
            // O-X-
            if ( listElement.next != null )
            {
                // O-X-O
                listElement.previous.next = listElement.next;
                listElement.next.previous = listElement.previous;
            }
            else
            {
                // O-X-|
                listElement.previous.next = null;
                last = listElement.previous;
            }
        }
        else
        {
            // |-X-
            if ( listElement.next != null )
            {
                // |-X-O
                first = listElement.next;
                listElement.next.previous = null;
            }
            else
            {
                // |-X-|
                first = null;
                last = null;
            }
        }

        listElement.previous = null;
        listElement.next = null;

    }


    /**
     * Removes an element and its list element from the list.
     * ATTENTION: Don't forget to remove the content element's link to
     * the DLLElement.
     * @param listElement  The list element to be removed.
     */
    public void removeListElement( OrderedShiftDLLElement listElement )
    {
        unlinkListElement( listElement );
        listElement.element = null;
        numberOfElements-- ;
    }


    /**
     * Updates a list element's position in the list.
     * @param listElement  The list element to be updated.
     */
    public void updateListElementPosition( OrderedShiftDLLElement listElement )
    {

        // Checking if an update if necessary

        boolean updateNecessaryPrevious = ( listElement.previous != null ) && ( compareElements( listElement.previous.element , listElement.element ) == -1 );
        boolean updateNecessaryNext = ( listElement.next != null ) && ( compareElements( listElement.element , listElement.next.element ) == -1 );

        // Executing the update if necessary

        if ( updateNecessaryPrevious || updateNecessaryNext )
        {
            unlinkListElement( listElement );
            OrderedShiftDLLElement listElementNext
                = determinePosition( listElement.element );
            insertListElement( listElement , listElementNext );
        }
    }


    /**
     * Compares two elements and returns -1 if element2 is to be ordered
     * before element1, 0 if the order does not matter, and 1 if element1
     * if to be ordered before element2.
     * @param element1  The first element to be compared.
     * @param element2  The second element to be compared.
     */
    protected abstract int compareElements( Object element1 ,
        Object element2 );

}
