package com.ai_center.planpresentation;


import java.awt.Graphics;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JComponent;




/** The default swimlane control. Responsible for painting lines over the 
 plan paint area.<p>
 
 Additional to painting connection lines, the swimlane control paints a
 dashed line for the current time and tracks "sort on top" commands,
 when an action is clicked.
 
 @author Daniel Weiss
 @version $Revision: 1.4 $
 */

public class DefaultSwimlaneControl implements SwimlaneControl
{


    /** Mapping from datamodel objects to swimlane connectors. */
    Hashtable connectorDirectory = new Hashtable();

    /** @see SwimlaneControl#getCTPercent */
    private double ctPercent;

    /** The glass pane, where all lines should be painted. */
    private GlassPaneArea glassPane;

    /** The plan paint area, where all views are contained. */
    private JComponent planPaintArea;

    /** The link to the PlanPresentationAPI. */
    private PlanPresentationAPI planPresentationAPI;


    /** The list of all swimlanes */
    List swimlaneList = new LinkedList();

    /** A mapping from owners to swimlanes */
    Hashtable swimlaneOwnerHash = new Hashtable();

    /**  The DefaultSwimlaneControl constructor.
     
     @param ppAPI The link to the PlanPresentationAPI.
     */
    public DefaultSwimlaneControl(PlanPresentationAPI ppAPI)
    {
        planPresentationAPI = ppAPI;
    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneControl#getConnector(java.lang.Object)
     */
    public SwimlaneConnector getConnector(Object o)
    {

        return (SwimlaneConnector) connectorDirectory.get(o);
    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneControl#getCTPercent()
     */
    public double getCTPercent()
    {

        return ctPercent;
    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneControl#getGlassPane()
     */
    public JComponent getGlassPane()
    {

        return glassPane;
    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneControl#getPlanPaintArea()
     */
    public JComponent getPlanPaintArea()
    {
        return planPaintArea;
    }

	/* (non-Javadoc)
	 * @see com.ai_center.planpresentation.SwimlaneControl#minVisibleY()
	 */
	public int minVisibleY() {
	
	    return (int) planPaintArea.getVisibleRect().getMinY();
	}
	
    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneControl#maxVisibleY()
     */
    public int maxVisibleY()
    {

        return (int) planPaintArea.getVisibleRect().getMaxY();
    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneControl#paintSwimlanes(java.awt.Graphics)
     */
    public synchronized void paintSwimlanes(Graphics g)
    {
        Iterator it = swimlaneList.iterator();

        while (it.hasNext())
        {

            Swimlane sw = (Swimlane) it.next();

            sw.paint(g);

        }


    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneControl#paintVerticalLine(int)
     */
    public void paintVerticalLine(int x)
    {

        glassPane.paintVerticalLine(x);

    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneControl#registerConnector(java.lang.Object, com.ai_center.planpresentation.SwimlaneConnector)
     */
    public synchronized void registerConnector(Object object,
            SwimlaneConnector sc)
    {
        if (object == null || sc == null)
        {
            throw new Error(
                    "SwimlaneControl: The object to represent or the gui element is null");
        }
        connectorDirectory.put(object, sc);

    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneControl#removeConnector(java.lang.Object)
     */
    public synchronized void removeConnector(Object object)
    {
        if (object == null)
        {
            throw new Error(
                    "Cannot remove a null object from the connectorDirectory");
        }
        connectorDirectory.remove(object);

    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneControl#removeSwimlane(com.ai_center.planpresentation.SwimlaneConnector)
     */
    public synchronized boolean removeSwimlane(SwimlaneConnector ownerConnector)
    {

        if (swimlaneOwnerHash.containsKey(ownerConnector))
        {
            // Find the swimlane
            Swimlane sw = (Swimlane) swimlaneOwnerHash.get(ownerConnector);

            // remove swimlane from the directory
            swimlaneOwnerHash.remove(ownerConnector);

            // walk through the list and remove the swimlane
            Iterator it = swimlaneList.iterator();
            while (it.hasNext())
            {
                Swimlane sw2 = (Swimlane) it.next();
                if (sw2 == sw)
                {
                    it.remove();
                    break;
                }
            }

        }
        return false;

    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneControl#requestSwimlane(int, com.ai_center.planpresentation.SwimlaneConnector, java.util.List)
     */
    public synchronized boolean requestSwimlane(int connectionType,
            SwimlaneConnector ownerConnector, List objectList)
    {
        List connectorList = objectList;

        if (connectorList.size() >= 2)
        {
            // we can only paint swimlane at least between two swimlaneConnectors.
            Swimlane swimlane = null;
            if (swimlaneOwnerHash.containsKey(ownerConnector))
            {
                swimlane = (Swimlane) swimlaneOwnerHash.get(ownerConnector);
            } else
            {
                // create a new swimlane
                swimlane = new Swimlane(this, ownerConnector);
                swimlaneOwnerHash.put(ownerConnector, swimlane);
                swimlaneList.add(swimlane);
            }

            swimlane.setConnectorList(connectorList);
            swimlane.setConnectionType(connectionType);
            return true;
        } else
        {
            return false;
        }
    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneControl#setCTPercent(double)
     */
    public void setCTPercent(double _ctPercent)
    {
        ctPercent = _ctPercent;

    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneControl#setGlassPane(com.ai_center.planpresentation.PlanPaintArea)
     */
    public void setGlassPane(GlassPaneArea _glassPane)
    {
        glassPane = _glassPane;
    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneControl#setPaintArea(javax.swing.JPanel)
     */
    public void setPaintArea(JComponent _planPaintArea)
    {

        planPaintArea = _planPaintArea;

    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneControl#setPriorityOffset(java.lang.Object, int)
     */
    public void setPriorityOffset(Object resourceConstraint, int offset)
    {
        SwimlaneConnector sc = (SwimlaneConnector) connectorDirectory
                .get(resourceConstraint);
        if (sc != null)
        {
            sc.setPriorityOffset(offset);
        }

    }


    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneControl#setSortMode(int)
     */
    public void setSortMode(int sortmode)
    {
        planPresentationAPI.setSortMode(sortmode);
    }



}