package com.ai_center.planpresentation;


import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;

import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;




/** The GlassPaneArea is a invisible glass pane on top of the JFrame window.<p>
 
 On this glass panes are all lines painted. Like the dashed line for the
 current time and all connection lines (swimlanes) between two and more
 elements.<p>
 
 @author Daniel Weiss
 @version $Revision: 1.2 $
 */
public class GlassPaneArea extends JComponent
        implements
            PlanPresentationConstants
{

    /** Indicates if a vertical line should be painted at the mouse cursor */
    private boolean mouseTrackerOn = false;

    /** The x-coordinate of the last mouse event */
    private int mouseX;

    /** If this counter is zero, the vertical line will be removed. Otherwise
      the vertical line would disappear to fast. */
    private int removeCounter;

    /** The link to the SwimlaneControl. */
    private SwimlaneControl swimlaneControl;

    public GlassPaneArea(SwimlaneControl _sc)
    {
        super();
        swimlaneControl = _sc;

    }


    /** Draws a dashed line.<p>
     	
       @param g The graphics context.
       @param x1 The first x-coordinate.
       @param y1 The first y-coordinate.
       @param x2 The second x-coordinate.
       @param y2 The secondy y-coordinate.
       @param dashlength The length of a single dash.
       @param spacelength The lenght of a single space between to dashes.
     */
    public void drawDashedLine(Graphics g, int x1, int y1, int x2, int y2,
            double dashlength, double spacelength)
    {
        if ((x1 == x2) && (y1 == y2))
        {
            g.drawLine(x1, y1, x2, y2);
            return;
        }
        double linelength = Math.sqrt((x2 - x1) * (x2 - x1) + (y2 - y1)
                * (y2 - y1));
        double yincrement = (y2 - y1)
                / (linelength / (dashlength + spacelength));
        double xincdashspace = (x2 - x1)
                / (linelength / (dashlength + spacelength));
        double yincdashspace = (y2 - y1)
                / (linelength / (dashlength + spacelength));
        double xincdash = (x2 - x1) / (linelength / (dashlength));
        double yincdash = (y2 - y1) / (linelength / (dashlength));
        int counter = 0;
        for (double i = 0; i < linelength - dashlength; i += dashlength
                + spacelength)
        {
            g.drawLine((int) (x1 + xincdashspace * counter),
                    (int) (y1 + yincdashspace * counter), (int) (x1
                            + xincdashspace * counter + xincdash), (int) (y1
                            + yincdashspace * counter + yincdash));
            counter++;
        }
        if ((dashlength + spacelength) * counter <= linelength)
            g.drawLine((int) (x1 + xincdashspace * counter),
                    (int) (y1 + yincdashspace * counter), x2, y2);
    }


    /** The paint function for the glass pane 
     
      @param g The graphics context.
     */
    public void paint(Graphics g)
    {
        // first paint the background (nothing to do)
        super.paint(g);

        // paint all swimlanes
        swimlaneControl.paintSwimlanes(g);
        
        
        JPanel planPaintArea = (JPanel) swimlaneControl.getPlanPaintArea();
        
        // compute the x-coordinate, where the current time line should be
        // painted
        double ctPercent = swimlaneControl.getCTPercent();
        int x = (int) (ctPercent * (planPaintArea.getWidth()))
                + RETRACT_BUTTON_WIDTH;
        
        // clip the dashes outside of the plan paint area away.
        Rectangle visibleRect = planPaintArea.getVisibleRect();
        if (x > RETRACT_BUTTON_WIDTH)
        {
            // the first point on the  top
            Point p1 = SwingUtilities.convertPoint(planPaintArea, x,
                    visibleRect.y, this);

            // the second point on the bottom
            Point p2 = SwingUtilities.convertPoint(planPaintArea, x,
                    visibleRect.y + visibleRect.height, this);
            drawDashedLine(g, p1.x, p1.y, p2.x, p2.y, 2.0, 4.0);
        }

        // check if a vertical line for the mousetracker should be painted
        if (mouseX != -1 && mouseX > RETRACT_BUTTON_WIDTH)
        {
            Point p1 = SwingUtilities.convertPoint(planPaintArea, mouseX,
                    visibleRect.y, this);

            Point p2 = SwingUtilities.convertPoint(planPaintArea, mouseX,
                    visibleRect.y + visibleRect.height, this);


            g.drawLine(p1.x, p1.y, p2.x, p2.y);
            
            // decrement the counter. If it is zero, the line will disappear.
            removeCounter--;
            if (removeCounter == 0)
            {
                mouseX = -1; // make the line disappear
            }
        }
    }


    
    /** Requests a vertical line, which should be painted at the next
     	paint-call.<p>
     	
       @param x The x-coordinate where the line should be painted.
     */
    public void paintVerticalLine(int x)
    {
        //  the line should be visible for 5 paint calls.
        removeCounter = 5;
        
        // set the mouse coordinates, where the line should be painted.
        mouseX = x;

    }
}