package com.ai_center.planpresentation;


import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Set;

import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JMenuItem;
import javax.swing.UIManager;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.EventListenerList;

import com.ai_center.dragonbreath.Constraint;
import com.ai_center.dragonbreath.GlobalSearchControl;
import com.ai_center.dragonbreath.RealTimeControl;
import com.ai_center.dragonbreath.constraintgraph.Node;
import com.ai_center.planpresentation.plugins.Plugin;
import com.ai_center.planpresentation.plugins.RetractableView;
import com.ai_center.planpresentation.utilities.ChangeMessage;
import com.ai_center.planpresentation.utilities.PluginInitConstant;




/** A graphical user interface for the EXCALIBUR project <p>

 This is the main class, which starts the gui and controls the plugins. The
 PlanPresentation class can be instantiated at any point in the planning
 process.<p>

 <pre>
 *
 * GlobalSearchControl gsc = new GlobalSearchControl(2);
 * PlanPresentation p = new PlanPresentation(gsc);
 *
 </pre>

 @author Daniel Weiss
 @version $Revision: 1.5 $
 */

public class PlanPresentation
        implements
            PlanPresentationConstants,
            PlanPresentationAPI,
            ChangeListener
{


    // /////////////////////////////////////
    // attributes

    /** A mapping from contraintTypeIDs to Plugins, which should be informed
     about added or deleted constraints of this type.
     */
    private HashMap constraintTypeRegistry = null;

    /** The current time of the planer */
    private long currentTime;

    /** The link to the GlobalSearchControl */
    private GlobalSearchControl gsc = null;

    /**
     Contains the subscribed EventListeners to track events like zooming,
     scrolling and sortmode changes.
     */
    protected EventListenerList listenerList = new EventListenerList();

    /** The main window */
    private PlanJFrame planGUI = null;

    /**
     * Represents the sort order for contraint views.
     * @see PlanPresentationConstants#SORTMODE_UPDATE
     * @see PlanPresentationConstants#SORTMODE_GROUPED
     */
    private int sortMode;

    /** The current zoomfactor */
    private Float zoomfactor;


    // /////////////////////////////////////
    // operations

    /**
     * Initiates the gui control system and opens a planning window.
     * @param gsc  The link to the GlobalSearchControl
     */
    public PlanPresentation( GlobalSearchControl _gsc )
    {
        gsc = _gsc;
        constraintTypeRegistry = new HashMap();

        // Adjust LookAndFeel to the current Platform
        try
        {
            UIManager.setLookAndFeel( UIManager.getSystemLookAndFeelClassName() );
        }
        catch ( Exception e )
        {
            e.printStackTrace();
        }

        // Prepare planning window.
        planGUI = new PlanJFrame( this );

        // load all plugins
        PluginLoader pluginInit = new PluginInitConstant();
        pluginInit.setPlanPresentationAPI( (PlanPresentationAPI)this );
        pluginInit.loadPlugins();

        // generate menues and controls
        planGUI.generateMenus();
        planGUI.generateControls();

        // read the existing csp graph
        readCSPGraph();

        // connect constraint creation and deletion listener to the gsc
        gsc.addChangeListener( (ChangeListener)this );

        // make the planning window visible
        planGUI.show();
    }


    /** Read the existings CSP Graph.<p>
     */
    private void readCSPGraph()
    {
        // get all constraints
        Set constraintSet = gsc.getConstraints();
        Iterator iteratorConstraint = constraintSet.iterator();
        Constraint constraint = null;

        // walk through all existing constraints
        while (iteratorConstraint.hasNext())
        {
            try
            {
                constraint = (Constraint) iteratorConstraint.next();
                // add new constraint
                updateConstraint(constraint);
            } catch (ClassCastException cce)
            {
                System.out
                        .println("ClassCastException: Constraint doesn't implement Constraint Interface. "
                                + cce.getMessage());
            }


        }

    }

    /** Fires a changeMessage.<p>

     This method supports normal ChangeListeners and MessageListeners.
     MessageListeners support filters and customized messages.<p>

     @param message The ChangeMessage.
     */
    protected void fireChange(ChangeMessage message)
    {
        // generate changeEvent
        ChangeEvent changeEvent = new ChangeEvent(message);

        //inform all listeners
        Object[] listeners = listenerList.getListenerList();
        ChangeListener listener = null;
        for (int i = listeners.length - 2; i >= 0; i -= 2)
        {
            if (listeners[i] == ChangeListener.class)
            {
                ((ChangeListener) listeners[i + 1]).stateChanged(changeEvent);
            }
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see com.ai_center.planpresentation.PluginController#absoluteY(int)
     */
    public int absoluteY(int relativeY)
    {
        return relativeY;
    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.PlanPresentationAPI#addChangeListener(javax.swing.event.ChangeListener)
     */
    public void addChangeListener(ChangeListener l)
    {
        listenerList.add(ChangeListener.class, l);
    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.PlanPresentationAPI#addControl(javax.swing.JComponent, int)
     */
    public void addControl(JComponent controlComponent, int order)
    {

        planGUI.addControl(controlComponent, order);

    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.PlanPresentationAPI#addMenu(java.lang.String, int)
     */
    public void addMenu(String menuname, int mnemonic , int order)
    {
        planGUI.addMenu(menuname, mnemonic,order);
    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.PlanPresentationAPI#addMenuItem(javax.swing.JMenuItem, java.lang.String)
     */
    public void addMenuItem(JMenuItem jMenuItem, String menuname)
    {
        planGUI.addMenuItem(jMenuItem, menuname);
    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.PlanPresentationAPI#addToolButton(javax.swing.JComponent, int)
     */
    public void addToolButton(JComponent toolButton, int priority)
    {
        planGUI.addToolButton(toolButton, priority);

    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.PlanPresentationAPI#addView(com.ai_center.planpresentation.plugins.RetractableView)
     */
    public void addView(RetractableView view)
    {
        planGUI.viewList.add(view);
        updateView();
    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.PlanPresentationAPI#addViewChangeListener(javax.swing.event.ChangeListener)
     */
    public void addViewChangeListener(ChangeListener changeListener)
    {
        planGUI.addViewChangeListener(changeListener);
    }


    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.PlanPresentationAPI#engineStop()
     */
    public void engineStop()
    {
        if (gsc.getRealTimeControl().getMode() != RealTimeControl.STEPPING)
        {
            gsc.getRealTimeControl().setMode(RealTimeControl.STEPPING);
        }
    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.PlanPresentationAPI#getGlobalSearchControl()
     */
    public GlobalSearchControl getGlobalSearchControl()
    {
        return gsc;
    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.PlanPresentationAPI#getGuiFrame()
     */
    public JFrame getGuiFrame()
    {
        return planGUI;
    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.PlanPresentationAPI#getSortMode()
     */
    public int getSortMode()
    {
        return sortMode;
    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.PlanPresentationAPI#getSwimlaneControl()
     */
    public SwimlaneControl getSwimlaneControl()
    {
        return planGUI.getSwimlaneControl();
    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.PlanPresentationAPI#getViewPanelWidth()
     */
    public int getViewPanelWidth()
    {
        return planGUI.planPaintArea.getWidth();
    }


    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.PlanPresentationAPI#registerConstraintType(int, com.ai_center.planpresentation.plugins.Plugin)
     */
    public void registerConstraintType(int constraintTypeID, Plugin plugin)
    {
        // get existing plugin List
        LinkedList pluginList = (LinkedList) constraintTypeRegistry
                .get(new Integer(constraintTypeID));

        // check if this list really exists
        if (pluginList == null)
        {
            // make new list
            pluginList = new LinkedList();
        }
        // add new plugin to the list
        pluginList.add(plugin);

        // put list back in the constraintTypeRegistry
        constraintTypeRegistry.put(new Integer(constraintTypeID), pluginList);
    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.PlanPresentationAPI#removeChangeListener(javax.swing.event.ChangeListener)
     */
    public void removeChangeListener(ChangeListener changeListener)
    {
        listenerList.remove(ChangeListener.class, changeListener);
    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.PlanPresentationAPI#removeView(com.ai_center.planpresentation.plugins.RetractableView)
     */
    public void removeView(RetractableView view)
    {
        // find the view in the planGUI.viewList
        Iterator it = planGUI.viewList.iterator();
        while (it.hasNext())
        {
            RetractableView retractableView = (RetractableView) it.next();
            if (view == retractableView)
            {
                // remove it from the planGUI.viewList
                it.remove();
                break;
            }
        }
        // request a view update
        updateView();
    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.PlanPresentationAPI#removeViewChangeListener(javax.swing.event.ChangeListener)
     */
    public void removeViewChangeListener(ChangeListener l)
    {
        planGUI.removeViewChangeListener(l);
    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.PlanPresentationAPI#scroll(int)
     */
    public void scroll(int scrollstep, boolean extendViewIfNecessary)
    {
        planGUI.scroll(scrollstep, extendViewIfNecessary);
    }


    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.PlanPresentationAPI#setCurrentTime(long)
     */
    public void setCurrentTime( long ct )
    {
        //set new currentTime
        currentTime = ct;

        // check if ct is bigger than the currentTimeModel maximum
        if ( planGUI.currentTimeModel.getMaximum() < ct )
        {
            // adjust the maximumg
            planGUI.currentTimeModel.setMaximum( (int)( ct + 100 ) );
        }

        if ( planGUI.mainViewModel.getMaximum() < ct )
        {
            planGUI.mainViewModel.setMaximum( (int)( ct + 100 ) );
        }
        // set currentTime in currentTimeModel (used in the progress bars)
        planGUI.currentTimeModel.setValue( (int)currentTime );

        // change time display
        planGUI.currentTimeJProgressBarNorth.setString( "Current Time: "
            + currentTime);
        planGUI.currentTimeJProgressBarSouth.setString( "Current Time: "
            + currentTime);

        // inform all listeners about a current time change
        fireChange( new ChangeMessage( CURRENT_TIME_CHANGED ) );
    }


    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.PlanPresentationAPI#setHelpText(java.lang.String)
     */
    public void setHelpText( String helpText )
    {
        planGUI.helpMessage.setText( helpText );
        planGUI.helpMessage.setToolTipText( helpText );
    }


    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.PlanPresentationAPI#setSortMode(int)
     */
    public void setSortMode( int _sortMode )
    {
        sortMode = _sortMode;
        fireChange( new ChangeMessage( SORTMODE_CHANGED ) );
        updateView();
    }


    /** Sets the zoomfactor<p>

     Sets the zoomfactor and informs all listeners about the change.
     @param _zoomfactor the new zoomfactor.
     */
    public void setZoomfactor(Float _zoomfactor)
    {
        zoomfactor = _zoomfactor;
        fireChange(new ChangeMessage(ZOOM_CHANGED));
    }

    /** Callback function for stateChanges in the GlobalSearchControl

     This is the main entry point for datachanges in the model.
     @param e The ChangeEvent.
     */
    public void stateChanged(ChangeEvent e)
    {
        if (e.getSource() instanceof Constraint)
        {
            updateConstraint((Constraint) e.getSource());
        }
    }

    /** Add or remove a constraint.<p>

       updateConstraint() checks, if the constraint is new (it is known
       by the datamodel) or if it should be removed (it is not known).
       @param constraint The constraint to be checked.
     */
    public void updateConstraint(Constraint constraint)
    {
        Plugin plugin = null;
        LinkedList pluginList = null;
        try
        {
            // get the plugins to be informed
            pluginList = (LinkedList) constraintTypeRegistry
                    .get(new Integer(constraint.getTypeID()));
        } catch (ClassCastException cce)
        {
            System.out
                    .println("ClassCastException: wrong Object in contraintTypeRegistry. "
                            + cce.getMessage());
        }

        if (pluginList != null)
        {
            Iterator it = pluginList.iterator();

            if (gsc.containsNode((Node) constraint))
            {
                // Constraint exists in datamodel, so add it to the gui
                while (it.hasNext())
                {
                    // inform all plugins
                    plugin = (Plugin) it.next();
                    plugin.addConstraint(constraint);
                }
            } else
            {
                // Constraint was removed from datamodel, so remove it from
                // the gui

                while (it.hasNext())
                { // inform all plugins
                    plugin = (Plugin) it.next();
                    plugin.removeConstraint(constraint);
                }
            } // end if
        }
    }

   /*
    *  (non-Javadoc)
    * @see com.ai_center.planpresentation.PlanPresentationAPI#updateView()
    */
    public void updateView()
    {
        planGUI.updateView = true;
    }


        /* (non-Javadoc)
         * @see com.ai_center.planpresentation.PlanPresentationAPI#moveView(com.ai_center.planpresentation.plugins.RetractableView, boolean)
         */
        public void moveView(RetractableView r, boolean up) {
                planGUI.moveView(r,up);

        }
} // end PlanPresentation
