package com.ai_center.planpresentation;


import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JMenuItem;
import javax.swing.event.ChangeListener;

import com.ai_center.dragonbreath.GlobalSearchControl;
import com.ai_center.planpresentation.plugins.Plugin;
import com.ai_center.planpresentation.plugins.RetractableView;




/** API for the PlanPresentation graphical user interface.<p>
 
 This interface declares the communication from plugins to the planpresentation
 application. Plugin programmer can access all necessary functions like addView,
 addMenu etc. with it<p>
 
 @author Daniel Weiss
 @version $Revision: 1.6 $
 */
public interface PlanPresentationAPI
{


    /** Returns the absolut y value of the coordinate.<p>
     
     absoluteY(relativeY) computes the absolute value of y values in the
     planning area.<p>
     
     @param relativeY The local y value in the planning area 
     @return the absolute y value in the JFrame.
     */
    public int absoluteY(int relativeY);


    /** Adds a ChangeListener to the PlanPresentation.<p>
     
     Informs about sort mode changes.<p> 	
     @param listener The listener to be added.
     */
    public void addChangeListener(ChangeListener listener);

    /** Adds a control element to the gui.<p>
     
     Adds a control element to the left column of the gui. Like zoomslider.<p> 	
     @param controlComponent Can be any JComponent.
     @param order the order in which the elements should appear. zero on top.
     */
    public void addControl(JComponent controlComponent, int order);


    /** Adds a menu to the menubar.<p>
     
     Creates a new menu if none with the same menu name exists. If you are not
     sure, if the menu already exist, call addMenu again, this function is
     idempotent.<p>
     @param menuname The name of the menu.
     @param mnemonic The shortcut to open the menu.
     @param order The order, zero on the left side.
     */
    public void addMenu(String menuname, int mnemonic, int order);

    /** Adds a menu item to a give menu.<p>

     @param jMenuItem The menu item to be added.
     @param menuname The menu name in which this item should be added.
     */
    public void addMenuItem(JMenuItem jMenuItem, String menuname);

    /** Adds a button in the toolbar below the menu.<p>
     
     The button can also be any other JComponent.<p> 	
     
     @param button The button to be added.
     @param order The order from left to right. Zero is left.
     */
    public void addToolButton(JComponent button, int order);


    /** Adds a new plugin view to the planning area.<p>
     
     @param view A RetractableView element.
     */
    public void addView(RetractableView view);

    /** Adds a ViewChangeListener.<p>
     
     The ViewChangeLister will be informed about all changes corresponding,
     to the view. Like scrolling or time stepping.<p> 	
     
     @param changeListener A changeListener which should be informed.
     */
    public void addViewChangeListener(ChangeListener changeListener);

    /** Pauses the engine.
     
     A restart is only possible with user interaction. (clicking on the
     play-button).<p>
     */
    public void engineStop();

    /** Returns the connected GlobalSearchControl.<p>
     
     @return GlobalSearchControl
     */
    public GlobalSearchControl getGlobalSearchControl();

    /** Returns the Frame of the mainapplication.
     
     @return The JFrame in which the application runs.
     */
    public JFrame getGuiFrame();

    /** Returns the current SortMode.
     
     @return The SortMode defined in PlanPresentationConstants.
     @see PlanPresentationConstants#SORTMODE_UPDATE
     @see PlanPresentationConstants#SORTMODE_GROUPED
     */
    public int getSortMode();

    /** Returns the SwimlaneControl
     
     @return The current SwimlaneControl
     */
    public SwimlaneControl getSwimlaneControl();

    /** Returns the width of the view.

     @return width 
     */
    public int getViewPanelWidth();

    /** Register for constraints with the given typeID.<p>
     
     The plugin will be informed about all new or removed constraints with
     this typeID.
     * @param typeID The typeID of the constraint.
     * @param plugin The plugin which should be informed.
     */
    public void registerConstraintType(int typeID, Plugin plugin);
    
    /** Removes the a changeListener
     
        @param changeListener The changeListener to be removed.
     */
    public void removeChangeListener(ChangeListener changeListener);

    /** Removes the view.<p>
     
     @param view The RetractableView to be removed.
     */
    public void removeView(RetractableView view);

    /** Removes the ViewChangeListener from the listenerList.
     
     @param changeListener The changeListener to be removed.
     */
    public void removeViewChangeListener(ChangeListener changeListener);

    /** Scrolls the view to the right or left in different speeds.<p>
	
	Normally scrolling is traced by the RetractableView and the plugin
	programmer does not need to take care about it.<p>
	
     @param scrollstep A high positive value scrolls fast to the right,
     a negative value scrolls to the right.
     @param extendViewIfNecessary If the planing horizont shall be adjusted if necessary.
     */
    public void scroll(int scrollstep, boolean extendViewIfNecessary);

    /** Sets the CurrentTime in the graphical user interface.<p>
       
       @param ct The new current time.
     */
    public void setCurrentTime(long ct);

    /** Sets the helpText on top of the planning area.<p>
     
       @param helpText The new helpText.
     */
    public void setHelpText(String helpText);


    /** Sets a new sortmode.<p>
      
      @param sortmode The values from the PlanPresentationConstants beginning with SORTMODE_
      @see PlanPresentationConstants#SORTMODE_UPDATE
      @see PlanPresentationConstants#SORTMODE_GROUPED
     */
    public void setSortMode(int sortmode);

   /** Requests a full view update.<p>
    
      e.g.when the vertical size of the view has changed is this call necessary.<p>
    */
    public void updateView();
    
    public void moveView(RetractableView r, boolean up);
}