package com.ai_center.planpresentation;


import java.awt.BasicStroke;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Stroke;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JComponent;
import javax.swing.SwingUtilities;




/** The instance of this class represents on line painted on the screen.<p>
 
 It is instantiated by the SwimlaneControl, knows on object of the datamodel
 as owner and has a liste of connected other objects.
 
 @author Daniel Weiss
 @version $Revision: 1.5 $
 */
public class Swimlane implements PlanPresentationConstants
{


    /** The link to the swimlaneControl */
    private SwimlaneControl swimlaneControl;

    /** The maximum y value. This is the y-value, where the vertical line 
     ends on the bottom. */
    private int maxY;

    /** The minimum y value. This is the y-value, where the vertical line 
     beginns on the top. */
    private int minY;

    /** The x-value of the connected-element which is furthest left of all
     connected objects. */
    private int minX;

    /** The x-value of the vertical line. This is NOT equal with the minX,
     because we want that the line is 5 pixel more left than the leftest
     object. */
    private int lineX;

    /** The connectionType, which is transmitted to the connected objects.
     */
    private int connectionType;

    /** The list of the connected gui objects. */
    private List connectorList = new LinkedList();

    /** The owner of this swimlane. */
    private SwimlaneConnector ownerConnector;

    /**Mapping from the SwimlaneConnector object to the corresponding 
     x coordinate. */
    private Hashtable xHash = new Hashtable();

    /** Mapping from the SwimlaneConnector object to the corresponding 
     x coordinate.*/
    private Hashtable yHash = new Hashtable();

    /** The link to the plan paint area. This is needed to convert 
     coordinates from the plan paint area to the glass pane. */
    private JComponent planPaintArea = null;

    /** The link to the glassPane. This is needed to convert 
     coordinates from the plan paint area to the glass pane. */
    private JComponent glassPane = null;

    private static float[] dashArray = {4.0f,1.0f};
    
    private static BasicStroke dashStroke = new BasicStroke(1.0f,1,2,1.0f,dashArray,1);
    /** Creates the swimlane
     
     @param _swimlaneControl The link to the swimlanecontrol.
     @param _ownerConnector The owner of the swimlane.
     */

    public Swimlane(SwimlaneControl _swimlaneControl,
            SwimlaneConnector _ownerConnector)
    {

        ownerConnector = _ownerConnector;
        swimlaneControl = _swimlaneControl;
        planPaintArea = swimlaneControl.getPlanPaintArea();
        glassPane = swimlaneControl.getGlassPane();


    }

    /** Paints the swimlane.<p>
     
     @param g The graphics context.
     */
    public void paint(Graphics g)
    {

        Graphics2D g2 = (Graphics2D) g;
        Stroke defaultStroke = g2.getStroke();
        
        // compute the minX, minY, maxY values
        computeCoordinates();

        // first paint vertical line from minY to maxY
        boolean verticalLinePainted = false;

        // check if the vertical line is visible
        if (lineX > RETRACT_BUTTON_WIDTH && minY != maxY && 
        		minY < swimlaneControl.maxVisibleY() && maxY > swimlaneControl.minVisibleY() // and not completely outside  
				) 
        {

            // clipping
          if (minY < swimlaneControl.minVisibleY())
            {
                minY = swimlaneControl.minVisibleY();
            }
        	  
            if (maxY > swimlaneControl.maxVisibleY())
                  //  + Math.abs(planPaintArea.getY()))
            {

                maxY = swimlaneControl.maxVisibleY();
                        //+ Math.abs(planPaintArea.getY());
            }
        	
            // convert points from planpaint area to the coordinate system
            // of the glass pane.
            Point p = SwingUtilities.convertPoint(planPaintArea, lineX, minY,
                    glassPane);

            Point p2 = SwingUtilities.convertPoint(planPaintArea, lineX, maxY,
                    glassPane);

            if (connectionType < 0) {
                g2.setStroke(dashStroke);
            }
            
            
          
            // draw the line
            g2.drawLine(p.x, p.y, p2.x, p2.y);

            // mark the vertical line drawed so that that the horizontal 
            // lines will be painted.
            verticalLinePainted = true;
            if (connectionType < 0) {
                g2.setStroke(defaultStroke);
            }
        }
       
        if (  lineX < RETRACT_BUTTON_WIDTH ) {
        	// the horizonal line begins "under" the retract button. Clip it and let it begin right of the retract
        	// button
        	lineX = RETRACT_BUTTON_WIDTH;
        }
        if ( verticalLinePainted) {
        // draw horizontal lines to every swimlaneConnector
        Iterator it = connectorList.iterator();
        while (it.hasNext())
        {
            Object o = it.next();
            SwimlaneConnector sc = (SwimlaneConnector) swimlaneControl
                    .getConnector(o);
            if (sc != null)
            {

                // check if the connector is visible on the screen and
                // if it should be visible from the point of view of 
                // the datamodel
                if (sc.isVisible(connectionType) && sc.shouldVisible())
                {

                    // get the coordinates from the hash 
                    // they were saved before by the computeCoordinates()
                    Integer xInteger = (Integer) xHash.get(sc);
                    Integer yInteger = (Integer) yHash.get(sc);

                    // check if the coordinate retrieving was successfull
                    if (xInteger != null && yInteger != null)
                    {
                        int x = xInteger.intValue();
                        int y = yInteger.intValue();

                        // check if the vertical line was painted and
                        // if this horizontal line will be visible or
                        // out of the view range.
                          //   && verticalLinePainted removed
                        if ((lineX < x)
                              
                                && (y > Math.abs(planPaintArea.getY()))
                                && (y < Math.abs(planPaintArea.getY())
                                        + swimlaneControl.maxVisibleY())
							)
                        {
                            // convert points to the correct coordinate 
                            // system
                            Point p = SwingUtilities.convertPoint(
                                    planPaintArea, lineX, y, glassPane);

                            Point p2 = SwingUtilities.convertPoint(
                                    planPaintArea, x, y, glassPane);

                            // draw the horizontal line
                            g.drawLine(p.x, p.y, p2.x, p2.y);
                        }

                    }
                }
            }
        }
        }

    }


    /** This function computes the coordinates of minX, minY, maxY.<p>
     
     	Additional it put the retrieved coordinates of every connector in a
     	hash, to cache the values.*/
    private void computeCoordinates()
    {
        // reset old data
        xHash = new Hashtable();
        yHash = new Hashtable();
        minX = -1;
        minY = -1;
        maxY = -1;
        lineX = -1;

        Iterator it = connectorList.iterator();

        while (it.hasNext())
        {
            Object o = it.next();
            SwimlaneConnector sc = (SwimlaneConnector) swimlaneControl
                    .getConnector(o);
            // check if a connector exists for this datamodel object
            if (sc != null)
            {
                // check if this gui object is visible
                if (sc.isVisible(connectionType) && sc.shouldVisible())
                {

                    int x = 0;
                    int y = 0;
                    
                    // get the point, where the line should be connected at
                    // a give connectionType
                    
                    x = sc.getConnectorX(connectionType);
                    y = sc.getConnectorY(connectionType);
                    
                    // check if x is the new minimum
                    if (x < minX || minX == -1)
                    {
                        minX = x;
                        int offset = 0;
                        // if we have a connection on the left or right side
                        // we want the vertical line to have a little distance
                        // so add a offset of -10 pixel to the left.
                       
                        if ((sc.getPosition(connectionType) == SwimlaneConnector.WEST)
                                || (sc.getPosition(connectionType) == SwimlaneConnector.EAST))
                        {
                            offset = -10;
                        }
                        // check if we have a new minimum value fpr the vertical
                        // line
                        if (minX + offset < lineX || lineX == -1)
                        {
                            lineX = minX + offset;
                        }

                    }
                    // cache the x coordinate for this connector
                    xHash.put(sc, new Integer(x));

                    // check if we have a new minimum y value
                    if (y < minY || minY == -1)
                    {
                        minY = y;
                    }
                    // check if we have a new maximum y value
                    if (y > maxY || maxY == -1)
                    {
                        maxY = y;
                    }
                    // cache the y coordinate for this connector
                    yHash.put(sc, new Integer(y));
                }
            }
        }
    }


   /** Sets a new connector list.<p>
    
      This function is called by the SwimlaneControl implementation, to
      change the content of the connector list.<p> 	
      @param _connectorList The new connector list.
    */
    public void setConnectorList(List _connectorList)
    {
        connectorList = _connectorList;
    }

    /** Sets a new conenction type.<p>
     
      This function is called by the SwimlaneControl implementation, to
      change the connection type.<p> 	
       @param _connectionType The new connectionType.
     */
    public void setConnectionType(int _connectionType)
    {
        connectionType = _connectionType;

    }
} 
