package com.ai_center.planpresentation;


/** The interface for all objects, which want a swimlane connected to it.<p>
 
 The SwimlaneConnector Interface specifies the communication  from the
 SwimlaneControl to the object, which wants a swimlane connected to it.
 <p>  
 
 @author Daniel Weiss
 @version $Revision: 1.3 $
 */
public interface SwimlaneConnector
{

    /** connection should be on the right side of the object */
    static final int EAST = 4;
    
    /** connection should be on the top of the object */
    static final int NORTH = 1;

    /** connection should be on the bottom of the object */
    static final int SOUTH = 2;
    
    /** connection should be on the left side of the object */
    static final int WEST = 3;

    /** Returns the absolute x-coordinate, where the connection should be
     attached.<p>
     
     @param connectionType One object can have more connections on. With this
     param the connection is identified.
     
     @return The absoulte x coordinate in the planning window.
     */
    public int getConnectorX(int connectionType);


    /** Returns the absolute y-coordinate, where the connection should be
     attached.<p>
     
     @param connectionType One object can have more connections on. With this
     param the connection is identified.
     
     @return The absoulte y coordinate in the planning window.
     */
    public int getConnectorY(int connectionType);

    /** Returns SwimlaneConnector.NORTH, SOUTH, WEST or EAST to represent, where the 
     connection should be attached.
     
     @param connectionType One object can have more connections on. With this
     param the connection is identified.
     
     @return SwimlaneConnector.NORTH, SOUTH, WEST or EAST
     @see #NORTH
     @see #SOUTH
     @see #WEST
     @see #EAST
     */
    public int getPosition(int connectionType);

    /** Returns if this is a outgoing connection or false, if it is a incoming
     connection.
     
     (Prepared for a later use.)
     
     @param connectionType One object can have more connections on. With this
     param the connection is identified.
     
     @return True if outgoing or false if incoming.
     */
    public boolean isOutgoing(int connectionType);

    /** Returns true if the connector is really visible and if the swimlane
     should be attached.<p>
     @param connectionType One object can have more connections on. With this
     param the connection is identified.
     
     @return True or false
     */
    public boolean isVisible(int connectionType);


    /** Sets the PriorityOffset to define the sort order of the views.
     
     e.g. used to sort all connected resources on top, when clicking on an
     action.<p>
     
     @param offset The offset which is added to the last access time.
     */
    public void setPriorityOffset(long offset);


    /** Sets if this object should be visible from the datamodel point of view.
     
     This function doesn't care about if the object is really visible on the
     screen. This only represents the point of view of the datamodel.
     
     @param shouldVisible If this object should visible.
     */
    public void setShouldVisible(boolean shouldVisible);


    /** Returns if this should visible.<p>
     
     See setShouldVisible for more information.<p>
     @return If this object should visible
     @see #setShouldVisible
     */
    public boolean shouldVisible();


    /** Terminates the object.
     
     i.e. makes shouldVisible false and removes unnecessary memory references. 
     */
    public void terminate();


}