package com.ai_center.planpresentation.plugins;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;

import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JPanel;

import com.ai_center.planpresentation.PlanPresentationAPI;
import com.ai_center.planpresentation.PlanPresentationConstants;
import com.ai_center.planpresentation.utilities.OvalBorder;

/**
 * Every Plugin has to implement the RetractableView class, to show information
 * in the plan paint area.
 * <p>
 * 
 * This class provides also a button to retract and expand a view correct.
 * Additional it shows the name of the view and gives the possibility to add
 * additional control elements in this view. (like buttons to add goals)
 * 
 * @author Daniel Weiss
 * @version $$
 */
public abstract class RetractableView extends JPanel implements
		MouseMotionListener, PlanPresentationConstants, Comparable {

	/** The width of all control components, like addGoal and editSensordata */
	int controlComponentsWidth = 5; // start with a margin of 5.

	/**
	 * The link to the controller. See presentation-abstraction-controll pattern
	 * and bottom level agent.
	 */
	public Controller controller;

	/** The class saves here the height, to recongize height-changes */
	private int oldheight;

	/** The button to retract and to expand the view. */
	JButton retractButton = new JButton();

	/** The state, if this view is retracted or not. */
	protected boolean retracted = false;

	/** The name of the view. */
	protected String viewName;

	/** The button, where the name of the view is displayed. */
	JButton viewNameButton = new JButton();

	/**
	 * Creates a view.
	 * <p>
	 * 
	 * @param _c
	 *            The link to a controller interface.
	 * @param _viewName
	 *            The name of the view.
	 */
	public RetractableView(Controller _c, String _viewName) {
		super(true);
		createRetractableView(_c, _viewName, true);

	}

	/**
	 * @param _c
	 * @param name
	 * @param b
	 */
	private void createRetractableView(Controller _c, String _viewName,
			boolean moveButtons) {

		controller = _c;
		viewName = _viewName;

		// register mouse listener to track autoscroll
		this.addMouseMotionListener(this);

		ActionListener buttonActionListener = new ButtonActionListener(this);
		// init retractbutton
		retractButton.setActionCommand("retract");
		retractButton.addActionListener(buttonActionListener);
		this.add(retractButton);

		// init the button which shows the view name
		viewNameButton.setText(viewName);
		//   viewNameButton.setBackground(new Color(253,196,169));
		viewNameButton.setBackground(this.getBackground());
		viewNameButton.setBorder(new OvalBorder(10, 10, Color.black,
				Color.black));
		//.setBorder(BorderFactory.createLineBorder(Color.black, 1));
		// -10 = margin top and bottom
		viewNameButton.setSize(200, HEADER_HEIGHT - 10);
		this.addControl(viewNameButton);

		if (moveButtons) {
			JButton b = new JButton();
			b.setIcon(new ImageIcon(controller.getPlanPresentationAPI()
					.getClass().getResource("resources/Down.gif")));
			b.setSize(20, 20);
			b.setOpaque(false);
			b.setBorderPainted(false);
			b.setToolTipText("Move this view below among the other view");
			b.addActionListener(buttonActionListener);
			b.setActionCommand("down");

			this.addControl(b);

			b = new JButton();
			b.setIcon(new ImageIcon(controller.getPlanPresentationAPI()
					.getClass().getResource("resources/Up.gif")));
			b.setSize(20, 20);
			b.setOpaque(false);
			b.setBorderPainted(false);
			b.setToolTipText("Move this view above other view");
			b.addActionListener(buttonActionListener);
			b.setActionCommand("up");

			this.addControl(b);
		}

	}

	/**
	 * @param controller2
	 * @param name
	 * @param moveButtons
	 */
	public RetractableView(Controller _c, String _viewName, boolean moveButtons) {
		super(true);
		createRetractableView(_c, _viewName, moveButtons);

	}

	public PlanPresentationAPI getPlanPresentationAPI() {
		return controller.getPlanPresentationAPI();
	}

	/**
	 * Adds a control element to the view.
	 * <p>
	 * 
	 * Like addGoal or edit Sensordata.
	 * <p>
	 * 
	 * @param component
	 *            A component, normally a JButton.
	 */
	public void addControl(JComponent component) {
		controlComponentsWidth += 10; // add margin
		component.setLocation(controlComponentsWidth, 5);
		this.add(component);
		controlComponentsWidth += component.getWidth();
	}

	/**
	 * Changes the displaymode from retracted to expanded and reverse.
	 * <p>
	 */
	public void changeDisplayMode() {
		// save height for a later comparement

		oldheight = this.getHeight();

		// check if is retracted or if it is expanded
		if (!retracted) {
			// it is expanded, so retract it
			retracted = true;
			int width = controller.getPanelWidth();

			//make the view height very little
			this.setBounds(0, 0, width, 15);
			this.setMinimumSize(new Dimension(width, 15));
			this.setPreferredSize(new Dimension(width, 15));

			// write the name of the view on the retract button (now very big)
			retractButton.setText(viewName);
			// set the width of the button to the width of the whole panel
			retractButton.setBounds(0, 0, this.getWidth(), this.getHeight());
		} else {
			// it is retracted, so expand it
			retracted = false;
			// remove the text from the retract button
			retractButton.setText(null);

			// make the button smaller
			retractButton.setBounds(0, 0, RETRACT_BUTTON_WIDTH, this
					.getHeight());

			// update the view size again.
			updateViewSize();
		}
	}

	/**
	 * This function compares to views in relation to the priority.
	 * <p>
	 * 
	 * The priority is the last access time added to the priorityOffset.
	 * <p>
	 * 
	 * @param o
	 *            The other retractableView to compare with.
	 */
	public int compareTo(Object o) {
		// check if we can compare these two objects
		if (o instanceof RetractableView) {

			RetractableView otherView = (RetractableView) o;
			int ret = -this.getPriority().compareTo(otherView.getPriority());
			// reverse order;
			return ret;
		} else {
			throw new ClassCastException();
		}
	}

	/**
	 * Returns the priority.
	 * <p>
	 * 
	 * The priority = accessTime + priorityOffset.
	 * <p>
	 * 
	 * @return The priority of this view.
	 */
	public Long getPriority() {

		return controller.getPriority();

	}

	/**
	 * Checks if the view is retracted or expanded
	 * 
	 * @return True if retracted.
	 */
	public boolean isRetracted() {
		return retracted;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.awt.event.MouseMotionListener#mouseDragged(java.awt.event.MouseEvent)
	 */
	public void mouseDragged(MouseEvent evt) {
		// Nothing to do, will be overwritten by a implementor, who uses it.
	}

	/**
	 * This function tracks MouseEvents to realize autoscroll.
	 * 
	 * @param evt
	 *            The MouseEvent.
	 */
	public void mouseMoved(MouseEvent evt) {
		int width = getWidth();
		Point location = evt.getPoint();
		int scrollstep = 0;

		//check for normal speed
		if (location.x > (width - SCROLL_NORMAL_AREA)) {
			scrollstep = SCROLL_NORMAL;
			// scroll to the right with normal speed
		} else if (location.x < SCROLL_NORMAL_AREA) {
			scrollstep = -SCROLL_NORMAL;
			// scroll to the lefft with normal speed
		}

		//check for fast speed
		if (location.x > (width - SCROLL_FAST_AREA)) {
			scrollstep = SCROLL_FAST; // scroll to the right with normal speed
		} else if (location.x < SCROLL_FAST_AREA) {
			scrollstep = -SCROLL_FAST; // scroll to the lefft with normal speed
		}

		if (scrollstep != 0) {
			controller.scroll(scrollstep, false);
		}

	}

	/**
	 * Paint the component.
	 * <p>
	 * 
	 * @param g
	 *            The Graphics context.
	 */
	public void paintComponent(Graphics g) {
		// check if the height has changed
		if (oldheight != this.getHeight()) {
			// the height has changed, we must adjust the retractbutton height.
			oldheight = this.getHeight();
			retractButton.setBounds(0, 0, RETRACT_BUTTON_WIDTH, this
					.getHeight());
		}
		viewNameButton.setBackground(this.getBackground());
		super.paintComponent(g);

	}

	/**
	 * Delegate function. Changes the PriorityOffset.
	 * <p>
	 * 
	 * @param _offset
	 *            The new priorityOffset
	 * @see com.ai_center.planpresentation.plugins.Controller#setPriorityOffset
	 */
	public void setPriorityOffset(long _offset) {
		controller.setPriorityOffset(_offset);

	}

	/**
	 * This function must be implmented by the plugin view.
	 * <p>
	 * It is called by the RetractableView, when someone clicks on the retract-
	 * button. This function has to adjust the view size to the actual required
	 * size, to show all information.
	 */
	public abstract void updateViewSize();

} // end RetractableView

/**
 * The ButtonActionListener tracks the button clicks on the retract-button.
 * <p>
 * 
 * @author Daniel Weiss
 * @version $Revision: 1.6 $
 */

class ButtonActionListener implements ActionListener {

	/** The link to the retractableView */
	RetractableView view = null;

	/**
	 * The constructor to create the ButtonActionListener.
	 * <p>
	 * 
	 * @param view
	 *            The link to the retractableView.
	 */
	public ButtonActionListener(RetractableView _view) {

		view = _view;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 */
	public void actionPerformed(ActionEvent e) {
		// switch the the displaymode from retracted to expanded and reverse

		if (e.getSource() instanceof JButton) {
			// remove the selected flag from the button.
			if (e.getActionCommand().equals("retract")) {
				view.changeDisplayMode();
				JButton klickedButton = (JButton) e.getSource();
				klickedButton.getModel().setSelected(false);
			} else if (e.getActionCommand().equals("up")) {
				view.getPlanPresentationAPI().moveView(view, true);
			} else if (e.getActionCommand().equals("down")) {
				view.getPlanPresentationAPI().moveView(view, false);
			}
		}

	}

}