package com.ai_center.planpresentation.utilities;


import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import com.ai_center.planpresentation.PlanPresentationAPI;
import com.ai_center.planpresentation.PlanPresentationConstants;
import com.ai_center.planpresentation.SwimlaneControl;
import com.ai_center.planpresentation.plugins.Controller;
import com.ai_center.planpresentation.plugins.Coordinator;




/** The DefaulResourceController represents one object of the datamodel. <p>
 
 This special controller supports a representation in a JList and properties
 for sorting purposes. Addition it implements most of the default function,
 which are required by the Controller Interface.<p>
 
 The programmer has to implement the function expandView, removeView,
 getAccessTime and stateChanged (when something in the datamodel has changed).
 
 @author Daniel Weiss
 @version $Revision: 1.3 $
 @see Controller
 */
public abstract class DefaultResourceController extends Observable
        implements
            Comparable,
            ResourceController,
            Controller,
            ChangeListener,
            PlanPresentationConstants
{


    /** The last change in the datamodel to this object. */
    protected long accessTime = 0;

    /** The uplink to the coordinator, which administrates this controller. */
    protected Coordinator coordinator;

    /** True if the view of this controller is expanded. */
    protected boolean isExpanded = false;


    /** The index of this controller in the JList.   */
    protected int listIndex = -1;

    /** The priorityOffset + the accessTime is the priority, on which the
     the sorting algorithm sorts the elements.  */
    protected long priorityOffset;

    /** The name of the resource. */
    protected String resourceName = null;

    /** Creates a DefaultResourceController.
     @param _c The reference to the coordinator.
     */
    public DefaultResourceController(Coordinator _c)
    {
        coordinator = _c;
    }

    /** Add a view to the plan paint area. */
    protected abstract void expandView();

    /** Removes the view from the plan panel. */
    protected abstract void removeView();

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.utilities.MultiLevelController#absoluteY(int)
     */
    public int absoluteY(int relativeY)
    {
        return coordinator.absoluteY(relativeY);
    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.plugins.Controller#addViewChangeListener(javax.swing.event.ChangeListener)
     */
    public void addViewChangeListener(ChangeListener l)
    {
        coordinator.addViewChangeListener(l);

    }

    /*
     *  (non-Javadoc)
     * @see java.lang.Comparable#compareTo(java.lang.Object)
     */
    public int compareTo(Object obj)
    {

        if ((obj instanceof ResourceController))
        {
            ResourceController rc2 = (ResourceController) obj;
            return (new Long(getAccessTime())).compareTo(new Long(rc2
                    .getAccessTime()));
        } else
        {
            throw new Error("Uncomparable objects " + obj.toString());
        }
    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.plugins.Controller#engineStop
     */
    public void engineStop()
    {
        coordinator.engineStop();

    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.plugins.Controller#getAccessTime()
     */
    public abstract long getAccessTime();

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.utilities.ResourceController#getName()
     */
    public String getName()
    {
        if (resourceName != null)
        {
            return resourceName;
        } else
        {
            return "";
        }
    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.plugins.Controller#getPanelWidth()
     */
    public int getPanelWidth()
    {

        return coordinator.getPanelWidth();
    }

    /** Returns a link to the planpresentationAPI.<p>

     @return A link to the planpresentationAPI.  */
    public PlanPresentationAPI getPlanPresentationAPI()
    {
        return coordinator.getPlanPresentationAPI();
    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.plugins.Controller#getPriority()
     */
    public Long getPriority()
    {

        return new Long(getAccessTime() + priorityOffset);
    }


    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.utilities.ResourceController#getSpeedMode()
     */
    public int getSpeedMode()
    {
        return getPlanPresentationAPI().getGlobalSearchControl().getRealTimeControl().getMode();
    }

    /** Returns a link to the swimlanecontrol.<p>
     
     @return The link to the SwimlaneControl.*/
    public SwimlaneControl getSwimlaneControl()
    {
        return coordinator.getSwimlaneControl();

    }

    /** This function is called, when someone clicks in the JList on this 
     controller. <p>     */
    public void onClick()
    {
        // check if the view is already expanded
        if (isExpanded)
        {
            // it is expanded so remove it
            removeView();
            isExpanded = false;
        } else
        {
            // it is not expanded, so expand it
            expandView();
            isExpanded = true;
        }
        fireChange(this);
    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.plugins.Controller#removeViewChangeListener(javax.swing.event.ChangeListener)
     */
    public void removeViewChangeListener(ChangeListener l)
    {
        coordinator.removeViewChangeListener(l);

    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.plugins.Controller#scroll(int)
     */
    public void scroll(int scrollstep,  boolean extendViewIfNecessary)
    {
        coordinator.getPlanPresentationAPI().scroll(scrollstep, extendViewIfNecessary);

    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.utilities.ResourceController#setName(java.lang.String)
     */
    public void setName(String name)
    {
        resourceName = name;
        fireChange(this);

    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.plugins.Controller#setPriorityOffset(long)
     */
    public void setPriorityOffset(long _priorityOffset)
    {
        priorityOffset = _priorityOffset;
    }

    /* (non-Javadoc)
     * @see javax.swing.event.ChangeListener#stateChanged(javax.swing.event.ChangeEvent)
     */
    public abstract void stateChanged(ChangeEvent e);

    /** Makes a string of this object to represent it.<p>
      
      This string represents this controller in the JList on the left side.
      If the view is expanded, it adds a minus to remove it again.<p>
      @return the name of the resource and a sign to recognize if it is 
      expanded or not.
     */
    public String toString()
    {
        String prefix = null;

        if (isExpanded)
        {
            prefix = "- ";
        } else
        {
            prefix = "+ ";
        }

        if (resourceName != null)
        {
            return prefix + resourceName;
        } else
        {
            return "Unknown";
        }

    }

    
    /** Delegate function. Requests a view update.<p>
     
       @see com.ai_center.planpresentation.PlanPresentationAPI#updateView
     */
    public void updateView()
    {
        coordinator.updateView();

    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.plugins.Controller#updateViewSize(int)
     */
    public void updateViewSize(int level)
    {
        coordinator.updateView();

    }


}