package com.ai_center.planpresentation.utilities;


import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import com.ai_center.dragonbreath.RealTimeControl;
import com.ai_center.dragonbreath.utilities.IntervalList;
import com.ai_center.dragonbreath.utilities.IntervalListElement;
import com.ai_center.planpresentation.PlanPresentationConstants;




/** This model builds a framework to copy IntervalLists.<p>
 
 The framework supports tracing of the last access time and other
 support-features. A implementor must overwrite the copyElement
 function.
 
 @author Daniel Weiss
 @version $Revision: 1.2 $
 */
public abstract class DefaultResourceModel
        implements
            ChangeListener,
            PlanPresentationConstants
{


    /** The last time, when something has changed in the model. */
    private long accessTime = 0;

    /** Uplink to the ResourceController. */
    private ResourceController controller;

    /** A map with the key as the copy and the value as the original.
     This is need to recognize elements. */
    protected Map copiedObjectToOriginalMap = new HashMap();

    /** The copy of the IntervaList.*/
    private IntervalList intervalListCopy = null;

    /** The time, when the copy was last updated. There a too often updates
     to handle all. */
    private long lastUpdate = 0;

    /**Maps the references of the original elements to the references of the
     gui elements.*/
    private Map originalIntervalElementToSemanticGraphicMap = new HashMap();

    /** Creates a default resource model.<p>
     
     @param _controller The link to the corresponding ResourceController. */
    public DefaultResourceModel(ResourceController _controller)
    {
        controller = _controller;
    }

    /** Copies a IntervalListElement.<p>
     
     Requires that the copyElement-function is overwritten.<p> 	
     @param ile The IntervalListElement to copy.
     @param ileCopyPrevious The previous IntervalListElement, needed to make
     the back and forward link in the IntervalListelements. 
     @return A copy of the IntervalListElement.
     */
    private IntervalListElement copyIntervalListElement(
            IntervalListElement ile, IntervalListElement ileCopyPrevious)
    {
        IntervalListElement copy = new IntervalListElement(ile.getBeginning(),
                ile.getEnd(), copyElement(ile.getElement()));

        return copy;

    }

    /** Copies a IntervalList.<p>
     
     */
    private synchronized void updateIntervalList()
    {
        // reset directory
        copiedObjectToOriginalMap = new HashMap();

        // get reference to intervallist
        IntervalList intervalList = (IntervalList) controller.getDatamodel();


        intervalListCopy = null;

        // walk through intervalList and make a copy of every Element
        if (intervalList.getFirst() != null)
        {

            IntervalListElement ile = intervalList.getFirst();
            IntervalListElement ileCopyPrevious = null;

            intervalListCopy = new IntervalList(ile.getBeginning(), ile
                    .getEnd(), copyElement(ile.getElement()));
            IntervalListElement ileCopy = intervalListCopy.getFirst();

            while (ile.getNext() != null)
            {
                ile = ile.getNext();
                ileCopyPrevious = ileCopy;
                ileCopy = copyIntervalListElement(ile, ileCopyPrevious);

                ileCopy.setPrevious(ileCopyPrevious);
                if (ileCopyPrevious != null)
                {
                    ileCopyPrevious.setNext(ileCopy);
                }

            }

        }
    }

    /** This function must be overwritten.<p>
     
     @param element The element to copy.
     @return The copied element.
     */
    protected abstract Object copyElement(Object element);

    /** Returns the last access time.<p>
     
     @return The last access tim*/
    public long getAccesTime()
    {
        return accessTime;
    }

    /** Returns a reference to the original element for a given copy..<p>
     
     This function is used, to recognize elements again. 
     @param copy The copied element.
     @return The reference to the original element, if existent.
     */
    public Object getOriginal(Object copy)
    {
        return copiedObjectToOriginalMap.get(copy);
    }

    /**Returns all states, tasks and preconditions in the interval between
     beginTick and endTick.<p>
     
     @param beginTick The beginTick on the left side.
     @param endTick The endTick on the right side of the view.
     @return List with IntervalListElements between the beginTick and the endTick.
     */
    public List getStatesAndTasksBetween(long beginTick, long endTick)
    {
        List intervalLinkedList = new LinkedList();

        // check if the copy exists
        if (intervalListCopy != null)
        {

            IntervalListElement currentILE = intervalListCopy.getFirst();
            IntervalListElement lastILE = null;

            boolean firstElement = true;

            while (currentILE != null)
            {
                synchronized (currentILE)
                {
                    //1.check if the element beginns in the interval  
                    //2. check if the element ends in the interval
                    //3. check if the element goes from the left side to
                    // the right side of the interval.
                    if ((currentILE.getBeginning() >= beginTick && currentILE
                            .getBeginning() <= endTick)
                            || (currentILE.getEnd() >= beginTick && currentILE
                                    .getEnd() <= endTick)
                            || (currentILE.getBeginning() < beginTick && currentILE
                                    .getEnd() > endTick))
                    {

                        if (firstElement)
                        {
                            firstElement = false;
                            // first element was found now.

                            //check if there is an element, with beginning outside the interval, 
                            //but ending inside.
                            if (lastILE != null)
                            {
                                intervalLinkedList.add(lastILE);
                                // add the predecessor element			
                            }
                        }

                        // this element begins in the interval
                        intervalLinkedList.add(currentILE);

                    }

                    lastILE = currentILE;
                    currentILE = currentILE.getNext();
                    // caution deadlock: two currentILE locked or mutualaccessexception
                } // end synchronized
            }
        }
        return intervalLinkedList;
    }


    /* (non-Javadoc)
     * @see javax.swing.event.ChangeListener#stateChanged(javax.swing.event.ChangeEvent)
     */
    public synchronized void stateChanged(ChangeEvent e)
    {
        accessTime = System.currentTimeMillis();
        // check if the lastUpdate was not in the last 100ms
        // or if we are in single-step mode (than we want all updates)
        if ((System.currentTimeMillis() - lastUpdate > 100)
                || controller.getSpeedMode() == RealTimeControl.STEPPING)
        {
            updateIntervalList();
            lastUpdate = System.currentTimeMillis();
        }

    }

}