package com.ai_center.planpresentation.utilities;


import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import com.ai_center.planpresentation.plugins.Controller;




/** This MultiLevelModel arranges SemanticGraphic-objects without any overlap.<p>
 
 If two object would overlap, it puts one object on a different level. 
 It works best in context with the MultiLevelView and the 
 MultiLevelController.<p>
 
 @author Daniel Weiss
 @version $Revision: 1.2 $
 @see com.ai_center.planpresentation.utilities.SemanticGraphic
 */
public class MultiLevelModel
{


    /** A counter for debug purposes. */
    private int addCount;

    /** A link to the controller . */
    private Controller controller;

    /** The levelcount is not equal to the level.
     The levelcount can be greater than the level, because the view should
     not downsize, when ever a level is deleted. The screen would flicker
     very much. Therefore the levelcount is only adjusted, when: <p> 
     <pre>
     * levelcount - level > maxLevelBuffer
     </pre>
     */
    private int levelcount = 1;

    /** This list has one element for every level. Every level consists of
     an TreeMap,where the key is beginCT of an action and the value is 
     the ActionSemanticGraphic. There is no overlap of actions one the 
     same level. (i.e. beginCT1 < endCT1 < beginCT2 < endCT2)    
     */
    private List levels = null;

    /** maxLevelBuffer is against flickering on the screen, when level 
     increasing/decreasing is to fast
     <pre> 
     * if (maxLevelBuffer > (levelcount-levels.getSize())) 
     *      levelcount = levels.getSize() + 1;
     </pre>
     */
    private final int maxLevelBuffer = 3;


    /** In the actionDirectory are all actions stored, to find their 
     SemanticGraphic instance. e.g. while removing an action.  */
    private Map objectDirectory = new HashMap();

    /** A counter for debug purposes. */
    private int removeCount;

    /** Maps the SemanticGraphic Objects to the beginCT time, where it is
     stored in the TreeMaps of the levels-list. In this directory are only
     SemanticGraphics which have a timespan and are therefore visible to 
     the gui.<p> 
     */
    private Map visibleObjectDirectory = null;

    /** Creates the MultiLevelModel
     @param _c The link to the controller.
     */
    public MultiLevelModel(Controller _c)
    {
        controller = _c;
        levels = new LinkedList();
        visibleObjectDirectory = new HashMap();

    }

    /** AdjustLevel checks if the adjustment of the levelcount is necessary.<p>
     
     The levelcount is not adjusted everytime, when the level number is 
     changed. This would result screen flickering. So the levelcount is only
     adjusted, when 
     <pre>level > levelcount</pre><p>
     or when <p<
     <pre> levelcount - level > maxLevelBuffer</pre><p>
     is.<p>
     
     @param newlevel The new level number.
     */
    private synchronized void adjustLevel(int newlevel)
    {
        if (newlevel >= levelcount)
        {
            // adjust level
            levelcount = newlevel + 2;
            // add additional buffer, so that next time, increasing is not neccessary
            controller.updateViewSize(levelcount);

        } else if ((levelcount - newlevel) > maxLevelBuffer)
        {
            // more than the maxLevelBuffer-Space is wasted. 
            // Adjust levelcount

            levelcount = newlevel + 2;
            controller.updateViewSize(levelcount);

        }

    }

    /** This function checks, if the there is enough place on 
     the time scala in this level.
     
     @param level The number of the level to check for free place on time 
     scala.
     @param beginCT The left border of the interval.
     @param endCT  The right border of the interval.
     @return returns true, if there is enough free place and false, if there isn't.
     */
    private synchronized boolean checkLevel(int level, long beginCT, long endCT)
    {

        TreeMap tempLevel = null;

        try
        {
            tempLevel = (TreeMap) levels.get(level);

        } catch (Exception e)
        {
            System.out
                    .println("MultiLevelModel.checkLevel(): level doesn't exist: "
                            + level);
            return false;
        }

        TreeMap subMap = null;
        Set beginCTAsgPairs = null;
        SemanticGraphic tempSg = null;

        // check endCT of predecessor
        subMap = new TreeMap(tempLevel.headMap(new Long(beginCT)));

        if (subMap.size() >= 1)
        {
            // there are actions before the beginCT in this level

            tempSg = (SemanticGraphic) subMap.get(subMap.lastKey());
            // get predecessor

            if ((tempSg.getBeginCT() + tempSg.getDuration()) > beginCT)
            {
                // endCT of predecessor < beginCT ?
                // there is an predecessor which collides with the beginCT of the action.
                // this level has no free place for the action
                return false;

            }
        }

        // check beginCT of successor
        subMap = new TreeMap(tempLevel.tailMap(new Long(beginCT)));
        if (subMap.size() >= 1)
        {
            // there are action after the endCT in this level
            tempSg = (SemanticGraphic) subMap.get(subMap.firstKey());
            // get successor

            if ((tempSg.getBeginCT()) < endCT)
            {
                // beginCT of successor < endCT
                // there is an successor which collides with the endCT of the action.
                // this level has no free place for the action
                return false;

            }

        }

        return true;

    }

    /** Walks through all levels and calls checkLevel() to find the lowest
     levelnumber,where the action can be placed. 
     @param beginCT Beginning time of the SemanticGraphic.
     @param endCT Ending time of the SemanticGraphic.
     @return level  The lowest levelnumber where the action can be placed or
     -1, if there is no place for this SemanticGraphic in any level.
     */
    private synchronized int findLevel(long beginCT, long endCT)
    {

        for (int level = 0; level < levels.size(); level++)
        {

            if (checkLevel(level, beginCT, endCT) == true)
            {
                // place for the action in this level found.
                return level;
            }

        }
        return -1; // no place in any level found.

    }

    /** Adds a object to the original-SemanticGraphic mapper.<p>
     
     This function does NOT add a object to the visible area. See
     addVisibleObject for this purpose.<p>
     
     @param obj The object in the datamodel.
     @param sg The corresponding SemanticGraphic
     */
    public synchronized void addObject(Object obj, SemanticGraphic sg)
    {
        objectDirectory.put(obj, sg);
    }


    /** Adds a new SemanticGraphic to the MultiLevelModel.<p>
     
     First it searches for a level, where enough place on the time scala is,
     to insert the SemanticGraphic. Then it puts the SemanticGraphic in this
     level.
     @param sg The SemanticGraphic to insert in the model.
     */
    public synchronized void addVisibleObject(SemanticGraphic sg)
    {
        addCount++;
        long beginCT = sg.getBeginCT();
        long durationCT = sg.getDuration();
        long endCT = beginCT + durationCT;

        // find a level, where the SemanticGraphic can be put.
        int level = findLevel(beginCT, endCT);


        if (level == -1)
        {
            // no level exists, where the action can be placed. Add a new level.	
            levels.add(new TreeMap());
            level = levels.size() - 1; // place action in the new level
            adjustLevel(level);
        }

        // put SemanticGraphic in level
        TreeMap tempLevel = (TreeMap) levels.get(level);
        tempLevel.put(new Long(beginCT), sg);

        // notice the beginCT of the SemanticGraphic to find it later again,
        // when beginCT had changed
        visibleObjectDirectory.put(sg, new Long(beginCT));
        sg.setSwimlaneVisibility(true);


    }

    /** Returns the actual level.<p>
     
     The actual level is the value of how many levels exist in parallel.<p>
     @return The actual level.
     */
    public synchronized int getLevel()
    {

        if (levels == null)
            return 0;
        else
            return levels.size();
    }

    /** The levelcount is not equal to the level.
     The levelcount can be greater than the level, because the view should
     not downsize, when ever a level is deleted. The screen would flicker
     very much. Therefore the levelcount is only adjusted, when: <p> 
     <pre>
     * levelcount - level > maxLevelBuffer
     </pre>
     @return The levelcount.
     */
    public int getLevelcount()
    {

        return levelcount;
    }

    /** Returns all object between beginCT and beginCT+durationCT.<p>
     
     @param beginCT The begin Tick.
     @param durationCT The duration.
     @return A Map of all SemanticGraphics in this interval.
     */
    public synchronized Map getObjects(long beginCT, long durationCT)
    {
        if ((beginCT < 0) || (durationCT < 0))
        {
            throw new Error(
                    "DefaultCoordinatorModel.getObjects() : illegal CurrentTime-Value");
        }

        TreeMap tempMap = new TreeMap();

        for (int level = 0; level < levels.size(); level++)
        {
            tempMap.putAll(getObjectsOnLevel(level, beginCT, durationCT));
        }

        return tempMap;
    }

    /** Returns all object between beginCT and beginCT+durationCT on the given
     level.<p>
     
     @param beginCT The begin Tick.
     @param durationCT The duration.
     @param The level to check.
     @return A Map of all SemanticGraphics in this interval.
     */
    public synchronized Map getObjectsOnLevel(int level, long beginCT,
            long durationCT)
    {

        if ((level < 0) || (level >= levels.size()))
        {
            return new TreeMap(); // the calling function has an old level-value
            // return a empty treeMap

        }

        /*
         *  make a clone of the treeMap instead of a reference, because the return value is
         * only for read access. Otherwise it wouldn't be thread safe, because the paint() 
         * command reads the TreeMap and the GSC-Thread could write the TreeMap the same time.
         * 
         */
        TreeMap tempMap = null;
        try
        {
            TreeMap refMap = (TreeMap) levels.get(level);
            if (refMap == null)
            {
                System.out.println("MultiLevelModel.refMap == null");
            }
            tempMap = (TreeMap) refMap.clone();
        } catch (Exception e)
        {
            System.out.println(e.getMessage());
        }


        Object firstKeyCT = null;
        Object lastKeyCT = null;

        // get predecessor element (=firstKeyCT)
        TreeMap headMap = new TreeMap(tempMap.headMap(new Long(beginCT - 1)));
        if (headMap.size() >= 1)
        {
            firstKeyCT = headMap.lastKey();
        }

        // get successor element (=lastKeyCT)
        TreeMap tailMap = new TreeMap(tempMap.tailMap(new Long(beginCT
                + durationCT)));
        if (tailMap.size() >= 1)
        {
            lastKeyCT = tailMap.firstKey();
        }

        if ((firstKeyCT != null) && (lastKeyCT != null))
        {

            // on the left and on the right are actions to clip. return submap 
            // with additional predecessor- and successor element
            return tempMap.subMap(firstKeyCT, lastKeyCT);

        } else if ((firstKeyCT == null) && (lastKeyCT != null))
        {

            // there are no predecessors, but some successor. return headMap
            return tempMap.headMap(lastKeyCT);

        } else if ((firstKeyCT != null) && (lastKeyCT == null))
        {

            // there are no successors, but some predecessors. return tailMap
            return tempMap.tailMap(firstKeyCT);

        } else
        { // firstKeyCT == null && lastKeyCT== null

            //there are no predecessors and successors return full list 
            return tempMap;

        }

    }

    /** Returns the corresponding SemanticGraphic element to the given
     original object.<p>
     
     @param obj The original object.
     @return The corresponding SemanticGraphic element. 
     */
    public synchronized SemanticGraphic getSemanticGraphic(Object obj)
    {
        return (SemanticGraphic) objectDirectory.get(obj);

    }

    /** Removes a object completely from the MulitLevelModel.<p>
     
     @param obj The object to remove.
     */
    public synchronized void removeObject(Object obj)
    {
        SemanticGraphic sg = (SemanticGraphic) objectDirectory.remove(obj);
        removeVisibleObject(sg);
        sg.terminate();

    }

    /** Remove a SemanticGraphic element only from the visible 
     area.<p>
     
     @param sg The SemanticGraphic to remove.
     */
    public synchronized void removeVisibleObject(SemanticGraphic sg)
    {
        removeCount++;
        Object beginCTObject = visibleObjectDirectory.get(sg);
        visibleObjectDirectory.remove(sg);

        if (beginCTObject != null)
        {
            // if the SemanticGraphic was visible, it is neccessary, 
            // to delete it from the level-datastructure 

            long beginCT = ((Long) beginCTObject).longValue();

            // search level
            for (int level = 0; level < levels.size(); level++)
            {
                TreeMap tempMap = (TreeMap) levels.get(level);

                if (tempMap.get(beginCTObject) != null)
                {
                    // search for the same beginCT
                    //probably the right beginCT found
                    if (tempMap.get(beginCTObject) == sg)
                    {
                        // right object found
                        tempMap.remove(beginCTObject); // delete from levelMap

                        if (tempMap.size() == 0)
                        {
                            // this level is now empty, remove it
                            levels.remove(level);
                            // remove level and shift all other to the left     
                            adjustLevel(levels.size());
                        }

                    }
                }
            } // end for
        }
        sg.setSwimlaneVisibility(false);
    }

    /** Updates a SemanticGraphic element, if the content has changed.<p>
     
     @param sg The SemanticGraphic element to update.*/
    public synchronized void updateVisibleObject(SemanticGraphic sg)
    {
        removeVisibleObject(sg);
        addVisibleObject(sg);

    }


}