package com.ai_center.planpresentation.utilities;


import java.awt.Dimension;
import java.awt.Graphics;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.swing.BoundedRangeModel;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import com.ai_center.planpresentation.PlanPresentationConstants;
import com.ai_center.planpresentation.SwimlaneConnector;
import com.ai_center.planpresentation.plugins.Controller;
import com.ai_center.planpresentation.plugins.RetractableView;
import com.ai_center.planpresentation.plugins.View;




/** MultiLevelView shows many SemanticGraphic objects in parallel. <p>
 
 It works best in context with a MultiLevelModel.
 
 @author Daniel Weiss
 @version $Revision: 1.8 $
 */
public class MultiLevelView extends RetractableView
        implements
            View,
            ChangeListener,
            PlanPresentationConstants,
            SwimlaneConnector
{


    /** The offset to add at every object, which is painted in this view.
     e.g. this offset is used, to make free space for the states at the 
     StateResourceView.*/
    private int baseOffset = 0;

    /** Indicates if a fullRepaint is necessary or not. */
    private boolean fullRepaint;


    /** A reference to the mainViewModel. */
    protected BoundedRangeModel mainViewModel;

    /** mainViewModel.getExtent() */
    private int mainViewModelExtent = 0;

    /** mainViewModel.getValue() */
    private int mainViewModelValue = 0;

    /** The link to the MultiLevelController. */
    public MultiLevelController multiLevelController;

    /** The old mainViewModel.getExtent() */
    private int oldMainViewModelExtent;

    /** The old mainViewModel.getValue() */
    private int oldMainViewModelValue;

    /** The cache for the SemanticGraphic objects. */
    private PanelCache panelCache = new PanelCache();

    /** A link to the SemanticGraphic object, which is dragged at the moment.*/
    private SemanticGraphic semanticGraphicInDragMode = null;

    /** Creates the multiLevelView.<p>
     
     @param _mlc The link to the MultiLevelController
     @param _viewName The name of the view.
     */
    public MultiLevelView(MultiLevelController _mlc, String _viewName)
    {
        super((Controller) _mlc, _viewName);
        multiLevelController = _mlc;


        
        int width = controller.getPanelWidth();
        this.setMinimumSize(new Dimension(width, ACTION_LEVEL_HEIGHT + 20
                + HEADER_HEIGHT));
        this.setPreferredSize(new Dimension(width, ACTION_LEVEL_HEIGHT + 20
                + HEADER_HEIGHT));
        this.setLayout(null);

    }

    /** Creates the multiLevelView.<p>
    
    @param _mlc The link to the MultiLevelController
    @param _viewName The name of the view.
    */
   public MultiLevelView(MultiLevelController _mlc, String _viewName, boolean moveButtons)
   {
       super((Controller) _mlc, _viewName,moveButtons);
       multiLevelController = _mlc;


       
       int width = controller.getPanelWidth();
       this.setMinimumSize(new Dimension(width, ACTION_LEVEL_HEIGHT + 20
               + HEADER_HEIGHT));
       this.setPreferredSize(new Dimension(width, ACTION_LEVEL_HEIGHT + 20
               + HEADER_HEIGHT));
       this.setLayout(null);

   }
    /** Removes the panels from the view.<p>
     
     @param panels The panels to remove.
     */
    private void removeRemainingPanels(Collection panels)
    {

        Iterator remainingIt = panels.iterator();
        while (remainingIt.hasNext())
        {
            Object nextObject = remainingIt.next();
            JPanel panel = (JPanel) nextObject;
            if (nextObject instanceof SwimlaneConnector)
            {
                SwimlaneConnector sc = (SwimlaneConnector) nextObject;
                sc.setShouldVisible(false); // make this object invisible
            }
            this.remove(panel);
        }

    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.plugins.View#absoluteY(int)
     */
    public int absoluteY(int i)
    {

        return multiLevelController.absoluteY(this.getY()) + i;
    }

    /** This function checks, if the actual screenarea has to be repainted. 
     @param beginCT The left border of the interval.
     @param endCT The right border of the interval.
     @return True, if the screen needs a repaint.
     */
    public boolean checkRevalidateMultiLevelView(int beginCT, int endCT)
    {
        if (((beginCT > mainViewModelValue) && (beginCT < mainViewModelValue
                + mainViewModelExtent))
                // beginCT on actual screen
                || ((endCT > mainViewModelValue) && (endCT < mainViewModelValue
                        + mainViewModelExtent))
                // endCT on actual screen)
                || ((beginCT < mainViewModelValue) && (endCT > mainViewModelValue
                        + mainViewModelExtent)))
        {
            // beginCT is on the left of the actual screen an endCT on the right

            return true; // repaint neccessary!
        } else
        {
            return false;
        }

    }

    /** Converts the x coordinate of a pixel to a absolute tick-value.<p>
     
     @param x The x-coordinate.
     @return The tick.
     */
    public long convertCoordinatesToTime(int x)
    {
        // mainViewModelValue + (x-pixel / pixelPerTimeUnit)
        int width = this.getWidth();
        double pixelPerTimeUnit = (double) this.getWidth()
                / (double) mainViewModelExtent;

        return (long) (mainViewModelValue + ((x - RETRACT_BUTTON_WIDTH) / pixelPerTimeUnit));

    }

    /** Converts a distance of pixel to a tick value.<p>
     
     @param pixel The pixel number to convert.
     @return The distance as tick.
     */
    public long convertPixelToCT(int pixel)
    {
        int width = this.getWidth();
        double pixelPerTimeUnit = (double) this.getWidth()
                / (double) mainViewModelExtent;
        return (long) (pixel / pixelPerTimeUnit);
    }

    /** Converts a tick to a x-coordinate.<p>
    
    @param tick The tick to convert.
    @return The x-coordinate.
    */
   public int convertTickToPixel(long tick)
   {
       int extent = mainViewModelExtent;
       int width = this.getWidth();

       double pixelPerTimeUnit = (double) width / (double) mainViewModelExtent;

       // avoid Java-Bug: Long.MAX_VALUE * 2.0 == - Long.MAX_VALUE;
       if (tick > 100000000)
       {
           tick = 100000000;
       }
       int x = RETRACT_BUTTON_WIDTH
               + (int) ((double) (tick - mainViewModelValue) * pixelPerTimeUnit);

       if (x < 0 && tick > mainViewModelValue && mainViewModelValue > 100)
       {
           System.err
                   .println("SymbolicStateResourceView.convertTickToPixel(): Illegal value. Resetting to zero");
           x = 0;
       }
       // protection against extreme coordinate values and the following jpanel memory bug
       if (x > 2 * width)
       {
           x = 2 * width;
       }
       if (x < (-2 * width))
       {
           x = -2 * width;
       }
       return x;

   }
    /* Converts a tick to a x-coordinate.<p>
     
     @param tick The tick to convert.
     @return The x-coordinate.
     /
    public int convertTickToPixel(long tick)
    {
        int extent = mainViewModelExtent;
        int width = this.getWidth();

        double pixelPerTimeUnit = (double) width / (double) mainViewModelExtent;
        int x = (int) ((double) (tick - mainViewModelValue) * pixelPerTimeUnit);

        // protection against extreme coordinate values and the following jpanel memory bug
        if (x > 2 * width)
        {
            x = 2 * width;
        }
        if (x < (-2 * width))
        {
            x = -2 * width;
        }
        return x + RETRACT_BUTTON_WIDTH;

    }*/

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.plugins.View#engineStop()
     */
    public void engineStop()
    {
        // do nothing, this function isn't called.

    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneConnector#getConnectorX(int)
     */
    public int getConnectorX(int connectionType)
    {
        return 0;
    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneConnector#getConnectorY(int)
     */
    public int getConnectorY(int connectionType)
    {
        return 0;
    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.plugins.View#getMainViewModelExtent()
     */
    public int getMainViewModelExtent()
    {
        return mainViewModelExtent;
    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.plugins.View#getMainViewModelValue()
     */
    public int getMainViewModelValue()
    {
        return mainViewModelValue;
    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneConnector#getPosition(int)
     */
    public int getPosition(int connectionType)
    {
        return absoluteY(0);
    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneConnector#isOutgoing(int)
     */
    public boolean isOutgoing(int connectionType)
    {

        return false;
    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneConnector#isVisible(int)
     */
    public boolean isVisible(int connectionType)
    {
        return this.isVisible();
    }

    /*
     *  (non-Javadoc)
     * @see javax.swing.JComponent#paintComponent(java.awt.Graphics)
     */
    public void paintComponent(Graphics g)
    {
        long localMaximumTime = 0;

        if (!retracted)
        {

            /*
             * Runtime cost are O(n^2), but there are not many  levels.
             */
            if (mainViewModel != null)
            {

                mainViewModelValue = mainViewModel.getValue();
                mainViewModelExtent = mainViewModel.getExtent();
            }
            int extent = mainViewModelExtent;
            int width = this.getWidth();

            double pixelPerTimeUnit = (double) width
                    / (double) mainViewModelExtent;

            // paint actions
            for (int level = 0; level < multiLevelController.getLevel(); level++)
            {
                //get all actions on this level
                Set objects = multiLevelController.getObjectsOnLevel(level,
                        mainViewModelValue, mainViewModelExtent).entrySet();
                Iterator it = objects.iterator();

                while (it.hasNext())
                {
                    // paint alle actions on this level
                    Map.Entry entry = (Map.Entry) it.next();

                    SemanticGraphic sg = (SemanticGraphic) entry.getValue();
                    sg.setShouldVisible(true);
                    if (!sg.isDragging() && sg.shouldVisible())
                    {
                        if ((sg.getBeginCT() + sg.getDuration()) > localMaximumTime)
                        {
                            localMaximumTime = sg.getBeginCT()
                                    + sg.getDuration();
                        }
                        double x1 = ((double) (sg.getBeginCT() - mainViewModelValue) * pixelPerTimeUnit);

                        if (x1 < 0)
                        {
                            x1 = -RETRACT_BUTTON_WIDTH;
                        } else if (x1 > width)
                        {
                            x1 = width - 1;
                        }
                        double y1 = (level * ACTION_LEVEL_HEIGHT) + 5
                                + HEADER_HEIGHT + baseOffset;
                        //+5 margin in top

                        double x2 = ((sg.getBeginCT() - mainViewModelValue + sg
                                .getDuration()) * pixelPerTimeUnit);

                        if (x2 < 1)
                        {
                            x2 = -RETRACT_BUTTON_WIDTH;
                        } else if (x2 > width)
                        {

                            x2 = width - 1;
                        }
                        double y2 = y1 + 5; // + ACTION_LEVEL_HEIGHT - 20 ;

                        // sg.setBackground(Color.green);
                        sg
                                .setSize((int) (x2 - x1 - 1),
                                        ACTION_LEVEL_HEIGHT - 5);
                        sg.setLocation((int) x1 + RETRACT_BUTTON_WIDTH,
                                (int) y1);

                        SemanticGraphic sgPanel = (SemanticGraphic) panelCache
                                .getObject(sg);
                        if (sgPanel == null)
                        {
                            // this is a new sematicGraphic
                            panelCache.addObject(sg, sg);
                            this.add((JPanel) sg);
                        }

                    } else
                    {
                        //action isDragging== true
                        semanticGraphicInDragMode = sg;
                    }
                }

                if (semanticGraphicInDragMode != null)
                {

                    if (semanticGraphicInDragMode.isDragging())
                    {

                        semanticGraphicInDragMode
                                .setSize((int) (semanticGraphicInDragMode
                                        .getDuration() * pixelPerTimeUnit),
                                        ACTION_LEVEL_HEIGHT - 5);
                        //  semanticGraphicInDragMode.setBackground(Color.green);

                        // mark this semanticGraphic as used. So that it will not deleted in this step.
                        panelCache.addObject(semanticGraphicInDragMode,
                                semanticGraphicInDragMode);

                        //    this.add((JPanel) semanticGraphicInDragMode, 0);

                    } else
                    {
                        semanticGraphicInDragMode = null;
                    }

                }

            }
            removeRemainingPanels(panelCache.getRemainingPanels());
        }

        if (localMaximumTime > mainViewModel.getMaximum())
        {

            SwingUtilities.invokeLater(new Runnable()
            {


                public void run()
                {
                    mainViewModel
                            .setMaximum((int) (mainViewModel.getMaximum() + 1000));

                }
            });
        }
        super.paintComponent(g); // paint background

    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.plugins.View#scroll(int)
     */
    public void scroll(int scrollstep,  boolean extendViewIfNecessary)
    {
        controller.scroll(scrollstep, extendViewIfNecessary);

    }

    /** Sets the baseOffset.<p>
     
     @param _offset The new baseOffset.
     */
    public void setBaseOffset(int _offset)
    {
        baseOffset = _offset;
    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneConnector#setShouldVisible(boolean)
     */
    public void setShouldVisible(boolean b)
    {
        // do nothing, this function isn't called.
    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneConnector#shouldVisible()
     */
    public boolean shouldVisible()
    {

        return false;
    }

    /*
     *  (non-Javadoc)
     * @see javax.swing.event.ChangeListener#stateChanged(javax.swing.event.ChangeEvent)
     */
    public void stateChanged(ChangeEvent e)
    {
        if (e.getSource() instanceof BoundedRangeModel)
        {
            BoundedRangeModel mainViewModelEvent = (BoundedRangeModel) e
                    .getSource();
            mainViewModel = mainViewModelEvent;
            if (!mainViewModelEvent.getValueIsAdjusting())
            {
                if ((mainViewModelEvent.getValue() != mainViewModelValue)
                        || (mainViewModelEvent.getExtent() != mainViewModelExtent))
                {
                    mainViewModelValue = mainViewModelEvent.getValue();
                    mainViewModelExtent = mainViewModelEvent.getExtent();
                    javax.swing.SwingUtilities.invokeLater(new Runnable()
                    {


                        public void run()
                        {
                            repaint();
                        }
                    });

                }
            }
        }
    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.plugins.View#terminate()
     */
    public void terminate()
    {
        // do nothing, this function isn't called

    }

    /*
     *  (non-Javadoc)
     * @see com.ai_center.planpresentation.plugins.RetractableView#updateViewSize()
     */
    public void updateViewSize()
    {
        updateViewSize(multiLevelController.getLevelcount());
    }

    /** Requests a view update.<p>
     
     Requests a update for an adjustment of a new height. The levelcount
     is used to check, if this update is really necessary.
     @param levelcount The actual number of levels.
     */
    public void updateViewSize(int level)
    {
        if (!retracted)
        {

            int width = controller.getPanelWidth();
            this.setMinimumSize(new Dimension(width, ACTION_LEVEL_HEIGHT
                    * level + 30 + HEADER_HEIGHT));
            this.setPreferredSize(new Dimension(width, ACTION_LEVEL_HEIGHT
                    * level + 30 + HEADER_HEIGHT));
        }
        multiLevelController.updateView();
    }


}