package com.ai_center.planpresentation.utilities;


import java.awt.Cursor;
import java.awt.Point;
import java.awt.event.MouseEvent;
import java.util.List;

import com.ai_center.planpresentation.PlanPresentationConstants;
import com.ai_center.planpresentation.SwimlaneConnector;
import com.ai_center.planpresentation.SwimlaneControl;
import com.ai_center.planpresentation.plugins.View;




/** TaskSemanticGraphic represents all TaskObjectConstraints.<p>
 
 
 @author Daniel Weiss
 @version $Revision: 1.4 $
 */
public abstract class TaskSemanticGraphic extends DragablePanel
        implements
            SwimlaneConnector,
            PlanPresentationConstants,
            SemanticGraphic
{


    /** The list of connected objects. */
    protected List connectorList;

    /** If the swimlane should be visible or not. */
    protected boolean swimlaneVisibility = true;

    /** The task, e.g. the preconditionTaskObjectConstraint. */
    protected Object task;

    /** A link to the view. */
    protected View view;


    /** Creates the TaskSemanticGraphic.
     
     @param _swimlaneControl A link to the SwimlaneControl.
     @param _task A link to the corresponding TaskObjectConstraint.
     @param _view A link to the view.
     */
    public TaskSemanticGraphic(SwimlaneControl _swimlaneControl, Object _task,
            View _view)
    {
        swimlaneControl = _swimlaneControl;
        task = _task;
        view = _view;
        swimlaneVisibility = true;

    }

    /** Converts the x coordinate of a pixel to a absolute tick-value.<p>
     
     @param x The x-coordinate.
     @return The tick.
     */
    protected long convertCoordinatesToTime(int x)
    {

        int width = view.getWidth();
        double pixelPerTimeUnit = (double) width
                / (double) view.getMainViewModelExtent();
        return (long) (view.getMainViewModelValue() + ((x - RETRACT_BUTTON_WIDTH) / pixelPerTimeUnit));

    }

    /** Returns the absolut y value of the coordinate.<p>
     
     absoluteY(relativeY) computes the absolute value of y values in the
     planning area.<p>
     
     @param relativeY The local y value in the planning area 
     @return the absolute y value in the JFrame.
     */
    public int absoluteY(int relativeY)
    {
        return view.absoluteY(this.getY()) + relativeY;
    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneConnector#getConnectorX(int)
     */
    public int getConnectorX(int connectionType)
    {
        return this.getX();

    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneConnector#getConnectorY(int)
     */
    public int getConnectorY(int connectionType)
    {
        return absoluteY(connectionType);

    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneConnector#getPosition(int)
     */
    public int getPosition(int connectionType)
    {
        return SwimlaneConnector.NORTH;

    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneConnector#isOutgoing(int)
     */
    public boolean isOutgoing(int connectionType)
    {
        return false;
    }

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneConnector#isVisible(int)
     */
    public boolean isVisible(int connectionType)
    {

        return true && !view.isRetracted();
    }

    
    /*
     *  (non-Javadoc)
     * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
     */
    public void mouseEntered(MouseEvent evt) {
      view.getPlanPresentationAPI().getGuiFrame().setCursor(
                Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
        super.mouseEntered(evt);
    }
    
    /*
     *  (non-Javadoc)
     * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
     */
    public void mouseExited(MouseEvent evt) {
        view.getPlanPresentationAPI().getGuiFrame().setCursor(
                Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
        super.mouseExited(evt);
        
    }
    
    
    /*
     *  (non-Javadoc)
     * @see java.awt.event.MouseMotionListener#mouseDragged(java.awt.event.MouseEvent)
     */
    public void mouseDragged(MouseEvent evt)
    {
   
        view.engineStop();
        //overwrite DragablePanel function
        super.mouseDragged(evt);
        int width = view.getWidth();
        Point location = this.getLocation();

        int scrollstep = 0;
        //check for normal speed
        if ((location.x + this.getWidth()) > (width - SCROLL_NORMAL_AREA))
        {
            scrollstep = SCROLL_NORMAL;
            // scroll to the right with normal speed
        } else if (location.x < SCROLL_NORMAL_AREA)
        {
            scrollstep = -SCROLL_NORMAL;
            // scroll to the left with normal speed
        }

        //check for fast speed
        if ((location.x + this.getWidth()) > (width - SCROLL_FAST_AREA))
        {
            scrollstep = SCROLL_FAST;
            // scroll to the right with normal speed
        } else if (location.x < SCROLL_FAST_AREA)
        {
            scrollstep = -SCROLL_FAST;
            // scroll to the lefft with normal speed
        }

        if (location.x < 0)
        {
            location.x = 0;

        }
        if (location.y < 0)
        {
            location.y = 0;

        }

        if (location.y + this.getHeight() > view.getHeight())
        {
            location.y = view.getHeight() - this.getHeight();
        }

        if (location.x + this.getWidth() > view.getWidth())
        {
            location.x = view.getWidth() - this.getWidth();
        }

        this.setLocation(location);

        if (scrollstep != 0)
        {
            view.scroll(scrollstep, true);
        }

    }

    /** Registers this TaskSemanticGraphic as Connector at the SwimlaneControl.
     */
    public void registerConnector()
    {
        if (task != null)
        {

            swimlaneControl.registerConnector(task, this);
        }
        swimlaneVisibility = false;
    }

    /** Requests a Swimlane at the SwimlaneControl.<p>
     
     @param connectionType The connectionType of the Swimlane.
     */
    public abstract void requestSwimlane(int connectionType);

    /* (non-Javadoc)
     * @see com.ai_center.planpresentation.SwimlaneConnector#terminate()
     */
    public void terminate()
    {
        swimlaneControl.removeSwimlane((SwimlaneConnector) this);
        if (task != null)
        {

            swimlaneControl.removeConnector(task);
        }
        swimlaneVisibility = false;
    }

}